#!/bin/bash
#
# Build script for 'BestCrypt for Linux'
#
# This script will build 'src' and 'rpm' packages and upload it to Binary Repository
#
# Our packages are signed by http://nsk.jetico.com/redmine/attachments/470/packages_jetico.com.gpgkey key.
# (page link: http://nsk.jetico.com/redmine/projects/best-crypt-for-linux/wiki )
set -e
set -u

BUILD_DEPENDS_DEBIAN_6_0="debian-archive-keyring pbuilder dpkg-dev openssh-client"



### Build System Bootstrap ###
# Get 'buildsys.func.inc' from Binary Repository
SOURCES_TOP_DIR="."
BUILDSYS_SCRIPT="buildsys.func.inc"
if [ ! -s "$SOURCES_TOP_DIR/$BUILDSYS_SCRIPT" ] ; then
  # Normally, location of Binary Repository should be specified via environment,
  # see http://nsk.jetico.com/redmine/projects/bc-wipe-total-wipe-o/wiki/Build_Host_Setup
  # Default location (in builder's home directory) is for test purpose.
  BIN_REPO_DIR="${BIN_REPO_DIR:-$HOME/binary-repository}"
  BIN_REPO_SSH="${BIN_REPO_SSH:-}"
  bootstrap_get_file()
  {
    local src_file="$BIN_REPO_DIR/$1"
    local dst_file="$SOURCES_TOP_DIR/$1"
    if [ -n "${BIN_REPO_SSH:-}" ] ; then
      scp -q -o "BatchMode yes" -p "$BIN_REPO_SSH:$src_file" $dst_file
    else
      cp -a "$src_file" $dst_file
    fi
  }
  echo "Bootstrap: getting '$BUILDSYS_SCRIPT' from $BIN_REPO_SSH $BIN_REPO_DIR"
  bootstrap_get_file "$BUILDSYS_SCRIPT"
fi
. $SOURCES_TOP_DIR/$BUILDSYS_SCRIPT
### end of bootstrap ###



### Configuration ###
ARCH=`uname -m`

BUILD_DIR="build_deb_tmp"



### Pre-build checks ###

# check binary repository availability
ping_repository
check_build_depends
set_source_control_variables



### Constants ###

# this should be an user account which has rights to put binari packages to APT/YUM
# repository. This user should have 'contrib/put_package_in_repo.sh' script in its home directory.
APT_REPO_SSH="two-uploader@umbrella"
APT_REPO_DIR="/home/alex/Public/packages"



### Get components depending on ###
TMP_SOURCES=`mktemp -d /tmp/src_bestcrypt_src_XXXXXX`
download_component_to_directory "src" "$TMP_SOURCES" "all"




### Build ###


# Put package installation instruction as index.html
INSTALL_HTML="index.html"
cat << EOF > "$INSTALL_HTML"
<html>
 <head>
  <title>BestCrypt installation instruction</title>
 </head>
 <body>
  <pre>
  `cat INSTALL.packages`
  </pre>
 </body>
</html>
EOF
scp "$INSTALL_HTML" "$APT_REPO_SSH:$APT_REPO_DIR/"

scp contrib/put_package_in_repo.sh "$APT_REPO_SSH:"



# Extract debian source package, fix version and build for specified distribution
#  Args:
#  $1 - distro name (precise, trusty, squeeze, wheezy)
build_and_upload_debian_package_with_pbuilder()
{
  local dist="$1"

  # clean from previous
  rm -rf $BUILD_DIR
  mkdir -p $BUILD_DIR

  echo
  echo
  echo
  echo "===> $dist: Extract debian source package..."
  echo
  echo
  DEB_SRC=`find $TMP_SOURCES/ -name 'bestcrypt_*'`
  $CP $DEB_SRC   $BUILD_DIR/
  (
    cd $BUILD_DIR
    DSC=`find -name '*.dsc'`
    test -n "$DSC" || { echo "Error: $0: unable to find debian source control file (.dsc)" ; exit 1 ; }
    dpkg-source -x $DSC
  )

  echo "===> $dist: Build debian binary package..."
  (
    cd $BUILD_DIR/
    DIR=`find -type d -maxdepth 1 -name 'bestcrypt-*'`
    test -n "$DIR" || { echo "Error: $0: unable to get exctracted debian package directory" ; exit 1 ; }
    cd $DIR
    # previous method: just build on host system: dpkg-buildpackage -B

    # this was tested on Debian 6.0
    # with following packages from 'sid' (jul 2014):
    #   * ubuntu-archive-keyring_2012.05.19-1_all.deb
    #   * debootstrap_1.0.60_all.deb

    # add $dist to package version to avoid .deb files clash in APT repository (pool/ folder)
    DEB_VER=`dpkg-parsechangelog -ldebian/changelog -o0 -c1 | awk '/^Version:/ { print $2 }'`
    sed -i "s/${DEB_VER}/${DEB_VER}.${dist}/" debian/changelog

    # Create pbuilder environment if it is not created for this dist yet
    BASETGZ="/var/cache/pbuilder/$dist.tgz"
    if [ ! -s "$BASETGZ" ] ; then

      if [ ! -s /usr/share/debootstrap/scripts/$dist ] ; then

cat <<EOF

Error: $0: seems you are using an old version of debootstrap

Hint: You should install the lates possible from 'sid' repository:

   https://packages.debian.org/debootstrap

EOF
        exit 1
      fi

      # hint: keep distribution codenames in sync with contrib/put_package_in_repo.sh
      case $dist in
        squeeze|wheezy|testing)
          MIRROR="http://mirror.yandex.ru/debian/"
          KEYRING="/usr/share/keyrings/debian-archive-keyring.gpg"
          ;;
        # 'raring' is out of stock (removed from main mirrors)
        raring)
          MIRROR="http://old-releases.ubuntu.com/ubuntu/"
          KEYRING="/usr/share/keyrings/ubuntu-archive-keyring.gpg"
          ;;
        lucid|saucy|precise|trusty|utopic)
          MIRROR="http://ru.archive.ubuntu.com/ubuntu/"
          KEYRING="/usr/share/keyrings/ubuntu-archive-keyring.gpg"
          ;;
        *)
          echo "Error: $0: please add MIRROR= definition to this script for dist '$dist'"
          exit 1
      esac

      if [ ! -s $KEYRING ] ; then
        cat <<EOF

Error: $0: there is no keyring file $KEYRING.

You should install appropriate debian-archive-keyring or
ubuntu-archive-keyring package.

On Debian you can get it directly from 'sid': https://packages.debian.org/ubuntu-archive-keyring

EOF
        exit 1
      fi

      sudo pbuilder create \
        --distribution $dist \
        --mirror $MIRROR \
        --debootstrapopts --keyring=$KEYRING \
        --basetgz "$BASETGZ"
    fi

    if ! sudo pdebuild -- --basetgz "$BASETGZ" --buildresult .. ; then
      # may be we need to do 'pbuilder update' to refresh the packages list.
      # we should do it anyway at least once per week.
      if find "$BASETGZ" -mtime +7 >/dev/null ; then
        echo
        echo "Info: $BASETGZ is older than 7 days. Try to update the system image and re-try build"
        echo
        sudo pbuilder update --basetgz "$BASETGZ"

        # and try to re-build
        sudo pdebuild -- --basetgz "$BASETGZ" --buildresult ..
      fi
    fi
  )
  DEB_PKG=`find $BUILD_DIR/ -name '*.deb'`

  # Get Debian binary packages
  DEB_PKG_FN=""
  for i in $DEB_PKG ; do
    DEB_PKG_FN="$DEB_PKG_FN `basename $i`"
  done



  ### Upload to Binary Repository ###
  echo "===> $dist: Upload to Binary Repository"
  TMP_COMP=`mktemp -d /tmp/bestcrypt_deb_XXXXXX`

  $CP $DEB_PKG $TMP_COMP/

  echo "$SOURCE_CONTROL_REV" > $TMP_COMP/revision.txt
  echo -e "Recent changes:\n\n${SOURCE_CONTROL_LOG}" > $TMP_COMP/recent_changes.txt

  # build host info
  write_buildinfo_file "$TMP_COMP/buildinfo.txt"

  FILELIST="$DEB_PKG_FN revision.txt recent_changes.txt buildinfo.txt"

  # create digest file
  ( cd $TMP_COMP ; md5sum $FILELIST > md5sums.txt )

  # TODO: do we need this binaries in our Binary Repository?
  # It does a little sense because it is better to install packages with APT.
  upload_component_from_directory "deb_$dist" $TMP_COMP



  # Upload binary packages into APT repository
  echo "===> $dist: Upload to APT repository"

  ssh $APT_REPO_SSH "mkdir -p incoming"
  ( cd $BUILD_DIR ; scp ${DEB_PKG_FN} ${APT_REPO_SSH}:incoming/ )
  ssh $APT_REPO_SSH "cd incoming ; ../put_package_in_repo.sh -d $dist -b $APT_REPO_DIR ${DEB_PKG_FN} ; rm ${DEB_PKG_FN}"
  if [ "$?" = "127" ] ; then
    cat <<EOF

Hint: Make sure you have script 'put_package_in_repo.sh' on $APT_REPO_SSH.
      This may fix the problem:
          $ cd svn/bestcrypt-linux
          $ scp contrib/put_package_in_repo.sh $APT_REPO_SSH:

EOF
    exit 1
  fi

  ### Cleanup ###
  rm -rf $TMP_COMP
  rm -rf $BUILD_DIR
}


# Build packages for supported version alive versions of Debian and Ubuntu
# exclude lucid (10.04):   Depends: debhelper (>= 8.0.0) but it is not installable
# exclude raring: removed from yandex mirror
for dist in squeeze wheezy testing saucy precise trusty utopic
do
  build_and_upload_debian_package_with_pbuilder $dist
done
