/////////////////////////////////////////////////////////////////
//
//      Copyright (c) 1994-2013 Jetico, Inc., Finland
//      All rights reserved.
//
//      File: cPEMCtx.h
//      Revision:
//      Created:
//			12.12.2013
//      Description: 
//			1. Parses PEM header and fills encryption context ( algorithm identiier and salt ).
//			2. Decrypts buffer according context by UTF8 password.
//				It generates key compatible with EVP_BytesToKey function from openssl.
//
//		PEM ENCRYPTION FORMAT:
//			From: http://www.openssl.org/docs/crypto/pem.html#PEM_ENCRYPTION_FORMAT
//	
//			This old PrivateKey routines use a non standard technique for encryption.
//
//			The private key (or other data) takes the following form:
//
//			 -----BEGIN RSA PRIVATE KEY-----
//			 Proc-Type: 4,ENCRYPTED
//			 DEK-Info: DES-EDE3-CBC,3F17F5316E2BAC89
//			 ...base64 encoded data...
//			 -----END RSA PRIVATE KEY-----
//			The line beginning DEK-Info contains two comma separated pieces of information: 
//			the encryption algorithm name as used by EVP_get_cipherbyname() and an 8 byte salt encoded as a set of hexadecimal digits.
//
//			After this is the base64 encoded encrypted data.
//
//			The encryption key is determined using EVP_bytestokey(), using salt and an iteration count of 1.
//			The IV used is the value of salt and *not* the IV returned by EVP_bytestokey().
//
/////////////////////////////////////////////////////////////////

#pragma once

#include "..\SMemory\cPassword.h"

class cPEMCtx
{
	cPTB m_ptbSalt;

	cPTB m_ptbKey;

	//
	//	bytesToKey
	//		Generates key to m_ptbKey from byte buffer ( really it is utf8 password )
	//		according parsed context.
	//
	int bytesToKey( const BYTE *bf, int bfSize );

	//
	//	decrypt
	//		Decrypts buffer by parsed m_algId, m_salt and m_ptbKey generated by bytesToKey
	//		Returns:
	//			ppe_ok if success.
	//			ppe_invalidPassword if no other eror but padding of encrypted block was not processed properly.
	//
	int decrypt( cPTB &decBlock, const unsigned char *bf, int bfSize );

public:

	BYTE m_algId;
	BYTE *m_salt;
	int m_saltSize;
	
	cPEMCtx()
	{
		m_algId = (BYTE)-1;
		m_salt = NULL;
		m_saltSize = 0;
	}

	//
	//	isPlainText
	//		returns true if parsed context is not encrypted.
	//
	bool isPlainText();

	//
	//	decodeBase64Block
	//
	//		decodes base64 armored text inside strings
	//		"-----BEGIN RSA PRIVATE KEY-----"
	//		"-----END RSA PRIVATE KEY-----"
	//
	//		Returns ppe_ok if success.
	//
	static int decodeBase64Block( cPTB &ptb, BYTE *ptr, int size, const char **pArmorHeaders = NULL );

	//
	//	parseHeaders
	//		1. parses headers wich get by PEM_decodeBase64Block.
	//			Sample:
	//				Proc-Type: 4,ENCRYPTED
	//				DEK-Info: DES-EDE3-CBC,3F17F5316E2BAC89
	//		2. fills 
	//			m_algId
	//			m_salt, m_saltSize
	//
	//		Note, if m_algId != pgpsaPlaintext (0) ( see isPlainText() ),
	//			so the decoded base 64 body has to be decrypted.
	//
	//		Returns ppe_ok if success.
	//
	int parseHeaders( const char *armorHeaders );

	//
	//	decrypt
	//		Decrypts buffer accoding parsed parameters by password to decBlock.
	//		Returns:
	//			ppe_ok if success.
	//			ppe_invalidPassword if no other eror but padding of encrypted block was not processed properly.
	//
	int decrypt( cPTB &decBlock, const cPassword &pwd, const unsigned char *bf, int bfSize );
};
