//
//	Copyright (c) 2000-2001 Jetico, Inc., Finland
//	All rights reserved.
//
//	File:			PGPPkts.h
//	Description:	PGP packets library..
//	Purpouses:		Parses and format PGP pakets 
//	Documents:		RFC 2440.
//
#ifndef __PGPPkts__
#define __PGPPkts__

#include "pktdefs.h"
#include "../SMemory/SBlocks.h"

typedef class cPKA * HPKA; // Handle to public/private key algorithm

//-----------------------------------------------------------------
//
//	Key ring functions...
//		pgp_EnumKeyIDs
//		pgp_FindKeyByID
//
//		pgp_RemoveKeyByID
//		pgp_AddPacketToKeyRing
//
//-----------------------------------------------------------------

//
//	Function:
//		pgp_FindKeyByID
//	Finds public or secret key with specified 8 BYTEs identificator keyID 
//	in a PGP packet or key ring.
//
int pgp_FindKeyByID(
	BYTE **pKeyBuff,	// OUT - pointer to position of the key in common packet.
	int  *pKeyBuffSize,	// OUT - pointer to size of the key packet
	BYTE *keyID,		// IN  - 8 BYTEs identificator of key to be imported from following key ring.
						// Use pgp_ImportEnumKeyIDs to get buffer with ID of available keys
	BYTE *keyRingBuff,  // IN  - Key ring packet.
	int keyRingBuffSize,// IN  - Key ring buffer size.

	bool isSecret = false, // if true finds secret key with specified ID otherwise finds public


	BYTE **pSignaturePkt = NULL,	// OUT - it is filled if the key ring contains a following signature
	int *pSignaturePktSize = NULL,  // OUT - the signature packet size.

	BYTE **pUserIDPkt = NULL,	 // OUT - it is filled if the key ring contains a following user information
	int *pUserIDPktSize = NULL,  // OUT - the user information packet size.

	BYTE **pImgPkt = NULL,		//	Image packet.
	int *pImgPktSize = NULL		// 
	);

class cpgp_pkt
{
public:

	BYTE *body;
	int size;

	cpgp_pkt *next;

	cpgp_pkt();

	cpgp_pkt( BYTE *p, int s, cpgp_pkt *n = NULL )
	{
		body = p;
		size = s;
		next = n;
	}

	~cpgp_pkt()
	{
		if ( next ) delete next;
	}

};

class cpgp_keyInfo
{
public:

	BYTE keyID[ pgp_keyIDSize ];			// Key ID buffer to be generated

	BYTE *keyPkt; // pointer to sPublicKeyV3 or sPublicKey
	int keyPktSize;

	cpgp_pkt *userID;
	cpgp_pkt *sig;
	cpgp_pkt *img;

	cpgp_keyInfo *next; // Subpacket.

	void init();

	void freeSubitems()
	{
		if ( next )		delete next;
		if ( userID )	delete userID;
		if ( sig )		delete sig;
		if ( img )		delete img;

		init();
	}

	cpgp_keyInfo()
	{
		init();
	}

	~cpgp_keyInfo()
	{
		freeSubitems();
	}
};

//
//	Function:
//		pgp_FindKeyByIDEx
//	It finds key with subkeys, even if a subkey ID was specified. 
//	User has to search subpackets if key was found but the master key ID 
//	is not equal to specified key. Or simple check pSubkeyInfo
//
int pgp_FindKeyByIDEx(
	cpgp_keyInfo *cKey,	// OUT - struct with class info to be filled. 
	BYTE *keyID,		// IN  - 8 BYTEs identificator of key to be imported from following key ring.
						// Use pgp_EnumKeyIDs to get buffer with ID of available keys
	BYTE *keyRingBuff,  // IN  - Key ring packet.
	int keyRingBuffSize,// IN  - Key ring buffer size.

	bool isSecret = false, // if true finds secret key with specified ID otherwise finds public
	cpgp_keyInfo **pSubkeyInfo = NULL // - Pointer to found subkey with specified key ID, the item is linked to common list to kiyInfo
	);

//
//	Function:
//		pgp_GetKeyUserID
//	Copies user ID from subKeyInfo if it is not NULL, otherwise
//	Writes 'Subkey of ' + user ID of master key
//
bool pgp_GetKeyUserID( 
	cPlainTextBlock *ptbUserID,
	cpgp_keyInfo *masterKeyInfo,
	cpgp_keyInfo *subKeyInfo
	);

//
//	Function:
//		pgp_FindKeyByUserID
//	It finds key with subkeys. 
//	User has to search subpackets if key was found but the master key ID 
//	is not equal to specified key.
//
int pgp_FindKeyByUserID(
	cpgp_keyInfo *cKey,	// OUT - struct with class info to be filled. 
	char *userID,		// IN  - null terminated string with user ID
	BYTE *keyRingBuff,  // IN  - Key ring packet.
	int keyRingBuffSize,// IN  - Key ring buffer size.

	bool isSecret = false // if true finds secret key with specified ID otherwise finds public
	);

//
//	Function:
//		pgp_EnumKeyIDs
//	Enumerates secret or public key ID's of all keys in specified packet or key ring.
//
int pgp_EnumKeyIDs(
	BYTE *keyIDBuff,		// OUT - buffer for key IDs, 
							//  it is filled as sequence of available keyIDs with keyIDSize.
							// If returns ppe_ok the *pKeyIDBuffSize / keyIDSize is number of 
							//	really filled key IDs
	int  *keyIDBuffSize,	// IN  - pointer to allocated size of key buffer
							// OUT - pointer to required size if ppe_bufferTooSmall returned

	BYTE *keyRingBuff,		// IN - Key ring packet to be parsed for available keys.
	int  keyRingBuffSize,
	bool secret = false
	);

//
//	Function:
//		pgp_RemoveKeyByID
//	Removes key by specified ID ( with all subpackets ) from key ring 
//		1. Finds main key with the ID. ( see pgp_FindKeyByIDEx descryption )
//		2. Walks all subkeys 
//		3. Copys next keys to the position.
//		4. Returns new decreased size.
//
int pgp_RemoveKeyByID(
	BYTE *keyRingBuff,		// Key ring packet.
	int  *pKeyRingBuffSize,	// [IN] Key ring buffer size.
							// [OUT] Key decreased size
	BYTE *keyID,
	bool isSecret
	);

//
//	Function:
//		pgp_AddPacketToKeyRing
//	Adds a packet to the end of key ring
//
int pgp_AddPacketToKeyRing(
	BYTE *keyRingBuff,		// Key ring packet.
	int  *pKeyRingBuffSize,	// [IN] Key ring buffer size.
							// [OUT] Key ring increased size
	BYTE *pkt,
	int pktSize
	);

//
//	Function:
//		pgp_GetEntireKeyByID
//	Gets entire key packet ( with all subpackets ) by specified ID from key ring 
//		1. Finds main key with the ID or subpacket with the ID
//		2. Gets entire packet position and size
//
int pgp_GetEntireKeyByID(
	BYTE **pEntireKeyPos,
	int *pEntireKeySize,
	BYTE *keyRingBuff,		// Key ring packet.
	int  keyRingBuffSize,	// Key ring buffer size.
	BYTE *keyID,
	bool isSecret
	);

//--------------------------------------------------------------------------------------
//
//	Session key packet format and parse function
//		pgp_FormatPublicKeyEncryptedSessionKeyPacket
//		pgp_ParsePublicKeyEncryptedSessionKeyPacket
//		pgp_DecryptPublicKeyEncryptedSessionKeyFromPacket
//
//---------------------------------------------------------------------------------------

//
//	Function : pgp_FormatPublicKeyEncryptedSessionKeyPacket
//	Prepares key block with signature and encrypts it.
//
int pgp_FormatPublicKeyEncryptedSessionKeyPacket(
	BYTE *pkt,		// OUT - Encrypted packet buffer.
	int  *pPktSize, // IN  - pointer to size of allocated buffer.
					// OUT - if no error it is pointer to size of really busied part
					//	 if error is ppe_bufferTooSmall it is pointer to required size .

	int  algID,		// = pgpsaBlowfish,
						// Identificator of symmetric algorithm to be used for encryption of a data with
						// following session key
	BYTE *sessionKey,	// Session key for the symmetric algorithm.
	int sessionKeySize, // Size of the session key

	int  hashAlgID,	// = pgphaSHA1,
						// Identificator of hash algorithm.
						// It is used for internal signature

	BYTE *pkPkt,		// Public key packet
	int pkPktSize	

	);

//
//	Function
//		pgp_ParsePublicKeyEncryptedSessionKeyPacket
//	Check integrity of public key encrypted session key packet.
//
int pgp_ParsePublicKeyEncryptedSessionKeyPacket(
	BYTE **pKeyID,			// [OUT] pointer to public key ID
	BYTE *pAlgID,			// [OUT] pointer to public key algorithm ID
	BYTE **pEncMpi,			// [OUT] pointer to encrypted Mpi buffer position
	int *pEncMpiSize,		// [OUT] pointer to size of encrypted Mpi buffer
	BYTE *pkt,				// [IN]The public key encrypted session key packet
	int	 pktSize
	);

//
//	Function pgp_DecryptPublicKeyEncryptedSessionKeyFromPacket
//	Parses the packet, decrypts with specified public key ID the session key block with signature and
//	verifies the signature. If signature is invalid the function returns ppe_invalidSignature.
//

int pgp_DecryptPublicKeyEncryptedSessionKeyFromPacket(
	BYTE *pAlgID,			// OUT - pointer to identificator of symmetic key algorithm to be used 
							//	 for decryption of a data or packets with following session key
	BYTE *sessionKey,		// OUT - plain text buffer with session key
	int  *pSessionKeySize,	// IN  - pointer to allocated size of key buffer
							// OUT - pointer to required size if ppe_bufferTooSmall returned

	BYTE *pkt,			// The public key encrypted session key packet
	int	 pktSize,

	BYTE *skBuff,		// Secret key buffer
	int  skBuffSize,

	const char *pwd		// Null terminated password string.
	);

//
//	Function: pgp_DecryptSecretKey
//	Parses session key packet, gets symmetric key algorithm ID,
//	 prepares key from password with parsed S2K method,
//	 encrypts session key to skBuffer.
//
int pgp_DecryptSecretKey(
	BYTE *skBuff,	// OUT - plaintext secret key buffer
	int *pSkBuff,	// IN  - pointer to size of allocated secret key buffer.
					// OUT - if no error it is pointer to size of really busied part
					//	 if error is ppe_bufferTooSmall it is pointer to required size .
	char *pwd,		// IN - password characters -  zero BYTE ended string.
	BYTE *skPkt,	// IN - encrypted secret key packet
	int skPktSize
	);

//
//	Function: pgp_VerifySignaturePacket
//
//		1. Parses signature packet.
//
//		2. Extracts issuer key ID.
//			2.1 if the keyPkt ID is same with extracted issuer ID, so uses the 
//				keyPkt for verification
//			2.2 otherwise finds the issuer public key into keyRingPkt 
//				by the extracted issuer keyID.
//
//		4. Prepares signature block.
//		5. Verifies it with issuer public key.
//
int pgp_VerifySignaturePacket(
	BYTE *sigPkt,
	int sigPktSize,
	cHashedList *hashList,
	BYTE *keyPkt = NULL,
	int keyPktSize = 0,
	BYTE *issuerKeyRing = NULL,
	int issuerKeyRingSize = 0,
	BYTE *issuerKeyIDBuff = NULL
	);

//
//	Function: pgp_MakeKeySignature_0x10_HashList
//		Adds the required packets to hash list accoding with signature type 0x10 
//			   0x10: Generic certification of a User ID and Public Key packet.
//
//	See RFC2440 5.2.4. Computing Signatures
//
int pgp_MakeKeySignature_0x10_HashList(
		cHashedList *hashList,	// List of pointers to packets and it's sizes to be hashed
		BYTE *keyPkt,			// body of key packet
		int keyPktSize,
		BYTE *userID,			// body of user ID packet
		int userIDSize,
		BYTE sigVersion			// signature version
		);

int pgp_GetSigType( BYTE *sigPtr, int signatureSize );

//
//	Function: pgp_AddKeyPktToHashList
//		Adds the key packet body with old style header
//
int pgp_AddKeyPktToHashList(
		cHashedList *hashList,	// List of pointers to pakcets and it's sizes to be hashed
		BYTE *keyPkt,			// body of key packet
		int keyPktSize
	);

int pgp_VerifyKey(
		BYTE *pk,	  int pkSize,		// public key
		BYTE *userID, int userIDSize,	// user ID
		BYTE *sig,	  int sigSize,		// signature packet
		BYTE *kr=NULL,int krSize = 0,	// user key ID == issuer key ID, so key ring does not use and can be NULL
		BYTE *issuerKeyIDBuff = NULL
	);

//
//	Functions for symmetric key encrypted message
//		pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket
//		pgp_FormatSymmetricKeyEncrytpedSessionKeyPacket
//
//	See RFC 2440
//		5.3. Symmetric-Key Encrypted Session-Key Packets (Tag 3)
//
//	Function.
//		pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket
//	Parses the packet.
//	Prepares symmetric key by s2k value and pwd.
//	If session key exists it will be decrypted to sessKeyBuff. 
//	Otherwise the s2k prepared key will be copied to the session key buffer.
//
int pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket(
		BYTE *pAlgID,			// pointer to symmetric algorithm identifier.
		BYTE *sessKeyBuff,
		int  *pSessKeyBuffSize, // Size of the buffer is small the function returns ppe_bufferTooSmall
								// and fills the value by required size.

		const BYTE *pkt,		// Symmetric-Key Encrypted Session-Key Packet
		int pktSize,

		const char *pwd			// Null terminated string.
	);

int pgp_ExtractSessionKeyFromSymmetricKeyEncrytpedPacket_Rudiment(
		BYTE *pAlgID,			// pointer to symmetric algorithm identifier.
		BYTE *sessKeyBuff,
		int  *pSessKeyBuffSize, // Size of the buffer is small the function returns ppe_bufferTooSmall
								// and fills the value by required size.

		const BYTE *pkt,		// Symmetric-Key Encrypted Session-Key Packet
		int pktSize,

		const char *pwd			// Null terminated string.
	);

int pgp_FormatSymmetricKeyEncrytpedSessionKeyPacket(
		BYTE *pkt,
		int  *pPktSize,		// Size of the buffer is small the function returns ppe_bufferTooSmall
							// and fills the value by required size.

		BYTE algID,				// Session algorithm ID

		const BYTE *sessKey,	// Session-Key buffer
		int sessKeySize,

		int keyAlgID,			// The algorithm for encryption of the key.

		const char *pwd,		// Null terminated string.

		BYTE hashAlgID = 2,		// s2k hash algorithm ( pgphaSHA1 == 2 )

		const BYTE *salt = NULL,// If NULL, it generates random 8 bytes salt

		int iterationCount = -1

	);


//
//	Function:
//		pgp_PacketHeader_Parse ( see also pgp_PacketHeader_Format )
//	Reads packet header and extracts:
//		1. packet tag
//		2. size of the packet header ( pointer to data if pkt + hdrSize
//		3. size of the data
//
//	Note, the function can read packets with partial data blocks
//
int pgp_PacketHeader_Parse(
			BYTE *pTag,		// [Out] Pointer to content tag
			int *pHdrSize,	// [Out] Pointer to the packet header size
			int *pDataSize,	// [Out] Pointer to the packet body size
			bool *pPartial,	// Pointer to partial value
							// [In ] False if read new packet,
							//		  ( *pTag is detected and changed )
							//		 True if read next part of partial packet, 
							//        ( *pTag is not detected and changed )
							// [Out] if True the data can be extracted again 
							//			form pkt + *pDataSize + *pHeaderSize;
							//		 if false the extracted data is last.
			BYTE *pkt,		// The packet with header.
			int pktSize		// Total size of available packet buffer.
			);

int pgp_PacketHeader_Format(
			BYTE *pkt,		// [Out]The packet with header.
			int  *pktSize,	// [In] Size of allcated buffer for the packet.
							// [Out] Space written or required if the buffer is too small.
							//	If last it returns error ppe_bufferTooSmall.
			BYTE tag,		// Content tag, 0 if it is partial tail. Will be packed without "Packet Tag" byte.
			BYTE *data,		// Packet body ( data )
			int	 size,		// The packet body size
			bool partialSize// Pack as partial length ( following packet has to be packed as partial type ( tag 0 ) ) 
					= false	// the length is a power of 2, from 1 to 1,073,741,824 (2 to the 30th power)
							// The first partial length MUST be at least 512 octets long ( rfc 2440 )
		);

//
//	Function:
//		pgp_GenerateKeyFromSeedBuffer
//	Generates key by string to key specification simple mode.
//		hashAlgID - hash algorithm for s2k specifier
//
int pgp_GenerateKeyFromSeedBuffer( BYTE *key, int keySize, const BYTE *seed, int seedSize, BYTE hashAlgID = pgphaSHA1 );
int pgp_SetRandomBuffer( BYTE *buff, int buffSize );

//
//	pgp_GetPKData
//	Extracts first packet data from key ring packet
//	if pTag != NULL and *pTag != 0 returns ppe_invalidBlockType if packed tag not equal to assigned *pTag
//	f.e. if *pTag == pkt_PublicKey it automatically check if the packed is Public Key
//
int pgp_GetPktData( BYTE *pTag, BYTE **pData, int *pDataSize, BYTE *pktBuff, int pktBuffSize, int *pParsed = NULL );

//
//	Function:
//		pgp_WritePublicKeyToPacketV4
//	Creates public key v4 packet from algorithm handle
//
int pgp_WritePublicKeyToPacketV4( BYTE *buff, int *pBuffSize, HPKA pka );
int pgp_WritePublicKeyToPacket( BYTE *buff, int *pBuffSize, HPKA pka, int version );

//
//	Function:
//		pgp_CompressedPacket_Parse
//	Retrieves compressed algorithm ID and pointers to compressed data.
//
//	RFC2440 pkt_CompressedData (Tag 8)
//	The body of this packet consists of:
//     - One octet that gives the algorithm used to compress the packet.
//     - The remainder of the packet is compressed data.
//
int pgp_CompressedPacket_Parse(
		BYTE *pAlgID,
		BYTE **ppData,
		int	 *pDataSize,
		const BYTE *pkt,
		int pktSize
	);

//
//	Function:
//		pgp_SymEncryptedAndIntegrityProtectedDataPacket_Parse
//
//	See: openpgp-rfc2440bis-22
//		Sym. Encrypted and Integrity Protected Data Packet (Tag 18)
//		pkt_SymEncryptedAndIntegrityProtectedData
//
int pgp_SymEncryptedAndIntegrityProtectedDataPacket_Parse(
		const BYTE **pPkt,	// [Out] position of plaintext data ( really 10 bytes shifted from packet beginnig.
		int *pPktSize,		// [Out] real size of plaintext data without preffered 10 bytes and appended digest
		int saBlockSize,	// [In]  size in bytes of the symmetric algorithm is used for the packet decrypting.
		const BYTE *bf,		// [In]  just encrypted packet
		int bfSize			// [In]  just encrypted packet size 
	);


//
//	Function:
//		pgp_SymEncryptedAndIntegrityProtectedDataPacket_Format
//
//	See: openpgp-rfc2440bis-22
//		Sym. Encrypted and Integrity Protected Data Packet (Tag 18)
//		pkt_SymEncryptedAndIntegrityProtectedData
//	1. Generates random saBlockSize bytes at packet beginning
//		( use the cRNG::setPool if the function returns ppe_randomBufferRequired )
//	2. Duplicates two last random bytes
//	3. Copies input data to output packet
//	4. Adds SHA1 hash of the output packet to the packet end.
//
int pgp_SymEncryptedAndIntegrityProtectedDataPacket_Format(
		BYTE *pPkt,			// = NULL [Out] 
		int *pPktSize,		// [In ] if is less the required set the required value and return bufferTooSmall.
							// [Out] if success set the value by filled size.
		int saBlockSize,	// [In ] size in bytes of the symmetric algorithm is used for the packet encrypting.
		const BYTE *bf,		// [In ] input data to be packed
		int bfSize			// [In ] input data size
	);

//
//	Function:
//		pgp_LiteralDataPacket_Format
//
//	See: openpgp-rfc2440bis-22
//
int pgp_LiteralDataPacket_Format(
		BYTE *pkt,		// [Out] 
		int *pPktSize,		// [Out] return required or filled packet size
		int codePage,		// [In ] CP_UTF8 and CP_ACP is used, all other f.e. -1 create bynary tag
		const char *specialName,// [In ] special name to written at packet beginning
		UINT date,		// [In ] date stamp to be written after special name
		const BYTE *bf,		// [In ] input data
		int bfSize		// [In ] input data size
	);

//
//	Function:
//		pgp_LiteralDataPacket_Parse
//
//	See: openpgp-rfc2440bis-22
//
int pgp_LiteralDataPacket_Parse(
		const BYTE **pPkt,	// [Out] packet body pointer inside input packet
		int *pPktSize,		// [Out] packet body size
		int *pCodePage,		// [Out] modify if first byte is 0x74 to CP_ACP, 0x75 to CP_UTF8
		char **pSpecialName,// [Out] pointer to special name
		UINT *pDate,		// [Out] pointer to date stamp
		const BYTE *bf,		// [In ] formatted packet
		int bfSize			// [In ] formatted packet size
	);

#endif // __PGPPkts__
