#ifndef __PKTSignature__
#define __PKTSignature__

#include "Packets.h"

//
//	Signature type enums...
//
//	5.2.1. Signature Types
//
//   There are a number of possible meanings for a signature, which are
//   specified in a signature type octet in any given signature. These
//   meanings are:
//
enum{
		sig_SignatureOfABinaryDocument = 0x00,
		//	Signature of a binary document.
		//	Typically, this means the signer owns it, created it, or
		//	certifies that it has not been modified.

		sig_SignatureOfACanonicalTextDocument = 0x01,
		//	Typically, this means the signer owns it, created it, or
		//	certifies that it has not been modified.  The signature is
		//	calculated over the text data with its line endings converted
		//	to <CR><LF> and trailing blanks removed.

		sig_StandaloneSignature = 	0x02,
		//	This signature is a signature of only its own subpacket
		//	contents. It is calculated identically to a signature over a
		//	zero-length binary document. Note that it doesn't make sense to
		//	have a V3 standalone signature.

		sig_GenericCertificationOfAUserIDAndPublicKeyPacket = 0x10,
		//	The issuer of this certification does not make any particular
		//	assertion as to how well the certifier has checked that the
		//	owner of the key is in fact the person described by the user
		//	ID.  Note that all PGP "key signatures" are this type of
		//	certification.

		sig_PersonaCertificationOfAUserIDAndPublicKeyPacket = 0x11,
		//	The issuer of this certification has not done any verification
		//	of the claim that the owner of this key is the user ID
		//	specified.

		sig_CasualCertificationOfAUserIDAndPublicKeyPacket = 0x12,
		//	The issuer of this certification has done some casual
		//	verification of the claim of identity.

		sig_PositiveCertificationOfAUserIDAndPublicKeyPacket = 0x13,
		//	The issuer of this certification has done substantial
		//	verification of the claim of identity.
		//		Please note that the vagueness of these certification claims is
		//	not a flaw, but a feature of the system. Because PGP places
		//	final authority for validity upon the receiver of a
		//	certification, it may be that one authority's casual
		//	certification might be more rigorous than some other
		//	authority's positive certification. These classifications allow
		//	a certification authority to issue fine-grained claims.

		sig_SubkeyBindingSignature = 0x18,
		//	This signature is a statement by the top-level signing key
		//	indicates that it owns the subkey. This signature is calculated
		//	directly on the subkey itself, not on any User ID or other
		//	packets.

		sig_SignatureDirectlyOnAKey = 0x1F,
		//	This signature is calculated directly on a key.  It binds the
		//	information in the signature subpackets to the key, and is
		//	appropriate to be used for subpackets that provide information
		//	about the key, such as the revocation key subpacket. It is also
		//	appropriate for statements that non-self certifiers want to
		//	make about the key itself, rather than the binding between a
		//	key and a name.

		sig_KeyRevocationSignature = 0x20,
		//	The signature is calculated directly on the key being revoked.
		//	A revoked key is not to be used.  Only revocation signatures by
		//	the key being revoked, or by an authorized revocation key,
		//	should be considered valid revocation signatures.

		sig_SubkeyRevocationSignature = 0x28,
		//	The signature is calculated directly on the subkey being
		//	revoked.  A revoked subkey is not to be used.  Only revocation
		//	signatures by the top-level signature key that is bound to this
		//	subkey, or by an authorized revocation key, should be
		//	considered valid revocation signatures.

		sig_CertificationRevocationSignature = 0x30,
		//	This signature revokes an earlier user ID certification
		//	signature (signature class 0x10 through 0x13). It should be
		//	issued by the same key that issued the revoked signature or an
		//	authorized revocation key The signature should have a later
		//	creation date than the signature it revokes.

		sig_TimestampSignature = 0x40
		//	This signature is only meaningful for the timestamp contained
		//	in it.
};

//
//	Values of signature subpacket type octet
//
enum{
	ssp_signatureCreationTime				=  2, // signature creation time
	ssp_signatureExpirationTime				=  3, // signature expiration time
	ssp_exportableCertification				=  4, // exportable certification
	ssp_trustSignature						=  5, // trust signature
	ssp_regularExpression					=  6, // regular expression
	ssp_revocable							=  7, // revocable
	ssp_keyExpirationTime					=  9, // key expiration time
	ssp_placeholderForBackwardCompatibility = 10, // placeholder for backward compatibility
	ssp_preferredSymmetricAlgorithms		= 11, // preferred symmetric algorithms
	ssp_revocationKey						= 12, // revocation key
	ssp_issuerKeyID							= 16, // issuer key ID
	ssp_notationData						= 20, // notation data
	ssp_preferredHashAlgorithms				= 21, // preferred hash algorithms
	ssp_preferredCompressionAlgorithms		= 22, // preferred compression algorithms
	ssp_keyServerPreferences				= 23, // key server preferences
	ssp_preferredKeyServer					= 24, // preferred key server
	ssp_primaryUserId						= 25, // primary user id
	ssp_policyURL							= 26, // policy URL
	ssp_keyFlags							= 27, // key flags
	ssp_signerUserId						= 28, // signer's user id
	ssp_reasonForRevocation					= 29 // reason for revocation
	// 100 to 110 = internal or user-defined
};

class cSigSubpacketParser
{
	BYTE *m_pkt;
	int m_cur;
	int m_size;

 public:

	int m_error;
	BYTE *m_spBody;
	BOOL m_spTypeCritical;
	BYTE m_spType;
	int m_spSize;


	cSigSubpacketParser( BYTE *pkt, int pktSize );
	BOOL GetSubPacket();
	static BOOL IsValidSubpacketType( BYTE *pkt, int size, BOOL *pCritical );
	static const char *GetTypeName( BYTE spType );
};

#pragma pack(1)

struct sSignatureV3
{
	//
	//	Start of data buffer to be hashed...
	//
	BYTE version;
	BYTE hashedLength; // Always 5
		BYTE type;
		sTime time;
	BYTE keyID[8];
	BYTE publicKeyAlgID;
	BYTE hashAlgID;
	BYTE hashAlders[2]; // left 16 bits of signed hash value.
	BYTE mpi[2];
//
//	Algorithm Specific Fields for RSA signatures:
//	 - multiprecision integer (MPI) of RSA signature value m**d.
//	Algorithm Specific Fields for DSA signatures:
//	 - MPI of DSA value r.
//	 - MPI of DSA value s.
//
};

struct sSignature
{
	//
	//	Start of data buffer to be hashed...
	//
	BYTE version;
	BYTE type;
	BYTE publicKeyAlgID;
	BYTE hashAlgID;
	// BYTE hashedCnt[2];
	//	[
	//		hashed data ( zero or more subpackets )
	//	]
	//
	//	End of data buffer to be hashed 
	//	( Note, all data from version byte down to the place is hashed )...	
	//
	// BYTE unhashedCnt[2];
	//	[
	//		unhashed data  ( zero or more subpackets )
	//	]
	// BYTE hashAlders[2]; // left 16 bits of signed hash value.
	// BYTE mpi[2];		   // Public key algorithm specific MPI buffer with signature.
};

#pragma pack()


//
//	Function:
//		PKT_Signature_Parse
//
//	5.2. Signature Packet (Tag 2)
//

//
//	Signature V4 data
//
struct sSignatureData
{
	BYTE *hashedSubPkt;
	WORD hashedSize;
	BYTE *unhashedSubPkt;
	WORD unhashedSize;
	BYTE *hashLeft16Bits;
	BYTE *mpi;
};

int PKT_Signature_Parse(
		sSignature *sig,
		int pktSize,
		BYTE **pHashedSubPkt,
		WORD *pHashedSize,
		BYTE **pUnhashedSubPkt,
		WORD *pUnhashedSize,
		BYTE **pHashLeft16Bits,
		BYTE **pMpi
		);

int PKT_Signature_Parse(
		sSignature *sig,
		int pktSize,
		sSignatureData *sigData
		);

int PKT_SignatureV3_Parse(
		sSignatureV3 *sig,
		int pktSize
	);

//
//	Functions:
//		PKT_VerifySignaturePacket
//		PKT_VerifySignaturePacketV3
//
//		1. Parse signature packet.
//
//		2. Extract issuer key ID.
//			2.1 if the key is self signed packet, so use the 
//				key for verification
//			2.2 otherwise find the issuer public key into keyRingPkt 
//				by the extracted issuer keyID.
//
//		4. Prepare signature block.
//		5. Verify it with issuer public key.
//
//
int PKT_VerifySignaturePacket(
	sSignature *sigPkt,
	int sigPktSize,
	cHashedList *hashList,
	BYTE *keyPkt,
	int keyPktSize,
	BYTE *issuerKeyRing = NULL,
	int issuerKeyRingSize = 0,
	BYTE *issuerKeyIDBuff = NULL
	);

int PKT_VerifySignaturePacketV3(
	sSignatureV3 *sigPkt,
	int sigPktSize,
	cHashedList *hashList,
	BYTE *keyPkt,
	int keyPktSize,
	BYTE *issuerKeyRing = NULL,
	int issuerKeyRingSize = 0
	);

int PKT_VerifySignature(
	BYTE hashAlgID,
	BYTE pkAlgID,
	cHashedList *hashList,
	BYTE *hashAlders,
	BYTE *mpi,
	int mpiSize,
	BYTE *issuerKeyPkt,
	int issuerKeySize
	);

int PKT_VerifySignatureEx(
	BYTE *sigPkt,
	int sigPktSize,
	cHashedList *hashList,
	BYTE *issuerKeyPkt,
	int issuerKeySize
	);

const char *PKT_GetHashAlgName( BYTE algID );
const char *PKT_Signature_GetTypeName( BYTE type );
BYTE *PKT_findIssuerKeyID( BYTE *sp, int spSize );
sTime *PKT_findSignatureTime( BYTE *sp, int spSize );

#endif // __PKTSignature__
