#ifndef __Packets__
#define __Packets__

#include "../publicKeyAlgs/publicKeyAlgs.h"
#include "pktdefs.h"
#include "string2key.h"

//
//	PGP structs...
//
#pragma pack (1)

//
//	3.5. Time fields
//
//   A time field is an unsigned four-octet number containing the number
//   of seconds elapsed since midnight, 1 January 1970 UTC.
//
struct sTime
{
	BYTE t0;
	BYTE t1;
	BYTE t2;
	BYTE t3;

	UINT getTime()
	{
		return (UINT(t0) << 24) | (UINT(t1) << 16) | (UINT(t2) << 8) | t3;
	}
	void setTime( UINT time )
	{
		t0 = BYTE ( time >> 24 );
		t1 = BYTE ( time >> 16 );
		t2 = BYTE ( time >> 8  );
		t3 = BYTE ( time );
	}
};

struct sDayCnt
{
	BYTE t0;
	BYTE t1;

	WORD gatDays()
	{
		return ( WORD(t0) << 8 ) | t1;
	}

	void setDays( WORD days )
	{
		t0 = BYTE ( days >> 8  );
		t1 = BYTE ( days );
	}
};

struct sPublicKeyEncryptedSessionKey
{
	BYTE version;	// 2 or 3
	BYTE keyID[8];	// An eight-octet number that gives the key ID of the public key
					// that the session key is encrypted to.
	BYTE algID;		// A number giving the public key algorithm used.
	BYTE mpi[2]; // at least 2 bytes size of first mpi
			// if RSA		one MPI = key^e mod p
			// if Elgamal	two MPIs: 1) g^k mod p, 2) m * (y^k) mod p
};

//
// Versions 2 and 3 have same structures
//
struct sPublicKeyV3
{
	BYTE	version;
	sTime	time;
	sDayCnt	days;
	BYTE	algID;
	BYTE	mpi[2];
//
// - A series of multi-precision integers comprising the key material:
//	- a multiprecision integer (MPI) of RSA public modulus n;
//	- an MPI of RSA public encryption exponent e.
//
};

//
//	Version 4 ( main ) public key structure
//
struct sPublicKey
{
	BYTE	version;
	sTime	time;
	BYTE	algID;
	BYTE	mpi[2];
//
//	Algorithm Specific Fields for RSA public keys:
//		- MPI of RSA public modulus n;
//		- MPI of RSA public encryption exponent e.
//
//	Algorithm Specific Fields for DSA public keys:
//		- MPI of DSA prime p;
//		- MPI of DSA group order q (q is a prime divisor of p-1);
//		- MPI of DSA group generator g;
//		- MPI of DSA public key value y (= g**x where x is secret).
//
//	Algorithm Specific Fields for Elgamal public keys:
//		- MPI of Elgamal prime p;
//		- MPI of Elgamal group generator g;
//		- MPI of Elgamal public key value y (= g**x where x is
//		secret).
//
	BYTE getAlgID()
	{
		if ( version > 3 )
			return algID;
		else
		{
			sPublicKeyV3 *pktV3 = ( sPublicKeyV3 * )this;
			return pktV3->algID;
		}
	}
};

struct sSecretKey : public sPublicKey
{
//
//	- One octet indicating string-to-key usage conventions.  0
//	indicates that the secret key data is not encrypted.  255
//	indicates that a string-to-key specifier is being given.  Any
//	other value is a symmetric-key encryption algorithm specifier.
//
//	- [Optional] If string-to-key usage octet was 255, a one-octet
//	symmetric encryption algorithm.
//
//	- [Optional] If string-to-key usage octet was 255, a string-to-key
//	specifier.  The length of the string-to-key specifier is implied
//	by its type, as described above.
//
//	- [Optional] If secret data is encrypted, eight-octet Initial
//	Vector (IV).
//
//	- Encrypted multi-precision integers comprising the secret key
//	data. These algorithm-specific fields are as described below.
//
//	- Two-octet checksum of the plaintext of the algorithm-specific
//	portion (sum of all octets, mod 65536).
//
//	Algorithm Specific Fields for RSA secret keys:
//
//	- multiprecision integer (MPI) of RSA secret exponent d.
//
//	- MPI of RSA secret prime value p.
//
//	- MPI of RSA secret prime value q (p < q).
//
//	- MPI of u, the multiplicative inverse of p, mod q.
//
//	Algorithm Specific Fields for DSA secret keys:
//
//	- MPI of DSA secret exponent x.
//
//	Algorithm Specific Fields for Elgamal secret keys:
//
//	- MPI of Elgamal secret exponent x.
//
};

#pragma pack ()



//
//	Function:
//		PKT_GetTagName.
//	Returns ppe_ok if valid tag.
//
int PKT_GetTagName( BYTE tag, const char **pStr );

//
//	Functions:
//		PKT_PublicKeyEncryptedSessionKey_Parse
//		PKT_PublicKeyEncryptedSessionKey_Format
//	See RFC 2440
//			5.1. Public-Key Encrypted Session Key Packets (Tag 1)
//
//	Checks:
//	 version, - The currently defined value for packet version is 3.
//	 alg validity,
//	 mpi buffer size.
//
int PKT_PublicKeyEncryptedSessionKey_Parse(
		 sPublicKeyEncryptedSessionKey *pkt,
		 int pktSize,
		 int *pKeyMpiSize
		 );

int PKT_PublicKeyEncryptedSessionKey_Format(
		sPublicKeyEncryptedSessionKey *pkt,
						// [OUT] Packet buffer
		int  *pPktSize, // [IN]  Pointer to size of allocated buffer
						// [OUT] space really busied or required for the packet
		BYTE *keyID,	// [IN]  Key indentificator
		BYTE algID,		// [IN]  Public key algorithm
		BYTE *keyMpi,	// [IN]  Encrypted key
						// if RSA		one MPI = key^e mod p
						// if Elgamal	two MPIs: 1) g^k mod p, 2) m * (y^k) mod p
		int  keyMpiSize	// [IN]  Size of encrypted key buffer.
		);

//
//	RSA message block with session key format
//
//    EB = 00 || BT || PS || 00 || D .  Where D = Session key alg ID || Check summ || Session key     
//

int RSA_SessionKeyBlock_Parse( 
		BYTE **pSessKey, // Position of session key in inBuf
		int *pSesKeySize,
		BYTE *pSessKeyAlg,
		BYTE *inBuf,
		int inBufSize
		);
int RSA_SessionKeyBlock_Format( 
		BYTE *outBuf,
		int outBufSize,
		BYTE algID,
		const BYTE *sessKey,
		int sesKeySize
		);
//
//	pgp_SessionKeyBlock_Format
//		See notes before the function body realisation.
//
int pgp_SessionKeyBlock_Format( 
		BYTE *outBuf,
		int outBufSize,
		BYTE algID,
		const BYTE *sessKey,
		int sessKeySize
	);


//
//	Function: 
//		RSA_PublicKeyEncryptionBlock_Format PKCS1 v2, the ambiguies block type 00 is excluded
//		PKT_PublicKeyEncryptionBlock_Format - rudimentary function, we should use the RSA_ instead
//
//	Formats public key encryption block.
//
//	PKCS-1.
//	8.1 Encryption-block formatting
//		Block type values are 00, 01, or 02.
//		For a private-key operation, the block type shall be 00 or 01.
//		For a public-key operation, it shall be 02.
//
//	Note, pka_getBlockSize or ska_getBlockSize functions are required to detect 
//	the output block size required for public/private encryption
//
int RSA_PublicKeyEncryptionBlock_Format(
		bool publicKeyEncryption,// if false the method is used for signatures by private -> block type 01
								//  if true the method is used for encryption by public   -> block type 02
		BYTE *outBuf,
		int outBufSize,
		const BYTE *data,
		int dataSize
		);
/*
int PKT_PublicKeyEncryptionBlock_Format(
		BYTE *outBuf,
		int outBufSize,
		const BYTE *data,
		int dataSize,
		BYTE blockType
		);
*/
//
//	Function:
//		RSA_PublicKeyEncryptionBlock_Parse works accoding with PKCS1 v2, the ambiguies block type 00 is excluded
//
//	Parses and extracts data from public key encrytion block.
//
int RSA_PublicKeyEncryptionBlock_Parse(
		bool publicKeyEncryption,// if false the method is used for signatures by private -> block type 01
								//  if true the method is used for encryption by public   -> block type 02
		const BYTE *inBuf,
		int inBufSize,
		const BYTE **pData = NULL,	 // position of data in input buffer
		int *pDataSize = NULL
		);

//		PKT_PublicKeyEncryptionBlock_Parse  - rudimentary function, we should use the RSA_ instead
int PKT_PublicKeyEncryptionBlock_Parse(
		BYTE *data,		// may be NULL
		int *pDataSize, // may be NULL for block checking
		BYTE *inBuf,
		int inBufSize
		);

// Used for compatibility with early then 1.01 versions only.
int PKT_PublicKeyEncryptionBlock_Parse_Rudiment(
		BYTE *data,
		int *pDataSize,
		BYTE *inBuf,
		int inBufSize );

//
//	Function:
//		PKT_PublicKey_Parse
//	See RFC 2440
//		5.5.1.1. Public Key Packet (Tag 6)
//
//	Checks:
//	 version,
//	 alg validity,
//	 mpi buffer size.
//
int PKT_PublicKey_Parse( sPublicKey *pkt, int pktSize, int *pParsed = NULL, BYTE *pAlgID = NULL );
int PKT_PublicKeyV3_Parse( sPublicKeyV3 *pkt, int pktSize, int *pParsed = NULL );

//
//	Function:
//		PKT_CreatePublicKeyAlgFromPacket
//	1. Parses and ckecks public key packet.
//	2. Allocates public key algorithm in pHSA value, 
//		the hadler can be used for encryption ( pka_encrypt ) and verifying ( pka_verify ),
//		it has to be freed by ( pka_free )
//
int PKT_CreatePublicKeyAlgFromPacket( HPKA *pHPKA, sPublicKey *pkPkt, int pkPktSize );


//
//	Function:
//		PKT_SecretKey_Parse
//	See RFC 2440
//		5.5.3. Secret Key Packet Formats
//
//	1. Checks public part:
//		version,
//		alg validity,
//		mpi buffer size.
//
//	2. Fills secret key specific fields.
//		s2kMode,
//		symmetric algorithm ID
//		Initial Vector.
//		secret mpi buffer,
//	
int PKT_SecretKey_Parse(
	sSecretKey *pkt,	// pointer to secret key packet
	int pktSize,		// total pkt size

	int *pPublicMpiSize,

	BYTE *pSymAlgID,	// pointer to identifier of symmetric the algorithm used for decryption of private part
	S2K  *pS2K,			// pointer to mode for string to key conversion for symmetrin algorithm
	BYTE **pIV,			// pointer to initial vector

	BYTE **pEncMpi,		// pointer to encrypted buffer with secret mpis
	int *pEncPmiSize	// pointer to size of buffer to be decrypted
//	WORD *pCheckSumm	// pointer to two byte of check summ
	);

//
//	Function:
//		PKT_CreateSecretKeyAlgFromPacket
//	1. Parses and ckecks secret key packet.
//	2. Extracts public data
//	3. Decrypt secret data with specified password.
//	4. Allocates secret key algorithm in pHSA value, 
//		the hadler can be used for decryption ( ska_decrypt ) and signing ( ska_sing ),
//		it has to be freed by ( ska_free )
//
int PKT_CreateSecretKeyAlgFromPacket( HPKA *pHSA, sSecretKey *skPkt, int skPktSize, const char *pwd );

//
//	Functions:
//		PKT_GetPublicKeyFingerprint
//		PKT_GetPublicKeyID
//	See RFC 2440
//		11.2. Key IDs and Fingerprints
//  Note, packet with 'tag' byte required here.
//
int PKT_GetPublicKeyFingerprint( BYTE *pk, int totalPktSize, BYTE *bf20Bytes );
int PKT_GetPublicKeyID( BYTE *pk, int totalPktSize, BYTE *bf8Bytes );

int PKT_GetPublicKeyIDInternal( sPublicKey *pk, int totalPktSize, BYTE *bf8Bytes );

int PKT_GetPublicKeyID( HPKA pka, BYTE *bf8Bytes );

//
//	Function:
//		PKT_CheckMPIBuffer
//	Returns count of really bytes parsed. If size to small returns 0.
//
int PKT_CheckMPIBuffer( BYTE *mpis, int cnt, int size );
int PKT_GetEncryptedMessageMPICount( BYTE algID );
int PKT_GetPublicKeyMPICount( BYTE algID );
int PKT_GetSignatureMPICount( BYTE algID );

//
//	It uses PGP defined IDs
//	See enum pgppka in pktdefs.h, see also PKT_GetHashAlgName and PKT_GetSymmetricAlgName
//
const char *PKT_GetPublicKeyAlgName( BYTE algID );
const char *PKT_GetSymmetricAlgName( BYTE algID );

//
//	5.3. Symmetric-Key Encrypted Session-Key Packets (Tag 3)
//

struct sSymmetricKeyEncryptedSessionKeyPkt
{
	BYTE version;	// Always 4.
	BYTE algID;		// A number describing the symmetric algorithm used.
	BYTE s2kSpec[1];// A string-to-key (S2K) specifier
	//
	//	BYTE *algAndKey;
	//
	//	  If the packet length == sizeof of the structure with s2k length
	//	 Key is produced from s2k salted
	//	 Otherwise it decrypted by the algorithm with the s2k produced key.
	//
	//	    The decryption result consists of a one-octet algorithm identifier
	//   that specifies the symmetric-key encryption algorithm used to encrypt
	//   the following Symmetrically Encrypted Data Packet, followed by the
	//   session key octets themselves.
	//
};

#endif // __Packets__
