/////////////////////////////////////////////////////////////////
//
//	Copyright (c) 1994-2011 Jetico Inc., Finland
//	All rights reserved.
//
//	File: SBlocks.h
//	Revision:
//		28.06.2011 - Certificate lib v. 1.03.4
//		cUniversalString::Set function is changed. See notes below.
//
//	Created:
//	Description: Safe memory blocks...
//
/////////////////////////////////////////////////////////////////

#ifndef __SBlocks__
#define __SBlocks__

#ifdef _WIN32
#include <windows.h>
#else
#include "../publicKeyAlgs/Power/Windefs.h"
#endif

#include <memory.h>
#include <string.h>
#include <wchar.h>

class cUniversalString
{
  public:
	cUniversalString()
	{
		memset( this, 0, sizeof( *this ) );
	}
	~cUniversalString();

  	//
	// Set functions
	//	if size == -1 it means that the string is null terminated
	//	All functions return number of allocated and filled bytes,
	//	the allocated buffer will include null terminated symbol,
	//	so if success the functions return non zero valus.
	//	Return zero if error.
	//
	int Set( const char *str, int size = -1 );
	int Set( const wchar_t *wcs, int wcsLen = -1 );
	int SetWCS( const wchar_t *wcs, int wcsLen = -1 ); // Set wcBuffer only.
	int SetUnicode( const wchar_t *wStr, int sizeInWideChars = -1 );
	int SetUnicode( const BYTE *utf16be, int sizeInBytes );
	//
	//	SetUnicode_CheckBigEndianInput
	//		Checks input buffer before copying to allocated buffer
	//		Assumes input as double byte buffer in little or big endian format.
	//		The allocated unicode buffer used for password generation, so it has to be filled according 
	//		PKCS 12 instructions( see comments at cUniversalString::wstr2unicodeBuff. )
	//		in big-endian format (most-significant byte first)
	//
	int SetUnicode_CheckBigEndianInput( BYTE *unicodeInput, int sizeInBytes );
	int maxI( int a, int b ) { return ( a > b ? a : b ); };
	int Set( cUniversalString *us )
	{
		//
		//	Certificate lib v. 1.03.4 - 28.06.2011
		//
		//	1. The Set( BYTE* ) function clears wcBuffer and unicodeBuffer it has to be called first but not last.
		//	2. SetUnicode( us->wcBuffer ) does not set wcBuffer but with the value unicodeBuffer
		//		The SetWCS( us->wcBuffer ) has to be used instead...

		int size  = ( ( us->buffer		  != NULL ) ? Set       ( us->buffer								) : 0 );
		int sizeW = ( ( us->wcBuffer	  != NULL ) ? SetWCS	( us->wcBuffer								) : 0 );
		int sizeU = ( ( us->unicodeBuffer != NULL ) ? SetUnicode( us->unicodeBuffer, us->unicodeBufferSize	) : 0 );

		return maxI( size, maxI( sizeU, sizeW ) );
	}

	char *buffer;
	int bufferSize;
	wchar_t *wcBuffer;
	int wcBufferSize;
	BYTE *unicodeBuffer; // utf16 BigEndian byte buffer. See comments at cUniversalString::SetUnicode
	int unicodeBufferSize;

	BYTE *AllocBuffer( int size );
	int AllocateUnicodeBuffer( int sizeInSymbols ); // returns size of current unicode buffer.
	void FreeBuffer( BYTE *buff );
	void ClearBuffers();
	bool IsEmpty() { return ( bufferSize == 0 ) || ( buffer[0] == '\0' ); }
	
	//
	// The conversions assume little endian unicode, as used in asn1tag_UnicodeString ( 0x1E ).
	//	It tryes to generate unicode symbol from wcBuffer.
	//	If wcBuffer is empty it uses buffer but here is umbiguity the buffer may be encoded as utf8 or not. 
	//	Moreover the procedure simple extened every symbol from byte to wchar_t.
	//	So Note, always fill wcBuffer before the function call.
	//
	int GenerateUnicodeString(); // returns unicodeBufferSize in bytes, it is equal (pwd length + 1) * 2
	int MakeFromUnicodeString();

	//
	//	Wide char string / unicode buffer conversions.
	//
	//	wstr2unicodeBuff 
	//		unicode buffer size has to be equal wcslen( ws ) * 2 bytes
	//		returns number of copied wide char symbols
	//
	static int wstr2unicodeBuff( BYTE *unicodeBuff, int unicodeBuffSize, const wchar_t *wStr, int len = -1 ); //, bool checkBigEndian = false );
	int wstr2unicodeBuff( const wchar_t *ws, int len = -1 ) //, bool checkBigEndian = false )
	{
		if ( len == -1 )
			len = (int)wcslen( ws );

		if ( AllocateUnicodeBuffer( len + 1 ) == 0 )
			return 0;
		
		return wstr2unicodeBuff( unicodeBuffer, unicodeBufferSize, ws, len ); //, checkBigEndian );
	}

	//
	//	unicodeBuff2wstr
	//		size of wc buffer has to be one symbol more of unicodeBuffByteSize/2
	//		for null termination symbol
	//
	static int unicodeBuff2wstr( wchar_t *ws, int wsSizeInByte, BYTE *unicodeBuff, int unicodeBuffByteSize );
	int unicodeBuff2wstr( wchar_t *ws, int wsSizeInByte )
	{
		int size = wsSizeInByte;
		if ( size > unicodeBufferSize )
			size = unicodeBufferSize;

		return unicodeBuff2wstr( ws, wsSizeInByte, unicodeBuffer, size );
	}
};

//
//	cPlainTextBlock class is used for allocation, storing and deletion of
//	keys and other sensitive information
//
class cPlainTextBlock
{
	//
	//	The call can be send as void handler ( see static WriteBytes ).
	//	But it has a signature value. So validity of the block can be checked.
	//

public:

	enum { sigVal = 0x00627470 }; // "ptb"
	UINT sig;
	bool checkSig()	{ return sig == sigVal; }

//	HANDLE m_heap;


	cPlainTextBlock()
	{
		memset( this, 0, sizeof( *this ) );
		sig = sigVal;
//		m_heap = GetProcessHeap();
	}
	~cPlainTextBlock();

	//
	//	The block supports sequential filling by means of putByte and writeBytes.
	//	It is useful to allocate required size first.
	//
	int allocatedSize;

	BYTE *buffer;
	int bufferSize;

	void ClearBuffer();
	void SeekToBegin();

	//
	//	WriteBytes
	//		if ok returns number of bytes written, otherwise returns 0.
	//		if data == NULL it adds zero bytes to buffer tail.
	//
	int WriteBytes( const BYTE *data, int dataSize, bool init = false );
	BYTE *AddBytes( const BYTE *data, int dataSize );
	int WriteInt( int data ) { return WriteBytes( (BYTE *)&data, sizeof( data ), false ); }
	static int WriteBytesStatic( const void *ptr, BYTE *data, int dataSize );
	int PutByte( BYTE bt );
	int WriteString( const char *str, bool addZero = false ) { return WriteBytes( (const BYTE*)str, (int)strlen( str ) + ( addZero ? 1 : 0 )  ); }
	int WriteString( const wchar_t *str, bool addZero = true ) { return WriteBytes( (const BYTE*)str, 2 * ( (int)wcslen( str ) + ( addZero ? 1 : 0 ) ) ); }
	void Free();
	
	//
	//	Note, ReAllocate does not change bufferSize - current filled data does not changed...
	//
	BYTE *ReAllocate( int size );
	BYTE *Allocate( int size );
	BYTE *Assign( const BYTE *inBuffer, int size );
};

typedef class cPlainTextBlock cPTB;

#endif // __SBlocks__
