/*******************************************************************
 *
 *	Copyright (c) 1994-2000 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	File: md5.cpp
 *	Revision: 
 *	Created:
 *	Description: implementation of MD5 Message Digest Algorithm
 *
 *******************************************************************/

#include <stddef.h>
#include "md5.h"
#include "../../SMemory/smemory.h"
#include <stdlib.h> 

#ifndef BYTE
typedef unsigned char BYTE;
#endif

#ifndef UINT
typedef unsigned int UINT;
#endif

struct MD5Context
{
	UINT buf[4];
	UINT bits[2];
	BYTE  block[64];
};


/**********************************************************************
 *
 * ByteReverse() solves the problem of Big and Little-Endianess,
 * calculating on the fly if we need to reverse bytes.
 *
 **********************************************************************/

static void ByteReverse( BYTE *buf, UINT longs)
{
  UINT t;

  if((*(UINT *)("ABCD") >> 24) == 'A') return; // Nothing to do
 
  do { t = (UINT)( (((UINT)buf[3] << 8 | buf[2]) << 16) |
                       ((UINT)buf[1] << 8 | buf[0])          );
       *(UINT *)buf = t;
       buf += 4;
     } while (--longs);
}


/**********************************************************************
 * 
 * MD5's set of non-linear functions
 *
 * According to Rich Schroeppel: 
 *    (X & Y) | (~X & Z) == Z ^ (X & (Y ^ Z)) 
 *          - the last does not require ~ operation
 *
 **********************************************************************/
 
#define F1(x, y, z) (z ^ (x & (y ^ z)))
#define F2(x, y, z) F1(z, x, y)
#define F3(x, y, z) (x ^ y ^ z)
#define F4(x, y, z) (y ^ (x | ~z))

/* This is the central step in the MD5 algorithm. */
#define MD5STEP(f, w, x, y, z, data, s) \
	( w += f(x, y, z) + data,  w = w<<s | w>>(32-s),  w += x )


/**********************************************************************
 *
 * MD5Transform() - the main procedure to transform the message 
 * digest (digest[4] array) in according to the current block 
 * of input data (block[16] array)
 *
 **********************************************************************/

static void MD5Transform(UINT digest[4], UINT const block[16])
{
	register UINT a, b, c, d;

    a = digest[0];
    b = digest[1];
    c = digest[2];
    d = digest[3];

    MD5STEP(F1, a, b, c, d, block[0] + 0xd76aa478, 7);
    MD5STEP(F1, d, a, b, c, block[1] + 0xe8c7b756, 12);
    MD5STEP(F1, c, d, a, b, block[2] + 0x242070db, 17);
    MD5STEP(F1, b, c, d, a, block[3] + 0xc1bdceee, 22);
    MD5STEP(F1, a, b, c, d, block[4] + 0xf57c0faf, 7);
    MD5STEP(F1, d, a, b, c, block[5] + 0x4787c62a, 12);
    MD5STEP(F1, c, d, a, b, block[6] + 0xa8304613, 17);
    MD5STEP(F1, b, c, d, a, block[7] + 0xfd469501, 22);
    MD5STEP(F1, a, b, c, d, block[8] + 0x698098d8, 7);
    MD5STEP(F1, d, a, b, c, block[9] + 0x8b44f7af, 12);
    MD5STEP(F1, c, d, a, b, block[10] + 0xffff5bb1, 17);
    MD5STEP(F1, b, c, d, a, block[11] + 0x895cd7be, 22);
    MD5STEP(F1, a, b, c, d, block[12] + 0x6b901122, 7);
    MD5STEP(F1, d, a, b, c, block[13] + 0xfd987193, 12);
    MD5STEP(F1, c, d, a, b, block[14] + 0xa679438e, 17);
    MD5STEP(F1, b, c, d, a, block[15] + 0x49b40821, 22);

    MD5STEP(F2, a, b, c, d, block[1] + 0xf61e2562, 5);
    MD5STEP(F2, d, a, b, c, block[6] + 0xc040b340, 9);
    MD5STEP(F2, c, d, a, b, block[11] + 0x265e5a51, 14);
    MD5STEP(F2, b, c, d, a, block[0] + 0xe9b6c7aa, 20);
    MD5STEP(F2, a, b, c, d, block[5] + 0xd62f105d, 5);
    MD5STEP(F2, d, a, b, c, block[10] + 0x02441453, 9);
    MD5STEP(F2, c, d, a, b, block[15] + 0xd8a1e681, 14);
    MD5STEP(F2, b, c, d, a, block[4] + 0xe7d3fbc8, 20);
    MD5STEP(F2, a, b, c, d, block[9] + 0x21e1cde6, 5);
    MD5STEP(F2, d, a, b, c, block[14] + 0xc33707d6, 9);
    MD5STEP(F2, c, d, a, b, block[3] + 0xf4d50d87, 14);
    MD5STEP(F2, b, c, d, a, block[8] + 0x455a14ed, 20);
    MD5STEP(F2, a, b, c, d, block[13] + 0xa9e3e905, 5);
    MD5STEP(F2, d, a, b, c, block[2] + 0xfcefa3f8, 9);
    MD5STEP(F2, c, d, a, b, block[7] + 0x676f02d9, 14);
    MD5STEP(F2, b, c, d, a, block[12] + 0x8d2a4c8a, 20);

    MD5STEP(F3, a, b, c, d, block[5] + 0xfffa3942, 4);
    MD5STEP(F3, d, a, b, c, block[8] + 0x8771f681, 11);
    MD5STEP(F3, c, d, a, b, block[11] + 0x6d9d6122, 16);
    MD5STEP(F3, b, c, d, a, block[14] + 0xfde5380c, 23);
    MD5STEP(F3, a, b, c, d, block[1] + 0xa4beea44, 4);
    MD5STEP(F3, d, a, b, c, block[4] + 0x4bdecfa9, 11);
    MD5STEP(F3, c, d, a, b, block[7] + 0xf6bb4b60, 16);
    MD5STEP(F3, b, c, d, a, block[10] + 0xbebfbc70, 23);
    MD5STEP(F3, a, b, c, d, block[13] + 0x289b7ec6, 4);
    MD5STEP(F3, d, a, b, c, block[0] + 0xeaa127fa, 11);
    MD5STEP(F3, c, d, a, b, block[3] + 0xd4ef3085, 16);
    MD5STEP(F3, b, c, d, a, block[6] + 0x04881d05, 23);
    MD5STEP(F3, a, b, c, d, block[9] + 0xd9d4d039, 4);
    MD5STEP(F3, d, a, b, c, block[12] + 0xe6db99e5, 11);
    MD5STEP(F3, c, d, a, b, block[15] + 0x1fa27cf8, 16);
    MD5STEP(F3, b, c, d, a, block[2] + 0xc4ac5665, 23);

    MD5STEP(F4, a, b, c, d, block[0] + 0xf4292244, 6);
    MD5STEP(F4, d, a, b, c, block[7] + 0x432aff97, 10);
    MD5STEP(F4, c, d, a, b, block[14] + 0xab9423a7, 15);
    MD5STEP(F4, b, c, d, a, block[5] + 0xfc93a039, 21);
    MD5STEP(F4, a, b, c, d, block[12] + 0x655b59c3, 6);
    MD5STEP(F4, d, a, b, c, block[3] + 0x8f0ccc92, 10);
    MD5STEP(F4, c, d, a, b, block[10] + 0xffeff47d, 15);
    MD5STEP(F4, b, c, d, a, block[1] + 0x85845dd1, 21);
    MD5STEP(F4, a, b, c, d, block[8] + 0x6fa87e4f, 6);
    MD5STEP(F4, d, a, b, c, block[15] + 0xfe2ce6e0, 10);
    MD5STEP(F4, c, d, a, b, block[6] + 0xa3014314, 15);
    MD5STEP(F4, b, c, d, a, block[13] + 0x4e0811a1, 21);
    MD5STEP(F4, a, b, c, d, block[4] + 0xf7537e82, 6);
    MD5STEP(F4, d, a, b, c, block[11] + 0xbd3af235, 10);
    MD5STEP(F4, c, d, a, b, block[2] + 0x2ad7d2bb, 15);
    MD5STEP(F4, b, c, d, a, block[9] + 0xeb86d391, 21);

    digest[0] += a;
    digest[1] += b;
    digest[2] += c;
    digest[3] += d;
}



/**********************************************************************
 * 
 * MD5Init() initializes a context of the message digest calculations.
 * The memory to store the context is reserved before and the pointer 
 * on the context's memory stored in void element 
 *
 **********************************************************************/

int MD5Init( MD5Context * ctx )
{
  ctx->buf[0] = 0x67452301;
  ctx->buf[1] = 0xefcdab89;
  ctx->buf[2] = 0x98badcfe;
  ctx->buf[3] = 0x10325476;
  ctx->bits[0] = 0;
  ctx->bits[1] = 0;
  return MD5_OK;
}


/**********************************************************************
 * 
 * MD5Process() is called multiple times (if needed) until our message 
 * buffer is not empty
 *
 **********************************************************************/

int MD5Process( MD5Context *ctx, const BYTE *data, int length )
{
  UINT t;

  /* Update bitcount: Shift left on 3 bites used because length is 
     in bytes, but our counter is in bites
  */
  t = ctx->bits[0];
  if ((ctx->bits[0] = t + ((UINT)length << 3)) < t)
	  ctx->bits[1]++;		           /* Carry from low to high */
  
  ctx->bits[1] += length >> 29;

  /* If the previous *data was not multiply of 64 bytes,
   * we currently store remaining of the previous Message 
   * in ctx->block[] array.
   * Calculate, how many bytes we currently store in ctx->block[]
   */
  
  /* calculate in t number of BUSY bytes in ctx->block[] buffer */
  t = (t >> 3) & 0x3f;

  if (t) 
  /* ctx->block[] already stores t bytes */
  {
    BYTE *p = ctx->block + t;

    t = 64 - t;
    if ((UINT)length < t) 
    /* even we add t bytes to ctx->block, 
       it is continue to be small for digest 
       calculating, everything OK, we just return 
	 */
    { SMemCpy(p, data, length);
      return MD5_OK;
    }
	SMemCpy(p, data, t); /* fill ctx->block and calculate digest */
	ByteReverse(ctx->block, 16);
	MD5Transform(ctx->buf, (UINT *)ctx->block);
	data += t;
	length -= t;
  }
  
  /* Process data in 64-BYTE blocks */
  while (length >= 64) 
  {
    SMemCpy(ctx->block, data, 64);
    ByteReverse(ctx->block, 16);
    MD5Transform(ctx->buf, (UINT *)ctx->block);
    data += 64;
    length -= 64;
  }

  /* Copy remaining bytes of data (it's size < 64 bytes) to the ctx->block */
  SMemCpy(ctx->block, data, length);
  return MD5_OK;
}



/**********************************************************************
 *
 * MD5Final(): All message buffer is transformed using MD5Process().
 * To get the final Message Digest, we have to call MD5Final()
 *
 **********************************************************************/

int MD5Final( MD5Context *ctx, BYTE *digest )
{
  UINT count;
  BYTE *p;

  /* Calculate number of BUSY bytes in ctx->block[] buffer */
  count = (ctx->bits[0] >> 3) & 0x3F;

  /* Set the first char of padding to 0x80.  This is safe since there is
     always at least one BYTE free */
  p = ctx->block + count;
  *p++ = 0x80;

  /* Calculate number of FREE bytes in sha_ctx->data buffer */
  count = 64 - 1 - count; /* ctx->data is 64 bytes, but we already used one BYTE for 0x80 */

  /*   If count is more than 8, it means that we have the place for 
   * two UINT to place the length of message. 
   *    If no, we have to pad the current 64-BYTE block by 0, re-calculate 
   * digest, then fill the first 56 bytes in sha_ctx->data by 0, place 
   * the length of Message to the last 8 bytes and calculate the final Digest
   */
  if (count < 8) 
  {
    SMemSet(p, 0, count);
    ByteReverse(ctx->block, 16);
    MD5Transform(ctx->buf, (UINT *)ctx->block);
    /* prepare to calculate the final digest, 
	   where the length will be written in last 8 bytes 
	*/
    SMemSet(ctx->block, 0, 56);
  } 
  else 
  { /* Pad block to 56 bytes */
    SMemSet(p, 0, count - 8);
  }
    
  ByteReverse(ctx->block, 14);

  /* Append length of the message in bits to the last 8 bytes and 
     calculate Final Digest 
  */
  ((UINT *)ctx->block)[14] = ctx->bits[0];
  ((UINT *)ctx->block)[15] = ctx->bits[1];

  MD5Transform(ctx->buf, (UINT *)ctx->block);
  ByteReverse((BYTE *)ctx->buf, 4);
  SMemCpy(digest, ctx->buf, 16);
  
  SMemSet(ctx, 0, sizeof(ctx));	/* Clear the MD5 context */

  return MD5_OK;
}



/**********************************************************************
 *
 *   MD5Update() - the size of BYTE array "data" must be 
 *  equal to the block size of MD5 algorithm, i.e. 512 bits = 64 bytes
 *
 **********************************************************************/

int MD5Update( MD5Context *ctx, BYTE *data)
{ 
   /* 16 * sizeof(UINT) == 64 bytes */
  SMemCpy( ctx->block, data, 64 ); 

  ByteReverse(ctx->block, 16);
  MD5Transform(ctx->buf, (UINT *)ctx->block);

  return MD5_OK;
}



/**********************************************************************
 *
 *   MD5Allocate() - allocates memory needed for context structure, and
 * the structures within (or pointed to by) context structure. Uses 
 * SMemory functions
 *
 **********************************************************************/

int MD5Allocate(MD5Context **ctx)
{
  *ctx = (MD5Context *)SMalloc( sizeof(MD5Context) );
  if (*ctx == NULL ) 
	return MD5_MALLOC_ERROR;
  return MD5_OK;
}

/**********************************************************************
 * 
 * MD5Free() - Frees resources allocated in MD5Allocate
 *
 **********************************************************************/

int MD5Free( MD5Context **ctx)
{
  SFree(*ctx);
  *ctx = 0;
  return MD5_OK;
}

/**********************************************************************
 * 
 * MD5MakeDigest() - Helpful function if the whole message is in the 
 * memory and we have to create digest for it.
 *
 **********************************************************************/

int MD5MakeDigest( const BYTE *message, int messageLength, BYTE *digest )
{
	MD5Context *context;
  
	if ( MD5Allocate( &context ) != MD5_OK )
		return MD5_ERROR_INTERNAL_ERROR;


    if ( MD5Init( context ) != MD5_OK)
	{	MD5Free( &context );
		return MD5_ERROR_INTERNAL_ERROR;
	}

    if ( MD5Process( context, message, messageLength ) != MD5_OK )
	{	MD5Free( &context );
		return MD5_ERROR_INTERNAL_ERROR;
	}

    if ( MD5Final( context, digest ) != MD5_OK )
	{	MD5Free( &context );
		return MD5_ERROR_INTERNAL_ERROR;
	}

    MD5Free( &context );
	return MD5_OK;
}
