/*******************************************************************
 *
 *      Copyright (c) 2014 Jetico, Inc., Finland
 *      All rights reserved.
 *
 *      File: sha3-512.cpp
 *      Revision:
 *      Created:
 *      Description: implementation of SHA3 (Keccak) 512 bit Secure Hash Algorithm
 *
 *******************************************************************/

#include <string.h>
#include <stdlib.h>
#include "../hashAlgs.h"
#include "sha3-512.h"
#include "round.h"
#include "keccak_f.h"

#ifdef _BIG_ENDIAN_DEFINED

static void revertByteBuffer( bc_u8 *bf, bc_32 bfSize )
{
	for( bc_32 i = 0; i < bfSize / 2; i++ )
	{
		bc_u8 bt = bf[i];
		bf[i] = bf[bfSize-i-1];
		bf[bfSize-i-1] = bt;
	}
}

static void revertInt64Buffer( bc_u64 *bf, bc_32 bfSize )
{
	for( int i = 0; i < bfSize; i++ )
		revertByteBuffer( (bc_u8*)(bf+i), sizeof( bc_u64 ) );
}
#endif


int SHA3_512_AutoTest()
{
	// http://en.wikipedia.org/wiki/SHA-3
	//	Keccak-512("")
	//	Out = 0x 0eab42de4c3ceb9235fc91acffe746b29c29a8c366b7c60e4e67c466f36a4304c00fa9caf9d87976ba469bcbe06713b435f091ef2769fb160cdab33d3670680e

	BYTE digest[64];

	const char *text = "";

	if ( SHA3_512_MakeDigest( (const bc_u8*)text, strlen(text), digest ) != SHA3_512_ERROR_NO )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	bc_u8 T[]= {
		0x0e, 0xab, 0x42, 0xde, 0x4c, 0x3c, 0xeb, 0x92,
		0x35, 0xfc, 0x91, 0xac, 0xff, 0xe7, 0x46, 0xb2,
		0x9c, 0x29, 0xa8, 0xc3, 0x66, 0xb7, 0xc6, 0x0e,
		0x4e, 0x67, 0xc4, 0x66, 0xf3, 0x6a, 0x43, 0x04,
		0xc0, 0x0f, 0xa9, 0xca, 0xf9, 0xd8, 0x79, 0x76,
		0xba, 0x46, 0x9b, 0xcb, 0xe0, 0x67, 0x13, 0xb4,
		0x35, 0xf0, 0x91, 0xef, 0x27, 0x69, 0xfb, 0x16,
		0x0c, 0xda, 0xb3, 0x3d, 0x36, 0x70, 0x68, 0x0e,
	};

	bool ok = ( memcmp( digest, T, sizeof( T ) ) == 0 );

	if ( !ok )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	text = "The quick brown fox jumps over the lazy dog";

	bc_u8 outT[] = {
		0xD1, 0x35, 0xBB, 0x84, 0xD0, 0x43, 0x9D, 0xBA,
		0xC4, 0x32, 0x24, 0x7E, 0xE5, 0x73, 0xA2, 0x3E,
		0xA7, 0xD3, 0xC9, 0xDE, 0xB2, 0xA9, 0x68, 0xEB,
		0x31, 0xD4, 0x7C, 0x4F, 0xB4, 0x5F, 0x1E, 0xF4,
		0x42, 0x2D, 0x6C, 0x53, 0x1B, 0x5B, 0x9B, 0xD6,
		0xF4, 0x49, 0xEB, 0xCC, 0x44, 0x9E, 0xA9, 0x4D,
		0x0A, 0x8F, 0x05, 0xF6, 0x21, 0x30, 0xFD, 0xA6,
		0x12, 0xDA, 0x53, 0xC7, 0x96, 0x59, 0xF6, 0x09
	};

	if ( SHA3_512_MakeDigest( (const bc_u8*)text, strlen(text), digest ) != SHA3_512_ERROR_NO )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	ok = ( memcmp( digest, outT, sizeof( outT ) ) == 0 );

	if ( !ok )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	return SHA3_512_ERROR_NO;

}

struct SHA3_512_CTX
{
	bc_u64 S[5][5];
};

int SHA3_512_Allocate( void **ctx )
{
	if ( ctx == NULL )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	*ctx = malloc( sizeof( SHA3_512_CTX ) );

	if ( *ctx == NULL )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	return SHA3_512_ERROR_NO;
}

int SHA3_512_Free(void **ctx)
{
	if ( ctx == NULL )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	if ( *ctx == NULL )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	memset( *ctx, 0, sizeof( SHA3_512_CTX ) );

	*ctx = NULL;

	return SHA3_512_ERROR_NO;
}

int SHA3_512_Init( void *ctx )
{
	if ( ctx == NULL )
		return SHA3_512_ERROR_INTERNAL_ERROR;

	memset( ctx, 0, sizeof( SHA3_512_CTX ) );

	return SHA3_512_ERROR_NO;
}

static bc_u8 *padding( const bc_u8* M, bc_u32* S)
{
	bc_32 i=*S;
	bc_32 newS=(*S+72-(*S%72));;
	bc_u8 *nM;

	nM=(bc_u8 *)malloc(*S+(72-(*S%72)));

	/*Copy string*/
	for(bc_u32 j=0;j<*S;j++){
		*(nM+j)=*(M+j);
	}

	*(nM+i)=0x01;
	i++;

	while(i<(newS-1)){
		*(nM+i)=0x00;
		i++;
	}

	*(nM+i)=0x80;
	i++;
	*S=i;
	return nM;
}

static const bc_u32 r=72;
static const bc_u32 w=8;

int SHA3_512_Process( void *ctx, const unsigned char *M, bc_u32 size )
{
	bc_u8 *mBuff = NULL;
	bc_u64 *nM;

	/*Padding*/
	if( ( size == 0 ) || ( (size%r)!=0) ) //r=72 bytes
	{
		mBuff = padding(M,&size);
		M = mBuff;		
	}
#ifdef _BIG_ENDIAN_DEFINED
	else
	{
		mBuff = (bc_u8 *)malloc( r );
		memcpy( mBuff, M, r );
		M = mBuff;
	}

	nM=(bc_u64 *)mBuff;
	revertInt64Buffer( nM, size / sizeof( bc_u64 ) );

#else
	nM=(bc_u64 *)M;
#endif

	SHA3_512_CTX &C = *(SHA3_512_CTX *)ctx;

	/*Absorbing Phase*/
	for(bc_u32 i=0;i<size/r;i++) //Each block has 72 bytes
	{
		for(bc_32 y=0;y<5;y++)
		{
			for(bc_32 x=0;x<5;x++)
			{
				if((x+5*y)<(r/w))
				{
					C.S[x][y] = C.S[x][y] ^ *(nM+i*9+x+5*y);
				}
			}
		}

		keccak_f( C.S );
	}

	if ( mBuff != NULL )
		free( mBuff );
	
	return SHA3_512_ERROR_NO;
}

//int SHA3_512_Update(void *ctx, unsigned char *data)
//{
//	return SHA3_512_ERROR_INTERNAL_ERROR;
//}

int SHA3_512_Final(void *ctx, unsigned char *digest)
{
	/*Squeezing phase*/

	bc_u64 *Z = (bc_u64 *)digest;
	memset( digest, 0, 64 );

	SHA3_512_CTX &C = *(SHA3_512_CTX *)ctx;

	for(bc_32 b=0, y=0;y<5;y++)
	{
		for(bc_32 x=0; (x<5) && ( b<8 ) ; x++ )
		{
			if((x+5*y)<(r/w))
			{
				*(Z+b)^=C.S[x][y];
				b++;
			}
		}
	}

#ifdef _BIG_ENDIAN_DEFINED
	revertInt64Buffer( Z, 8 );
#endif

	return SHA3_512_ERROR_NO;
}

int SHA3_512_MakeDigest( const unsigned char *message, bc_u32 messageLength, unsigned char *digest )
{
	void *context = NULL;

    if ( SHA3_512_Allocate(&context) != SHA3_512_ERROR_NO)
		return SHA3_512_ERROR_INTERNAL_ERROR;
  
    if ( SHA3_512_Init(context) != SHA3_512_ERROR_NO)
		return SHA3_512_ERROR_INTERNAL_ERROR;

    if ( SHA3_512_Process(context, message, messageLength) != SHA3_512_ERROR_NO )
		return SHA3_512_ERROR_INTERNAL_ERROR;

    if ( SHA3_512_Final(context, digest) != SHA3_512_ERROR_NO )
		return SHA3_512_ERROR_INTERNAL_ERROR;
	
    if ( SHA3_512_Free(&context) != SHA3_512_ERROR_NO)
		return SHA3_512_ERROR_INTERNAL_ERROR;

	return SHA3_512_ERROR_NO;
}
