#include "../publicKeyAlgs/Power/Windefs.h"
#include "hashAlgs.h"
#include "cHashAlg.h"
#include "cHashSHA1.h"
#include "cHashMD5.h"
#include "cHashRipemd.h"
#include "cHashSHA256.h"
#include "cHashSHA512.h"
#include "cHashAlgSHA3-512.h"
#include "cHashWhirlpool512.h"
#include "cHashSkein512.h"
#include "../PGPPkts/pktdefs.h"

typedef cHashAlg *typeCreateAlg();

const char *GetHashAlgName( BYTE algID )
{	switch ( algID )
	{
		case pgphaMD5	: return "MD5";
		case pgphaSHA1	: return "SHA1";
		case pgphaRIPEMD160: return "RIPEMD160";
		case pgphaMD2	: return "MD2";
		case pgphaSHA256: return "SHA-256";
		case pgphaSHA384: return "SHA-384";
		case pgphaSHA512: return "SHA-512";
		case pgphaSHA224: return "SHA-224";

		default : return NULL;
	}
}

const char *ha_getHashAlgName( BYTE algID )
{
	if ( algID < bchaMinAlgID )
		return GetHashAlgName( algID );

	switch( algID )
	{
		case bchaSHA256			: return GetHashAlgName( pgphaSHA256 );
		case bchaWhirlpool512	: return "Whirlpool-512";
		case bchaSHA3_512		: return "SHA3-512";
		case bchaSkein512		: return "Skein-512";
		default : return NULL;
	}
}


//
//	9.4. Hash Algorithms
//
//	ID           Algorithm                              Text Name
//	--           ---------                              ---- ----
//	1          - MD5                                    "MD5"
//	2          - SHA-1                                  "SHA1"
//	3          - RIPE-MD/160                            "RIPEMD160"
//	4          - Reserved for double-width SHA (experimental)
//	5          - MD2                                    "MD2"
//	6          - Reserved for TIGER/192                 "TIGER192"
//	7          - Reserved for HAVAL (5 pass, 160-bit)
//	"HAVAL-5-160"
//	100 to 110 - Private/Experimental algorithm.
//
//	Implementations MUST implement SHA-1. Implementations SHOULD
//	implement MD5.
//

static int IDsTab[] = {
	pgphaMD5,
	pgphaSHA1,
	pgphaRIPEMD160,
	bchaSHA256,
	pgphaSHA512,
	bchaWhirlpool512,
	bchaSHA3_512,
	bchaSkein512
};

int ha_getIDsOfAvailableAlgAlgorithms( int *pIDsCount, int **ppIDsTab )
{
	if ( pIDsCount != NULL )
		*pIDsCount = sizeof( IDsTab ) / sizeof( *IDsTab );

	if ( ppIDsTab != NULL )
		*ppIDsTab = IDsTab;

	return sae_ok;
}



struct sHashAlgTabItem { int algID;	typeCreateAlg *createFunc; int digestSize; }
	hashAlgTbl[] = 
		{
			{ pgphaMD5,			cHashMD5			::createAlg, MD5_DIGEST_SIZE },
			{ pgphaSHA1,		cHashSHA1			::createAlg, SHA1_DIGEST_SIZE },
			{ pgphaRIPEMD160,	cHashRipemd			::createAlg, RIPEMD_DIGEST_SIZE_BYTES },
			{ bchaSHA256,		cHashSHA256			::createAlg, SHA256_DIGEST_SIZE },
			{ pgphaSHA256,		cHashSHA256			::createAlg, SHA256_DIGEST_SIZE },
			{ pgphaSHA512,		cHashSHA512			::createAlg, SHA512_DIGEST_SIZE },
			{ bchaWhirlpool512,	cHashWhirlpool512	::createAlg, Whirlpool512_DIGEST_SIZE },
			{ bchaSkein512,		cHashSkein512		::createAlg, Skein512_DIGEST_SIZE },
			{ bchaSHA3_512,		cHashSHA3_512		::createAlg, SHA3_512_DIGEST_SIZE },
		};


int ha_createAlgorithm( HHA *pHHA, int algID )
{
	int q = sizeof( hashAlgTbl ) / sizeof( hashAlgTbl[0] );

	while ( q-- )
	{
		if ( hashAlgTbl[q].algID == algID )
		{
			*pHHA = hashAlgTbl[q].createFunc();
			return hae_ok;
		}
	}

	return hae_invalidAlgorithmID;
}

int ha_init( cHashAlg *ha )
{
	return ha->Init();
}

void ha_free( cHashAlg *ha )
{
	delete ha;
}

int ha_getDigestSize( cHashAlg *ha )
{
	return ha->getDigestSize();
}

int ha_getDigestSize( int *pSize, int algID )
{
	int q = sizeof( hashAlgTbl ) / sizeof( hashAlgTbl[0] );

	while ( q-- )
	{
		if ( hashAlgTbl[q].algID == algID )
		{
			*pSize = hashAlgTbl[q].digestSize;
			return hae_ok;
		}
	}

	return hae_invalidAlgorithmID;
}

int ha_makeDigest(
	cHashAlg *ha,	// Handle of hash algorithm
	BYTE *digest,	// Output digest buffer.
					// Note, size of the buffer has not to be less then size
					// the ha_getDigestSize returns.
	BYTE *message,	// Input message buffer. 
	int messageSize	// Size of message buffer.
)
{
	return ha->makeDigest( digest, message, messageSize );
}

int ha_makeDigestEx(
	int algID,		// Hash algorithm ID
	BYTE *digest,	// Output digest buffer.
					// Note, size of the buffer has not to be less then size
					// the ha_getDigestSize returns.
	BYTE *message,	// Input message buffer. 
	int messageSize	// Size of message buffer.
)
{
	if ( digest == NULL )
		return pkae_invalidInputData;

	cHashAlg *ha;
	int res = ha_createAlgorithm( &ha, algID );
	if ( res != hae_ok ) return res;

	res = ha->makeDigest( digest, message, messageSize );

	ha_free( ha );

	return res;
}


int ha_process( cHashAlg *ha, const BYTE *data, int length )
{
	return ha->Process( data, length );
}

int ha_final( cHashAlg *ha, BYTE *digest )
{
	return ha->Final( digest );
}
