/*
 	Copyright (c) 1994-2000 Jetico, Inc., Finland
 	All rights reserved.

	File: ripemd.cpp
	Description: RIPEMD-160 message-digest algorithm.
*/

/* 
 * On the Intel processors Circular Shift Left operation 
 * (Rotation Left) may be done as one command for processor
 */

/* #define __INTEL_PLATFORM */

#ifdef __INTEL_PLATFORM
  #define ROL(X, n) _asm rol X, n
#else
  #define ROL( X, n )	( ( ( X ) << n ) | ( ( X ) >> ( 32 - n ) ) )
#endif

#include <stddef.h>
#include "ripemd.h"
#include "../../SMemory/smemory.h"

#ifndef UINT
typedef unsigned int UINT;
#endif

#define RIPEMD_DIGEST_SIZE_UINT32B (RIPEMD_DIGEST_SIZE_BYTES / sizeof(UINT))
#define RIPEMD_BLOCK_SIZE_BYTES   64


/* The structure for storing the RIPE-MD context */
struct RipemdContext
{    UINT digest[RIPEMD_DIGEST_SIZE_UINT32B]; /* Message digest */
     UINT countLo, countHi;                   /* 64-bit BYTE count */
     UINT data[16];                           /* RIPE-MD data buffer */
};


/* 
 * The RIPEMD-160 initial values 
 */

#define H0	0x67452301
#define H1	0xEFCDAB89
#define H2	0x98BADCFE
#define H3	0x10325476
#define H4	0xC3D2E1F0


 /* 
 * The RIPEMD-160 constants 
 * In the RIPEMD-160 description the terminology K ... and K'
 * is used. We will use K ... and K_ ...
 */
 
/* constants for rounds  0-15 */
#define K1		0x00000000
#define K1_		0x50A28BE6
 /* constants for rounds 16-31 */
#define K2		0x5A827999
#define K2_ 	0x5C4DD124
 /* constants for rounds 32-47 */
#define K3		0x6ED9EBA1
#define K3_		0x6D703EF3
 /* constants for rounds 48-63 */
#define K4		0x8F1BBCDC
#define K4_		0x7A6D76E9
 /* constants for rounds 64-79 */
#define K5		0xA953FD4E
#define K5_		0x00000000


/* 
 * RIPEMD-160 set of nonlinear functions.
 *
 * According to Rich Schroeppel: 
 *    (X & Y) | (~X & Z) == Z ^ (X & (Y ^ Z)) // the last does not require ~ operation
 */

/* For step  0-15 */
#define f1(X,Y,Z)	((X) ^ (Y) ^ (Z)) 
/* For step 16-31: Schroeppel's optimization used */
#define f2(X,Y,Z)	((Z) ^ ((X) & ((Y) ^ (Z)))) 
/* For step 32-47 */
#define f3(X,Y,Z)   (((X) | (~(Y))) ^ (Z))
/* For step 48-63: Schroeppel's optimization used also, but X, Y, Z rotated */
#define f4(X,Y,Z)	((Y) ^ ((Z) & ((X) ^ (Y)))) 
/* For step 64-79 */
#define f5(X,Y,Z)   ((X) ^ ((Y) | (~(Z))))


/* 
  According to the RIPEMD-160 description in the article of H. Dobbertin, 
  A. Bosselaers, B. Preneel "RIPEMD-160, a strengthened version of RIPEMD.",
  the main RIPEMD-160 loop uses a, b, c, d, e (and a', b', c', d', e' in 
  parallel) numbers. Only a and c are changed on each step of the loop. 
  b, d, e are swapped between each other:

  if F(a,b,c,d,e) = ((a + f(b,c,d) + data + K) <<< s ) + e, then 

  a1 = e0
  b1 = F(a0,b0,c0,d0,e0)
  c1 = b0
  d1 = c0<<<10
  e1 = d0

  Hence, for first 3 steps we get:

  b1 = F(a0,b0,c0,d0,e0) = F( a0, b0,                c0,                d0,      e0)
  b2 = F(a1,b1,c1,d1,e1) = F( e0, F(a0,b0,c0,d0,e0), b0,                c0<<<10, d0)
  b3 = F(a2,b2,c2,d2,e2) = F( d0, F(a1,b1,c1,d1,e1), F(a0,b0,c0,d0,e0), b0<<<10, c0)

  From one hand, in each step we need one Temporary local variable to 
  store the result of F(), from the other hand, we need a0 only on the 
  first step. Hence, we can use a0 as this temporary variable. Then, if in 
  each step we will shift left c variable, the next step will look 
  simpler. The first 3 steps will look like:

  1: a0'  = F(a0, b0,  c0,  d0 , e0 ); c0'  = c0  <<< 10
  2: e0'  = F(e0, a0', b0,  c0', d0 ); b0'  = b0  <<< 10
  3: d0'  = F(d0, e0,  a0', b0', c0'); a0'' = a0' <<< 10 

  Hence, the first 3 rounds in RIPEMD-160 loop will look like:

  Round0( A, B, C, D, E) = { A = F(A,B,C,D,E); C = C << 10; }
  Round1( E, A, B, C, D) = { E = F(E,A,B,C,D); B = B << 10; }
  Round2( D, E, A, B, C) = { D = F(D,E,A,B,C); A = A << 10; }

  So, we can make the following #define :
*/

/* (mask 0xffffffff is needed if our machine is 64-bits) */

#define MASK32( x )  x

#define Round( a, b, c, d, e, f, s, k, data ) \
	a += f( b, c, d ) + data + k; \
	a = MASK32( ROL( MASK32( a ),s ) + e ); \
	c = MASK32( ROL( c, 10 ) )

/*

#define Round( A, B, C, D, E, f, s, K, data ) \
 A = ((ROL(((A + f( B, C, D ) + data + K) & 0xffffffff), s) + E) & 0xffffffff); \
 C = ((ROL(C, 10)) & 0xffffffff)
*/

static void ripemdUpdateDigest( UINT *Digest, UINT *MessageBlock )
 {
   /* In the RIPEMD-160 description the terminology A, B ... and A', B'...
    * is used. We will use a, b ... and a1 , b1 ...
    */
	UINT a, b, c, d, e, a1, b1, c1, d1, e1, tmp;

	a = a1 = Digest[0];
	b = b1 = Digest[1];
	c = c1 = Digest[2];
	d = d1 = Digest[3];
	e = e1 = Digest[4];

    /* 
	 * 80 rounds to calculate a, b, c, d, e 
	 */
	
	/* 0-15 rounds */
	Round( a, b, c, d, e, f1, 11, K1, MessageBlock[ 0] );
	Round( e, a, b, c, d, f1, 14, K1, MessageBlock[ 1] );
	Round( d, e, a, b, c, f1, 15, K1, MessageBlock[ 2] );
	Round( c, d, e, a, b, f1, 12, K1, MessageBlock[ 3] );
	Round( b, c, d, e, a, f1,  5, K1, MessageBlock[ 4] );
	Round( a, b, c, d, e, f1,  8, K1, MessageBlock[ 5] );
	Round( e, a, b, c, d, f1,  7, K1, MessageBlock[ 6] );
	Round( d, e, a, b, c, f1,  9, K1, MessageBlock[ 7] );
	Round( c, d, e, a, b, f1, 11, K1, MessageBlock[ 8] );
	Round( b, c, d, e, a, f1, 13, K1, MessageBlock[ 9] );
	Round( a, b, c, d, e, f1, 14, K1, MessageBlock[10] );
	Round( e, a, b, c, d, f1, 15, K1, MessageBlock[11] );
	Round( d, e, a, b, c, f1,  6, K1, MessageBlock[12] );
	Round( c, d, e, a, b, f1,  7, K1, MessageBlock[13] );
	Round( b, c, d, e, a, f1,  9, K1, MessageBlock[14] );
	Round( a, b, c, d, e, f1,  8, K1, MessageBlock[15] );

	/* 16-31 rounds */
	Round( e, a, b, c, d, f2,  7, K2, MessageBlock[ 7] );
	Round( d, e, a, b, c, f2,  6, K2, MessageBlock[ 4] );
	Round( c, d, e, a, b, f2,  8, K2, MessageBlock[13] );
	Round( b, c, d, e, a, f2, 13, K2, MessageBlock[ 1] );
	Round( a, b, c, d, e, f2, 11, K2, MessageBlock[10] );
	Round( e, a, b, c, d, f2,  9, K2, MessageBlock[ 6] );
	Round( d, e, a, b, c, f2,  7, K2, MessageBlock[15] );
	Round( c, d, e, a, b, f2, 15, K2, MessageBlock[ 3] );
	Round( b, c, d, e, a, f2,  7, K2, MessageBlock[12] );
	Round( a, b, c, d, e, f2, 12, K2, MessageBlock[ 0] );
	Round( e, a, b, c, d, f2, 15, K2, MessageBlock[ 9] );
	Round( d, e, a, b, c, f2,  9, K2, MessageBlock[ 5] );
	Round( c, d, e, a, b, f2, 11, K2, MessageBlock[ 2] );
	Round( b, c, d, e, a, f2,  7, K2, MessageBlock[14] );
	Round( a, b, c, d, e, f2, 13, K2, MessageBlock[11] );
	Round( e, a, b, c, d, f2, 12, K2, MessageBlock[ 8] );
                                                       
	/* 32-47 rounds */
	Round( d, e, a, b, c, f3, 11, K3, MessageBlock[ 3] );
	Round( c, d, e, a, b, f3, 13, K3, MessageBlock[10] );
	Round( b, c, d, e, a, f3,  6, K3, MessageBlock[14] );
	Round( a, b, c, d, e, f3,  7, K3, MessageBlock[ 4] );
	Round( e, a, b, c, d, f3, 14, K3, MessageBlock[ 9] );
	Round( d, e, a, b, c, f3,  9, K3, MessageBlock[15] );
	Round( c, d, e, a, b, f3, 13, K3, MessageBlock[ 8] );
	Round( b, c, d, e, a, f3, 15, K3, MessageBlock[ 1] );
	Round( a, b, c, d, e, f3, 14, K3, MessageBlock[ 2] );
	Round( e, a, b, c, d, f3,  8, K3, MessageBlock[ 7] );
	Round( d, e, a, b, c, f3, 13, K3, MessageBlock[ 0] );
	Round( c, d, e, a, b, f3,  6, K3, MessageBlock[ 6] );
	Round( b, c, d, e, a, f3,  5, K3, MessageBlock[13] );
	Round( a, b, c, d, e, f3, 12, K3, MessageBlock[11] );
	Round( e, a, b, c, d, f3,  7, K3, MessageBlock[ 5] );
	Round( d, e, a, b, c, f3,  5, K3, MessageBlock[12] );

	/* 48-63 rounds */
	Round( c, d, e, a, b, f4, 11, K4, MessageBlock[ 1] );
	Round( b, c, d, e, a, f4, 12, K4, MessageBlock[ 9] );
	Round( a, b, c, d, e, f4, 14, K4, MessageBlock[11] );
	Round( e, a, b, c, d, f4, 15, K4, MessageBlock[10] );
	Round( d, e, a, b, c, f4, 14, K4, MessageBlock[ 0] );
	Round( c, d, e, a, b, f4, 15, K4, MessageBlock[ 8] );
	Round( b, c, d, e, a, f4,  9, K4, MessageBlock[12] );
	Round( a, b, c, d, e, f4,  8, K4, MessageBlock[ 4] );
	Round( e, a, b, c, d, f4,  9, K4, MessageBlock[13] );
	Round( d, e, a, b, c, f4, 14, K4, MessageBlock[ 3] );
	Round( c, d, e, a, b, f4,  5, K4, MessageBlock[ 7] );
	Round( b, c, d, e, a, f4,  6, K4, MessageBlock[15] );
	Round( a, b, c, d, e, f4,  8, K4, MessageBlock[14] );
	Round( e, a, b, c, d, f4,  6, K4, MessageBlock[ 5] );
	Round( d, e, a, b, c, f4,  5, K4, MessageBlock[ 6] );
	Round( c, d, e, a, b, f4, 12, K4, MessageBlock[ 2] );
                                                       
	/* 64-79 rounds */
	Round( b, c, d, e, a, f5,  9, K5, MessageBlock[ 4] );
	Round( a, b, c, d, e, f5, 15, K5, MessageBlock[ 0] );
	Round( e, a, b, c, d, f5,  5, K5, MessageBlock[ 5] );
	Round( d, e, a, b, c, f5, 11, K5, MessageBlock[ 9] );
	Round( c, d, e, a, b, f5,  6, K5, MessageBlock[ 7] );
	Round( b, c, d, e, a, f5,  8, K5, MessageBlock[12] );
	Round( a, b, c, d, e, f5, 13, K5, MessageBlock[ 2] );
	Round( e, a, b, c, d, f5, 12, K5, MessageBlock[10] );
	Round( d, e, a, b, c, f5,  5, K5, MessageBlock[14] );
	Round( c, d, e, a, b, f5, 12, K5, MessageBlock[ 1] );
	Round( b, c, d, e, a, f5, 13, K5, MessageBlock[ 3] );
	Round( a, b, c, d, e, f5, 14, K5, MessageBlock[ 8] );
	Round( e, a, b, c, d, f5, 11, K5, MessageBlock[11] );
	Round( d, e, a, b, c, f5,  8, K5, MessageBlock[ 6] );
	Round( c, d, e, a, b, f5,  5, K5, MessageBlock[15] );
	Round( b, c, d, e, a, f5,  6, K5, MessageBlock[13] );

    /* 
	 * 80 rounds to calculate a', b', c', d', e' 
	 */
	
	/* 0-15 rounds */
	Round( a1, b1, c1, d1, e1, f5,  8, K1_, MessageBlock[ 5] );
	Round( e1, a1, b1, c1, d1, f5,  9, K1_, MessageBlock[14] );
	Round( d1, e1, a1, b1, c1, f5,  9, K1_, MessageBlock[ 7] );
	Round( c1, d1, e1, a1, b1, f5, 11, K1_, MessageBlock[ 0] );
	Round( b1, c1, d1, e1, a1, f5, 13, K1_, MessageBlock[ 9] );
	Round( a1, b1, c1, d1, e1, f5, 15, K1_, MessageBlock[ 2] );
	Round( e1, a1, b1, c1, d1, f5, 15, K1_, MessageBlock[11] );
	Round( d1, e1, a1, b1, c1, f5,  5, K1_, MessageBlock[ 4] );
	Round( c1, d1, e1, a1, b1, f5,  7, K1_, MessageBlock[13] );
	Round( b1, c1, d1, e1, a1, f5,  7, K1_, MessageBlock[ 6] );
	Round( a1, b1, c1, d1, e1, f5,  8, K1_, MessageBlock[15] );
	Round( e1, a1, b1, c1, d1, f5, 11, K1_, MessageBlock[ 8] );
	Round( d1, e1, a1, b1, c1, f5, 14, K1_, MessageBlock[ 1] );
	Round( c1, d1, e1, a1, b1, f5, 14, K1_, MessageBlock[10] );
	Round( b1, c1, d1, e1, a1, f5, 12, K1_, MessageBlock[ 3] );
	Round( a1, b1, c1, d1, e1, f5,  6, K1_, MessageBlock[12] );

	/* 16-31 rounds */
	Round( e1, a1, b1, c1, d1, f4,  9, K2_, MessageBlock[ 6] );
	Round( d1, e1, a1, b1, c1, f4, 13, K2_, MessageBlock[11] );
	Round( c1, d1, e1, a1, b1, f4, 15, K2_, MessageBlock[ 3] );
	Round( b1, c1, d1, e1, a1, f4,  7, K2_, MessageBlock[ 7] );
	Round( a1, b1, c1, d1, e1, f4, 12, K2_, MessageBlock[ 0] );
	Round( e1, a1, b1, c1, d1, f4,  8, K2_, MessageBlock[13] );
	Round( d1, e1, a1, b1, c1, f4,  9, K2_, MessageBlock[ 5] );
	Round( c1, d1, e1, a1, b1, f4, 11, K2_, MessageBlock[10] );
	Round( b1, c1, d1, e1, a1, f4,  7, K2_, MessageBlock[14] );
	Round( a1, b1, c1, d1, e1, f4,  7, K2_, MessageBlock[15] );
	Round( e1, a1, b1, c1, d1, f4, 12, K2_, MessageBlock[ 8] );
	Round( d1, e1, a1, b1, c1, f4,  7, K2_, MessageBlock[12] );
	Round( c1, d1, e1, a1, b1, f4,  6, K2_, MessageBlock[ 4] );
	Round( b1, c1, d1, e1, a1, f4, 15, K2_, MessageBlock[ 9] );
	Round( a1, b1, c1, d1, e1, f4, 13, K2_, MessageBlock[ 1] );
	Round( e1, a1, b1, c1, d1, f4, 11, K2_, MessageBlock[ 2] );
                                                       
	/* 32-47 rounds */
	Round( d1, e1, a1, b1, c1, f3,  9, K3_, MessageBlock[15] );
	Round( c1, d1, e1, a1, b1, f3,  7, K3_, MessageBlock[ 5] );
	Round( b1, c1, d1, e1, a1, f3, 15, K3_, MessageBlock[ 1] );
	Round( a1, b1, c1, d1, e1, f3, 11, K3_, MessageBlock[ 3] );
	Round( e1, a1, b1, c1, d1, f3,  8, K3_, MessageBlock[ 7] );
	Round( d1, e1, a1, b1, c1, f3,  6, K3_, MessageBlock[14] );
	Round( c1, d1, e1, a1, b1, f3,  6, K3_, MessageBlock[ 6] );
	Round( b1, c1, d1, e1, a1, f3, 14, K3_, MessageBlock[ 9] );
	Round( a1, b1, c1, d1, e1, f3, 12, K3_, MessageBlock[11] );
	Round( e1, a1, b1, c1, d1, f3, 13, K3_, MessageBlock[ 8] );
	Round( d1, e1, a1, b1, c1, f3,  5, K3_, MessageBlock[12] );
	Round( c1, d1, e1, a1, b1, f3, 14, K3_, MessageBlock[ 2] );
	Round( b1, c1, d1, e1, a1, f3, 13, K3_, MessageBlock[10] );
	Round( a1, b1, c1, d1, e1, f3, 13, K3_, MessageBlock[ 0] );
	Round( e1, a1, b1, c1, d1, f3,  7, K3_, MessageBlock[ 4] );
	Round( d1, e1, a1, b1, c1, f3,  5, K3_, MessageBlock[13] );
                                                       
	/* 48-63 rounds */
	Round( c1, d1, e1, a1, b1, f2, 15, K4_, MessageBlock[ 8] );
	Round( b1, c1, d1, e1, a1, f2,  5, K4_, MessageBlock[ 6] );
	Round( a1, b1, c1, d1, e1, f2,  8, K4_, MessageBlock[ 4] );
	Round( e1, a1, b1, c1, d1, f2, 11, K4_, MessageBlock[ 1] );
	Round( d1, e1, a1, b1, c1, f2, 14, K4_, MessageBlock[ 3] );
	Round( c1, d1, e1, a1, b1, f2, 14, K4_, MessageBlock[11] );
	Round( b1, c1, d1, e1, a1, f2,  6, K4_, MessageBlock[15] );
	Round( a1, b1, c1, d1, e1, f2, 14, K4_, MessageBlock[ 0] );
	Round( e1, a1, b1, c1, d1, f2,  6, K4_, MessageBlock[ 5] );
	Round( d1, e1, a1, b1, c1, f2,  9, K4_, MessageBlock[12] );
	Round( c1, d1, e1, a1, b1, f2, 12, K4_, MessageBlock[ 2] );
	Round( b1, c1, d1, e1, a1, f2,  9, K4_, MessageBlock[13] );
	Round( a1, b1, c1, d1, e1, f2, 12, K4_, MessageBlock[ 9] );
	Round( e1, a1, b1, c1, d1, f2,  5, K4_, MessageBlock[ 7] );
	Round( d1, e1, a1, b1, c1, f2, 15, K4_, MessageBlock[10] );
	Round( c1, d1, e1, a1, b1, f2,  8, K4_, MessageBlock[14] );
                                                       
	/* 64-79 rounds */
	Round( b1, c1, d1, e1, a1, f1,  8, K5_, MessageBlock[12] );
	Round( a1, b1, c1, d1, e1, f1,  5, K5_, MessageBlock[15] );
	Round( e1, a1, b1, c1, d1, f1, 12, K5_, MessageBlock[10] );
	Round( d1, e1, a1, b1, c1, f1,  9, K5_, MessageBlock[ 4] );
	Round( c1, d1, e1, a1, b1, f1, 12, K5_, MessageBlock[ 1] );
	Round( b1, c1, d1, e1, a1, f1,  5, K5_, MessageBlock[ 5] );
	Round( a1, b1, c1, d1, e1, f1, 14, K5_, MessageBlock[ 8] );
	Round( e1, a1, b1, c1, d1, f1,  6, K5_, MessageBlock[ 7] );
	Round( d1, e1, a1, b1, c1, f1,  8, K5_, MessageBlock[ 6] );
	Round( c1, d1, e1, a1, b1, f1, 13, K5_, MessageBlock[ 2] );
	Round( b1, c1, d1, e1, a1, f1,  6, K5_, MessageBlock[13] );
	Round( a1, b1, c1, d1, e1, f1,  5, K5_, MessageBlock[14] );
	Round( e1, a1, b1, c1, d1, f1, 15, K5_, MessageBlock[ 0] );
	Round( d1, e1, a1, b1, c1, f1, 13, K5_, MessageBlock[ 3] );
	Round( c1, d1, e1, a1, b1, f1, 11, K5_, MessageBlock[ 9] );
	Round( b1, c1, d1, e1, a1, f1, 11, K5_, MessageBlock[11] );

	tmp       = (Digest[1] + c + d1) & 0xffffffff;
	Digest[1] = (Digest[2] + d + e1) & 0xffffffff;
	Digest[2] = (Digest[3] + e + a1) & 0xffffffff;
	Digest[3] = (Digest[4] + a + b1) & 0xffffffff;
	Digest[4] = (Digest[0] + b + c1) & 0xffffffff;
	Digest[0] = tmp;

 }


/************************************************************************
 *
 * Two useful functions: 
 * ReverseUINT32B        - worries about Big and Little Endians
 * copyDigestToByteArray - may be, we want to copy digest from Context to 
 *                         independent BYTE array
 *************************************************************************/

static void copyDigestToByteArray(UINT *dwInp, BYTE *bOut )
{ UINT i, j;

  for(i = 0, j = 0; j < 20; i++, j += 4) /* 20 - is the Digest length in BYTEs */
   { bOut[j]   = (BYTE)( dwInp[i]        & 0xff);
     bOut[j+1] = (BYTE)((dwInp[i] >> 8)  & 0xff);
     bOut[j+2] = (BYTE)((dwInp[i] >> 16) & 0xff);
     bOut[j+3] = (BYTE)((dwInp[i] >> 24) & 0xff);
   }
}

/* The function automatically defines if we are Big or Little Endian 
 * The BYTE reversing in RIPEMD differs from used in SHA-1 and MD5 !
 */

static void ReverseUINT32B( UINT *Data, UINT Length_UINT )
{ UINT i;
  UINT dwTmp;

  if((*(UINT *)("ABCD") >> 24) == 'D') return; /* Nothing to do */

  for(i=0; i<Length_UINT; i++)
  { dwTmp = (Data[i] << 16) | (Data[i] >> 16);
    Data[i] = ((dwTmp & 0xFF00FF00L) >> 8) | ((dwTmp & 0x00FF00FFL) << 8);
  }
}



/**********************************************************************
 * 
 * RipemdInit() initializes a context of the message digest calculations.
 *
 **********************************************************************/

int RipemdInit( void *self )
{
  RipemdContext *rmd_ctx = (RipemdContext *) self;
  
  rmd_ctx->digest[0] = H0;
  rmd_ctx->digest[1] = H1;
  rmd_ctx->digest[2] = H2;
  rmd_ctx->digest[3] = H3;
  rmd_ctx->digest[4] = H4;

  /* Initialise bit count */
  rmd_ctx->countLo = rmd_ctx->countHi = 0;

  return RIPEMD_OK;
}


/**********************************************************************
 * 
 * RipemdProcess() is called multiple times (if needed) until our message 
 * buffer is not empty
 *
 **********************************************************************/

int RipemdProcess( void *self, const BYTE *data, int length )
{
  UINT       dwTmp;
  int		  dataCount, len;
  BYTE        *bPtr;
  const BYTE  *msgPtr;
  RipemdContext *rmd_ctx = (RipemdContext *)self;

  /* Update BYTE count */
  /* Check if the overflow occured in the countLo */
  dwTmp = rmd_ctx->countLo;

  if ((rmd_ctx->countLo = dwTmp + length) < dwTmp)
     (rmd_ctx->countHi)++;

  /* If the previous Message was not multiply of 64 BYTEs (RIPEMD_BLOCK_SIZE_BYTES),
   * we store remain of the previous Message in rmd_ctx->data[] array.
   * Calculate, how many BYTEs we currently store in rmd_ctx->data[]
   */
  
  /* calculate in dataCount number of BUSY BYTEs in rmd_ctx->data buffer */
  dataCount = dwTmp & 0x3F; 

  msgPtr = data;
  len = length;

  if( dataCount )
   { bPtr = (BYTE *)rmd_ctx->data + dataCount;

     /* calculate in dataCount number of FREE BYTEs in rmd_ctx->data buffer */
     dataCount = RIPEMD_BLOCK_SIZE_BYTES - dataCount;
     if( length < dataCount )
       { SMemCpy( bPtr, data, length );
         return RIPEMD_OK;
       }
     
	 /* in dataCount - number of free BYTEs in rmd_ctx->data buffer */
     SMemCpy( bPtr, data, dataCount );
     ReverseUINT32B( rmd_ctx->data, 16 );
	 ripemdUpdateDigest( rmd_ctx->digest, rmd_ctx->data );
     msgPtr += dataCount;
     len    -= dataCount;
   }

  /* Process data in 64-BYTEs (512 bites) blocks (RIPEMD_BLOCK_SIZE_BYTES) */
  while( len >= RIPEMD_BLOCK_SIZE_BYTES )
   { SMemCpy( rmd_ctx->data, msgPtr, RIPEMD_BLOCK_SIZE_BYTES );
     ReverseUINT32B( rmd_ctx->data, 16 );
     ripemdUpdateDigest( rmd_ctx->digest, rmd_ctx->data );
     msgPtr += RIPEMD_BLOCK_SIZE_BYTES;
     len -= RIPEMD_BLOCK_SIZE_BYTES;
   }

  /* Save remaining BYTEs of Message. */
  if (len > 0) SMemCpy( rmd_ctx->data, msgPtr, len );

  return RIPEMD_OK;
}



/**********************************************************************
 *
 * RipemdFinal(): All message buffer is transformed using RipemdProcess().
 * To get the final Message Digest, we have to call RipemdFinal()
 *
 **********************************************************************/

int RipemdFinal( void *self, BYTE *digest ) /* 'digest' array must be >= 20 BYTEs */
{
  UINT count;
  BYTE  *dataPtr;
  RipemdContext *rmd_ctx = (RipemdContext *)self;

  /* Calculate number of BUSY BYTEs in rmd_ctx->data buffer */
  count = rmd_ctx->countLo & 0x3F;

  /* Set the first char of padding to 0x80.  This is safe since there is
	 always at least one BYTE free */
  dataPtr = (BYTE *)rmd_ctx->data + count;
  *dataPtr++ = 0x80;

  /* Calculate number of FREE BYTEs in rmd_ctx->data buffer */
  count = 63 - count; /* rmd_ctx->data is 64 BYTEs, but we already used one BYTE for 0x80 */

  /*   If count is more than 8, it means that we have the place for 
   * 2 UINTS to place the length of message. 
   *   If no, we have to pad the current 64-BYTE block by 0, re-calculate 
   * digest, then fill the first 56 BYTEs in rmd_ctx->data by 0, place 
   * the length of Message to the last 8 BYTEs and calculate final Digest
   */
  if( count < 8 )
   { SMemSet( dataPtr, 0, count );
     ReverseUINT32B( rmd_ctx->data, 16 );
     ripemdUpdateDigest( rmd_ctx->digest, rmd_ctx->data );
	 /* prepare to calculate final digest, where the length will be written in last 8 BYTEs */
     SMemSet( rmd_ctx->data, 0, 56 ); 
   }
  else if (count > 8) SMemSet(dataPtr, 0, count - 8);

  /* The words where we have to place lenght of the message i
   * in RIPEMD differs from used in SHA-1 and MD5 !
   */

  rmd_ctx->data[15] = (((rmd_ctx->countHi) << 3) &0xffffffff) | (((rmd_ctx->countLo) >> 29) & 0x7);
  rmd_ctx->data[14] = ((rmd_ctx->countLo) << 3) & 0xffffffff;

  ReverseUINT32B( rmd_ctx->data, 14 ); // Do not reverse last 8 BYTEs
  ripemdUpdateDigest( rmd_ctx->digest, rmd_ctx->data );

  copyDigestToByteArray( rmd_ctx->digest, digest );
 
  return RIPEMD_OK;
}



/**********************************************************************
 * 
 *  RipemdUpdate() - the size of BYTE array "data" must be 
 *  equal to the block size of RIPEMD-160 algorithm, i.e. 512 bits = 64 BYTEs
 *
 **********************************************************************/

int RipemdUpdate( void *self, const BYTE *data )
{ 
  RipemdContext *rmd_ctx = (RipemdContext *)self;

   /* 16 * sizeof(UINT) == 64 BYTEs */
  SMemCpy( rmd_ctx->data, data, sizeof( rmd_ctx->data ) ); // 16 ); 
  
  ReverseUINT32B( rmd_ctx->data, 16 );
  ripemdUpdateDigest( rmd_ctx->digest, rmd_ctx->data );

  return RIPEMD_OK;
}

/**********************************************************************
 *
 * RipemdAllocate() - allocates memory needed for RipemdContext structure.
 * Uses SMemory functions
 *
 **********************************************************************/

int RipemdAllocate( void **pAlgorithm)
{
	*pAlgorithm = SMalloc( sizeof(RipemdContext) );
	if ( *pAlgorithm == NULL ) 
		return RIPEMD_MALLOC_ERROR;

	return RIPEMD_OK;
}


/**********************************************************************
 * 
 * RipemdFree() - Frees resources allocated in RipemdAllocate
 *
 **********************************************************************/

int RipemdFree( void **pAlgorithm)
{
	SFree(*pAlgorithm);
	*pAlgorithm = NULL;
	return RIPEMD_OK;
}

