#include "../Power/Windefs.h"
#include "../Power/power.h"
#include "../../RNG/cRNG.h"
#include "ElGamal.h"
#include "../../x509pkts/oidFuncs.h"
#include <assert.h>

//
//  dhpublicnumber OBJECT IDENTIFIER ::= 
//		{ iso(1) member-body(2) us(840) ansi-x942(10046) number-type(2) 1 }
//
static const char *alg_ElG_OID = "1.2.840.10046.2.1";

enum {
	ElGamalKeyType		= (UINT('E') << 24) | (UINT('l') << 16) | (UINT('G') << 8),
	ElGamalPublicKey	=  ElGamalKeyType | UINT('P'),
	ElGamalSecretKey	=  ElGamalKeyType | UINT('S')
};

struct sElgPublicKey
{
	int type;

	pw_int p;	// random prime
	pw_int g;	// random g < p;
	pw_int y;	// y = g^x mod p;

	//
	//	RFC 2459        Internet X.509 Public Key Infrastructure    January 1999
	//  7.3.2  Diffie-Hellman Key Exchange Key
	//	
	pw_int q; // INTEGER, -- factor of p-1, It is not used in this implementation.

	sElgPublicKey()
	{
		type = ElGamalPublicKey;
		q.setu( 0 ); // It is not used and inited at all by creation procedures. ( So we init it to avoid assigmedt errors )
	}

	~sElgPublicKey()
	{
		p.clear();
		g.clear();
		y.clear();

		q.clear();

		memset( this, 0, sizeof( *this ) );
	}
};

struct sElgSecretKey: public sElgPublicKey
{
	pw_int x;	// random x < p;

	sElgSecretKey()
	{
		type = ElGamalSecretKey;
	}

	~sElgSecretKey()
	{
		p.clear();
		g.clear();
		y.clear();
		x.clear();

		q.clear();

		memset( this, 0, sizeof( *this ) );
	}
};

int ElGamal_GetModulusBitSize( ElGamal_KeyHandle hKey )
{
	sElgPublicKey *pk = (sElgPublicKey *)hKey;

	if ( ( pk == NULL ) || ((pk->type != ElGamalPublicKey) && (pk->type != ElGamalSecretKey)))
		return 0;

	return pw_getSizeInBits( &pk->p );
}


//
//	Function:
//		generateK
//	Generates random k relatively prime with p-1
//
static int generateK( pw_int *k, pw_int *p )
{
	int bufferSize = ( pw_getSizeInBits( p ) + 7 ) / 8;

	BYTE *buffer = new BYTE[ bufferSize ];

	if ( buffer == NULL )
		return pw_MemoryAllocationError;

	int res = cRNG::generate( buffer, bufferSize ); 	if ( res != pw_OK ) return res;

	res = pw_bytes2int( k, buffer, bufferSize );		if ( res != pw_OK ) return res;

	pw_int pm1;

	res = pw_subu( &pm1, p, 1 );						if ( res != pw_OK )	return res;
	res = pw_mod( k, k, &pm1 );							if ( res != pw_OK )	return res;

	// k will at least twice less then p-1
	res = pw_div_exp2( k, k, 1 );						if ( res != pw_OK )	return res;
	res = pw_setBit( k, 0 );							if ( res != pw_OK )	return res;

	pw_int gcd;

	for(;;)
	{
		res = pw_gcd( &gcd, k, &pm1 );	if ( res != pw_OK )	return res;

		if ( pw_cmpu( &gcd, 1 ) == 0 )
			break;

		res = pw_addu( k, k, 2 );		if ( res != pw_OK )	return res;
	}

	return pw_OK;
}

ElGamal_KeyHandle ElGamal_GenerateKey( int keyBitSize )
{
	if ( keyBitSize == 0 )
		return NULL;

	return NULL;
}

ElGamal_KeyHandle ElGamal_DerivePublicKey( ElGamal_KeyHandle hSecKey )
{
	sElgSecretKey *sk = ( sElgSecretKey * )hSecKey;

	if ( sk->type != ElGamalSecretKey )
	{
		assert( 0 );
		return NULL;
	}

	sElgPublicKey *pk = new sElgPublicKey;

	if ( pk == NULL )
		return NULL;

	if ( ( sk->y.data == NULL ) || ( pw_cmpu( &(sk->y), 0 ) == 0 ) )
		pw_powm( &(sk->y), &(sk->g), &(sk->x), &(sk->p) ); // y = g^x mod p

	if (( pk->p.set( &sk->p ) != pw_OK ) ||
		( pk->g.set( &sk->g ) != pw_OK ) ||
		( pk->y.set( &sk->y ) != pw_OK ) ||
		( pk->q.set( &sk->q ) != pw_OK )  )
	{	
		ElGamal_FreeKey( pk );
		return NULL;
	}

	return (ElGamal_KeyHandle)pk;
}


//
//RFC 2440                 OpenPGP Message Format            November 1998
//Callas, et. al.             Standards Track                    [Page 35]
//
//       Algorithm Specific Fields for Elgamal public keys:
//
//         - MPI of Elgamal prime p;
//
//         - MPI of Elgamal group generator g;
//
//         - MPI of Elgamal public key value y (= g**x where x is
//           secret).
//

static int readPublicPart( sElgPublicKey *pk, const BYTE *buffer, int bufferSize )
{
	int read;
	int res;

	res = pw_mpi2int( &pk->p, buffer, bufferSize );
	if ( res == 0 ) 
		return 0;

	read = res;

	res = pw_mpi2int( &pk->g, buffer + read, bufferSize - read);
	if ( res == 0 ) 
		return 0;

	read += res;

	res = pw_mpi2int( &pk->y, buffer + read, bufferSize - read);
	if ( res == 0 ) 
		return 0;

	return read + res;
}

ElGamal_KeyHandle ElGamal_ReadPublicKeyFromMpiBuffer( const BYTE *buffer, int bufferSize )
{
	sElgPublicKey *pk = new sElgPublicKey;

	if ( pk == NULL )
		return NULL;

	int read = readPublicPart( pk, buffer, bufferSize );

	if ( read == 0 ) 
	{
		delete pk;
		return NULL;
	}

	return (ElGamal_KeyHandle) pk;
}

//
//
// Callas, et. al.             Standards Track                    [Page 37]
// RFC 2440                 OpenPGP Message Format            November 1998
//	
//	The Public information has to be read before.
//
//       Algorithm Specific Fields for Elgamal secret keys:
//
//       - MPI of Elgamal secret exponent x.
//
ElGamal_KeyHandle ElGamal_ReadSecretKeyFromMpiBuffer( const BYTE *mpi, int mpiSize, const BYTE *secretMpi, int secretMpiSize )
{
	sElgSecretKey *sk = new sElgSecretKey;

	if ( sk == NULL )
		return NULL;

	int res = readPublicPart( sk, mpi, mpiSize );

	if ( res == 0 ) 
	{
		delete sk;
		return NULL;
	}
	
	//
	//	Reads secret exponent..
	//
	res = pw_mpi2int( &sk->x, secretMpi, secretMpiSize );
	if ( res == 0 )
	{	delete sk;
		return NULL;
	}

	return (ElGamal_KeyHandle)sk;
}

void ElGamal_FreeKey( ElGamal_KeyHandle hKey )
{
	sElgPublicKey *pk = (sElgPublicKey *)hKey;

	if ( pk->type == ElGamalPublicKey )
		delete pk;
	else
	if ( pk->type == ElGamalSecretKey )
	{		
		sElgSecretKey *sk = (sElgSecretKey *)hKey;
		delete sk;
	}
}

//
//	Encryption functions.
//

//
//	ElGamal encryption - static function.
//	See description Bruce Schneier - Applied Cryptography SE p.478
//
static int encrypt(
	pw_int *a, // cipher text value a
	pw_int *b, // cipher text value b
	
	pw_int *m, // plain text message
	sElgPublicKey *pk )
{
	int res;

	//
	//	Generates random k < p and k is relativle prime to p-1
	//
	pw_int k;
	res = generateK( &k, &pk->p );
	if ( res != pw_OK ) return res;

	//	
	//	a = g^k mod p
	//	
	res = pw_powm( a, &pk->g, &k, &pk->p );
	if ( res != pw_OK ) return res;

	//
	//	b = ( y^k * M ) mod p == ((( y^k ) mod p ) * M ) mod p
	//
	res = pw_powm( b, &pk->y, &k, &pk->p );
	if ( res != pw_OK ) return res;

	return pw_mulm( b, b, m, &pk->p );
}

//
//	ElGamal decryption - static function.
//	See description Bruce Schneier - Applied Cryptography SE p.478
//
static int decrypt( 
	pw_int *m, // plain text message

	pw_int *a, // cipher text value a
	pw_int *b, // cipher text value b
	
	sElgSecretKey *sk )
{
	int res;

	pw_int tmp;

	//	
	//	M = b/(a^x) mod p
	//
	
	res = pw_powm( &tmp, a, &sk->x, &sk->p );
	if ( res != pw_OK ) return res;

	res = pw_invert( &tmp, &tmp, &sk->p );
	if ( res != pw_OK ) return res;

	return pw_mulm( m, &tmp, b, &sk->p );
}


//
//	Function:
//		ElGamal_Encrypt.
//
//	InBuf is the Plain text, it is to be padded as described in  PKCS-1 block type 02 [RFC2313]
//	 8.1 Encryption-block formatting.
//		See implemention PKEB_formatBlock() in Packets.h
//	OutBuf
//
int ElGamal_Encrypt( ElGamal_KeyHandle hPKey, BYTE *outMpiBuf, int *pOutMpiBufSize, const BYTE *inBuf, int inBufSize )
{
	int sizeRequired = (ElGamal_GetModulusBitSize(hPKey) + 7) / 8 + 2; // One Mpi

	sizeRequired *=2; // Two mpi a and b

	if ( *pOutMpiBufSize < sizeRequired )
	{
		*pOutMpiBufSize = sizeRequired;
		return pkae_bufferSizeTooSmall;
	}

	sElgPublicKey *pk = (sElgPublicKey *)hPKey;

	//
	//	Check key handle...
	//
	if ( ( pk == NULL ) || ( pk->type != ElGamalPublicKey ) )
		return pw_InvalidKeyBlock;

	pw_int a,b,m;

	int res = pw_bytes2int( &m, inBuf, inBufSize );
	if ( res != pw_OK )	return res;

	res = encrypt( &a, &b, &m, pk );
	if ( res != pw_OK )	return res;

	int pos = 0;

	if ( !pw_writeMpiBuffer( outMpiBuf, *pOutMpiBufSize, &pos, &a ) ||
		 !pw_writeMpiBuffer( outMpiBuf, *pOutMpiBufSize, &pos, &b ) )
			return pw_InternalError;

	*pOutMpiBufSize = pos;

	return pw_OK;
}

//
//	Seems the encrypted buffer
//
//	Assumes encryption result as
//	SEQUENCE {
//		a	INTEGER,
//		b	INTEGER
//	}
//
int ElGamal_Encrypt_X509( ElGamal_KeyHandle hPKey, BYTE *outBuf, int *pOutBufSize, const BYTE *inBuf, int inBufSize )
{
	x509_struct s, sInt;

	int maxRequiredSize; // Unfortunately we cannot encrypt the data first 
						 // and calculate the result value, because we 
						 // did not get same result if the data will be encrypted again,
						 // because the generateK function uses random data

	maxRequiredSize = 
		s.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, 
			2*sInt.prepareForWritting( asn1tag_INTEGER,
				( ( ElGamal_GetModulusBitSize(hPKey) + 7 ) / 8 )
			)
		);

	if (*pOutBufSize < maxRequiredSize )
	{
		*pOutBufSize = maxRequiredSize;
		return x509e_bufferTooSmall;
	}

	sElgPublicKey *pk = (sElgPublicKey *)hPKey;

	//
	//	Check key handle...
	//
	if ( ( pk == NULL ) || ( pk->type != ElGamalPublicKey ) )
		return pw_InvalidKeyBlock;

	pw_int a,b,m;

	int res = pw_bytes2int( &m, inBuf, inBufSize );
	if ( res != pw_OK )	return res;

	res = encrypt( &a, &b, &m, pk );
	if ( res != pw_OK )	return res;

	int sizeRequired = 
		s.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			pw_getASN1Size( &a ) + 
			pw_getASN1Size( &b )
		);

	if (*pOutBufSize < sizeRequired )
	{
		return x509e_internalError;
	}

	if ( !s.writeHeader( outBuf, sizeRequired ) )
		return x509e_internalError;

	int pos = 0;

	if ( !pw_writeASN1Buffer( s.B, s.L, &pos, &a ) ||
		 !pw_writeASN1Buffer( s.B, s.L, &pos, &b ) )
			return pw_InternalError;

	if ( pos != (int)s.L )
		return x509e_internalError;

	*pOutBufSize = s.TL;

	return x509e_ok;
}

int ElGamal_Decrypt( ElGamal_KeyHandle hSKey, BYTE *outBuf, int *pOutBufSize, const BYTE *inMpiBuf, int inMpiBufSize )
{
	int sizeRequired = ( ElGamal_GetModulusBitSize(hSKey) + 7 ) / 8;

	if ( *pOutBufSize < sizeRequired )
	{
		*pOutBufSize = sizeRequired;
		return pkae_bufferSizeTooSmall;
	}

	sElgSecretKey *sk = (sElgSecretKey *)hSKey;

	//
	//	Check key handle...
	//
	if ( ( sk == NULL ) || ( sk->type != ElGamalSecretKey ) )
		return pw_InvalidKeyBlock;

	pw_int a,b,m;

	int pos = 0;

	if ( !pw_readMpiBuffer( &a, inMpiBuf, inMpiBufSize, &pos ) ||
		 !pw_readMpiBuffer( &b, inMpiBuf, inMpiBufSize, &pos ) )
		return pw_InternalError;

	int res = decrypt( &m, &a, &b, sk );

	if ( res != pw_OK )
		return res;

	res = pw_int2bytes( outBuf, *pOutBufSize, &m );

	return res;
}

int ElGamal_Decrypt_X509( ElGamal_KeyHandle hSKey,
		BYTE *outBuf, int *pOutBufSize,
		const BYTE *inBuf, int inBufSize // ASN.1 packed buffer
		)
{
	int sizeRequired = ( ElGamal_GetModulusBitSize(hSKey) + 7 ) / 8;

	if ( *pOutBufSize < sizeRequired )
	{
		*pOutBufSize = sizeRequired;
		return pkae_bufferSizeTooSmall;
	}

	sElgSecretKey *sk = (sElgSecretKey *)hSKey;

	//
	//	Check key handle...
	//
	if ( ( sk == NULL ) || ( sk->type != ElGamalSecretKey ) )
		return pw_InvalidKeyBlock;

	x509_struct s;

	int res = s.parse( inBuf, inBufSize, NULL, NULL, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	pw_int a,b,m;

	int pos = 0;

	if ( !pw_readASN1Buffer( &a, s.B, s.L, &pos ) ||
		 !pw_readASN1Buffer( &b, s.B, s.L, &pos ) ||
		 ( pos != (int)s.L ) )
		return x509e_invalidPkt;

	res = decrypt( &m, &a, &b, sk );

	if ( res != pw_OK )
		return res;

	res = pw_int2bytes( outBuf, *pOutBufSize, &m );

	return res;
}

//
//	Signature functions.
//

//
//	ElGamal signature - static function.
//	See description Bruce Schneier - Applied Cryptography SE pp.476-477
//
static int sign( sElgSecretKey *sk, pw_int *a, pw_int *b, pw_int *m )
{
	int res;

	//
	//	Generates random k < p and k is relativle prime to p-1
	//
	pw_int k;
	res = generateK( &k, &sk->p );
	if ( res != pw_OK ) return res;

	//
	//	a = g^k mod p
	//	
	res = pw_powm( a, &sk->g, &k, &sk->p );
	if ( res != pw_OK ) return res;

	//
	//	Find such b, M = ( xa + kb ) mod ( p - 1 )
	//		==> b = (( M - xa ) * 1/k ) mod ( p - 1 )
	//

	pw_int pm1, tmp;

	res = pw_subu( &pm1, &sk->p, 1 );
	if ( res != pw_OK ) return res;

	res = pw_mulm( &tmp, a, &sk->x, &pm1 );
	if ( res != pw_OK ) return res;

	res = pw_add( b, m, &pm1 ); // To be sure that  M - ax will positive...
	if ( res != pw_OK ) return res;

	res = pw_subm( b, b, &tmp, &pm1 );
	if ( res != pw_OK ) return res;

	res = pw_invert( &tmp, &k, &pm1 );
	if ( res != pw_OK ) return res;

	return pw_mulm( b, b, &tmp, &pm1 );
}

//
//	ElGamal verification - static function.
//	See description Bruce Schneier - Applied Cryptography SE p.477
//	return pw_OK if (y^a)*(a^b) mod p == g^M mod p
//	otherwise return error code.
//
static int verify( sElgPublicKey *pk, pw_int *m, pw_int *a, pw_int *b )
{
	//
	//	Checks if (y^a)*(a^b) mod p == g^M mod p
	//

	int res;
	pw_int t1,t2;

	res = pw_powm( &t1, &pk->y, a, &pk->p );
	if ( res != pw_OK ) return res;

	res = pw_powm( &t2, a, b, &pk->p );
	if ( res != pw_OK ) return res;

	res = pw_mulm( &t1, &t1, &t2, &pk->p );
	if ( res != pw_OK ) return res;

	res = pw_powm( &t2, &pk->g, m, &pk->p );
	if ( res != pw_OK ) return res;

	if ( pw_cmp( &t1, &t2 ) != 0 )
		return pw_InvalidKeyBlock;

	return pw_OK;
}


int ElGamal_Sign( ElGamal_KeyHandle hSKey , BYTE *sigMpiBuf, int *pSigMpiBufSize, const BYTE *digestBuf, int digestBufSize )
{
	int sizeRequired = (ElGamal_GetModulusBitSize(hSKey) + 7) / 8 + 2; // One Mpi

	sizeRequired *=2; // Two mpis a and b

	if ( *pSigMpiBufSize < sizeRequired )
	{
		*pSigMpiBufSize = sizeRequired;
		return pkae_bufferSizeTooSmall;
	}

	sElgSecretKey *sk = (sElgSecretKey *)hSKey;

	//
	//	Check key handle...
	//
	if ( ( sk == NULL ) || ( sk->type != ElGamalSecretKey ) )
		return pw_InvalidKeyBlock;

	pw_int a,b,m;

	int res = pw_bytes2int( &m, digestBuf, digestBufSize );
	if ( res != pw_OK )	return res;

	res = sign( sk, &a, &b, &m );
	if ( res != pw_OK )	return res;

	int pos = 0;

	if ( !pw_writeMpiBuffer( sigMpiBuf, *pSigMpiBufSize, &pos, &a ) ||
		 !pw_writeMpiBuffer( sigMpiBuf, *pSigMpiBufSize, &pos, &b ) )
			return pw_InternalError;

	*pSigMpiBufSize = pos;

	return pw_OK;
}

int ElGamal_Sign_X509(
		ElGamal_KeyHandle hSKey,
		BYTE *sigBuf, int *pSigBufSize, // ASN.1 buffer
		const BYTE *digestBuf, int digestBufSize )
{
	x509_struct s, sInt;

	int maxRequiredSize; // Unfortunately we cannot encrypt the data first 
						 // and calculate the result value, because we 
						 // did not get same result if the data will be encrypted again,
						 // because the generateK function uses random data

	maxRequiredSize = 
		s.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, 
			2*sInt.prepareForWritting( asn1tag_INTEGER,
				( ( ElGamal_GetModulusBitSize(hSKey) + 7 ) / 8 )
			)
		);

	if (*pSigBufSize < maxRequiredSize )
	{
		*pSigBufSize = maxRequiredSize;
		return x509e_bufferTooSmall;
	}

	sElgSecretKey *sk = (sElgSecretKey *)hSKey;

	//
	//	Check key handle...
	//
	if ( ( sk == NULL ) || ( sk->type != ElGamalSecretKey ) )
		return pw_InvalidKeyBlock;

	pw_int a,b,m;

	int res = pw_bytes2int( &m, digestBuf, digestBufSize );
	if ( res != pw_OK )	return res;

	res = sign( sk, &a, &b, &m );
	if ( res != pw_OK )	return res;

	//
	//	Tries again with real size
	//	The size has to be less then maxRequiredSize
	//
	int sizeRequired = 
		s.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			pw_getASN1Size( &a ) + 
			pw_getASN1Size( &b )
		);

	if ( *pSigBufSize < sizeRequired )
	{
		return x509e_internalError;
	}

	if ( !s.writeHeader( sigBuf, sizeRequired ) )
		return x509e_internalError;

	int pos = 0;

	if ( !pw_writeASN1Buffer( s.B, s.L, &pos, &a ) ||
		 !pw_writeASN1Buffer( s.B, s.L, &pos, &b ) )
			return pw_InternalError;

	if ( pos != (int)s.L )
		return x509e_internalError;

	*pSigBufSize = s.TL;

	return x509e_ok;
}

int ElGamal_Verify( ElGamal_KeyHandle hPKey, const BYTE *sigMpiBuf, int sigMpiBufSize, const BYTE *digestBuf, int digestBufSize )
{
	sElgPublicKey *pk = (sElgPublicKey *)hPKey;

	//
	//	Check key handle...
	//
	if ( ( pk == NULL ) || ( pk->type != ElGamalPublicKey ) )
		return pw_InvalidKeyBlock;

	pw_int a,b,m;

	int pos = 0;

	if ( !pw_readMpiBuffer( &a, sigMpiBuf, sigMpiBufSize, &pos ) ||
		 !pw_readMpiBuffer( &b, sigMpiBuf, sigMpiBufSize, &pos ) )
			return pw_InternalError;

	int res = pw_bytes2int( &m, digestBuf, digestBufSize );
	if ( res != pw_OK )	return res;

	return verify( pk, &m, &a, &b );
}

int ElGamal_Verify_X509(
		ElGamal_KeyHandle hPKey, 
		const BYTE *sigBuf, int sigBufSize, // ASN.1 buffer
		const BYTE *digestBuf, int digestBufSize )
{
	sElgPublicKey *pk = (sElgPublicKey *)hPKey;

	//
	//	Check key handle...
	//
	if ( ( pk == NULL ) || ( pk->type != ElGamalPublicKey ) )
		return pw_InvalidKeyBlock;

	x509_struct s; // Sequence

	int res = s.parse( sigBuf, sigBufSize, NULL, NULL, asn1tag_SEQUENCE );
	if( res != x509e_ok ) return res;

	pw_int a,b,m;

	int pos = 0;

	if ( !pw_readASN1Buffer( &a, s.B, s.L, &pos ) ||
		 !pw_readASN1Buffer( &b, s.B, s.L, &pos ) ||
		 ( (int)s.L != pos )
		 )
			return x509e_internalError;

	res = pw_bytes2int( &m, digestBuf, digestBufSize );
	if ( res != pw_OK )	return res;

	return verify( pk, &m, &a, &b );
}

/*

Housley, et. al.            Standards Track                    [Page 61]

RFC 2459        Internet X.509 Public Key Infrastructure    January 1999

 7.3.2  Diffie-Hellman Key Exchange Key

   The Diffie-Hellman OID supported by this profile is defined by ANSI
   X9.42 [X9.42].

        dhpublicnumber OBJECT IDENTIFIER ::= { iso(1) member-body(2)
                  us(840) ansi-x942(10046) number-type(2) 1 }


   The dhpublicnumber OID is intended to be used in the algorithm field
   of a value of type AlgorithmIdentifier. The parameters field of that
   type, which has the algorithm-specific syntax ANY DEFINED BY
   algorithm, have the ASN.1 type DomainParameters for this algorithm.

        DomainParameters ::= SEQUENCE {
              p       INTEGER, -- odd prime, p=jq +1
              g       INTEGER, -- generator, g
              q       INTEGER, -- factor of p-1
              j       INTEGER OPTIONAL, -- subgroup factor
              validationParms  ValidationParms OPTIONAL }

        ValidationParms ::= SEQUENCE {
              seed             BIT STRING,
              pgenCounter      INTEGER }

      DHPublicKey ::= INTEGER -- public key, y = g^x mod p
      DHPrivateKey ::= INTEGER -- private key, x
*/


//
//	ElGamal_ReadPublicKeyFromASN1Buffer
//	Uses RFC 2459 description.
//
int ElGamal_ReadPublicKeyFromASN1Buffer( ElGamal_KeyHandle *hk, x509_publicKeyInfo *pki )
{
	int res = x509e_ok;

	switch ( pki->algOID.params.T & asn1tag_TYPE )
	{
		case asn1tag_SEQUENCE: break;
		default:
			return x509e_invalidPktTag;
	}

	sElgPublicKey *key = new sElgPublicKey;

	if ( key == NULL )
		return x509e_memoryAllocation;


	int parsed = 0;

	if (!pw_readASN1Buffer( &(key->p), pki->algOID.params.B, pki->algOID.params.L, &parsed ) ||
		!pw_readASN1Buffer( &(key->g), pki->algOID.params.B, pki->algOID.params.L, &parsed )
		)
			res = x509e_invalidPkt;
	else
	{
		if ( ( parsed != (int)pki->algOID.params.L ) &&
			(
			//	pkcs #3 means the third parameter as length of private value, it is OPTIONAL
			//  x9.42 notation means the third parameter as q where p = j*q + 1,  ( one factor of p - 1 )
			//	the implementation does not use the value, so it can be ignored.
				!pw_readASN1Buffer( &(key->q), pki->algOID.params.B, pki->algOID.params.L, &parsed ) 
			//  x9.42  also has two optionals
			//
			//		j       INTEGER OPTIONAL, -- subgroup factor ( second factor of p-1 ) 
			//		validationParms  ValidationParms OPTIONAL }
			//
				)
			)
				res = x509e_invalidPkt;

		parsed = 0;

		if ( !pw_readASN1Buffer( &(key->y), pki->publicKey.B, pki->publicKey.L, &parsed ) ||
		   ( parsed != (int)pki->publicKey.L )
			)
				res = x509e_invalidPkt;
	}

	if ( res == x509e_ok )
		*hk = (ElGamal_KeyHandle) key;

	return res;
}

int ElGamal_ReadSecretKeyFromASN1Buffer( ElGamal_KeyHandle *hk, x509_privateKeyInfo *pki )
{
	int res = x509e_ok;

	switch ( pki->algOID.params.T & asn1tag_TYPE )
	{
		case asn1tag_SEQUENCE: break;
		default:
			return x509e_invalidPktTag;
	}

	sElgSecretKey *key = new sElgSecretKey;

	if ( key == NULL )
		return x509e_memoryAllocation;


	int parsed = 0;

	if (!pw_readASN1Buffer( &(key->p), pki->algOID.params.B, pki->algOID.params.L, &parsed ) ||
		!pw_readASN1Buffer( &(key->g), pki->algOID.params.B, pki->algOID.params.L, &parsed )
		)
			res = x509e_invalidPkt;
	else
	{
		if ( ( parsed != (int)pki->algOID.params.L ) &&
			(
			//	pkcs #3 means the third parameter as length of private value, it is OPTIONAL
			//  x9.42 notation means the third parameter as q where p = j*q + 1,  ( one factor of p - 1 )
			//	the implementation does not use the value, so it can be ignored.
				!pw_readASN1Buffer( &(key->q), pki->algOID.params.B, pki->algOID.params.L, &parsed ) 
			//  x9.42  also has two optionals
			//
			//		j       INTEGER OPTIONAL, -- subgroup factor ( second factor of p-1 ) 
			//		validationParms  ValidationParms OPTIONAL }
			//
				)
			)
				res = x509e_invalidPkt;

		parsed = 0;

		if ( !pw_readASN1Buffer( &(key->x), pki->privateKey.B, pki->privateKey.L, &parsed ) ||
		   ( parsed != (int)pki->privateKey.L )
			)
				res = x509e_invalidPkt;
	}
	
	pw_powm( &(key->y), &(key->g), &(key->x), &(key->p) ); // y = g^x mod p

	if ( res == x509e_ok )
		*hk = (ElGamal_KeyHandle) key;

	return res;
}

int ElGamal_WritePublicKeyToASN1Buffer(	ElGamal_KeyHandle hKey, BYTE *buff, int *pBuffSize )
{
	sElgPublicKey *key = (sElgPublicKey *)hKey;

	if ( ( key == NULL ) || ((key->type != ElGamalPublicKey) && (key->type != ElGamalSecretKey)))
		return x509e_internalError;

	x509_publicKeyInfo pki;

	//
	//	Forms public key info structure and calculates required size
	//
	//	DomainParameters ::= SEQUENCE {
	//	  p       INTEGER, -- odd prime, p=jq +1
	//	  g       INTEGER, -- generator, g
	//	  q       INTEGER, -- factor of p-1
	//	  j       INTEGER OPTIONAL, -- subgroup factor
	//	  validationParms  ValidationParms OPTIONAL }
	//
	// ValidationParms ::= SEQUENCE {
	//	  seed             BIT STRING,
	//	  pgenCounter      INTEGER }
	//
	int required = 
		pki.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			pki.algOID.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				pki.algOID.oid.prepareForWritting( alg_ElG_OID )
				+
				pki.algOID.params.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
					pw_getASN1Size( &(key->p) )
					+
					pw_getASN1Size( &(key->g) )
					+
					pw_getASN1Size( &(key->q) )
				)
			)
			+
			pki.publicKey.prepareForWritting( pw_getASN1Size( &(key->y) )
			)
		);

	if ( required > *pBuffSize )
	{
		*pBuffSize = required;
		return x509e_bufferTooSmall;
	}

	*pBuffSize = required;

	//
	//	Fills buffers.
	//
	if ( pki.writeHeaders( buff, required ) == 0 )
		return x509e_internalError;

	//
	//	Fills data.
	//
	int oidLen = pki.algOID.oid.L;
	if ( string2oid( pki.algOID.oid.B, &oidLen, alg_ElG_OID ) != (int)pki.algOID.oid.L )
		return x509e_internalError;

	int pos = 0;
	if ( !pw_writeASN1Buffer( pki.algOID.params.B, pki.algOID.params.L, &pos , &(key->p) ) ||
		 !pw_writeASN1Buffer( pki.algOID.params.B, pki.algOID.params.L, &pos , &(key->g) ) ||
		 !pw_writeASN1Buffer( pki.algOID.params.B, pki.algOID.params.L, &pos , &(key->q) ) ||
		 ( pos != (int)pki.algOID.params.L )
		 )
			return x509e_internalError;

	pos = 0;

	if ( !pw_writeASN1Buffer( pki.publicKey.B, pki.publicKey.L, &pos , &(key->y) )  )
			return x509e_internalError;

	return x509e_ok;
}

int ElGamal_WriteSecretKeyToASN1Buffer(	ElGamal_KeyHandle hKey, BYTE *buff, int *pBuffSize )
{
	sElgSecretKey *key = (sElgSecretKey *)hKey;

	if ( ( key == NULL ) || ( key->type != ElGamalSecretKey ) )
		return x509e_internalError;

	x509_privateKeyInfo pki;

	//
	//	Forms pprivate key info structure and calculates required size
	//
	//	DomainParameters ::= SEQUENCE {
	//	  p       INTEGER, -- odd prime, p=jq +1
	//	  g       INTEGER, -- generator, g
	//	  q       INTEGER, -- factor of p-1
	//	  j       INTEGER OPTIONAL, -- subgroup factor
	//	  validationParms  ValidationParms OPTIONAL }
	//
	// ValidationParms ::= SEQUENCE {
	//	  seed             BIT STRING,
	//	  pgenCounter      INTEGER }
	//
	int required = 
		pki.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			pki.algOID.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				pki.algOID.oid.prepareForWritting( alg_ElG_OID )
				+
				pki.algOID.params.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
					pw_getASN1Size( &(key->p) )
					+
					pw_getASN1Size( &(key->g) )
					+
					pw_getASN1Size( &(key->q) )
				)
			)
			+
			pki.privateKey.prepareForWritting( asn1tag_OCTET_STRING, pw_getASN1Size( &(key->x) )
			)
		);

	if ( required > *pBuffSize )
	{
		*pBuffSize = required;
		return x509e_bufferTooSmall;
	}

	*pBuffSize = required;

	//
	//	Fills buffers.
	//
	if ( pki.writeHeaders( buff, required ) == 0 )
		return x509e_internalError;

	//
	//	Fills data.
	//
	int oidLen = pki.algOID.oid.L;
	if ( string2oid( pki.algOID.oid.B, &oidLen, alg_ElG_OID ) != (int)pki.algOID.oid.L )
		return x509e_internalError;

	int pos = 0;
	if ( !pw_writeASN1Buffer( pki.algOID.params.B, pki.algOID.params.L, &pos , &(key->p) ) ||
		 !pw_writeASN1Buffer( pki.algOID.params.B, pki.algOID.params.L, &pos , &(key->g) ) ||
		 !pw_writeASN1Buffer( pki.algOID.params.B, pki.algOID.params.L, &pos , &(key->q) ) ||
		 ( pos != (int)pki.algOID.params.L )
		 )
			return x509e_internalError;

	pos = 0;

	if ( !pw_writeASN1Buffer( pki.privateKey.B, pki.privateKey.L, &pos , &(key->x) )  )
			return x509e_internalError;

	return x509e_ok;
}


/*
pkcs-3.txt


9. Object identifier

This standard defines two object identifiers: pkcs-3 and
dhKeyAgreement.

The object identifier pkcs-3 identifies this standard.

pkcs-3 OBJECT IDENTIFIER ::=
  { iso(1) member-body(2) US(840) rsadsi(113549)
      pkcs(1) 3 }

The object identifier dhKeyAgreement identifies the Diffie-
Hellman key agreement method defined in Sections 7 and 8.

dhKeyAgreement OBJECT IDENTIFIER ::= { pkcs-3 1 }

The dhKeyAgreement object identifier is intended to be used
in the algorithm field of a value of type
AlgorithmIdentifier. The parameters field of that type,
which has the algorithm-specific syntax ANY DEFINED BY
algorithm, would have ASN.1 type DHParameter for this
algorithm.

DHParameter ::= SEQUENCE {
  prime INTEGER, -- p
  base INTEGER, -- g
  privateValueLength INTEGER OPTIONAL }

The fields of type DHParameter have the following meanings:

     o    prime is the prime p.

     o    base is the base g.

     o    privateValueLength is the optional private-value
          length l.
*/

