//
//	Montgomery method is used for power function.
//
//	The idea is:
//		Choose R -> 2^k > modulo m.
//		Compute the RInv = R^-1 mod m.
//		Convert every number a,b,c to Montgomery form a', b', c', like:
//			a' = a * R mod m.
//			b' = b * R mod m.
//			c' = c * R mod m.
//		If c == a * b, so the result 
//			c' = a * b * R mod m
//			c' = a * b * R * ( R * RInv ) mod m
//			c' = a'* b'* RInv mod m
//		Now a'* b' = X is more then R, so we multiply the numbers 
//		and write the result as:
//			X = x1 * R + x0, ==> every number before mutiplying to RInv can be reduced
//			X * RInv mod m = ( x1 * R + x0 ) * RInv mod m = 
//				= x1 * R * RInv + x0 * RInv = x1 + x0 * RInv mod m
//

#include "power.h"
#include <assert.h>

//
//	Montgomery context
//
class mgm_ctx
{
public:
	pw_int *m;		// Modulo
	int rDim;		// Dimention of choosen R in limbs

	bool isNegative; // Euclid algorithm gives two results
					 //  1. R * rE - m * mE = 1, is false
					 //  2. m * mE - R * rE = 1. is true

	pw_int rE;	//	Euclid algorithm value see isNegative
	pw_int mE;	//  Euclid algorithm value see isNegative

	pw_int rInv; // Real multiplicative inverse of R in modulo m  
				 // rInv = isNegative ? m - rE : rE.


	mgm_ctx()
	{
		m = NULL;
		rDim = 0;
	}

	int create( pw_int *pMod )
	{
		int res;

		m = pMod;

		rDim = pw_getLastLimb( m ) + 1;

		pw_int R;

		for(;;)
		{
			res = R.setu( 1 );
			if ( res != pw_OK ) return res;

			res = pw_mul_exp2( &R, &R, rDim * U64_BIT_Q );
			if ( res != pw_OK ) return res;

			res = pw_invertEx( &rE, &mE, &R, m, &isNegative );

			if ( res == pw_InversionError )
			{
				rDim++;		// We can choose R more if initial value is not mutial prime with module m
				continue;
			}
			else
			if ( res == pw_OK )
				break;
			else
				return res;
		}

		if ( isNegative )
			res = pw_sub( &rInv, m, &rE );
		else
			res = rInv.setVirtual( &rE, 0 );

#ifdef _DEBUG

		pw_int t;

		res = pw_mulm( &t, &R, &rInv, m );
		if ( res != pw_OK ) return res;

		if ( pw_cmpu( &t, 1 ) != 0 )
			return pw_InternalError;

		pw_int mInv;

		if ( !isNegative )
			res = pw_sub( &mInv, &R, &mE );
		else
			res = mInv.setVirtual( &mE, 0 );

		res = pw_mulm( &t, m, &mInv, &R );
		if ( res != pw_OK ) return res;

		if ( pw_cmpu( &t, 1 ) != 0 )
			return pw_InternalError;

#endif // _DEBUG

		return pw_OK;
	}
};


//
//	Function:
//	 mgm_restore
//		Converts from montgomery representation to real value.
//			r = x * rInv mod m.
//
static int mgm_restore( pw_int *r, pw_int *x, mgm_ctx *ctx )
{
	return pw_mulm( r, x, &ctx->rInv, ctx->m );
}

//
//	Function:
//	 mgm_convert
//		Prepares montgomery representation
//			r = x * R mod m.
//
static int mgm_convert( pw_int *r, pw_int *x, mgm_ctx *ctx )
{
	int res = pw_mul_exp2( r, x, ctx->rDim * U64_BIT_Q );
	if ( res != pw_OK ) return res;

	res = pw_mod( r, r, ctx->m );
	if ( res != pw_OK ) return res;

#ifdef _DEBUG

	pw_int t;
	res = mgm_restore( &t, r, ctx );
	if ( res != pw_OK ) return res;

	if ( pw_cmp( &t, x ) != 0 )
		return pw_InternalError;

#endif // _DEBUG

	return res;
}


static void clearAlderLimbs( pw_int *p, int limbNum )
{
#ifdef _UseAsm_

	UINT *ptr = (UINT *)&(p->data[ limbNum ]);

	int dwSize = ( p->m_ullSize - limbNum ) * 2;

	while( dwSize --  > 0 )
		*(ptr++) = 0;

#else

	int last = pw_getLastLimb( p );

	for( ; last >= limbNum; last-- )
		p->data[ last ] = 0;

#endif // _UseAsm_
}

//
//	mgm_pro_neg( ma, mb ), negative case -> m * mInv - R * rInv == 1;
//	1. t = ma * mb,
//	2. p = t * mE mod R,
//  3. u = ( p * m - t ) / R,
//	 u = u mod m, but u < 2 * m, so
//	4. u = u > m ? u - m : u
//	But u = t * rE mod m, rE = rInv * (-1) mod m,
//	So u = (-1) * ma * mb * rInv mod m, u < m
//	5. r = (-1) * u mod m;
//
static int mgm_pro_neg( pw_int *mr, pw_int *ma, pw_int *mb, mgm_ctx *ctx )
{
	int res;
	pw_int t,p;

#ifdef _DEBUG

	pw_int c;

	//
	//	mc = ma * mb * rInv mod m;
	//
	res = pw_mulm( &c, ma, mb, ctx->m );
	if ( res != pw_OK ) return res;

	res = pw_mulm( &c, &c, &ctx->rInv, ctx->m );
	if ( res != pw_OK ) return res;

#endif // _DEBUG

	//
	//	1. t = ma * mb,
	//
	res = pw_mul( &t, ma, mb );
	if ( res != pw_OK ) return res;

	//
	//	2. p = t * mE mod R,
	//
	res = p.set( &t );
	if ( res != pw_OK ) return res;
	clearAlderLimbs( &p, ctx->rDim );
	res = pw_mul( &p, &p, &ctx->mE );
	if ( res != pw_OK ) return res;
	clearAlderLimbs( &p, ctx->rDim );

	//
	//  3. t = ( p * m - t ) / R,
	//
	res = pw_mul( &p, &p, ctx->m );
	if ( res != pw_OK ) return res;

	//
	//	Check if p > t to make valid sub
	//
	bool sig = pw_cmp( &t, &p ) > 0;

	if ( !sig )
		res = pw_sub( &t, &p, &t );
	else
		res = pw_sub( &t, &t, &p );

	if ( res != pw_OK ) return res;

	pw_int u;

	res = u.setVirtual( &t, ctx->rDim ); // u = t / R
	if ( res != pw_OK ) return res;

	//
	//	4. u = u > m ? u - m : u
	//
	if ( pw_cmp( &u, ctx->m ) >= 0 )
		res = pw_sub( mr, &u, ctx->m );
	else
		res = pw_set( mr, &u );

	if ( res != pw_OK ) return res;

	//
	//	5. r = (-1) * u mod m;
	//	But if sig r = -1 * -1 * u = u
	//	So we'll change the result only if !sig.
	//
	if ( !sig && ( pw_cmpu( mr, 0 ) != 0 ) )
	{
		res = pw_sub( mr, ctx->m, mr );
		if ( res != pw_OK ) return res;
	}

#ifdef _DEBUG

	if ( pw_cmp( &c, mr ) != 0 )
		return pw_InternalError;

#endif // _DEBUG

	return res;
}

//
//	mgm_pro( ma, mb ), positive case e.t. R * rInv - m * mInv == 1;
//	1. t = ma * mb,
//	2. p = t * mE mod R,
//  3. u = ( t + p * m ) / R,
//	r = u mod m, but u < 2 * m, so
//	4. r = u > m ? u - m : u
//
static int mgm_pro( pw_int *mr, pw_int *ma, pw_int *mb, mgm_ctx *ctx )
{
	if ( ctx->isNegative )
		return mgm_pro_neg( mr, ma, mb, ctx );

	int res;
	pw_int t,p;

#ifdef _DEBUG

	pw_int c;

	//
	//	c = ma * mb * rInv mod m;
	//
	res = pw_mulm( &c, ma, mb, ctx->m );
	if ( res != pw_OK ) return res;

	res = pw_mulm( &c, &c, &ctx->rInv, ctx->m );
	if ( res != pw_OK ) return res;

#endif // _DEBUG

	//
	//	1. t = ma * mb,
	//
	res = pw_mul( &t, ma, mb );
	if ( res != pw_OK ) return res;

	//
	//	2. p = t * mE mod R
	//
	res = p.set( &t );
	if ( res != pw_OK ) return res;
	clearAlderLimbs( &p, ctx->rDim );
	res = pw_mul( &p, &p, &ctx->mE );
	if ( res != pw_OK ) return res;
	clearAlderLimbs( &p, ctx->rDim );

	//
	//  3. u = ( t + p * m ) / R,
	//
	res = pw_mul( &p, &p, ctx->m );
	if ( res != pw_OK ) return res;
	res = pw_add( &t, &t, &p );
	if ( res != pw_OK ) return res;

	pw_int u;

	res = u.setVirtual( &t, ctx->rDim ); // u = t / R
	if ( res != pw_OK ) return res;

	//
	//	4. r = u > m ? u - m : u
	//
	if ( pw_cmp( &u, ctx->m ) >= 0 )
		res = pw_sub( mr, &u, ctx->m );
	else
		res = pw_set( mr, &u );

#ifdef _DEBUG

	if ( pw_cmp( &c, mr ) != 0 )
		return pw_InternalError;

#endif // _DEBUG

	return res;
}

int pw_powm_montgomery( pw_int *r, pw_int *x, pw_int *exp, pw_int *mod )
{
	int res;

	mgm_ctx ctx;

	res = ctx.create( mod );
	if ( res != pw_OK ) return res;

	pw_int tmp;
	pw_int mx;

	int bitExpNum = pw_getLastBit( exp );

	if ( bitExpNum == 0 ) // Zero exponent;
	{
		return r->setu( 1 );
	}

	bitExpNum--; // One based result; if only first bit -> bitExpNum = 1

	res = mgm_convert( &mx, x, &ctx );
	if ( res != pw_OK ) return res;

	res = tmp.set( &mx );
	if ( res != pw_OK ) return res;

	for( ; bitExpNum--; )
	{
		res = mgm_pro( &tmp, &tmp, &tmp, &ctx );
		if ( res != pw_OK ) return res;

		BOOL currentBit = pw_getBit( exp, bitExpNum );
		if ( currentBit )
		{
			res = mgm_pro( &tmp, &tmp, &mx, &ctx  );
			if ( res != pw_OK ) return res;
		}
	}

	return mgm_restore( r, &tmp, &ctx ); 
}

int pw_powm_montgomery_windows( pw_int *r, pw_int *x, pw_int *exp, pw_int *mod )
{
	int res;

	mgm_ctx ctx;

	res = ctx.create( mod );
	if ( res != pw_OK ) return res;

	pw_int tmp;
	pw_int mx;

	int bitExpNum = pw_getLastBit( exp );

	if ( bitExpNum == 0 ) // Zero exponent;
		return r->setu( 1 );

	res = mgm_convert( &mx, x, &ctx );
	if ( res != pw_OK ) return res;

	const int windowSize = 6; //6 is optimal for 2048 bit, 7 is optimal for 4096 bits

	//
	//	ptab is  table with x powers
	//	t[0] = 1 
	//	t[1] = x
	//	t[2] = x^2
	//	t[i] = x^i
	//	t[31] = x^31  1111
	//
	pw_int ptab[ 1 << windowSize ];
/*
	res = tmp.setu( 1 );
	if ( res != pw_OK )	return res;

	res = mgm_convert( &ptab[i], &tmp, &ctx );
	if ( res != pw_OK ) return res;
*/
	res = ptab[1].setVirtual( &mx, 0 );
	if ( res != pw_OK )	return res;

	for ( unsigned int i = 2; i < sizeof(ptab) / sizeof( pw_int ); i++ )
	{
		res = mgm_pro( &ptab[i], &ptab[i-1], &mx, &ctx );
		if ( res != pw_OK )	return res;
	}

	res = tmp.set( &mx );
	if ( res != pw_OK ) return res;

	bitExpNum--; // One based result; if only first bit -> bitExpNum = 1

	for( ; bitExpNum > 0 ; )
	{
		int q = 0;
		int c = 0;
		for( ; (q < windowSize) && bitExpNum; q++)
		{
			bitExpNum--;

			res = mgm_pro( &tmp, &tmp, &tmp, &ctx );
			if ( res != pw_OK ) return res;

			BOOL currentBit = pw_getBit( exp, bitExpNum );
			c = ( c<<1 ) + currentBit;
		}

		if ( c != 0 )
		{
			res = mgm_pro( &tmp, &tmp, &ptab[c], &ctx );
			if ( res != pw_OK ) return res;
		}
	}

	return mgm_restore( r, &tmp, &ctx ); 
}


///////////////////////////

/*
	The descryption in PGP sources.

 * Montgomery reduce n, modulo mod.  This reduces modulo mod and divides by
 * 2^(32*mlen).  Returns the result in the *top* mlen words of the argument n.
 * This is ready for another multiplication using bniMul_32.
 *
 * Montgomery representation is a very useful way to encode numbers when
 * you're doing lots of modular reduction.  What you do is pick a multiplier
 * R which is relatively prime to the modulus and very easy to divide by.
 * Since the modulus is odd, R is closen as a power of 2, so the division
 * is a shift.  In fact, it's a shift of an integral number of words,
 * so the shift can be implicit - just drop the low-order words.
 *
 * Now, choose R *larger* than the modulus m, 2^(32*mlen).  Then convert
 * all numbers a, b, etc. to Montgomery form MMul(a), MMul(b), etc using the
 * relationship MMul(a) = a*R mod m, MMul(b) = b*R mod m, etc.  Note that:
 * - The Montgomery form of a number depends on the modulus m.
 *   A fixed modulus m is assumed throughout this discussion.
 * - Since R is relaitvely prime to m, multiplication by R is invertible;
 *   no information about the numbers is lost, they're just scrambled.
 * - Adding (and subtracting) numbers in this form works just as usual.
 *   MMul(a+b) = (a+b)*R mod m = (a*R + b*R) mod m = (MMul(a) + MMul(b)) mod m
 * - Multiplying numbers in this form produces a*b*R*R.  The problem
 *   is to divide out the excess factor of R, modulo m as well as to
 *   reduce to the given length mlen.  It turns out that this can be
 *   done *faster* than a normal divide, which is where the speedup
 *   in Montgomery division comes from.
 *
 * Normal reduction chooses a most-significant quotient digit q and then
 * subtracts q*m from the number to be reduced.  Choosing q is tricky
 * and involved (just look at bniDiv_32 to see!) and is usually
 * imperfect, requiring a check for correction after the subtraction.
 *
 * Montgomery reduction *adds* a multiple of m to the *low-order* part
 * of the number to be reduced.  This multiple is chosen to make the
 * low-order part of the number come out to zero.  This can be done
 * with no trickery or error using a precomputed inverse of the modulus.
 * In this code, the "part" is one word, but any width can be used.
 *
 * Repeating this step sufficiently often results in a value which
 * is a multiple of R (a power of two, remember) but is still (since
 * the additions were to the low-order part and thus did not increase
 * the value of the number being reduced very much) still not much
 * larger than m*R.  Then implicitly divide by R and subtract off
 * m until the result is in the correct range.
 *
 * Since the low-order part being cancelled is less than R, the
 * multiple of m added must have a multiplier which is at most R-1.
 * Assuming that the input is at most m*R-1, the final number is
 * at most m*(2*R-1)-1 = 2*m*R - m - 1, so subtracting m once from
 * the high-order part, equivalent to subtracting m*R from the
 * while number, produces a result which is at most m*R - m - 1,
 * which divided by R is at most m-1.
 *
 * To convert *to* Montgomery form, you need a regular remainder
 * routine, although you can just compute R*R (mod m) and do the
 * conversion using Montgomery multiplication.  To convert *from*
 * Montgomery form, just Montgomery reduce the number to
 * remove the extra factor of R.
 *
 * TODO: Change to a full inverse and use Karatsuba's multiplication
 * rather than this word-at-a-time.
 */
