#include "../Power/Windefs.h"
#include "../Power/power.h"
#include "../../SMemory/smemory.h"
#include "Rsa.h"
#include <assert.h>
#include "../../x509pkts/oidFuncs.h"
#include "../../PGPPkts/pktdefs.h"

//
// PKCS#1 recommend public exponent 0x03 or F4 ( 65537 0x10001 )
//
#define DefaultPublicExponent 0x10001

//
// Generate a safe prime p where (p - 1) / 2 it also prime...
//
//static bool useSafePrime = true;

int pw_init_set( pw_int **x, pw_int *y )
{
	*x = new pw_int;

	if ( *x == NULL )
		return pw_MemoryAllocationError;

	return (*x)->set( y );
}

int pw_init( pw_int **x )
{
	*x = new pw_int;

	if ( *x == NULL )
		return pw_MemoryAllocationError;

	return (*x)->setu( 0 );
}

int pw_clear( pw_int *x )
{
	delete x;
	return pw_OK;
}

#ifdef _DEBUG
#include <stdio.h>
#endif


static int RSAAllocatePublicKey( sRSAPublicKey **key )
{
	*key = (sRSAPublicKey *) SMalloc( sizeof( sRSAPublicKey ) );

	if (*key == NULL)
		return RSA_MALLOC_ERROR;

	pw_init( &(*key)->n );	/* public modulus */
	pw_init( &(*key)->e );	/* public exponent */

	if ( SMemoryError )
			return RSA_MALLOC_ERROR;

	(*key)->keyBlock = NULL;

	return RSA_OK;
}

int RSADerivePublicKey(
	sRSAPrivateKey *privKey,
	sRSAPublicKey **publicKey )
{
	int res;

	sRSAPublicKey *key;

	*publicKey = NULL;

	res = RSAAllocatePublicKey( &key );

	if ( res != RSA_OK ) return res;

	key->sizeInBits = privKey->sizeInBits;

	pw_set( key->n , privKey->n);
	pw_set( key->e , privKey->e);

	*publicKey = key;

	return RSA_OK;
}

/* Perform safe deallocation of allocated resources */
int RSAFreePublicKey(
	sRSAPublicKey **key
)
{
	pw_clear( (*key)->n );
	pw_clear( (*key)->e );

	if ( (*key)->keyBlock )
		SFree( (*key)->keyBlock );

	SFree( *key );
	*key = 0;

	return RSA_OK;
}

/*
 *	Procedure: makePrime
 *		while p is not prime p = (p+1) mod 2 exp (initial p length);
 */
static int makePrime( pw_int * p, bool safePrime = false, pw_int *publicExp = NULL, cProgressActor *actor = NULL )
{
//	assert( false );

	pw_int *r, *t;
	size_t bitQ;

	pw_init_set( &r , p );

	//
	//	Point 1.
	//	r is even number, the fact is used in Point 2, see below.
	//
	pw_clearBit ( r , 0 );

	pw_init( &t );

	bitQ = pw_getSizeInBits( r );

	//
	//	Only if prime table was inited here, it will be freed below ( see call of pw_primeTableFree )...
	//
	bool primeTableWasInited = ( pw_primeTableInit( bitQ, actor ) == pw_OK );

	int res = 0;

	for(;; pw_addu( r , r  , 2 ), pw_clearBit ( r , bitQ ) )
	{
		if ( safePrime )
			pw_setBit ( r , 1 );

		//
		//	If public exponent is assigned,
		//	checks if it is relatively prime with p-1, see PKSC #1
		//
		if ( publicExp != NULL )
		{
			//
			//	Point 2.
			//	r now is p - 1, see point 1 above and point 3 below
			//

			res = pw_gcd( t, r, publicExp );
			if ( res != pw_OK )
				break;

			if ( pw_cmpu( t, 1 ) != 0 )
				continue;
		}

		pw_set ( p , r );

		//
		//	Point 3.
		//	p will odd and equal to r + 1. Because r is even number, see point 1 above.
		//	That is required to check the prime relation to public exponent. See point 2.
		//
		pw_setBit ( p , 0 );

		int probCount = 0;
		int is_prime = 0;

		if ( safePrime )
		{	// Tests it once( by table only ), p will be tested 25 times when it will checked that it is safe prime
			// The trick significantly save time because it is not required to perform strong test
			// if p-1/2 was not checked even by dividers from prime table.
			is_prime = pw_probab_prime( p, 0, &probCount );

			if ( is_prime ) // p is prime by table...
			{
				//
				// Safe prime test...
				// Checks if (t - 1) / 2 is also prime...
				//
				res = pw_subu( t, p, 1 );
				if ( res != pw_OK )
					break;

				res = pw_div_exp2( t, t, 1 );
				if ( res != pw_OK )
					break;

				is_prime = pw_probab_prime( t, 1, &probCount );

				if ( is_prime )
				{
					is_prime = pw_probab_prime( p, 25, &probCount ); // see comments above

					if ( is_prime )
						is_prime = pw_probab_prime( t, 24, &probCount ); // One test was performed above..
				}
			}
		}
		else
			is_prime = pw_probab_prime( p, 25, &probCount );

		if ( actor != NULL )
		{
			int q = actor->func( NULL, res, probCount );
			if ( q != 0 )
			{
				res = q;
				break;
			}
		}

		if ( is_prime )
			break;
	}

	pw_clear( r );
	pw_clear( t );

	if ( primeTableWasInited )
		pw_primeTableFree();

	return res;
}

static int RSAAllocatePrivateKey(
	sRSAPrivateKey **privateKey     /* private key to be generated */
)
{
	sRSAPrivateKey *key;

	*privateKey = (sRSAPrivateKey *)SMalloc( sizeof( sRSAPrivateKey ));
	if ( *privateKey == NULL)
		return RSA_MALLOC_ERROR;

	key = *privateKey;

	pw_init( &key->n );	/* public modulus */
	pw_init( &key->e );	/* public exponent */
	pw_init( &key->p );	/* prime number, larger factor of n */
	pw_init( &key->q );	/* prime number, smaller factor of n */
	pw_init( &key->d );	/* private exponent */
	pw_init( &key->c );	/* multiplicative inverse of p mod q */
	pw_init( &key->e1 );	/* exp1 = d mod (p-1) */
	pw_init( &key->e2 );	/* exp2 = d mod (q-1) */

	key->keyBlock = NULL;

	return RSA_OK;
}
/*
static bool CheckPrivateKeyOld( sRSAPrivateKey *key )
{
	pw_int *t, *enc, *dec;

	int i;
	int length = ( key->sizeInBits + 7 ) / 8;

	pw_init( &enc );
	pw_init( &dec );
	pw_init( &t );

	pw_random( t, length - 1 );

	pw_powm( enc, t , key->e , key->n );

	pw_powm( dec , enc , key->d , key->n );

	i = pw_cmp( t , dec );

	if ( i != 0 )
	{
//		printf( "Error in CheckPrivatekey...\n" );
		assert( 0 );
	}

	pw_clear( enc );
	pw_clear( dec );
	pw_clear( t );

	return i==0;
}
*/

bool CheckPrivateKey( sRSAPrivateKey *key )
{
	byte *t, *enc, *dec;

	unsigned int i;
	unsigned int length = ( key->sizeInBits + 7 ) / 8;

	t	= new byte[ length ];
	enc = new byte[ length ];
	dec = new byte[ length ];

	if ( !t || !enc || !dec )
		return false;

	for( i = 1 ; i < length; i++ )
		t[i] = (byte)(i + 'A');

	t[0] = 0; // It has to be less then module.

	// pw_powm( enc, t , key->e , key->n );

	sRSAPublicKey *publicKey;

	RSADerivePublicKey( key, &publicKey );

	RSAPublicUpdate( publicKey, t, enc );

	RSAFreePublicKey( &publicKey );


	// pw_powm( dec , enc , key->d , key->n );

	RSAPrivateUpdate( key, enc, dec );

	// i = pw_cmp( t , dec );

	i = (byte)memcmp( t, dec, length );

	if ( i != 0 )
	{
//		printf( "Error in CheckPrivatekey...\n" );
		assert( 0 );
	}

	delete t;
	delete enc;
	delete dec;

	return i==0;
}

//////////////////////////////////////////
// ExtendPrivateKey...
// Note:
//  Private key has initialized only e,p,q.
//  Other value will generated by the function.

static int ExtendPrivateKey(	sRSAPrivateKey *key  )
{
/*	Create Private exponent d == e^-1 mod (p-1)(q-1) */
	pw_subu( key->c , key->q ,1 );	// c = q-1
	pw_subu( key->n , key->p ,1 );	// n = p-1
	pw_mul( key->c , key->c , key->n ); /*  c = q-1 and n = p-1 => c=(q-1)(p-1)*/


	if ( pw_invert( key->d, key->e, key->c ) != pw_OK )
		return RSA_INIT_ERROR;


/*	Compute c = q^(-1) mod p if q < p */
	/* Used to speed up private cipher operation by Chinese remainder theorem */

	if ( pw_invert( key->c, key->q, key->p ) != pw_OK )
		return RSA_INIT_ERROR;


/*	Compute n = p*q */
	pw_mul( key->n , key->p , key->q );
	key->sizeInBits = (uint16b)pw_getSizeInBits( key->n );

/*	Prepare e1 = d mod (p - 1) */
	pw_subu( key->e1 , key->p , 1 );		/* e1 == p - 1 */
	pw_mod( key->e1 , key->d , key->e1 );	/* e1 == d mod e1 */

/*	Prepare e2 = d mod (q - 1) */
	pw_subu( key->e2 , key->q , 1 );		/* e2 == q - 1 */
	pw_mod( key->e2 , key->d , key->e2 );	/* e2 == d mod e2 */

	return RSA_OK;
}


#ifndef __UseRandomBuffer__

static int generateRandomPrime(
	pw_int * p,
	sSHA1Random *rnd,
	size_t bufferSize,
	BYTE *seedBuffer,
	size_t seedBufferSize,
	bool safePrime = false )
{
	byte *buffer;
	if ( bufferSize == 0 )
		return RSA_INIT_ERROR;

	buffer = (byte *)SMalloc( bufferSize );
	if ( buffer == NULL )
		return RSA_MALLOC_ERROR;

	SHA1RandomGenerateBytesEx( rnd, buffer, bufferSize, seedBuffer, seedBufferSize );
	buffer[0] |= 0x80; /* set alder bit */
	pw_bytes2int( p, buffer, bufferSize ); //bytes2mpzint( p, buffer , bufferSize );

	SFree( buffer );

	makePrime( p, safePrime );
 
	return RSA_OK;
}

/*
 *	RSAGeneratePrivateKey:
 *	Note:
 *	 Random generator required,
 *	 Public exponent can be specified as 0x03 or 0xF4
 */

int RSAGeneratePrivateKey(
	sRSAPrivateKey **privateKey,
	size_t bits,
	BYTE *seedBuffer,
	size_t seedBufferSize,
	size_t publicExp, // 0 == Undefined
	bool useSafePrime
	)
{
	int res;
	sRSAPrivateKey *key;
	int cmp;
	size_t primeLengthInBytes;

	*privateKey = NULL;

/*	1) Random generator check */

	sSHA1Random *rng;

	res = SHA1RandomAllocate( &rng );

	if ( res != RNG_OK )
		return res;

	SHA1RandomInit( rng );

	if ( res != RNG_OK )
		return res;

//	SHA1RandomUpdate( rng, seedBuffer, seedBufferSize );


/* 3) Allocate private key */
	res = RSAAllocatePrivateKey( &key );

	if ( res != RSA_OK )
	{
		SHA1RandomFinal( &rng );
		return res;
	}

	*privateKey = key;

/* 4) Key generation */

	if ( publicExp != 0 ) 
	{	key->e->setu( publicExp );
	/*  Check public exponent  0x03 != e != 0xF4 */
		if (( pw_cmpu( key->e , 0x03 ) != 0 )
			&& ( pw_cmpu( key->e , 0xF4 ) != 0 ))
		{
			SHA1RandomFinal( &rng );

			return RSA_INIT_ERROR;
		}
	}
	else
	{	key->e->setu( DefaultPublicExponent );	/* PKCS#1 recommend public exponent 0x3 or 0xF4 */
		publicExp = (size_t)key->e->data[0];
	}

	key->sizeInBits = ((bits + 7) & ~7);/* PKCS#1 recommend to use octet length*/

	pw_int gcd;

/*	Generate random primes p and q */
	primeLengthInBytes = (key->sizeInBits + 7) / 8 ;
	do
	{
		generateRandomPrime( key->p , rng , primeLengthInBytes / 2, seedBuffer, seedBufferSize, useSafePrime );

		res = pw_subu( key->n , key->p ,1 ); /* here n == p-1 , it must be relatively prime to e PKSC #1*/

		if ( res != pw_OK )
		{
			SHA1RandomFinal( &rng );
			return res;
		}

		res = pw_gcd( &gcd , key->n , key->e );

		if ( res != pw_OK )
		{
			SHA1RandomFinal( &rng );
			return res;
		}

	}
	while ( pw_cmpu( &gcd, 1) != 0 );

	do
	{	generateRandomPrime( key->q , rng , primeLengthInBytes - (primeLengthInBytes / 2), seedBuffer, seedBufferSize, useSafePrime );

		cmp = pw_cmp( key->q , key->p ); /* distinct odd primes p and q */

		if ( cmp == 0 )
			continue;

		res = pw_subu( key->c , key->q ,1 ); /* here c == q-1 , it must be relatively prime to e PKSC #1*/

		if ( res != pw_OK )
		{
			SHA1RandomFinal( &rng );
			return res;
		}

		res = pw_gcd( &gcd , key->c , key->e );

		if ( res != pw_OK )
		{
			SHA1RandomFinal( &rng );
			return res;
		}

	} while ( pw_cmpu( &gcd, 1) != 0 ) ;

	if ( cmp > 0 ) 
	{	/* swap p and q to be q < p */
		/* Used to speed up private cipher operation by Chinese remainder theorem */
		pw_set( key->e1 , key->p );
		pw_set( key->p , key->q );
		pw_set( key->q , key->e1 );
	}

	res = ExtendPrivateKey( key );

	if ( res == pw_OK )
		if ( !CheckPrivateKey( key , rng ) )
			res = RSA_INVALID_KEY_BLOCK;

	SHA1RandomFinal( &rng );

	return res;
}

#else // __UseRandomBuffer__

//
//	RSAPreparePrimesFromRandomBuffer:
//	Helper function.
//	It finds nearest primes for random in the buffer.
//	So we can very quikly generate a real private key from the primes.
//
int RSAPreparePrimesFromRandomBuffer(
	BYTE *buffer,
	size_t bufferSize,
	size_t publicExp,
	bool useSafePrime
	)
{
	pw_int p,q,e;

	if ( publicExp == 0 )
		publicExp = DefaultPublicExponent;

	int res = e.setu( publicExp );

	if ( res != pw_OK )
		return res;

	//
	//	Only if prime table was inited here, it will be freed below ( see call of pw_primeTableFree )...
	//
	bool primeTableWasInited = ( pw_primeTableInit( bufferSize * 8 ) == pw_OK );

	//
	//	Generate first prime
	//

	//
	//	set alder bit.
	//
	buffer[ 0 ] |= 0x80;
	pw_bytes2int( &p, buffer, bufferSize / 2 );

	makePrime( &p, useSafePrime, &e );

	pw_int2bytes( buffer, bufferSize / 2, &p );

	//
	//	Generate second prime
	//

	//
	//	set alder bit.
	//
	buffer[ bufferSize / 2 ] |= 0x80;
	pw_bytes2int( &q, buffer + bufferSize / 2, bufferSize - bufferSize / 2 );

	for (;;)
	{
		makePrime( &q, useSafePrime, &e );

		//
		//	The distinct odd primes p and q are required...
		//
		int cmp = pw_cmp( &q , &p ); 

		if ( cmp != 0 )
			break;

		//
		//	Increment current q state upto 2 and continue prime generation...
		//
		int bitQ = pw_getSizeInBits( &q );
		pw_addu( &q , &q  , 2 );
		pw_clearBit ( &q , bitQ );
	}

	pw_int2bytes( buffer + bufferSize / 2, bufferSize - bufferSize / 2, &q );

	if ( primeTableWasInited )
		pw_primeTableFree();

	return pw_OK;
}

/*
 *	RSAGeneratePrivateKeyFromRandomBuffer:
 *	Note:
 *	 Public exponent can be specified as 0x03 or F4 (65537 0x10001 )
 */

int RSAGeneratePrivateKeyFromRandomBuffer(
	sRSAPrivateKey **privateKey,
	BYTE *buffer,
	size_t bufferSize,
	size_t publicExp, // 0 == Undefined
	cProgressActor *actor,
	bool useSafePrime
	)
{
	int res;
	sRSAPrivateKey *key;
	int cmp;

	*privateKey = NULL;


/* 1) Allocate private key */
	res = RSAAllocatePrivateKey( &key );

	if ( res != RSA_OK )
		return res;

	*privateKey = key;

/* 2) Key generation */

	if ( publicExp != 0 ) 
	{	key->e->setu( publicExp );
	/*  Check public exponent  0x03 != e != 0x10001 */
//		if (( pw_cmpu( key->e , 0x03 ) != 0 )
//			&& ( pw_cmpu( key->e , 0x10001 ) != 0 ))
//		{
//				return RSA_INIT_ERROR;
//		}
	}
	else
	{	key->e->setu( DefaultPublicExponent );		/* PKCS#1 recommend public exponent 0x3 or F4( 65537 0x10001 ) */
		publicExp = (size_t)key->e->data[0];
	}

	key->sizeInBits = (uint16b)(bufferSize * 8); /* PKCS#1 recommend to use octet length*/

//	pw_int gcd;

/*	Generate random primes p and q */

	buffer[ 0 ] |= 0x80; /* set alder bit */
	pw_bytes2int( key->p, buffer, bufferSize / 2 );

	if ( actor != NULL )
	{
		int q = actor->func( "Generates prime table\n", 0, bufferSize * 8 );
		if ( q != 0 )
			return q;
	}
	bool primeTableWasInited = ( pw_primeTableInit( bufferSize * 8 ) == pw_OK );

//	do
//	{
		if ( actor != NULL ) 
		{
			int q = actor->func( "Finds prime P\n", 0, 0 );
			if ( q != 0 )
				return q;
		}

		makePrime( key->p, useSafePrime, key->e, actor );
/*
		res = pw_subu( key->n , key->p ,1 ); // here n == p-1 , it must be relatively prime to e PKSC #1

		if ( res != pw_OK )
			return res;

		res = pw_gcd( &gcd , key->n , key->e );

		if ( res != pw_OK )
			return res;
	}
	while ( pw_cmpu( &gcd, 1) != 0 );
*/
	buffer[ bufferSize / 2 ] |= 0x80; /* set alder bit */
	pw_bytes2int( key->q, buffer + bufferSize / 2, bufferSize - bufferSize / 2 );

	for (;;)
	{
		if ( actor != NULL )
		{
			int q = actor->func( "Finds prime Q\n", 0, 0 );
			if ( q != 0 )
				return q;
		}

		makePrime( key->q, useSafePrime, key->e, actor );

		cmp = pw_cmp( key->q , key->p ); /* distinct odd primes p and q */

		if ( cmp != 0 )
			break;

		//
		//	Increment current q state upto 2 and continue prime generation...
		//
		int bitQ = pw_getSizeInBits( key->q );
		pw_addu( key->q , key->q  , 2 );
		pw_clearBit ( key->q , bitQ );

	}
/*
		if ( cmp == 0 )
			continue;

		res = pw_subu( key->c , key->q ,1 ); // here c == q-1 , it must be relatively prime to e PKSC #1

		if ( res != pw_OK )
			return res;

		res = pw_gcd( &gcd , key->c , key->e );

		if ( res != pw_OK )
			return res;

	} while ( pw_cmpu( &gcd, 1) != 0 ) ;
*/
	if ( cmp > 0 ) 
	{
		// swap p and q to be q < p 
		// Used to speed up private cipher operation by Chinese remainder theorem
		pw_set( key->e1 , key->p );
		pw_set( key->p , key->q );
		pw_set( key->q , key->e1 );
	}

	res = ExtendPrivateKey( key );

	if ( primeTableWasInited )
		pw_primeTableFree();

	if ( res == pw_OK )
		if ( !CheckPrivateKey( key ) )
			return RSA_INVALID_KEY_BLOCK;

	return res;
}

#endif // __UseRandomBuffer__

/* Perform safe deallocation of allocated resources */
int RSAFreePrivateKey(
	sRSAPrivateKey **privateKey
)
{
	sRSAPrivateKey *key;
	key = *privateKey;
	pw_clear( key->n );
	pw_clear( key->e );
	pw_clear( key->p );
	pw_clear( key->q );
	pw_clear( key->d );
	pw_clear( key->c );
	pw_clear( key->e1 );
	pw_clear( key->e2 );

	if ( key->keyBlock )
		SFree( key->keyBlock );

	SFree( *privateKey );
	*privateKey = 0;

	return RSA_OK;
}


/* Private key operation with using Chinese remainder theorem */
int
	RSAPrivateUpdate(
		sRSAPrivateKey *key,
		const byte *inputBlock,	/* input data */
		byte *outputBlock,	/* output data */
		int flags
		)
{
	pw_int *cP;				/* cP = (c mod p)^dQ mod p, where dQ = d mod (p-1) */
	pw_int *cQ;				/* cQ = (c mod q)^dQ mod q, where dQ = d mod (q-1) */

	pw_int *data;

	size_t blockSize;

	blockSize = ( key->sizeInBits + 7 ) / 8;

	pw_init( &data );
	pw_init( &cP );
	pw_init( &cQ );

	if ( flags & useMpiIn )
	{
		int res = pw_mpi2int( data, inputBlock, blockSize + 2 );
		if ( res == 0 )
			return RSA_INIT_ERROR;
	}
	else
		pw_bytes2int( data, inputBlock , blockSize );

/* Compute cP = (c mod p)^dQ mod p, where dQ = d mod (p-1), for Chinese remainder theorem */
	pw_mod( cP , data , key->p );
	pw_powm( cP , cP , key->e1, key->p );

/* Compute cQ = (c mod q)^dQ mod q, where dQ = d mod (q-1), for Chinese remainder theorem */
	pw_mod( cQ , data , key->q );
	pw_powm( cQ , cQ , key->e2, key->q );


/* data = ((((cP - cQ) mod p) * qInv) mod p) * q) + cQ ) , TRUE if q < p */


	//
	//	Calculate ((cP - cQ) mod p)
	//
	if ( pw_cmp( cP, cQ ) < 0 )
	{
		//
		//  cP < cQ
		//
		// ((cP - cQ) mod p) == ( cP + p - cQ ) mod p
		//
		//	cQ < q , see cQ generation, p > q ==> cQ < p; ==> cP + P > cQ

		pw_add( cP, cP, key->p );
	}

	pw_sub( data , cP , cQ );
	pw_mod( data , data , key->p );


	pw_mul( data , data , key->c ); /* c == qInv */
	pw_mod( data , data , key->p);
	pw_mul( data , data , key->q );
	pw_add( data , data , cQ );

	if ( flags & useMpiOut )
	{
		int res = pw_int2mpi( outputBlock, blockSize + 2, data );
		if ( res == 0 )
			return RSA_INIT_ERROR;
	}
	else
		pw_int2bytes( outputBlock , blockSize , data );

	pw_clear( data );
	pw_clear( cQ );
	pw_clear( cP );

	if ( SMemoryError )
		return RSA_MALLOC_ERROR;
	return RSA_OK;
}
/*

'Damn' hacker atack sample:

Encrypted signature:
00137B58  04 81 80 00 00 00 00 00  .__.....
00137B60  00 00 00 00 00 00 00 00  ........
00137B68  00 00 00 00 00 00 00 00  ........
00137B70  00 00 00 00 00 00 00 00  ........
00137B78  00 00 00 00 00 00 00 00  ........
00137B80  00 00 00 00 00 00 00 00  ........
00137B88  00 00 00 00 00 00 00 00  ........
00137B90  00 00 00 00 00 00 00 00  ........
00137B98  00 00 00 00 00 00 00 00  ........
00137BA0  00 00 00 00 00 00 00 00  ........
00137BA8  00 00 00 00 00 00 00 00  ........
00137BB0  01 39 17 FD FC F7 FB F7  .9.?????
00137BB8  FC AD BD DA 12 DD 07 82  ?-_?.?.'
00137BC0  66 24 29 8F 7A D5 E0 C0  f$)_z???
00137BC8  94 7C 9A 9E BA 30 3D 5F  "|__?0=_
00137BD0  18 BA 79 03 06 60 AD 3D  .?y..`-=
00137BD8  B8 CE F3

Public updated ( exp = 3 ) plain text result 0x80 bytes:
Valid 01 based block ( but with random padding :)
The small value allows analising of small ( f.e. 3 ) exponential result 
( it is less then module ) and intruder uses binary convergence
to find required values.

00137BF8  00 01 D4 52 0A 01 D5 9F  ..?R..?_
00137C00  E5 C7 57 D0 EA 55 49 C7  ??W??UI?
00137C08  63 8E A7 97 CE 2A 20 09  c_�-?*
00137C10  F2 32 5E AA 28 A6 EA 5B  ?2^?(�?[
00137C18  65 B7 6F 31 38 BC EA E1  e�o18_??
00137C20  C9 78 D1 DC 31 56 D8 0B  ?x??1V?.
00137C28  3F 6A 9D 23 41 DB 39 19  ?j_#A?9.
00137C30  F2 05 92 6E 0B 38 7F A5  ?.'n.8._
00137C38  35 CE 02 B4 C3 4C 51 C3  5?._?LQ?
00137C40  CF 4B 35 1F 87 AC 2A 76  ?K5.++*v
00137C48  86 96 23 F9 47 24 C5 D2  +-#?G$??
00137C50  BE 8B FA 92 B2 B2 69 29  _<?'__i)
00137C58  3B FE C0 00 00 00 00 00  ;??.....
00137C60  00 00 00 00 E7 FA 46 D3  ....??F?
00137C68  AF 0F 51 3C A5 EF 68 6D  ?.Q<_?hm
00137C70  33 8A 2A 20 3C 3A EC 6B  3_* <:?k
*/

/*
 *	RSAPublicUpdate;
 *		The function can be used as for public key ecryption 
 *		so for public decryption operation.
 */
int
	RSAPublicUpdate(
		sRSAPublicKey *key,
		const byte *inputBlock,		/* input data */
		byte *outputBlock,		/* output data */
		int flags
		)
{
	size_t blockSize;
	pw_int *data;

	blockSize = ( key->sizeInBits + 7 ) / 8;

	pw_init( &data );

	if ( flags & useMpiIn )
	{
		int res = pw_mpi2int( data, inputBlock, blockSize + 2 );
		if ( res == 0 )
			return RSA_INIT_ERROR;
	}
	else
		pw_bytes2int( data, inputBlock , blockSize );

	size_t dataSize = pw_getSizeInBytes( data );

	pw_powm( data , data , key->e , key->n );

	if ( flags & useMpiOut )
	{
		int res = pw_int2mpi( outputBlock, blockSize + 2, data );
		if ( res == 0 )
			return RSA_INIT_ERROR;
	}
	else
		pw_int2bytes( outputBlock , blockSize , data );

	pw_clear( data );

	//
	//	Damn hacker atack.
	//	He used small values to be public updated and generates valid block. 
	//	See sample and description above.
	//
	bool smallExp = pw_cmpu( key->e, 0x1001 ) < 0;

	if ( smallExp  && 
		( dataSize < (( blockSize / 3 ) * 2 ) )
		)
		return RSA_SIGNATURE_ATACK_SUSSPECTION;

	if ( SMemoryError )
		return RSA_MALLOC_ERROR;
	return RSA_OK;
}


////////////
//
//  Public/Private Key operations CreateIndirect from and Save to byte buffer ( key block )
//

//  Key Block Type...
enum { RSA_PublicKey = 0, RSA_PrivateKey };

////////////////////////////
// RSACreatePrivateKeyIndirect
// Create private key from byte block
// Byte block structure is:
//  size_t blockType
//	size_t keyLengthInBits
//  size_t public exponet
//  byte[ primeLengthInBytes / 2  ] q - always less then p see Chinize reminder theorem
//  byte[ primeLengthInBytes - primeLengthInBytes / 2 ] p - always more then q.
//  size_t blockType

int RSACreatePrivateKeyIndirect( 
		sRSAPrivateKey **pKey,
		const byte *keyBlock)
{
	int res;
	sRSAPrivateKey *key;
	size_t primeLengthInBytes;
	size_t byteNum = 0;

	res = RSAAllocatePrivateKey( &key );
	if( res != RSA_OK )
		return res;

// Check block type...
	if ( *(size_t *)(keyBlock + byteNum) != RSA_PrivateKey )
		return RSA_INVALID_BLOCK_TYPE;

	byteNum = sizeof( size_t );

// Get size...
	key->sizeInBits = (uint16b)*(size_t *)(keyBlock + byteNum);
	byteNum += sizeof( size_t );

// Get public exponet...
	key->e->setu( *(size_t *)(keyBlock + byteNum) );
	byteNum += sizeof( size_t );

// Get prime q...
	primeLengthInBytes = (key->sizeInBits + 7) / 8 ;
	pw_bytes2int( key->q, keyBlock + byteNum , primeLengthInBytes / 2  );
	byteNum += primeLengthInBytes /2 ;

// Get prime p...

	pw_bytes2int( key->p, keyBlock + byteNum , primeLengthInBytes - primeLengthInBytes / 2);
	byteNum += primeLengthInBytes - primeLengthInBytes /2 ;

// Check end block type...
	if ( *(size_t *)(keyBlock + byteNum) != RSA_PrivateKey )
		return RSA_INVALID_KEY_BLOCK;

	ExtendPrivateKey( key );

	*pKey = key;

	return RSA_OK;
}

int RSASavePrivateKey( 
		sRSAPrivateKey *key,
		byte **pKeyBlock,
		size_t *pKeyBlockSize)
{
	byte *keyBlock;
	size_t primeLengthInBytes;
	size_t byteNum = 0;

	*pKeyBlockSize = (((key->sizeInBits + 7 ) / 8) + 4 * sizeof( size_t ));

	keyBlock = (byte *)SMalloc( *pKeyBlockSize );
	if ( keyBlock == NULL )
		return RSA_MALLOC_ERROR;

	if ( key->keyBlock )
		SFree( key->keyBlock );

	key->keyBlock = keyBlock;


// Write block type...
	*(size_t *)(keyBlock + byteNum) = RSA_PrivateKey;
	byteNum += sizeof( size_t );

// Write size...
	*(size_t *)(keyBlock + byteNum) = key->sizeInBits;
	byteNum += sizeof( size_t );

// Write public exponet...
	*(size_t *)(keyBlock + byteNum) = (size_t)key->e->data[0];
	byteNum += sizeof( size_t );

// Write prime q...
	primeLengthInBytes = (key->sizeInBits + 7) / 8 ;
	pw_int2bytes( keyBlock + byteNum , primeLengthInBytes / 2, key->q );
	byteNum += primeLengthInBytes / 2 ;

// Write prime p...
	pw_int2bytes( keyBlock + byteNum , primeLengthInBytes - primeLengthInBytes / 2, key->p );
	byteNum += primeLengthInBytes - primeLengthInBytes /2 ;

// Write block type...
	*(size_t *)(keyBlock + byteNum) = RSA_PrivateKey;
	byteNum += sizeof( size_t );

	if ( *pKeyBlockSize != byteNum )
		return RSA_INIT_ERROR;

	*pKeyBlock = keyBlock;

	return RSA_OK;
}


int RSACreatePublicKeyIndirect( 
		sRSAPublicKey **pKey,
		const byte *keyBlock )
{
	int res;
	sRSAPublicKey *key;
	size_t primeLengthInBytes;
	size_t byteNum = 0;

	res = RSAAllocatePublicKey( &key );
	if( res != RSA_OK )
		return res;

// Check block type...
	if ( *(size_t *)(keyBlock + byteNum) != RSA_PublicKey )
		return RSA_INVALID_BLOCK_TYPE;

	byteNum = sizeof( size_t );

// Get size...
	key->sizeInBits = (uint16b)*(size_t *)(keyBlock + byteNum);
	byteNum += sizeof( size_t );

// Get public exponet...
	key->e->setu( *(size_t *)(keyBlock + byteNum) );
	byteNum += sizeof( size_t );

// Get module n...
	primeLengthInBytes = (key->sizeInBits + 7) / 8 ;
	pw_bytes2int( key->n, keyBlock + byteNum , primeLengthInBytes  );
	byteNum += primeLengthInBytes;

// Check end block type...
	if ( *(size_t *)(keyBlock + byteNum) != RSA_PublicKey )
		return RSA_INVALID_KEY_BLOCK;

	*pKey = key;

	return RSA_OK;
}

int RSASavePublicKey( 
		sRSAPublicKey *key,
		byte **pKeyBlock,
		size_t *pKeyBlockSize)
{
	byte *keyBlock;
	size_t primeLengthInBytes;
	size_t byteNum = 0;

	*pKeyBlockSize = (((key->sizeInBits + 7 ) / 8) + 4 * sizeof( size_t ));

	keyBlock = (byte *)SMalloc( *pKeyBlockSize );
	if ( keyBlock == NULL )
		return RSA_MALLOC_ERROR;

	if ( key->keyBlock )
		SFree( key->keyBlock );

	key->keyBlock = keyBlock;

// Write block type...
	*(size_t *)(keyBlock + byteNum) = RSA_PublicKey;
	byteNum += sizeof( size_t );

// Write size...
	*(size_t *)(keyBlock + byteNum) = key->sizeInBits;
	byteNum += sizeof( size_t );

// Write public exponet...
	*(size_t *)(keyBlock + byteNum) = (size_t)key->e->data[0];
	byteNum += sizeof( size_t );

// Write module n...
	primeLengthInBytes = (key->sizeInBits + 7) / 8 ;
	pw_int2bytes( keyBlock + byteNum , primeLengthInBytes , key->n );
	byteNum += primeLengthInBytes;

// Write block type...
	*(size_t *)(keyBlock + byteNum) = RSA_PublicKey;
	byteNum += sizeof( size_t );

	if ( *pKeyBlockSize != byteNum )
		return RSA_INIT_ERROR;

	*pKeyBlock = keyBlock;

	return RSA_OK;
}

int RSACreatePublicKeyFromMPIBuffer( 
		sRSAPublicKey **pKey,
		const byte *mpi,
		int mpiSize )
{
	int res;
	sRSAPublicKey *key;
//	size_t byteNum = 0;

	res = RSAAllocatePublicKey( &key );
	if( res != RSA_OK )
		return res;

	int parsed = 0;

	//
	//	Read module n.
	//
	if (!pw_readMpiBuffer( key->n, mpi, mpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	//
	//	Read public exponent e.
	//
	if (!pw_readMpiBuffer( key->e, mpi, mpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	key->sizeInBits = (uint16b)pw_getSizeInBits( key->n );

	*pKey = key;

	return RSA_OK;
}

//////////////////////////////////////////
// CheckAndExtendPrivateKeyByRFC2440...
//	See RFC 2440 - 5.5.3. Secret Key Packet Formats
//
// Note:
//  Private key is absolutely defined by e,p,q. ( This RSA implementation means p > q )
//  Other values ( n,d,c ( RFC calls it u ) ) can be generated, so they will be tested in debug mode.
//	Release version skips the generation and uses the provided values
//
static int CheckAndExtendPrivateKeyByRFC2440(	sRSAPrivateKey *key  )
{
	//
	//	( This RSA implementation means p > q )
	//
	if( pw_cmp( key->p, key->q ) < 0 )
	{
		//
		// swap the values
		//
		pw_int *swp;

		swp	   = key->q;
		key->q = key->p;
		key->p = swp;
	}

//
//	It was encounterred when the d value from the key was not really inverted with e
//	so we have to recalculate the value.
//
//#ifdef _DEBUG

	/*	Check Private exponent d == e^-1 mod (p-1)(q-1) */

	pw_int t0,t1;

	pw_subu( &t0 , key->q ,1 );	// t0 = q-1
	pw_subu( &t1 , key->p ,1 );	// t1 = p-1
	pw_mul( &t0 , &t0 , &t1 ); /*  t0 = q-1 and t1 = p-1 => t0=(q-1)(p-1)*/

	pw_mulm( &t1, key->d, key->e, &t0 ); // t1 has to be ONE.
	if ( pw_cmpu( &t1, 1 ) != 0 )
	{
//		assert( 0 );

		if ( pw_invert( key->d, key->e, &t0 ) != pw_OK ) // d = e^-1 mod t0 
			return RSA_INIT_ERROR;
	}


	/*	Check c = q^(-1) mod p if q < p */
	/* Used to speed up private cipher operation by Chinese remainder theorem */

	pw_mulm( &t1, key->c, key->q, key->p ); // t1 has to be ONE.
	if ( pw_cmpu( &t1, 1 ) != 0 )
	{
//		assert( 0 );

		if ( pw_invert( key->c, key->q, key->p ) != pw_OK )
			return RSA_INIT_ERROR;
	}


	/*	Check n = p*q */

	pw_mul( &t1 , key->p , key->q );
	if ( pw_cmp( key->n, &t1 ) != 0 )
	{
//		assert( 0 );

		pw_mul( key->n , key->p , key->q );
	}

// #endif // _DEBUG

	key->sizeInBits = (uint16b)pw_getSizeInBits( key->n );

/*	Prepare e1 = d mod (p - 1) */
	pw_subu( key->e1 , key->p , 1 );		/* e1 == p - 1 */
	pw_mod( key->e1 , key->d , key->e1 );	/* e1 == d mod e1 */

/*	Prepare e2 = d mod (q - 1) */
	pw_subu( key->e2 , key->q , 1 );		/* e2 == q - 1 */
	pw_mod( key->e2 , key->d , key->e2 );	/* e2 == d mod e2 */

	return RSA_OK;
}

//
//	Function:
//		RSACreateSecretKeyFromMPIBuffer
//
//	See, RFC 2440 $ 5.5.3. Secret Key Packet Formats
//
int RSACreateSecretKeyFromMPIBuffer( 
		sRSAPrivateKey **pKey,
		const byte *mpi,
		int mpiSize,
		const byte *privateMpi,
		int privateMpiSize
		)
{
	int res;
	sRSAPrivateKey *key;
//	size_t byteNum = 0;

	res = RSAAllocatePrivateKey( &key );
	if( res != RSA_OK )
		return res;

	int parsed = 0;

	//
	//	Read module n.
	//
	if (!pw_readMpiBuffer( key->n, mpi, mpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	//
	//	Read public exponent e.
	//
	if (!pw_readMpiBuffer( key->e, mpi, mpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	//
	//	Read private part...
	//

	parsed = 0;

	//
	// - multiprecision integer (MPI) of RSA secret exponent d.
	//
	if (!pw_readMpiBuffer( key->d, privateMpi, privateMpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	//
	//	- MPI of RSA secret prime value p.
	//
	if (!pw_readMpiBuffer( key->p, privateMpi, privateMpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	//
	//	- MPI of RSA secret prime value q (p < q).
	//
	if (!pw_readMpiBuffer( key->q, privateMpi, privateMpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	//
	//	- MPI of u, the multiplicative inverse of p, mod q.
	//
	if (!pw_readMpiBuffer( key->c, privateMpi, privateMpiSize, &parsed ) )
		return pw_InvalidKeyBlock;

	res = CheckAndExtendPrivateKeyByRFC2440( key );

	if ( res != RSA_OK )
		return res;

	*pKey = key;

	return RSA_OK;
}

int RSAWritePublicKeyToMpiBuffer(
	sRSAPublicKey *key,
	byte *mpiBuff, 
	int *pMpiBuffSize )
{

	int required = pw_getMpiSize( key->n ) + pw_getMpiSize( key->e );

	if ( required > *pMpiBuffSize )
	{
		*pMpiBuffSize = required;
		return RSA_INVALID_BLOCK_SIZE;
	}

	*pMpiBuffSize = required;

	int pos = 0;

	if ( !pw_writeMpiBuffer( mpiBuff, *pMpiBuffSize, &pos, key->n ) ||
		 !pw_writeMpiBuffer( mpiBuff, *pMpiBuffSize, &pos, key->e ) )
			return RSA_INIT_ERROR;

	return RSA_OK;
}

int RSAWriteSecretKeyToMpiBuffer(
	sRSAPrivateKey *key,
	byte *mpiBuff, 
	int *pMpiBuffSize )
{

	int required = 
		pw_getMpiSize( key->d ) +
		pw_getMpiSize( key->q ) +  // here q < p
		pw_getMpiSize( key->p ) +
		pw_getMpiSize( key->c );

	if ( required > *pMpiBuffSize )
	{
		*pMpiBuffSize = required;
		return RSA_INVALID_BLOCK_SIZE;
	}

	*pMpiBuffSize = required;

	int pos = 0;

	if ( !pw_writeMpiBuffer( mpiBuff, *pMpiBuffSize, &pos, key->d ) ||
		 !pw_writeMpiBuffer( mpiBuff, *pMpiBuffSize, &pos, key->q ) ||
		 !pw_writeMpiBuffer( mpiBuff, *pMpiBuffSize, &pos, key->p ) ||
		 !pw_writeMpiBuffer( mpiBuff, *pMpiBuffSize, &pos, key->c ) )
			return RSA_INIT_ERROR;

	return RSA_OK;
}

////////////
//
//	ASN.1 RSA read and write functions.
//
//	RSACreatePublicKeyFromASN1Buffer
//	RSACreateSecretKeyFromASN1Buffer
//
//	RSAWritePublicKeyToASN1Buffer
//	RSAWriteSecretKeyToASN1Buffer

//
//	Function:
//		RSACreatePublicKeyFromASN1Buffer
//
//	See, PKCS-1
//	7.1 Public-key syntax
//
int RSACreatePublicKeyFromASN1Buffer( 
		sRSAPublicKey **pKey,
		x509_publicKeyInfo *pki
		)
{
	int res;

// RSAPublicKey ::= SEQUENCE {
//   modulus INTEGER, -- n
//   publicExponent INTEGER -- e }

	x509_struct s; // SEQUENCE

	int size = pki->publicKey.L;

	res = s.parse( pki->publicKey.B, size, NULL, &size, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	sRSAPublicKey *key;

	res = RSAAllocatePublicKey( &key );
	if( res != RSA_OK )
		return res;

	int parsed = 0;

	//
	//	Read module n.
	//
	if (!pw_readASN1Buffer( key->n, s.B, s.L, &parsed ) )
		res = pw_InvalidKeyBlock;
	else
	//
	//	Read public exponent e.
	//
	if (!pw_readASN1Buffer( key->e, s.B, s.L, &parsed ) )
		res = pw_InvalidKeyBlock;
	else
	if ( (int)s.L != parsed )
		res = pw_InvalidKeyBlock;

	if ( res != RSA_OK )
	{
		RSAFreePublicKey( &key );
		return res;
	}

	key->sizeInBits = (uint16b)pw_getSizeInBits( key->n );
	*pKey = key;

	return RSA_OK;
}

static const char *algRSA_OID = "1.2.840.113549.1.1.1";

int RSAWritePublicKeyInfoToASN1Buffer( 
	sRSAPublicKey *key,
	byte *buff, 
	int *pBuffSize )
{
//
//	PublicKeyInfo  ::=  SEQUENCE  {
//		algorithm     AlgorithmIdentifier,
//		publicKey     BIT STRING
//	}
//

//
//	AlgorithmIdentifier ::= SEQUENCE{
//		    algorithm       OBJECT IDENTIFIER,
//			parameters      ANY DEFINED BY algorithm OPTIONAL ( or NULL )
//	}
//

//
//	RSAPublicKey ::= SEQUENCE {
//		modulus INTEGER, -- n
//		publicExponent INTEGER -- e
//	}
//

	x509_publicKeyInfo pki;

	x509_struct RSAPublicKey;

	//
	//	Forms public key info structure and calculates required size
	//
	int required = 
		pki.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			pki.algOID.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				pki.algOID.oid.prepareForWritting( algRSA_OID )
				+
				pki.algOID.params.prepareForWritting( asn1tag_NULL, 0 )
			)
			+
			pki.publicKey.prepareForWritting( 
				RSAPublicKey.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
					pw_getASN1Size( key->n )
					+
					pw_getASN1Size( key->e )
				)
			)
		);

	if ( required > *pBuffSize )
	{
		*pBuffSize = required;
		return RSA_INVALID_BLOCK_SIZE;
	}

	*pBuffSize = required;

	//
	//	Fills buffers.
	//
	if ( pki.writeHeaders( buff, required) == 0 )
		return RSA_INIT_ERROR;

	if ( RSAPublicKey.writeHeader( pki.publicKey.B, pki.publicKey.L ) == 0 )
		return RSA_INIT_ERROR;

	//
	//	Fills data.
	//
	if ( pki.algOID.oid.writeBody() != ppe_ok )
		return RSA_INIT_ERROR;

	int pos = 0;

	if ( !pw_writeASN1Buffer( RSAPublicKey.B, RSAPublicKey.L, &pos, key->n ) ||
		 !pw_writeASN1Buffer( RSAPublicKey.B, RSAPublicKey.L, &pos, key->e ) )
			return RSA_INIT_ERROR;

	return RSA_OK;
}

//
//	ANS.1 RSA Private key functions
//

//
//	See 7.2 Private-key syntax
//
//	RSAPrivateKey ::= SEQUENCE {
//	  version Version,
//	  modulus INTEGER, -- n
//	  publicExponent INTEGER, -- e
//	  privateExponent INTEGER, -- d
//	  prime1 INTEGER, -- p
//	  prime2 INTEGER, -- q
//	  exponent1 INTEGER, -- d mod (p-1)
//	  exponent2 INTEGER, -- d mod (q-1)
//	  coefficient INTEGER -- (inverse of q) mod p }
//
int RSACreateSecretKeyFromASN1Buffer( 
		sRSAPrivateKey **pKey,
		x509_privateKeyInfo *pki
		)
{

	if ( ( pki->algOID.tu->pAlg != pgppkaRSA ) &&
		 ( pki->algOID.tu->pAlg != pgppkaRSA_e ) &&
		 ( pki->algOID.tu->pAlg != pgppkaRSA_s )  )
		 return RSA_INVALID_KEY_BLOCK;
		 

	x509_struct s; // SEQUENCE

	int size = pki->privateKey.L;

	int res;
	res = s.parse( pki->privateKey.B, pki->privateKey.L, NULL, &size, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
		return x509e_invalidPktSize;

	sRSAPrivateKey *key;

	res = RSAAllocatePrivateKey( &key );
	if( res != RSA_OK )
		return res;

	int parsed = 0;

	pw_int version;

	if ( !pw_readASN1Buffer( &version, s.B, s.L, &parsed ) )
		res = pw_InvalidKeyBlock;
	else
	if ( pw_cmpu( &version, 0 ) != 0 )
		res = x509e_invalidPktVersion;
	else
	if (
		!pw_readASN1Buffer( key->n, s.B, s.L, &parsed ) ||
		!pw_readASN1Buffer( key->e, s.B, s.L, &parsed ) ||
		!pw_readASN1Buffer( key->d, s.B, s.L, &parsed ) ||
		!pw_readASN1Buffer( key->p, s.B, s.L, &parsed ) ||
		!pw_readASN1Buffer( key->q, s.B, s.L, &parsed ) ||
		!pw_readASN1Buffer( key->e1, s.B, s.L, &parsed ) ||
		!pw_readASN1Buffer( key->e2, s.B, s.L, &parsed ) ||
		!pw_readASN1Buffer( key->c, s.B, s.L, &parsed ) ||
		( (int)s.L != parsed ) 
	)
		res = pw_InvalidKeyBlock;
	else
		res = CheckAndExtendPrivateKeyByRFC2440( key );

	if ( res != RSA_OK )
	{
		RSAFreePrivateKey( &key );
		return res;
	}

	*pKey = key;

	return RSA_OK;
}

//
//	Function:
//		RSAWriteSecretKeyInfoToASN1Buffer
//
//	PKCS #8: Private-Key Information Syntax Standard
//	see x509_privateKeyInfo
//
int RSAWriteSecretKeyInfoToASN1Buffer(
	sRSAPrivateKey *key,
	byte *buff, 
	int *pBuffSize )
{

	x509_privateKeyInfo pki;

	x509_struct RSAPrivateKey;

	pw_int version;
	version.setu( 0 );

	//
	//	Forms private key info structure and calculates required size
	//
	int required = 
		pki.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
//			pki.version.prepareForWritting( asn1tag_INTEGER, pw_getSizeInBytes( &version ) )
			pki.version.prepareForWritting( 0 )
			+
			pki.algOID.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				pki.algOID.oid.prepareForWritting( algRSA_OID )
				+
				pki.algOID.params.prepareForWritting( asn1tag_NULL, 0 )
			)
			+
			pki.privateKey.prepareForWritting( asn1tag_OCTET_STRING,
				RSAPrivateKey.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
					pw_getASN1Size( &version ) + 
					pw_getASN1Size( key->n ) + 
					pw_getASN1Size( key->e ) +
					pw_getASN1Size( key->d ) +
					pw_getASN1Size( key->p ) +
					pw_getASN1Size( key->q ) +
					pw_getASN1Size( key->e1 ) +
					pw_getASN1Size( key->e2 ) +
					pw_getASN1Size( key->c )
				)
			)
			// No attributes...
		);

	if ( required > *pBuffSize )
	{
		*pBuffSize = required;
		return RSA_INVALID_BLOCK_SIZE;
	}

	*pBuffSize = required;

	//
	//	Fills headers.
	//
	if ( pki.writeHeaders( buff, required ) == 0 )
		return RSA_INIT_ERROR;

	if ( RSAPrivateKey.writeHeader( pki.privateKey.B, pki.privateKey.L ) == 0 )
		return RSA_INIT_ERROR;

	//
	//	Fills data.
	//
	if ( pki.algOID.oid.writeBody() != ppe_ok ) 
		return RSA_INIT_ERROR;

	if ( pki.version.writeBody() != ppe_ok )
		return RSA_INIT_ERROR;

	int pos = 0;

	if ( !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, &version ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->n ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->e ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->d ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->p ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->q ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->e1 ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->e2 ) ||
		 !pw_writeASN1Buffer( RSAPrivateKey.B, RSAPrivateKey.L, &pos, key->c )
	)
			return RSA_INIT_ERROR;

	return RSA_OK;

}
