#ifndef __RSA_H__
#define __RSA_H__

#include "../Power/power.h"
#include "../../x509pkts/x509_structs.h"
#include "../../crtLibErrs.h"
#include "Defs.h"

//
//	Definition...
//
//	__UseRandomBuffer__
//
//	Purpouse:
//		With the define we can exclude Random and SHA1 files from the project.
//
#define __UseRandomBuffer__

//typedef unsigned int size_t;

/* See PKCS1 */
struct sRSAPublicKey {
  uint16b sizeInBits;
  pw_int *n;                     /* Modulus */
  pw_int *e;                     /* Public exponent */

  byte *keyBlock;				// Key block pointer used for deallocation if not NULL;
};

/* See PKCS1 */
struct sRSAPrivateKey {
  uint16b sizeInBits;
  pw_int *n;                     /* Modulus */
  pw_int *e;                     /* Public exponent */
  pw_int *d;                     /* Private exponent */
  pw_int *p;                     /* The prime factor p of n */
  pw_int *q;                     /* The prime factor q of n */
  pw_int *e1;                    /* exponent1 is d mod (p-1) */
  pw_int *e2;                    /* exponent2 is d mod (q-1) */
  pw_int *c;                     /* coefficient is the Chinese Remainder
                                  Theorem coefficient q-1 mod p */

  byte *keyBlock;				// Key block pointer used for deallocation if not NULL;
};

//
//	RSADerivePublicKey
//		Derives public key from private key. Application has to use
//		RSAFreePublicKey when finishes work with the public key...
//
int RSADerivePublicKey(
	sRSAPrivateKey *privKey,
	sRSAPublicKey **publicKey );

#ifndef __UseRandomBuffer__

#include "random.h"

/*
 *	RSAGeneratePrivateKey:
 *	Note:
 *	 Random generator required,
 *	 Public exponent can be specified as 0x03 or 0xF4
 */

int RSAGeneratePrivateKey(
	sRSAPrivateKey **privateKey,
	size_t bits,
	BYTE *seedBuffer,
	size_t seedBufferSize,
	size_t publicExp = 0, // 0 == Undefined
	bool useSafePrime = false
	);
#else // __UseRandomBuffer__

/*
 *	RSAGeneratePrivateKeyFromRandomBuffer:
 *	Note:
 *	 Public exponent can be specified as 0x03 and F4 ( 65537 0x10001)
 */

class cProgressActor;

int RSAGeneratePrivateKeyFromRandomBuffer(
	sRSAPrivateKey **privateKey,
	BYTE *buffer,
	size_t bufferSize,
	size_t publicExp = 0, // 0 == Undefined 
	cProgressActor *actor = NULL,
	bool useSafePrime = false
	);

//
//	RSAPreparePrimesFromRandomBuffer:
//	Helper function.
//	It finds nearest primes for random in the buffer.
//	So we can very quikly generate a real private key from the primes.
//
int RSAPreparePrimesFromRandomBuffer(
	BYTE *buffer,
	size_t bufferSize,
	size_t publicExp = 0,
	bool useSafePrime = false
	);

#endif // __UseRandomBuffer__

////////////////////////////////////////
// Encrypt / Decrypt functions
// Notes:
//  1) Block's sizes are to equal key->sizeInBits / 8
//  2) First byte of plane text data is to be less 0x80 
//   -> all value is to be less then module.
//	3) Both update function can be used as for encryption 
//   so for decryption operation.
//  4) Input block can be used as output.

/* Private key operation with using Chinese remainder theorem */

//
//	flags emuns
//
enum { useMpiIn = 1, useMpiOut = 2 };

int
	RSAPrivateUpdate(
		sRSAPrivateKey *key,
		const byte *inputBlock,	/* input data */
		byte *outputBlock,	/* output data */
		int flags = 0 // see emuns above
		);
int
	RSAPublicUpdate(
		sRSAPublicKey *key,
		const byte *inputBlock,		/* input data */
		byte *outputBlock,		/* output data */
		int flags = 0 // see emuns above
		);


/* Perform safe deallocation of allocated resources */
int RSAFreePublicKey( sRSAPublicKey **key );
int RSAFreePrivateKey( sRSAPrivateKey **Key );


//
//	MPI Buffer funcs.
//	
int RSACreatePublicKeyFromMPIBuffer( 
		sRSAPublicKey **pKey,
		const byte *mpi,
		int mpiSize );

int RSACreateSecretKeyFromMPIBuffer( 
		sRSAPrivateKey **pKey,
		const byte *mpi,
		int mpiSize,
		const byte *privateMpi,
		int privateMpiSize
		);

int RSAWritePublicKeyToMpiBuffer(
	sRSAPublicKey *key,
	byte *mpiBuff, 
	int *pMpiBuffSize );

int RSAWriteSecretKeyToMpiBuffer(
	sRSAPrivateKey *key,
	byte *mpiBuff, 
	int *pMpiBuffSize );

//
//	ASN.1 package funcs.
//
int RSACreatePublicKeyFromASN1Buffer( sRSAPublicKey **pKey,	x509_publicKeyInfo *pki	);
int RSACreateSecretKeyFromASN1Buffer( sRSAPrivateKey **pKey, x509_privateKeyInfo *pki	);

int RSAWritePublicKeyInfoToASN1Buffer( sRSAPublicKey *key, byte *buff, int *pBuffSize );
int RSAWriteSecretKeyInfoToASN1Buffer( sRSAPrivateKey *key,byte *buff, int *pBuffSize );

#endif /* __RSA_H__ */

