#ifndef __cDSA__
#define __cDSA__

#include "DSA/Dsa.h"

class cPublicDSA : public cPKA
{
 public:

	DSA_KeyHandle key;

	int moduleSize;

	cPublicDSA()
	{
		key = NULL;
		moduleSize = 0;
	}

	~cPublicDSA()
	{
		if ( key )
			DSA_FreeKey( &key );
	}

	static class cPKA *createAlg( BYTE *mpiBf, int bfSize )
	{
		cPublicDSA *alg = new cPublicDSA;

		alg->key = DSA_ReadPublicKeyFromMpiBuffer( mpiBf, bfSize );

		if ( alg->key == NULL )
		{
			delete alg;
			return NULL;
		}

		alg->id = 17;
		alg->moduleSize = ( DSA_GetModulusBitSize( alg->key ) + 7 ) / 8;

		return alg;

	}

	static int createAlg_X509( cPKA **pPKA, x509_publicKeyInfo *pki )
	{
		cPublicDSA *alg = new cPublicDSA;

		int res = DSA_ReadPublicKeyFromASN1Buffer( &(alg->key), pki );

		if ( res != x509e_ok )
		{
			delete alg;
			return x509e_invalidPkt;
		}

		alg->id = 17;
		alg->moduleSize = ( DSA_GetModulusBitSize( alg->key ) + 7 ) / 8;

		*pPKA = alg;

		return x509e_ok;

	}

	virtual int getModuleSize()
	{
		return moduleSize;
	}

	virtual int getBlockSize()
	{
		return DSA_SMALL_PRIME_BYTE_SIZE;
	}

	int verify(
			const BYTE *sigBuff,// IN - signature MPI buffer 
			int  sigBuffSize,	// IN  - size of allocated buffer.
			const BYTE *digestBuff,
								// IN  - formatted digest buffer. 
								// This value is padded as described in PKCS-1
			int digestBuffSize	// Size of the formatted digest buffer to be verified with signature. 
			)
	{
		int res = DSA_Verify( key, sigBuff, sigBuffSize, digestBuff, digestBuffSize );

		if ( res != 0 )
			return pkae_invalidSignature;

		return pkae_ok;
	}

	int verify_X509( const BYTE *sigBuf, int   sigBufSize, const BYTE *digestBuf, int digestBufSize )
	{
		int res = DSA_Verify_X509( key, sigBuf, sigBufSize, digestBuf, digestBufSize );

		if ( res != 0 )
			return pkae_invalidSignature;

		return pkae_ok;
	}

	int writeKeyToASN1Buffer(
		BYTE *buff,		// buffer for the key, x509_publicKeyInfo
		int  *pBuffSize	// IN  - pointer to size of buffer,
						// OUT - pointer to required or busied size.
						// the value is filled even if the error 
						// pkae_bufferSizeTooSmall returned
	)
	{
		return DSA_WritePublicKeyToASN1Buffer( key, buff, pBuffSize );
	}
};

class cPrivateDSA : public cPKA
{
	DSA_KeyHandle key;

	int moduleSize;

 public:

	cPrivateDSA()
	{
		key = NULL;
		moduleSize = 0;
	}

	~cPrivateDSA()
	{
		if ( key )
			DSA_FreeKey( &key );
	}

#ifdef _WIN32
#pragma warning( disable : 4100 )
#endif

	static class cPKA *createAlg( BYTE *mpiBf, int bfSize, BYTE *mpiPrivateBf, int privateBfSize  )
	{
		cPrivateDSA *alg = new cPrivateDSA;

		alg->key = DSA_ReadSecretKeyFromMpiBuffer( mpiBf, bfSize, mpiPrivateBf, privateBfSize );

		if ( alg->key == NULL )
		{
			delete alg;
			return NULL;
		}

		alg->id = 17;
		alg->moduleSize = ( DSA_GetModulusBitSize( alg->key ) + 7 ) / 8;

		return alg;

	}

	static int createAlg_X509( cPKA **pPKA, x509_privateKeyInfo *pki )
	{
		cPrivateDSA *alg = new cPrivateDSA;

		int res = DSA_ReadSecretKeyFromASN1Buffer( &(alg->key), pki );

		if ( res != x509e_ok )
		{
			delete alg;
			return res;
		}

		alg->id = 17;
		alg->moduleSize = ( DSA_GetModulusBitSize( alg->key ) + 7 ) / 8;

		return x509e_ok;
	}

	virtual int getModuleSize()
	{
		return moduleSize;
	}

	virtual int getBlockSize()
	{
		return DSA_SMALL_PRIME_BYTE_SIZE;
	}

	int sign(
		BYTE *sigBuff,		// OUT - signature MPI buffer 
		int  *pSigBuffSize,	// IN  - pointer pointer size of allocated buffer.
							// OUT - ointer to required or really busied size.
		const BYTE *digestBuff,	// IN  - formatted digest buffer to be signed. 
							// This value is padded as described in PKCS-1
		int digestBuffSize	// Size of the formatted digest buffer. 
							// It has to be equal to module size in bytes 
	)
	{
		int res = DSA_Sign( key, sigBuff, pSigBuffSize, digestBuff, digestBuffSize );

		if ( res == pw_BufferSizeTooSmall )
			return pkae_bufferSizeTooSmall;

		if ( res != pw_OK )
			return pkae_internalError;

		return pkae_ok;
	}

	int sign_X509(  BYTE *sigBuf, int *pSigBufSize, const BYTE *digestBuf, int digestBufSize )
	{
		int res = DSA_Sign_X509( key, sigBuf, pSigBufSize, digestBuf, digestBufSize );

		if ( res == pw_BufferSizeTooSmall )
			return pkae_bufferSizeTooSmall;

		return res;
	}

	static int generateKey(
		int	 keyLength,				// Key size in byte,
		BYTE *publicMpiBuff,		// Public key MPI buffer,
		int  *pPublicMpiBuffSize,	// IN  - pointer to size of MPI buffer,
									// OUT - pointer to required or busied size.
									// the value is filled even if the error 
									// pkae_bufferSizeTooSmall returned
		BYTE *secretMpiBuff,		// Secret key MPI buffer,
		int  *pSecretMpiBuffSize	// IN  - pointer to size of MPI buffer,
									// OUT - pointer to required or busied size.
									// the value is filled even if the error 
									// pkae_bufferSizeTooSmall returned
		)
	{
		return pkae_internalError;
	}

	int writeKeyToASN1Buffer(
		BYTE *buff,		// buffer for the key, x509_privateKeyInfo
		int  *pBuffSize	// IN  - pointer to size of buffer,
						// OUT - pointer to required or busied size.
						// the value is filled even if the error 
						// pkae_bufferSizeTooSmall returned
	)
	{
		return DSA_WriteSecretKeyToASN1Buffer( key, buff, pBuffSize );
	}

	virtual int derivePublicKeyAlgorithm(
		HPKA *pHPka	// OUT - pointer to the derived public key algorithm.
		)
	{
		cPublicDSA *publicAlg = new cPublicDSA;

		if ( publicAlg == NULL )
			return pkae_memoryAllocation;

		int res = DSA_DerivePublicKey( key, &(publicAlg->key) );

		if ( res != RSA_OK )
		{
			delete publicAlg;
			return pkae_memoryAllocation;
		}

		publicAlg->moduleSize = ( DSA_GetModulusBitSize( publicAlg->key ) + 7 ) / 8;
			//( publicAlg->key->sizeInBits + 7 ) / 8;

		*pHPka = publicAlg;

		return pkae_ok;
	}
#ifdef _WIN32
#pragma warning( default : 4100 )
#endif
};

#endif // __cDSA__
