#ifndef __cAlgRSA__
#define __cAlgRSA__

#include "RSA/Rsa.h"
#include "../RNG/cRNG.h"
#include "../PGPPkts/Packets.h"

class cPublicRSA : public cPKA
{
 public:
	sRSAPublicKey *key;

	int moduleSize;

	cPublicRSA()
	{
		key = NULL;
		moduleSize = 0;
	}

	~cPublicRSA()
	{
		if ( key )
			RSAFreePublicKey( &key );
	}

	static class cPKA *createAlg( BYTE *mpiBf, int bfSize )
	{
		cPublicRSA *alg = new cPublicRSA;

		int res = RSACreatePublicKeyFromMPIBuffer( &(alg->key), mpiBf, bfSize );

		if ( res != RSA_OK )
		{
			delete alg;
			return NULL;
		}

		alg->moduleSize = ( alg->key->sizeInBits + 7 ) / 8;
		alg->id = 1;

		return alg;

	}

	static int createAlg_X509( cPKA **pPka, x509_publicKeyInfo *pki )
	{
		cPublicRSA *alg = new cPublicRSA;

		int res = RSACreatePublicKeyFromASN1Buffer( &(alg->key), pki );

		if ( res != RSA_OK )
		{
			delete alg;
			return x509e_invalidPkt;
		}

		alg->moduleSize = ( alg->key->sizeInBits + 7 ) / 8;
		alg->id = 1;

		*pPka = alg;

		return res;
	}

	int getModuleSize()
	{
		return moduleSize;
	}

	int encrypt(
			BYTE *encBuff,		// OUT - MPI buffer of encrypted message
			int  *encBuffSize,	// IN  - pointer to size of allocated buffer.
								// OUT - pointer to required or busied size.
			const BYTE *messageBuff,
								// IN  - formatted message buffer. 
								// This value is padded as described in PKCS-1
			int messageBuffSize	// Size of the message buffer to be encrypted. 
								// It has to be equal to module size in bytes 
			)
	{
		if ( *encBuffSize < moduleSize + 2 ) // Size with length of mpi
		{
			*encBuffSize = moduleSize + 2;
			return pkae_bufferSizeTooSmall;
		}

		*encBuffSize = moduleSize + 2;

		if ( messageBuffSize != moduleSize )
		{
			return pkae_invalidInputData;
		}

		int res = RSAPublicUpdate( key, messageBuff, encBuff, useMpiOut );

		if ( res != RSA_OK )
			return pkae_internalError;

		return pkae_ok;
	}

	int verify(
			const BYTE *sigBuff,		// IN - signature MPI buffer 
			int  sigBuffSize,	// IN  - size of allocated buffer.
			const BYTE *digestBuff,	// IN  - formatted digest buffer. 
								// This value is padded as described in PKCS-1
			int digestBuffSize	// Size of the formatted digest buffer to be verified with signature. 
			)
	{
		if ( sigBuffSize < moduleSize + 2 ) // Size with length of mpi
		{
			return pkae_bufferSizeTooSmall;
		}

		if ( digestBuffSize > moduleSize )
			return pkae_invalidInputData;

		return verifySignatureBlock( sigBuff, digestBuff, digestBuffSize, useMpiIn );

		//byte *decryptedSig = new byte[ moduleSize ];

		//if ( decryptedSig == NULL )
		//	return pkae_memoryAllocation;

		//int res = RSAPublicUpdate( key, sigBuff, decryptedSig, useMpiIn );

		//// v. 1.04.1 Checks entire formated block. Prev vesions check digest only.
		//const BYTE *pos;
		//int size;

		//if ( res == RSA_OK )
		//	res = RSA_PublicKeyEncryptionBlock_Parse( false, decryptedSig, moduleSize, &pos, &size );

		//if ( ( res == RSA_OK ) &&
		//	 ( size != digestBuffSize ) )
		//		res = ppe_invalidPacketSize;

		//if ( res == RSA_OK )
		//	res = memcmp( pos, digestBuff, digestBuffSize );

		//delete decryptedSig;

		//if ( res != 0 )
		//	return pkae_invalidSignature;

		//return pkae_ok;
	}

	int decrypt(
			const BYTE *encBuff,// IN - signature MPI buffer 
			int  encBuffSize,	// IN  - size of allocated buffer.
			BYTE *decBuff,		// OUT decrypted buffer. 
			int *pDecBuffSize 	// OUT required size of decrypted buffer. size of allocated buffer
			)
	{
		if ( encBuffSize < moduleSize + 2 ) // Size with length of mpi
		{
			return pkae_invalidInputData;
		}

		if ( *pDecBuffSize < moduleSize )
		{
			*pDecBuffSize = moduleSize;
			return pkae_bufferSizeTooSmall;
		}

		*pDecBuffSize = moduleSize;

		return RSAPublicUpdate( key, encBuff, decBuff, useMpiIn );
	}

	int encrypt_X509(
			BYTE *encBuff,		// OUT - ASN.1 buffer of encrypted message
			int  *encBuffSize,	// IN  - pointer to size of allocated buffer.
								// OUT - pointer to required or busied size.
			const BYTE *messageBuff,
								// IN  - formatted message buffer. 
								// This value is padded as described in PKCS-1
			int messageBuffSize	// Size of the message buffer to be encrypted. 
								// It has to be equal to module size in bytes 
			)
	{
		if ( messageBuffSize != moduleSize )
		{
			return pkae_invalidInputData;
		}

		x509_struct s;

		if ( s.prepareForWritting( asn1tag_OCTET_STRING, moduleSize ) == 0 )
			return x509e_internalError;

		if ( *encBuffSize < (int)s.TL )
		{
			*encBuffSize = s.TL;
			return pkae_bufferSizeTooSmall;
		}
		
		if ( s.writeHeader( encBuff, *encBuffSize ) == 0 )
			return x509e_internalError;

		*encBuffSize = s.TL;

		int res = RSAPublicUpdate( key, messageBuff, s.B );

		if ( res != RSA_OK )
			return pkae_invalidSignature;

		return pkae_ok;
	}

	int verify_X509(
			const BYTE *sigBuff,// IN - signature ASN.1 buffer 
			int  sigBuffSize,	// IN  - size of allocated buffer.
			const BYTE *digestBuff,	// IN  - formatted digest buffer. 
								// This value is padded as described in PKCS-1
			int digestBuffSize	// Size of the formatted digest buffer to be verified with signature. 
			)
	{
		if ( digestBuffSize > moduleSize )
			return pkae_invalidInputData;

		x509_struct s;

		int L;
		BYTE *B;

		int res;

		//
		//	RSA recommend to pack signature tp OCTET STRING
		//

		if ( ( res = s.parse( sigBuff, sigBuffSize, 0, 0, asn1tag_OCTET_STRING ) ) == x509e_ok )
		{
			L = s.L;
			B = s.B;
		}
		else

		//
		//	RFC 2459. Uses signature value for X509 certificate as BIT STRING
		//
		if ( ( res = s.parse( sigBuff, sigBuffSize, 0, 0, asn1tag_BIT_STRING ) ) == x509e_ok )
		{
			L = s.L - 1;
			B = s.B + 1;
		}
		else
			return res;

		if ( L != moduleSize  )
			return pkae_invalidInputData;

		return verifySignatureBlock( B, digestBuff, digestBuffSize );
	}

	int verifySignatureBlock( const BYTE *encryptedSig, const BYTE *digestBuff, int digestBuffSize, int mpiFlags = 0 )
	{
		byte *decryptedSig = new byte[ moduleSize ];

		if ( decryptedSig == NULL )
			return pkae_memoryAllocation;

		int res = RSAPublicUpdate( key, encryptedSig, decryptedSig, mpiFlags );

		// v. 1.04.1 Checks entire formated block. Prev vesions check digest only.
		const BYTE *pos = NULL;
		int size = 0;

		if ( res == RSA_OK )
			res = RSA_PublicKeyEncryptionBlock_Parse( false, decryptedSig, moduleSize, &pos, &size );

		if ( ( res == RSA_OK ) && ( size != digestBuffSize ) )
		{
			// - 06.03.2014
			// signature may be packed as x509 digest info. Parses it before checking.
			x509_digestInfo di;

			res = di.parse( (BYTE*)pos, size );
			if ( res == x509e_ok )
			{
				pos = di.sValue.B;
				size = di.sValue.L;

				if ( size != digestBuffSize )
					res = pkae_invalidSignature;
			}
		}

		if ( res == RSA_OK )
			res = memcmp( pos, digestBuff, digestBuffSize );

		memset( decryptedSig, 0, moduleSize );
		delete decryptedSig;

		if ( res != 0 )
			return pkae_invalidSignature;

		return pkae_ok;
	}

	int decrypt_X509(
			const BYTE *encBuff,// IN - encrypted ASN.1 buffer 
			int  encBuffSize,	// IN - size of encrypted buffer.
			BYTE *decBuff,		// OUT decrypted buffer. 
			int *pDecBuffSize 	// OUT required size of decrypted buffer. size of allocated buffer
			)
	{
		x509_struct s;

		int res = s.parse( encBuff, encBuffSize, 0,0, asn1tag_OCTET_STRING );
		if ( res != x509e_ok ) return res;

		if ( (int)s.L != moduleSize )
			return pkae_invalidInputData;

		if ( *pDecBuffSize < moduleSize )
		{
			*pDecBuffSize = moduleSize;
			return pkae_bufferSizeTooSmall;
		}

		*pDecBuffSize = moduleSize;

		return RSAPublicUpdate( key, s.B, decBuff );
	}

	int writeKeyToMpiBuffer(
		BYTE *mpiBuff,		// Public key MPI buffer,

		int  *pMpiBuffSize	// IN  - pointer to size of MPI buffer,
							// OUT - pointer to required or busied size.
							// the value is filled even if the error 
							// pkae_bufferSizeTooSmall returned
		)
	{
		return RSAWritePublicKeyToMpiBuffer( key, mpiBuff, pMpiBuffSize );
	}

	int writeKeyToASN1Buffer(
		BYTE *buff,		// buffer for the key, x509_publicKeyInfo
		int  *pBuffSize	// IN  - pointer to size of buffer,
						// OUT - pointer to required or busied size.
						// the value is filled even if the error 
						// pkae_bufferSizeTooSmall returned
	)
	{
		return RSAWritePublicKeyInfoToASN1Buffer( key, buff, pBuffSize );
	}
};

class cPrivateRSA : public cPKA
{
	sRSAPrivateKey *key;

	int moduleSize;

 public:
	static bool m_generateSafePrime;

	cPrivateRSA()
	{
		key = NULL;
		moduleSize = 0;
	}

	~cPrivateRSA()
	{
		if ( key )
			RSAFreePrivateKey( &key );
	}

	static class cPKA *createAlg( BYTE *mpiBf, int bfSize, BYTE *mpiPrivateBf, int privateBfSize  )
	{
		cPrivateRSA *alg = new cPrivateRSA;

		int res = RSACreateSecretKeyFromMPIBuffer( &(alg->key), mpiBf, bfSize, mpiPrivateBf, privateBfSize );

		if ( res != RSA_OK )
		{
			delete alg;
			return NULL;
		}

		alg->moduleSize = ( alg->key->sizeInBits + 7 ) / 8;
		alg->id = 1;

		return alg;

	}

	static int createAlg_X509( cPKA **pSka, x509_privateKeyInfo *pki )
	{
		cPrivateRSA *alg = new cPrivateRSA;

		int res = RSACreateSecretKeyFromASN1Buffer( &(alg->key), pki );

		if ( res != RSA_OK )
		{
			delete alg;
			return x509e_invalidPkt;
		}

		alg->moduleSize = ( alg->key->sizeInBits + 7 ) / 8;
		alg->id = 1;

		*pSka = alg;

		return res;
	}

	int getModuleSize()
	{
		return moduleSize;
	}

	int decrypt(
		BYTE *decBuff,		// OUT - buffer for decrypted message
		int  *decBuffSize,	// IN  - pointer to size of allocated buffer. ( required size == module size in byte )
							// OUT - pointer to required or busied size.
		const BYTE *encBuff,		// IN  - encrypted MPI buffer. 
		int encBuffSize	
	)
	{
		if ( *decBuffSize < moduleSize )
		{
			*decBuffSize = moduleSize;
			return pkae_bufferSizeTooSmall;
		}

		//
		//	Notes for v. 1.02.3 - 03.12.2007
		//
		//	if ( encBuffSize != moduleSize + 2 )
		//	Note, encrypted buffer mpi size may be less then module size + 2
		//	Becouse encrypted result any mpi ( even zero ).
		//	So we may check only encBuffSize > moduleSize + 2 for error here
		//
		if ( encBuffSize > moduleSize + 2 )
		{
			return pkae_invalidInputData;
		}

		int res = RSAPrivateUpdate( key, encBuff, decBuff, useMpiIn );

		if ( res != RSA_OK )
			return pkae_internalError;

		return pkae_ok;
	}

	int sign(
		BYTE *sigBuff,		// OUT - signature MPI buffer 
		int  *sigBuffSize,	// IN  - pointer pointer size of allocated buffer.
							// OUT - ointer to required or really busied size.
		const BYTE *digestBuff,	// IN  - formatted digest buffer to be signed. 
							// This value is padded as described in PKCS-1
		int digestBuffSize	// Size of the formatted digest buffer. 
							// It has to be equal to module size in bytes 
	)
	{
		if ( *sigBuffSize < moduleSize + 2 ) // Size with length of mpi
		{
			*sigBuffSize = moduleSize + 2;
			return pkae_bufferSizeTooSmall;
		}

		if ( digestBuffSize != moduleSize )
		{
			return pkae_invalidInputData;
		}
		

		int res = RSAPrivateUpdate( key, digestBuff, sigBuff, useMpiOut );


		if ( res != RSA_OK )
			return pkae_invalidSignature;

		return pkae_ok;
	}

	int decrypt_X509(
		BYTE *decBuff,		// OUT - buffer for decrypted message
		int  *decBuffSize,	// IN  - pointer to size of allocated buffer. ( required size == module size in byte )
							// OUT - pointer to required or busied size.
		const BYTE *encBuff,// IN  - encrypted ASN.1 buffer. 
		int encBuffSize	
	)
	{
		if ( *decBuffSize < moduleSize )
		{
			*decBuffSize = moduleSize;
			return pkae_bufferSizeTooSmall;
		}

		x509_struct s;

		int size = encBuffSize;

		int res = s.parse( encBuff, size, NULL, &size, asn1tag_OCTET_STRING );
		if ( res != x509e_ok ) return res;			

		if ( (int)s.L != moduleSize  )
		{
			return pkae_invalidInputData;
		}

		res = RSAPrivateUpdate( key, s.B, decBuff );

		if ( res != RSA_OK )
			return pkae_internalError;

		return pkae_ok;
	}

	int sign_X509(
		BYTE *sigBuff,		// OUT - signature ASN.1 buffer 
		int  *sigBuffSize,	// IN  - pointer pointer size of allocated buffer.
							// OUT - ointer to required or really busied size.
		const BYTE *digestBuff,	// IN  - formatted digest buffer to be signed. 
							// This value is padded as described in PKCS-1
							// Use the PKT_PublicKeyEncryptionBlock_Format function 
							//	to prepare formated block.
							// Note , for a private-key operation, the block type shall be 00 or 01.
		int digestBuffSize	// Size of the formatted digest buffer. 
							// It has to be equal to module size in bytes 
	)
	{
		
		x509_struct s;

		if ( s.prepareForWritting( asn1tag_OCTET_STRING, moduleSize ) == 0 )
			return x509e_internalError;

		if ( *sigBuffSize < (int)s.TL )
		{
			*sigBuffSize = s.TL;
			return pkae_bufferSizeTooSmall;
		}

		if ( digestBuffSize > moduleSize )
		{
			return pkae_invalidInputData;
		}
		else
		if ( digestBuffSize == moduleSize ) // digest buffer is already packed
		{
			// So we check the packet validity
			int res = RSA_PublicKeyEncryptionBlock_Parse( false, digestBuff, digestBuffSize );
			if ( res != ppe_ok )
				return res;
		}

		if ( s.writeHeader( sigBuff, *sigBuffSize ) == 0 )
			return x509e_internalError;

		*sigBuffSize = s.TL;

		BYTE *formattedBlock = NULL;

		if ( digestBuffSize < moduleSize )
		{
			// signature block was not formatted
			// return pkae_invalidInputData;

			formattedBlock = new BYTE[ moduleSize ];
			if ( formattedBlock == NULL )
				return RSA_MALLOC_ERROR;

			int res = RSA_PublicKeyEncryptionBlock_Format( false, formattedBlock, moduleSize, digestBuff, digestBuffSize ); // publicKeyEncryption == false
			if ( res != ppe_ok )
			{
				delete formattedBlock;
				return res;
			}

			digestBuff = formattedBlock;
		}

		int res = RSAPrivateUpdate( key, digestBuff, s.B );

		if ( formattedBlock != NULL )
		{
			memset( formattedBlock, 0, moduleSize );
			delete( formattedBlock );
		}

		if ( res != RSA_OK )
			return pkae_invalidSignature;

		return pkae_ok;
	}

	static int generateKeyAndCreateAlgorithm(
		HPKA *pSka,		// OUT - pointer to the algorithm with new generated key.
		int	 keyLength,	// Key size in byte
		cProgressActor *actor = NULL
		)
	{
		*pSka = NULL;

		BYTE *rndBuff = cRNG::get( keyLength );

		if ( rndBuff == NULL )
			return pkae_randomBufferRequired;

		cPrivateRSA *alg = new cPrivateRSA;

		if ( alg == NULL )
			return pkae_memoryAllocation;

		int res = RSAGeneratePrivateKeyFromRandomBuffer( &(alg->key), rndBuff, keyLength, 0, actor, m_generateSafePrime );

		if ( res != RSA_OK )
		{
			delete alg;
			return pkae_internalError;
		}

		alg->moduleSize = ( alg->key->sizeInBits + 7 ) / 8;
		alg->id = 1;

		*pSka = alg;

		return pkae_ok;
	}


	int derivePublicKeyAlgorithm(
		HPKA *pHPka	// OUT - pointer to the derived public key algorithm.
		)
	{
		cPublicRSA *publicAlg = new cPublicRSA;

		if ( publicAlg == NULL )
			return pkae_memoryAllocation;

		int res = RSADerivePublicKey( key, &(publicAlg->key) );

		if ( res != RSA_OK )
		{
			delete publicAlg;
			return pkae_memoryAllocation;
		}

		publicAlg->moduleSize = ( publicAlg->key->sizeInBits + 7 ) / 8;
		publicAlg->id = 1;

		*pHPka = publicAlg;

		return pkae_ok;
	}

	int writeKeyToMpiBuffer(
		BYTE *mpiBuff,		// Public key MPI buffer,

		int  *pMpiBuffSize	// IN  - pointer to size of MPI buffer,
							// OUT - pointer to required or busied size.
							// the value is filled even if the error 
							// pkae_bufferSizeTooSmall returned
		)
	{
		return RSAWriteSecretKeyToMpiBuffer( key, mpiBuff, pMpiBuffSize );
	}

	int writeKeyToASN1Buffer(
		BYTE *buff,		// buffer for the key, x509_privateKeyInfo
		int  *pBuffSize	// IN  - pointer to size of buffer,
						// OUT - pointer to required or busied size.
						// the value is filled even if the error 
						// pkae_bufferSizeTooSmall returned
	)
	{
		return RSAWriteSecretKeyInfoToASN1Buffer( key, buff, pBuffSize );
	}
};



#endif // __cAlgRSA__
