#include "publicKeyAlgs.h"
#include "../PGPPkts/pktdefs.h"
#include "../hashAlgs/hashAlgs.h"
#include <assert.h>
#include "cPKA.h"
#include "cAlgRSA.h"
#include "cAlgDSA.h"
#include "cAlgElgamal.h"
#include "../RNG/cRNG.h"

typedef cPKA *typePublicCreateAlg( BYTE *mpiBf, int bfSize );
typedef cPKA *typePrivateCreateAlg( BYTE *mpiBf, int bfSize, BYTE *mpiPrivateBf, int privateBfSize );
typedef int typePublicCreateAlg_X509( HPKA *pPka, x509_publicKeyInfo *pki );
typedef int typePrivateCreateAlg_X509( HPKA *pSka, x509_privateKeyInfo *pki );
typedef int typeGenerateKeyAndCreateAlgorithm( cPKA **pSka, int keyLength, cProgressActor *actor );

int pka_setRandomBuffer( BYTE *buffer, int bufferSize )
{
	cRNG::setPool( buffer, bufferSize );
	return pkae_ok;
}

struct sPublicAlgTabItem {	int algID;	typePublicCreateAlg *createFunc; typePublicCreateAlg_X509 *createX509Func; }
	algPublicTbl[] = 
		{
			{  1, cPublicRSA::createAlg,		cPublicRSA::createAlg_X509 },
			{  2, cPublicRSA::createAlg,		cPublicRSA::createAlg_X509 },
			{  3, cPublicRSA::createAlg,		cPublicRSA::createAlg_X509 },
			{ 17, cPublicDSA::createAlg,		cPublicDSA::createAlg_X509 },
			{ 16, cPublicElgamal::createAlg,	cPublicElgamal::createAlg_X509 },
			{ 20, cPublicElgamal::createAlg,	cPublicElgamal::createAlg_X509 }
		};

struct sPrivateAlgTabItem { int algID; 	typePrivateCreateAlg *createFunc; typePrivateCreateAlg_X509 *createX509Func; }
	algPrivateTbl[] = 
		{
			{  1, cPrivateRSA::createAlg, cPrivateRSA::createAlg_X509 },
			{  2, cPrivateRSA::createAlg, cPrivateRSA::createAlg_X509 },
			{  3, cPrivateRSA::createAlg, cPrivateRSA::createAlg_X509 },
			{ 17, cPrivateDSA::createAlg, cPrivateDSA::createAlg_X509 },
			{ 16, cPrivateElgamal::createAlg, cPrivateElgamal::createAlg_X509 },
			{ 20, cPrivateElgamal::createAlg, cPrivateElgamal::createAlg_X509 }
		};

struct sGenerateKeyTabItem { int algID; typeGenerateKeyAndCreateAlgorithm *generateKeyAndCreateAlgorithmFunc; }
	generateKeyTbl[] = 
		{
			{ 1, cPrivateRSA::generateKeyAndCreateAlgorithm },
			{ 2, cPrivateRSA::generateKeyAndCreateAlgorithm },
			{ 3, cPrivateRSA::generateKeyAndCreateAlgorithm }

//			{ 16, cPrivateElgamal::generateKeyAndCreateAlgorithmFunc },
//			{ 17, cPrivateDSA::generateKeyAndCreateAlgorithmFunc },
//			{ 20, cPrivateElgamal::generateKeyAndCreateAlgorithmFunc }
		};

int pka_createAlgorithm(
	HPKA *hPka,
	int algID,				// Identificator of public key algorithm
	BYTE *keyBuff,			// Public key MPI buffer,
	int keyBuffSize
	)
{
	for ( int q = sizeof( algPublicTbl ) / sizeof( sPublicAlgTabItem ); q--; )
	{
		if ( algPublicTbl[q].algID == algID )
		{	*hPka = algPublicTbl[q].createFunc( keyBuff, keyBuffSize );
			if ( *hPka == NULL )
				return pkae_invalidInputData;

			(*hPka)->id = algID;
			return pkae_ok;
		}
	}

	return pkae_invalidAlgorithmID;
}

int pka_getModuleSize( class cPKA *pka )
{
	return pka->getModuleSize();
}

int ska_getModuleSize( class cPKA *pka )
{
	return pka->getModuleSize();
}

int pka_getBlockSize( cPKA *pka )
{
	return pka->getBlockSize();
}
int ska_getBlockeSize( cPKA *pka )
{
	return pka->getBlockSize();
}

void pka_free( class cPKA *pka )
{
	delete pka;
}

void ska_free( class cPKA *pka )
{
	delete pka;
}

int pka_encrypt(
	class cPKA *pka,	// Public key algorithm
	BYTE *encBuff,		// OUT - MPI buffer of encrypted message
	int  *encBuffSize,	// IN  - pointer to size of allocated buffer.
						// OUT - pointer to required or busied size.
	BYTE *messageBuff,	// IN  - formatted message buffer. 
						// This value is padded as described in PKCS-1
	int messageBuffSize	// Size of the message buffer to be encrypted. 
						// It has to be equal to module size in bytes 
)
{
	return pka->encrypt( encBuff, encBuffSize, messageBuff, messageBuffSize );
}

int pka_encrypt_X509(
	class cPKA *pka,	// Public key algorithm
	BYTE *encBuff,		// OUT - ASN.1 buffer of encrypted message
	int  *encBuffSize,	// IN  - pointer to size of allocated buffer.
						// OUT - pointer to required or busied size.
	BYTE *messageBuff,	// IN  - formatted message buffer. 
						// This value is padded as described in PKCS-1
	int messageBuffSize	// Size of the message buffer to be encrypted. 
						// It has to be equal to module size in bytes 
)
{
	return pka->encrypt_X509( encBuff, encBuffSize, messageBuff, messageBuffSize );
}

//
//	Function:
//		pka_verify
//	Verifies signed MPI buffer with digest..
//
int pka_verify(
	class cPKA *pka,	// Public key algorithm
	BYTE *sigBuff,		// IN - signature MPI buffer 
	int  sigBuffSize,	// IN  - size of allocated buffer.
	BYTE *digestBuff,	// IN  - formatted digest buffer. 
	int digestBuffSize	// Size of the formatted digest buffer to be verified with signature. 
)
{
	return pka->verify( sigBuff, sigBuffSize, digestBuff, digestBuffSize );
}

int pka_verify_X509(
	class cPKA *pka,	// Public key algorithm
	BYTE *sigBuff,		// IN - signature ASN.1 buffer 
	int  sigBuffSize,	// IN  - size of allocated buffer.
	BYTE *digestBuff,	// IN  - formatted digest buffer. 
	int digestBuffSize	// Size of the formatted digest buffer to be verified with signature. 
)
{
	return pka->verify_X509( sigBuff, sigBuffSize, digestBuff, digestBuffSize );
}

int pka_decrypt_X509(
	class cPKA *pka,	// Public key algorithm
	BYTE *decBuff,		// OUT decrypted buffer. 
	int *pDecBuffSize, 	// OUT required size of decrypted buffer. size of allocated buffer
	const BYTE *encBuff,// IN - encrypted ASN.1 buffer 
	int  encBuffSize	// IN - size of encrypted buffer.
)
{	return pka->decrypt_X509( encBuff, encBuffSize, decBuff, pDecBuffSize ); }

int pka_decrypt(
	class cPKA *pka,	// Public key algorithm
	BYTE *decBuff,		// OUT decrypted buffer. 
	int *pDecBuffSize, 	// OUT required size of decrypted buffer. size of allocated buffer
	const BYTE *encBuff,// IN - signature MPI buffer 
	int  encBuffSize	// IN  - size of allocated buffer.
)
{	return pka->decrypt( encBuff, encBuffSize, decBuff, pDecBuffSize ); }

int ska_getAlgID( class cPKA *pSka ) { return (pSka)->getAlgID(); }
int pka_getAlgID( class cPKA *pPka ) { return (pPka)->getAlgID(); }

////////////////////////////////////////////////////
//
//	Private key funcs...
//

int ska_createAlgorithm(
	class cPKA **pSka,		// Pointer to handle of secret key algorithm
	int algID,				// Identificator of public key algorithm
	BYTE *mpiBuff,			// Public key MPI buffer,
	int mpiBuffSize,
	BYTE *privateMpiBuff,	// Private key MPI buffer,
	int privateMpiBuffSize
	)
{
	for ( int q = sizeof( algPrivateTbl ) / sizeof( sPrivateAlgTabItem ); q--; )
	{
		if ( algPrivateTbl[q].algID == algID )
		{
			*pSka = algPrivateTbl[q].createFunc( mpiBuff, mpiBuffSize, privateMpiBuff, privateMpiBuffSize );
			if ( *pSka != NULL )
				(*pSka)->id = algID;

			return pkae_ok;
		}
	}

	return pkae_invalidAlgorithmID;
}

int ska_decrypt(
	class cPKA *ska,	// Handle of secret key algorithm
	BYTE *decBuff,		// OUT - buffer for decrypted message
	int  *decBuffSize,	// IN  - pointer to size of allocated buffer. ( required size == module size in byte )
						// OUT - pointer to required or busied size.
	BYTE *encBuff,		// IN  - encrypted MPI buffer. 
	int encBuffSize
)
{
	return ska->decrypt( decBuff, decBuffSize, encBuff,encBuffSize );
}

int ska_decrypt_X509(
	class cPKA *ska,	// Handle of secret key algorithm
	BYTE *decBuff,		// OUT - buffer for decrypted message
	int  *decBuffSize,	// IN  - pointer to size of allocated buffer.
						// OUT - pointer to required or busied size.
	const BYTE *encBuff,// IN  - encrypted ASN.1 buffer. 
	int encBuffSize
)
{
	return ska->decrypt_X509( decBuff, decBuffSize, encBuff,encBuffSize );
}

int ska_sign(
	class cPKA *ska,	// Handle of secret key algorithm
	BYTE *sigBuff,		// OUT - signature MPI buffer 
	int  *sigBuffSize,	// IN  - pointer pointer size of allocated buffer.
						// OUT - ointer to required or really busied size.
	BYTE *digestBuff,	// IN  - formatted digest buffer to be signed. 
	int digestBuffSize	// Size of the formatted digest buffer. 
						// It has to be equal to module size in bytes 
)
{
	return ska->sign( sigBuff, sigBuffSize, digestBuff, digestBuffSize );
}

//
//	Generates signature to ASN.1 OCTET STRING ( PKCS-1 requirement )
//
int ska_sign_X509(
	class cPKA *ska,	// Handle of secret key algorithm
	BYTE *sigBuff,		// OUT - signature ASN.1 buffer 
	int  *pSigBuffSize,	// IN  - pointer pointer size of allocated buffer.
						// OUT - ointer to required or really busied size.
	BYTE *digestBuff,	// IN  - formatted digest buffer to be signed. 
	int digestBuffSize	// Size of the formatted digest buffer. 
						// It has to be equal to module size in bytes 
)
{
	return ska->sign_X509( sigBuff, pSigBuffSize, digestBuff, digestBuffSize );
}

//
//	Generates signature to ASN.1 BIT STRING ( RFC-2459 requirement )
//
int ska_sign_X509_BIT_STRING(
	class cPKA *ska,	// Handle of secret key algorithm
	BYTE *sigBuff,		// OUT - signature ASN.1 buffer 
	int  *pSigBuffSize,	// IN  - pointer pointer size of allocated buffer.
						// OUT - pointer to required or really busied size.
	BYTE *digestBuff,	// IN  - formatted digest buffer to be signed. 
	int digestBuffSize	// Size of the formatted digest buffer. 
						// It has to be equal to module size in bytes 
)
{
	BYTE *ptr = sigBuff;
	if ( ptr != NULL )
	{
		ptr++;
		(*pSigBuffSize)--;
	}

	int res = ska->sign_X509( ptr, pSigBuffSize, digestBuff, digestBuffSize );

	// Converts OCTET STRING to BIT STRING

	(*pSigBuffSize)++;

	if ( res != pkae_ok )
		return res;

	x509_struct os;
	res = os.parse( ptr, *pSigBuffSize - 1, NULL, NULL, asn1tag_OCTET_STRING );
	if ( res == x509e_ok )
	{
		// RSA octet string will be converted to bit string

		x509_bitString bs;
		if ( bs.prepareForWritting( os.L ) != *pSigBuffSize )
			return x509e_internalError;

		if ( bs.writeHeader( sigBuff, *pSigBuffSize ) == 0 )
			return x509e_internalError;

		return x509e_ok;
	}

	// DSA signature packed as sequence
	x509_struct seq;
	res = seq.parse( ptr, *pSigBuffSize - 1, NULL, NULL, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res == x509e_ok )
	{
		x509_bitString bs;
		if ( bs.prepareForWritting( seq.L ) != *pSigBuffSize )
			return x509e_internalError;

		if ( bs.writeHeader( sigBuff, *pSigBuffSize ) == 0 )
			return x509e_internalError;

		return x509e_ok;
	}

	return res;

}
///////////////////////////////////////////////////////
//
//	4. Key pairs generation function...
//		Note, random buffer has to be assigned before,
//		use the pka_setRandomBuffer function.
//

bool cPrivateRSA::m_generateSafePrime = true;

void ska_RSA_generateSafePrime( bool safe )
{	cPrivateRSA::m_generateSafePrime = safe;
}

int ska_generateKeyAndCreateAlgorithm(
	class cPKA **pSka,			// Pointer to handle of secret key algorithm
	int  algID,					// Identificator of public key algorithm
	int	 keyLength,				// Key size in byte
	cProgressActor *actor
	)
{
	for ( int q = sizeof( generateKeyTbl ) / sizeof( sGenerateKeyTabItem ); q--; )
	{
		if ( generateKeyTbl[q].algID == algID )
		{
			int res = generateKeyTbl[q].generateKeyAndCreateAlgorithmFunc( pSka, keyLength, actor );

			if ( *pSka != NULL )
				(*pSka)->id = algID;

			return res;
		}
	}

	return pkae_invalidAlgorithmID;
}

int ska_derivePublicKeyAlgorithm(
	HPKA *pHPka,	// OUT - pointer to the derived public key algorithm.
	class cPKA *ska		// IN  - secret key algorithm.
	)
{
	if ( ska == NULL )
		return pkae_internalError;

	int res = ska->derivePublicKeyAlgorithm( pHPka );

	if ( res != pkae_ok )
		return res;

	if ( *pHPka != NULL )
	{
		(*pHPka)->id = ska->getAlgID();
		(*pHPka)->time = ska->time;
		(*pHPka)->days = ska->days;
	}

	return pkae_ok;
}

int pka_writeKeyToMpiBuffer(
	HPKA hPka,			// Handler of public key algorithm
						// The handler can be derived from secret key algorithm
						// See ska_derivePublicKeyAlgorithm.

	BYTE *mpiBuff,		// Public key MPI buffer,

	int  *pMpiBuffSize	// IN  - pointer to size of MPI buffer,
						// OUT - pointer to required or busied size.
						// the value is filled even if the error 
						// pkae_bufferSizeTooSmall returned
	)
{
	return hPka->writeKeyToMpiBuffer( mpiBuff, pMpiBuffSize );
}

int ska_writeKeyToMpiBuffer(
	HPKA hSka,			// Handler of secret key algorithm
						// The handler can be generated by
						// ska_generateKeyAndCreateAlgorithm.

	BYTE *mpiBuff,		// Secret key MPI buffer,

	int  *pMpiBuffSize	// IN  - pointer to size of MPI buffer,
						// OUT - pointer to required or busied size.
						// the value is filled even if the error 
						// pkae_bufferSizeTooSmall returned
	)
{
	return hSka->writeKeyToMpiBuffer( mpiBuff, pMpiBuffSize );
}

int pka_createAlgorithm_X509( HPKA *pPka, x509_publicKeyInfo *pki )
{
	int algID = pki->algOID.tu->pAlg;

	for ( int q = sizeof( algPublicTbl ) / sizeof( sPublicAlgTabItem ); q--; )
	{
		if ( algPublicTbl[q].algID == algID )
		{
			return algPublicTbl[q].createX509Func( pPka, pki );
		}
	}

	return pkae_invalidAlgorithmID;
}

int ska_createAlgorithm_X509( HPKA *hSka, x509_privateKeyInfo *pki )
{
	int algID = pki->algOID.tu->pAlg;

	for ( int q = sizeof( algPrivateTbl ) / sizeof( sPrivateAlgTabItem ); q--; )
	{
		if ( algPrivateTbl[q].algID == algID )
		{
			return algPrivateTbl[q].createX509Func( hSka, pki );
		}
	}

	return pkae_invalidAlgorithmID;
}

int pka_writeKeyToASN1Buffer( HPKA hPka, BYTE *buff, int *pBuffSize )
{
	return hPka->writeKeyToASN1Buffer( buff, pBuffSize );
}

int ska_writeKeyToASN1Buffer( HPKA hSka, BYTE *buff, int *pBuffSize )
{
	return hSka->writeKeyToASN1Buffer( buff, pBuffSize );
}

//
//	pka_generateLocalKeyID
//	1. Makes ASN1 buffer with public key
//	2. Generates SHA1 digest of the buffer t output buffer
//
int pka_generateLocalKeyID(
		HPKA hPka,		// Public key algorithm.
		BYTE *buff,		// Output buffer for local key ID 
		int *pBuffSize,	// Buffer size required
		int hashID
		)
{
	 if ( hashID == 0 )
		hashID = pgphaSHA1;

	int digetSize = 0;
	int res = ha_getDigestSize( &digetSize, hashID );
	if ( res != hae_ok )
		return res;

	if ( *pBuffSize < digetSize )
	{
		*pBuffSize = digetSize;
		return pkae_bufferSizeTooSmall;
	}

	*pBuffSize = digetSize;

	int size = 0;

	res = pka_writeKeyToASN1Buffer( hPka, NULL, &size );
	if ( res != RSA_INVALID_BLOCK_SIZE )
		return res;

	BYTE *pkBuff = (BYTE *)SMalloc( size );

	if ( pkBuff == NULL )
		return pkae_memoryAllocation;

	res = pka_writeKeyToASN1Buffer( hPka, pkBuff, &size );

	if ( res == pkae_ok )
		res = ha_makeDigestEx( hashID, buff, pkBuff, size );

	SFree( pkBuff );

	return res;
}

//
//	ska_generateLocalKeyID
//	1. Derives public key from the secret key
//	2. Uses pka_generateLocalKeyID to generate local key ID to output buffer
//	3. Frees the derived public key.
//
int ska_generateLocalKeyID(
		HPKA hSka,		// Secret key algorithm.
		BYTE *buff,		// Buffer for local key ID 
		int *pBuffSize	// Buffer size required
		)
{
	HPKA hPka;
	int res = ska_derivePublicKeyAlgorithm( &hPka, hSka );

	if ( res != pkae_ok )
		return res;

	res = pka_generateLocalKeyID( hPka, buff, pBuffSize );

	pka_free( hPka );

	return res;
}

