#include "shamir.h"
#include "..\pkcipher\pkcommon.h"
#include "fsrandom.h"

#include "fscmemory.h"

FSCRYPTRESULT SSSAllocateShares( SSShares **s , size_t number)
{
	*s = FSCMalloc( sizeof(SSShares) );
	if ( *s == NULL ) return FSCRYPT_MALLOC_ERROR;
	
	(*s)->array = FSCMalloc( sizeof(intPair) * number );
	if ( (*s)->array == NULL )
	{	FSCFree( *s );
		return FSCRYPT_MALLOC_ERROR;
	}

	(*s)->arraySize = number;

	mpz_init( (*s)->prime );

	while( number--)
	{
		mpz_init((*s)->array[number].x);
		mpz_init((*s)->array[number].y);
	}
	
	if (FSCMemoryError)
		return FSCRYPT_MALLOC_ERROR;

	return FSCRYPT_OK;
}

FSCRYPTRESULT SSSFree( SSShares ** s)
{
	int number = (*s)->arraySize;

	while( number-- )
	{
		mpz_clear((*s)->array[number].x);
		mpz_clear((*s)->array[number].y);
	}

	mpz_clear( (*s)->prime );

	FSCFree( (*s)->array );

	FSCFree( (*s) );

	*s = NULL;

	return FSCRYPT_OK;
}

static bool generateRandom( mpz_t r , size_t size , FSRandomNumberGenerator *rng )
{
	byte *buff;

	buff = FSCMalloc( size );
	if ( buff == NULL ) return FALSE;

	if (FSCryptGetRandomBytes( rng , size , buff ) != FSCRYPT_OK)
	{
		FSCFree( buff );
		return FALSE;
	}

	bytes2mpzint( r , buff , size );

	FSCFree( buff );

	return TRUE;
}

static bool generateRandomPrime( mpz_t p , size_t size , FSRandomNumberGenerator * rng)
{
	bool isPrime;

	if (! generateRandom( p , size , rng )) return FALSE;

	mpz_setbit( p , 0 );

	while( TRUE )
	{
		mpz_setbit( p , size * 8 - 1 );
		mpz_clrbit( p , size * 8 );

		isPrime = mpz_probab_prime_p ( p , 25 );
		if ( isPrime ) break;

		mpz_add_ui( p , p , 2 );
	}

	return TRUE;
}

typedef struct {
	size_t size;
	mpz_t *array;
} sPolinom;


static bool generatePolinom( sPolinom *P , size_t size , byte *key , size_t keySize, FSRandomNumberGenerator *rng )
{
	P->size = size;

	P->array = FSCMalloc( sizeof( mpz_t ) * size );
	if ( P->array == NULL ) return FALSE;

	while( --size )
	{	
		mpz_init( P->array[size] );
		if (! generateRandom( P->array[size] , keySize , rng )) return FALSE;
	}

	mpz_init( P->array[0]);
	bytes2mpzint( P->array[0] , key , keySize );
}

static void freePolinom( sPolinom *P )
{
	int i = P->size;
	while( i-- )
		mpz_clear( P->array[i] );

	FSCFree( P->array );
}

/* Check with 0 and with previous X */
static bool notEqualToPrevious( SSShares *s , int num )
{
	int i;

	if (mpz_cmp_ui( s->array[num].x, 0 ) == 0) return FALSE;

	for( i = num; i--; )
		if (mpz_cmp_ui( s->array[num].x, s->array[i].x) == 0 ) return FALSE;

	return TRUE;
}

static bool prepareRandomX( SSShares *s , size_t size , FSRandomNumberGenerator *rng )
{
	size_t i;

	for ( i = 0 ; i < s->arraySize; i++ )
	{
		generateRandom( s->array[i].x , size , rng );
		while( TRUE )
		{
			if ( notEqualToPrevious( s , i ))
				break;

			mpz_add_ui( s->array[i].x , s->array[i].x, 1 );
		}
	}

	return TRUE;
}

static bool fillShares( SSShares *s , sPolinom *P )
{
	int i,n;
	mpz_t t;
	bool res = TRUE;

	mpz_init( t );

	for( i = s->arraySize; i-- ; )
	{
		if (!notEqualToPrevious( s , i ))
			res = FALSE;

		mpz_set( s->array[i].y , P->array[0] );

		for( n = P->size; --n; )
		{
			mpz_powm_ui( t , s->array[i].x , n , s->prime );
			mpz_mul( t , t , P->array[n] );
			mpz_mod( t , t, s->prime );
			mpz_add( s->array[i].y , s->array[i].y , t );
			mpz_mod( s->array[i].y , s->array[i].y , s->prime );
		}
	}

	mpz_clear( t );

	return res;
}

/*
	Function : SSSGenerateShares
	1) Generate prime p more the key and total shares number;
	2) Generate (recNum - 1) degree polinom P(x) = Summ( Ai*(x exp i)) , where 0 <= i <= t-1,
	3) Fill shares array by Xi and Yi where P(Xi) == Yi;
 */

FSCRYPTRESULT SSSGenerateShares(
	SSShares *s,		/* Previously allocated shares */
	size_t recNumber ,	/* Required number of participants those can compute key */
	byte *key,			/* Byte buffer of key to be shared */
	size_t keySize		/* size of key */
	)
{
	FSCRYPTRESULT res;
	FSRandomNumberGenerator *rng;
	sPolinom P;
	char *rngName;

/*	1) Generate prime p more the key and total shares number; */

	/*	Random generator allocation */
	rngName = "SHA1_RANDOM";
	res = FSCryptInitRandoms();
	if ( res != FSCRYPT_OK ) return res;

	res = FSCryptRequestRandom( rngName, &rng);
	if ( res != FSCRYPT_OK ) return res;

	FSCryptRandomInit( rng , NULL );
	FSCryptRandomizeFnc( rng , keySize , key );


	generateRandomPrime( s->prime , keySize + 1 , rng ); /* key must be < prime */


/*	2) Generate (recNum - 1) degree polinom P(x);
	P(x) = Summ( Ai*(x exp i)) , where 0 <= i <= recNum-1,*/

	generatePolinom( &P , recNumber , key, keySize , rng );

/*	3) Fill shares array by Xi and Yi where P(Xi) == Yi; */

	if (!fillShares( s , &P ))
		res = FSCRYPT_INIT_ERROR;

	freePolinom( &P );

	/* Random free */

	FSCryptRandomFinal( rng );

	FSCryptReleaseRandom(&rng);

	FSCryptFreeRandoms();

	return res;
}


static bool prepareB( mpz_t Bj , SSShares *s , int j )
{
	int i;
	mpz_t xInv;

	mpz_init( xInv );

	mpz_set_ui( Bj , 1 );

	for( i = s->arraySize; i--; )
	{
#ifdef _DEBUG
		char s1[512],s2[512];
		mpz_get_str( s1 , 16 , s->array[i].x );
		mpz_get_str( s2 , 16 , s->array[j].x );
#endif

		if ( j == i ) continue;

		mpz_sub( xInv , s->array[i].x , s->array[j].x );

		mpz_mod( xInv , xInv , s->prime ); /* can be ommited */

		if ( !makeInvert( xInv , xInv , s->prime ) )
			return FALSE;

		mpz_mul( Bj , Bj , s->array[i].x );
		mpz_mod( Bj , Bj , s->prime );
		mpz_mul( Bj , Bj , xInv );
		mpz_mod( Bj , Bj , s->prime );
	}

	mpz_clear( xInv );
	return TRUE;
}

/*
	Lagrange interpolation formula used for x == 0;
	pp.329-330 D.R. Stinson, CRYPTOGRAPHY Theory and Practice.
 */

FSCRYPTRESULT SSSRestoreKey( SSShares *s , byte *key , size_t size )
{
	size_t t;
	mpz_t K, B;

	t = mpz_sizeinbase( s->prime , 2 );

	if ( t != ( (size + 1) * 8))
		return FSCRYPT_INIT_ERROR;

	mpz_init( K );
	mpz_init( B );

	for( t = s->arraySize; t--; )
	{
		prepareB( B , s , t );
		mpz_mul( B , B , s->array[t].y );
		mpz_add( K , K , B );
		mpz_mod( K , K , s->prime );
	}
	
	mpzint2bytes( key , size , K );

	mpz_clear( K );
	mpz_clear( B );

	return FSCRYPT_OK;
}


