#include "shamir.h"
#include "../../SMemory/smemory.h"
#include "../../crtLibErrs.h"
#include "../../RNG/cRNG.h"

int SSSAllocateShares( SSShares **s , size_t number)
{
	*s = (SSShares *)SMalloc( sizeof(SSShares), true );
	if ( *s == NULL ) return RSA_MALLOC_ERROR;

	(*s)->array = (intPair *)SMalloc( sizeof(intPair) * number, true );
	if ( (*s)->array == NULL )
	{	SFree( *s );
		return RSA_MALLOC_ERROR;
	}

	(*s)->arraySize = number;
	(*s)->curPos = 0;

	return RSA_OK;
}

int SSSFree( SSShares ** s)
{
	int number = (*s)->arraySize;

	(*s)->prime.freeData();

	while( number--)
	{
		(*s)->array[number].x.freeData();
		(*s)->array[number].y.freeData();
	}

	SFree( (*s)->array );

	SFree( (*s) );

	*s = NULL;

	return RSA_OK;
}

static bool generateRandom( pw_int *r , size_t size , cSimpleRng *rng )
{
	BYTE *buff = rng->get( size );

	if ( buff == NULL )
		return false;

	pw_bytes2int( r, buff, size );

	return TRUE;
}

static bool generateRandomPrime( pw_int *p , size_t size , cSimpleRng * rng, cProgressBlock *progress = NULL )
{
	cProgressBlock emptyProgress;

	if ( progress == NULL )
		progress = &emptyProgress;

	if ( !generateRandom( p , size , rng ) )
		return false;

	pw_setBit( p , 0 );

	progress->SetText( "Fill prime table" );

	pw_primeTableInit( size * 8 );

	progress->SetText( "Generate random prime" );

	bool cancelled = false;

	while( !cancelled )
	{
		pw_setBit( p , size * 8 - 1 );
		pw_clearBit( p , size * 8 );

		int count = 0;

		if ( pw_probab_prime( p , 25, &count )  )
		{
			progress->SetCurrentStepResult( cProgressBlock::stepResult_True );
			break;
		}

		pw_addu( p , p , 2 );

		cancelled = 
			!progress->
				SetCurrentStepResult( 
					count < 0 ? 
						( cProgressBlock::stepResult_False_ByTable ) :
						( cProgressBlock::stepResult_False_ByPrimeTest )
					);
	}

	pw_primeTableFree();

	return !cancelled;
}

typedef struct {
	size_t size;
	pw_int *array;
} sPolinom;


static bool generatePolinom( sPolinom *P , size_t size , BYTE *key , size_t keySize, cSimpleRng *rng )
{
	P->size = size;

	P->array = ( pw_int * )SMalloc( sizeof( pw_int ) * size, true );
	if ( P->array == NULL ) return false;

	// Set random alder koeffs only, Xo koeffs is the key see below.
	while( --size )
	{
//		mpz_init( P->array[size] );
		if ( !generateRandom( &P->array[size] , keySize , rng )) return false;
	}

//	mpz_init( P->array[0]);
	pw_bytes2int( &P->array[0] , key , keySize );

	return true;
}

static void freePolinom( sPolinom *P )
{
	int i = P->size;
	while( i-- )
		P->array[i].freeData();

	SFree( P->array );
}

// Check with 0 and with previous X
static bool notEqualToPrevious( SSShares *s , int num )
{
	int i;

	if ( pw_cmpu( &s->array[num].x, 0 ) == 0) return FALSE;

	for( i = num; i--; )
		if ( pw_cmp( &s->array[num].x, &s->array[i].x) == 0 ) return FALSE;

	return TRUE;
}

bool prepareOrderedX( SSShares *s )
{
	if ( s == NULL )
		return false;

	for ( size_t i = 0 ; i < s->arraySize; i++ )
		s->array[i].x.setu( i + 1 );

	return true;
}

bool prepareRandomX( SSShares *s , size_t size , cSimpleRng *rng )
{
	size_t i;

	for ( i = 0 ; i < s->arraySize; i++ )
	{
		generateRandom( &s->array[i].x , size , rng );
		for(;;)
		{
			if ( notEqualToPrevious( s , i ))
				break;

			pw_addu( &s->array[i].x , &s->array[i].x, 1 );
		}
	}

	return TRUE;
}

static bool fillShares( SSShares *s , sPolinom *P )
{
	int i,n;
	pw_int t;
	bool res = true;

	for( i = s->arraySize; i-- ; )
	{
		if (!notEqualToPrevious( s , i ))
			res = false;

		pw_set( &s->array[i].y , &P->array[0] );

		for( n = P->size; --n; )
		{
			pw_powmu( &t , &s->array[i].x , n , &s->prime );
			pw_mulm( &t , &t , &P->array[n], &s->prime );
			pw_addm( &s->array[i].y , &s->array[i].y , &t, &s->prime );
		}
	}

	return res;
}

//
//	Function : SSSGenerateShares
//	1) Generate prime p more the key and total shares number;
//	2) Generate (reqNumber - 1) degree polinom P(x) = Summ( Ai*(x exp i)) , where 0 <= i <= t-1,
//	3) Fill shares array by Xi and Yi where P(Xi) == Yi;
//

int SSSGenerateShares(
	SSShares *s,		// Previously allocated shares
	size_t reqNumber,	// Required number of participants those can compute key

	BYTE *key,			// BYTE buffer of key to be shared
	size_t keySize,		// size of key

	BYTE *seed,			// Seed buffer for random generator
	int *pSeedSize,		// if seed buffer is to small, it returns ppe_randomBufferRequired
						// and fills required *pSeedSize value
	cProgressBlock *progress
	)
{
	cProgressBlock emptyProgress;

	if ( progress == NULL )
		progress = &emptyProgress;

	int res = ppe_ok;
	cSimpleRng rng;
	sPolinom P;

	// the values may be simple non zero order number of share ( see prepareOrderedX )
	bool X_valuesIsAssigned = ( s->array[0].x.data != NULL ) && ( pw_cmpu( &s->array[0].x, 0 ) != 0 );

// Checks size of seed buffer
	int requeredRandomBufferSize = 
			keySize + 1	+					// Prime size ( one byte more then key )
			( reqNumber - 1 ) * keySize;		// Polinom koeffs 

	if ( !X_valuesIsAssigned )
		requeredRandomBufferSize +=
			s->arraySize * keySize;			// Shares X

	if ( requeredRandomBufferSize * 2 > *pSeedSize )
	{
		*pSeedSize = requeredRandomBufferSize * 2;
		return ppe_randomBufferRequired;
	}

	progress->SetText( "Generate random pool" );
//	Random generator initialisation
	if ( !rng.generatePool( requeredRandomBufferSize, seed, *pSeedSize ) )
		return ppe_internalError;

//	1) Generate prime p more the key and total shares number; 
	if ( !generateRandomPrime( &s->prime , keySize + 1 , &rng, progress ) ) // key must be < prime
		return ppe_internalError;

	progress->SetText( "Generate polinom" );
//	2) Generate (reqNumber - 1) degree polinom P(x);
//	P(x) = Summ( Ai*(x exp i)) , where 0 <= i <= reqNumber - 1,
	if ( !generatePolinom( &P , reqNumber , key, keySize , &rng ) )
		return ppe_internalError;

	progress->SetText( "Generate X values" );
//	3) Fill shares array by Xi and Yi where P(Xi) == Yi;
	if ( !X_valuesIsAssigned )
		prepareRandomX( s, keySize, &rng );

	progress->SetText( "Generate shares" );
	if (!fillShares( s , &P ))
		res = ppe_internalError;

	freePolinom( &P );

	return res;
}

static bool prepareB( pw_int * Bj , SSShares *s , int j )
{
	int i;
	pw_int xInv;

	Bj->setu( 1 );

	for( i = s->arraySize; i--; )
	{
		if ( j == i ) continue;

		pw_subm( &xInv , &s->array[i].x , &s->array[j].x, &s->prime );

		if ( pw_invert( &xInv , &xInv , &s->prime ) != pw_OK )
			return false;

		pw_mulm( Bj , Bj , &s->array[i].x, &s->prime );
		pw_mulm( Bj , Bj , &xInv, &s->prime );
	}

	return true;
}

//
//	SSSAddShare
//		Adds decrypted share info to share's array
//
int SSSAddShare( SSShares *s, pw_int *x, pw_int *y )
{
	if ( s->curPos >= s->arraySize )
		return pw_OK;

	for( size_t i = 0; i < s->curPos; i++ )
	{
		if ( pw_cmp( &s->array[ i ].x, x ) == 0 )
			return pw_InvalidKeyBlock;
	}

	int res = s->array[ s->curPos ].x.set( x );
	if ( res != pw_OK )
		return pw_InternalError;

	if ( s->array[ s->curPos ].y.set( y ) != pw_OK )
		return pw_InternalError;

	if (  ++(s->curPos) >= s->arraySize )
		return pw_OK;

	return pw_BufferSizeTooSmall;
}

//
//	Lagrange interpolation formula used for x == 0;
//	pp.329-330 D.R. Stinson, CRYPTOGRAPHY Theory and Practice.
//
int SSSRestoreKey( SSShares *s , BYTE *key , size_t size )
{
	size_t t;
	pw_int K, B;

	K.setu( 0 );

	t = pw_getLastBit( &s->prime );

	if ( t != ( (size + 1) * 8) )
		return ppe_internalError;

	for( t = s->arraySize; t--; )
	{
		if ( !prepareB( &B , s , t ) )
			return ppe_internalError;

		pw_mul( &B , &B , &s->array[t].y );
		pw_addm( &K , &K , &B, &s->prime );
	}

	pw_int2bytes( key , size , &K );

	return ppe_ok;
}


