
// I retain copyright in this code but I encourage its free use provided
// that I don't carry any responsibility for the results. I am especially 
// happy to see it used in free and open source software. If you do use 
// it I would appreciate an acknowledgement of its origin in the code or
// the product that results and I would also appreciate knowing a liitle
// about the use to which it is being put.
//
// Dr B. R. Gladman <brg@gladman.uk.net> 23rd January 2001.

// This is an implementation of the AES encryption algorithm (Rijndael)
// designed by Joan Daemen and Vincent Rijmen. This version is designed
// to provide both fixed and dynamic block and key lengths and can also 
// run with either big or little endian internal byte order (see aes.h). 
// It inputs block and key lengths in bytes with the legal values being 
// 16, 24 and 32.

// CONFIGURATION OPTIONS (see also aes.h)
// 1. define COMPACT for a compact but slow version of the cipher (the 
//    other options below do not apply to the compact version)
// 2. define UNROLL for loop unrolling in encrypt and decrypt
// 3. define FIXED_TABLES for static tables built within the executable
//    image - otherwise tables are built dynamically on first use
// 4. define FOUR_TABLES for high speed but increased memory use (uses four
//    tables in place of one to reduce inner loop instruction counts)
// 5. define LAST_ROUND_TABLES for ultimate speed but still more memory use
//    (this uses tables for the last round)
// 6. define SMALL_PTABLE for slower field multiply but smaller table

//#define COMPACT
#define UNROLL
#define FIXED_TABLES
#define FOUR_TABLES
#define LAST_ROUND_TABLES
//#define SMALL_PTABLE

// End of configuration options

#include "AES.H"

#if !defined(AES_DLL) || !defined(TEST)

// the finite field modular polynomial and elements

#define ff_poly 0x011b
#define ff_hi   0x80
#define ff_02   0x02
#define ff_03   0x03
#define ff_09   0x09
#define ff_0b   0x0b
#define ff_0d   0x0d
#define ff_0e   0x0e

// finite field multiply of a and b

#if defined(SMALL_PTABLE)
#define FFmul(a, b) (a && b ? \
        FFpow[(u = FFlog[a], v = FFlog[b] + u, v + (v < u ? 1 : 0))] : 0)
#else
#define FFmul(a, b) (a && b ? FFpow[FFlog[a] + FFlog[b]] : 0)
#endif

// multiply four bytes in GF(2^8) by 'x' {02} in parallel

#define m1  0x80808080
#define m2  0x7f7f7f7f
#define m3  0x1b1b1b1b
#define FFmulX(x) (u = (x) & m1, ((x) & m2) << 1) ^ ((u - (u >> 7)) & m3)

// perform column mix operation on four bytes in parallel

#define mix_col(x) (f2 = FFmulX(x), f2 ^ rot3(x ^ f2) ^ rot2(x) ^ rot1(x))

// perform inverse column mix operation on four bytes in parallel

#define inv_mix_col(x)                                                      \
    (f9 = (x),f2 = FFmulX(f9), f4 = FFmulX(f2), f8 = FFmulX(f4), f9 ^= f8,  \
    f2 ^ f4 ^ f8 ^ rot3(f2 ^ f9) ^ rot2(f4 ^ f9) ^ rot1(f9))

#if defined(FIXED_TABLES) || defined(COMPACT)

#include "aes_tab.h"

#else

#if defined(SMALL_PTABLE)
static byte  FFpow[256];        // powers of generator (0x03) in GF(2^8)
#else
static byte  FFpow[512];
#endif
static byte  FFlog[256];        // log: map element to power of generator

static byte  s_box[256];        // the S box
static byte  inv_s_box[256];    // the inverse S box
static word  rcon_tab[30];      // table of round constants (can be reduced
                                // to a length of 10 for 128-bit blocks)
#if defined(FOUR_TABLES)
static word  ft_tab[4][256];
static word  it_tab[4][256];
#else
static word  ft_tab[256];
static word  it_tab[256];
#endif

#if defined(LAST_ROUND_TABLES)
#if defined(FOUR_TABLES)
    static word  fl_tab[4][256];
    static word  il_tab[4][256];
#else
    static word  fl_tab[256];
    static word  il_tab[256];
#endif
#endif

static void gen_tabs(void)
{   word  i, w;
    byte  b;
#if defined(SMALL_PTABLE)
    byte  u, v;
#endif

    // log and power tables for GF(2^8) finite field with
    // 0x011b as modular polynomial - the simplest primitive
    // root is 0x03, used here to generate the tables

    i = 0; w = 1; 
    do
    {   
        FFpow[i] = (byte)w;
#if !defined(SMALL_PTABLE)
        FFpow[i + 255] = (byte)w;
#endif
        FFlog[w] = (byte)i++;
        w ^=  (w << 1) ^ (w & ff_hi ? ff_poly : 0);
    }
    while (w != 1);

    FFpow[255] = 1;

    for(i = 0, w = 1; i < 28; ++i)
    {
        rcon_tab[i] = bytes2word(w, 0, 0, 0);
        w = (w << 1) ^ (w & ff_hi ? ff_poly : 0);
    }

    for(i = w = 0; i < 256; ++i)
    {
        w ^= (w << 1) ^ (w << 2) ^ (w << 3) ^ (w << 4);
        w ^= 0x63 ^ (w >> 8);
        s_box[i] = (byte)w;
        inv_s_box[(byte)w] = (byte)i;
        w = FFpow[255 - FFlog[i + 1]];
    }

    for(i = 0; i < 256; ++i)
    {
        b = s_box[i];
#if defined(LAST_ROUND_TABLES)
        w = bytes2word(b, 0, 0, 0);
#if defined(FOUR_TABLES)
        fl_tab[0][i] = w;
        fl_tab[1][i] = rot1(w);
        fl_tab[2][i] = rot2(w);
        fl_tab[3][i] = rot3(w);
#else
        fl_tab[i] = w;
#endif
#endif
        w = bytes2word(FFmul(ff_02, b), b, b, FFmul(ff_03, b));
#if defined(FOUR_TABLES)
        ft_tab[0][i] = w;
        ft_tab[1][i] = rot1(w);
        ft_tab[2][i] = rot2(w);
        ft_tab[3][i] = rot3(w);
#else
        ft_tab[i] = w;
#endif
        b = inv_s_box[i];
#if defined(LAST_ROUND_TABLES)
        w = bytes2word(b, 0, 0, 0);
#if defined(FOUR_TABLES)
        il_tab[0][i] = w;
        il_tab[1][i] = rot1(w);
        il_tab[2][i] = rot2(w);
        il_tab[3][i] = rot3(w);
#else
        il_tab[i] = w;
#endif
#endif
        w = bytes2word(FFmul(ff_0e, b), FFmul(ff_09, b),
                            FFmul(ff_0d, b), FFmul(ff_0b, b));
#if defined(FOUR_TABLES)
        it_tab[0][i] = w;
        it_tab[1][i] = rot1(w);
        it_tab[2][i] = rot2(w);
        it_tab[3][i] = rot3(w);
#else
        it_tab[i] = w;
#endif
    }
}

#endif  // !FIXED_TABLES && !COMPACT

#if defined(COMPACT) || !defined(LAST_ROUND_TABLES)

#define ls_box(x) bytes2word(           \
    s_box[byte0(x)], s_box[byte1(x)],   \
    s_box[byte2(x)], s_box[byte3(x)])

#elif defined(FOUR_TABLES)

#define ls_box(x)       \
 (  fl_tab[0][byte0(x)] \
  ^ fl_tab[1][byte1(x)] \
  ^ fl_tab[2][byte2(x)] \
  ^ fl_tab[3][byte3(x)] )

#else

#define ls_box(x)           \
 (  fl_tab[byte0(x)]        \
  ^ rot1(fl_tab[byte1(x)])  \
  ^ rot2(fl_tab[byte2(x)])  \
  ^ rot3(fl_tab[byte3(x)]) )

#endif

// Subroutne to set the block size (if variable) in bytes, legal
// values being 16, 24 and 32.

#if defined(BLOCK_SIZE_R)
#define n_col   (Ncol)
#define fr_sh(n,r)  (((n) + Shr##r) % n_col)
#define ir_sh(n,r)  (((n) + n_col - Shr##r) % n_col)
#else
#define n_col       (f_dat(cx,Ncol))
#define sh(x)       (f_dat(cx,Shr)[x])
#define fr_sh(n,r)  (f_dat(cx,shf##r)[n])
#define ir_sh(n,r)  (f_dat(cx,shr##r)[n])

#if defined(AES_IN_CPP)
rval aes::set_blk(const word n_bytes)
#else
cf_dec set_blk(const word n_bytes, aes *cx)     
#endif
{   word    i, j;

    if(n_bytes != 16 && n_bytes != 24 && n_bytes != 32) 
        
        return (n_bytes ? aes_bad : (rval)n_col);
    
#if !defined(FIXED_TABLES)
    if(!(f_dat(cx,mode) & 0x08)) { gen_tabs(); mode = 0x08; }
#endif

    f_dat(cx,mode) |= 0x04;

    n_col = n_bytes >> 2;
    
    for(i = 0; i < 8; ++i)
    { 
        j = 1;
        f_dat(cx,shf1)[i] = (byte)((i + j) % n_col);
        f_dat(cx,shr1)[i] = (byte)((n_col + i - j) % n_col);
        j = (n_col == 8 ? 3 : 2);
        f_dat(cx,shf2)[i] = (byte)((i + j) % n_col);
        f_dat(cx,shr2)[i] = (byte)((n_col + i - j) % n_col);
        ++j;
        f_dat(cx,shf3)[i] = (byte)((i + j) % n_col);
        f_dat(cx,shr3)[i] = (byte)((n_col + i - j) % n_col);
    }

    return aes_good;
}
#endif

// Initialise the key schedule from the user supplied key. The key
// length is now specified in bytes - 16, 24 or 32 as appropriate.
// This corresponds to bit lengths of 128, 192 and 256 bits, and
// to Nk values of 4, 6 and 8 respectively.

#define mx(t,f)     *t++ = inv_mix_col(*f++);

#if !defined(BLOCK_SIZE_R)

#define cpy(t,f)                        \
switch(n_col)                           \
{   case 8: *t++ = *f++; *t++ = *f++;   \
    case 6: *t++ = *f++; *t++ = *f++;   \
    case 4: *t++ = *f++; *t++ = *f++;   \
            *t++ = *f++; *t++ = *f++;   \
}

#define mix(t,f)                \
switch(n_col)                   \
{   case 8: mx(t,f); mx(t,f);   \
    case 6: mx(t,f); mx(t,f);   \
    case 4: mx(t,f); mx(t,f);   \
            mx(t,f); mx(t,f);   \
}

#elif   BLOCK_SIZE_R == 16
#define cpy(t,f)    *t++ = *f++; *t++ = *f++; *t++ = *f++; *t++ = *f++
#define mix(t,f)    {   mx(t,f); mx(t,f); mx(t,f); mx(t,f); }
#elif BLOCK_SIZE_R == 24
#define cpy(t,f)    *t++ = *f++; *t++ = *f++; *t++ = *f++; *t++ = *f++; \
                    *t++ = *f++; *t++ = *f++
#define mix(t,f)    {   mx(t,f); mx(t,f); mx(t,f); mx(t,f); \
                        mx(t,f); mx(t,f); }
#elif BLOCK_SIZE_R == 32
#define cpy(t,f)    *t++ = *f++; *t++ = *f++; *t++ = *f++; *t++ = *f++; \
                    *t++ = *f++; *t++ = *f++; *t++ = *f++; *t++ = *f++
#define mix(t,f)    {   mx(t,f); mx(t,f); mx(t,f); mx(t,f); \
                        mx(t,f); mx(t,f); mx(t,f); mx(t,f); }
#endif

#if defined(AES_IN_CPP)
rval aes::set_key(const byte in_key[], const word n_bytes, const enum aes_key f)
#else
cf_dec set_key(const byte in_key[], const word n_bytes, const enum aes_key f, aes *cx)
#endif
{   word    t, *k1, *kt, *rcp;

    if(n_bytes != 16 && n_bytes != 24 && n_bytes != 32) return aes_bad;

#if !defined(FIXED_TABLES)
    if(!(f_dat(cx,mode) & 0x08)) { gen_tabs(); mode = 0x08; }
#endif

#if !defined(BLOCK_SIZE_R)
#if defined(AES_IN_CPP)
    if(!(f_dat(cx,mode) & 0x04)) set_blk(16);
#else
    if(!(f_dat(cx,mode) & 0x04)) set_blk(16, cx);
#endif
#endif

    f_dat(cx,mode) |= (byte)f & 3;                  // encryption mode = enc, dec or both
    f_dat(cx,Nkey) = n_bytes >> 2;
    f_dat(cx,Nrnd) = (f_dat(cx,Nkey) > n_col ? f_dat(cx,Nkey) : n_col) + 6; 

    f_dat(cx,e_key)[0] = word_in(in_key     );
    f_dat(cx,e_key)[1] = word_in(in_key +  4);
    f_dat(cx,e_key)[2] = word_in(in_key +  8);
    f_dat(cx,e_key)[3] = word_in(in_key + 12);

    k1 = f_dat(cx,e_key); 
    kt = k1 + n_col * (f_dat(cx,Nrnd) + 1) - f_dat(cx,Nkey); 
    rcp = rcon_tab;

    switch(f_dat(cx,Nkey))
    {
    case 4: do
            {   t = rot3(k1[3]);
                k1[4] = k1[0] ^ ls_box(t) ^ *rcp++;
                k1[5] = k1[1] ^ k1[4];
                k1[6] = k1[2] ^ k1[5];
                k1[7] = k1[3] ^ k1[6];
                k1 += 4;
            }
            while(k1 < kt);
            break;

    case 6: f_dat(cx,e_key)[4] = word_in(in_key + 16);
            f_dat(cx,e_key)[5] = word_in(in_key + 20);
            do
            {   t = rot3(k1[5]);
                k1[ 6] = k1[0] ^ ls_box(t) ^ *rcp++;
                k1[ 7] = k1[1] ^ k1[ 6];
                k1[ 8] = k1[2] ^ k1[ 7];
                k1[ 9] = k1[3] ^ k1[ 8];
                k1[10] = k1[4] ^ k1[ 9];
                k1[11] = k1[5] ^ k1[10];
                k1 += 6;
            }
            while(k1 < kt);
            break;

    case 8: f_dat(cx,e_key)[4] = word_in(in_key + 16);
            f_dat(cx,e_key)[5] = word_in(in_key + 20);
            f_dat(cx,e_key)[6] = word_in(in_key + 24);
            f_dat(cx,e_key)[7] = word_in(in_key + 28);
            do
            {   t = rot3(k1[7]);
                k1[ 8] = k1[0] ^ ls_box(t) ^ *rcp++;
                k1[ 9] = k1[1] ^ k1[ 8];
                k1[10] = k1[2] ^ k1[ 9];
                k1[11] = k1[3] ^ k1[10];
                k1[12] = k1[4] ^ ls_box(k1[11]);
                k1[13] = k1[5] ^ k1[12];
                k1[14] = k1[6] ^ k1[13];
                k1[15] = k1[7] ^ k1[14];
                k1 += 8;
            }
            while (k1 < kt);
            break;
    }

#if !defined(COMPACT)

    if(f_dat(cx,mode) != enc)
    {   word    i, u, f2, f4, f8, f9;
        
        k1 = f_dat(cx,d_key); kt = f_dat(cx,e_key);
        cpy(k1, kt);
        for(i = 1; i < f_dat(cx,Nrnd); ++i) 
            mix(k1, kt);
        cpy(k1, kt);
    }

#endif

    return aes_good;
}

#if defined(COMPACT)

#define sbx_row(bi,n) bytes2word    (   \
    s_box[byte0(bi[n])],                \
    s_box[byte1(bi[fr_sh(n,1)])],       \
    s_box[byte2(bi[fr_sh(n,2)])],       \
    s_box[byte3(bi[fr_sh(n,3)])])

#define inv_sbx_row(bi,n)  bytes2word(  \
    inv_s_box[byte0(bi[n])],            \
    inv_s_box[byte1(bi[ir_sh(n,1)])],   \
    inv_s_box[byte2(bi[ir_sh(n,2)])],   \
    inv_s_box[byte3(bi[ir_sh(n,3)])])

#else

#if defined(FOUR_TABLES)

#define f_rnd(x,n)                      \
  ( ft_tab[0][byte0(x[n])]              \
  ^ ft_tab[1][byte1(x[fr_sh(n,1)])] \
  ^ ft_tab[2][byte2(x[fr_sh(n,2)])] \
  ^ ft_tab[3][byte3(x[fr_sh(n,3)])] )

#define i_rnd(x,n)                      \
  ( it_tab[0][byte0(x[n])]              \
  ^ it_tab[1][byte1(x[ir_sh(n,1)])] \
  ^ it_tab[2][byte2(x[ir_sh(n,2)])] \
  ^ it_tab[3][byte3(x[ir_sh(n,3)])] )

#else

#define f_rnd(x,n)                      \
  ( ft_tab[byte0(x[n])]                 \
  ^ rot1(ft_tab[byte1(x[fr_sh(n,1)])])  \
  ^ rot2(ft_tab[byte2(x[fr_sh(n,2)])])  \
  ^ rot3(ft_tab[byte3(x[fr_sh(n,3)])])  )

#define i_rnd(x,n)                      \
  ( it_tab[byte0(x[n])]                 \
  ^ rot1(it_tab[byte1(x[ir_sh(n,1)])])  \
  ^ rot2(it_tab[byte2(x[ir_sh(n,2)])])  \
  ^ rot3(it_tab[byte3(x[ir_sh(n,3)])])  )

#endif

#if !defined(LAST_ROUND_TABLES)

#define lf_rnd(x,n)    bytes2word(  \
    s_box[byte0(x[n])],             \
    s_box[byte1(x[fr_sh(n,1)])],    \
    s_box[byte2(x[fr_sh(n,2)])],    \
    s_box[byte3(x[fr_sh(n,3)])])

#define li_rnd(x,n)    bytes2word(      \
    inv_s_box[byte0(x[n])],             \
    inv_s_box[byte1(x[ir_sh(n,1)])],    \
    inv_s_box[byte2(x[ir_sh(n,2)])],    \
    inv_s_box[byte3(x[ir_sh(n,3)])])

#elif defined(FOUR_TABLES)

#define lf_rnd(x,n)                     \
  ( fl_tab[0][byte0(x[n])]              \
  ^ fl_tab[1][byte1(x[fr_sh(n,1)])] \
  ^ fl_tab[2][byte2(x[fr_sh(n,2)])] \
  ^ fl_tab[3][byte3(x[fr_sh(n,3)])] )

#define li_rnd(x,n)                     \
  ( il_tab[0][byte0(x[n])]              \
  ^ il_tab[1][byte1(x[ir_sh(n,1)])] \
  ^ il_tab[2][byte2(x[ir_sh(n,2)])] \
  ^ il_tab[3][byte3(x[ir_sh(n,3)])] )

#else

#define lf_rnd(x,n)                     \
  ( fl_tab[byte0(x[n])]                 \
  ^ rot1(fl_tab[byte1(x[fr_sh(n,1)])])  \
  ^ rot2(fl_tab[byte2(x[fr_sh(n,2)])])  \
  ^ rot3(fl_tab[byte3(x[fr_sh(n,3)])])  )

#define li_rnd(x,n)                     \
  ( il_tab[byte0(x[n])]                 \
  ^ rot1(il_tab[byte1(x[ir_sh(n,1)])])  \
  ^ rot2(il_tab[byte2(x[ir_sh(n,2)])])  \
  ^ rot3(il_tab[byte3(x[ir_sh(n,3)])])  )

#endif
#endif // COMPACT

#define si(b,k,n)       b[n] = word_in(in_blk + 4 * n) ^ k[n]
#define so(b,n)         word_out(out_blk + 4 * n, b[n])
#define cf(bo,bi,k,n)   bo[n] = mix_col(sbx_row(bi, n)) ^ k[n]
#define ci(bo,bi,k,n)   bo[n] = inv_mix_col(inv_sbx_row(bi, n) ^ k[n])
#define lf(bo,bi,k,n)   bo[n] = sbx_row(bi, n) ^ k[n]
#define li(bo,bi,k,n)   bo[n] = inv_sbx_row(bi, n) ^ k[n]
#define fr(bo,bi,k,n)   bo[n] = f_rnd(bi,n) ^ (k)[n]
#define ir(bo,bi,k,n)   bo[n] = i_rnd(bi,n) ^ (k)[n]
#define fl(bo,bi,k,n)   bo[n] = lf_rnd(bi,n) ^ (k)[n]
#define il(bo,bi,k,n)   bo[n] = li_rnd(bi,n) ^ (k)[n]

#if !defined(BLOCK_SIZE_R)

#define state_in(b,k)               \
switch(n_col)                       \
{   case 8: si(b,k,7); si(b,k,6);   \
    case 6: si(b,k,5); si(b,k,4);   \
    case 4: si(b,k,3); si(b,k,2);   \
            si(b,k,1); si(b,k,0);   \
}

#define state_out(b)            \
switch(n_col)                   \
{   case 8: so(b,7); so(b,6);   \
    case 6: so(b,5); so(b,4);   \
    case 4: so(b,3); so(b,2);   \
            so(b,1); so(b,0);   \
}

#if defined(COMPACT)

#define cf_round(bo,bi,k)                   \
switch(n_col)                               \
{   case 8: cf(bo,bi,k,7); cf(bo,bi,k,6);   \
    case 6: cf(bo,bi,k,5); cf(bo,bi,k,4);   \
    case 4: cf(bo,bi,k,3); cf(bo,bi,k,2);   \
            cf(bo,bi,k,1); cf(bo,bi,k,0);   \
}

#define ci_round(bo,bi,k)                   \
switch(n_col)                               \
{   case 8: ci(bo,bi,k,7); ci(bo,bi,k,6);   \
    case 6: ci(bo,bi,k,5); ci(bo,bi,k,4);   \
    case 4: ci(bo,bi,k,3); ci(bo,bi,k,2);   \
            ci(bo,bi,k,1); ci(bo,bi,k,0);   \
}

#define lcf_round(bo,bi,k)                  \
switch(n_col)                               \
{   case 8: lf(bo,bi,k,7); lf(bo,bi,k,6);   \
    case 6: lf(bo,bi,k,5); lf(bo,bi,k,4);   \
    case 4: lf(bo,bi,k,3); lf(bo,bi,k,2);   \
            lf(bo,bi,k,1); lf(bo,bi,k,0);   \
}

#define lci_round(bo,bi,k)                  \
switch(n_col)                               \
{   case 8: li(bo,bi,k,7); li(bo,bi,k,6);   \
    case 6: li(bo,bi,k,5); li(bo,bi,k,4);   \
    case 4: li(bo,bi,k,3); li(bo,bi,k,2);   \
            li(bo,bi,k,1); li(bo,bi,k,0);   \
}

#else

#define f_round(bo,bi,k)                    \
switch(n_col)                               \
{   case 8: fr(bo,bi,k,7); fr(bo,bi,k,6);   \
    case 6: fr(bo,bi,k,5); fr(bo,bi,k,4);   \
    case 4: fr(bo,bi,k,3); fr(bo,bi,k,2);   \
            fr(bo,bi,k,1); fr(bo,bi,k,0);   \
}

#define i_round(bo,bi,k)                    \
switch(n_col)                               \
{   case 8: ir(bo,bi,k,7); ir(bo,bi,k,6);   \
    case 6: ir(bo,bi,k,5); ir(bo,bi,k,4);   \
    case 4: ir(bo,bi,k,3); ir(bo,bi,k,2);   \
            ir(bo,bi,k,1); ir(bo,bi,k,0);   \
}

#define lf_round(bo,bi,k)                   \
switch(n_col)                               \
{   case 8: fl(bo,bi,k,7); fl(bo,bi,k,6);   \
    case 6: fl(bo,bi,k,5); fl(bo,bi,k,4);   \
    case 4: fl(bo,bi,k,3); fl(bo,bi,k,2);   \
            fl(bo,bi,k,1); fl(bo,bi,k,0);   \
}

#define li_round(bo,bi,k)                   \
switch(n_col)                               \
{   case 8: il(bo,bi,k,7); il(bo,bi,k,6);   \
    case 6: il(bo,bi,k,5); il(bo,bi,k,4);   \
    case 4: il(bo,bi,k,3); il(bo,bi,k,2);   \
            il(bo,bi,k,1); il(bo,bi,k,0);   \
}

#endif

#elif BLOCK_SIZE_R == 16

#define state_in(b,k)   si(b,k,0); si(b,k,1); si(b,k,2); si(b,k,3)
#define state_out(b)    so(b,0); so(b,1); so(b,2); so(b,3)
#if defined(COMPACT)
#define cf_round(bo,bi,k)   cf(bo,bi,k,0); cf(bo,bi,k,1); cf(bo,bi,k,2); cf(bo,bi,k,3)
#define ci_round(bo,bi,k)   ci(bo,bi,k,3); ci(bo,bi,k,2); ci(bo,bi,k,1); ci(bo,bi,k,0)
#define lcf_round(bo,bi,k)  lf(bo,bi,k,0); lf(bo,bi,k,1); lf(bo,bi,k,2); lf(bo,bi,k,3)
#define lci_round(bo,bi,k)  li(bo,bi,k,3); li(bo,bi,k,2); li(bo,bi,k,1); li(bo,bi,k,0)
#else
#define f_round(bo,bi,k)    fr(bo,bi,k,0); fr(bo,bi,k,1); fr(bo,bi,k,2); fr(bo,bi,k,3)
#define i_round(bo,bi,k)    ir(bo,bi,k,3); ir(bo,bi,k,2); ir(bo,bi,k,1); ir(bo,bi,k,0)
#define lf_round(bo,bi,k)   fl(bo,bi,k,0); fl(bo,bi,k,1); fl(bo,bi,k,2); fl(bo,bi,k,3)
#define li_round(bo,bi,k)   il(bo,bi,k,3); il(bo,bi,k,2); il(bo,bi,k,1); il(bo,bi,k,0)
#endif

#elif BLOCK_SIZE_R == 24

#define state_in(b,k)   si(b,k,0); si(b,k,1); si(b,k,2); si(b,k,3); si(b,k,4); si(b,k,5)
#define state_out(b)    so(b,0); so(b,1); so(b,2); so(b,3); so(b,4); so(b,5)
#if defined(COMPACT)
#define cf_round(bo,bi,k)   cf(bo,bi,k,0); cf(bo,bi,k,1); cf(bo,bi,k,2); cf(bo,bi,k,3); \
                            cf(bo,bi,k,4); cf(bo,bi,k,5)
#define ci_round(bo,bi,k)   ci(bo,bi,k,5); ci(bo,bi,k,4); ci(bo,bi,k,3); ci(bo,bi,k,2); \
                            ci(bo,bi,k,1); ci(bo,bi,k,0)
#define lcf_round(bo,bi,k)  lf(bo,bi,k,0); lf(bo,bi,k,1); lf(bo,bi,k,2); lf(bo,bi,k,3); \
                            lf(bo,bi,k,4); lf(bo,bi,k,5)
#define lci_round(bo,bi,k)  li(bo,bi,k,5); li(bo,bi,k,4); li(bo,bi,k,3); li(bo,bi,k,2); \
                            li(bo,bi,k,1); li(bo,bi,k,0)
#else
#define f_round(bo,bi,k)    fr(bo,bi,k,0); fr(bo,bi,k,1); fr(bo,bi,k,2); fr(bo,bi,k,3); \
                            fr(bo,bi,k,4); fr(bo,bi,k,5)
#define i_round(bo,bi,k)    ir(bo,bi,k,5); ir(bo,bi,k,4); ir(bo,bi,k,3); ir(bo,bi,k,2); \
                            ir(bo,bi,k,1); ir(bo,bi,k,0)
#define lf_round(bo,bi,k)   fl(bo,bi,k,0); fl(bo,bi,k,1); fl(bo,bi,k,2); fl(bo,bi,k,3); \
                            fl(bo,bi,k,4); fl(bo,bi,k,5)
#define li_round(bo,bi,k)   il(bo,bi,k,5); il(bo,bi,k,4); il(bo,bi,k,3); il(bo,bi,k,2); \
                            il(bo,bi,k,1); il(bo,bi,k,0)
#endif

#elif BLOCK_SIZE_R == 32

#define state_in(b,k)   si(b,k,0); si(b,k,1); si(b,k,2); si(b,k,3); si(b,k,4); si(b,k,5); si(b,k,6); si(b,k,7)
#define state_out(b)    so(b,0); so(b,1); so(b,2); so(b,3); so(b,4); so(b,5); so(b,6); so(b,7)
#if defined(COMPACT)
#define cf_round(bo,bi,k)   cf(bo,bi,k,0); cf(bo,bi,k,1); cf(bo,bi,k,2); cf(bo,bi,k,3); \
                            cf(bo,bi,k,4); cf(bo,bi,k,5); cf(bo,bi,k,6); cf(bo,bi,k,7)
#define ci_round(bo,bi,k)   ci(bo,bi,k,7); ci(bo,bi,k,6); ci(bo,bi,k,5); ci(bo,bi,k,4); \
                            ci(bo,bi,k,3); ci(bo,bi,k,2); ci(bo,bi,k,1); ci(bo,bi,k,0)
#define lcf_round(bo,bi,k)  lf(bo,bi,k,0); lf(bo,bi,k,1); lf(bo,bi,k,2); lf(bo,bi,k,3); \
                            lf(bo,bi,k,4); lf(bo,bi,k,5); lf(bo,bi,k,6); lf(bo,bi,k,7)
#define lci_round(bo,bi,k)  li(bo,bi,k,7); li(bo,bi,k,6); li(bo,bi,k,5); li(bo,bi,k,4); \
                            li(bo,bi,k,3); li(bo,bi,k,2); li(bo,bi,k,1); li(bo,bi,k,0);
#else
#define f_round(bo,bi,k)    fr(bo,bi,k,0); fr(bo,bi,k,1); fr(bo,bi,k,2); fr(bo,bi,k,3); \
                            fr(bo,bi,k,4); fr(bo,bi,k,5); fr(bo,bi,k,6); fr(bo,bi,k,7)
#define i_round(bo,bi,k)    ir(bo,bi,k,7); ir(bo,bi,k,6); ir(bo,bi,k,5); ir(bo,bi,k,4); \
                            ir(bo,bi,k,3); ir(bo,bi,k,2); ir(bo,bi,k,1); ir(bo,bi,k,0)  
#define lf_round(bo,bi,k)   fl(bo,bi,k,0); fl(bo,bi,k,1); fl(bo,bi,k,2); fl(bo,bi,k,3); \
                            fl(bo,bi,k,4); fl(bo,bi,k,5); fl(bo,bi,k,6); fl(bo,bi,k,7)
#define li_round(bo,bi,k)   il(bo,bi,k,7); il(bo,bi,k,6); il(bo,bi,k,5); il(bo,bi,k,4); \
                            il(bo,bi,k,3); il(bo,bi,k,2); il(bo,bi,k,1); il(bo,bi,k,0)

#endif

#else
#error Illegal Block Size
#endif

#if defined(AES_IN_CPP)
rval aes::encrypt(const byte in_blk[], byte out_blk[]) const
#else
cf_dec encrypt(const byte in_blk[], byte out_blk[], const aes *cx)
#endif
{   word        b0[Mcol], b1[Mcol];
    const word  *kp = f_dat(cx,e_key);
    
    if(!(f_dat(cx,mode) & 0x01)) return aes_bad;

    state_in(b0, kp);
    kp += n_col;

#if defined(COMPACT)
    {   word    *p0 = b0, *p1 = b1, *t, rnd, u, f2;

        for(rnd = 0; rnd < f_dat(cx,Nrnd) - 1; ++rnd)
        {
            cf_round(p1, p0, kp);
            t = p0, p0 = p1, p1 = t; kp += n_col;
        }

        lcf_round(p1, p0, kp);
    }   
#elif defined(UNROLL)
    {
        if(f_dat(cx,Nrnd) > 12)
        {
            f_round(b1, b0, kp);
            f_round(b0, b1, kp + n_col); kp += 2 * n_col;
        }

        if(f_dat(cx,Nrnd) > 10)
        {
            f_round(b1, b0, kp);
            f_round(b0, b1, kp + n_col); kp += 2 * n_col;
        }

        f_round(b1, b0, kp);            f_round(b0, b1, kp +     n_col);
        f_round(b1, b0, kp + 2 * n_col); f_round(b0, b1, kp + 3 * n_col);
        f_round(b1, b0, kp + 4 * n_col); f_round(b0, b1, kp + 5 * n_col);
        f_round(b1, b0, kp + 6 * n_col); f_round(b0, b1, kp + 7 * n_col);
        f_round(b1, b0, kp + 8 * n_col); kp += 9 * n_col;

        lf_round(b0, b1, kp);
    }
#else
    {   word    rnd;

        for(rnd = 0; rnd < (f_dat(cx,Nrnd) >> 1) - 1; ++rnd)
        {
            f_round(b1, b0, kp);
            f_round(b0, b1, kp + n_col); kp += 2 * n_col;
        }

        f_round(b1, b0, kp); kp += n_col;

        lf_round(b0, b1, kp);
    }
#endif

    state_out(b0);
    return aes_good;
}

#if defined(AES_IN_CPP)
rval aes::decrypt(const byte in_blk[], byte out_blk[]) const
#else
cf_dec decrypt(const byte in_blk[], byte out_blk[], const aes *cx)
#endif
{   word        b0[Mcol], b1[Mcol];

#if defined(COMPACT)
    const word  *kp = f_dat(cx,e_key) + n_col * f_dat(cx,Nrnd);
#else
    const word  *kp = f_dat(cx,d_key) + n_col * f_dat(cx,Nrnd);
#endif

    if(!(f_dat(cx,mode) & 0x02)) return aes_bad;
    state_in(b0, kp);
    kp -= n_col;

#if defined(COMPACT)
    {   word *p0 = b0, *p1 = b1, *t, rnd, u, f2, f4, f8, f9; 

        for(rnd = 0; rnd < f_dat(cx,Nrnd) - 1; ++rnd)
        {
            ci_round(p1, p0, kp);
            t = p0, p0 = p1, p1 = t; kp -= n_col;
        }

        lci_round(p1, p0, kp);
    }
#elif defined(UNROLL)
    {
        if(f_dat(cx,Nrnd) > 12)
        {
            i_round(b1, b0, kp);
            i_round(b0, b1, kp - n_col); kp -= 2 * n_col;
        }

        if(f_dat(cx,Nrnd) > 10)
        {
            i_round(b1, b0, kp);
            i_round(b0, b1, kp - n_col); kp -= 2 * n_col;
        }

        i_round(b1, b0, kp);            i_round(b0, b1, kp -     n_col);
        i_round(b1, b0, kp - 2 * n_col); i_round(b0, b1, kp - 3 * n_col);
        i_round(b1, b0, kp - 4 * n_col); i_round(b0, b1, kp - 5 * n_col);
        i_round(b1, b0, kp - 6 * n_col); i_round(b0, b1, kp - 7 * n_col);
        i_round(b1, b0, kp - 8 * n_col); kp -= 9 * n_col;

        li_round(b0, b1, kp);
    }
#else
    {   word rnd;

        for(rnd = 0; rnd < (f_dat(cx,Nrnd) >> 1) - 1; ++rnd)
        {
            i_round(b1, b0, kp);
            i_round(b0, b1, kp - n_col); kp -= 2 * n_col;
        }

        i_round(b1, b0, kp); kp -= n_col;

        li_round(b0, b1, kp);
    }
#endif

    state_out(b0);
    return aes_good;
}

#endif // !defined(AES_DLL) || !defined(TEST)

// Testing Code that is not required for AES (Rijndael) implementation

#if defined(TEST)

#include <memory.h>

#if defined(AES_IN_CPP)

#include <iostream>
#include <iomanip>

void oblk(char m[], byte v[], word n = 16)
{
    std::cout << std::hex << std::setfill('0') << '\n' << m;

    for(int i = 0; i < n; ++i)
    
        std::cout << std::setw(2) << static_cast<word>(v[i]);
}

void message(char *s)   { std::cout << s; };

#else

#include <stdio.h>

void oblk(char m[], byte v[], word n)
{   word i;

    printf("\n%s", m);

    for(i = 0; i < n; ++i)
    
        printf("%02x", v[i]);
}

void message(char *s)   { printf(s); };

#endif

#if defined(AES_DLL)

#include "windows.h"

HINSTANCE init_dll(fn_ptrs *fn)
{   HINSTANCE   h_dll;

    if(!(h_dll = LoadLibrary(dll_path)))
    {
        message("\n\nDynamic link Library AES_DLL not found\n\n"); return 0;
    }

    fn->fn_set_key = (f_ky*)GetProcAddress(h_dll, "_set_key@16");
    fn->fn_encrypt = (f_ed*)GetProcAddress(h_dll, "_encrypt@12");
    fn->fn_decrypt = (f_ed*)GetProcAddress(h_dll, "_decrypt@12");
    fn->fn_set_blk = (f_bl*)GetProcAddress(h_dll, "_set_blk@8");

#if !defined(BLOCK_SIZE_R)
    if(!fn->fn_set_key || !fn->fn_encrypt || !fn->fn_decrypt || !fn->fn_set_blk)
#else
    if(!fn->fn_set_key || !fn->fn_encrypt || !fn->fn_decrypt)
#endif
    {
        message("\n\nRequired DLL Entry Point(s) not found\n\n"); 
        FreeLibrary(h_dll); 
        return 0;
    }

    return h_dll;
}

#endif // AES_DLL

byte pih[32] = // hex digits of pi
{
    0x32, 0x43, 0xf6, 0xa8, 0x88, 0x5a, 0x30, 0x8d,
    0x31, 0x31, 0x98, 0xa2, 0xe0, 0x37, 0x07, 0x34,
    0x4a, 0x40, 0x93, 0x82, 0x22, 0x99, 0xf3, 0x1d,
    0x00, 0x82, 0xef, 0xa9, 0x8e, 0xc4, 0xe6, 0xc8
};

byte exh[32] =  // hex digits of e
{
    0x2b, 0x7e, 0x15, 0x16, 0x28, 0xae, 0xd2, 0xa6,
    0xab, 0xf7, 0x15, 0x88, 0x09, 0xcf, 0x4f, 0x3c,
    0x76, 0x2e, 0x71, 0x60, 0xf3, 0x8b, 0x4d, 0xa5,
    0x6a, 0x78, 0x4d, 0x90, 0x45, 0x19, 0x0c, 0xfe
};

byte res[3][3][32] = 
{ 
    {
        { 0x39, 0x25, 0x84, 0x1d, 0x02, 0xdc, 0x09, 0xfb, 
          0xdc, 0x11, 0x85, 0x97, 0x19, 0x6a, 0x0b, 0x32 
        },
        { 0xf9, 0xfb, 0x29, 0xae, 0xfc, 0x38, 0x4a, 0x25, 
          0x03, 0x40, 0xd8, 0x33, 0xb8, 0x7e, 0xbc, 0x00
        },
        { 0x1a, 0x6e, 0x6c, 0x2c, 0x66, 0x2e, 0x7d, 0xa6, 
          0x50, 0x1f, 0xfb, 0x62, 0xbc, 0x9e, 0x93, 0xf3
        }
    },
    {
        { 0xb2, 0x4d, 0x27, 0x54, 0x89, 0xe8, 0x2b, 0xb8, 
          0xf7, 0x37, 0x5e, 0x0d, 0x5f, 0xcd, 0xb1, 0xf4, 
          0x81, 0x75, 0x7c, 0x53, 0x8b, 0x65, 0x14, 0x8a
        },
        { 0x72, 0x5a, 0xe4, 0x3b, 0x5f, 0x31, 0x61, 0xde, 
          0x80, 0x6a, 0x7c, 0x93, 0xe0, 0xbc, 0xa9, 0x3c,
          0x96, 0x7e, 0xc1, 0xae, 0x1b, 0x71, 0xe1, 0xcf
        },
        { 0x0e, 0xba, 0xcf, 0x19, 0x9e, 0x33, 0x15, 0xc2, 
          0xe3, 0x4b, 0x24, 0xfc, 0xc7, 0xc4, 0x6e, 0xf4, 
          0x38, 0x8a, 0xa4, 0x75, 0xd6, 0x6c, 0x19, 0x4c
        }
    },
    {
        { 0x7d, 0x15, 0x47, 0x90, 0x76, 0xb6, 0x9a, 0x46, 
          0xff, 0xb3, 0xb3, 0xbe, 0xae, 0x97, 0xad, 0x83, 
          0x13, 0xf6, 0x22, 0xf6, 0x7f, 0xed, 0xb4, 0x87, 
          0xde, 0x9f, 0x06, 0xb9, 0xed, 0x9c, 0x8f, 0x19
        },
        { 0x5d, 0x71, 0x01, 0x72, 0x7b, 0xb2, 0x57, 0x81, 
          0xbf, 0x67, 0x15, 0xb0, 0xe6, 0x95, 0x52, 0x82, 
          0xb9, 0x61, 0x0e, 0x23, 0xa4, 0x3c, 0x2e, 0xb0, 
          0x62, 0x69, 0x9f, 0x0e, 0xbf, 0x58, 0x87, 0xb2
        },
        { 0xa4, 0x94, 0x06, 0x11, 0x5d, 0xfb, 0x30, 0xa4, 
          0x04, 0x18, 0xaa, 0xfa, 0x48, 0x69, 0xb7, 0xc6, 
          0xa8, 0x86, 0xff, 0x31, 0x60, 0x2a, 0x7d, 0xd1, 
          0x9c, 0x88, 0x9d, 0xc6, 0x4f, 0x7e, 0x4e, 0x7a
        }
    }
};

#if !defined(BLOCK_SIZE_R)
#define STR 0
#define CNT 3
#elif BLOCK_SIZE_R == 16
#define STR 0
#define CNT 1
#elif BLOCK_SIZE_R == 24
#define STR 1
#define CNT 2
#elif BLOCK_SIZE_R == 32
#define STR 2
#define CNT 3
#else
#error Illegal block size
#endif

int main(void)
{   byte        out[32], ret[32], err = 0, i, j;
    aes         alg;

#if defined(AES_DLL)
    HINSTANCE   h_dll;
    fn_ptrs     fn;

    if(!(h_dll = init_dll(&fn))) return -1;
#endif

#if !defined(AES_IN_CPP)
    alg.mode = 0;
#endif

    for(i = STR; i < CNT; ++i)
    {
#if !defined(BLOCK_SIZE_R)
        f_set_blk(&alg, 16 + 8 * i);
#elif defined(AES_DLL)
        if(fn.fn_set_blk) f_set_blk(&alg, BLOCK_SIZE_R);
#endif
        for(j = 0; j < 3; ++j)
        {
            f_set_key(&alg, exh, 16 + 8 * j, both); 
            oblk("\nkey     = ", exh, 16 + 8 * j); 
            oblk("input   = ", pih, 16 + 8 * i); 
            f_encrypt(&alg, pih, out); 
            oblk("enc     = ", out, 16 + 8 * i);
            if(memcmp(out, res[i][j], 16 + 8 * i)) err++;
            f_decrypt(&alg, out, ret); 
            oblk("dec     = ", ret, 16 + 8 * i); 
            if(memcmp(ret, pih, 16 + 8 * i)) err++;
        }
    }

    message("\n\nThese values are ");
    message(err ? "in error\n\n" : "correct\n\n");

#if defined(AES_DLL)
    if(h_dll) FreeLibrary(h_dll);
#endif

    return 0;
}

#endif // TEST
