/***************************************************************************************
 * File: bc_rijn.c
 *  
 * Purpose: Using the Rijndael algorithm in CBC and CFB mode
 *          We use it with 16-bytes blocks and 256-bit key
 *
 *  Rijndael algorithm's authors:
 *		Joan Daemen
 *		Vincent Rijmen
 *
 * The code is based on Dr B. R. Gladman <brg@gladman.uk.net> algorithm implementation.
 *
 ***************************************************************************************/


//#include <windows.h>   /* the string MUST be commented if */
                       /* we compile Windows 95 or NT drivers */

//#ifndef __WIN95__
//#include <windef.h>
//#else
//#include <basedef.h>
//#endif

#include "BC_RIJN.H"
#include "AES.H"

/************ RIJNDAEL 32-bit CFB mode encryption *****************/

VOID
RIJN_encrypt_CFB( UINT *IVector, 
             UINT *KeyAddress,
             UINT *SrcBuffer,
             UINT *DstBuffer,
             UINT Length )    /* in bytes */
{
  UINT i,j,dwLength;
  UINT shReg[4], shRegEnc[4];

  for(i=0; i<(BLOCK_SIZE/32); i++) shReg[i] = IVector[i];

  dwLength = Length>>2; /* Length of buffer in 32-bit UINTs */

  for(i=0; i<dwLength; i++)
  {
    for(j=0; j<(BLOCK_SIZE/32); j++) shRegEnc[j] = shReg[j];

	encrypt((const byte *)shRegEnc, (byte *)shRegEnc, (const aes *)KeyAddress );

	DstBuffer[i] = SrcBuffer[i] ^ shRegEnc[ (BLOCK_SIZE/32 - 1) ];

/*  This code when aggressively optimized by gcc produces incorrect indexes
    for(j=BLOCK_SIZE/32; (j--)>0 ;)
	{  shReg[ j ] = shReg[ j - 1];
    }
*/

    shReg[3] = shReg[2];
    shReg[2] = shReg[1];
    shReg[1] = shReg[0];
	shReg[0] = DstBuffer[i];
  }
}


/************ RIJNDAEL 32-bit CFB mode decryption *****************/

VOID
RIJN_decrypt_CFB( UINT *IVector, 
             UINT *KeyAddress,
             UINT *SrcBuffer,
             UINT *DstBuffer,
             UINT Length )    /* in bytes */
{
  UINT i,j,dwLength;
  UINT shReg[4], shRegEnc[4], srcSave;

  for(i=0; i<(BLOCK_SIZE/32); i++) shReg[i] = IVector[i];

  dwLength = Length>>2; /* Length of buffer in 32-bit UINTs */

  for(i=0; i<dwLength; i++)
  {
	for(j=0; j<(BLOCK_SIZE/32); j++) shRegEnc[j] = shReg[j];

	srcSave = SrcBuffer[i];
	  
	encrypt((const byte *)shRegEnc, (byte *)shRegEnc, (const aes *)KeyAddress );

	DstBuffer[i] = SrcBuffer[i] ^ shRegEnc[ (BLOCK_SIZE/32 - 1) ];

/*  This code when aggressively optimized by gcc produces incorrect indexes
    for(j=BLOCK_SIZE/32; (j--)>0 ;)
	{  shReg[ j ] = shReg[ j - 1];
    }
*/ 

    shReg[3] = shReg[2];
    shReg[2] = shReg[1];
    shReg[1] = shReg[0];
	shReg[0] = srcSave;
  }
}




/************ RIJNDAEL CBC mode encryption *****************/
 
VOID
RIJN_Encrypt( UINT *IVector, 
         UINT *KeyAddress,
         UINT *SrcBuffer,
         UINT *DstBuffer,
         UINT Length )    /* in bytes */
{ UINT i,l;
  UINT ivLocal[4];

  for(i=0; i<4; i++) ivLocal[i] = IVector[i&1];

  if (Length < (BLOCK_SIZE/8))
  {
    RIJN_encrypt_CFB(ivLocal,KeyAddress,SrcBuffer,DstBuffer,Length);
    return;
  }

  /* CBC encryption */
  for(i=0, l=0; l<(Length>>4); l++, i=i+4)
    { 
	  // do ECB encryption of (Initial_Vector XOR Data)
      ivLocal[0] ^= SrcBuffer[i  ];
      ivLocal[1] ^= SrcBuffer[i+1];
      ivLocal[2] ^= SrcBuffer[i+2];
      ivLocal[3] ^= SrcBuffer[i+3];

      encrypt((const byte *)ivLocal, (byte *)ivLocal, (const aes *)KeyAddress );
            
      DstBuffer[i  ] = ivLocal[0];
      DstBuffer[i+1] = ivLocal[1];
      DstBuffer[i+2] = ivLocal[2];
      DstBuffer[i+3] = ivLocal[3];
    }
}


/************** RIJNDAEL CBC mode decryption ****************/

VOID
RIJN_Decrypt( UINT *IVector, 
         UINT *KeyAddress,
         UINT *SrcBuffer,
         UINT *DstBuffer,
         UINT Length )    // in bytes
{ UINT i,j,l;
  UINT ivOld[4], ivNew[4], buf[4];

  for(i=0; i<4; i++) ivOld[i] = IVector[i&1];

  if (Length < (BLOCK_SIZE/8))
  {
    RIJN_decrypt_CFB(ivOld,KeyAddress,SrcBuffer,DstBuffer,Length);
    return;
  }

  /* CBC decryption */
  for(i=0, l=0; l<(Length>>4) ; l++, i = i+4)
    { 
	  // Encrypted Data -> new IV, 
	  // then do ECB decryption of Encrypted Data, 
      // then XOR decrypted data with old IV

      for(j=0; j<4; j++) buf[j] = ivNew[j] = SrcBuffer[i+j];

      decrypt((const byte *)buf, (byte *)buf, (const aes *)KeyAddress );

      for(j=0; j<4; j++) 
	  { DstBuffer[i+j] = ivOld[j] ^ buf[j];
	    ivOld[j] = ivNew[j];
	  }
    }
}


////////////////////////////////////////////////////////

BOOL
RIJN_KeyExtend(
		PUCHAR KeySource,		//32 8-bits bytes - 256-bit key
		PUINT KeyDestination )
{
	if (set_key((const byte*)KeySource, KEY_LENGTH / 8, both, (aes *)KeyDestination ) != aes_good)
		return FALSE;

	return TRUE;
}

//
//	ECB block encryption
//
VOID
RIJN_encrypt_ECB( BYTE *KeyAddress,
             const BYTE *inBlock,
             BYTE *outBlock )
{
	encrypt( inBlock, outBlock, (const aes *)KeyAddress );
}

VOID
RIJN_decrypt_ECB( BYTE *KeyAddress,
             const BYTE *inBlock,
             BYTE *outBlock )
{
	decrypt( inBlock, outBlock, (const aes *)KeyAddress );
}
