#include "x509_tags.h"
#include "x509_structs.h"
#include "pkcs12_funcs.h"
#include "oidFuncs.h"
#include "../symmetricAlgs/symmetricAlgs.h"
#include "../SMemory/cPassword.h"
#include "../PGPPkts/pktdefs.h"
#include "../hashAlgs/hashAlgs.h"
#include "../RNG/cRNG.h"
#include "../publicKeyAlgs/publicKeyAlgs.h"
#include <stdlib.h>
#include <time.h>
#include <stdio.h> // fo debuging only.
//
//	Function : addBytesM
//		Adds byte buffer x to byte buffer r, both buffers are to be m size.
//
static int addBytesM( BYTE *r, BYTE *x, BYTE carry, int m )
{
	unsigned int a;

	while( m-- )
	{
		a = (unsigned int) r[m] + x[m] + carry;
		carry = BYTE( a >> 8 );
		r[m] = BYTE( a );
	}

	return carry;
}

//
//	Function: P12_producePseudoRandomBytes
//
//	See - pkcs #12, B.2 General method. (p.14)
//
//	Note, use password with last zero symbol. Otherwise  so the result does not depend on pwd length
//	if password contains same symbols only :(
//
//	The cPassword->GenerateUnicodeString uses last zero byte to geneate required unicode buffer.
//	See sample below.
//
int P12_producePseudoRandomBytes(
			BYTE *outBuff,
			int outBuffSize, 
			BYTE *salt,
			int saltSize,
			int iterations,
			BYTE *pwd,		// Unicode buffer required.
			int pwdSize,
			BYTE ID,		// See P12prgm enums above
			BYTE haID		// = -1 ( means pgphaSHA1 );
	)
{
	if ( outBuffSize == 0 )
		return x509e_ok;

	if ( haID == BYTE(-1) )
		haID = pgphaSHA1;

	int v = 64; // 512 bits

	int u;
	int res = ha_getDigestSize( &u, haID );

	if ( res != hae_ok )
		return res;

	// 1. Constructs a "Deversifier" string D
	cPlainTextBlock D;

	int i;

	for( i = v; i--; )
	{
		if ( D.PutByte( ID ) == 0 )
			return x509e_memoryAllocation;
	}

	// 2. Concatenates copies of salt to string S
	cPlainTextBlock I;
	for( i = v * (( saltSize + v - 1 ) / v ); ; i -= saltSize )
	{
		if ( i > saltSize )
		{
			if ( I.WriteBytes( salt, saltSize ) != saltSize )
				return x509e_memoryAllocation;
		}
		else
		{
			if ( I.WriteBytes( salt, i ) != i )
				return x509e_memoryAllocation;

			break;
		}
	}

	// 3. Concatenates copies of password to string P
	for( i = v * (( pwdSize + v - 1 ) / v ); ; i -= pwdSize )
	{
		if ( i > pwdSize )
		{
			if ( I.WriteBytes( pwd, pwdSize ) != pwdSize )
				return x509e_memoryAllocation;
		}
		else
		{
			if ( I.WriteBytes( pwd, i ) != i )
				return x509e_memoryAllocation;

			break;
		}
	}

	// 4. I == S||P, it is already so due to previous steps.

	// 5. Set c
	int c = ( outBuffSize + u - 1 ) / u;

	cPlainTextBlock A;
	if ( A.Allocate( c * u ) == NULL )
		return x509e_memoryAllocation;

	cPlainTextBlock B;
	if ( B.Allocate( v ) == NULL )
		return x509e_memoryAllocation;

	HHA ha;

	res = ha_createAlgorithm( &ha, haID );
	if ( res != hae_ok ) return res;

	// 6. For i = 1, .. c, does the following:

	for( i = 0; i < c; i++ )
	{
		// a. Set A[i] = H(H(H(...D||I ))), r times

		BYTE *Ai = A.buffer + i * u;

		ha_init( ha );
		ha_process( ha, D.buffer, D.bufferSize );
		ha_process( ha, I.buffer, I.bufferSize );
		ha_final( ha, Ai );

		for( int h = 1; h < iterations; h++ )
		{
			ha_init( ha );
			ha_process( ha, Ai, u );
			ha_final( ha, Ai );
		}

		if ( i + 1 == c ) // It is the last pass, so following steps are not required.
			break;

		// b. Concatenates copies of the A[i] to create string B of length v

		B.SeekToBegin();

		for( int p = v; ; p -= u )
		{
			if ( p > u )
				B.WriteBytes( Ai, u );
			else
			{
				B.WriteBytes( Ai, p );
				break;
			}
		}

		// c. Lets I = Io, ...Ik-1 v-byte block, Ij = ( Ij + B + 1 ) mod 2^v

		for ( int j = 0; j < I.bufferSize; j += v )
		{
			if ( j + v > I.bufferSize )		// I.bufferSize == k * v
				return x509e_internalError; 

			BYTE *Ij = I.buffer + j;

			addBytesM( Ij, B.buffer, 1, v ); // Ij = ( Ij + B + 1 ) mod 2^v
		}
	}

	// 7. Concatenates A1, A2... Ac to form pseudo-random string

	SMemCpy( outBuff, A.buffer, outBuffSize );

	ha_free( ha );

	return x509e_ok;
}

#ifdef _DEBUG

int pwdTest()
{

//	cPassword pwd( "smeg" );
//	int keyLength = 24;
//	int ID = P12_prgm4key;
//	int iterations = 1;
//	BYTE salt[] = { 0x0A, 0x58, 0xCF, 0x64, 0x53, 0x0D, 0x82, 0x3F };
	// result = 8AAAE6297B6CB04642AB5B077851284EB7128F1A2A7FBCA3
	// ----

//	cPassword pwd( "smeg" );
//	int keyLength = 8;
//	int ID = P12_prgm4IV;
//	int iterations = 1;
//	BYTE salt[] = { 0x0A, 0x58, 0xCF, 0x64, 0x53, 0x0D, 0x82, 0x3F };
	// result = 79993DFE048D3B76
	// ----

	cPassword pwd( "queeg" );
	int keyLength = 20;
	int ID = P12_prgm4MAC;
	int iterations = 1000;
	BYTE salt[] = { 0x26, 0x32, 0x16, 0xFC, 0xC2, 0xFA, 0xB3, 0x1C };
	// result = 5EC4C7A80DF652294C3925B6489A7AB857C83476
	// ----

	if ( pwd.GenerateUnicodeString() == 0 )
		return x509e_memoryAllocation;

	cPlainTextBlock ptbKey;

	if ( ptbKey.Allocate( keyLength ) == NULL )
		return x509e_memoryAllocation;

	int res = P12_producePseudoRandomBytes(
		ptbKey.buffer,
		ptbKey.bufferSize,
		salt, sizeof( salt ),
		iterations,
		pwd.unicodeBuffer, pwd.unicodeBufferSize,
		ID );

	return res;
}

#endif // _DEBUG


/*

  P12 safe bag sample ( with PKCS8ShroudedKeyBag type )
  see PKCS-12 4.2 The SafeBag type

0000  30 82 03 43   835: SEQUENCE
0004  30 82 03 3F   831: . SEQUENCE
0008  06 0B          11: . . OID { pkcs-12-bagType-PKCS8ShroudedKeyBag }
0015  A0 82 02 A5   677: . . [0]
0019  30 82 02 A1   673: . . . SEQUENCE
001D  30 1B          27: . . . . SEQUENCE
001F  06 0A          10: . . . . . OID { pkcs-12-pbe-pbeWithSHAAnd3-KeyTripleDES-CBC }
002B  30 0D          13: . . . . . SEQUENCE
002D  04 08           8: . . . . . . OCTET STRING
                        47 67 96 1C 16 2A 57 8B
0037  02 01           1: . . . . . . INTEGER 1 ( 1 )
003A                   : . . . . . EOF SEQUENCE -> 002B - 13
003A                   : . . . . EOF SEQUENCE -> 001D - 27
003A  04 82 02 80   640: . . . . OCTET STRING
                        C6 B1 77 7A B2 C5 C0 5D DF 91 C1 EC 59 BE 06 13
02BE                   : . . . EOF SEQUENCE -> 0019 - 673
02BE                   : . . EOF OPTIONAL[0] -> 0015 - 677
02BE  31 81 86      134: . . SET
02C1  30 5F          95: . . . SEQUENCE
02C3  06 09           9: . . . . OID { pkcs-9-friendlyName }
02CE  31 52          82: . . . . SET
02D0  1E 50          80: . . . . . Unicode String - crypt@crypt.nsk.ru's GlobalS
ign nv-sa ID
0322                   : . . . . EOF SET -> 02CE - 82
0322                   : . . . EOF SEQUENCE -> 02C1 - 95
0322  30 23          35: . . . SEQUENCE
0324  06 09           9: . . . . OID { pkcs-9-localKeyId }
032F  31 16          22: . . . . SET
0331  04 14          20: . . . . . OCTET STRING
                        26 7B 0E 0C D8 32 60 97 5A 0E 16 ED C0 C8 14 51
0347                   : . . . . EOF SET -> 032F - 22
0347                   : . . . EOF SEQUENCE -> 0322 - 35
0347                   : . . EOF SET -> 02BE - 134
0347                   : . EOF SEQUENCE -> 0004 - 831
0347                   : EOF SEQUENCE -> 0000 - 835




 PKCS-12.v1.  4.2.2 The PKCS-8ShroudedKeyBag type
 PKCS8ShroudedKeyBag ::= EncryptedPrivateKeyInfo

 PKCS-8.v1.2	7. Encrypted private-key information syntax

	EncryptedPrivateKeyInfo ::= SEQUENCE {
	  encryptionAlgorithm EncryptionAlgorithmIdentifier,
	  encryptedData EncryptedData }

	EncryptionAlgorithmIdentifier ::= AlgorithmIdentifier
	EncryptedData ::= OCTET STRING

	AlgorithmIdentifier  ::=  SEQUENCE  {
		 algorithm               OBJECT IDENTIFIER,
		 parameters              ANY DEFINED BY algorithm OPTIONAL  }

	pkcs-12PbeParams ::= SEQUENCE {
		salt		OCTET STRING,
		iterations	INTEGER
	}



SafeBag ::= SEQUENCE {
  	bagId	    	BAG-TYPE.&id ({PKCS12BagSet})
  	bagValue  	[0] EXPLICIT BAG-TYPE.&Type({PKCS12BagSet}{@bagId}),
  	bagAttributes 	SET OF PKCS12Attribute OPTIONAL
}

PKCS12Attribute ::= SEQUENCE {
	attrId		ATTRIBUTE.&id ({PKCS12AttrSet}),
	attrValues	SET OF ATTRIBUTE.&Type ({PKCS12AttrSet}{@attrId})
} -- This type is compatible with the X.500 type 'Attribute'

PKCS12AttrSet ATTRIBUTE ::= {
	friendlyName |	-- from PKCS #9
	localKeyId, 	-- from PKCS #9
	... -- Other attributes are allowed
}

*/

const char *oidP12PbeWithSHAAnd3KeyTripleDES_CBC = "1.2.840.113549.1.12.1.3"; // { pkcs-12 pbe pbeWithSHAAnd3KeyTripleDES_CBC }

//
//	If out - encrypts data and fills buffer including octet string...
//	otherwise fills only pOutSize value.
//
/*
	EncryptedPrivateKeyInfo ::= SEQUENCE {
	  encryptionAlgorithm EncryptionAlgorithmIdentifier,
	  encryptedData EncryptedData }

	EncryptionAlgorithmIdentifier ::= AlgorithmIdentifier
	EncryptedData ::= OCTET STRING

	AlgorithmIdentifier  ::=  SEQUENCE  {
		 algorithm               OBJECT IDENTIFIER,
		 parameters              ANY DEFINED BY algorithm OPTIONAL  }

	pkcs-12PbeParams ::= SEQUENCE {
		salt		OCTET STRING,
		iterations	INTEGER
	}
*/

//
//	Function P12_makePBEncryptedPkt - Password Based Encryption of specified data
//
int P12_makePBEncryptedPkt(
		BYTE *out,	int *pOutSize,
		BYTE *data,	int dataSize,

		BYTE *salt, int saltSize,
		cPassword *cPwd, //const char *pwd,
		int iterations,
		const char *pbeAlgOID // = oidP12PbeWithSHAAnd3KeyTripleDES_CBC
	)
{
	//
	//	1. Finds pbe algorithm by oid and get its info
	//
	tUnit *tu;

	if ( generateOIDName( NULL, pbeAlgOID, &tu ) == 0 )
		return x509e_undefinedOIDType;

	BYTE haID;		// hash algorithm ID;
	BYTE saID;		// symmetric algorithm ID;
	BYTE saMode;	// symmetric algorithm mode ( usually CBC );

	haID = tu->hAlg;
	saID = tu->sAlg;
	saMode = tu->mode;

	if ( ( haID == BYTE(-1) ) || ( saID == BYTE(-1) ) || ( saMode == BYTE(-1) ) )
		return x509e_undefinedAlgorithm;

	int keyLength = sa_getKeySize( saID );

	if ( keyLength == 0 )
		return sae_invalidAlgorithmID;

	int blockSize = sa_getBlockSize( saID );
	int addedSize = blockSize - ( dataSize % blockSize ); // 0 < addedSize <= blockSize
	int paddedSize = dataSize + addedSize;
	//
	//	2. Checks required out buffer size
	//

	x509_struct seq;
	x509_struct		seqAlgID;
	x509_OID			oid;
	x509_struct			seqParams;
	x509_struct				octSalt;
	x509_integer			intIterations;
	x509_struct		octEncData;

	int required = 
		seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			seqAlgID.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				oid.prepareForWritting( pbeAlgOID  )
				+
				seqParams.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
					octSalt.prepareForWritting( asn1tag_OCTET_STRING, saltSize )
					+
					intIterations.prepareForWritting( iterations )
				)
			)
			+
			octEncData.prepareForWritting( asn1tag_OCTET_STRING, paddedSize )
		);

	if ( required > *pOutSize )
	{
		*pOutSize = required;
		return x509e_bufferTooSmall;
	}

	//
	//	3. Fills out buffer
	//

	// 3.1 Writes headers

	int pos = 0;
	int algIDPos = 0;
	int paramsPos = 0;

	*pOutSize = required;
	if ( seq.writeHeader( out, required ) == 0 ) return x509e_internalError;
		if ( seqAlgID.writeHeader( seq.B, seq.L, &pos ) == 0 ) return x509e_internalError;
			if ( oid.writeHeader( seqAlgID.B, seqAlgID.L, &algIDPos ) == 0 ) return x509e_internalError;
			if ( seqParams.writeHeader( seqAlgID.B, seqAlgID.L, &algIDPos ) == 0 ) return x509e_internalError;
				if ( octSalt.writeHeader( seqParams.B, seqParams.L, &paramsPos ) == 0 ) return x509e_internalError;
				if ( intIterations.writeHeader( seqParams.B, seqParams.L, &paramsPos ) == 0 ) return x509e_internalError;
		if ( octEncData.writeHeader( seq.B, seq.L, &pos ) == 0 ) return x509e_internalError;

	if ( ( (int)seq.L != pos ) ||
		 ( (int)seqAlgID.L != algIDPos ) ||
		 ( (int)seqParams.L != paramsPos ) )
			return x509e_internalError;

	// 3.2 Fills bodies

	if ( oid.writeBody() != ppe_ok ) return x509e_internalError;
	if ( octSalt.writeBody( salt, saltSize ) != ppe_ok ) return x509e_internalError;
	if ( intIterations.writeBody( ) != ppe_ok ) return x509e_internalError;

	//
	//	4. Encrypts data
	//

	// 4.1 Prepares key and IV

	int ivLength;
	switch ( saID )
	{
//		case pgpsa128BitRC4:
//		case pgpsa40BitRC4:
//		case pgpsa128BitRC2:
//		case pgpsa40BitRC2:
//			ivLength = 0;
//			break;
		case pgpsa40BitRC2:
		default:
			ivLength = blockSize;
			if ( ivLength == 0 )
				return sae_invalidAlgorithmID;
	}

	cPlainTextBlock ptbKey;

	if ( ptbKey.Allocate( keyLength ) == NULL )
		return x509e_memoryAllocation;

	cPlainTextBlock ptbIV;
	if ( ( ivLength != 0 ) && ( ptbIV.Allocate( ivLength ) == NULL ) )
		return x509e_memoryAllocation;

//	cPassword cPwd( pwd );

	if ( cPwd->GenerateUnicodeString() == 0 )
		return x509e_memoryAllocation;

	int res = P12_producePseudoRandomBytes( ptbKey.buffer, ptbKey.bufferSize, octSalt.B, octSalt.L, iterations, cPwd->unicodeBuffer, cPwd->unicodeBufferSize, P12_prgm4key );
	if ( res != x509e_ok ) return res;

	if ( ivLength != 0 )
	{
		res = P12_producePseudoRandomBytes( ptbIV.buffer, ptbIV.bufferSize, octSalt.B, octSalt.L, iterations, cPwd->unicodeBuffer, cPwd->unicodeBufferSize, P12_prgm4IV );
		if ( res != x509e_ok ) return res;
	}

	// 4.2 Makes padded block

	cPlainTextBlock ptbPadded;

	if ( ptbPadded.Allocate( paddedSize ) == NULL )
		return x509e_memoryAllocation;

	ptbPadded.WriteBytes( data, dataSize, true );

	BYTE padV = BYTE( addedSize );
	for( int i = addedSize; i--; )
		ptbPadded.PutByte( padV );

	// 4.3 Creates algorithm

	HSA sa;
	res = sa_createAlgorithm( &sa, saID, ptbKey.buffer, ptbKey.bufferSize );
	if ( res != sae_ok ) return res;

	// 4.4 Encrypts padded block

	if ( (int)octEncData.L != ptbPadded.bufferSize ) return x509e_internalError;

	switch ( saMode  )
	{
		case pgpsamodeCBC : res = sa_encryptCBC( sa, ptbIV.buffer, octEncData.B, ptbPadded.buffer, ptbPadded.bufferSize ); break;
		case pgpsamodeECB : res = sa_encrypt( sa, octEncData.B, ptbPadded.buffer, ptbPadded.bufferSize ); break;
		case pgpsamodeCFB :
		{
			res = sa_initCFB( sa, ptbIV.buffer );
			if ( res == sae_ok )
				res = sa_encryptCFB( sa, octEncData.B, ptbPadded.buffer, ptbPadded.bufferSize ); break;
			break;
		}
		default:
			res = sae_invalidAlgorithmID;
	}

	sa_free( sa );

	if ( res != sae_ok ) return res;

	return x509e_ok;
}

//	PKCS #7 
//	9.1 SignedData type
//
//	SignedData ::= SEQUENCE {
//	  version Version,
//	  digestAlgorithms DigestAlgorithmIdentifiers,
//	  contentInfo ContentInfo,
//	  certificates
//    	[0] IMPLICIT ExtendedCertificatesAndCertificates OPTIONAL,
//	  crls
//		[1] IMPLICIT CertificateRevocationLists OPTIONAL,
//	  signerInfos SignerInfos }
//	DigestAlgorithmIdentifiers ::=
//	  SET OF DigestAlgorithmIdentifier
//	SignerInfos ::= SET OF SignerInfo

//000F  30 80        3075: . . SEQUENCE
//0011  02 01           1: . . . INTEGER 1 ( 1 )
//0014  31 0B          11: . . . SET
//0016  30 09           9: . . . . SEQUENCE
//0018  06 05           5: . . . . . OID { secSIG-SHA1 }
//001F  05 00           0: . . . . . NULL
//0021  30 80          13: . . . SEQUENCE
//0023  06 09           9: . . . . OID { pkcs-7-data }
//002E  00 00           0: . . . . END OF CONTENTS MARKER	// here empty content
//0030  A0 82 08 D4  2260: . . . [0]
//0034  30 82 04 66  1126: . . . . SEQUENCE
//0038  30 82 02 4E   590: . . . . . SEQUENCE
//003C  A0 03           3: . . . . . . [0]
//003E  02 01           1: . . . . . . . INTEGER 2 ( 2 )

int p7_signedData::parse( const BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res != x509e_ok ) return res;

	BYTE *pos = B;
	int size = L;

	res = version.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( version.compare( 1 ) != 0 )
		return x509e_invalidPktVersion;

	res = digestAlgorithms.parse( pos, size, &pos, &size, asn1tag_CONSTRUCTED_SET );
	if ( res != x509e_ok ) return res;

	res = contentInfo.parse( pos, size, &pos, &size, asn1tag_CONSTRUCTED_SEQUENCE);
	if ( res != x509e_ok ) return res;

	x509_struct optional;
	res = optional.parse( pos, size );
	if ( res != x509e_ok ) return res;

	if ( optional.isOptional() && ( optional.getOptionalNum() == 0 ) )
	{
		res = certificates.parse( pos, size, &pos, &size );
		if ( res != x509e_ok ) return res;
	}

	res = optional.parse( pos, size );
	if ( res != x509e_ok ) return res;

	if ( optional.isOptional() && ( optional.getOptionalNum() == 1 ) )
	{
		res = crls.parse( pos, size, &pos, &size );
		if ( res != x509e_ok ) return res;
	}

	res = signerInfos.parse( pos, size, &pos, &size, asn1tag_CONSTRUCTED_SET );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
	{
		x509_struct eocm;
		res = eocm.parse( pos, size, &pos, &size, asn1tag_END_OF_CONTENTS_MARKER );
		if ( res != x509e_ok ) return res;
	}

	if ( size != 0 ) return x509e_invalidPktSize;

	return x509e_ok;
}

//	SafeBag ::= SEQUENCE {
//		bagId           BAG-TYPE.&id ({PKCS12BagSet}),
//		bagValue        [0] EXPLICIT BAG-TYPE.&Type({PKCS12BagSet}{@bagId}),
//		bagAttributes   SET OF PKCS12Attribute OPTIONAL
//	}
int p12_safeBag::parse( const BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize, const char **oidT )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res != x509e_ok )
		return res;

	BYTE *pos = B;
	int size = L;

	res = oid.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( ( oidT != NULL ) && ( !oid.check( oidT ) ) )
		return x509e_unexpectedOIDType;

	res = value.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( size == 0 ) // skip optional
		return x509e_ok;

	res = attr.parse( pos, size, &pos, &size );

	if ( size != 0 )
		return x509e_invalidPktSize;

	return res;
}

const char *oidP9FriendlyNameOID			= "1.2.840.113549.1.9.20";		// { pkcs-9 friendlyName }
const char *oidP9LocalKeyIdOID				= "1.2.840.113549.1.9.21";		// { pkcs-9 localKeyId }
const char *oidP12BtPKCS8SroudedKeyBagOID	= "1.2.840.113549.1.12.10.1.2";	// { pkcs-12 bagTypes PKCS8SroudedKeyBag }
const char *oidP12BtCertBagOID				= "1.2.840.113549.1.12.10.1.3";	// { pkcs-12 bagTypes certBag }

//
//	P12_createAttributeList
//		Helpful function to create standard pkcs-12 safe bag attribute list.
//
p12_attribute *P12_createAttributeList(
		BYTE *friendlyName, int friendlyNameSize,
		BYTE *localKeyId,	int localKeyIdSize
	)
{
	p12_attribute *list = new p12_attribute;
	if ( list == NULL ) return NULL;

	list->next = new p12_attribute;

	if ( list->next == NULL ) { delete list; return NULL; }

	list->oidStr = oidP9FriendlyNameOID;
	list->type = asn1tag_UnicodeString;
	list->value = friendlyName;
	list->size = friendlyNameSize;


	list->next->oidStr = oidP9LocalKeyIdOID;
	list->next->type = asn1tag_OCTET_STRING;
	list->next->value = localKeyId;
	list->next->size = localKeyIdSize;

	return list;
}

int P12_makeSafeKeyBagPkt( 
		BYTE *out,	int *pOutSize,
		const char *safeBagOID, // f.e. PKCS8SroudedKeyBag_OID
		BYTE *data,	int dataSize,
		p12_attribute *attrList
		)
{

	//
	//	1. Gets required size.
	//

	// 1.a Gets size of attributes

	int attrSize = 0;
	p12_attribute *attrUnit;

	for( attrUnit = attrList; attrUnit; attrUnit = attrUnit->next )
	{
		x509_struct seq;
		x509_struct  oid;
		x509_struct  set;
		x509_struct   val;

		attrSize += seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( asn1tag_OBJECT_IDENTIFIER, string2oid( NULL, 0, attrUnit->oidStr ) )
			+
			set.prepareForWritting( asn1tag_CONSTRUCTED_SET, 
				val.prepareForWritting( attrUnit->type, attrUnit->size )
			)
		);
	}

	// 1.b Gets common size

	p12_safeBag bag;

	int required = 
	bag.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
		bag.oid.prepareForWritting( safeBagOID )
		+
		bag.value.prepareForWritting( asn1tag_CONSTRUCTED_OPTIONAL, dataSize )
		+
		bag.attr.prepareForWritting( asn1tag_CONSTRUCTED_SET, attrSize )
	);

	if ( required > *pOutSize )
	{
		*pOutSize = required;
		return x509e_bufferTooSmall;
	}

	//
	//	2. Writes required buffer
	//

	*pOutSize = required;

	if ( bag.writeHeader( out, required ) == 0 ) return x509e_internalError;

	int pos = 0;

	// 2.a Writes oid

	if ( bag.oid.writeHeader( bag.B, bag.L, &pos ) == 0 ) return x509e_internalError;
	if ( string2oid( bag.oid.B, 0, safeBagOID ) != (int)bag.oid.L ) return x509e_internalError;

	// 2.b. Writes data

	if ( bag.value.writeHeader( bag.B, bag.L, &pos ) == 0 ) return x509e_internalError;
	if ( bag.value.writeBody( data, dataSize ) != x509e_ok ) return x509e_internalError;
	if ( (int)bag.value.L != dataSize ) return x509e_internalError;

	// 2.c. Writes attributes

	if ( bag.attr.writeHeader( bag.B, bag.L, &pos ) == 0 ) return x509e_internalError;

	if ( (int)bag.L != pos ) return x509e_internalError;

	pos = 0;

	int lPos;

	for( attrUnit = attrList; attrUnit; attrUnit = attrUnit->next )
	{
		x509_struct seq;
		x509_struct  oid;
		x509_struct  set;
		x509_struct   val;

		attrSize = seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( asn1tag_OBJECT_IDENTIFIER, string2oid( NULL, 0, attrUnit->oidStr ) )
			+
			set.prepareForWritting( asn1tag_CONSTRUCTED_SET, 
				val.prepareForWritting( attrUnit->type, attrUnit->size )
			)
		);

		lPos = 0;

		if ( seq.writeHeader( bag.attr.B, bag.attr.L, &pos ) == 0 )				return x509e_internalError;
		 if ( oid.writeHeader( seq.B, seq.L, &lPos ) == 0 )						return x509e_internalError;
		  if ( string2oid( oid.B, 0, attrUnit->oidStr ) != (int)oid.L )			return x509e_internalError;
		 if ( set.writeHeader( seq.B, seq.L, &lPos ) == 0 )						return x509e_internalError;
		  if ( val.writeHeader( set.B, set.L ) == 0 )							return x509e_internalError;
		   if ( val.writeBody( attrUnit->value, attrUnit->size ) != x509e_ok )	return x509e_internalError;

		if ( lPos != (int)seq.L ) return x509e_internalError;
	}

	if ( pos != (int)bag.attr.L ) return x509e_internalError;

	return ppe_ok;
}

int P12_makePKCS8ShroudedKeyBagPkt( 
		BYTE *out,	int *pOutSize,
		BYTE *data,	int dataSize,

		BYTE *salt, int saltSize,
		cPassword *cPwd, //const char *pwd,
		int iterations,
		const char *pbeAlgOID,

		p12_attribute *attrList
	)
{

	//
	//	1. Gets required size.
	//

	// 1.a Gets size of encrypted data.

	int valueSize = 0;

	int res = P12_makePBEncryptedPkt( // P8_makeEncryptedPrivateKeyInfoPkt( 
		NULL, &valueSize, 
		NULL, dataSize, 
		salt, saltSize, 
		cPwd,
		iterations,
		pbeAlgOID );

	if ( res != x509e_bufferTooSmall )
		return res;

	// 1.b Gets size of attributes

	int attrSize = 0;
	p12_attribute *attrUnit;

	for( attrUnit = attrList; attrUnit; attrUnit = attrUnit->next )
	{
		x509_struct seq;
		x509_struct  oid;
		x509_struct  set;
		x509_struct   val;

		attrSize += seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( asn1tag_OBJECT_IDENTIFIER, string2oid( NULL, 0, attrUnit->oidStr ) )
			+
			set.prepareForWritting( asn1tag_CONSTRUCTED_SET, 
				val.prepareForWritting( attrUnit->type, attrUnit->size )
			)
		);
	}

	// 1.c Gets common size

	p12_safeBag bag;

	int required = 
	bag.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
		bag.oid.prepareForWritting( oidP12BtPKCS8SroudedKeyBagOID )
		+
		bag.value.prepareForWritting( asn1tag_CONSTRUCTED_OPTIONAL, valueSize )
		+
		bag.attr.prepareForWritting( asn1tag_CONSTRUCTED_SET, attrSize )
	);

	if ( required > *pOutSize )
	{
		*pOutSize = required;
		return x509e_bufferTooSmall;
	}

	//
	//	2. Writes required buffer
	//

	*pOutSize = required;

	if ( bag.writeHeader( out, required ) == 0 ) return x509e_internalError;

	int pos = 0;

	// 2.a Writes oid

	if ( bag.oid.writeHeader( bag.B, bag.L, &pos ) == 0 ) return x509e_internalError;
	if ( bag.oid.writeBody() != ppe_ok ) return x509e_internalError;

	// 2.b. Writes encrypted value 

	if ( bag.value.writeHeader( bag.B, bag.L, &pos ) == 0 ) return x509e_internalError;
	res = P12_makePBEncryptedPkt( // P8_makeEncryptedPrivateKeyInfoPkt(
			bag.value.B, &valueSize, 
			data, dataSize, 
			salt, saltSize, 
			cPwd,
			iterations, 
			pbeAlgOID );

	if ( res != ppe_ok ) return res;
	if ( (int)bag.value.L != valueSize ) return x509e_internalError;

	// 2.c. Writes attributes

	if ( bag.attr.writeHeader( bag.B, bag.L, &pos ) == 0 ) return x509e_internalError;

	if ( (int)bag.L != pos ) return x509e_internalError;

	pos = 0;

	int lPos;

	for( attrUnit = attrList; attrUnit; attrUnit = attrUnit->next )
	{
		x509_struct seq;
		x509_OID	oid;
		x509_struct  set;
		x509_struct   val;

		attrSize = seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			oid.prepareForWritting( attrUnit->oidStr )
			+
			set.prepareForWritting( asn1tag_CONSTRUCTED_SET, 
				val.prepareForWritting( attrUnit->type, attrUnit->size )
			)
		);

		lPos = 0;

		if ( seq.writeHeader( bag.attr.B, bag.attr.L, &pos ) == 0 )		return x509e_internalError;
		 if ( oid.writeHeader( seq.B, seq.L, &lPos ) == 0 )				return x509e_internalError;
		  if ( oid.writeBody() != ppe_ok )								return x509e_internalError;
		 if ( set.writeHeader( seq.B, seq.L, &lPos ) == 0 )				return x509e_internalError;
		  if ( val.writeHeader( set.B, set.L ) == 0 )					return x509e_internalError;
		   if ( val.writeBody( attrUnit->value, attrUnit->size ) != ppe_ok ) return x509e_internalError;

		if ( lPos != (int)seq.L ) return x509e_internalError;
	}

	if ( pos != (int)bag.attr.L ) return x509e_internalError;

	return ppe_ok;
}

int P12_makeCertificateKeyBagPkt( BYTE *out, int *pOut, BYTE *cert, int certSize )
{
	return P12_makeSafeKeyBagPkt( out, pOut, oidP12BtCertBagOID, cert, certSize, NULL );
}


////////////////////////////
//
//	Parser functions
//

//
//	Function: P12_parsePBEAlgorithmID
//
//   AlgorithmIdentifier  ::=  SEQUENCE  {
//        algorithm               OBJECT IDENTIFIER,
//        parameters              ANY DEFINED BY algorithm OPTIONAL  }
//
//
//	pkcs-12PbeParams ::= SEQUENCE {
//		salt		OCTET STRING,
//		iterations	INTEGER
//	}
//

//	SEQUENCE
//	. OID { pkcs-12-pbe OID }
//	. SEQUENCE
//	. . OCTET STRING
//	. . INTEGER 1 ( 1 )


int P12_parsePBEAlgorithmID( BYTE *pos, int size, unsigned int *pIterations, x509_struct *salt, BYTE *haID, BYTE *saID, BYTE *saMode )
{
	x509_OID algOID;

	int res = algOID.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	const char *checkNameT[] = { "pkcs", "12", "pbe" };

	if ( !algOID.checkN( checkNameT, 3 ) )
		return x509e_unexpectedOIDType;

	tUnit *tu;

	if ( generateOIDName( NULL, algOID.oidStr, &tu ) == 0 )
		return x509e_undefinedOIDType;

	*haID = tu->hAlg;
	*saID = tu->sAlg;
	*saMode = tu->mode;

	if ( ( *haID == BYTE(-1) ) || ( *saID == BYTE(-1) ) || ( *saMode == BYTE(-1) ) )
		return x509e_undefinedAlgorithm;

	x509_struct sPar; // SEQUENCE pbeParams

	res = sPar.parse( pos, size, &pos, &size, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
		return x509e_invalidPktSize;

	res = salt->parse( sPar.B, sPar.L, &pos, &size, asn1tag_OCTET_STRING );
	if ( res != x509e_ok ) return res;

	x509_integer sInt; // Iterations count.
	res = sInt.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
		return x509e_invalidPktSize;

	res = sInt.getUIValue( pIterations );
	if ( res != x509e_ok ) return res;

	return x509e_ok;
}


//
//	Info: pkcs #12 v1.0 p.7 - PFX PDU syntax, 
//
//	PDU - Protocol Data Unit.
//	PFX - The top level exchange PDU.
//
// PFX ::= SEQUENCE {
//   	version		INTEGER {v3(3)}(v3,...), ( v == 3 for 1.0 version of pkcs #12 document )
//    	authSafe	ContentInfo,     ( data or signed-data )
//    	macData    	MacData OPTIONAL ( only in password integrity )
// }
//
// MacData ::= SEQUENCE {
// 	mac 		DigestInfo,
//	macSalt		OCTET STRING,
//	iterations	INTEGER DEFAULT 1
// }
//

//
// SafeContents ::= SEQUENCE OF SafeBag
//
// SafeBag ::= SEQUENCE {
// 	bagId	    	BAG-TYPE.&id ({PKCS12BagSet})
// 	bagValue  		[0] EXPLICIT BAG-TYPE.&Type({PKCS12BagSet}{@bagId}),
// 	bagAttributes 	SET OF PKCS12Attribute OPTIONAL
// }
//
// PKCS12Attribute ::= SEQUENCE {
//	attrId		ATTRIBUTE.&id ({PKCS12AttrSet}),
//	attrValues	SET OF ATTRIBUTE.&Type ({PKCS12AttrSet}{@attrId})
// } -- This type is compatible with the X.500 type 'Attribute'
//
// PKCS12AttrSet ATTRIBUTE ::= {
// 	friendlyName |	-- from PKCS #9
//	localKeyId, 	-- from PKCS #9
//	... -- Other attributes are allowed
// }
//

//
//	PKCS #7. General syntax.
//
//	ContentInfo ::= SEQUENCE {
//	  contentType ContentType,
//		content
//		[0] EXPLICIT ANY DEFINED BY contentType OPTIONAL }
//
//	ContentType ::= OBJECT IDENTIFIER
//
static int extractP7Content_Data( BYTE *pos, int size, cPlainTextBlock *ptbContent )
{
	//
	//	DATA := OCTET STRING, it can be CONSTRAINT
	//
	x509_struct octetString;

	int res = octetString.parse( pos, size, NULL, NULL, asn1tag_OCTET_STRING );
	if ( res != x509e_ok ) return res;

	BYTE *dataOctetPos;
	int dataOctetSize;

	if ( octetString.T & asn1tag_CONSTRUCTED )
	{
		dataOctetPos = octetString.B;
		dataOctetSize = octetString.L;
	}
	else
	{
		dataOctetPos = pos;
		dataOctetSize = size;
	}

	while( dataOctetSize )
	{
		x509_struct s;
		res = s.parse( 
				dataOctetPos, dataOctetSize,
				&dataOctetPos, &dataOctetSize
				);

		if ( res != x509e_ok )
			return res;

		if ( s.L == 0 ) continue;

		if ( ( s.T & asn1tag_TYPE ) == asn1tag_OCTET_STRING )
		{	if ( ptbContent->WriteBytes( s.B, s.L ) != (int)s.L )
			return x509e_memoryAllocation;
		}
		else
		if ( ( ( s.T & asn1tag_TYPE ) != asn1tag_END_OF_CONTENTS_MARKER ) // Only end of content marker is allowed
			&& ( dataOctetSize != 0 ) )										// And only in the end of parsed structure.
				return x509e_invalidPktTag;
	}

	return x509e_ok;
}

//
//	PKCS #7, v 1.5
//
//	EncryptedData ::= SEQUENCE {
//	version Version,
//	encryptedContentInfo EncryptedContentInfo }
//
//	The fields of type EncryptedData have the following meanings:
//		�	version is the syntax version number. It shall be 0 for this version of the standard.
//		*	encryptedContentInfo is the encrypted content information, as in Section 10.
//
static int extractP7Content_EncryptedData( BYTE *pos, int size, cPlainTextBlock *ptbContent )
{
	x509_struct seq;
	int res = seq.parse( pos, size, NULL, NULL, asn1tag_SEQUENCE );
	if ( res != x509e_ok )
		return res;

	pos = seq.B;
	size = seq.L;

	x509_integer ver;
	res = ver.parse( pos, size, &pos, &size );
	if ( res != x509e_ok )
		return res;

	UINT verRequired = 0;
	if ( ver.compare( verRequired ) != 0 )
		return x509e_invalidPktVersion;

	//	EncryptedContentInfo ::= SEQUENCE {
	//	contentType ContentType,
	//	contentEncryptionAlgorithm
	//		ContentEncryptionAlgorithmIdentifier,
	//	encryptedContent
	//		[0] IMPLICIT EncryptedContent OPTIONAL }
	//	EncryptedContent ::= OCTET STRING

	x509_struct ci;
	res =  ci.parse( pos, size,  &pos, &size, asn1tag_SEQUENCE );
	if ( res != x509e_ok )
		return res;

	if ( size != 0 )
		return x509e_invalidPkt;

	x509_OID contentType;
	res = contentType.parse( ci.B, ci.L, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( ptbContent->WriteBytes( pos, size ) != size )
		return x509e_memoryAllocation;

	return x509e_treatAsEncryptedAlgorithmContent;
}

static int extractP7Content( BYTE *pos, int size, cPlainTextBlock *ptbContent )
{
	x509_OID contentType;
	int res = contentType.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	x509_struct optional;

	res = optional.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	const char *pkcs7dataT[] = { "pkcs", "7", "data", NULL };
	const char *pkcs7encryptedDataT[] = { "pkcs", "7", "encryptedData", NULL };

	if ( contentType.check( pkcs7dataT ) )
	{
		return extractP7Content_Data( optional.B, optional.L, ptbContent );
	}
	else
	if ( contentType.check( pkcs7encryptedDataT ) )
	{
		return extractP7Content_EncryptedData( optional.B, optional.L, ptbContent );
	}

	return x509e_unexpectedOIDType;
}

// Smple of PKCS12AttrSet - Safe bag attributes.

//02BE  31 81 86      134: . . SET
//02C1  30 5F          95: . . . SEQUENCE
//02C3  06 09           9: . . . . OID { pkcs-9-friendlyName }
//02CE  31 52          82: . . . . SET
//02D0  1E 50          80: . . . . . Unicode String - crypt@crypt.nsk.ru's GlobalSign nv-sa ID
//0322                   : . . . . EOF SET -> 02CE - 82
//0322                   : . . . EOF SEQUENCE -> 02C1 - 95
//0322  30 23          35: . . . SEQUENCE
//0324  06 09           9: . . . . OID { pkcs-9-localKeyId }
//032F  31 16          22: . . . . SET
//0331  04 14          20: . . . . . OCTET STRING
//                        26 7B 0E 0C D8 32 60 97 5A 0E 16 ED C0 C8 14 51
//0347                   : . . . . EOF SET -> 032F - 22
//0347                   : . . . EOF SEQUENCE -> 0322 - 35
//0347                   : . . EOF SET -> 02BE - 134

int P12_attrSet::getAttributeValue( x509_struct *s, const char **checkNameT )
{
	if ( s != NULL )
	{
		s->B = NULL;
		s->L = 0;
	}

	if ( L == 0 ) return x509e_ok;
	if ( B == NULL )	return x509e_internalError;

	x509_struct seq, set;
	x509_OID oid;
	BYTE *pkt = B;
	int pktSize = L;
	int res = 0;

	while( pktSize != 0 )
	{
		res = seq.parse( pkt, pktSize, &pkt, &pktSize, asn1tag_CONSTRUCTED_SEQUENCE );
		if ( res != x509e_ok )
			return res;

		BYTE *pos = seq.B;
		int size = seq.L;

		res = oid.parse( pos, size, &pos, &size );
		if ( res != x509e_ok )	return res;

		if ( !oid.check( checkNameT ) )
			continue;

		res = set.parse( pos, size, &pos, &size, asn1tag_CONSTRUCTED_SET );
		if ( res != x509e_ok )	return res;

		if ( s == NULL )
			return x509e_ok;

		res = s->parse( set.B, set.L );

		return res;
	}

	// The attributes package is optional so it returns ok if the attribute is absent
	return x509e_ok; 
}

// see pkcs9_T table.

int P12_attrSet::getFriendlyName( x509_struct *s )
{
	const char *checkNameT[] = { "pkcs", "9", "friendlyName", NULL };
	return getAttributeValue( s, checkNameT );
}

int P12_attrSet::getLocalKeyId( x509_struct *s )
{
	const char *checkNameT[] = { "pkcs", "9", "localKeyId", NULL };
	return getAttributeValue( s, checkNameT );
}

//
//	P12_PKCS8ShroudedKeyBag_parse
//
//	PKCS8ShroudedKeyBag := EncryptedPrivateKeyInfo // PKCS-8
//
// EncryptedPrivateKeyInfo ::= SEQUENCE {
//   encryptionAlgorithm EncryptionAlgorithmIdentifier,
//   encryptedData EncryptedData }
// EncryptionAlgorithmIdentifier ::= AlgorithmIdentifier
// EncryptedData ::= OCTET STRING
//
//	Sample:
//0004  30 82 03 32   818: . SEQUENCE
//0008  06 0B          11: . . OID { pkcs-12-bagType-PKCS8ShroudedKeyBag }
//0015  A0 82 02 A5   677: . . [0]
//0019  30 82 02 A1   673: . . . SEQUENCE
//001D  30 1B          27: . . . . SEQUENCE
//001F  06 0A          10: . . . . . OID { pkcs-12-pbe-pbeWithSHAAnd3-KeyTripleDES-CBC }
//002B  30 0D          13: . . . . . SEQUENCE
//002D  04 08           8: . . . . . . OCTET STRING
//                       5F 03 59 F4 42 72 CD 14
//0037  02 01           1: . . . . . . INTEGER 1 ( 1 )
//003A                   : . . . . . EOF SEQUENCE -> 002B - 13
//003A                   : . . . . EOF SEQUENCE -> 001D - 27
//003A  04 82 02 80   640: . . . . OCTET STRING
//                        20 03 45 BE 2C 9C 06 FD D9 2C 6F F1 86 51 0C D3
//02BE                   : . . . EOF SEQUENCE -> 0019 - 673
//02BE                   : . . EOF OPTIONAL[0] -> 0015 - 677
//02BE  31 81 86      134: . . SET
//02C1  30 5F          95: . . . SEQUENCE
//02C3  06 09           9: . . . . OID { pkcs-9-friendlyName }
//02CE  31 52          82: . . . . SET
//02D0  1E 50          80: . . . . . Unicode String - crypt@crypt.nsk.ru's GlobalSign nv-sa ID
//0322                   : . . . . EOF SET -> 02CE - 82
//0322                   : . . . EOF SEQUENCE -> 02C1 - 95
//0322  30 23          35: . . . SEQUENCE
//0324  06 09           9: . . . . OID { pkcs-9-localKeyId }
//032F  31 16          22: . . . . SET
//0331  04 14          20: . . . . . OCTET STRING
//                        26 7B 0E 0C D8 32 60 97 5A 0E 16 ED C0 C8 14 51
//0347                   : . . . . EOF SET -> 032F - 22
//0347                   : . . . EOF SEQUENCE -> 0322 - 35
//0347                   : . . EOF SET -> 02BE - 134

const char *namePKCS12_bt_PKCS8ShroudedKeyBagT[] = { "pkcs", "12", "bagType", "PKCS8ShroudedKeyBag", NULL };

static int EncryptedPrivateKeyInfo_parse(
		BYTE *sEPKI_B, int sEPKI_L,
		cPlainTextBlock *ptbInfo,
		cPassword *pwd		 // = NULL
	)
{
	BYTE *pos = NULL;
	int size = 0;

	//
	//	001D SEQUENCE -> encryptionAlgorithm
	//

	x509_struct sEA;

	int res = sEA.parse( sEPKI_B, sEPKI_L, &pos, &size );
	if ( res != x509e_ok ) return res;

	BYTE haID;		// hash algorithm ID;
	BYTE saID;		// symmetric algorithm ID;
	BYTE saMode;	// symmetric algorithm mode ( usually CBC );

	unsigned int iterations;
	x509_struct salt;

	res = P12_parsePBEAlgorithmID( sEA.B, sEA.L, &iterations, &salt, &haID, &saID, &saMode );
	if ( res != x509e_ok ) return res;

	int keyLength = sa_getKeySize( saID );

	if ( keyLength == 0 )
		return sae_invalidAlgorithmID;

	int blockSize = sa_getBlockSize( saID );
	int ivLength;
	switch ( saID )
	{
//		case pgpsa128BitRC4:
//		case pgpsa40BitRC4:
//		case pgpsa128BitRC2:
//		case pgpsa40BitRC2:
//			ivLength = 0;
//			break;
		case pgpsa40BitRC2:
		default:
			ivLength = blockSize;
			if ( ivLength == 0 )
				return sae_invalidAlgorithmID;
	}

	cPlainTextBlock ptbKey;

	if ( ptbKey.Allocate( keyLength ) == NULL )
		return x509e_memoryAllocation;

	cPlainTextBlock ptbIV;
	if ( ( ivLength != 0 ) && ( ptbIV.Allocate( ivLength ) == NULL ) )
		return x509e_memoryAllocation;

	if ( pwd == NULL )
		return x509e_emptyParameters;

	if ( pwd->GenerateUnicodeString() == 0 )
		return x509e_memoryAllocation;

	res = P12_producePseudoRandomBytes( ptbKey.buffer, ptbKey.bufferSize, salt.B, salt.L, iterations, pwd->unicodeBuffer, pwd->unicodeBufferSize, P12_prgm4key );
	if ( res != x509e_ok ) return res;

	if ( ivLength != 0 )
	{
		res = P12_producePseudoRandomBytes( ptbIV.buffer, ptbIV.bufferSize, salt.B, salt.L, iterations, pwd->unicodeBuffer, pwd->unicodeBufferSize, P12_prgm4IV );
		if ( res != x509e_ok ) return res;
	}

	//
	//	003A OCTET STRING -> encryptedData
	//
	//	Note encryptedContent for EnvelopedData may be parsed by the function too
	//	but with openSSL implemetation it has tag 80 - ( asn1tag_OPTIONAL only ), so we do not check the tag here.
	//
	x509_struct osED;

	res = osED.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
		return x509e_invalidPkt;

	if ( ptbInfo->Allocate( osED.L ) == NULL )
		return x509e_memoryAllocation;

	HSA sa;
	res = sa_createAlgorithm( &sa, saID, ptbKey.buffer, ptbKey.bufferSize );
	if ( res != sae_ok ) return res;

	switch ( saMode  )
	{
		case pgpsamodeCBC : res = sa_decryptCBC( sa, ptbIV.buffer, ptbInfo->buffer, osED.B, osED.L ); break;
		case pgpsamodeECB : res = sa_decrypt( sa, ptbInfo->buffer, osED.B, osED.L ); break;
		case pgpsamodeCFB :
		{
			res = sa_initCFB( sa, ptbIV.buffer );
			if ( res == sae_ok )
				res = sa_decryptCFB( sa, ptbInfo->buffer, osED.B, osED.L );
			break;
		}
		default:
			res = sae_invalidAlgorithmID;
	}

	sa_free( sa );

	//
	//	Check padding...
	//

	BYTE padV = ptbInfo->buffer[ ptbInfo->bufferSize - 1 ];
	if ( ( padV > blockSize ) || ( padV == 0 ) )
		return ppe_invalidPassword; // x509e_invalidPkt;
	else
	{
		for ( int i = padV; i; i-- )
		{
			if ( ptbInfo->buffer[ ptbInfo->bufferSize - i ] != padV )
				return x509e_invalidPkt;
		}

		ptbInfo->bufferSize -= padV;
	}

	return res;
}

//
//	Function:
//		P12_AuthenticatedSafe_parse
//
//	AuthenticatedSafe ::= SEQUENCE OF ContentInfo
//	-- Data if unencrypted
//	-- EncryptedData if password-encrypted
//	-- EnvelopedData if public key-encrypted
//

static const char *namePKCS7T[] = { "pkcs", "7" };

int P12_PKCS8ShroudedKeyBag_parse(
		BYTE *pkt, int pktSize,
		cPlainTextBlock *ptbInfo,
		cPassword *pwd,		 // = NULL
		cPlainTextBlock *attrSet // = NULL
	)
{
	// pkt position for this sample == 0008

	BYTE *pos = pkt;
	int size = pktSize;

	if ( attrSet != NULL )
		attrSet->SeekToBegin();

	x509_OID contentType;
	int res = contentType.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;


	if ( !contentType.check( namePKCS12_bt_PKCS8ShroudedKeyBagT ) )
		return x509e_internalError;

	//
	//	0015 . . [0]
	//
	x509_struct optional;

	res = optional.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( !optional.isOptional() || ( optional.getOptionalNum() != 0 ) )
		return x509e_invalidPkt;

	if ( size != 0 )
	{
		x509_struct atSet; // bagAttributes 	SET OF PKCS12Attribute OPTIONAL

		res = atSet.parse( pos, size, &pos, &size, asn1tag_SET );
		if ( res != x509e_ok ) return res;

		if ( size != 0 )
			return x509e_invalidPkt;

		if ( attrSet != NULL )
			attrSet->WriteBytes( atSet.H, atSet.TL, true );
	}

	//
	//	0019 SEQUENCE -> EncryptedPrivateKeyInfo
	//
	x509_struct sEPKI;

	res = sEPKI.parse( optional.B, optional.L, &pos, &size, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
		return x509e_invalidPkt;

	return EncryptedPrivateKeyInfo_parse( sEPKI.B, sEPKI.L, ptbInfo, pwd );
}

int P12_AuthenticatedSafe_parse( BYTE *pos, int size, cPlainTextBlock *ptbInfo, cPassword *pwd, cPlainTextBlock *attrSet )
{
	x509_struct s;
	int res = s.parse( pos, size, &pos, &size, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	if ( size != 0 )
		return x509e_invalidPktSize;

	pos = s.B;
	size = s.L;

	int mainRes = x509e_ok;

	while( size )
	{
		x509_struct ci; // content info

		res = ci.parse( pos, size, &pos, &size, asn1tag_SEQUENCE );
		if ( res != x509e_ok ) return res;

		x509_OID cOID; // Content type

		res = cOID.parse( ci.B, ci.L );
		if ( res != x509e_ok ) return res;

		if ( cOID.checkN( namePKCS7T, 2 ) )
		{
			cPlainTextBlock ptbContent;

			res = extractP7Content( ci.B, ci.L, &ptbContent );

			if ( res == x509e_treatAsEncryptedAlgorithmContent )
			{
				res = EncryptedPrivateKeyInfo_parse( ptbContent.buffer, ptbContent.bufferSize, ptbInfo, pwd );

				if ( res == x509e_ok )
					return res;

				//
				//	Note, SLL P12 file contains the pkcs-7-data block copy with octed string too, so we may skip
				//	the sequence if a error encounterred.
				//

//				if ( mainRes == x509e_ok ) // saves first result
				mainRes = res;
				continue;
			}


			if ( res != x509e_ok ) return res;

			res = P12_AuthenticatedSafe_parse( ptbContent.buffer, ptbContent.bufferSize, ptbInfo, pwd, attrSet );
			if ( res != x509e_ok ) return res;

			//
			//	PtbInfo was properly encrypted and filled
			//	Or attributes was properly readed
			//
			if ( ( ptbInfo->bufferSize != 0 ) 
				|| ( ( attrSet != NULL ) && ( attrSet->bufferSize != 0 ) ) )
					return res;
		}
		else
		if ( cOID.check( namePKCS12_bt_PKCS8ShroudedKeyBagT ) )
		{
			res = P12_PKCS8ShroudedKeyBag_parse( ci.B, ci.L, ptbInfo, pwd, attrSet );
			if ( res != x509e_ok ) return res;

//			WriteToFile( "ptbInfo.p12", ptbInfo->buffer, ptbInfo->bufferSize );

		}
		else
			return x509e_unexpectedOIDType;

	}

	return mainRes;
}

int P12_PFX_PDU_parse( BYTE *pkt, int pktSize, cPlainTextBlock *ptbInfo, cPassword *pwd, cPlainTextBlock *attrSet )
{
	//
	// Parse PFX SECUENCE
	//

	x509_struct pfx;

	int res = pfx.parse( pkt, pktSize, 0,0, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	BYTE *pos = pfx.B;
	int  size = pfx.L;

	//
	//	Version = 3 for 1.0 version of pkcs #12 document
	//

	x509_integer ver;

	res = ver.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	if ( ver.compare( 3 ) != 0 )
		return x509e_invalidPktVersion;

	//
	//	Sequence, PKSC #7 general sytax.
	//
	x509_struct s;

	res = s.parse( pos, size, &pos, &size, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	x509_OID contentType;

	res = contentType.parse( s.B, s.L );

	if ( res != x509e_ok ) return res;

	cPlainTextBlock ptbContent;

	if ( contentType.checkN( namePKCS7T, 2 ) )
	{
	//	cPlainTextBlock ptbContent;

		res = extractP7Content( s.B, s.L, &ptbContent );
		if ( res != x509e_ok ) return res;

		res = P12_AuthenticatedSafe_parse( ptbContent.buffer, ptbContent.bufferSize, ptbInfo, pwd, attrSet );
		if ( res != x509e_ok ) return res;
	}
	else
		return x509e_unexpectedOIDType;

	return x509e_ok;
}
//
//	P12_PFX_PDU_makeP7Data
//	 the contentInfo can be prepared f.e. by P12_makeSafeContents.
//

const char *szP7data = "1.2.840.113549.1.7.1";

struct p12_safeBag_p7data : public p12_safeBag
{
	x509_struct	octetStr;

	int prepareForWritting( int contentInfoSize )
	{	return 
			x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				oid.prepareForWritting( szP7data )
				+
				value.prepareForWritting( asn1tag_CONSTRUCTED_OPTIONAL,
					octetStr.prepareForWritting( asn1tag_OCTET_STRING, contentInfoSize )
				)
			);
	}

	int writeHeader( BYTE *buff, int buffSize, int *pPos = NULL )
	{
		if ( x509_struct::writeHeader( buff, buffSize, pPos ) == 0 ) 
			return x509e_internalError;

		int pos = 0;
		oid.writeHeader( B, L, &pos );
		oid.writeBody();
		value.writeHeader( B, L, &pos );
		return octetStr.writeHeader( value.B, value.L );
	}

	int writeBody( BYTE *contentInfo, int ciSize )
	{
		return octetStr.writeBody( contentInfo, ciSize );
	}
};

int P12_PFX_PDU_makeP7Data( BYTE *pfx, int *pSize, BYTE *contentInfo, int ciSize )
{
	p12_pfx_pdu_p7data s;

	int required = s.prepareForWritting( ciSize );

	if ( required > *pSize )
	{
		*pSize = required;
		return x509e_bufferTooSmall;
	}

	*pSize = required;
	s.writeHeader( pfx, required );
	return s.octetStr.writeBody( contentInfo, ciSize );
}

int p12_pfx_pdu_p7data::prepareForWritting( int contentInfoSize )
{
	return 
		x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			ver.prepareForWritting( 3 )
			+
			seqOfContInfo.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				p7dataOID.prepareForWritting( szP7data )
				+
				optional.prepareForWritting( asn1tag_CONSTRUCTED_OPTIONAL,
					octetStr.prepareForWritting( asn1tag_OCTET_STRING, contentInfoSize )
				)
			)
		);
}

int p12_pfx_pdu_p7data::writeHeader( BYTE *buff, int buffSize, int *pPos )
{
	if ( x509_struct::writeHeader( buff, buffSize, pPos ) == 0 ) 
		return x509e_internalError;

	int pos = 0;
	ver.writeHeader( B, L, &pos );
	ver.writeBody();

	seqOfContInfo.writeHeader( B, L, &pos );

	pos = 0;
	p7dataOID.writeHeader( seqOfContInfo.B, seqOfContInfo.L, &pos );
	p7dataOID.writeBody();

	optional.writeHeader( seqOfContInfo.B, seqOfContInfo.L, &pos );
	return octetStr.writeHeader( optional.B, optional.L );
}

int p12_pfx_pdu_p7data::writeBody( BYTE *contentInfo, int ciSize )
{
	return octetStr.writeBody( contentInfo, ciSize );
}

//
// P12_makeSafeContents
//	the safeBagsT has to be prepared before, f.e by P12_makePKCS8ShroudedKeyBagPkt
//
int P12_makeSafeContents( BYTE *SC, int *pSCSize, cPlainTextBlock *safeBagsT, int safeBagsTSize )
{
	p12_safeContent sc;

	int required = sc.prepareForWritting( safeBagsT, safeBagsTSize );

	if ( *pSCSize < required )
	{
		*pSCSize = required;
		return x509e_bufferTooSmall;
	}

	*pSCSize = required;

	sc.writeHeader( SC, required );

	return sc.writeBody( safeBagsT, safeBagsTSize );
}

int p12_safeContent::prepareForWritting( cPlainTextBlock *safeBagsT, int safeBagsTSize )
{
	int required = 0;
	for( int i = 0; i < safeBagsTSize; i++ )
	{
		x509_struct safeBag;
		int res = safeBag.parse( safeBagsT[i].buffer, safeBagsT[i].bufferSize, 0, 0, asn1tag_CONSTRUCTED_SEQUENCE );
		if ( res != x509e_ok ) return res;

		required += safeBag.TL;
	}

	return x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, required );
}

int p12_safeContent::prepareForWritting( x509_struct *safeBagsT, int safeBagsTSize )
{
	int required = 0;
	for( int i = 0; i < safeBagsTSize; i++ )
	{
		required += safeBagsT->TL;
	}

	return x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, required );
}

int p12_safeContent::writeBody( cPlainTextBlock *safeBagsT, int safeBagsTSize )
{
	int pos = 0;
	for( int i = 0; i < safeBagsTSize; i++ )
	{
		x509_struct safeBag;
		int res = safeBag.parse( safeBagsT[i].buffer, safeBagsT[i].bufferSize, 0, 0, asn1tag_CONSTRUCTED_SEQUENCE );
		if ( res != x509e_ok ) return res;

		BYTE *sbBody = safeBag.B;

		safeBag.writeHeader( B, L, &pos );
		safeBag.writeBody( sbBody, safeBag.L );
	}

	return ppe_ok;
}
/*
//
//	Safe bag
//
struct p12_safeBag : public x509_struct
{
	x509_OID	  OID;
	x509_struct   optional;
	x509_struct		octetStr;
	x509_struct	  attr; // SET OF PKCS12Attribute OPTIONAL

	int prepareForWritting( const char *contentOID, int contentInfoSize )
	{	return 
			x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				OID.prepareForWritting( contentOID )
				+
				optional.prepareForWritting( asn1tag_CONSTRUCTED_OPTIONAL,
					octetStr.prepareForWritting( asn1tag_OCTET_STRING, contentInfoSize )
				)
			);
	}

	int writeHeader( BYTE *buff, int buffSize, int *pPos = NULL )
	{
		if ( x509_struct::writeHeader( buff, buffSize, pPos ) == 0 ) 
			return x509e_internalError;

		int pos = 0;
		OID.writeHeader( B, L, &pos );
		OID.writeBody();
		optional.writeHeader( B, L, &pos );
		return octetStr.writeHeader( optional.B, optional.L );
	}

	int writeBody( BYTE *contentInfo, int ciSize )
	{
		return octetStr.writeBody( contentInfo, ciSize );
	}
};
*/

/////////////////////////////////
//
//	Secret key packet creation sample 
//
int P12_MakeSecretKeyPkt(
	cPlainTextBlock *pfxPduPtb,
	HPKA hSka, 
	wchar_t *friendlyNameBuff,
	int friendlyNameBuffSizeInBytes,
	cPassword *pwd, //const char *pwd,
	const BYTE *pSalt,
	int iterations )
{
	if ( hSka == NULL )
		return x509e_internalError;


	BYTE salt[ 8 ] = { 0x5F, 0x03, 0x59, 0xF4, 0x42, 0x72, 0xCD, 0x14 };
	int res = 0;

	if ( pSalt == NULL )
	{
		//
		//	Generates random salt from current time and clock;
		//
		time_t tm = time(0) ^ clock();
		res = generateRandomKey( salt, sizeof( salt ), (BYTE *)&tm, sizeof( tm ) );

		if ( res != ppe_ok )
			return res;
	}
	else
		memcpy( salt, pSalt, sizeof( salt ) );


	//
	//	Prepares secret key SEQUENCE
	//
	int keySize = 0;
	cPlainTextBlock ptbSKey;

	res = ska_writeKeyToASN1Buffer( hSka, NULL, &keySize );
	if ( res != pkae_bufferSizeTooSmall )
		return res;

	if ( ptbSKey.Allocate( keySize ) == NULL )
		return x509e_memoryAllocation;

	//
	//	Generates local key ID
	//
	int localKeyIDSize = 0;

	res = ska_generateLocalKeyID( hSka, NULL, &localKeyIDSize );

	if ( res != pkae_bufferSizeTooSmall )
		return res;

	cPlainTextBlock ptbLocalKeyID;

	if( ptbLocalKeyID.Allocate( localKeyIDSize ) == NULL )
		return x509e_memoryAllocation;

	res = ska_generateLocalKeyID( hSka, ptbLocalKeyID.buffer, &localKeyIDSize );
	if ( res != x509e_ok ) return res;

	//
	//	Creates secret key attributes
	//

	p12_attribute *attrList = 
			P12_createAttributeList(
				(BYTE*)friendlyNameBuff, friendlyNameBuffSizeInBytes, 
				ptbLocalKeyID.buffer, ptbLocalKeyID.bufferSize );

	//
	//	Makes shrouded secret key bag
	//

	int size = 0;

	res = P12_makePKCS8ShroudedKeyBagPkt(
		NULL, &size, 
		NULL, keySize, 
		salt, sizeof( salt ),
		NULL,
		iterations,
		oidP12PbeWithSHAAnd3KeyTripleDES_CBC,
		attrList );

	if ( res != pkae_bufferSizeTooSmall )
	{
		p12_attribute::freeList( attrList );
		return res;
	}

	cPlainTextBlock shroudedSecretKeyPtb;

	if ( shroudedSecretKeyPtb.Allocate( size ) == NULL )
	{
		p12_attribute::freeList( attrList );
		return ppe_memoryAllocation;
	}

	res = ska_writeKeyToASN1Buffer( hSka, ptbSKey.buffer, &keySize );
	if ( res != pkae_ok )
	{
		p12_attribute::freeList( attrList );
		return res;
	}

	res = P12_makePKCS8ShroudedKeyBagPkt( 
		shroudedSecretKeyPtb.buffer, &size, 
		ptbSKey.buffer, keySize, 
		salt, sizeof( salt ),
		pwd,
		iterations,
		oidP12PbeWithSHAAnd3KeyTripleDES_CBC,
		attrList );

	p12_attribute::freeList( attrList );

	p12_safeBag_p7data safeBag_p7data;
	p12_safeContent sc_1;

	safeBag_p7data.prepareForWritting( sc_1.prepareForWritting( &shroudedSecretKeyPtb, 1 ) );

	p12_pfx_pdu_p7data pfxpdu;
	p12_safeContent sc;

	int pfxPduSize =
		pfxpdu.prepareForWritting( sc.prepareForWritting( &safeBag_p7data, 1 ) );

	if ( pfxPduPtb->Allocate( pfxPduSize ) == NULL )
		return ppe_memoryAllocation;

	pfxpdu.writeHeader( pfxPduPtb->buffer, pfxPduSize );
	sc.writeHeader( pfxpdu.octetStr.B, pfxpdu.octetStr.L );

	safeBag_p7data.writeHeader( sc.B, sc.L );
	sc_1.writeHeader( safeBag_p7data.octetStr.B, safeBag_p7data.octetStr.L );

	return sc_1.writeBody(  &shroudedSecretKeyPtb, 1 );
}

//
//      ConvertToUnicodeString
//              Converts ansi buffer to unicode.
//              buff can the same str buffer
//              Returns really filled bytes in buffer
//
int ConvertToUnicodeString( wchar_t *buff, int buffSize, char *str )
{
        char *p = (char *)malloc( strlen( str ) + 1 );

        if ( p == NULL )
                return 0;

        int converted = 0;

        wchar_t wch = 0;

        for( ; converted < buffSize; converted ++ )
        {
                BYTE *pb = (BYTE *) &wch;
                pb[1] = p[ converted ];

                buff[ converted ] = wch;

                if ( p[ converted ] == 0 )
                {
//                      converted --; // Null char does not includes
                        break;
                }
        }

        free( p );

        return converted * sizeof( wchar_t );
}
