#include "base64funcs.h"
#include <string.h>

static int findString( const BYTE *pkt, int pktSize, const char *str )
{
	int checkLength = strlen( str );
	int checkCount = 0;
	int pos = 0;

	BOOL ok = FALSE;

	for(; pos + checkCount < pktSize ;)
	{
		if ( pkt[ pos + checkCount ] == str[ checkCount ] )
		{
			if ( ++checkCount == checkLength )
			{
				return pos;
			}
		}
		else	// Reset counter
		{
			checkCount = 0;
			pos ++;
		}
	}

	return -1;
}

int findEncodedBlock(
		const BYTE **pBlock,		// OUT - Pointer to byte position of encoded block
		int *pBlockSize,	// OUT - Pointer to finded size of encoded block
		const BYTE *pkt,	// IN - the packet with encoded block to be find.
		int pktSize,		// IN - the packet size
		const char *szBlockStart,
		const char *szBlockEnd,
		char **pArmorHeaders// OUT pointer to position of Armor headers
							// the headers are separated by single 0x0A symbol
							// and are terminated by double 0x0A symbols ( a blank line )
		)
{
	int posStart = findString( pkt, pktSize, szBlockStart );

	if ( posStart == -1 )
		return ppe_invalidBlock;

	posStart += strlen( szBlockStart );

	int posEnd = findString( pkt + posStart, pktSize - posStart, szBlockEnd );

	if ( posEnd == -1 )
		return ppe_invalidBlock;

	//
	//	Finds header lines in armored message.
	//
	BYTE blankLine[3] = { 0x0A, 0x0A, 0 };

	int posBody = findString( pkt + posStart, posEnd, (char *)blankLine );

	if ( posBody == -1 )
	{
		BYTE blankLine2[5] = { 0x0D, 0x0A, 0x0D, 0x0A, 0 };
		posBody = findString( pkt + posStart, posEnd, (char *)blankLine2 );
	}

	if ( posBody == -1 ) // No headers found
		posBody = posStart;
	else
	{
		posBody += strlen( (char *)blankLine );
		posEnd -= posBody; // Position from body position
		posBody = posStart + posBody; // Posotion from pkt begin

	}

	if ( pArmorHeaders != NULL )
		*pArmorHeaders = (char *)pkt + posStart + 1; // +1 skips 0x0A symbols before headers.

	*pBlock = pkt + posBody;
	*pBlockSize = posEnd;

	return ppe_ok;
}

//
//	ASCII Armor
//	6.2.     Forming ASCII Armor                                      42
//	RFC 2440                 OpenPGP Message Format            November 1998
// 
// 
//          +--first octet--+-second octet--+--third octet--+
//          |7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|7 6 5 4 3 2 1 0|
//          +-----------+---+-------+-------+---+-----------+
//          |5 4 3 2 1 0|5 4 3 2 1 0|5 4 3 2 1 0|5 4 3 2 1 0|
//          +--1.index--+--2.index--+--3.index--+--4.index--+
// 
//    Each 6-bit group is used as an index into an array of 64 printable
//    characters from the table below. The character referenced by the
//    index is placed in the output string.
// 
//	   Special processing is performed if fewer than 24 bits are available
//	   at the end of the data being encoded. There are three possibilities:
//
//	    1. The last data group has 24 bits (3 octets). No special
//	       processing is needed.
//
//	    2. The last data group has 16 bits (2 octets). The first two 6-bit
//	       groups are processed as above. The third (incomplete) data group
//	       has two zero-value bits added to it, and is processed as above.
//	       A pad character (=) is added to the output.
//
//	    3. The last data group has 8 bits (1 octet). The first 6-bit group
//	       is processed as above. The second (incomplete) data group has
//	       four zero-value bits added to it, and is processed as above. Two
//	       pad characters (=) are added to the output.
//

const char *base64T = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";
const int base64TSize = 64;

static int getCode( BYTE ch )
{
	int code = base64TSize + 1;

	for (; code--;)
	{
		if( base64T[ code ] == ch )
			break;
	}

	return code;
}

static int getDecodedSizeAndCheck( 
		const BYTE *encBlock,		// IN - encoded buffer
		int encBlockSize )
{
	int code;

	int size = 0;
	int padQ = 0;
	int count = 0;

	bool newLine = false;

	for(; count < encBlockSize; count++ )
	{
		BYTE ch = encBlock[ count ];

		code = getCode( ch );

		if (( ch == 0x0A ) || ( ch == 0x0D ) )
			newLine = true;

		if ( code == base64TSize )
		{
			if ( newLine )	break; // We encountered with check summ.

			padQ++;
		}
		else
		if ( code >= 0 )
		{
			newLine = false;
			size++;
		}
	}

	size = ( size / 4 ) * 3;

	if ( padQ == 1 )
		size += 2;
	else
	if ( padQ == 2 )
		size += 1;
	else
	if ( padQ != 0 )
		return (-1);

	return size;
}

int base64Decode(
		BYTE *decBlock,		// OUT - buffer for decoded data.
		int *pDecBlockSize,	// IN - pointer for decoded size, if size of 
							// the output block is too small return ppe_bufferTooSmall
							// OUT - required or really filled size 
		const BYTE *encBlock,		// IN - encoded buffer
		int encBlockSize )
{
	int size = getDecodedSizeAndCheck( encBlock, encBlockSize );

	if ( size == -1 )
		return ppe_invalidBlock;

	if ( size > *pDecBlockSize )
	{
		*pDecBlockSize = size;
		return ppe_bufferTooSmall;
	}

	int posD = 0, posE = 0;

	DWORD dw24 = 0;

	int dwPos = 0;

	bool newLine = false;

	for(; posE < encBlockSize; )
	{
		BYTE ch = encBlock[ posE++ ];

		if ( ( ch == 0x0A ) || ( ch == 0x0D ) )
		{
			newLine = true;
			continue;
		}

		int code = getCode( ch );

		if ( code < 0 )
			continue;

		if ( code == base64TSize ) // pad -- '='
		{
			if ( newLine )	// 0x0A, "=QWER" It is a check summ just after newline
				break;

			dw24 = ( dw24 << 6 );
			continue;
		}
		else
		{
			dw24 = ( dw24 << 6 ) | (DWORD) code;
			newLine = false;
		}

		if ( ++dwPos == 4 )
		{
			dwPos = 0;
			if ( posD + 2 >=  size )
				return ppe_invalidBlock;

			decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
			decBlock[ posD++ ] = (BYTE)( dw24 >> 8  );
			decBlock[ posD++ ] = (BYTE)( dw24       );

			dw24 = 0;
		}
	}

	if ( dwPos == 3 )
	{
		decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
		decBlock[ posD++ ] = (BYTE)( dw24 >> 8  );
	}
	else
	if ( dwPos == 2 )
	{
		decBlock[ posD++ ] = (BYTE)( dw24 >> 16 );
	}

	*pDecBlockSize = posD;

	return ppe_ok;
}

int base64Encode(
		BYTE *encBlock,		// OUT - buffer for encoded data.
		int *pEncBlockSize,	// IN - pointer for encoded size, if size of 
							// the output block is too small return ppe_bufferTooSmall
							// OUT - required or really filled size 
		const BYTE *decBlock,		// IN - plain text buffer
		int decBlockSize,

		const char *szBlockStart,	// If the strings are not NULL,
		const char *szBlockEnd		// so encloses the encoded block by the strings
		)
{
	int encodedSize = ( decBlockSize / 3 ) * 4;
	int tail = decBlockSize % 3;

	int padded = 0;

	if ( tail == 2 )
		padded = 1;
	else
	if ( tail == 1 )
		padded = 2;

	if ( tail )
		encodedSize += 4;

	int lines = encodedSize / 64;

	bool useBoundStrings = (( szBlockStart != NULL ) && ( szBlockEnd != NULL ));

	if ( useBoundStrings )
	{
		lines += 2;
		encodedSize += ( strlen( szBlockStart ) + strlen( szBlockEnd ) );
	}

	encodedSize += lines * 2; // "\r\n"

	if ( *pEncBlockSize < encodedSize )
	{
		*pEncBlockSize = encodedSize;
		return ppe_bufferTooSmall;
	}

	int bl; // block number

	int ePos = 0;

	if ( useBoundStrings )
	{
		int len = strlen( szBlockStart );
		memcpy( encBlock + ePos, szBlockStart, len );
		ePos += len;
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
	}

	for ( bl = 0; bl < ( decBlockSize / 3 ); bl++ )
	{
		DWORD dw24 =  DWORD( decBlock[ 3 * bl + 2 ] )         | 
					( DWORD( decBlock[ 3 * bl + 1 ] ) << 8  ) |
					( DWORD( decBlock[ 3 * bl     ] ) << 16 );

		encBlock[ ePos++ ] = base64T[ ( dw24 >> 18 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >> 12 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >>  6 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24       ) & 0x3F ];

		if ( bl && ( ( bl % 16 ) == 0 ) ) // new line required;
		{
			encBlock[ ePos++ ] = '\r';
			encBlock[ ePos++ ] = '\n';
		}
	}

	if ( tail )
	{
		DWORD dw24 =  DWORD( decBlock[ 3 * bl ] ) << 16;
		if ( tail == 2 )
			dw24 |= ( DWORD( decBlock[ 3 * bl + 1 ] ) << 8 );

		encBlock[ ePos++ ] = base64T[ ( dw24 >> 18 ) & 0x3F ];
		encBlock[ ePos++ ] = base64T[ ( dw24 >> 12 ) & 0x3F ];

		if ( tail == 2 )
			encBlock[ ePos++ ] = base64T[ ( dw24 >> 6 ) & 0x3F ];
		else
			encBlock[ ePos++ ] = base64T[ 0x40 ];

		encBlock[ ePos++ ] = base64T[ 0x40 ];
	}

	if ( useBoundStrings )
	{
		encBlock[ ePos++ ] = '\r';
		encBlock[ ePos++ ] = '\n';
		int len = strlen( szBlockEnd );
		memcpy( encBlock + ePos, szBlockEnd, len );
		ePos += len;
	}

	if ( ePos != encodedSize )
		ppe_internalError;

	*pEncBlockSize = encodedSize;

	return ppe_ok;
}
