#include "..\crtLibErrs.h"
#include "..\PGPPkts\pktdefs.h."
#include "oidFuncs.h"
#include <stdio.h>
#include <string.h>

//
//	getSignatureAlgOidStr
//		returns specified algorithm oid string
//
const char *getSignatureAlgOidStr(
		unsigned char hAlg,	// Hash alg id				f.e. pgphaSHA1
		unsigned char pAlg	// Public key alg id		f.e. pgppkaRSA
	)
{
	if ( ( hAlg == pgphaMD2 ) && ( pAlg == pgppkaRSA ) ) return "1.2.840.113549.1.1.2";
	if ( ( hAlg == pgphaMD4 ) && ( pAlg == pgppkaRSA ) ) return "1.2.840.113549.1.1.3";
	if ( ( hAlg == pgphaMD5 ) && ( pAlg == pgppkaRSA ) ) return "1.2.840.113549.1.1.4";
	if ( ( hAlg == pgphaSHA1 ) && ( pAlg == pgppkaRSA ) ) return "1.2.840.113549.1.1.5";
	if ( ( hAlg == pgphaSHA1 ) && ( pAlg == pgppkaDSA ) ) return "1.2.840.10040.4.3"; // id-dsa-with-sha1 ID 

	return NULL;
}

//
//	getDigestAlgOidStr
//		returns specified algorithm oid string
//
const char *getDigestAlgOidStr(
		unsigned char hAlg	// Hash alg id				f.e. pgphaSHA1
	)
{
	if ( hAlg == pgphaMD2  ) return "1.2.840.113549.2.2";
	if ( hAlg == pgphaMD4  ) return "1.2.840.113549.2.4";
	if ( hAlg == pgphaMD5  ) return "1.2.840.113549.2.5";
	if ( hAlg == pgphaSHA1 ) return "1.3.14.3.2.26";

	return NULL;
}

//
//	Function: oid2string
//
//	converts a DER encoding ASN.1 object-identifier into dot-string format
// 
char *oid2string( char *buf, unsigned char *oidVal, unsigned int oidSize )
{
	unsigned int ioff;
	unsigned int ibeg;
	int a1;
	int t;

	if ( oidVal[0] < 40 )
		a1 = 0;
	else if ( oidVal[0] < 80 )
		a1 = 1;
	else
		a1 = 2;

    sprintf( buf, "%d.%d", a1, oidVal[0] - 40 * a1 );
    
    for ( ioff = 1; ioff < oidSize; ioff++ )
    {
	    ibeg = ioff;
	    while ( oidVal[ioff] & 0x80 && ioff < oidSize )
	        ioff++;
	    for ( t = 0; ibeg <= ioff; ibeg++ )
	        t += (int)((oidVal[ibeg] & 0x7f)) * ( int(1) << ( 7 * (ioff - ibeg) ));

		int len = strlen( buf );

		if ( len >= maxOIDStrSize - 10 )
			return 0;

	    sprintf( buf + len, ".%d", t );
    }

    return buf;
}

//
//	Function: string2oid
//		Converts OID string to unsigned char buffer.
//		Returns -1 if *pOidSize too small, otherwise return filled size.
//
int string2oid( unsigned char *oid, int *pOidSize, const char *str )
{
    int num;
    int count = 0;
    int first = 0;
    int offset = 0;
    unsigned char buf[256];

    while (*str)
    {
		for ( num = 0; *str && *str != '.'; str++ )
		{
			num *= 10;
			num += *str - '0';
		}
		if ( *str )
			str++;
		if ( count == 0 )
			first = num;
		else
		{
			if ( count == 1 )
				num += 40 * first;
			if ( num > 0x200000 ) /* 0x80^3 */
			{
				buf[offset++] = 0x80 | (num / 0x200000);
				num -= 0x200000 * (num / 0x200000);
			}
			if ( num > 0x4000 ) /* 0x80^2 */
			{
				buf[offset++] = 0x80 | (num / 0x4000);
				num -= 0x4000 * (num / 0x4000);
			}
			if ( num > 128 )
			{
				buf[offset++] = 0x80 | (num / 0x80);
				num -= 0x80 * (num / 0x80);
			}
			buf[offset++] = num;
		}
		count++;
    }

	if ( pOidSize != NULL )
	{
		if ( *pOidSize < offset )
		{
			*pOidSize = offset;
			return -1;
		}

		*pOidSize = offset;
	}

	if ( oid != NULL )
	    memcpy ( oid, buf, offset );

    return offset;
}

static int findUnit( tUnit *t, int q, const char *oidString )
{
	while( q-- )
	{
		if ( strncmp( t[q].oid, oidString, strlen( t[q].oid ) ) == 0 )
			break;
	}

	return q;
}

//
//	Function: generateOIDName
//		Parses the oid string and collects name from table.name.
//		If succeful returns length of generated name.
//		If specified oid is not found returns 0;
//
int generateOIDName( 
		char *name,
		const char *oidString,
		tUnit **pUnit // ultimate unit, it contains an additional information see tUnit
	)
{
	tUnit *t = OID_T;
	int q = OID_Q;
	int n;
	const char *s = oidString;

	if ( name ) *name = '\0';

	while ( true )
	{
		n = findUnit( t, q, s );

		if ( n == -1 ) // not found;
			return NULL;

		const char *cur = "UDF";
		if ( t[n].name ) cur = t[n].name;
		else
		if ( t[n].desc ) cur = t[n].desc;

		if ( name ) strcat( name, cur );

		int oidLen = strlen( t[n].oid );
		int strLen = strlen( s );

		if ( strLen == oidLen )
		{
			if ( pUnit )	*pUnit	= &(t[n]);

			return strlen( name ? name : cur );
		}
		else
		if (
//			( strLen < oidLen ) ||
			( strLen <= oidLen + 1) ||
			( t[n].uT == NULL )
		   )
			return 0;

		q = t[n].uQ;
		t = t[n].uT;
		s = s + oidLen + 1; // skip current oid string with following dot.
		if ( name ) strcat( name, "-" );
	}
}

//
//	Function: checkOIDName
//		Parses the oid string, gets name from table.name and checks 
//			if the names are compatible with corresponded names from nameT.
//		If sucessful returns x509e_ok.
//		Otherwise returns x509e_undefinedOIDType
//
int checkOIDName( 
		char *oidString,
		char **nameT, // NULL terminated table with szNames.
		int nameTSize	// If nameTSize == 0, nameT has to be NULL terminated and the 
						// function will sucessed only if whole OID name checked.
						// Otherwise checking successfully stops if the nameTSize is achieved
	)
{
	tUnit *t = OID_T;
	int q = OID_Q;
	int n;
	char *s = oidString;

	for( int c = 0; nameT[c]; c++ )
	{
		n = findUnit( t, q, s );

		if ( n == -1 ) // not found;
			return x509e_undefinedOIDType;

		const char *cur = 0;

		if ( t[n].name ) cur = t[n].name;
		else
		if ( t[n].desc ) cur = t[n].desc;

		if ( cur == 0 )
			return x509e_undefinedOIDType;

		if ( strcmp( cur, nameT[c] ) != NULL )
			return x509e_undefinedOIDType;

		int oidLen = strlen( t[n].oid );
		int strLen = strlen( s );

		if ( nameTSize && ( nameTSize == c + 1 ) ) // Success.
		{
			return x509e_ok;
		}

		if ( strLen == oidLen )
		{
			if ( nameT[c+1] == NULL )	// Success.
				return x509e_ok;
			else
				return x509e_undefinedOIDType;
		}
		else
		if (
//			( strLen < oidLen ) ||
			( strLen <= oidLen + 1) ||
			( t[n].uT == NULL )
		   )
			return x509e_undefinedOIDType;

		q = t[n].uQ;
		t = t[n].uT;
		s = s + oidLen + 1; // skip current oid string with following dot.
	}

	return x509e_internalError;
}