#include "x509_structs.h"
#include "oidFuncs.h"
#include "..\PGPPkts\pktdefs.h"

#include <time.h>
#include <stdio.h>

const char *getTagString( BYTE T )
{
	if ( T & asn1tag_OPTIONAL )
	{
		static char tmp[80];
		sprintf( tmp, "OPTIONAL[%d]", T & 0xF );
		return tmp;
	}


	switch ( T & asn1tag_TYPE )
	{
		case asn1tag_END_OF_CONTENTS_MARKER	: return "END OF CONTENTS MARKER";
		case asn1tag_BOOLEAN				: return "BOOLEAN";
		case asn1tag_INTEGER				: return "INTEGER";
		case asn1tag_BIT_STRING				: return "BIT STRING";
		case asn1tag_OCTET_STRING			: return "OCTET STRING";
		case asn1tag_NULL					: return "NULL";
		case asn1tag_OBJECT_IDENTIFIER		: return "OID";
		case asn1tag_REAL					: return "REAL";
		case asn1tag_ENUMERATED				: return "ENUMERATED";
		case asn1tag_SEQUENCE				: return "SEQUENCE";
		case asn1tag_SET					: return "SET";


		case asn1tag_NumericString		: return "Numeric String";
		case asn1tag_PrintableString 	: return "Printable String";
		case asn1tag_TelextexString		: return "Telextex String";
		case asn1tag_VideotexString		: return "Videotex String";
		case asn1tag_IA5String			: return "IA5 String";

		case asn1tag_UTCTime			: return "UTC Time";
		case asn1tag_GeneralizedTime	: return "Generalized Time";
		case asn1tag_GraphicString		: return "Graphic String";
		case asn1tag_VisibleString		: return "Visible String";
		case asn1tag_GeneralString		: return "General String";

		case asn1tag_UnicodeString		: return "Unicode String";

		default : return "Undefined ASN.1 TAG";
	}
}

///////////////////////////////////////////////////////////
//
//	struct 
//		x509_time & x509_validity
//

//
//	Function
//		x509_parseNumericString
//	checks if characters is in range '0' up to '9' and 
//	accumulates the integer value in res
//
static int x509_parseNumericString(
	int *res,	// pointer to resulting integer value
	BYTE *str,	// string buffer
	int q		// the character quantity to be parsed
	)
{
	if ( q > sizeof( int ) )
		return x509e_internalError;

	DWORD val = 0;

	while( q-- )
	{
		if ( ( *str < '0' ) || ( *str > '9' ) )
			return x509e_invalidPkt;

		val = ( val * 10 ) + ( *(str++) - '0' );
	}

	*res = val;

	return x509e_ok;
}


int x509_time::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );

	if ( res != x509e_ok )
		return res;

	BOOL bUTCTime = ( T == asn1tag_UTCTime );

	DWORD requiredLength = bUTCTime ? 0x0d : 0x0f;

	if ( requiredLength != L )
		return x509e_invalidPktSize;

	// B = "YYYYMMDDHHMMSSZ"

	if ( B[ requiredLength - 1 ] != 'Z' )
		return x509e_invalidPkt;

	if ( 
		x509_parseNumericString( &second, B + requiredLength -  3, 2 ) ||
		x509_parseNumericString( &minute, B + requiredLength -  5, 2 ) ||
		x509_parseNumericString( &hour  , B + requiredLength -  7, 2 ) ||
		x509_parseNumericString( &day   , B + requiredLength -  9, 2 ) ||
		x509_parseNumericString( &month , B + requiredLength - 11, 2 ) ||
		x509_parseNumericString( &year	 , B + requiredLength - 13, 2 )
		)
			return x509e_invalidPkt;

	int centenary;

	if ( bUTCTime )
	{
		// 4.1.2.5.1  UTCTime - RFC 2459

		if ( year >= 50 )
			centenary = 19;
		else
			centenary = 20;
	}
	else
	if ( x509_parseNumericString( &centenary	, B, 2 ) )
			return x509e_invalidPkt;

	year += centenary * 100;

	return x509e_ok;
}

int x509_time::compare( int p_year, int p_month, int p_day, int p_hour, int p_minute, int p_second )
{
	if ( p_year < year )	return 1;
	if ( p_year > year )	return -1;
	if ( p_month < month )	return 1;
	if ( p_month > month )	return -1;
	if ( p_day < day ) return 1;
	if ( p_day > day ) return -1;
	if ( p_hour < hour ) return 1;
	if ( p_hour > hour ) return -1;
	if ( p_minute < minute ) return 1;
	if ( p_minute > minute ) return -1;
	if ( p_second < second ) return 1;
	if ( p_second > second ) return -1;

	return 0;
}

int x509_time::prepareForWritting( int ye, int mo, int da, int ho, int mi, int se )
{
	year = ye;
	month = mo;
	day = da;
	hour = ho;
	minute = mi;
	second = se;

	bool isUTCTime = year < 2049;

	int requiredLength = 0x0f;
	BYTE tag = asn1tag_GeneralizedTime;

	if ( isUTCTime )
	{
		requiredLength = 0x0d;
		tag = asn1tag_UTCTime;
	}

	return x509_struct::prepareForWritting( tag, requiredLength );
}

int x509_time::writeBody()
{
	char time[ 0x10 ];

	bool isUTCTime = year < 2049;

	if ( year > 4000 )
	{
		assert( false );
		return x509e_internalError;
	}

	if ( isUTCTime )
	{
		int ye;
		if ( year >= 2000 )
			ye = year - 2000;
		else
			ye = year - 1900;
		// YYMMDDHHMMSSZ - asn1tag_GeneralizedTime
		sprintf( time, "%02d%02d%02d%02d%02d%02dZ", ye, month, day, hour, minute, second );
	}
	else
	{
		// YYYYMMDDHHMMSSZ - asn1tag_GeneralizedTime

		char pattern[] = "%04d%02d%02d%02d%02d%02dZ";
		sprintf( time, pattern, year, month, day, hour, minute, second );
	}

	if ( strlen( time ) != (int)L ) return x509e_internalError;

	memcpy( B, time, L );

	return x509e_ok;
}

int x509_validity::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );

	if ( res != x509e_ok )
		return res;

	if ( ( T & asn1tag_TYPE ) != asn1tag_SEQUENCE )
		return x509e_invalidPkt;

	res = notBefore.parse( B, L );
	if ( res != x509e_ok )
		return res;

	return notAfter.parse( B + notBefore.TL, L - notBefore.TL );
}

bool x509_validity::isValid()
{
	time_t t = time(0);
	struct tm *gmt = gmtime( &t );

	if ( notBefore.compare( 1900 + gmt->tm_year, gmt->tm_mon + 1, gmt->tm_mday, gmt->tm_hour, gmt->tm_min, gmt->tm_sec ) > 0 )
		return false;
	if ( notAfter.compare( 1900 + gmt->tm_year, gmt->tm_mon + 1, gmt->tm_mday, gmt->tm_hour, gmt->tm_min, gmt->tm_sec ) < 0 )
		return false;

	return true;    
}


//
//	end of x509_validity
//
//////////////////

///////////////////////////////////////////////////////////
//
//	struct 
//		x509_bitString
//

int x509_bitString::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );

	if ( res != x509e_ok )
		return res;

	if ( T != asn1tag_BIT_STRING )
		return x509e_invalidPkt;

//	By samples 
//		if first byte >= 8 so unused bit 
//		does not use

	if ( B[0] < 8  )
	{
		unusedBits = B[0];
		B += 1;
		L -= 1;
	}
	else
		unusedBits = 0;

	return x509e_ok;
}

int x509_bitString::prepareForWritting( unsigned int bodySize )
{
	return x509_struct::prepareForWritting( asn1tag_BIT_STRING, bodySize + 1 );
}

int x509_bitString::writeBody( BYTE *body, unsigned int bodySize )
{
	if ( B == NULL ) return x509e_internalError;

	if ( bodySize != L ) return x509e_internalError;

	memcpy( B, body, bodySize );

	return ppe_ok;
}

int x509_bitString::writeHeader(
	BYTE *buff,		// Buffer for header writting
	int buffSize,	// Size of allocated buffer
	int *pPos		// Offset position for next unit
	)
{
	int res = x509_struct::writeHeader( buff, buffSize, pPos );
	if ( ( res != 0 ) && ( B != NULL ) )
	{
		// Write unused bites count.
		*B = 0; B++; L--;
	}

	return res;
}

//
//	end of x509_bitString
//
//////////////////

///////////////////////////////////////////////////////////
//
//	struct 
//		x509_struct
//

//
// Function: findEndOfContentMarker.
//	Parses the constructed sequence or another field with undefined length,
//	 the sequence has to be finished by corresponded END_OF_CONTENTS_MARKER,
//	Stops when the asn1tag_END_OF_CONTENTS_MARKER is found.
//
static int findEndOfContentMarker( BYTE *ptr, int size, int *pParsedSize )
{
	x509_struct s;
	int res = 0;

	int parsed = 0;

	while (true)
	{
		res = s.parse( ptr, size );

		if ( res != x509e_ok ) return res;

		ptr += s.TL;
		size -= s.TL;

		parsed += s.TL;

		if ( s.T == asn1tag_END_OF_CONTENTS_MARKER )
			break;
	}

	*pParsedSize = parsed;

	return x509e_ok;

}

int x509_struct::prepareForWritting( BYTE tag, unsigned int bodySize )
{
	T = tag;
	L = bodySize;
	TL = makeHeader( L ) + L;
	return TL;
}

bool x509_struct::isOptional()
{
	assert( T != 0 );
	bool optional = (( T & asn1tag_OPTIONAL ) != 0 );
	return optional;
}

int x509_struct::getOptionalNum()
{
	if ( !isOptional() )
		return -1;

	return int ( T & 0x0F );
}

int x509_struct::writeBody( BYTE *body, unsigned int bodySize )
{
	if ( bodySize != L ) return x509e_internalError;
	if ( B == NULL ) return x509e_internalError;

	memcpy( B, body, L );
	return x509e_ok;
}

int x509_struct::writeHeader( BYTE *buff, int buffSize, int *pPos  )
{
	int pos = 0;

	if ( pPos )
		pos = *pPos;

	if ( (int)TL + pos > buffSize )
		return 0;

	if ( pPos )
		*pPos = pos + TL;

	int hdrSize = makeHeader( L, T, buff + pos );

	B = buff + pos + hdrSize;

	//
	// if pPos == NULL so we are not going to write other structures to the buffer,
	//	so all buffer has to filled by the structure only,
	//	so TL == bufferSize
	//
	if ( ( pPos == NULL ) && ( TL != buffSize ) )
	{
		assert( false );
		return 0;
	}

	return hdrSize;
}

int x509_struct::makeHeader(
		unsigned int bodySize,
		BYTE tag,
		BYTE *hdr // If not NULL is poiter to byte buffer maxHdrSize
		)
{
	if ( hdr ) *(hdr++) = tag;

	int hdrSize = 2;

	if ( bodySize < 0x80 )
	{
		if ( hdr ) *(hdr++) = bodySize;
	}
	else
	{
		int ll = 0; // Length of length

		for ( unsigned int count = bodySize; count; count = count >> ( 8 ) )
			ll++;

		hdrSize += ll;

		if ( hdr )
		{
			*(hdr++) = 0x80 | (BYTE)ll;
			while( ll-- )
			{
				*(hdr++) = BYTE ( bodySize >> (ll * 8) );
			}
		}
	}

	return hdrSize;
}


int x509_struct::parse( const BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize, int checkTag )
{
	if ( pktSize < 2 )
		return x509e_invalidPktSize;

	H = (BYTE *)pkt;

	BYTE l = 0; // length of length.

	T = pkt[0];

	L = pkt[1];

	if ( L > 0x84 )
		return x509e_invalidPkt;

	TL = 2;

	if ( L == 0x80 )
	{ 
		uL = true;
//		TL = pktSize;
//		L = pktSize - 2;
		B = (BYTE *)pkt + 2;

		int res = findEndOfContentMarker( B, pktSize - 2, (int *)&L );

		if ( res != x509e_ok )
			return res;

		TL = L + 2;

		if ( pPkt ) *pPkt = (BYTE *)pkt + TL;
		if ( pPktSize ) *pPktSize = pktSize  - TL;
	}
	else
	{
		uL = false;
		if ( L > 0x80 )
		{
			l = BYTE(L) & ~0x80;

			TL += l;

			L = 0;

			for( int i = 0 ; i < l; i++ )
			{
				L = ( L << 8 ) | pkt[ 2 + i ];
			}
		}


		TL += L;

		if ( (int)TL > pktSize )
			return x509e_invalidPktSize;

		B = (BYTE *)pkt + l + 2;

		if ( pPkt ) *pPkt = (BYTE *)pkt + TL;
		if ( pPktSize ) *pPktSize = pktSize  - TL;
//		if ( pPkt ) *pPkt += TL;
//		if ( pPktSize ) *pPktSize -= TL;
	}

	if ( ( checkTag != -1 ) && ( checkTag != T ) && ( checkTag != ( T & asn1tag_TYPE ) ) )
		return x509e_invalidPktTag;

	return x509e_ok;
}

x509_struct::x509_struct()
{
	clear();
}

//
//	end of x509_struct
//
//////////////////

///////////////////////////////////////////////////////////
//
//	struct 
//		x509_integer
//

int x509_integer::prepareForWritting( unsigned int val )
{
	lastNonZero = sizeof( unsigned int ) - 1; // If val == 0 write one zero byte.

	for( int i = sizeof( unsigned int ); i--; )
	{
		body[i] = BYTE(val);
		if ( body[i] )
			lastNonZero = i;

		val = val >> 8;
	}

	return x509_struct::prepareForWritting( asn1tag_INTEGER, sizeof( unsigned int ) - lastNonZero );
}

int x509_integer::writeBody()
{
	return x509_struct::writeBody( body + lastNonZero, sizeof( unsigned int ) - lastNonZero );
}

int x509_integer::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	return x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_INTEGER );
}

int x509_integer::compare( unsigned int x )
{
	unsigned int p = 0;
	unsigned int v = 0;

	if ( L > sizeof( v ) )
	{
		for( ; p < (L - sizeof( v )); p++ )
			if ( B[p] != 0 )
				return 1;

		p = L - sizeof( v );
	}
	else
		p = 0;


	for ( ; p < L; p++ )
		v = (v << 8) + B[p];

	if ( v > x )
		return 1;
	else
	if ( v < x )
		return -1;
	else
		return 0;
}

int x509_integer::compare( x509_integer *x )
{
	unsigned int p = 0;

	if ( L > x->L )
	{
		for ( ; p < ( L - x->L ); p++ )
			if ( B[p] != 0 )
				return 1;

		return memcmp( B + p, x->B, x->L );
	}
	else
	{
		for ( ; p < ( x->L - L ); p++ )
			if ( x->B[p] != 0 )
				return -1;

		return memcmp( x->B + p, B, L );
	}
}

int x509_integer::getUIValue( unsigned int *pVal )
{
	if ( L > sizeof( *pVal ) )
		return x509e_bufferTooSmall;

	*pVal = 0;

	for( unsigned int i = 0; i < L; i++ )
	{
		*pVal = ( *pVal << 8 ) | B[i];
	}

	return x509e_ok;
}

//
//	end of x509_integer
//
//////////////////


///////////////////////////////////////////////////////////
//
//	struct 
//		x509_algorithmID
//

int x509_algorithmID::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	tu = NULL;

	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	BYTE *pos = B;
	int size = L;

	res = oid.parse( pos, size, &pos, &size );
	if ( res != x509e_ok )	return res;

	if ( generateOIDName( NULL, oid.oidStr, &tu ) == 0 )
		return x509e_undefinedAlgorithm;

	if ( size != 0 )
	{
		res = params.parse( pos, size, &pos, &size );
		if ( res != x509e_ok )	return res;

		if ( size != 0 )
			return x509e_invalidPktSize;
	}

	return x509e_ok;
}

//
//	end of x509_algorithmID
//
//////////////////

