#include "x509_structs.h"
#include "oidFuncs.h"
#include "../hashAlgs/hashAlgs.h"
#include "../publicKeyAlgs/publicKeyAlgs.h"
#include "../SMemory/SBlocks.h"
#include "../PGPPkts/pktdefs.h"


///////////////////////////////////////////////////////////
//
//	struct 
//		x509_tbsCertificate
//

int x509_tbsCertificate::getSubjectKeyID( BYTE **pKeyIDPtr, int *pKeyIDSize )
{
	return extensions.getSubjectKeyID( pKeyIDPtr, pKeyIDSize );
}

int x509_tbsCertificate::generateSubjectKeyID( BYTE *pKeyIDPtr, int *pKeyIDSize )
{
	int hashID = pgphaSHA1;

	int digetSize = 0;
	int res = ha_getDigestSize( &digetSize, hashID );
	if ( res != hae_ok )
		return res;

	if ( *pKeyIDSize < digetSize )
	{
		*pKeyIDSize = digetSize;
		return pkae_bufferSizeTooSmall;
	}

	*pKeyIDSize = digetSize;

	res = ha_makeDigestEx( hashID, pKeyIDPtr, subjectPKInfo.H, subjectPKInfo.TL );

	return res;
}



int x509_tbsCertificate::getIssuerKeyID( BYTE **pKeyIDPtr, int *pKeyIDSize )
{
	return extensions.getAuthorityKeyID( pKeyIDPtr, pKeyIDSize );
}

int x509_tbsCertificate::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );

	if ( res != x509e_ok )
		return res;

	BYTE *pos;
	int size;

	pos  = B;
	size = L;

	//
	//	Parses version number - OPTIONAL
	//

	x509_struct tmp;

	res = tmp.parse( pos, size );
	if ( res != x509e_ok ) return res;

	if ( tmp.isOptional() )
	{
		if ( tmp.getOptionalNum() != 0 )
			return x509e_internalError;

		pos += tmp.TL;
		size -= tmp.TL;

		res = tmp.parse( tmp.B, tmp.L );
		if ( res != x509e_ok ) return res;

		if ( tmp.L > 1 )
			return x509e_internalError;

		version = *(tmp.B) + 1;
	}
	else
		version = 1;

	res = serialNumber.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = signatureAlg.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = issuerName.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

#ifdef _DEBUG

	int nameBuffSize = 0;

	res = issuerName.getAttributesString( NULL, &nameBuffSize );

	if ( res != x509e_bufferTooSmall )
		return res;

	char *nameBuff = new char[ nameBuffSize ];

	if ( nameBuff == NULL )
		return x509e_memoryAllocation;

	res = issuerName.getAttributesString( nameBuff, &nameBuffSize );

	delete nameBuff;

	if ( res != x509e_ok ) return res;

#endif


	res = validity.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

//	bool valid = validity.isValid();

	res = subjectName.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = subjectPKInfo.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	while( size )
	{
		res = tmp.parse( pos, size, &pos, &size );
		if ( res != x509e_ok ) return res;

		if ( !tmp.isOptional() )
			return x509e_internalError;

		switch ( tmp.getOptionalNum() )
		{
			case 1: res = issuerUniqueID.parse( tmp.B, tmp.L ); break;
			case 2: res = subjectUniqueID.parse( tmp.B, tmp.L ); break;
			case 3: 
			{
				res = extensions.parse( tmp.B, tmp.L );
				if ( res != x509e_ok )
					return res;

				extensions.checkBasicConstraint();
				extensions.getAuthorityKeyID();
				extensions.getSubjectKeyID();
				break;
				}
			default: res = x509e_internalError;
		}

		if ( res != x509e_ok ) return res;
	}

	return x509e_ok;
}

int x509_tbsCertificate::createPacket(
	BYTE *buff, int *pBuffSize,
	HPKA pkaSubject, HPKA skaIssuer,
	sNameAttribute *rdnSubject, sNameAttribute *rdnIssuer, 
	int nbYe, int nbMo, int nbDa, // Not before Year, Month, Day
	int naYe, int naMo, int naDa, // Not after Year, Month, Day
	const char *signatureAlgOID, int iSerialNumber )
{
/*
	BYTE				version; // OPTIONAL [0]
	x509_integer		serialNumber;
	x509_algorithmID	signatureAlg;
	x509_name			issuerName; // non-empty distinguished name (DN)
	x509_validity		validity;
	x509_name			subjectName; // non-empty distinguished name (DN)
	x509_publicKeyInfo	subjectPKInfo;
	x509_bitString		issuerUniqueID; //	OPTIONAL [1]
	x509_bitString		subjectUniqueID;//	OPTIONAL [2]
	x509_extensions		extensions;		//	OPTIONAL [3]
*/
	issuerName.setList( rdnIssuer );
	subjectName.setList( rdnSubject );

	// 1. Prepares structs for writting and checks size.

	x509_struct opt[4];
	x509_extension sExt[3];
	x509_integer sVerInteger;
	int verNum = 2; // Value 2 means Version 3, version 3 allows using of extensions field.

	int keyBlockSize = 0;
	int res = pka_writeKeyToASN1Buffer( pkaSubject, NULL, &keyBlockSize );
	if ( res != pkae_bufferSizeTooSmall )
		return res;

	// Prepare public keys' UIDs

	int issuerUIDSize = 0;	cPlainTextBlock ptbIssuerUID;

	res = ska_generateLocalKeyID( skaIssuer, NULL, &issuerUIDSize );
	if ( res != pkae_bufferSizeTooSmall ) return res;
	if ( ptbIssuerUID.Allocate( issuerUIDSize ) == NULL ) return x509e_memoryAllocation;
	res = ska_generateLocalKeyID( skaIssuer, ptbIssuerUID.buffer, &issuerUIDSize );
	if ( res != pkae_ok ) return res;

	int subjectUIDSize = 0;	cPlainTextBlock ptbSubjectUID;

	res = pka_generateLocalKeyID( pkaSubject, NULL, &subjectUIDSize );
	if ( res != pkae_bufferSizeTooSmall ) return res;
	if ( ptbSubjectUID.Allocate( subjectUIDSize ) == NULL ) return x509e_memoryAllocation;
	res = pka_generateLocalKeyID( pkaSubject, ptbSubjectUID.buffer, &subjectUIDSize );
	if ( res != pkae_ok ) return res;

	int required = 
		x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			opt[0].prepareForWritting( makeOptionalTag(0), sVerInteger.prepareForWritting( verNum ) ) +
			serialNumber.prepareForWritting( iSerialNumber ) + 
			signatureAlg.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				signatureAlg.oid.prepareForWritting( signatureAlgOID ) +
				signatureAlg.params.prepareForWritting( asn1tag_NULL, 0 )
				) +
			issuerName.prepareForWritting()+ 
			validity.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				validity.notBefore.prepareForWritting( nbYe, nbMo, nbDa ) +
				validity.notAfter.prepareForWritting( naYe, naMo, naDa )
			) +
			subjectName.prepareForWritting() +
			keyBlockSize +
//			opt[1].prepareForWritting( makeOptionalTag(1), issuerUniqueID.prepareForWritting( issuerUIDSize ) ) +
//			opt[2].prepareForWritting( makeOptionalTag(2), subjectUniqueID.prepareForWritting( subjectUIDSize ) )
			opt[3].prepareForWritting( makeOptionalTag(3), 
				extensions.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
					sExt[0].prepareForWritting( 
						extensions.subjectKeyIdentifier.getOid(),
						extensions.subjectKeyIdentifier.prepareForWritting( subjectUIDSize )
					) +
					sExt[1].prepareForWritting(
						extensions.authorityKeyIdentifier.getOid(),
						extensions.authorityKeyIdentifier.prepareForWritting(
							issuerUIDSize, // Authority key ID size,
							rdnIssuer,		// Authority RDN
							iSerialNumber )
					)+
					sExt[2].prepareForWritting(
						extensions.basicConstraints.getOid(),
						extensions.basicConstraints.prepareForWritting( true )
					)
				)
			)
		);

	if ( *pBuffSize < required )
	{
		*pBuffSize = required;
		return x509e_bufferTooSmall;
	}

	// 2. Writes headers.

	*pBuffSize = required;

	if ( writeHeader( buff, required ) == 0 ) return x509e_internalError;

	int pos = 0;

	if ( opt[0].writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		if ( sVerInteger.writeHeader( opt[0].B, opt[0].L ) == 0 ) return x509e_internalError;
			if ( sVerInteger.writeBody() != ppe_ok ) return x509e_internalError;
	if ( serialNumber.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		if ( serialNumber.writeBody() != ppe_ok ) return x509e_internalError;

	if ( signatureAlg.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		int lPos = 0;
		if ( signatureAlg.oid.writeHeader( signatureAlg.B, signatureAlg.L, &lPos ) == 0 ) return x509e_internalError;
			if ( string2oid( signatureAlg.oid.B, 0, signatureAlgOID ) != (int)signatureAlg.oid.L )	return x509e_internalError;
		if ( signatureAlg.params.writeHeader( signatureAlg.B, signatureAlg.L, &lPos ) == 0 ) return x509e_internalError;

	if ( issuerName.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		if ( issuerName.writeBody() != x509e_ok ) return x509e_internalError;

	if ( validity.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		lPos = 0;
		if ( validity.notBefore.writeHeader( validity.B, validity.L, &lPos ) == 0 ) return x509e_internalError;
			if ( validity.notBefore.writeBody() != x509e_ok ) return x509e_internalError;
		if ( validity.notAfter.writeHeader( validity.B, validity.L, &lPos ) == 0 ) return x509e_internalError;
			if ( validity.notAfter.writeBody() != x509e_ok ) return x509e_internalError;
	if ( subjectName.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		if ( subjectName.writeBody() != x509e_ok ) return x509e_internalError;

	// Writes subject PKI
	if ( keyBlockSize > (int)( L - pos ) ) return x509e_internalError;
	res = pka_writeKeyToASN1Buffer( pkaSubject, B + pos, &keyBlockSize );
	if ( res != x509e_ok ) return x509e_internalError;
	pos += keyBlockSize;
/*
	if ( opt[1].writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		if ( issuerUniqueID.writeHeader( opt[1].B, opt[1].L ) == 0 ) return x509e_internalError;
			if ( issuerUniqueID.writeBody( ptbIssuerUID.buffer, issuerUIDSize ) != x509e_ok ) return x509e_internalError;
	if ( opt[2].writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
		if ( subjectUniqueID.writeHeader( opt[2].B, opt[2].L ) == 0 ) return x509e_internalError;
			if ( subjectUniqueID.writeBody( ptbSubjectUID.buffer, subjectUIDSize ) != x509e_ok ) return x509e_internalError;
*/
	if ( opt[3].writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
	if ( pos != (int)L )	return x509e_internalError;

		if ( extensions.writeHeader( opt[3].B, opt[3].L ) == 0 ) return x509e_internalError;
			pos = 0;
			if ( sExt[0].writeHeader( extensions.B, extensions.L, &pos ) == 0 ) return x509e_internalError;
			if ( extensions.subjectKeyIdentifier.	writeHeader( sExt[0].octet.B, sExt[0].octet.L ) == 0 ) return x509e_internalError;
			res = extensions.subjectKeyIdentifier.writeBody(ptbSubjectUID.buffer, subjectUIDSize); if ( res != x509e_ok ) return res;

			if ( sExt[1].writeHeader( extensions.B, extensions.L, &pos ) == 0 ) return x509e_internalError;
			if ( extensions.authorityKeyIdentifier.	writeHeader( sExt[1].octet.B, sExt[1].octet.L ) == 0 ) return x509e_internalError;
			res = extensions.authorityKeyIdentifier.writeBody(ptbIssuerUID.buffer, issuerUIDSize); if ( res != x509e_ok ) return res;

			if ( sExt[2].writeHeader( extensions.B, extensions.L, &pos ) == 0 ) return x509e_internalError;
			if ( extensions.basicConstraints.		writeHeader( sExt[2].octet.B, sExt[2].octet.L ) == 0 ) return x509e_internalError;
			res = extensions.basicConstraints.writeBody(); if ( res != x509e_ok ) return res;

			if ( pos != (int)(extensions.L) ) return x509e_internalError;

	return x509e_ok;
}

//
//	end of x509_tbsCertificate
//
//////////////////


///////////////////////////////////////////////////////////
//
//	struct 
//		x509_certificate
//

int x509_certificate::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );

	if ( res != x509e_ok )
		return res;

	BYTE *pos = B;
	int size = L;

	res = tbsCertificate.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = signatureAlg.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = signatureValue.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	return res;
}

int x509_certificate::checkValidityPeriod()
{
	if ( B == NULL )
		return x509e_internalError;

	if ( tbsCertificate.validity.isValid() )
		return x509e_ok;

	return x509e_validityPeriodFinished;
}

int x509_certificate::checkPurpose( int purpose, BOOL *pCritical )
{
	BOOL critical = FALSE;

	const char *extKeyUsageOID = "2.5.29.37";

	BYTE *buf;
	int bufSize;

	int res = tbsCertificate.extensions.getExtensionValue( extKeyUsageOID, &critical, &buf, &bufSize );

	if ( res != x509e_ok ) // Key usage extension is not present.
		return res;

	x509_struct s;

	res = s.parse( buf, bufSize, NULL, NULL, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	cPlainTextBlock ptbOID;

	ptbOID.WriteString( "1.3.6.1.5.5.7.3.X", true ); // see pkix_kp_T.

	char *purposeOID = (char *)ptbOID.buffer;

	if ( purposeOID == NULL )
	    return x509e_memoryAllocation;
	    
	// strncpy( purposeOID, "1.3.6.1.5.5.7.3.X", sizeof( purposeOID ) ); // see pkix_kp_T.

	int numPos = strlen( purposeOID ) - 1;

	switch( purpose )
	{
		case kp_serverAuth:		purposeOID[ numPos ] = '1'; break;
		case kp_clientAuth: 		purposeOID[ numPos ] = '2'; break;
		case kp_codeSigning:		purposeOID[ numPos ] = '3'; break;
		case kp_emailProtection:	purposeOID[ numPos ] = '4'; break;
		case kp_timeStamping:		purposeOID[ numPos ] = '8'; break;

		default : return x509e_internalError; // Undefined purpose specified.
	}

	x509_OID oid;

	res = oid.parse( s.B, s.L );
	if ( res != x509e_ok ) return res;

	if ( strcmp( oid.oidStr, purposeOID ) == 0 )
	{
		if ( pCritical )
			*pCritical = critical;

		return x509e_ok;
	}
	else
		return x509e_extensionIsNotPresent;
}


int x509_certificate::verify( BYTE haID, BYTE pkID, HPKA pka )
{
	(void)pkID; // Unreferenced formal parameter.
	//
	//	1. Make digest.
	//
	HHA ha;
	int res = ha_createAlgorithm( &ha, haID );
	if ( res != hae_ok ) return res;

	int digestSize = ha_getDigestSize( ha );

	BYTE *digest = new BYTE[ digestSize ];
	if ( digest == NULL )
	{
		ha_free( ha );
		return x509e_memoryAllocation;
	}

	res = ha_makeDigest( ha, digest, this->B, tbsCertificate.TL );

	ha_free( ha );

	if ( res != hae_ok )
	{
		delete digest;
		return res;
	}

	//
	//	2. Verification
	//

	res = pka_verify_X509( pka, signatureValue.H, signatureValue.TL, digest, digestSize );

	delete digest;

	return res;
}

int x509_certificate::verify( x509_publicKeyInfo *issuerPKI )
{
	tUnit *tu;

	if ( generateOIDName( NULL, signatureAlg.oid.oidStr, &tu ) == 0 )
		return x509e_undefinedOIDType;

	BYTE haID = tu->hAlg;
	BYTE pkID = tu->pAlg;

	if ( issuerPKI->algOID.tu->pAlg != pkID )
		return x509e_invalidPkt;

	HPKA pka;

	int res = pka_createAlgorithm_X509( &pka, issuerPKI );
	if ( res != pkae_ok )
		return res;

	res = verify( haID, pkID, pka );

	pka_free( pka );

	return res;
}


//int x509_certificate::verify( x509_publicKeyInfo *issuerPKI )
//{
//	tUnit *tu;
//
//	if ( generateOIDName( NULL, signatureAlg.oid.oidStr, &tu ) == 0 )
//		return x509e_undefinedOIDType;
//
//	BYTE haID = tu->hAlg;
//	BYTE pkID = tu->pAlg;
//
//	if ( issuerPKI->algOID.tu->pAlg != pkID )
//	{
//		return x509e_invalidPkt;
//	}
//
//	//
//	//	1. Make digest.
//	//
//	HHA ha;
//	int res;
//	res = ha_createAlgorithm( &ha, haID );
//	if ( res != hae_ok ) return res;
//
//	int digestSize = ha_getDigestSize( ha );
//
//	BYTE *digest = new BYTE[ digestSize ];
//	if ( digest == NULL )
//	{
//		ha_free( ha );
//		return x509e_memoryAllocation;
//	}
//
//	res = ha_makeDigest( ha, digest, this->B, tbsCertificate.TL );
//
//	ha_free( ha );
//
//	if ( res != hae_ok )
//	{
//		delete digest;
//		return res;
//	}
//
//	//
//	//	2. Verification
//	//
//
//	HPKA pka;
//
//	res = pka_createAlgorithm_X509( &pka, issuerPKI );
//	if ( res != pkae_ok )
//	{
//		delete digest;
//		return res;
//	}
//
//	res = pka_verify_X509( pka, signatureValue.H, signatureValue.TL, digest, digestSize );
//
//	pka_free( pka );
//
//	return res;
//}

int x509_certificate::createPacket(
	BYTE *buff, int *pBuffSize, 
	HPKA pkaSubject, HPKA skaIssuer,
	sNameAttribute *rdnSubject, sNameAttribute *rdnIssuer, 
	int nbYe, int nbMo, int nbDa, // Not before Year, Month, Day
	int naYe, int naMo, int naDa, // Not after Year, Month, Day
	const char *signatureAlgOID, int iSerialNumber )
{
	//
	//	1. Checks signature algorithm value
	//
	tUnit *tuSigAlg;
	if ( generateOIDName( NULL, signatureAlgOID, &tuSigAlg )== 0 )
		return x509e_undefinedOIDType;

	if ( tuSigAlg->pAlg != ska_getAlgID( skaIssuer ) )
		return x509e_unexpectedOIDType;

	//
	//	2. Gets required size.
	//
	int required = 0;

	int res = tbsCertificate.createPacket(
		NULL, &required,
		pkaSubject, skaIssuer,
		rdnSubject, rdnIssuer, 
		nbYe, nbMo, nbDa,
		naYe, naMo, naDa,
		signatureAlgOID,
		iSerialNumber );


	if ( res != x509e_bufferTooSmall )
		return res;

	assert( (int)tbsCertificate.TL == required );

	int digestBuffSize = 0;
	res = ha_getDigestSize( &digestBuffSize,  tuSigAlg->hAlg );

	if ( res != hae_ok )
		return res;

	int sigSize = 0;
	res = ska_sign_X509_BIT_STRING( skaIssuer, NULL, &sigSize, NULL, 0 );

	if ( res != pkae_bufferSizeTooSmall )
		return res;

	required =
		prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
			tbsCertificate.TL
			+
			signatureAlg.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				signatureAlg.oid.prepareForWritting( signatureAlgOID ) +
				signatureAlg.params.prepareForWritting( asn1tag_NULL, 0 )
				)
			+
			sigSize
			);

	if ( *pBuffSize < required )
	{
		*pBuffSize = required;
		return x509e_bufferTooSmall;
	}

	//
	//	3. Writes required buffer
	//

	*pBuffSize = required;

	if ( writeHeader( buff, required ) == 0 ) return x509e_internalError;

	int pos = required;

	res = tbsCertificate.createPacket(
		B, &pos,
		pkaSubject, skaIssuer,
		rdnSubject, rdnIssuer, 
		nbYe, nbMo, nbDa,
		naYe, naMo, naDa,
		signatureAlgOID,
		iSerialNumber );

	if ( res != x509e_ok )
		return res;

//	if ( tbsCertificate.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
	if ( signatureAlg.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
	 int lPos = 0;
	 if ( signatureAlg.oid.writeHeader( signatureAlg.B, signatureAlg.L, &lPos ) == 0 ) return x509e_internalError;
	  if ( string2oid( signatureAlg.oid.B, 0, signatureAlgOID ) != (int)signatureAlg.oid.L )	return x509e_internalError;
	 if ( signatureAlg.params.writeHeader( signatureAlg.B, signatureAlg.L, &lPos ) == 0 ) return x509e_internalError;

	//
	//	3. Makes digest and generaes signature.
	//		we can do it now because tbsCertificate is already filled.
	//

	HHA ha;
	res = ha_createAlgorithm( &ha, tuSigAlg->hAlg );
	if ( res != hae_ok )
		return res;

    if ( digestBuffSize != ha_getDigestSize( ha ) )
	{
		ha_free( ha );
		return x509e_internalError;
	}

//	x509_struct sDigestPackage; // SEQUENCE
//	 x509_algorithmID sAlgID;
//	 x509_struct	sValue; // OCTET_STRING

	int digestPackageSize = 
		sDigestPackage.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, 
			sAlgID.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				sAlgID.oid.prepareForWritting( getDigestAlgOidStr( tuSigAlg->hAlg ) ) +
				sAlgID.params.prepareForWritting( asn1tag_NULL, 0 ) 
			) +
			sValue.prepareForWritting( asn1tag_OCTET_STRING, digestBuffSize )
		);

	cPlainTextBlock ptb;
	if ( ptb.Allocate( digestPackageSize ) == NULL )
	{
		ha_free( ha );
		return x509e_memoryAllocation;
	}

	if ( sDigestPackage.writeHeader( ptb.buffer, digestPackageSize ) == 0 ) return x509e_internalError;
	 lPos = 0;
	 if ( sAlgID.writeHeader( sDigestPackage.B, sDigestPackage.L, &lPos ) == 0 ) return x509e_internalError;
	  int lPos0 = 0;
	  if ( sAlgID.oid.writeHeader( sAlgID.B, sAlgID.L, &lPos0 ) == 0 ) return x509e_internalError;
	   res = sAlgID.oid.writeBody(); if ( res != x509e_ok ) return res;
	  if ( sAlgID.params.writeHeader( sAlgID.B, sAlgID.L, &lPos0 ) == 0 ) return x509e_internalError;
	  if ( lPos0 != (int)(sAlgID.L) ) return x509e_internalError;
	 if ( sValue.writeHeader( sDigestPackage.B, sDigestPackage.L, &lPos ) == 0 ) return x509e_internalError;
	  if ( (int)(sDigestPackage.L) != lPos ) return x509e_internalError;
	  
	if ( (int)(sValue.L) != digestBuffSize ) return x509e_internalError;
	res = ha_makeDigest( ha, sValue.B, B, tbsCertificate.TL );

	ha_free( ha );

	if ( res != hae_ok )
		return res;

	int newSigSize = sigSize;

	res = ska_sign_X509_BIT_STRING( skaIssuer, B + pos, &newSigSize, ptb.buffer, ptb.bufferSize );
	if ( res != x509e_ok )
		return res;

	return ppe_ok;
}

//
//	end of x509_certificate
//
//////////////////
