#include "x509_structs.h"

//
//	Structure x509_extensions
//

int x509_extensions::getAuthorityKeyID(
		BYTE **pKeyIDPtr, int *pKeyIDSize,
		int *pGeneralNameChoise,
		BYTE **pGeneralName, int *pGeneralNameSize,
		int *pSerialNumber )
{
	const char *extOID = "2.5.29.35";
	BOOL bCritical = false;
	BYTE *ptr;
	int size;

	// All the fields are optional so if a filed does not exist it's value is NULL.

	if ( pKeyIDPtr != NULL ) *pKeyIDPtr = NULL;
	if ( pKeyIDSize != NULL )*pKeyIDSize = 0;
	if ( pGeneralNameChoise != NULL ) *pGeneralNameChoise = 0;
	if ( pGeneralName != NULL ) *pGeneralName = NULL;
	if ( pGeneralNameSize != NULL ) *pGeneralNameSize = 0;
	if ( pSerialNumber != NULL ) *pSerialNumber = 0;

	int res = getExtensionValue( extOID, &bCritical, &ptr, &size );
	if ( res != x509e_ok )
		return res;

	x509_struct seq;
	res = seq.parse( ptr, size, NULL, NULL, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res != x509e_ok ) return res;

	ptr = seq.B;
	size = seq.L;

	while( size )
	{
		x509_struct tmp;

		res = tmp.parse( ptr, size, &ptr, &size );
		if ( res != x509e_ok ) return res;

		if ( !tmp.isOptional() )
			return x509e_internalError;

		int num = tmp.getOptionalNum();

		switch( num )
		{
			case 0: // Key identifier
			{
				res = getKeyID( tmp.B, tmp.L, pKeyIDPtr, pKeyIDSize );
				if ( res != x509e_ok )
					return res;
				break;
			}
			case 1: // General names
			{
				x509_struct sChoice;

				res = sChoice.parse( tmp.B, tmp.L );
				if ( res != x509e_ok ) return res;

				if ( !sChoice.isOptional() )
					return x509e_internalError;

				if ( pGeneralNameChoise )
					*pGeneralNameChoise = sChoice.getOptionalNum();

				if ( pGeneralName != NULL ) *pGeneralName = sChoice.B;
				if ( pGeneralNameSize != NULL ) *pGeneralNameSize = sChoice.L;

				if ( tmp.L != sChoice.TL )
					return x509e_internalError;

				break;
			}
			case 2: // Serial number
			{
				if ( pSerialNumber != NULL )
				{
					for ( int i = 0; i < (int)sizeof( int ); i++ )
					{
						int pos = tmp.L - ( i + 1 );
						if ( pos < 0 )
							break;

						*pSerialNumber += ( tmp.B[ pos ] << (i * 8) );
					}
				}
				break;
			}
			default: 
				return x509e_internalError;
		}
	}

	return x509e_ok;
}

int x509_extensions::getKeyID( BYTE *ptr, int size, BYTE **pKeyIDPtr, int *pKeyIDSize )
{
	if ( size < 20 )
		return x509e_internalError;

	if ( size > 20 )
	{
		if ( ( size == 22 ) && ( ptr[0] == 0x04 ) && ( ptr[1] == 20 ) )
		{
			ptr += 2;
			size -= 2;
		}
		else
			return x509e_internalError;
	}

	if ( pKeyIDPtr != NULL ) *pKeyIDPtr = ptr;
	if ( pKeyIDSize != NULL ) *pKeyIDSize = size;

	return x509e_ok;
}

int x509_extensions::getSubjectKeyID( BYTE **pKeyIDPtr, int *pKeyIDSize )
{
	const char *extID = "2.5.29.14";
	BOOL bCritical = false;
	BYTE *ptr;
	int size;

	int res = getExtensionValue( extID, &bCritical, &ptr, &size );
	if ( res != x509e_ok ) return res;

	return getKeyID( ptr, size, pKeyIDPtr, pKeyIDSize );
}

int x509_extensions::checkBasicConstraint( BYTE *pBoolCA, unsigned int *pPathLength )
{
	const char *extBasicConstraint = "2.5.29.19";
	BOOL bCritical = false;
	BYTE *ptr;
	int size;

	int res = getExtensionValue( extBasicConstraint, &bCritical, &ptr, &size );
	if ( res != x509e_ok )
		return res;

	x509_struct seqBC;
	res = seqBC.parse( ptr, size, NULL, NULL, asn1tag_CONSTRUCTED_SEQUENCE );
	if ( res != x509e_ok ) return res;

	x509_struct sBoolCA;
	unsigned int pathLength = (unsigned int)(-1);

	res = sBoolCA.parse( seqBC.B, seqBC.L, &ptr, &size, asn1tag_BOOLEAN );
	if ( res != x509e_ok ) return res;

	if ( pBoolCA != NULL )
		*pBoolCA = sBoolCA.B[0];

	if ( size != 0 )
	{
		x509_integer sPathLength;
		res = sPathLength.parse( ptr, size, &ptr, &size );
		if ( res != x509e_ok ) return res;

		res = sPathLength.getUIValue( &pathLength );
		if ( res != x509e_ok ) return res;
	}

	if ( pPathLength != NULL )
		*pPathLength = pathLength;

	if ( size != 0 )
		return x509e_internalError;

	return x509e_ok;
}

//
//	Parses all extensions and searchs specified extension.
//	returns x509e_ok if the extension found.
//
int x509_extensions::getExtensionValue(
	const char *oid,
	BOOL *pCritical,
	BYTE **pValue,		// pointer to value body ( body of OCTET STRING ).
	int *pValueSize		// pointer to size
)
{
	int res;

	x509_struct ext;

	BYTE *pos = B;
	int size = L;

	while( size )
	{
		res = ext.parse( pos, size, &pos, &size, asn1tag_SEQUENCE );
		if ( res != x509e_ok ) return res;

		x509_OID sOID;

		BYTE *nextP;
		int nextS;

		res = sOID.parse( ext.B, ext.L, &nextP, &nextS );

		if ( strcmp( sOID.oidStr, oid ) != 0 )
			continue;

		x509_struct st;

		res = st.parse( nextP, nextS, &nextP, &nextS );
		if ( res != x509e_ok ) return res;

		if ( st.T == asn1tag_BOOLEAN )
		{
			if ( pCritical )
				*pCritical = st.B[0];

			res = st.parse( nextP, nextS, &nextP, &nextS );
			if ( res != x509e_ok ) return res;
		}

		if ( st.T != asn1tag_OCTET_STRING )
			return x509e_invalidPkt;

		*pValue = st.B;
		*pValueSize = st.L;

		return x509e_ok;
	}

	return x509e_extensionIsNotPresent;
}

//
//	End of x509_extensions structure
//
/////////////////////////////////

//
//	Structure x509_extension
//

int x509_extension::prepareForWritting( const char *oidStr, int bodySize, bool critical )
{
	return x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
		oid.prepareForWritting( oidStr ) +
		( critical ? bCritical.prepareForWritting( asn1tag_BOOLEAN, 1 ) : 0 ) +
		octet.prepareForWritting( asn1tag_OCTET_STRING, bodySize )
	);
}

int x509_extension::writeHeader( BYTE *buff, int buffSize, int *pPos )
{
	int res = x509_struct::writeHeader( buff, buffSize, pPos );
	if ( res == 0 ) return 0;

	int pos = 0;

	res = oid.writeHeader( B, L, &pos );
	if ( res == 0 ) return 0;
	res = oid.writeBody();
	if ( res != x509e_ok ) return 0;

	if ( bCritical.T == asn1tag_BOOLEAN ) // was inited at prepareForWritting
	{
		res = bCritical.writeHeader( B, L, &pos );
		if ( res == 0 ) return 0;
	}

	res = octet.writeHeader( B, L, &pos );
	if ( res == 0 ) return 0;

	if ( (int)L != pos )
		return 0;

	if ( bCritical.T == asn1tag_BOOLEAN ) // was inited at prepareForWritting
	{
		bCritical.B[0] = 0xFF;
	}

	return TL - L; // octet.writeBody( body, bodySize );
}
//
// end of x509_extension structure
//
///////////////////////////////////////

//
//	Structure x509_ce_authorityKeyIdentifier
//

int x509_ce_authorityKeyIdentifier::writeBody( BYTE *keyID, int keyIDSize )
{

	int res = opt[0].writeBody( keyID, keyIDSize );
	if ( res != x509e_ok ) return res;

	res = name.writeBody();
	if ( res != x509e_ok ) return res;

	return opt2.writeBody();
}

int x509_ce_authorityKeyIdentifier::writeHeader(
	BYTE *buff,		// Buffer for header writting
	int buffSize,	// Size of allocated buffer
	int *pPos		// Offset position for next unit
)
{
	int res = x509_struct::writeHeader( buff, buffSize, pPos );
	if ( res == 0 ) return res;

	int pos = 0; 

	opt[0].writeHeader( B, L, &pos );
	opt[1].writeHeader( B, L, &pos );
		choice.writeHeader( opt[1].B, opt[1].L );
		name.writeHeader( choice.B, choice.L );
	opt2.writeHeader( B, L, &pos );

	if ( pos != (int)L ) return 0;

	return res;
}
int x509_ce_authorityKeyIdentifier::prepareForWritting( int keyIDSize, sNameAttribute *rdn, int serNum )
{
	name.setList( rdn );

	int res = x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
		opt[0].prepareForWritting( makeOptionalTag(0) & ~asn1tag_CONSTRUCTED, keyIDSize ) +
		opt[1].prepareForWritting( makeOptionalTag(1), 
			choice.prepareForWritting( makeChoiceTag(4), // choice directoryName
				name.prepareForWritting() )
			) +
		opt2.prepareForWritting( serNum ) );

	opt2.T = makeOptionalTag(2) & ~asn1tag_CONSTRUCTED;

	return res;
}
//
//	End of x509_ce_authorityKeyIdentifier structure
//
/////////////////////////////////////////////////////

//
//	Structure x509_ce_basicConstraints
//

int x509_ce_basicConstraints::prepareForWritting( bool CA, int pathLengthConstraint )
{
	bCA = CA;

	return 
		x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, 
			sCA.prepareForWritting( asn1tag_BOOLEAN, 1 ) +
			( pathLengthConstraint == -1 ? 0: // default
				opt.prepareForWritting( makeOptionalTag(0), 
					pathLength.prepareForWritting( pathLengthConstraint ) )
			)
		);
}

int x509_ce_basicConstraints::writeBody()
{
	int pos = 0;
	int res = sCA.writeHeader( B, L, &pos );
	if ( res == 0 ) 	return x509e_internalError;

	sCA.B[0] = ( bCA ? 0xFF : 0 );

	if ( opt.T == 0 ) // Path Length Constrain was not inited.
		return x509e_ok;

	res = opt.writeHeader( B, L, &pos );
	if ( res == 0 ) 	return x509e_internalError;

	res = pathLength.writeHeader( opt.B, opt.L );
	if ( res == 0 ) 	return x509e_internalError;

	return pathLength.writeBody();
}

//
//	End of x509_ce_basicConstraints structure
//
///////////////////////////////////
