///////////////////////////
//
//	struct x509_name
//

#include "x509_structs.h"
#include "oidFuncs.h"
#include "../SMemory/SBlocks.h"
#include <string.h>

#ifndef _WIN32
#define _stricmp strcmp
#endif

/*
RFC 2459        Internet X.509 Public Key Infrastructure    January 1999
Housley, et. al.            Standards Track                    [Page 18]


   Name ::= CHOICE {
     RDNSequence }

   RDNSequence ::= SEQUENCE OF RelativeDistinguishedName

   RelativeDistinguishedName ::=
     SET OF AttributeTypeAndValue

   AttributeTypeAndValue ::= SEQUENCE {
     type     AttributeType,
     value    AttributeValue }

   AttributeType ::= OBJECT IDENTIFIER

   AttributeValue ::= ANY DEFINED BY AttributeType
*/

sNameAttribute::~sNameAttribute()
{
	if ( szValue != NULL )
		delete szValue;
	if ( szOID != NULL )
		delete szOID;
/*
	if ( next != NULL )
	{
		delete next;
		next = NULL;
	}
*/
}
const char * sNameAttribute::getName()
{
	const char *name;

	tUnit *unit;

	if ( generateOIDName( NULL, szOID, &unit ) == 0 )
		name = szOID;
	else name = unit->name ? unit->name : unit->desc;

	return name;
}

int sNameAttribute::getIdByName( const char *name )
{
	for ( int id = id_first; id <= id_last; id++ )
	{
		tUnit *unit;
		if ( generateOIDName( NULL, getOid( at_type(id) ), &unit ) == 0 )
			return -1;

		if ( _stricmp( unit->name, name ) == 0 )
			return id;
	}

	return -1;
}

int sNameAttribute::getIdByOID( const char *szOID )
{
	for ( int id = id_first; id <= id_last; id++ )
	{
		const char *tmpOID = getOid( at_type(id) );

		if ( _stricmp( tmpOID, szOID ) == 0 )
		{
			// Checks if unit with the oid presents in our table 
			tUnit *unit;
			if ( generateOIDName( NULL, tmpOID, &unit ) == 0 )
				return -1;

			return id;
		}
	}

	return -1;
}
//
//	Function: getOid
//	Gets oid string for a name attribute from from enum list
//
const char *sNameAttribute::getOid( at_type idNum )
{
	switch( idNum )
	{
		case id_at_commonName: return "2.5.4.3";
		case id_at_surname: return "2.5.4.4";
		case id_at_serialNumber: return "2.5.4.5";
		case id_at_countryName: return "2.5.4.6";
		case id_at_localityName: return "2.5.4.7";
		case id_at_stateOrProvinceName: return "2.5.4.8";
		case id_at_streetAddress: return "2.5.4.9";
		case id_at_organizationName: return "2.5.4.10";
		case id_at_organizationalUnitName: return "2.5.4.11";
		case id_at_title: return "2.5.4.12";
		case id_at_description: return "2.5.4.13";
		case id_at_postalAddress: return "2.5.4.16";
		case id_at_postOfficeBox: return "2.5.4.18";
		case id_at_telephoneNumber: return "2.5.4.20";
		case id_at_facsimileTelephoneNumber: return "2.5.4.23";
		case id_at_name: return "2.5.4.41";
		case id_at_givenName: return "2.5.4.42";
		case id_at_initials: return "2.5.4.43";
		case id_at_generationQualifier: return "2.5.4.44";
		case id_at_uniqueIdentifier: return "2.5.4.45";
		case id_at_dnQualifier: return "2.5.4.46";
		case pkcs9_emailAddress: return "1.2.840.113549.1.9.1";
	}

	return NULL;
}

sNameAttribute *sNameAttribute::create( const char *oidStr, const char *value, BYTE valueTag )
{
	//
	//	Creates empty unit
	//
	sNameAttribute *na = new sNameAttribute();
	if ( na == NULL ) return NULL;

	na->szOID = new char[ strlen( oidStr ) + 1 ];

	na->szValue = new char[ strlen( value ) + 1 ];

	if ( ( na->szValue == NULL ) || ( na->szOID == NULL ) )
	{
		delete na;
		return NULL;
	}

	strcpy( na->szValue, value );

	strcpy( na->szOID, oidStr );
	na->tag = valueTag;

	return na;
}

sNameAttribute *sNameAttribute::create( BYTE *OID, int OIDSize, BYTE *value, int valueSize, BYTE valueTag )
{
	char buf[ maxOIDStrSize ];
	if ( oid2string( buf, OID, OIDSize ) == NULL )
		return NULL;

	char *szValue = new char[ valueSize + 1 ];
	if ( szValue == NULL )
		return NULL;

	memcpy( szValue, value, valueSize );
	szValue[ valueSize ] = '\0';

	sNameAttribute *res = create( buf, szValue, valueTag );

	delete szValue;

	return res;
}


x509_name::x509_name()
{
}

x509_name::~x509_name()
{
/*
	for( ; head; )
	{
		sNameAttribute *t = head;
		head = head->next;
		delete t;
	}
*/
}

/*
4.1.2.4  Issuer

   The issuer field identifies the entity who has signed and issued the
   certificate.  The issuer field MUST contain a non-empty distinguished
   name (DN).  The issuer field is defined as the X.501 type Name.
   [X.501] Name is defined by the following ASN.1 structures:

   Name ::= CHOICE {
     RDNSequence }

   RDNSequence ::= SEQUENCE OF RelativeDistinguishedName

   RelativeDistinguishedName ::=
     SET OF AttributeTypeAndValue

   AttributeTypeAndValue ::= SEQUENCE {
     type     AttributeType,
     value    AttributeValue }

   AttributeType ::= OBJECT IDENTIFIER

   AttributeValue ::= ANY DEFINED BY AttributeType
*/

//
//	Parses the RDNSequence ::= SEQUENCE OF RelativeDistinguishedName
//
int x509_name::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );
	if ( res != x509e_ok )
		return res;

	if ( ( T & asn1tag_TYPE ) != asn1tag_SEQUENCE )
		return x509e_invalidPktTag;

	return readAttributes( B, L );
}

//
//	Parses the RelativeDistinguishedNames ::= SETs OF AttributeTypeAndValue
//
int x509_name::readAttributes( BYTE *pkt, int pktSize )
{
	while( pktSize )
	{
//
//0021 31 0b         11: . . . SET
//0023 30 09          9: . . . . SEQUENCE
//0025 06 03          3: . . . . . OID 2.5.4.6: C
//                     : 55 04 06
//0030 13 02          2: . . . . . PrintableString  'US'
//                     : 55 53

		//
		//	SET
		//
		x509_struct sSET;

		int res = sSET.parse( pkt, pktSize );
		if ( res != x509e_ok )
			return res;

		if ( ( sSET.T & asn1tag_TYPE ) != asn1tag_SET )
			return x509e_invalidPktTag;


		//
		//	SEQUENCE
		//
		x509_struct sSEQUENCE;

		res = sSEQUENCE.parse( sSET.B, sSET.L );
		if ( res != x509e_ok )
			return res;

		if ( ( sSEQUENCE.T & asn1tag_TYPE ) != asn1tag_SEQUENCE )
			return x509e_invalidPktTag;

		//
		//	OID
		//
		x509_struct sOID;

		res = sOID.parse( sSEQUENCE.B, sSEQUENCE.L );
		if ( res != x509e_ok )
			return res;

		if ( sOID.T != asn1tag_OBJECT_IDENTIFIER )
			return x509e_invalidPktTag;

		//
		//	OID's PrintableString
		//
		x509_struct sPrintableString;

		res = sPrintableString.parse( sSEQUENCE.B + sOID.TL, sSEQUENCE.L - sOID.TL );
		if ( res != x509e_ok )
			return res;

		if (( sPrintableString.T < asn1tag_PrintableString ) || ( sPrintableString.T > asn1tag_IA5String ) )
		{
			// v. 1.04.4 - 25.11.2014
			if ( sPrintableString.T != asn1tag_UndefinedString ) // Windows certificate viewer allows using of this tag as CN attribute of RND
				return x509e_invalidPktTag;
		}

		//
		//	Creation of the attribute.
		//
		sNameAttribute *t;

		t = sNameAttribute::create( sOID.B, sOID.L, sPrintableString.B, sPrintableString.L, sPrintableString.T );

		if ( t == NULL )
			return x509e_memoryAllocation;

		//
		//	Linking.
		//
		t->next = head;
		head = t;


		//
		//	Jump to next SET
		//

		pkt += sSET.TL;		// Adds total parsed length by the SET.
		pktSize -= sSET.TL;
	}

	return x509e_ok;
}


int x509_name::setList( sNameAttribute *nameAttributeList )
{
	if ( head )
		freeList();

	for ( sNameAttribute *at = nameAttributeList; at; at = at->next )
	{
		if ( AddCopy( at ) == NULL )
			return x509e_internalError;
	}

	return x509e_ok;
}

int x509_name::prepareForWritting()
{
//	head = nameAttributeList;

	int bodySize = 0;

	for ( sNameAttribute *at = head; at; at = at->next )
	{
		x509_struct set;
		x509_struct  seq;
		x509_struct	  oid;
		x509_struct   str;

		int byteLen = strlen( at->szValue );

		set.prepareForWritting( asn1tag_CONSTRUCTED_SET,
			seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				oid.prepareForWritting( asn1tag_OBJECT_IDENTIFIER, string2oid( NULL, 0, at->szOID ) )
				+
				str.prepareForWritting( at->tag, byteLen )
			)
		);

		bodySize += set.TL;
	}

	return x509_struct::prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE, bodySize );
}

int x509_name::writeBody()
{
	int pos = 0;

	for ( sNameAttribute *at = head; at; at = at->next )
	{
		x509_struct set;
		x509_struct  seq;
		x509_struct	  oid;
		x509_struct   str;

		int byteLen = strlen( at->szValue );

		set.prepareForWritting( asn1tag_CONSTRUCTED_SET,
			seq.prepareForWritting( asn1tag_CONSTRUCTED_SEQUENCE,
				oid.prepareForWritting( asn1tag_OBJECT_IDENTIFIER, string2oid( NULL, 0, at->szOID ) )
				+
				str.prepareForWritting( at->tag, byteLen )
			)
		);

		int lpos = 0;
		if ( set.writeHeader( B, L, &pos ) == 0 ) return x509e_internalError;
			if ( seq.writeHeader( set.B, set.L ) == 0 ) return x509e_internalError;
				if ( oid.writeHeader( seq.B, seq.L, &lpos ) == 0 ) return x509e_internalError;
				if ( str.writeHeader( seq.B, seq.L, &lpos ) == 0 ) return x509e_internalError;

				if ( lpos != (int)seq.L ) return x509e_internalError;

				if ( string2oid( oid.B, 0, at->szOID ) != (int)oid.L ) return x509e_internalError;
				memcpy( str.B, at->szValue, str.L );

//		pos += set.TL;
	}

	if ( pos != (int)L ) return x509e_internalError;

	return x509e_ok;
}

//
//	end of struct 
//		x509_name
//
///////////////////////////////////////////////////////////



///////////////////////////
//
//	struct x509_OID
//

int x509_OID::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );

	if ( res != x509e_ok )
		return res;

	if ( T != asn1tag_OBJECT_IDENTIFIER )
		return x509e_invalidPkt;

	char buf[ maxOIDStrSize ];
	if ( oid2string( buf, B, L ) == 0 )
		return x509e_internalError;

	int len = strlen( buf );

	if ( oidStr )
		delete oidStr;

	oidStr = new char[ len + 1 ];

	strcpy( oidStr, buf );

	return x509e_ok;
}

bool x509_OID::checkN( const char **nameT, int count )
{
	return checkOIDName( oidStr, nameT, count ) == x509e_ok;
}

bool x509_OID::check( const char **nameT )
{
	return checkOIDName( oidStr, nameT ) == x509e_ok;
}

x509_OID::x509_OID() : x509_struct()
{
	oidStr = NULL;
}

x509_OID::~x509_OID()
{
	if ( oidStr )
	{
		delete oidStr;
		oidStr = NULL;
	}
}

int x509_OID::prepareForWritting( const char *szOid )
{
	BYTE tag = asn1tag_OBJECT_IDENTIFIER;
	constOID = szOid;

	unsigned int bodySize = string2oid( NULL, NULL, constOID );

	return x509_struct::prepareForWritting( tag, bodySize );
}

int x509_OID::writeBody()
{
	unsigned int bodySize = string2oid( B, NULL, constOID );

	if ( bodySize != L ) return x509e_internalError;

	return ppe_ok;
}
//
//	end of struct 
//		x509_OID
//
///////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////
//
//	sNameAttributeList
//

bool sNameAttributeList::isEmpty()
{
	return head == NULL;
}

int sNameAttributeList::GenerateName( char *buff, int *pBuffSize )
{
	return getAttributesString( buff, pBuffSize );
/*
	sNameAttribute *tmp = head;
	if ( *pBuffSize > 0 )
		buff[0] = '\0';

	int pos = 0;

	for(; tmp != NULL; tmp = tmp->next )
	{
		pos += strlen( tmp->szValue ) + 1; // ","
	}

	if ( pos > *pBuffSize )
	{
		*pBuffSize = pos;
		return x509e_bufferTooSmall;
	}

	pos = 0;

	for(; tmp != NULL; )
	{
		strcpy( buff + pos, tmp->szValue );
		pos += strlen( tmp->szValue ); 
		tmp = tmp->next;
		if ( tmp != NULL )
			strcpy( buff + pos, "," );
		pos++;
	}

	*pBuffSize = pos;
	return x509e_ok;
*/
}

sNameAttribute *sNameAttributeList::Add( sNameAttribute::at_type idNum, const char *value, BYTE valueTag )
{
	if ( idNum == sNameAttribute::pkcs9_emailAddress )
		valueTag = asn1tag_IA5String;

	const char *oidStr = sNameAttribute::getOid( idNum );
	sNameAttribute *tmp = sNameAttribute::create( oidStr, value, valueTag );

	if ( tmp == NULL )
		return NULL;

	return Add( tmp );
}


sNameAttributeList::~sNameAttributeList()
{
	freeList();
}

void sNameAttributeList::freeList()
{
	sNameAttribute *tmp;
	while ( head != NULL )
	{
		tmp = head;
		head = head->next;

		delete tmp;
	}

	last = NULL;
}

sNameAttributeList::sNameAttributeList()
{
	head = NULL;
	last = NULL;
}

sNameAttribute *sNameAttributeList::Add( sNameAttribute *att )
{
	if ( head == NULL )
	{
		head = last = att;
	}
	else
	{
		last->next = att;
		last = att;
	}

	return att;
}

sNameAttribute *sNameAttributeList::AddCopy( sNameAttribute *att )
{
	sNameAttribute *tmp = sNameAttribute::create( att->szOID, att->szValue, att->tag );

	if ( tmp == NULL )
		return NULL;

	return Add( tmp );
}

int sNameAttributeList::AddCopyList( sNameAttribute *attList )
{
	int q;

	for( q = 0; attList; q++, attList = attList->next )
	{
		if ( !AddCopy( attList ) )
			return -1;
	}

	return q;
}

int sNameAttributeList::ParseString( const char *str, char separator )
{
	// "E = crypt@crypt.nsk.ru;CN = crypt.nsk.ru;OU = crypt.nsk.ru"
	int len = strlen( str );
	char *bf = new char[ len + 1 ];
	strcpy( bf, str );

	int pos = len - 1;

	while ( pos != 0 )
	{
		char *value;
		char *name;

		// Skips all probels
		while( bf[pos] == ' ' )
			if ( --pos < 0 )
				return -1;	// Empty string detected

		bf[pos+1] = '\0'; // Separates the probels from value

		if ( bf[pos] == '=' )
			return -1; // Value is empty

		// Gets value

		while( bf[pos] != '=' )
			if ( --pos < 0 )
				return -1;	// There is no '=' symbol

		// here bf[pos] == '='

		value = bf + pos + 1; // next char after '='
		while( *value == ' ' ) // Removes initial probels
			value++;

		// Gets name

		pos--;

		// Skips all probels
		while( bf[pos] == ' ' )
			if ( --pos < 0 )
				return -1;	// Empty string detected
		bf[pos+1] = '\0'; // Separates the probels from name

		for( ;; pos-- )
		{
			if ( bf[pos] == separator )
			{
				name = bf + pos + 1;
				pos--;
				break;
			}

			if ( pos == 0 )
			{
				name = bf;
				break;
			}
		}

		while( *name == ' ' ) // Removes initial probels
			name++;

		int id = sNameAttribute::getIdByName( name );
		if ( id == -1 )
			return -1;

		Add( sNameAttribute::at_type( id ), value );
	}

	return 0;
}

static int x509_name_getRequiredSize( sNameAttribute *t, const char *separator = ", " )
{
	int valueLen;

	valueLen = strlen( t->szValue );

	int size = strlen( t->getName() ) + strlen( "=" ) + valueLen + strlen( separator );

	return size;
}

static int x509_name_fillBuffer( char *buff, int size, sNameAttribute *t, const char *separator = ", " )
{
	if ( t == NULL )
	{
		buff[0] = 0;
		return x509e_ok;
	}
	else
	{
		int res = x509_name_fillBuffer( buff, size, t->next, separator );

		if ( res != x509e_ok )
			return res;
	}

	int len = strlen( buff );

	if ( len + x509_name_getRequiredSize( t, separator ) >= size )
		return x509e_internalError;

	strcat( strcat( strcat( strcat( buff, t->getName() ), "=" ), t->szValue ), separator );

	return x509e_ok;
}

char *sNameAttributeList::getAttributeValue( const char *szOID )
{
	for( sNameAttribute *t = head; t; t = t->next )
	{
		if ( strcmp( szOID, t->szOID ) == 0 )
			return t->szValue;
	}

	return NULL;
}

int sNameAttributeList::getAttributesString( char *szBuff, int *buffSize, const char *separator )
{
	int rSize = 1;

	for( sNameAttribute *t = head; t; t = t->next )
	{
		rSize += x509_name_getRequiredSize( t, separator );
	}

	if ( rSize > *buffSize )
	{
		*buffSize = rSize;
		return x509e_bufferTooSmall;
	}

	*buffSize = rSize;

	int res = x509_name_fillBuffer( szBuff, rSize, head, separator );

	int len = strlen( szBuff );

	int sepLen = strlen( separator );

	if ( len > sepLen )
		szBuff[ len - sepLen ] = 0; // Clear last separator in name path.

	return res;
}
