#include "x509_structs.h"
#include "../PGPPkts/pktdefs.h"

//
//	PKCS #8: Private-Key Information Syntax Standard
//
//	PrivateKeyInfo ::= SEQUENCE {
//	  version Version,
//
//	  privateKeyAlgorithm PrivateKeyAlgorithmIdentifier,
//	  privateKey PrivateKey,
//	  attributes [0] IMPLICIT Attributes OPTIONAL }
//
//	Version ::= INTEGER
//
//	PrivateKeyAlgorithmIdentifier ::= AlgorithmIdentifier
//
//	PrivateKey ::= OCTET STRING
//
//	Attributes ::= SET OF Attribute
//
// Sample of RSA private key:
//	SEQUENCE
//	. INTEGER 0 ( 0 )
//	. SEQUENCE
//	. . OID { pkcs-1-rsaEncryption }
//	. . NULL
//	. OCTET STRING
//	30 82 02 5C 02 01 00 02 81 81 00 B4 73 4B E2 A6
//
int x509_privateKeyInfo::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	BYTE *pos = B;
	int size = L;

	res = version.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = algOID.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = privateKey.parse( pos, size, &pos, &size, asn1tag_OCTET_STRING );
	if ( res != x509e_ok ) return res;

	if ( size )
	{
		res = attributes.parse( pos, size, &pos, &size );
		if ( res != x509e_ok ) return res;

		if ( size != 0 )
			return x509e_invalidPktSize;

		if ( !attributes.isOptional() || ( attributes.getOptionalNum() != 0 ) )
			return x509e_invalidPkt;
	}

	return x509e_ok;

//	return ska_createAlgorithmX509(	&ska, this );
}

//
//	Function: writeHeaders
//		writes prepared headers,
//			see x509_struct::prepareForWritting
//			see publicKeyAlgs -> writeKeyInfoASN1Buffer
//		returns tatal privateKeyInfo size.
//
int x509_privateKeyInfo::writeHeaders( BYTE *buff, int buffSize )
{
	if ( (int)TL > buffSize ) return 0;

	if ( writeHeader( buff, buffSize ) )
	{
		int pos = 0;

		if ( version.writeHeader( B , L, &pos ) && 
			 algOID.writeHeader( B, L, &pos  ) )
		{
			int pos = 0;

			if ( !algOID.oid.writeHeader( algOID.B, algOID.L, &pos ) )	return 0;
			if ( !algOID.params.writeHeader( algOID.B, algOID.L, &pos ) ) return 0;
		}
		else
			return 0;

		if ( !privateKey.writeHeader( B, L, &pos) ) return 0;

		if ( attributes.TL 
			&& !attributes.writeHeader( B, L, &pos ) ) return 0;

		if ( pos != (int)L )
			return 0;
	}
	else
		return 0;

	return TL;
}


//
//	RFC 2459		PKIX
//
//	   PublicKeyInfo  ::=  SEQUENCE  {
//	        algorithm            AlgorithmIdentifier,
//	        subjectPublicKey     BIT STRING  }
//
int x509_publicKeyInfo::parse( BYTE *pkt, int pktSize, BYTE **pPkt, int *pPktSize )
{
	int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_SEQUENCE );
	if ( res != x509e_ok ) return res;

	BYTE *pos = B;
	int size = L;

	res = algOID.parse( pos, size, &pos, &size );
	if ( res != x509e_ok ) return res;

	res = publicKey.parse( pos, size );
	if ( res != x509e_ok ) return res;

	return x509e_ok;

// 	return pka_createAlgorithmX509( pka, this );
}

int x509_publicKeyInfo::writeHeaders( BYTE *buff, int buffSize )
{
	if ( (int)TL > buffSize ) return 0;

	if ( writeHeader( buff, buffSize ) )
	{
		int pos = 0;

		if ( algOID.writeHeader( B, L, &pos ) )
		{
			int pos = 0;

			if ( !algOID.oid.writeHeader( algOID.B, algOID.L, &pos ) ) return 0;
			if ( !algOID.params.writeHeader( algOID.B, algOID.L, &pos ) ) return 0;
		}

		if ( !publicKey.writeHeader( B, L, &pos  ) ) return 0;

		if ( pos != (int)L )
			return 0;
	}

	return TL;
}
