/////////////////////////////////////////////////////////////////
//
//	Copyright (c) 1994-2008 Jetico Inc., Finland
//	All rights reserved.
//
//	File: x509_structs.h
//	Revision: 
//	Created:
//	Description:
//
/////////////////////////////////////////////////////////////////

#ifndef __x509structs__
#define __x509structs__

//#include <windows.h>
#include "../publicKeyAlgs/Power/Windefs.h"
#include <assert.h>
#include "x509_tags.h"
#include "../crtLibErrs.h"
#include "oidStrings.h"
#include <memory.h>
#include <string.h>

typedef class cPKA * HPKA; // Handle to public/private key algorithm

struct x509_struct
{
	BYTE	T;	// tag
	UINT	L;	// length
	BYTE   *B;	// pointer to body
	BYTE   *H;	// pointer to header
	UINT	TL;	// total length( size busied by the structure including tag, length, body )

	bool uL; // indefiniteLenght - L = 0x80, we have to find end of contents marker to stop searching.

	//
	//	Checks if the struct tag is asn1tag_OPTIONAL
	//
	bool isOptional();

	//
	//	Returns number of the optional field.
	//
	int getOptionalNum();

	//
	//	Checks packet size, parses tag and length, fills pointer to body and total length values.
	//
	int parse(
		const BYTE *pkt,	// pointer to packet begining
		int pktSize,// total size of parsed packet
		BYTE **pPkt = NULL,		// next position value for updating
		int *pPktSize = NULL,	// size left value for updating
		int checkTag = -1		// Tag value to be checked
		);

	enum {
		maxHdrSize = 
				1 + // Tag
				1 + // Length tag or one byte length of lengt
				4	// max available length
	};

	//
	//	
	//	Return required size of header for new ANS1 unit to pack the body.
	//
	static int makeHeader(
		unsigned int bodySize,
		BYTE tag = 0,
		BYTE *hdr = NULL // If not NULL is poiter to byte buffer maxHdrSize
		);

	//
	//	Fills the structure tag and length
	//	If successful returns total length busied by the structure, otherwise returns 0.
	//
	int prepareForWritting( BYTE tag, unsigned int bodySize );

	//
	//	Writes the previously prepared header ( tag and length ).
	//	Checks if size of allocated buffer is large enough for the unit.
	//	If successful returns header size, otherwise returns 0.
	//	Note, if pPos is NULL, 
	//		==> so we are not going to write other structures to the buffer,
	//		==> all buffer has to filled by the structure only,
	//		so if TL != bufferSize returns 0
	//
	int writeHeader(
		BYTE *buff,		// Buffer for header writting
		int buffSize,	// Size of allocated buffer
		int *pPos = NULL// Offset position for next unit
		);

	//
	// Write the body to prepared by writeHeader buffer
	//	return ppe_ok if written;
	//
	int writeBody( BYTE *body, unsigned int bodySize );

	x509_struct();

	void clear() { memset( this, 0, sizeof( *this ) ); }
};


struct x509_OID : public x509_struct
{
	char *oidStr;
	const char *constOID; // used by prepareForWritting and writeBody
	
	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );

	bool checkN( const char **nameT, int count );
	bool check( const char **nameT );
	
	x509_OID();
	~x509_OID();

	int prepareForWritting( const char *szOid );

	int writeBody();
};

//
//	AlgorithmIdentifier ::= SEQUENCE{
//	        algorithm       OBJECT IDENTIFIER,
//	        parameters      ANY DEFINED BY algorithm OPTIONAL ( or NULL )
//	}
//
struct x509_algorithmID : public x509_struct
{
	x509_OID	oid;
	x509_struct params;

	tUnit		*tu;	// The ultimate unit of object identifier tables. 
						// It has to be filled during parsing

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );
};

struct x509_bitString : public x509_struct
{
	BYTE unusedBits;
	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );
	int prepareForWritting( unsigned int bodySize );
	int writeBody( BYTE *body, unsigned int bodySize );
	int writeHeader(
		BYTE *buff,		// Buffer for header writting
		int buffSize,	// Size of allocated buffer
		int *pPos = NULL// Offset position for next unit
		);
};

struct x509_time : public x509_struct
{
	int year;
	int month;
	int day;
	int hour;
	int minute;
	int second;

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );
	int compare( int year, int month, int day, int hour, int minute, int second );

	int prepareForWritting( int ye, int mo, int da, int ho = 0, int mi = 0, int se = 0 );
	int writeBody();

	SYSTEMTIME ConvertToSystemTime()
	{
		SYSTEMTIME st;
		st.wYear		= (WORD)year;
		st.wMonth		= (WORD)month;
		st.wDayOfWeek	= (WORD)0;
		st.wDay			= (WORD)day;
		st.wHour		= (WORD)hour;
		st.wMinute		= (WORD)minute;
		st.wSecond		= (WORD)second;
		st.wMilliseconds= (WORD)0;

		return st;
	}
};

struct x509_validity : public x509_struct
{
	x509_time notBefore;
	x509_time notAfter;
	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );
	bool isValid( time_t validityTime = 0 );
};

struct sNameAttribute
{
	sNameAttribute *next;
	char *szOID;
	char *szValue;
	BYTE tag;

	sNameAttribute()
	{
		szOID = NULL;
		szValue = NULL;
		next = NULL;
		tag = 0; // asn1tag_PrintableString;
	}

	~sNameAttribute();

	static sNameAttribute *create( BYTE *OID, int OIDSize, BYTE *value, int valueSize, BYTE valueTag = asn1tag_PrintableString);
	static sNameAttribute *create( const char *oidStr, const char *value, BYTE valueTag );

	enum at_type {
		id_first,
		id_at_commonName = id_first,
		id_at_surname,
		id_at_serialNumber,
		id_at_countryName,
		id_at_localityName,
		id_at_stateOrProvinceName,
		id_at_streetAddress,
		id_at_organizationName,
		id_at_organizationalUnitName,
		id_at_title,
		id_at_description,
		id_at_postalAddress,
		id_at_postOfficeBox,
		id_at_telephoneNumber,
		id_at_facsimileTelephoneNumber,
		id_at_name,
		id_at_givenName,
		id_at_initials,
		id_at_generationQualifier,
		id_at_uniqueIdentifier,
		id_at_dnQualifier,
		pkcs9_emailAddress,
		id_last = pkcs9_emailAddress
	};

	//
	//	Function: getOid
	//	Gets oid string for a name attribute from from enum list
	//
	static const char *getOid( at_type idNum );

	//
	//	If comptible ID is found so return the at_type
	//	if error return -1
	//
	static int getIdByName( const char *name );
	static int getIdByOID( const char *szOID );

	const char *getName();

//	static void freeList( sNameAttribute *head ) { if ( head ) delete head; }
};

struct sNameAttributeList
{
	sNameAttribute *head, *last;

	sNameAttributeList();
	sNameAttribute *Add( sNameAttribute *att );
	sNameAttribute *AddCopy( sNameAttribute *att );
	int AddCopyList( sNameAttribute *attList );
	sNameAttribute *Add( sNameAttribute::at_type idNum, const char *value, BYTE valueTag = asn1tag_PrintableString );
	int ParseString( const char *str, char separator = ';' );
	~sNameAttributeList();
	void freeList();
	bool isEmpty();

	//
	//	getAttributeValue
	//		returns values for name attribute with specified OID
	//		see the sNameAttribute::getOid( at_type idNum ) to get OID for required name value ;
	//
	char *getAttributeValue( const char *szOID );
	char *getCN()    { return getAttributeValue( sNameAttribute::getOid( sNameAttribute::id_at_commonName ) ); }
	char *getEMail() { return getAttributeValue( sNameAttribute::getOid( sNameAttribute::pkcs9_emailAddress ) ); }

	int getAttributesString( char *szBuff, int *buffSize, const char *separator = ", " );
	int GenerateName( char *buff, int *pBuffSize );
};

struct x509_name : public x509_struct, public sNameAttributeList
{
	x509_name();
	~x509_name();

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );
	int readAttributes( BYTE *pkt, int pktSize );

	int setList( sNameAttribute *nameAttributeList );
	int prepareForWritting();
	int writeBody();
};

struct x509_integer : public x509_struct
{
	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );
	int compare( unsigned int x );
	int compare( x509_integer *x );
	int getUIValue( unsigned int *pVal );

	BYTE body[ sizeof( unsigned int ) ];
	int lastNonZero;

	int prepareForWritting( unsigned int val );
	int writeBody();
};

//
// X509 Extensions...
//
// RFC 2459  [Page 24]
//	At a minimum, applications conforming to this profile MUST recognize
//	the extensions which must or may be critical in this specification.
//	These extensions are:  key usage (see sec. 4.2.1.3), certificate
//	policies (see sec. 4.2.1.5), the subject alternative name (see sec.
//	4.2.1.7), basic constraints (see sec. 4.2.1.10), name constraints
//	(see sec. 4.2.1.11), policy constraints (see sec. 4.2.1.12), and
//	extended key usage (see sec. 4.2.1.13).
//
//
//	Extensions  ::=  SEQUENCE SIZE (1..MAX) OF Extension
//	
//	Extension  ::=  SEQUENCE  {
//		extnID      OBJECT IDENTIFIER,
//		critical    BOOLEAN DEFAULT FALSE,
//		extnValue   OCTET STRING  }
//

//
//	x509_extension
//	package for extension.
//
struct x509_extension : public x509_struct
{
	x509_OID oid;
	x509_struct bCritical;
	x509_struct octet;

	int prepareForWritting( const char *oidStr, int bodySize, bool critical = false );
	int writeHeader( BYTE *buff, int buffSize, int *pPos );
};

struct x509_ce_subjectKeyIdentifier: public x509_struct
{
	const char *getOid() { return "2.5.29.14"; }
	int prepareForWritting( int bodySize ) 
	{ return x509_struct::prepareForWritting( asn1tag_OCTET_STRING, bodySize ); }
};

//
//	RFC 2459        Internet X.509 Public Key Infrastructure    January 1999
//
//	GeneralName ::= CHOICE {
//		otherName                       [0]     OtherName,
//		rfc822Name                      [1]     IA5String,
//		dNSName                         [2]     IA5String,
//		x400Address                     [3]     ORAddress,
//		directoryName                   [4]     Name,
//		ediPartyName                    [5]     EDIPartyName,
//		uniformResourceIdentifier       [6]     IA5String,
//		iPAddress                       [7]     OCTET STRING,
//		registeredID                    [8]     OBJECT IDENTIFIER}
//
//	OtherName ::= SEQUENCE {
//		type-id    OBJECT IDENTIFIER,
//		value      [0] EXPLICIT ANY DEFINED BY type-id }
//
//	Housley, et. al.            Standards Track                    [Page 34]
//
struct x509_generalName: public x509_struct
{
	BYTE choiceNum;

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL )
	{
		int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize );
		if ( res != x509e_ok )
			return res;

		if ( ( T & 0xF0 ) != asn1tag_CONSTRUCTED_OPTIONAL )
			return x509e_invalidPktTag;
		
		choiceNum = T & 0x0F;

		return x509e_ok;
	}

	static const char *getDescription( int choiceNum )
	{
		switch ( choiceNum )
		{
			case 0: return "otherName";
			case 1: return "rfc822Name";
			case 2: return "dNSName";
			case 3: return "x400Address";
			case 4: return "directoryName";
			case 5: return "ediPartyName";
			case 6: return "uniformResourceIdentifier";
			case 7: return "iPAddress";
			case 8: return "registeredID";
			default: return "Undefined";
		}
	}

	//
	//	readDirectoryName
	//		reads directory name as RDN
	//
	int readDirectoryName( x509_name *sName ) { return readDirectoryName( sName, B, L ); }

/*

Sample :: GeneralName->directoryName ( part of 'Sample :: x509_ce_authorityKeyIdentifier' )

001C  A4 81 99      153: . . [4]; 					GeneralName :== CHOICE [4] ->directoryName
001F  30 81 96      150: . . . SEQUENCE
0022  31 0B          11: . . . . SET
0024  30 09           9: . . . . . SEQUENCE
0026  06 03           3: . . . . . . OID { id-at-C }
002B  13 02           2: . . . . . . Printable String->RU
002F  31 10          16: . . . . SET
0031  30 0E          14: . . . . . SEQUENCE
....

*/
	static int readDirectoryName(
			x509_name *sName,		// Result RDN
			BYTE *pkt, int pktSize	// Choice body
		)
	{
		x509_struct sSequence;

		int res = sSequence.parse( pkt, pktSize );
		if ( res != x509e_ok )
			return res;

		if ( ( sSequence.T != asn1tag_SEQUENCE )
			&& ( sSequence.T != asn1tag_CONSTRUCTED_SEQUENCE ) )
				return x509e_invalidPktTag;

		return sName->readAttributes( sSequence.B, sSequence.L );
	}
};
/*

Sample :: x509_ce_authorityKeyIdentifier

0000  30 81 B8      184: SEQUENCE
0003  80 14          20: . [0]; 					Key identifier = SHA-1 hash of public key
			42 ED 1D 68 82 8B 8D 30 48 4E 62 98 E4 23 E8 B1 0A C0 5D 85
0019  A1 81 9C      156: . [1]; 					General Names
001C  A4 81 99      153: . . [4]; 					GeneralName :== CHOICE [4] ->directoryName
001F  30 81 96      150: . . . SEQUENCE
0022  31 0B          11: . . . . SET
0024  30 09           9: . . . . . SEQUENCE
0026  06 03           3: . . . . . . OID { id-at-C }
002B  13 02           2: . . . . . . Printable String->RU
002F  31 10          16: . . . . SET
0031  30 0E          14: . . . . . SEQUENCE
0033  06 03           3: . . . . . . OID { id-at-ST }
0038  13 07           7: . . . . . . Printable String->Siberia
0041  31 14          20: . . . . SET
0043  30 12          18: . . . . . SEQUENCE
0045  06 03           3: . . . . . . OID { id-at-L }
004A  13 0B          11: . . . . . . Printable String->Novosibirsk
0057  31 0E          14: . . . . SET
0059  30 0C          12: . . . . . SEQUENCE
005B  06 03           3: . . . . . . OID { id-at-O }
0060  13 05           5: . . . . . . Printable String->Crypt
0067  31 15          21: . . . . SET
0069  30 13          19: . . . . . SEQUENCE
006B  06 03           3: . . . . . . OID { id-at-OU }
0070  13 0C          12: . . . . . . Printable String->crypt.nsk.ru
007E  31 15          21: . . . . SET
0080  30 13          19: . . . . . SEQUENCE
0082  06 03           3: . . . . . . OID { id-at-CN }
0087  13 0C          12: . . . . . . Printable String->crypt.nsk.ru
0095  31 21          33: . . . . SET
0097  30 1F          31: . . . . . SEQUENCE
0099  06 09           9: . . . . . . OID { pkcs-9-E-mail 			}
00A4  16 12          18: . . . . . . IA5 String->crypt@crypt.nsk.ru
00B8  82 01           1: . [2] 0 ( 0 ) ; 				Authority Certificate Serial number
*/
struct x509_ce_authorityKeyIdentifier: public x509_struct
{
	x509_struct opt[2];
	x509_integer opt2;
	x509_struct choice;
	x509_name	name;

	const char *getOid() { return "2.5.29.35"; }
	int prepareForWritting( int keyIDSize, sNameAttribute *rdn, int serNum );
	int writeHeader(
		BYTE *buff,		// Buffer for header writting
		int buffSize,	// Size of allocated buffer
		int *pPos = NULL// Offset position for next unit
	);

	int writeBody( BYTE *keyID, int keyIDSize );

};

struct x509_ce_basicConstraints: public x509_struct
{
	bool bCA;
	x509_struct sCA;
	x509_struct opt;
	x509_integer pathLength;

	const char *getOid() { return "2.5.29.19"; }
	int prepareForWritting( bool CA = false, int pathLengthConstraint = -1 );
	int writeBody();
};

struct x509_extensions : public x509_struct
{
	x509_ce_subjectKeyIdentifier	subjectKeyIdentifier;
	x509_ce_authorityKeyIdentifier	authorityKeyIdentifier;
	x509_ce_basicConstraints		basicConstraints;

	//
	//	Parses all extensions and searchs specified extension.
	//	returns x509e_ok if the extension found.
	//	Returns x509e_extensionIsNotPresent if the extesions was not found.
	//
	int getExtensionValue(
		const char *oid,
		BOOL *pCritical,
		BYTE **pValue,		// pointer to value body ( body of OCTET STRING ).
		int *pValueSize		// pointer to size
	);

	//
	//	There are several certificate extensions, 
	//	see RFC 2459 p.4.2 Standard Certificate Extensions.
	//	The checkBasicConstraint functions is a sample of processing the 
	//	id-ce-basicConstraint certificate extension
	//
	int checkBasicConstraint( BYTE *pBoolCA = NULL, unsigned int *pPathLength = NULL );

	//
	//	Function getSubjectKeyID
	//	Processes id-ce-subjectKeyIdentifier
	//	Returns x509e_ok if the extension found and assigns pointers to key ID Ptr and it size
	//	Note, it does not copy the key id.
	//
	int getSubjectKeyID( BYTE **pKeyIDPtr = NULL, int *pKeyIDSize = NULL );

	//
	//	Function getAuthorityKeyID
	//	Processes id-ce-authorityKeyIdentifier
	//	see RFC 2459, p.4.2.1.1
	//	It fills pointers to key Id buffer, to general name, and gets serial number.
	//	Note all the fields are optional, so the pointers may be assigned to NULL if an 
	//	optional filed does not exist 
	//
	int getAuthorityKeyID(
			BYTE **pKeyIDPtr = NULL, int *pKeyIDSize = NULL,
			int *pGeneralNameChoise = NULL,
			BYTE **pGeneralName = NULL, int *pGeneralNameSize = NULL,
			int *pSerialNumber = NULL );

	//
	//	Extension mas two variants of key ID, see RFC 2459, p.4.2.1.2
	//	The function check both variants and return valid key ID
	//
	static int getKeyID( BYTE *ptr, int size, BYTE **pKeyIDPtr, int *pKeyIDSize );
};

//
//	RFC 2459		PKIX
//	
//	   SubjectPublicKeyInfo  ::=  SEQUENCE  {
//	        algorithm            AlgorithmIdentifier,
//	        subjectPublicKey     BIT STRING  }
//
struct x509_publicKeyInfo : public x509_struct
{
	x509_algorithmID	algOID;
	x509_bitString		publicKey;

	//
	//	Function: parse
	//		fills internal structures
	//		checks validity of publicKeyInfo SECUENSE
	//	If the function succeful, the x509_publicKeyInfo can be used by
	//	pka_createAlgorithm_X509 to create corresponded public key algorithm
	//
	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );

	//
	//	Function: writeHeaders
	//		writes prepared headers,
	//			see x509_struct::prepareForWritting
	//			see publicKeyAlgs -> writeKeyInfoASN1Buffer
	//		returns total publicKeyInfo size.
	//
	int writeHeaders( BYTE *buff, int buffSize );
};

struct x509_tbsCertificate : public x509_struct
{
	BYTE				version; // OPTIONAL [0]
	x509_integer		serialNumber;
	x509_algorithmID	signatureAlg;
	x509_name			issuerName; // non-empty distinguished name (DN)
	x509_validity		validity;
	x509_name			subjectName; // non-empty distinguished name (DN)
	x509_publicKeyInfo	subjectPKInfo;
	x509_bitString		issuerUniqueID; //	OPTIONAL [1]
	x509_bitString		subjectUniqueID;//	OPTIONAL [2]
	x509_extensions		extensions;		//	OPTIONAL [3]

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );

	//
	//	getSubjectKeyID
	//		Checks extension only and fills pointers to parsed keyID.
	//
	int getSubjectKeyID( BYTE **pKeyIDPtr = NULL, int *pKeyIDSize = NULL );

	//
	//	generateSubjectKeyID
	//		If correponded extension is not present, the subject local key ID may 
	//		generated from subjectPKInfo.
	//
	int generateSubjectKeyID( BYTE *pKeyIDPtr, int *pKeyIDSize );

	int getIssuerKeyID( BYTE **pKeyIDPtr = NULL, int *pKeyIDSize = NULL );

	int createPacket(
		BYTE *buff, int *pBuffSize,
		HPKA pkaSubject, HPKA skaIssuer,
		sNameAttribute *rdnSubject, sNameAttribute *rdnIssuer, 
		int nbYe, int nbMo, int nbDa, // Not before Year, Month, Day
		int naYe, int naMo, int naDa, // Not after Year, Month, Day
		const char *signatureAlgOID, int iSerialNumber );
};

struct x509_certificate : public x509_struct
{
	x509_tbsCertificate tbsCertificate;
	x509_algorithmID	signatureAlg; //    This field MUST contain the same algorithm identifier as the
										//   signature field in the sequence tbsCertificate (see sec. 4.1.2.3, RFC 2459 ).
	x509_bitString		signatureValue;

	//
	//	sDigestPackage PKCS1 v.2, p 9.2.1 EMSA-PKCS1-v1_5
	//	DigestInfo ::= SEQUENCE {
	//		digestAlgorithm AlgorithmIdentifier,
	//		digest OCTET STRING }
	//
	x509_struct sDigestPackage; // SEQUENCE
	 x509_algorithmID sAlgID;
	 x509_struct	sValue; // OCTET_STRING

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );

	//
	//	Certificate check functions...
	//
	int checkValidityPeriod();

	//
	//	Checks intended purposes..
	//
	//	RFC 2459 - 4.2.1.13  Extended key usage field
	//
	//	Key purpose enum
	enum {
		kp_serverAuth = 1,	// if the purpose presents, it is required to check 
							// server name for compatibility with CN
		kp_clientAuth = 2,
		kp_codeSigning = 3,
		kp_emailProtection = 4,
		kp_timeStamping = 8
	};
	int checkPurpose( int purpose, BOOL *pCritical = NULL );

	//
	//	Verifies certificate entire signature value,
	//	 from header signatureValue.H upto total length signatureValue.TL ).
	//
	int verify( BYTE haID, BYTE pkID, HPKA pka );
	int verify( x509_publicKeyInfo *issuerPKI );

	int createPacket(
		BYTE *buff, int *pBuffSize, 
		HPKA pkaSubject, HPKA skaIssuer,
		sNameAttribute *rdnSubject, sNameAttribute *rdnIssuer, 
		int nbYe, int nbMo, int nbDa, // Not before Year, Month, Day
		int naYe, int naMo, int naDa, // Not after Year, Month, Day
		const char *signatureAlgOID, int iSerialNumber );
};


//
//	PKCS #8: Private-Key Information Syntax Standard
//
//	PrivateKeyInfo ::= SEQUENCE {
//	  version Version,
//	
//	  privateKeyAlgorithm PrivateKeyAlgorithmIdentifier,
//	  privateKey PrivateKey,
//	  attributes [0] IMPLICIT Attributes OPTIONAL
//	}
//	
//	Version ::= INTEGER
//	
//	PrivateKeyAlgorithmIdentifier ::= AlgorithmIdentifier
//	
//	PrivateKey ::= OCTET STRING
//	
//	Attributes ::= SET OF Attribute
//
struct x509_privateKeyInfo : public x509_struct
{
//	x509_struct			version;		// INTEGER
	x509_integer		version;		// INTEGER
	x509_algorithmID	algOID;			// PrivateKeyAlgorithmIdentifier
	x509_struct			privateKey;		// OCTET STRING
	x509_struct			attributes;		// IMPLICIT Attributes OPTIONAL

	//
	//	Function: parse
	//		fills internal structures
	//		checks validity of privateKeyInfo SECUENSE
	//	If the function successful, the x509_privateKeyInfo can be used by
	//	ska_createAlgorithm_X509 to create corresponded private key algorithm
	//
	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL );

	//
	//	Function: writeHeaders
	//		writes prepared headers,
	//			see x509_struct::prepareForWritting
	//			see publicKeyAlgs -> writeKeyInfoASN1Buffer
	//		returns tatal privateKeyInfo size.
	//
	int writeHeaders( BYTE *buff, int buffSize );
};

struct x509_digestInfo : public x509_struct
{
	//
	//	sDigestPackage PKCS1 v.2, p 9.2.1 EMSA-PKCS1-v1_5
	//	DigestInfo ::= SEQUENCE {
	//		digestAlgorithm AlgorithmIdentifier,
	//		digest OCTET STRING }
	//
	x509_algorithmID sAlgID;
	x509_struct	sValue; // OCTET_STRING

	int parse( BYTE *pkt, int pktSize, BYTE **pPkt = NULL, int *pPktSize = NULL )
	{
		int res = x509_struct::parse( pkt, pktSize, pPkt, pPktSize, asn1tag_SEQUENCE );
		if ( res != x509e_ok )	return res;

		res = sAlgID.parse( B, L );
		if ( res != x509e_ok )	return res;

		return sValue.parse( B + sAlgID.TL, L - sAlgID.TL, 0, 0, asn1tag_OCTET_STRING );
	}
};

int x509_test( BYTE *ptr, int size, bool skipOctetParsing );


#endif // __x509structs__
