/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 */

#ifndef BC_CIPHER_H
#define BC_CIPHER_H

#include <cstdlib>
#include <exception>
#include <assert.h>

#include "kernel/BCKernel.h"


/**
 * Exceptions thrown when kernel cipher implementation misbehaves.
 */
class BCCipherException : public std::exception
{
public:
	
	BCCipherException(bc_kern_result rc) throw() : m_rc(rc) {};
	
	virtual const char* what() const throw() {
		return BCKernelGetErrorString(m_rc);
	}

	virtual ~BCCipherException() throw() {
	}
	
private:
	
	bc_kern_result m_rc;
};


/**
 * Wrapper over kernel crypto service.
 * Combines mode and algorithm and does some error checking before entering kernel.
 * It is safe to copy this object anyway you want.
 * Throws exceptions when communicating with the kernel service fails. If you prefer not to deal with exceptions, use raw c interface declared in BCKernel.h
 */
class BCCipher
{
public:
	
	BCCipher();
		
	/**
	 * Initialize cipher configuration.
	 * This is safe to call anytime to reset cipher configuration reusing the same object.
	 * @alg_id	Kernel crypto service algorithm id.
	 * @mode_id	Kernel crypto service mode id.
	 * @return	kBCCipher_UnsupportedMode if given cipher does not support given mode.
	 *		kBCCipher_UnknownAlg if algorithm is not registered in kernel service.
	 */
	int init(bc_alg_id alg_id, bc_mode_id mode_id);
	int init(BCAlgInfo* alg, bc_mode_id mode_id);
	
	/// Get cipher block size in bytes.
	size_t blockSize() const {
		return m_alg.block_size;
	}

	/// Get block algorithm key size in bits.
	size_t keyLength() const {
		return m_alg.key_length;
	}

	/// Get cipher key size in bytes which is combined key size of algorithm and mode for this algorithm. 
	size_t keySize() const {
		switch (m_mode) {
		case kBCMode_CBC: return m_alg.cbc_key_size;
		case kBCMode_LRW: return m_alg.lrw_key_size;
		case kBCMode_XTS: return m_alg.xts_key_size;
		default: assert(0); return 0;
		}
	}

	/// Get cipher algorithm id.
	bc_alg_id algId() const {
		return m_alg.id;
	}
	
	/// Get cipher mode id.
	bc_mode_id modeId() const {
		return m_mode;
	}

	/// Get cipher algorithm name.
	const char* algName() const {
		return m_alg.name;
	}
	
	/// Get cipher mode name.
	const char* modeName() const {
		switch (m_mode) {
		case kBCMode_CBC: return "CBC";
		case kBCMode_LRW: return "LRW";
		case kBCMode_XTS: return "XTS";
		default: assert(0); return NULL;
		}
	}
	
	
	/**
	 * Process data using key handle stored in kernel service memory.
	 * @iv		Initial vector.
	 * @type	Encrypt/decrypt.
	 * @key_handle	Handle to the encryption key in kernel memory.
	 * @pool	Random data pool to xor with data, leave NULL for zero-filled pool.
	 * @data	Data to process.
	 * @nbytes	Number of bytes of data. Should be a multiple of cipher block size.
	 * 
	 * @throws	BCCipherException on user/kernel communication problems and other things.
	 */
	void process(uint64_t iv, bc_proc_type type, bc_key_handle key_handle, const char* pool, void* data, size_t nbytes);

	
	/**
	 * Process data using explicit key data.
	 * Use this carefully when you need some data processed with a one-time key.
	 * @iv		Initial vector.
	 * @type	Encrypt/decrypt.
	 * @key		Key data. Will not be stored in kernel memory.
	 * @keysize	Number of bytes in key data.
	 * @pool	Random data pool to xor with data, leave NULL for zero-filled pool.
	 * @data	Data to process.
	 * @nbytes	Number of bytes of data. Should be a multiple of cipher block size.
	 * 
	 * @throws	BCCipherException on user/kernel communication problems and other things.
	 */
	void process(uint64_t iv, bc_proc_type type, const char* key, size_t keysize, const char* pool, void* data, size_t nbytes);

	
	/// Encrypt block of data using stored key handle.
	void encrypt(uint64_t iv, bc_key_handle key_handle, const char* pool, void* data, size_t nbytes) {
		this->process(iv, kBCProcess_Encrypt, key_handle, pool, data, nbytes);
	}
	
	/// Encrypt block of data using supplied key data.
	void encrypt(uint64_t iv, const char* key, size_t key_size, const char* pool, void* data, size_t nbytes) {
		this->process(iv, kBCProcess_Encrypt, key, key_size, pool, data, nbytes);
	}

	/// Decrypt block of data using stored key handle.
	void decrypt(uint64_t iv, bc_key_handle key_handle, const char* pool, void* data, size_t nbytes) {
		this->process(iv, kBCProcess_Decrypt, key_handle, pool, data, nbytes);
	}
	
	/// Decrypt block of data using supplied key data.
	void decrypt(uint64_t iv, const char* key, size_t key_size, const char* pool, void* data, size_t nbytes) {
		this->process(iv, kBCProcess_Decrypt, key, key_size, pool, data, nbytes);
	}
	

private:
	
    BCAlgInfo   m_alg;
    bc_mode_id  m_mode;
};


/**
 * Enumeration of all supported algorithms.
 */
struct BCEnumAlgorithms {
	BCAlgInfo* algs;
	size_t nalgs;
	
	BCEnumAlgorithms();
	~BCEnumAlgorithms();
	
	BCAlgInfo* get(unsigned idx) {
		if(idx >= nalgs) {
			return NULL;
		}
		
		return algs + idx;
	}
	
	unsigned total() const {
		return nalgs;
	}
};


/**
 * Enumeration of all supported modes.
 */
struct BCEnumModes {
	enum {
		kBCTotalModes = 3,
	};
	
	bc_mode_id modes[kBCTotalModes];
	
	BCEnumModes();
	
	bc_mode_id get(unsigned idx) {
		assert(idx < kBCTotalModes);
		return modes[idx];
	}
	
	unsigned total() const {
		return kBCTotalModes;
	}
};


#endif
