/*
 * Copyright 2012 Jetico, Inc. All rights reserved.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

#include "container/container.h"
#include "container/user.h"
#include "container/file.h"
#include "container/password.h"
#include "container/cipher.h"
#include "kernel/BCKernel.h"



//////////////////////////////////////////////////////////////////////////
//
// Known keygens 
//
//////////////////////////////////////////////////////////////////////////



extern struct BCKeygen kg_sha;
extern struct BCKeygen kg_ghost;

static const struct BCKeygen* gKeygens[] = {
    &kg_sha,
    &kg_ghost,
};

static const struct BCKeygen* gDefaultKeygen = &kg_ghost;

#define BC_TOTAL_KEYGENS (sizeof(gKeygens) / sizeof(*gKeygens))



//////////////////////////////////////////////////////////////////////////
//
// User sessions 
//
//////////////////////////////////////////////////////////////////////////



// session flags
enum {
    kBCUserSession_Default          = 0,        // Default password verification session.
    kBCUserSession_MainPartOnly     = 0x01,     // Additionally verify that password is for main part.
    kBCUserSession_HiddenPartOnly   = 0x02      // Additionally verify that password is for hidden part.
};


// User session context to verify and store container password
typedef struct {
    const char*     file;       // container file
    IBCHeader*      header;     // on success the keygen that verified the password
    int             flags;      // kBCUserSession flags
} VerifyPasswordCtx;

//
//  Verify existing passwords
//

static const char* verifyPasswordCallback(const char* password, unsigned /*pwdlen*/, void* ctx_)
{
    VerifyPasswordCtx* ctx = (VerifyPasswordCtx*) ctx_;
    IBCHeader *header = ctx->header;
    const BCKeygen *keygenStruct = ctx->header ? ctx->header->keygen : NULL;
    const char *result = NULL;

    // If no header is specified, try to parse the container.
    if (!header) {
        for (unsigned i = 0; i < BC_TOTAL_KEYGENS; ++i) {

            header = gKeygens[i]->kgload(ctx->file);
            if(!header) {
                continue;
            }

            keygenStruct = gKeygens[i];
            break;
        }
    }
    if (!header)
        return "Cannot find an acceptable key generator.";

    BCVolumeInfo info;
    if (kBCContainer_Success == header->verifyPassword(password, &info)) {

        if ((ctx->flags & kBCUserSession_MainPartOnly) && (info.hidden)) {
            result = "Main part password required.";
        }
        else if ((ctx->flags & kBCUserSession_HiddenPartOnly) && (!info.hidden)) {
            result = "Hidden part password required.";
        }
        else {
            ctx->header = header;
            ctx->header->keygen = keygenStruct;
            result = NULL;
        }
    }
    else {
        result = "Incorrect password.";
    }

    // End of the function.
    if (result && !ctx->header) {
        // Need to remove created header.
        delete header;
    }

    return result;
}


static IBCHeader* verifyPassword(const char* file, const char* query, int flags,
                                 BCMountParams *mountParams, BCPassword* out_pwd)
{
    VerifyPasswordCtx ctx = { file, NULL, flags };
    
    BCPassword tmppwd;
    if (!out_pwd) {
        out_pwd = &tmppwd;
    }

    if(kBCUserOk == BCUserVerifyPassword(query, verifyPasswordCallback, &ctx,
                                         mountParams, out_pwd)) {
        return ctx.header;
    } else {
        return NULL;
    }
}


static IBCHeader* verifyPassword(const char* file, const char* query, BCMountParams *mountParams = NULL, BCPassword* out_pwd = NULL)
{
    return verifyPassword(file, query, kBCUserSession_Default, mountParams, out_pwd);
}


static IBCHeader* verifyMainPartPassword(const char* file, const char* query, BCMountParams *mountParams = NULL, BCPassword* pwd = NULL)
{
    return verifyPassword(file, query, kBCUserSession_MainPartOnly, mountParams, pwd);
}

//
//  Verify new passwords
//

static const char* verifyNewPasswordCallback(const char* password, unsigned /*pwdlen*/, void* arg)
{
    VerifyPasswordCtx* ctx = (VerifyPasswordCtx*) arg;
    BCVolumeInfo info;
    if(kBCContainer_Success == ctx->header->verifyPassword(password, &info)) {
        return "Password is already in use";
    }

    return NULL;
}


static bool askNewPassword(IBCHeader* header, const char* query, BCPassword* pwd)
{
    VerifyPasswordCtx ctx = {NULL, header, kBCUserSession_Default};
    return (kBCUserOk == BCUserCreatePassword(query, verifyNewPasswordCallback, &ctx, pwd));
}

static bool createPassword(const char* query, BCPassword* pwd)
{
    return (kBCUserOk == BCUserCreatePassword(query, NULL, NULL, pwd));
}


/**
  @brief Function is used as callback for protecting existing hidden parts during creating another one.
  Asks the password from an existing hidden part, decodes it's key and substracts the blocks from
  the free space map.
*/
static BCContainerResult protectHiddenPartCallback(void *header_, const char* file, SpaceMap *freeSpaceMap)
{
    IBCHeader *header = (IBCHeader*)header_;
    BCUserResult userResult;
    sm_result spaceMapResult;
    BCVolumeInfo volumeInfo;
    int key;
    char pool[512] = { 0 }; // Hidden parts are encrypted with zero pool.
    VerifyPasswordCtx ctx;
    BCPassword password;

    // Get the password from an existing hidden part.
    ctx.file = file;
    ctx.flags = kBCUserSession_HiddenPartOnly;
    ctx.header = header;

    userResult = BCUserVerifyPassword("Enter password for existing hidden part:", verifyPasswordCallback,
                                      (void*)&ctx, NULL, &password);
    if (userResult != kBCUserOk)
        return kBCContainer_Cancelled;

    // Decode hidden part information.
    BCContainerResult result = header->decodeKey(password.buf, &volumeInfo, &key, pool);

    if (kBCContainer_Success != result)
        return result;

    if (!volumeInfo.hidden)
        return kBCContainer_InvalidPassword;

    // Add an extent with the space map itself. It should also be protected.
    spaceMapResult = SpaceMapAddNew(volumeInfo.map, volumeInfo.offset, 4096);
    if (kSpaceMap_Success != spaceMapResult)
    {
        if (kSpaceMap_InvalidArg == spaceMapResult)
            return kBCContainer_InvalidArgument;
        if (kSpaceMap_NoMem == spaceMapResult)
            return kBCContainer_NoMemory;
        return kBCContainer_InternalError;
    }

    // Substract hidden part blocks from the free space map.
    sm_result smres = SpaceMapSubstract(freeSpaceMap, volumeInfo.map);
    if (kSpaceMap_Success != smres)
    {
        if (kSpaceMap_NoMem == smres)
            return kBCContainer_NoMemory;
        return kBCContainer_InternalError;
    }

    return kBCContainer_Success;
}


//////////////////////////////////////////////////////////////////////////
//
// Interface
//
//////////////////////////////////////////////////////////////////////////



BCContainerResult BCCreateContainer(const char* path, int algid, int modeid,
                                    bc_u64 size, const char* description,
                                    BCProgressCallback progress,
                                    void* arg, BCVolume* out_volume)
{
    BCPassword pwd;
    if (!createPassword("Enter new container password: ", &pwd)) {
        return kBCContainer_Cancelled;
    }

    BCHeaderHandle header(gDefaultKeygen->kgcreate());
    if (!header) {
        return kBCContainer_InternalError;
    }

    BCContainerResult error = header->init(algid, modeid, size, description, pwd.buf,
                                           progress, arg);
    if (kBCContainer_Success != error) {
        return error;
    }
     
    BCVolumeInfo info;
    int key_handle;
    char pool[512];

    error = header->decodeKey(pwd.buf, &info, &key_handle, pool);
    if (kBCContainer_Success != error) {
        return error;
    }

    error = BCInitVolume(&info, path, key_handle, pool, out_volume);

    if (kBCContainer_Success != error) {
        BCKernelFreeKey(key_handle);
        return error;
    }

    error = BCRandomizeVolumeBody(out_volume, progress, arg);
    if (kBCContainer_Success != error) {
        BCFileDelete(path);
        BCPutVolume(out_volume);
        return error;
    }

    if (kBCContainer_Success != header->store(path)) {
        BCFileDelete(path);
        BCPutVolume(out_volume);
        return kBCContainer_InternalError;
    }

    return kBCContainer_Success;
}


bool BCParseContainerFile(const char* file, BCVolumeInfo* out_info)
{
    for (unsigned i = 0; i < BC_TOTAL_KEYGENS; ++i) {
        if (kBCContainer_Success == gKeygens[i]->kgparse(file, out_info)) {
            return true;
        }
    }

    return false;
}


BCContainerResult BCDecodeVolume(const char* file, const char* query, bool mainPartOnly,
                                 BCMountParams *mountParams, BCPassword *out_password,
                                 BCVolume *out_volume)
{
    memset(out_volume, 0, sizeof(*out_volume));
    BCPassword tmpPassword;

    if (!out_password)
        out_password = &tmpPassword;

    BCHeaderHandle header(verifyPassword(file, query, 
            mainPartOnly? kBCUserSession_MainPartOnly : kBCUserSession_Default, 
            mountParams, out_password));
    
    if (!header) { 
        return kBCContainer_InvalidPassword;
    }

    BCVolumeInfo info;
    int handle;
    char pool[512];

    BCContainerResult error = header->decodeKey(out_password->buf, &info, &handle, pool);
    if (error != kBCContainer_Success) {
        return error;
    }

    error = BCInitVolume(&info, file, handle, pool, out_volume);
    if (error != kBCContainer_Success) {
        BCKernelFreeKey(handle);
        return error;
    }

    return kBCContainer_Success;
}

BCContainerResult BCDecodeVolumeByPassword(const char *file, BCPassword *password,
                                           BCVolume *out_volume)
{
    if (!file || !password || !out_volume)
        return kBCContainer_InvalidArgument;

    VerifyPasswordCtx ctx;

    ctx.file = file;
    ctx.flags = 0;
    ctx.header = NULL;

    // Get the header of the main part.
    if (verifyPasswordCallback(password->buf, password->length, (void*)&ctx)) {
        return kBCContainer_InvalidPassword;
    }

    BCHeaderHandle header(ctx.header);
    if (!header) {
        return kBCContainer_InvalidPassword;
    }

    BCVolumeInfo info;
    int handle;
    char pool[512];

    BCContainerResult error = header->decodeKey(password->buf, &info, &handle, pool);
    if (error != kBCContainer_Success) {
        return error;
    }

    error = BCInitVolume(&info, file, handle, pool, out_volume);
    if (error != kBCContainer_Success) {
        BCKernelFreeKey(handle);
        return error;
    }

    return kBCContainer_Success;
}


BCContainerResult BCAddPassword(const char* container)
{
    BCPassword password;
    BCHeaderHandle header(verifyPassword(container, "Enter existing container password:",
                                         NULL, &password));
    if (!header) {
        return kBCContainer_InvalidPassword;
    }

    BCPassword newPassword;
    if (!askNewPassword(header.ptr, "Enter new container password:", &newPassword)) {
        return kBCContainer_Cancelled;
    }

    BCContainerResult error = header->addPassword(password.buf, newPassword.buf);
    if (!error) {
        header->store(container);
    }

    return error;
}


BCContainerResult BCChangePassword(const char* container)
{
    BCPassword password;
    BCHeaderHandle header(verifyPassword(container, "Enter container password you would like to change:",
                                         NULL, &password));
    if (!header) {
        return kBCContainer_InvalidPassword;
    }

    BCPassword newPassword;
    if (!askNewPassword(header.ptr, "Enter new container password:", &newPassword)) {
        return kBCContainer_Cancelled;
    }

    BCContainerResult error = header->changePassword(password.buf, newPassword.buf);
    if (!error) {
        header->store(container);
    }

    return error;
}


BCContainerResult BCRemovePassword(const char* container)
{
    BCPassword password;
    BCHeaderHandle header(verifyPassword(container, "Enter container password you would like to remove:", NULL, &password));
    if (!header) {
        return kBCContainer_InvalidPassword;
    }

    BCContainerResult error = header->removePassword(password.buf);
    if(!error) {
        header->store(container);
    }

    return error;
}


BCContainerResult BCAddSecretSharingScheme(const char* /*container*/, BCVolume* /*out_volume*/)
{
    return kBCContainer_Unsupported;
}


BCContainerResult BCAddKeyPair(const char* /*container*/, BCVolume* /*out_volume*/)
{
    return kBCContainer_Unsupported;
}


BCContainerResult BCBackupHeader(const char* container, const char* backup)
{
    BCHeaderHandle header(verifyPassword(container, "Enter container password to create backup:"));
    if (!header) {
        return kBCContainer_InvalidPassword;
    }

    return header->store(backup);
}


BCContainerResult BCRestoreHeader(const char* container, const char* backup)
{
    BCHeaderHandle header(verifyPassword(backup, "Enter container password to restore backup:"));
    if (!header) {
        return kBCContainer_InvalidPassword;
    }

    return header->store(container);
}


BCContainerResult BCEncryptHeader(const char* container)
{
    BCPassword pwd;
    BCHeaderHandle header(verifyMainPartPassword(container, "Enter container password to encrypt header:", NULL, &pwd));
    if (!header) {
        return kBCContainer_Cancelled;
    }

    if (header->isEncrypted()) {
        return kBCContainer_Success;
    }

    BCContainerResult error = header->encrypt(pwd.buf);
    if (!error) {
        header->store(container);
    }

    return error;
}


BCContainerResult BCDecryptHeader(const char* container)
{
    BCPassword pwd;
    BCHeaderHandle header(verifyMainPartPassword(container, "Enter container password to decrypt header:", NULL, &pwd));
    if (!header) {
        return kBCContainer_Cancelled;
    }

    if (!header->isEncrypted()) {
        return kBCContainer_Success;
    }

    BCContainerResult error = header->decrypt(pwd.buf);
    if(!error) {
        header->store(container);
    }

    return error;
}


BCContainerResult BCUpdateDescription(const char* container, const char* description)
{
    BCPassword pwd;
    BCHeaderHandle header(verifyMainPartPassword(container, "Enter container password to update description:",
                                                 NULL, &pwd));
    if(!header) {
        return kBCContainer_InvalidPassword;
    }

    bool encrypted = header->isEncrypted();
    if(encrypted) {
        header->decrypt(pwd.buf);
    }

    header->setDescription(description);

    if(encrypted) {
        header->encrypt(pwd.buf);
    }

    return kBCContainer_Success;
}


static SpaceMap *TransformToSubmapOfSize(SpaceMap *map, bc_u64 size)
{
    if (size != 0)
    {
        // Check that the spacemap has enough space.
        bc_u64 totalFree = 0;
        sm_result result = SpaceMapGetLength(map, &totalFree);

        if (result != kSpaceMap_Success)
        {
            return NULL;
        }

        if (totalFree < size)
        {
            return NULL;
        }

        // Getting submap of desired size.
        SpaceMap *mapOfSize = 0;
        result = SpaceMapGetMapOfSize(map, size, &mapOfSize);

        if (kSpaceMap_Success != result)
        {
            return NULL;
        }

        return mapOfSize;
    }

    return NULL;
}


BCContainerResult BCCreateHiddenPart(const char* file, bc_u64 size,
            SpaceMap *freeSpace, BCPassword *mainPassword, BCProgressCallback progress,
            void *progressArg, BCPassword *out_hiddenPassword, char *fsType)
{
    if (!out_hiddenPassword || !mainPassword || !freeSpace || !file)
        return kBCContainer_InvalidArgument;

    BCUserResult userResult;
    VerifyPasswordCtx ctx;
    SpaceMap *transformedMap = NULL;

    ctx.file = file;
    ctx.flags = kBCUserSession_MainPartOnly;
    ctx.header = NULL;

    // Get the header of the main part.
    if (verifyPasswordCallback(mainPassword->buf, mainPassword->length, (void*)&ctx)) {
        return kBCContainer_InvalidPassword;
    }

    BCHeaderHandle header(ctx.header);
    if (!header) {
        return kBCContainer_InvalidPassword;
    }

    // Get parameters of a new hidden part from user.
    userResult = BCUserProtectHiddenPart((void*)header.get(), file,
                                         protectHiddenPartCallback, freeSpace,
                                         fsType, &transformedMap);

    if (kBCUserOk != userResult)
        return kBCContainer_Cancelled;

    if (size && !transformedMap)
    {
        // If user specified the hidden part size, we need to get
        // a submap from the free space map.
        transformedMap = TransformToSubmapOfSize(freeSpace, size);

        if (!transformedMap)
            return kBCContainer_NoMemory;
    }

    if (transformedMap)
    {
        freeSpace = transformedMap;
    }

    // Ask for the password for a new hidden part.
    if (!askNewPassword(header.get(), "Enter password for new hidden part:", out_hiddenPassword))
    {
        SpaceMapFree(transformedMap);
        return kBCContainer_Cancelled;
    }

    // Create a key for the new hidden part.
    BCContainerResult containerResult = header->createHiddenKey(freeSpace, mainPassword,
                                                    out_hiddenPassword, progress, progressArg);
    SpaceMapFree(transformedMap);

    if (kBCContainer_Success != containerResult)
    {
        return containerResult;
    }

    // Store header in the container file.
    containerResult = header->store(file);

    if (kBCContainer_Success != containerResult)
        return containerResult;

    return kBCContainer_Success;
}


////////////////////////////////////////////////////////////////////////////////////
//
//  Reencryption
//
////////////////////////////////////////////////////////////////////////////////////

    
static bool initRRStateBackup(const char* file)
{
    bc_file_handle fd;
    if (kBCFileResult_Success != BCFileOpen(file, kBCFileMode_ReadWrite, &fd)) {
        return false;
    }

    bc_u64 offset = 0;
    BCFileSeek(&fd, 0, kBCFileSeek_End);
    if(sizeof(offset) != BCFileWrite(&fd, &offset, sizeof(offset))) {
        BCFileClose(&fd);
        return false;
    }

    BCFileSync(&fd);
    BCFileClose(&fd);
    return true;
}

static bool readRRState(const char* file, bc_u64* offset)
{
    bc_file_handle fd;
    if (kBCFileResult_Success != BCFileOpen(file, kBCFileMode_Read, &fd)) {
        return false;
    }

    bc_u64 size = BCFileSize(file);
    if (sizeof(*offset) != BCFileReadWithOffset(&fd, offset, sizeof(*offset), size - sizeof(*offset))) {
        BCFileClose(&fd);
        return false;
    }

    BCFileSync(&fd);
    BCFileClose(&fd);
    return true;
}


static bool writeRRState(const char* file, bc_u64 offset)
{
    bc_file_handle fd;
    if (kBCFileResult_Success != BCFileOpen(file, kBCFileMode_ReadWrite, &fd)) {
        return false;
    }

    bc_u64 size = BCFileSize(file);
    if (sizeof(offset) != BCFileWriteWithOffset(&fd, &offset, sizeof(offset), size - sizeof(offset))) {
        BCFileClose(&fd);
        return false;
    }

    BCFileSync(&fd);
    BCFileClose(&fd);
    return true;
}

#define BC_VOLUME_BUFFER_SIZE (1024 * 1024)


static BCContainerResult reencryptVolume(BCVolume* src, BCVolume* dst, bc_u64 start, const char* file,
                                         const char* backup, BCProgressCallback progress, void* arg)
{
    BCCipher src_cipher;
    if(kBCKernel_Success != src_cipher.init(src->info.alg, src->info.mode)) {
        return kBCContainer_InvalidAlgorithm;
    }

    BCCipher dst_cipher;
    if(kBCKernel_Success != dst_cipher.init(dst->info.alg, dst->info.mode)) {
        return kBCContainer_InvalidAlgorithm;
    }

    bc_file_handle fd;
    if(kBCFileResult_Success != BCFileOpen(file, kBCFileMode_ReadWrite, &fd)) {
        return kBCContainer_IOError;
    }

    char* buffer = (char*) malloc(BC_VOLUME_BUFFER_SIZE);
    if(!buffer) {
        BCFileClose(&fd);
        return kBCContainer_NoMemory;
    }

    bc_u64 offset = start;
    bc_u64 size = src->info.size;
    bc_u64 iv;
    BCContainerResult error = kBCContainer_Success;

    while(offset < size) {
        size_t nbytes = size - offset;
        nbytes = nbytes > BC_VOLUME_BUFFER_SIZE ? BC_VOLUME_BUFFER_SIZE : nbytes;

        iv = offset / 512 + 1;

        if(nbytes != BCFileReadWithOffset(&fd, buffer, nbytes, src->info.offset + offset)) {
            error = kBCContainer_IOError;
            break;
        }

        src_cipher.decrypt(iv, src->key_handle, src->pool, buffer, nbytes);
        
        dst_cipher.encrypt(iv, dst->key_handle, dst->pool, buffer, nbytes);

        if(nbytes != BCFileWriteWithOffset(&fd, buffer, nbytes, dst->info.offset + offset)) {
            error = kBCContainer_IOError;
            break;
        }

        BCFileSync(&fd);

        offset += nbytes;

        if (!writeRRState(backup, offset)) {
            error = kBCContainer_IOError;
            break;
        }

        if(progress && !progress("Reencrypting", offset, size, arg)) {
            fprintf(stderr, "Reencryption terminated at offset %lu\n",
                    (unsigned long)offset);
            error = kBCContainer_Cancelled;
            break;
        }
    }

    BCFileClose(&fd);
    free(buffer);

    return error;
}


BCContainerResult BCReencrypt(const char* container, int algid, int modeid,
                              BCProgressCallback progress, void* arg)
{
    BCVolume src;
    BCVolume dst;

    // Get main part password
    BCPassword pwd;
    BCHeaderHandle header(verifyMainPartPassword(container, "Please enter password to reencrypt container:", NULL, &pwd));
    if (!header) {
        return kBCContainer_Cancelled;
    }

    char backupFile[PATH_MAX] = {0};
    snprintf(backupFile, sizeof(backupFile) - 1, "%s.rrkbb", container);

    const BCKeygen* keygen = header->keygen;
    BCContainerResult error = kBCContainer_Success;

    if (header->isMarkedForReencryption()) {

        // looks like previous reencryption process was interrupted
        // container header we've just read has new key and algorithm 
        // we need to get old key and algorithm from the backup file with the same password and restore the state to continue

        BCHeaderHandle oldHeader(keygen->kgload(backupFile));
        if (!oldHeader) {
            error = kBCContainer_UnknownFormat;
            goto OUT;
        }

        error = oldHeader->decodeKey(pwd.buf, &src.info, &src.key_handle, src.pool);
        if (error) {
            goto OUT;
        }

        error = header->decodeKey(pwd.buf, &dst.info, &dst.key_handle, dst.pool);
        if (error) {
            goto OUT;
        }

        bc_u64 offset;
        if (!readRRState(backupFile, &offset)) {
            error = kBCContainer_IOError;
            goto OUT;
        }

        fprintf(stderr, "Resuming reencryption from offset %lu\n", (unsigned long)offset);
        error = reencryptVolume(&src, &dst, offset, container, backupFile, progress, arg);
        if (!error) {
            header->markForReencryption(false);
            header->store(container);
        }

    } else {

        // begin new reencryption process
        error = header->decodeKey(pwd.buf, &src.info, &src.key_handle, src.pool);
        if (error) {
            return error;
        }

        error = header->store(backupFile);
        if (error) {
            goto OUT;
        }

        BCHeaderHandle newHeader(keygen->kgcreate());
        if (!newHeader) {
            error = kBCContainer_NoMemory;
            goto OUT;
        }

        error = newHeader->init(algid, modeid, src.info.size, src.info.description, pwd.buf,
                                progress, arg);
        if (error) {
            goto OUT;
        }

        assert(!src.info.hidden); // we can't reencrypt hidden parts

        error = newHeader->decodeKey(pwd.buf, &dst.info, &dst.key_handle, dst.pool);
        if (error) {
            goto OUT;
        }

        newHeader->markForReencryption(true);

        error = newHeader->store(container);
        if (error) {
            goto OUT;
        }

        if (!initRRStateBackup(backupFile)) {
            error = kBCContainer_IOError;
            goto OUT;
        }

        // Begin reencryption
        error = reencryptVolume(&src, &dst, 0, container, backupFile, progress, arg);
        if (!error) {
            newHeader->markForReencryption(false);
            newHeader->store(container);
        }
    }
    
    if (!error) {
        BCFileDelete(backupFile);
    }
        
OUT:
    BCPutVolume(&src);
    BCPutVolume(&dst);

    return error;
}








