/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * Container library interface
 */

#ifndef BC_CONTAINER_H
#define BC_CONTAINER_H

#include "bctypes.h"
#include "keygen.h"
#include "volume.h"
#include "error.h"

#define BC_MAX_PATH 4096

/// @brief Mount parameters to pass to user dialog.
typedef struct _BCMountParams
{
    bool mountReadOnly;             ///< Will container need to be mounted as read-only?
    bool useDefaultMountPoint;      ///< Should we use user-specified mount point or default (ignored on Linux).
    char mountPoint[BC_MAX_PATH];   ///< User-specified mount point.
} BCMountParams;

#ifdef __cplusplus
extern "C" {
#endif


    /**
     * Create a new container file
     * @param path          File path
     * @param algid         Encryption algorithm id
     * @param modeid        Encryption mode id
     * @param kgid          Key generator id
     * @param size          Container data size in bytes
     * @param description   Optional description string
     * @param progress      Progress callback
     * @param arg           Callback caller cookie
     */
    BCContainerResult BCCreateContainer(const char* path, int algid, int modeid, bc_u64 size,
                                        const char* description, BCProgressCallback progress,
                                        void* arg, BCVolume* out_vol);


    /**
     * Create new hidden part inside container.
     * @param file          Name of the container file.
     * @param freeSpace     Space map containing blocks for future hidden part.
     * @param out_hidden    Initialized new hidden volume on success.
     * @param progress      Progress callback.
     * @param progressArg   Progress callback cookie.
     * @param fsType        Parameter for GUI to transfer fs type from hidden part protection dialog.
     */
    BCContainerResult BCCreateHiddenPart(const char *file, bc_u64 size, SpaceMap *freeSpace,
                            BCPassword *mainPassword, BCProgressCallback progress,
                            void *progressArg, BCPassword *out_hiddenPassword, char *fsType);


    /**
     * Parse container file and return its main encoded volume description on success.
     * This procedure will not ask passwords so it will fail for containers with encrypted header.
     */
    bool BCParseContainerFile(const char* file, BCVolumeInfo* out_info);

    
    /**
     * Create container header backup
     * @param container     Path to container file
     * @param backup        Path to backup file
     */
    BCContainerResult BCBackupHeader(const char* container, const char* backup);


    /**
     * Restore container header from backup.
     * @param container     Path to container file
     * @param backup        Path to backup file
     */
    BCContainerResult BCRestoreHeader(const char* container, const char* backup);


    /**
     * Encrypt container header.
     * @param container     Path to container file
     */
    BCContainerResult BCEncryptHeader(const char* container);


    /**
     * Decrypt container header.
     * @param container     Path to container file
     */
    BCContainerResult BCDecryptHeader(const char* container);


    /**
     * Change one of the existing container passwords.
     * @param container     Path to container file
     * @param out_volume    Decoded volume info whose password was changed, can be NULL.
     */
    BCContainerResult BCChangePassword(const char* container);


    /**
     * Remove one of the existing container passwords.
     * Last password cannot be removed.
     * @param container     Path to container file
     * @param out_volume    Decoded volume info whose password was removed, can be NULL.
     */
    BCContainerResult BCRemovePassword(const char* container);


    /**
     * Add another container password.
     * @param container     Path to container file
     * @param out_volume    Decoded volume info that had a new password added, can be NULL.
     */
    BCContainerResult BCAddPassword(const char* container);


    /**
     * Add another secret sharing scheme.
     * @param container     Path to container file
     * @param out_volume    Decoded volume info that had a new sharing scheme added, can be NULL.
     */
    BCContainerResult BCAddSecretSharingScheme(const char* container, BCVolume* out_volume);


    /**
     * Add another private/public key pair.
     * @param container     Path to container file
     * @param out_volume    Decoded volume info that had a new key pair added, can be NULL.
     */
    BCContainerResult BCAddKeyPair(const char* container, BCVolume* out_volume);


    /**
     * Change container description.
     * Current policy is to ask container password.
     * @param container     Path to container file
     * @param description   New container description.
     */
    BCContainerResult BCUpdateDescription(const char* container, const char* description);


    /**
     * Reencrypt container body using new algorithm, mode and key.
     * Hidden part reencryption is not supported.
     * @param container     Path to container file
     * @param algid         New encryption algorithm id
     * @param modeid        New encryption mode id
     * @param progress      Progress reporting callback
     * @param arg           Progress callback argument
     */
    BCContainerResult BCReencrypt(const char* container, int algid, int modeid, BCProgressCallback progress, void* arg);


    /**
     * Ask password and decode volume encoded in container.
     * @param file          Container file path.
     * @param query         Password query to display to the user.
     * @param mainPartOnly  Have to parse only main volume.
     * @param password      Struct to store volume password. Can be NULL.
     * @param out_volume    Decoded volume on success. Should be released with BCPutVolume.
     */
    BCContainerResult BCDecodeVolume(const char* file, const char* query, bool mainPartOnly,
                                     BCMountParams *mountParams, BCPassword *out_password,
                                     BCVolume* out_volume);


    BCContainerResult BCDecodeVolumeByPassword(const char *file, BCPassword *password,
                                     BCVolume *out_volume);

#ifdef __cplusplus
}
#endif

#endif

