#ifndef BC_FILE_H
#define BC_FILE_H

#include "intsize.h"

#ifdef BC_LINUX
#include <linux/limits.h>
typedef int bc_file_handle;
#else
    #error "Not supported"
#endif
#define BC_FILE_ERR ((bc_u64)-1)
/**
 * List of modes that can be used while opening file.
 */
enum
{
    kBCFileMode_Read        = 0x01,
    kBCFileMode_ReadWrite   = 0x02,

    kBCFileMode_Exclude     = 0x10 // do not create file if it does not exist and return an error instead
};

typedef int bc_file_mode;

/**
 * List of positions from which seek operation can be performed.
 */
typedef enum
{
    kBCFileSeek_Begin       = 0,
    kBCFileSeek_Current     = 1,
    kBCFileSeek_End         = 2
} bc_file_seek;

typedef enum
{
    kBCFileResult_Success    = 0, ///< Operation successful.
    kBCFileResult_NoAccess,       ///< File doesn't exists or access denied.
    kBCFileResult_InternalError   ///< Unspecified error. In this case errno and stderr-output can be checked.
} bc_file_result;

/*#ifdef __cplusplus
extern "C"
{
#endif*/
    /**
     * Opens specified file.
     * @param fileName      Name of file to open.
     * @param mode          Opening mode.
     * @return              Possible returns:
     *                      - kBCFileResult_Success, if succeeded, 
     *                      - kBCFileResult_NoFile, if file was not found,
     *                      - kBCFileResult_NoAccess, if file access denied,
     *                      - kBCFileResult_InternalError, if internal error occured. 
     *                        In this case errno will be set.
     */
    bc_file_result BCFileOpen(const char *fileName, bc_file_mode mode, bc_file_handle *handle);
    
    /**
     * Repositions the offset of the open file.
     * @param handle        File handle returned by BCFileOpen call.
     * @param offset        Offset.
     * @param seek          Describes position, from which offset is defined.
     * @return              Current position of pointer from the beginning of file 
     *                      or (bc_u64)-1, if error occured.
     */
    bc_u64 BCFileSeek(bc_file_handle *handle, bc_u64 offset, bc_file_seek seek);

    /**
     * Returns current offset location from the beginning of file.
     * @param handle        File handle returned by BCFileOpen call.
     * @return              Current position of pointer from the beginning of file
     *                      or (bc_u64)-1, if error occured.
     */
    static inline bc_u64 BCFileGetPosition(bc_file_handle *handle) 
    {
        return BCFileSeek(handle, 0, kBCFileSeek_Current);
    }

    /**
     * Reads a part of file with offset from the beginning of the file.
     * @param handle        File handle returned by BCFileOpen call.
     * @param buffer        Data buffer.
     * @param byteCount     Data buffer size in bytes.
     * @param offset        Number of bytes from the beginning of file 
     *                      that must be skipped before starting reading.
     * @return              Number of bytes actually read or (bc_u64)-1, if error occured.
     */
    bc_u64 BCFileReadWithOffset(bc_file_handle *handle, void *buffer, bc_u64 byteCount, bc_u64 offset);

    /**
     * Reads a part of file starting from current position.
     * @param handle        File handle returned by BCFileOpen call.
     * @param buffer        Data buffer. 
     * @param byteCount     Data buffer size in bytes.
     * @return              Number of bytes actually read or (bc_u64)-1, if error occured.
     */
    bc_u64 BCFileRead(bc_file_handle *handle, void *buffer, bc_u64 byteCount);

    /**
     * Writes buffer to the file starting from the offset from the beginning of file.
     * @param handle File handle returned by BCFileOpen call.
     * @param buffer        Data buffer to write.
     * @param byteCount     Buffer size in bytes.
     * @param offset        Number of bytes from the beginning of file that must be skipped.
     * @return              0 if success, otherwise -1.
     */
    bc_u64 BCFileWriteWithOffset(bc_file_handle *handle, const void *buffer, bc_u64 byteCount, bc_u64 offset);

    /**
     * Writes buffer to the file starting from current location.
     * @param handle        File handle returned by BCFileOpen call.
     * @param buffer        Data buffer to write.
     * @param byteCount     Buffer size in bytes.
     * @return              0, if succeeded, otherwise -1.
     */
    bc_u64 BCFileWrite(bc_file_handle *handle, const void *buffer_, bc_u64 byteCount);

    /**
     * Closes specified file handle.
     * @param handle        Handle to close.
     * @return              0, if succeeded, otherwise -1.
     */
    int BCFileClose(bc_file_handle *handle);

    /**
     * Get file size in bytes.
     */
    bc_u64 BCFileSize(const char* path);

    /**
     * Sync file contents with the disk
     */
    void BCFileSync(bc_file_handle* fd);

    /**
     * Delete given file.
     * @param path          File system path.
     */
    void BCFileDelete(const char* path);

    /**
     * Returns string description for the operation result.
     * @param result        Result of the operation.
     * @return              String representation of the result.
     */
    const char *BCFileGetStringResult(bc_file_result result);

/*#ifdef __cplusplus
}
#endif*/
#endif //BC_FILE_H

