#include "../file.h"
#include "../exception.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <unistd.h>

#define SHOW_SYS_ERR(msg) fprintf(stderr, msg": %s\n\n", strerror(errno));

/// Convers errno code to our own error.
/// If no appropriate type exists, prints
/// error message to stderr and returns InternalError.
static bc_file_result ConvertError(int errnoCode)
{
    switch (errnoCode)
    {
        case EACCES:
        case ENOENT:
            return kBCFileResult_NoAccess;        
        default:
            SHOW_SYS_ERR("Internal error");
            return kBCFileResult_InternalError;
    }
}

static bc_u32 ConvertMode(bc_file_mode mode)
{
    bc_u32 systemMode = O_LARGEFILE | O_CREAT;
    
    if (mode & kBCFileMode_ReadWrite) {
        systemMode |= O_RDWR;
    } else if (mode & kBCFileMode_Read) {
        systemMode |= O_RDONLY;
    } else {
        assert(0 && "unsupported mode");
    }

    if (mode & kBCFileMode_Exclude) {
        systemMode |= O_EXCL;
    }

    return systemMode;
}

static int ConvertSeek(bc_file_seek seek)
{
    switch(seek)
    {
        case kBCFileSeek_Begin:
            return SEEK_SET;

        case kBCFileSeek_Current:
            return SEEK_CUR;

        case kBCFileSeek_End:
            return SEEK_END;

        default:
            assert(0 && "Unsupported mode");
    }
}

bc_file_result BCFileOpen(const char *fileName, bc_file_mode mode, bc_file_handle *handle)
{
    int fd = open(fileName, ConvertMode(mode), S_IRWXU);
    if (fd < 0) {
        return ConvertError(errno);
    }

    *handle = fd;
    return kBCFileResult_Success;
}

bc_u64 BCFileSeek(bc_file_handle *handle, bc_u64 offset, bc_file_seek seek) 
{
    assert(sizeof(off_t) == 8);
    int whence = ConvertSeek(seek);
    int fd = *handle;
    off_t resultOffset = lseek(fd, offset, whence);
    if (((off_t)-1) != resultOffset)
        return resultOffset;
    SHOW_SYS_ERR("Error while seeking");
    return BC_FILE_ERR;
}

bc_u64 BCFileRead(bc_file_handle *handle, void *buffer, bc_u64 byteCount)
{
    if (!buffer)
    {
        fprintf(stderr, "Can't read to null buffer.");
        return BC_FILE_ERR;
    }
    if (0 == byteCount)
        return 0;
 
    int fd = (*handle);

    // Save starting position to return to it in case of a error.
    bc_u64 startPosition = BCFileGetPosition(handle);
    if (BC_FILE_ERR == startPosition)
        return BC_FILE_ERR;

    ssize_t bytesRead = 0;
    ssize_t bytesLeft = byteCount;
    int error = 0;
    int tryCount = 0;
    const int maxTryCount = 100;

    do
    {
        bytesRead = read(fd, buffer, bytesLeft);

        if (-1 == bytesRead)
        {
            error = errno;

            // Offset is undefined in some error cases, set it to start position.
            if (BC_FILE_ERR == BCFileSeek(handle, startPosition, kBCFileSeek_Begin))
                break;

            if ((EINTR == error) || (EAGAIN == error) || (EWOULDBLOCK == error))
            {
                tryCount++;
                if (tryCount >= maxTryCount)
                    break;
                continue;
            }
            else
            {
                fprintf(stderr, "Read error: %s\n", strerror(error));
                break;
            }
        }
        buffer = (char*)buffer + bytesRead;
        bytesLeft -= bytesRead;
    }
    while (bytesRead && bytesLeft);

    // On error try to set file pointer to start position before return.
    if (-1 == bytesRead)
    {
        return BC_FILE_ERR;
    }
    return (bc_u64)bytesRead;

}

bc_u64 BCFileReadWithOffset(bc_file_handle *handle, void *buffer, bc_u64 byteCount, bc_u64 offset) 
{
    if (0 == byteCount)
        return 0;

    if (!buffer)
        return BC_FILE_ERR;

    bc_u64 startPosition = BCFileGetPosition(handle);
    if (BC_FILE_ERR == startPosition)
        return BC_FILE_ERR;
    bc_u64 seekResult = BCFileSeek(handle, offset, kBCFileSeek_Begin);
    if (BC_FILE_ERR == seekResult)
        return BC_FILE_ERR;
    bc_u64 bytesRead = BCFileRead(handle, buffer, byteCount); // No need to check for error here.
   
    seekResult = BCFileSeek(handle, startPosition, kBCFileSeek_Begin);
    if (BC_FILE_ERR == seekResult)
        return BC_FILE_ERR;
    return bytesRead;
}

bc_u64 BCFileWrite(bc_file_handle *handle, const void *buffer_, bc_u64 byteCount)
{
    if (!buffer_)
        return BC_FILE_ERR;
    if (0 == byteCount)
        return 0;
    
    int fd = (*handle);
    ssize_t bytesWritten = 0;
    int error = 0;
    ssize_t bytesLeft = byteCount;
    bc_u64 startPosition = BCFileGetPosition(handle);
    bc_u64 offset = startPosition;
    char *buffer = (char*)buffer_;
    int tryCount = 0;
    const int maxTryCount = 100;

    // Try to write buffer to the file.
    do
    {
        bytesWritten = write(fd, buffer, bytesLeft);

        if (-1 == bytesWritten)
        {
            // Some error happened. Pointer might be positioned, so move it to the offset.
            error = errno;

            if (BC_FILE_ERR == BCFileSeek(handle, offset, kBCFileSeek_Begin))
                break;

            if ((EINTR == error) || (EAGAIN == error) || (EWOULDBLOCK == error))
            {
                // Analyse count of failed writings.
                tryCount++;
                if (tryCount >= maxTryCount)
                    break;
                continue;
            }
            else
                break;
        }

        buffer += bytesWritten;
        bytesLeft -= bytesWritten;
        offset += bytesWritten;
        tryCount = 0; // If writing was succeeded, reset try-count.
    }
    while (bytesWritten && bytesLeft);

    if (-1 == bytesWritten)
    {
        fprintf(stderr, "BCFileWrite: %s\n", strerror(errno));
        return -1;
    }

    return bytesWritten;
}

bc_u64 BCFileWriteWithOffset(bc_file_handle *handle, const void *buffer, bc_u64 byteCount, bc_u64 offset)
{
    if (0 == byteCount)
        return 0;

    if (!buffer)
        return -1;

    bc_u64 startPosition = BCFileGetPosition(handle);
    if (BC_FILE_ERR == startPosition)
        return -1;

    bc_u64 seekResult = BCFileSeek(handle, offset, kBCFileSeek_Begin);
    if (BC_FILE_ERR == seekResult)
        return -1;

    int writeResult = BCFileWrite(handle, buffer, byteCount);
    if (BC_FILE_ERR == BCFileSeek(handle, startPosition, kBCFileSeek_Begin))
        return -1;

    return writeResult;
}

void BCFileSync(bc_file_handle* handle)
{
    fsync(*handle);
}

int BCFileClose(bc_file_handle *handle) 
{
    int fd = *handle;
    fsync(fd);

    if (close(fd))
    {
        fprintf(stderr, "Descriptor closing error: %s\n", strerror(errno));
        return -1; 
    }

    return 0;
}

bc_u64 BCFileSize(const char* path)
{
    bc_file_result result;
    bc_file_handle handle;
    result = BCFileOpen(path, kBCFileMode_Read, &handle);
    if (kBCFileResult_Success != result) {
        fprintf(stderr, "Error opening file: %s", BCFileGetStringResult(result));
        return 0;
    }
    bc_u64 seekResult = BCFileSeek(&handle, 0, kBCFileSeek_End);
    if (BC_FILE_ERR == seekResult)
        return 0;
    BCFileClose(&handle);
    return seekResult;
}

void BCFileDelete(const char* path)
{
    int error = unlink(path);
    if(error) {
        fprintf(stderr, "unlink(%s) failed: %s\n", path, strerror(errno));
    }
}


const char *BCFileGetStringResult(bc_file_result result)
{
    switch (result) {
    case kBCFileResult_Success:
        return "Success";
    case kBCFileResult_NoAccess:
        return "Access denied";
    case kBCFileResult_InternalError:
        return "Internal error";
    default:
        return "Value is not implemented for result.";
    }
}
