/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * Linux user auth session implementation.
 */

#ifndef BC_LINUX_USER_H
#define BC_LINUX_USER_H

#include <stdlib.h>
#include <stdio.h>
#include <stdio_ext.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <iconv.h>
#include <locale.h>
#include <langinfo.h>
#include <termios.h>

#include "container/user.h"
#include "container/intsize.h"
#include "session.h"

#define BC_INVALID_ICONV_RESULT ((iconv_t)-1)

static void* SetEchoOff()
{
    struct termios *term_flags = (struct termios*)malloc(sizeof(struct termios));
    if (tcgetattr(fileno(stdin), term_flags)) {
        fprintf(stderr, "tcgetattr failed: %s\n", strerror(errno));
        free(term_flags);
        return NULL;
    }

    tcflag_t old_flags = term_flags->c_lflag;
    term_flags->c_lflag &= ~ECHO;

    if (tcsetattr(fileno(stdin), TCSAFLUSH, term_flags)) {
        fprintf(stderr, "tcsetattr failed: %s\n", strerror(errno));
        free(term_flags);
        return NULL;
    }

    term_flags->c_lflag = old_flags;

    return term_flags;
}

static void SetEchoOn(void *state)
{
    struct termios *term_flags = (struct termios*)state;
    if (tcsetattr(fileno(stdin), TCSAFLUSH, term_flags)) {
        fprintf(stderr, "tcsetattr failed: %s\n", strerror(errno));
    }

    free(term_flags);
}

static int getPasswordString(char *buffer, size_t bufferSize)
{
    void *state = SetEchoOff();

    memset(buffer, '\0', bufferSize);
    // Get the password.
    unsigned i = 0;
    char c = '\0';

    for (; (i < bufferSize) && (c != '\n') && (c != EOF); i++) {
        c = fgetc(stdin);
        if ((c != '\n') && (c != EOF)) {
            buffer[i] = c;
        }
    }

    printf("\n");

    // Restore an old state.
    SetEchoOn(state);

    return 0;
}

static size_t convertString(char *msg, size_t msgSize, char *out_encodedMsg, size_t bufferSize, const char *fromEncoding, const char *toEncoding)
{
    iconv_t descriptor = iconv_open(toEncoding, fromEncoding);

    if (((iconv_t)-1) == descriptor) {
        fprintf(stderr, "Error creating conversion descriptor: %s\n", strerror(errno));
        return 0;
    }

    size_t nbytesLeft = bufferSize;
    if (-1 == (int)iconv(descriptor, &msg, &msgSize, &out_encodedMsg, &nbytesLeft)) {
        fprintf(stderr, "Error while converting: %s\n", strerror(errno));
        return 0;
    }

    iconv_close(descriptor);
    return bufferSize - nbytesLeft;
}

static size_t convertToUTF8(char *msg, size_t length, char *out_resultBuffer, size_t bufferSize)
{
    // Set program's current locale to system locale and use result as source encoding.
    setlocale(LC_ALL, "");
    const char *currentEncoding = nl_langinfo(CODESET);
    const char *targetEncoding = "UTF-8";

    return convertString(msg, length, out_resultBuffer, bufferSize, currentEncoding, targetEncoding);
}


// Will get password converted to UTF8 on success
static bool askPassword(const char* query, BCPassword* out_pwd)
{
    out_pwd->clear();

    printf("Empty password will cancel the operation.\n");
    printf("%s", query);
    char* pwd = (char*)malloc(kBCPasswordMax);

    if (!pwd) {
        return false;
    }

    if (getPasswordString(pwd, kBCPasswordMax)) {
        free(pwd);
        return false;
    }

    out_pwd->length = convertToUTF8(pwd, strlen(pwd), out_pwd->buf, sizeof(out_pwd->buf));
    memset(pwd, 0, strlen(pwd));
    free(pwd);
    return (out_pwd->length != 0);
}


BCUserResult BCUserVerifyPassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCMountParams * /*mountParams*/, BCPassword* out_pwd)
{
	if(!query) {
        query = BC_DEFAULT_PASSWORD_QUERY;
    }

    // Try password from environment if any
    if (const char *env_pwd = getenv("BESTCRYPT_PASSWORD")) {
        if (callback) {
            out_pwd->set(env_pwd);
            const char* error = callback(out_pwd->buf, out_pwd->length, cookie);
            if (!error) {
                return kBCUserOk;
            }
        }
    }

	while(1) {

        if (!askPassword(query, out_pwd)) {
            return kBCUserCancel;
        }

        if (callback) {
            const char* error = callback(out_pwd->buf, out_pwd->length, cookie);
            if (error) {
                printf("%s\n", error);
                continue;
            }
        }

        return kBCUserOk;
	}
}		


BCUserResult BCUserCreatePassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCPassword* out_pwd)
{
    if(!query) {
        query = BC_DEFAULT_PASSWORD_QUERY; 
    }

    BCPassword verpwd;

	while(1) {

        if (!askPassword(query, out_pwd)) {
            return kBCUserCancel;
        }

        if (out_pwd->length < kBCPasswordMin) {
			printf("Password should be at least %d characters.\n", kBCPasswordMin);
            continue;
        }
            
        if (!askPassword("Verify password:", &verpwd)) {
            return kBCUserCancel;
        }

        if (0 != strncmp(out_pwd->buf, verpwd.buf, kBCPasswordMax)) {
            printf("Passwords do not match.\n");
            continue;
        }

        if (callback) {
            const char* error = callback(out_pwd->buf, out_pwd->length, cookie);
            if (error) {
                printf("%s\n", error);
                continue;
            }
        }
        
		return kBCUserOk;
	}
}


static unsigned GenerateRandom(const char *message, char *out_buf, unsigned total,
                               BCProgressCallback callback, void *callbackArg, int strong)
{
    FILE *f;

    if (strong) {
        f = fopen("/dev/random", "r");
    }
    else {
        f = fopen("/dev/urandom", "r");
    }

    if(!f) {
        fprintf(stderr, "Error: open /dev/%s failed: %s\n", (strong)? "random" : "urandom",
                strerror(errno));
        return 0;
    }

    void *state = SetEchoOff();

    unsigned nread = 0;
    if (message) {
        printf("%s\n", message);
    }

    if (strong) {
        printf("Please press random keys and move your mouse (if any) to speed up the process...\n");
    }
    do {
        out_buf[nread++] = fgetc(f);
        if (callback)
        {
            callback("Random generation", nread, total, callbackArg);
        }
    } while(nread != total);

    SetEchoOn(state);

    fclose(f);
    return nread;
}


unsigned BCUserGenerateSeed(const char *message, char *out_buf,
                            unsigned total, BCProgressCallback callback, void *callbackArg)
{
#ifdef BC_DEBUG
    fprintf(stderr, "********************** Warning! ************************\n");
    fprintf(stderr, "Random generator was disabled for testing! \nDo not use containers created while you see this message!\n");
    fprintf(stderr, "********************************************************\n");
    memset(out_buf, 0, total);
    return total;
#else
    return GenerateRandom(message, out_buf, total, callback, callbackArg, 1);
#endif
}


void BCUserMessage(const char* msg, BCUserMessageLevel level)
{
	const char* prefix = "";
	switch (level) {
	case kBCUserMessage_Warning: 
		prefix = "Warning: "; 
		break;

	case kBCUserMessage_Error:	
		prefix = "Error: ";
		break;
	
	default:
		prefix = "";
		break;
	};

	printf("%s%s\n", prefix, msg);
}


static const char* getIconvEncoding(BCUserEncoding tag) {
    switch (tag) {
    case kBCUser_UTF8: return "UTF-8";
    case kBCUser_UTF16LE: return "UTF-16LE";
    case kBCUser_UTF16BE: return "UTF-16BE";
    default: assert(0); return NULL;
    };
}

size_t BCUserConvertString(char* in, size_t inlen, char* out, size_t outlen, BCUserEncoding fromEncoding, BCUserEncoding toEncoding)
{
    const char* from = getIconvEncoding(fromEncoding);
    const char* to = getIconvEncoding(toEncoding);
    return convertString(in, inlen, out, outlen, from, to);
}


size_t BCUserUTF16ToUTF8(char* in_utf16, size_t inlen, char* out_utf8, size_t outlen) {
    return convertString(in_utf16, inlen, out_utf8, outlen, "UTF-16BE", "UTF-8");
}

size_t BCUserUTF8ToUTF16(char* in_utf8, size_t inlen, char* out_utf16, size_t outlen) {
    return convertString(in_utf8, inlen, out_utf16, outlen, "UTF-8", "UTF-16BE");
}


BCUserResult BCUserProtectHiddenPart(void *header, const char *container,
                  BCProtectExistingHiddenPartCallback protectCallback, SpaceMap *hiddenMap,
                  char */*fsType*/, SpaceMap **/*out_resizedMap*/)
{
    BCUserResult result = kBCUserCancel;
    BCContainerResult containerResult;
    //BCPassword tmpPwd;
    char ans = '\0';

    // Request user to protect all existing hidden parts.
    if (protectCallback)
    {
        printf("Warning: All unprotected hidden parts can be destroyed!\n");
        do
        {
            printf("Do you want to protect an existing hidden part? ([y]/n)");

            ans = fgetc(stdin);
            if (ans != '\n')
                fgetc(stdin); // Read and ignore Enter char.

            if ((ans != 'N') && (ans != 'n'))
            {
                containerResult = protectCallback(header, container, hiddenMap);
                if (kBCContainer_Success != containerResult)
                {
                    BCUserMessage("Failed to protect existing hidden part.\nYou shouldn't continue to create a new hidden part until all existing hidden parts will not be protected!\n", kBCUserMessage_Error);
                    continue;
                }
            }
            else
            {
                return kBCUserOk;
            }
        }
        while ((ans != 'n') || (ans != 'N'));
    }

    return result;
}

#endif // #ifndef BC_LINUX_USER_H

