/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * Common memory and buffer related types.
 */

#ifndef BC_MEMORY_H
#define BC_MEMORY_H

#include <stdlib.h>
#include <string.h>
#include <assert.h>

/**
 * Secure static buffer.
 * Used to contain secrets (passwords, keys, etc) and automatically shred data on scope exit.
 */ 
template < unsigned long size >
struct BCSecureBuffer {
	char 	data[size];
	
	BCSecureBuffer (){
		this->shred();
	}

	~BCSecureBuffer() {
		this->shred();
	}

	void shred() {
		memset(data, 0, size);
	}
};


/**
 * Auto-deallocation at the end of scope pointer.
 */
template < class T >
class BCScopedPtr {
public:

    T* ptr;

    BCScopedPtr() : ptr(NULL) {}

    explicit BCScopedPtr( T* pointee ) : ptr(pointee) {}
	
    ~BCScopedPtr() throw() {
        if(ptr) {
            // this checks that T is known at this point.
            assert(sizeof(T));
            delete ptr;
        }
    }

	T* operator -> () const {
        assert(ptr);
		return ptr;
	}

	T& operator * () const {
        assert(ptr);
		return *ptr;
	}

	T* get() const {
        return ptr;
	}

    operator bool () {
        return ptr != NULL;
    }

protected:
    
    BCScopedPtr(const BCScopedPtr<T>& rhv) {}
    BCScopedPtr<T>& operator = (const BCScopedPtr<T>& rhv) { return *this; }
};

template < class T >
bool operator == (const BCScopedPtr<T>& lhv, const BCScopedPtr<T>& rhv) {
	return lhv.get() == rhv.get();
}

#endif

