/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * User authentication session interface.
 */

#ifndef BC_USER_H
#define BC_USER_H

#include <stdlib.h>

#include "intsize.h"
#include "password.h"
#include "container.h"

// Default verify password query
// TODO: Use "const char * " type to make it visible for debugger
#define BC_DEFAULT_PASSWORD_QUERY "Please enter container password:";

// Default new password query
#define BC_DEFAULT_NEW_PASSWORD_QUERY "Please enter new container password:";

/**
 * User session return codes.
 */
typedef enum {
	kBCUserOk = 0,
    kBCUserCancel
} BCUserResult;


/**
 * User message level.
 */
typedef enum {
	kBCUserMessage_Info = 0,
	kBCUserMessage_Warning,
    kBCUserMessage_Error
} BCUserMessageLevel;


// Encoding tags for conversion
typedef enum {
    kBCUser_UTF8    = 0x08,
    kBCUser_UTF16LE = 0x06,
    kBCUser_UTF16BE = 0x60
} BCUserEncoding;


/** 
 * Callback supplied by caller to verify a password.
 * Used to verify existing or new password.
 * Each function that accepts the callback can accept NULL to not perform any verification
 *
 * @param pwd		Password string.
 * @param pwdlen	Length of the password string.
 * @param ctx		Cookie passed by the caller.
 * @return	        Error message to display if password is invalid, NULL otherwise.
 */
typedef const char* (*BCVerifyPasswordCallback) (const char* utf8_pwd, unsigned pwdlen, void* ctx);

/**
 * Callback used to decrypt and protect existing hidden part by substracting it's blocks
 * from free space got from the container.
 * @param header        Container header structure.
 * @param password      Existing hidden part password.
 * @param freeSpaceMap  Space map with extents of new hidden part.
 * @return              BCContainerResult type describing result of operation.
*/
typedef BCContainerResult (*BCProtectExistingHiddenPartCallback) (void *header, const char* file, SpaceMap *freeSpaceMap);


#ifdef __cplusplus
extern "C" {
#endif

	/**
	 * Run verify password session.
     * @param query         Password query, default query will be used if NULL
	 * @param callback	    Password verification callback that will be called each time a user tries to enter a password.
	 * @param cookie		Password verification callback cookie.
     * @param out_pwd       Password that was successfully verified.
	 * @return	            kBCUserCancel if user gave up to provide a correct password, kBCUserOk otherwise.
	 */
    BCUserResult BCUserVerifyPassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCMountParams *mountParams, BCPassword* out_pwd);
	
	
	/**
	 * Run new password session.
     * @param query         New password query, default query will be used if NULL
	 * @param callback	    Password verification callback that will be called each time a user tries to enter a password.
     * @param cookie		Password verification callback cookie.
     * @param out_pwd       Password that was successfully verified.
	 * @return	            kBCUserCancel if user gave up to provide a correct password, kBCUserOk otherwise.
	 */
	BCUserResult BCUserCreatePassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCPassword* out_pwd);
	
	
	/**
	 * Fire up the seed generation progress dialog.
	 * @param total	        Total bytes of seed value that needs to be generated.
     * @return	            Number of bytes of seed generated, or 0, if failed/cancelled.
	 */
    unsigned BCUserGenerateSeed( const char *message, char *out_buf, unsigned total,
                                 BCProgressCallback callback, void *callbackArg);

    /**
     * Change string encoding.
     * @param msg                   Source message to change encoding of.
     * @param msgSize               Source message size in bytes.
     * @param out_encodedMsg        Buffer to contain target message with changed encoding.
     * @param bufferSize            Buffer length in bytes.
     * @param fromEncoding          Source message encoding.
     * @param toEncoding            Target message encoding.
     * @return                      0, if conversion succeeded, otherwise errno of error occurred.
     */
    size_t BCUserConvertString(char* in, size_t inlen, char* out, size_t outlen, BCUserEncoding fromEncoding, BCUserEncoding toEncoding);

    /**
     * Convert UTF16 string to UTF8.
     * @param in_utf16      UTF16LE encoded input string.
     * @param inlen         Input string buffer size.
     * @param out_utf8      Buffer to hold utf8 converted string.
     * @param out_len       Output buffer size.
     * @return              Number of bytes in output buffer used for converted string or 0 on error.
     */
    size_t BCUserUTF16ToUTF8(char* in_utf16, size_t inlen, char* out_utf8, size_t outlen);


    /**
     * Same as above but reverse.
     */
    size_t BCUserUTF8ToUTF16(char* in_utf8, size_t inlen, char* out_utf16, size_t outlen);


	/**
	 * Display a message to the user.
     * @param msg       User message text.
     * @param level     User message type level.
	 */
	void BCUserMessage(const char* msg, BCUserMessageLevel level);

    /**
     * Gets information about a new hidden part. Allows an user to protect existing hidden parts
     * by substracting their extents from the hidden map.
     * @param header            Parameter to pass to protectCallback.
     * @param protectCallback   Used to protect existing hidden part by substracting it's blocks from hiddenMap.
     * @param hiddenMap         Map with extents of future hidden part.
     * @param fsType            Crunch parameter for GUI, will contain a correct file system. Must be BC_MAX_PATH length. Not used by console.
     * @param out_resizedMap    Crunch parameter for GUI, will contain a resized space map.
     */
    BCUserResult BCUserProtectHiddenPart(void *header, const char* container,
                    BCProtectExistingHiddenPartCallback protectCallback, SpaceMap *hiddenMap,
                    char *fsType, SpaceMap **out_resizedMap);
	

#ifdef __cplusplus
}
#endif


#endif
