/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * Handling of encrypted volume stored in container files
 */

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include "volume.h"
#include "intsize.h"
#include "file.h"
#include "error.h"
#include "isaac.h"
#include "cipher.h"
#include "user.h"

#include "kernel/BCKernel.h"

#define BC_VOLUME_BUFFER_SIZE (1024 * 1024) // Buffer size for randomization and reencryption
#define BC_VOLUME_SEED_SIZE (256)           // Seed size for Isaac in container body randomizer

BCContainerResult BCInitVolume(struct BCVolumeInfo* info, const char* file, int key_handle, char* pool, struct BCVolume* out_volume)
{
    memset(out_volume, 0, sizeof(*out_volume));

    out_volume->file = strdup(file);
    if(!out_volume->file) {
        return kBCContainer_NoMemory;
    }

    out_volume->key_handle = key_handle;
    memcpy(&out_volume->info, info, sizeof(BCVolumeInfo));
    memcpy(out_volume->pool, pool, sizeof(out_volume->pool));
    out_volume->info.map = SpaceMapCreateCopy(info->map);

    if (!out_volume->info.map)
    {
        free(out_volume->file);
        out_volume->file = NULL;
        return kBCContainer_NoMemory;
    }

    return kBCContainer_Success;
}


BCContainerResult BCReadVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data)
{
    bc_u64 offset = sector * 512 + vol->info.offset;
    bc_u64 iv = sector + 1;

    bc_file_handle fd;
    if(kBCFileResult_Success != BCFileOpen(vol->file, kBCFileMode_Read | kBCFileMode_Exclude, &fd)) {
        return kBCContainer_IOError;
    }

    bc_u64 nbytes = BCFileReadWithOffset(&fd, data, sizeof(data), offset);
    BCFileClose(&fd);
    if(nbytes != sizeof(data)) {
        return kBCContainer_IOError;
    }

    if(kBCKernel_Success != BCKernelProcess(vol->info.alg, vol->info.mode, kBCProcess_Decrypt, vol->key_handle, iv, vol->pool, data, sizeof(data))) {
        return kBCContainer_InvalidAlgorithm;
    }

    return kBCContainer_Success;
}


BCContainerResult BCWriteVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data)
{
    bc_u64 offset = sector * 512 + vol->info.offset;
    bc_u64 iv = sector + 1;

    if(kBCKernel_Success != BCKernelProcess(vol->info.alg, vol->info.mode, kBCProcess_Encrypt, vol->key_handle, iv, vol->pool, data, sizeof(data))) {
        return kBCContainer_InvalidAlgorithm;
    }

    bc_file_handle fd;
    if(kBCFileResult_Success != BCFileOpen(vol->file, kBCFileMode_ReadWrite | kBCFileMode_Exclude, &fd)) {
        return kBCContainer_IOError;
    }

    bc_u64 nbytes = BCFileWriteWithOffset(&fd, data, sizeof(data), offset);
    BCFileClose(&fd);
    if(nbytes != sizeof(data)) {
        return kBCContainer_IOError;
    }

    return kBCContainer_Success;
}


BCContainerResult BCRandomizeVolumeBody(struct BCVolume* vol, BCProgressCallback progress,
                                        void* progressArg)
{
    bc_file_handle fd;
    if(kBCFileResult_Success != BCFileOpen(vol->file, kBCFileMode_ReadWrite, &fd)) {
        return kBCContainer_IOError;
    }

    bc_u64 seekResult = BCFileSeek(&fd, vol->info.offset, kBCFileSeek_Begin);
    if (BC_FILE_ERR == seekResult) {
        return kBCContainer_IOError;
    }

    char seed[BC_VOLUME_SEED_SIZE];
    unsigned seedSize = BCUserGenerateSeed("Generating seed for randomizer...\n", seed, BC_VOLUME_SEED_SIZE, progress, progressArg);
    if (!seedSize) {
        BCFileClose(&fd);
        return kBCContainer_Cancelled;
    }

    BCContainerResult error = kBCContainer_Success;
    bc_u64 size = vol->info.size;
    bc_u64 offset = 0;

    char* buffer = (char*) malloc(BC_VOLUME_BUFFER_SIZE);
    memset(buffer, 0, BC_VOLUME_BUFFER_SIZE);

    if(!buffer) {
        BCFileClose(&fd);
        return kBCContainer_NoMemory;
    }

    void *isaac = IsaacInit(seed, BC_VOLUME_SEED_SIZE);

    while (offset < size) {
        bc_u64 bytesLeft = size - offset;
        unsigned nbytes = bytesLeft > BC_VOLUME_BUFFER_SIZE ?
                    BC_VOLUME_BUFFER_SIZE : (unsigned)bytesLeft;

        IsaacGenerate(isaac, buffer, nbytes);

        if(nbytes != (unsigned)BCFileWrite(&fd, buffer, nbytes)) {
            error = kBCContainer_IOError;
            break;
        }

        offset += nbytes;

        if(progress && !progress("Randomizing container", offset, size, progressArg)) {
            error = kBCContainer_Cancelled;
            break;
        }
    }

    IsaacRelease(isaac);

    sync();

    BCFileClose(&fd);
    free(buffer);

    return error;
}


BCContainerResult BCReencryptVolume(struct BCVolume* src, struct BCVolume* dst, bc_u64 start,
                                    BCProgressCallback progress, void* arg)
{
    BCCipher src_cipher;
    if(kBCKernel_Success != src_cipher.init(src->info.alg, src->info.mode)) {
        return kBCContainer_InvalidAlgorithm;
    }

    BCCipher dst_cipher;
    if(kBCKernel_Success != dst_cipher.init(dst->info.alg, dst->info.mode)) {
        return kBCContainer_InvalidAlgorithm;
    }

    bc_file_handle fd;
    if(kBCFileResult_Success != BCFileOpen(src->file, kBCFileMode_ReadWrite | kBCFileMode_Exclude, &fd)) {
        return kBCContainer_IOError;
    }

    char* buffer = (char*) malloc(BC_VOLUME_BUFFER_SIZE);
    if(!buffer) {
        BCFileClose(&fd);
        return kBCContainer_NoMemory;
    }

    bc_u64 offset = start;
    bc_u64 size = src->info.size;
    bc_u64 iv;
    BCContainerResult error = kBCContainer_Success;

    while(offset < size) {
        size_t nbytes = size - offset;
        nbytes = nbytes > BC_VOLUME_BUFFER_SIZE ? BC_VOLUME_BUFFER_SIZE : nbytes;

        iv = offset / 512 + 1;

        if(nbytes != BCFileReadWithOffset(&fd, buffer, nbytes, src->info.offset + offset)) {
            error = kBCContainer_IOError;
            break;
        }

        src_cipher.decrypt(iv, src->key_handle, src->pool, buffer, nbytes);
        
        dst_cipher.encrypt(iv, dst->key_handle, dst->pool, buffer, nbytes);

        if(nbytes != BCFileWriteWithOffset(&fd, buffer, nbytes, dst->info.offset + offset)) {
            error = kBCContainer_IOError;
            break;
        }

        sync();

        offset += nbytes;

        if(progress && !progress("Reencrypting", offset, size, arg)) {
            error = kBCContainer_Cancelled;
            break;
        }
    }

    BCFileClose(&fd);
    free(buffer);

    return error;
}

        
void BCPutVolume(struct BCVolume* vol)
{
    BCKernelFreeKey(vol->key_handle);
    free(vol->file);
    SpaceMapFree(vol->info.map);
    memset(vol, 0, sizeof(*vol));
}

