/*
 * Copyright 2012 Jetico, Inc. All rights reserved.
 * Handling of encrypted volume stored in container files
 *
 * Author: Evgeny Yakovlev, Vasily Loginov
 *
 */

#ifndef BC_VOLUME_H
#define BC_VOLUME_H

#include "bctypes.h"
#include "keygen.h"
#include "error.h"

#include <stdio.h>

/// Volume format version.
typedef enum {
    kBCVolume_V7 = 7,       // V7 compatibility format
    kBCVolume_V8 = 8,       // V8 current main format
    kBCVolume_Unknown = 0   // Unknown format version
} BCVolumeFormat;


/**
 * Structure describes encrypted volume stored in container file.
 * Each container file encodes at least one main volume and a possible number of additional hidden ones.
 */
struct BCVolume
{
    struct BCVolumeInfo     info;
    char*                   file;
    int                     key_handle;
    char                    pool[512];

    BCVolume()
    {
        file = NULL;
        key_handle = -1;
        memset(pool, 0, 512);
    }
};


typedef char bc_sector_t[512];

#ifdef __cplusplus
extern "C" {
#endif


    /**
     * Init volume structure.
     * @param info          Volume info structure
     * @param file          Backing store file name, copied
     * @param key_handle    Encryption key handle in kernel memory
     * @param pool          Random data pool (optional)
     * @param out_volume    Initialized volume structure on success.
     */
    BCContainerResult BCInitVolume(struct BCVolumeInfo* info, const char* file,
                                   int key_handle, char* pool, struct BCVolume* out_volume);


    /**
     * Read and decrypt a volume sector
     * @param vol       Volume
     * @param sector    Sector number
     * @param data      Sector data buffer to store decrypted data
     */
    BCContainerResult BCReadVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data);


    /**
     * Encrypt and write a volume sector
     * @param vol       Volume
     * @param sector    Sector number
     * @param data      Sector data buffer to encrypt and write
     */
    BCContainerResult BCWriteVolumeSector(struct BCVolume* vol, bc_u64 sector, bc_sector_t data);


    /**
     * Randomize volume data
     * @param vol       Volume
     * @param progress  Caller progress callback
     * @param arg       Progress callback argument
     */
    BCContainerResult BCRandomizeVolumeBody(struct BCVolume* vol, BCProgressCallback progress,
                                            void* progressArg);


    /**
     * Reencrypt volume data
     * @param src           Current volume configuration
     * @param dst           Reencrypted volume configuration
     * @param start         Volume offset where to start reencryption from
     * @param progress      Progress callback
     * @param arg           Progress callback argument
     */
    BCContainerResult BCReencryptVolume(struct BCVolume* src, struct BCVolume* dst, bc_u64 start,
                                        BCProgressCallback progress, void* arg);


    /**
     * Release volume resources.
     */
    void BCPutVolume(struct BCVolume* vol);

#ifdef __cplusplus
}
#endif


#endif


