/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pthread.h>
#include <limits.h>
#include <assert.h>
#include <errno.h>

#include <list>

#include "disk.h"
#include "container/exception.h"
#include "kernel/kmod/bc_ctl.h"


static int getBlockMajor(void) 
{
    char namebuf[PATH_MAX] = {0};
    snprintf(namebuf, PATH_MAX - 1, "/sys/class/misc/%s/block_major", BC_CTL_NODE_NAME);

    FILE* f = fopen(namebuf, "r");
    if(!f) {
        fprintf(stderr, "Could not open %s for reading: %s\n", namebuf, strerror(errno));
        return -1;
    }

    int block_major;
    if(1 != fscanf(f, "%d", &block_major)) {        
        fprintf(stderr, "fscanf failed: %s\n", strerror(errno));
        fclose(f);        
        return -1;
    }

    fclose(f);
    return block_major;
}


///////////////////////////////////////////////////////////////////////
//
//  Event listener
//
///////////////////////////////////////////////////////////////////////


typedef struct {
    BCDiskEventCallback callback;
    int                 mask;
    void*               arg;
} EventSubscriber;

typedef std::list<EventSubscriber>  subscriber_list_t;
typedef subscriber_list_t::iterator subscriber_iter_t;

static subscriber_list_t gSubscribers;
static pthread_t gListenerThread;


#ifdef BC_DEBUG
static void logDevice(struct udev_device* dev)
{
    dev_t devnum = udev_device_get_devnum(dev);

    printf("Got device %s:\n", udev_device_get_devnode(dev));
    printf(" subsystem: %s\n", udev_device_get_subsystem(dev));
    printf(" driver:    %s\n", udev_device_get_driver(dev));
    printf(" devtype:   %s\n", udev_device_get_devtype(dev));
    printf(" action:    %s\n", udev_device_get_action(dev));
    printf(" devnum:    %d, %d\n", major(devnum), minor(devnum));
        
    if(0 == strcmp("disk", udev_device_get_devtype(dev))) {
        printf(" backing store: %s\n", udev_device_get_sysattr_value(dev, "backing_store"));
    }

    printf("\n");
}

# define BC_LOG_DEVICE(dev) logDevice(dev)
#else
# define BC_LOG_DEVICE(dev) 
#endif


void BCRegisterForDiskEvents(BCDiskEventCallback callback, int event_mask, void* arg) 
{
    EventSubscriber subscriber = {callback, event_mask, arg};
    gSubscribers.push_back(subscriber);
}


void BCUnregisterForDiskEvents(BCDiskEventCallback callback) 
{
    subscriber_iter_t i = gSubscribers.begin();
    subscriber_iter_t end = gSubscribers.end();

    while(i != end) {
        if(i->callback == callback) {
            gSubscribers.erase(i);
            break;
        }

        ++i;
    }
}


static void broadcastEvent(struct udev_device* dev)
{
    int event = 0;
    const char* action = udev_device_get_action(dev);
    if (0 == strcmp(action, "add")) {
        event = kBCDiskAppeared;
    } else if (0 == strcmp(action, "remove")) {
        event = kBCDiskDisappeared;
    } else if (0 == strcmp(action, "change")) {
        event = kBCDiskChanged;
    } else {
        fprintf(stderr, "Unknown event \"%s\"\n", action);
        return;
    }

    BCDiskRef diskref(dev);

    subscriber_iter_t i = gSubscribers.begin();
    subscriber_iter_t end = gSubscribers.end();
    
    while(i != end) {
        if(i->mask & event) {
            i->callback(&diskref, event, i->arg);
        }

        ++i;
    }
}

// Listener thread cleanup proc when thread is cancelled
static void threadCleanup(void* arg)
{
    struct udev* udev = (struct udev*) arg;
    udev_unref(udev);
}

static void* eventListenerThread(void* /*arg*/)
{
    printf("Device listening thread started.\n");
    int block_major = getBlockMajor();
    if(-1 == block_major) {
        fprintf(stderr, "Could not read bestcrypt block device major number\n");
        return NULL;
    }

    struct udev* udev = udev_new();
    if(!udev) {
        fprintf(stderr, "Could not allocate udev\n");
        return NULL;
    }

    struct udev_monitor* mon = udev_monitor_new_from_netlink(udev, "udev");
    if(!mon) {
        udev_unref(udev);
        fprintf(stderr, "Could not create new monitor\n");
        return NULL;
    }

    udev_monitor_filter_add_match_subsystem_devtype(mon, "block", NULL);
    udev_monitor_enable_receiving(mon);
    struct udev_device* dev = NULL;

    pthread_cleanup_push(threadCleanup, udev);

    while(1) {
        dev = udev_monitor_receive_device(mon); // blocks
        if(!dev) {
            fprintf(stderr, "Could not receive device from monitor\n");
            break;
        }

        dev_t devnum = udev_device_get_devnum(dev);
        if((int)major(devnum) == block_major) {
            BC_LOG_DEVICE(dev);
            // TODO: if callback will not complete before
            // new event came, we'll miss it, that's not good.
            broadcastEvent(dev);
        }

        udev_device_unref(dev);
    }

    pthread_cleanup_pop(1);
    return NULL;
}


int BCStartDiskEventListener(void) 
{
    if(gListenerThread) {
        return 0;
    }

    pthread_attr_t attr;
    pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	pthread_create(&gListenerThread, &attr, &eventListenerThread, NULL);
	pthread_attr_destroy(&attr);

    return 1;
}


void BCStopDiskEventListener(void) 
{
    if(!gListenerThread) {
        return;
    }

    pthread_cancel(gListenerThread);
    pthread_join(gListenerThread, NULL);

    gListenerThread = 0; 
}


///////////////////////////////////////////////////////////////////////
//
//  Disk ref
//
///////////////////////////////////////////////////////////////////////


// Find device using a predicate function
static bool forEachDevice(bool (*predicate) (struct udev_device* dev, void* arg), void* arg, BCDiskRef *out_disk)
{
    assert(out_disk);
    int major = getBlockMajor();
    if(-1 == major) {
        return false;
    }

    struct udev* udev = udev_new();
    if(!udev) {
        return false;
    }

    struct udev_enumerate* enumeration = udev_enumerate_new(udev);
    if(!enumeration) {
        udev_unref(udev);
        return false;
    }

    udev_enumerate_add_match_subsystem(enumeration, "block");
    udev_enumerate_scan_devices(enumeration);


    struct udev_list_entry* devices = udev_enumerate_get_list_entry(enumeration);
    struct udev_list_entry* entry;
    udev_list_entry_foreach(entry, devices) {
        const char* path = udev_list_entry_get_name(entry);
		struct udev_device* dev = udev_device_new_from_syspath(udev, path);
        if(!dev) {
            break;
        }
        dev_t devnum = udev_device_get_devnum(dev);

        if((int) major(devnum) == major) {
            if(predicate(dev, arg)) {
                //(void) udev_device_get_devnode(dev);

                // Load data before library context is unloaded.
                out_disk->loadData(dev);
                udev_device_unref(dev);
                udev_enumerate_unref(enumeration);
                udev_unref(udev);
                return true;
            }
        }

        udev_device_unref(dev);
    }

    // Nothing was found or error.
    udev_enumerate_unref(enumeration);
    udev_unref(udev);

    return false;
}


// Predicate for searching device by its minor number
static bool isMinorPredicate(struct udev_device* dev, void* arg)
{
    int minor = *(int*)arg;
    dev_t devnum = udev_device_get_devnum(dev);
    return (minor == (int)minor(devnum));
}


// Predicate for searching device by its node name
static bool isNodeNamePredicate(struct udev_device* dev, void* arg)
{
    const char* nodeName = (const char*) arg;
    return (0 == strcmp(nodeName, udev_device_get_devnode(dev)));
}


// Predicate for searching device by its backing store name
static bool isBackingStorePredicate(struct udev_device* dev, void* arg)
{
    const char* storeName = (const char*) arg;
    return (0 == strcmp(storeName, udev_device_get_sysattr_value(dev, "backing_store")));
}


// Not really a predicate, just copies device reference and returns false
static bool copyDevice(struct udev_device* dev, void* arg)
{
    std::vector<BCDiskRef>* list = (std::vector<BCDiskRef>*) arg;
    list->push_back(BCDiskRef(dev));
    return false;
}


// static 
bool BCDiskRef::findDiskWithMinor(int minor, BCDiskRef* out_ref)
{
    return forEachDevice(isMinorPredicate, &minor, out_ref);
}


// static
bool BCDiskRef::findDiskWithDeviceName(const char* devName, BCDiskRef* out_ref)
{
    return forEachDevice(isNodeNamePredicate, (void*)devName, out_ref);
}


// static 
bool BCDiskRef::findDiskWithBackingStore(const char* backingStore, BCDiskRef* out_ref)
{
    return forEachDevice(isBackingStorePredicate, (void*)backingStore, out_ref);
}


// static 
void BCDiskRef::listDisks(std::vector<BCDiskRef> * out_list)
{
    assert(out_list);
    out_list->clear();
    BCDiskRef fakeRef;
    (void) forEachDevice(copyDevice, out_list, &fakeRef);
}


// static
int BCDiskRef::blockMajor() {
    return getBlockMajor();
}


BCDiskRef::BCDiskRef() {
    this->setNull();
}


BCDiskRef::BCDiskRef(struct udev_device* ref) {
    this->setNull();
    this->loadData(ref);
}


BCDiskRef::BCDiskRef(BCDiskRef const & disk) {
    this->setNull();
    this->internalInit(disk.nodeName_, disk.backingStore_, disk.minor_);
}


BCDiskRef& BCDiskRef::operator = (BCDiskRef const & rhv) {
    this->internalInit(rhv.nodeName_, rhv.backingStore_, rhv.minor_);
    return *this;
}


void BCDiskRef::loadData(struct udev_device* ref) {
    (void) udev_device_get_devnode(ref);
    const char* node = udev_device_get_devnode(ref);
    const char* store = udev_device_get_sysattr_value(ref, "backing_store");
    dev_t devnum = udev_device_get_devnum(ref);
    this->internalInit(node, store, devnum);
}


void BCDiskRef::freeData()
{
    if (this->nodeName_)
    {
        delete this->nodeName_;
        this->nodeName_ = NULL;
    }
    if (this->backingStore_)
    {
        delete this->backingStore_;
        this->backingStore_ = NULL;
    }
}


void BCDiskRef::internalInit(const char *node, const char *store, int devnum)
{
    this->freeData();

    int nodeNameLen = strlen(node) + 1;
    this->nodeName_ = new char[nodeNameLen];
    strncpy(this->nodeName_, node, nodeNameLen);

    if (store)
    {
        int storeNameLen = strlen(store) + 1;
        this->backingStore_ = new char[storeNameLen];
        strncpy(this->backingStore_, store, storeNameLen);
    }

    this->minor_ = devnum;
}


void BCDiskRef::setNull()
{
    this->backingStore_ = NULL;
    this->nodeName_ = NULL;
    this->minor_ = -1;
}


BCDiskRef::~BCDiskRef()
{
    this->freeData();
}
