/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * BestCrypt virtual disk interface
 */

#ifndef BC_DISK_H
#define BC_DISK_H

#include <vector>
#ifdef __cplusplus
extern "C" {
#endif
#include <libudev.h>
#ifdef __cplusplus
}
#endif

/**
 * Reference to BestCrypt virtual disk.
 */
class BCDiskRef {
private:
    char    *nodeName_;
    char    *backingStore_;
    int     minor_;

    void freeData();
    void internalInit(const char *node, const char *store, int devnum);
    void setNull();

public:
    // Find and init a disk ref with a given minor number
    static bool findDiskWithMinor(int minor, BCDiskRef* out_ref);
    static bool findDiskWithDeviceName(const char* dev, BCDiskRef* out_ref);
    static bool findDiskWithBackingStore(const char* backingStore, BCDiskRef* out_ref);
    static void listDisks(std::vector<BCDiskRef> * out_list);

    // return common major number for bestcrypt devices
    static int blockMajor();

    // Create a blank disk reference
    BCDiskRef();

    // Create new disk with udev ref
    BCDiskRef(struct udev_device* ref);

    // Copy disk ref.
    BCDiskRef(BCDiskRef const & disk);

    // Assign disk ref
    BCDiskRef& operator = (BCDiskRef const & rhv);

    // Drop disk ref
    ~BCDiskRef();

    // Loads all necessary data from specified udev reference.
    void loadData(struct udev_device* ref);

    // Get disk minor number
    int minor() const
    {
        return this->minor_;
    }

    // Get disk bsd device node name
    const char* nodeName() const
    {
        return this->nodeName_;
    }

    // Get disk backing store (container file path)
    const char* backingStore() const
    {
        return this->backingStore_;
    }

};


// Disk event types
enum {
    kBCDiskAppeared     = 0x01,     // New disk has appeared
    kBCDiskDisappeared  = 0x02,     // Old disk has been removed
    kBCDiskChanged      = 0x04,     // Disk has changed (most likely size or mount point)

    kBCDiskAllEvents    = kBCDiskAppeared | kBCDiskDisappeared | kBCDiskChanged
};


/**
 * Callback for disk event notification.
 * @param disk        Reference to a disk in question. Will be destroyed once callback exits, make a copy if you need to extend the lifetime
 * @param event       Event type
 * @param arg         Caller argument.
 */
typedef void (*BCDiskEventCallback) (BCDiskRef* disk, int event, void* arg);


/**
 * Register a callback for disk events.
 * @param callback    Callback function.
 * @param event_mask  Mask of events caller is interested in
 * @param arg         Caller callback argument.
 */
void BCRegisterForDiskEvents(BCDiskEventCallback callback, int event_mask, void* arg);


/**
 * Remove event callback.
 */
void BCUnregisterForDiskEvents(BCDiskEventCallback callback);


/**
 * Start listening for disk events.
 * Dispatches a new thread to listen to new disk events.
 */
int BCStartDiskEventListener(void);


/**
 * Stop listening for disk events.
 */
void BCStopDiskEventListener(void);


#endif

