/*
 *  exfat.c
 *  Copyright 2011 Jetico, Inc. All rights reserved.
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <inttypes.h>
#include <assert.h>

#include "exfat.h"
#include "fslib.h"
#include "utf.h"

#define ROUND_UP(val, block) ( ((val) + (block) - 1) & ~((block) - 1) )


static inline exfat_bootsect_t* bootsect(struct fs_volume* vol) {
    return &((exfat_superblock_t*) vol->info)->bs;
}

static inline size_t ssize(exfat_bootsect_t* bs) {
	return (1 << bs->ssize_bit);
}

static inline size_t csize(exfat_bootsect_t* bs) {
	return (1 << bs->csize_bit);
}

static inline size_t bsize(exfat_bootsect_t* bs) {
    return csize(bs) * ssize(bs);
}

// mapped data cluster index -> absolute partition offset
static inline off_t c2off(exfat_bootsect_t* bs, uint32_t cluster) {
    return (cluster * csize(bs) + bs->clustsector) * ssize(bs);
}


// One directional fat cluster chain iterator
typedef struct {
    uint32_t*   fat;
    uint32_t    cluster;
} fat_iter_t;


static int exfat_init_iterator(struct fs_volume* volume, fat_iter_t* iter, uint32_t cluster)
{
    exfat_bootsect_t* bs = bootsect(volume);
    
    assert(cluster != EXFAT_CLUSTER_FREE);
    assert(cluster < bs->numclusters);
    
    memset(iter, 0, sizeof(*iter));
    size_t fatsize = bs->numfatsectors * ssize(bs);
    
    char* data = (char*) valloc(fatsize);
    if(!data) {
        fprintf(stderr, "ExFAT: Failed allocating fat table size %lu\n", (unsigned long)fatsize);
		return 0;
    }
    
    if(fatsize != fs_read_data(volume, data, fatsize, bs->numfatsectors * ssize(bs))) {
		fprintf(stderr, "ExFAT: Failed reading fat table\n");
        return 0;
    }
    
    iter->fat = (uint32_t*) data;
    iter->cluster = cluster;
    
    return 1;
}


static void exfat_free_iterator(fat_iter_t* iter)
{
    free(iter->fat);
}


// Return next cluster in chain. 
// EXFAT_CLUSTER_END means chain has ended and further
static uint32_t exfat_next_cluster(fat_iter_t* iter)
{
    assert(iter->cluster != EXFAT_CLUSTER_END);
    assert(iter->cluster != EXFAT_CLUSTER_FREE);
    
    iter->cluster = iter->fat[iter->cluster];
    return iter->cluster;
}


static const char* entry_name(struct exfat_entry* entry)
{
    switch (entry->type) {
        case EXFAT_ENTRY_BITMAP:    return "BITMAP";
        case EXFAT_ENTRY_LABEL:     return "LABEL";
        case EXFAT_ENTRY_FILE:      return "FILE";
        case EXFAT_ENTRY_FILE_INFO: return "FILE INFO";
        case EXFAT_ENTRY_FILE_NAME: return "FILE NAME";
        case EXFAT_ENTRY_UPCASE:    return "UPCASE";
        default:                    return "UNKNOWN!";
    }
}


//static int entry_is_valid(struct exfat_entry* entry)
//{
//    switch (entry->type) {
//        case EXFAT_ENTRY_BITMAP:
//        case EXFAT_ENTRY_LABEL:
//        case EXFAT_ENTRY_FILE:
//        case EXFAT_ENTRY_FILE_INFO:
//        case EXFAT_ENTRY_FILE_NAME:
//        case EXFAT_ENTRY_UPCASE:    return 1;
//        default:                    return 0;
//    }
//}


static void display_bitmap_entry(struct exfat_entry* entry) 
{
    assert(entry->type == EXFAT_ENTRY_BITMAP);
    struct exfat_entry_bitmap* bitmap = (struct exfat_entry_bitmap*) entry->data;
    fprintf(stderr, "bitmap size %llu, start cluster %u\n",
            (unsigned long long)bitmap->size, bitmap->start_cluster);
}


static void display_name_entry(struct exfat_entry* entry)
{
    assert(entry->type == EXFAT_ENTRY_FILE_NAME);
    struct exfat_entry_name* name = (struct exfat_entry_name*) entry->data;
    
    char utf8_name[0xff] = {0};
    utf16_to_utf8(name->name, sizeof(name->name), utf8_name, sizeof(utf8_name));
    
    fprintf(stderr, "file name \"%s\"\n", utf8_name);
}


//static void entry_display(struct exfat_entry* entry)
//{
//    switch (entry->type) {
//        case EXFAT_ENTRY_BITMAP:    display_bitmap_entry(entry); break;
//        case EXFAT_ENTRY_LABEL:     break;
//        case EXFAT_ENTRY_FILE:      break;
//        case EXFAT_ENTRY_FILE_INFO: break;
//        case EXFAT_ENTRY_FILE_NAME: display_name_entry(entry); break;
//        case EXFAT_ENTRY_UPCASE:    break;
//        default:                    break;
//    }
//}


static void exfat_display_volume(fs_volume_t* volume)
{
	exfat_superblock_t* super = (exfat_superblock_t*) volume->info;
	exfat_bootsect_t* bs = &super->bs;
	
	printf("OEMID: %s\n", bs->oemname);
    printf("Sector size (bytes): %lu\n", (unsigned long)ssize(bs));
    printf("Cluster size (sectors): %lu\n", (unsigned long)csize(bs));
    printf("Total sectors: %llu\n", (unsigned long long)bs->numsectors);
    printf("Total mapped clusters: %u (%lu sectors)\n", bs->numclusters,
           (unsigned long)(bs->numclusters * csize(bs)));
	printf("Reserved: 0 - %d\n", bs->fatsector - 1);
	printf("FAT Tables: %d - %d\n", bs->fatsector, bs->fatsector + bs->numfatsectors * bs->numfats - 1);
	
	unsigned i;
	for(i = 0; i < bs->numfats; ++i) {
		uint32_t cur_fat_off = bs->fatsector + (i * bs->numfatsectors);
		printf("-> FAT %d: %d - %d\n", i + 1, cur_fat_off, cur_fat_off + bs->numfatsectors - 1);
	}
	
    printf("Data region: %d - %lu\n", bs->clustsector, (unsigned long)(bs->numclusters - 1) * csize(bs) + bs->clustsector);
    printf("-> Root dirent: %lu (data cluster %u)\n",
           (unsigned long)(bs->rootcluster * csize(bs) + bs->clustsector), bs->rootcluster);
}


static fs_extent_t* exfat_parse_bitmap(struct fs_volume* volume, struct exfat_entry* entry) 
{
    assert(entry->type == EXFAT_ENTRY_BITMAP);

    exfat_bootsect_t* bs = bootsect(volume);
    struct exfat_entry_bitmap* bitmap = (struct exfat_entry_bitmap*) entry->data;
    
    char* data = fs_get_data(volume, c2off(bs, bitmap->start_cluster - EXFAT_FIRST_CLUSTER), ROUND_UP(bitmap->size, ssize(bs)));
    char* fat = data;
    if(!fat) {
        fprintf(stderr, "exfat: failed to allocate bitmap data size %llu at %llu\n",
                (unsigned long long)bitmap->size, (unsigned long long)c2off(bs, bitmap->start_cluster) / 512);
        return NULL;
    }
        
    unsigned size = csize(bs) * ssize(bs);

    fs_extent_t* freespace = NULL;
    fs_extent_t* cur_ext = NULL;
    uint32_t cluster = 0;
    int in_region = 0;

    for(cluster = 0; cluster < bs->numclusters; ) {
        char bitset = *fat++;
        
        int i;
        for(i = 0; (i < 8) && (cluster < bs->numclusters); ++i, ++cluster) {
            
            if(bitset & (1 << i)) {
                // free space region ended (if ever began)
                in_region = 0; 
                continue;
            } 
            
            if(!in_region) {
                // new free space region started
                fs_extent_t* ext = (fs_extent_t*) malloc(sizeof(fs_extent_t));
                if(!ext) {
                    fprintf(stderr, "ExFAT: Failed to allocate new free space extent\n");
                    fs_put_data(data);
                    fs_put_extent_list(freespace);
                    return NULL;
                }
                    
                ext->start = c2off(bs, cluster);
                ext->size = size;
                ext->next = NULL; // was missing!!! what a shame!!!
                
                if(!cur_ext) {
                    freespace = cur_ext = ext;
                } else {
                    fprintf(stderr, "ExFAT: Completed free space region %llu:%llu\n", cur_ext->start, cur_ext->size);
                    cur_ext->next = ext;
                    cur_ext = ext;
                }
                
                in_region = 1;
            } else {
                // continuing inside the region
                cur_ext->size += size;
            }
        }
    }
    
    assert(cluster == bs->numclusters);
    
    fs_put_data(data);
    return freespace;
}


// Read bitmap and parse free space extents from it
static struct fs_extent* exfat_get_free_space(struct fs_volume* volume)
{
	exfat_bootsect_t* bs = bootsect(volume);
	fs_extent_t* freespace = NULL;
    
    size_t size = csize(bs) * ssize(bs);
    uint32_t cluster = bs->rootcluster - EXFAT_FIRST_CLUSTER;

    char* cluster_data = (char*) valloc(size);
    if(!cluster_data) {
        fprintf(stderr, "ExFAT: Could not allocate cluster data buffer size %lu\n",
                (unsigned long)size);
        return NULL;
    }
    
    // Bitmap entry is stored in root dirent. Read it, find bitmap entry and read it's data.
    fat_iter_t iter;
    if(!exfat_init_iterator(volume, &iter, cluster)) {
        fprintf(stderr, "ExFAT: Could not read root entry\n");
        goto OUT;
    }
        
    do {
        // read current root dirent cluster and try to find bitmap entry
        if(size != fs_read_data(volume, cluster_data, c2off(bs, cluster), size)) {
            fprintf(stderr, "ExFAT: Failed reading root dirent cluster %u\n", cluster);
            goto OUT;
        }
        
        struct exfat_entry* entry = (struct exfat_entry*) cluster_data;
        while(entry->type != EXFAT_ENTRY_EOD) {
            if(entry->type == EXFAT_ENTRY_BITMAP) {
                fprintf(stderr, "ExFAT: bitmap entry found in cluster %u\n", cluster);
                freespace = exfat_parse_bitmap(volume, entry);
                goto OUT;
            }
            
            ++entry;
        }
    } while (EXFAT_CLUSTER_END != (cluster = exfat_next_cluster(&iter)));

    
    fprintf(stderr, "ExFAT: Bitmap entry not found!\n");
    
OUT:
    if(cluster_data) free(cluster_data);
    if(iter.fat) exfat_free_iterator(&iter);
    
    return freespace;
}


/*
static uint16_t calc_name_hash(const struct exfat* ef, const le16_t* name)
{
	size_t i;
	size_t length = utf16_length(name);
	uint16_t hash = 0;
    
	for (i = 0; i < length; i++)
	{
		uint16_t c = le16_to_cpu(name[i]);
        
		// convert to upper case 
		if (c < ef->upcase_chars)
			c = le16_to_cpu(ef->upcase[c]);
        
		hash = ((hash << 15) | (hash >> 1)) + (c & 0xff);
		hash = ((hash << 15) | (hash >> 1)) + (c >> 8);
	}
	return cpu_to_le16(hash);
}



typedef struct {
    char name[PATH_MAX];
} path_component_t;


static const char* next_path_component(const char* path, path_component_t* out_comp)
{
    memset(out_comp->name, 0, sizeof(out_comp->name));
    
    const char* begin = strstr(path, "/");
    if(begin) {
        ++begin;
        const char* end = strstr(begin, "/");
        unsigned long len = (end ? end - begin : PATH_MAX);
        strncpy(out_comp->name, begin, len);
    }
    
    return begin;
}


// Lookup a path component in parent directory
static int exfat_lookup(struct fs_volume* vol, fat_iter_t* iter, const char* path_component, struct exfat_entry* out_entry)
{
    
}


// Start from root dirent and scan for components until we hit our file or die trying
static int exfat_get_file_by_name(struct fs_volume* volume, const char* fname, fs_file_t* out_file)
{
    exfat_bootsect_t* bs = bootsect(volume);
    
    size_t size = csize(bs) * ssize(bs);
    uint32_t cluster = bs->rootcluster - EXFAT_FIRST_CLUSTER;
    
    char* cluster_data = (char*) valloc(size);
    if(!cluster_data) {
        fprintf(stderr, "ExFAT: Could not allocate cluster data buffer size %lu\n", size);
        goto OUT;
    }
    
    fat_iter_t iter;
    if(!exfat_init_iterator(volume, &iter, cluster)) {
        fprintf(stderr, "ExFAT: Could not read root entry\n");
        goto OUT;
    }

    char* root = strstr(fname, "/");
    if(!root) {
        fprintf(stderr, "ExFAT: Invalid file path\n");
        goto OUT;
    }
    
    struct exfat_entry_file_info* file_info = NULL; // will store found info here on success.
    
    path_component_t component;
    while (NULL != (fname = next_path_component(fname, &component))) {
        unsigned continuations = 0; // parts of the file meta data to read in separate entries.
        
        // read parts of the current dirent from disk
        do {
            if(size != fs_read_data(volume, cluster_data, c2off(bs, cluster), size)) {
                fprintf(stderr, "ExFAT: Failed reading root dirent cluster %u\n", cluster);
                goto OUT;
            }
            
            // find the file entry for the current component in current cluster
            struct exfat_entry* entry = (struct exfat_entry*) cluster_data;
            while(entry->type != EXFAT_ENTRY_EOD) {
                
                switch (entry->type) {
                    case EXFAT_ENTRY_FILE:          // file entry tells us how much parts of the info and name entries are to follow.
                        struct exfat_entry_file* file = (struct exfat_entry_file*) entry->data;
                        continuations = file->continuations;
                        break;
                        
                    case EXFAT_ENTRY_FILE_INFO:     // file info entry tells us the file location and name hash 
                        if(!continuations) {
                            fprintf(stderr, "ExFAT: Invalid continuations for a file entry!\n");
                            goto OUT;
                        }
                        
                        --continuations;
                        
                        struct exfat_entry_file_info* info = (struct exfat_entry_file_info*) entry->data;
                        if(info->name_hash != name_hash) {
                            continue;
                        }
                        
                        break;
                        
                    case EXFAT_ENTRY_FILE_NAME:     // accumulate all file name parts for final comparison if hash matched
                        if(!continuations) {
                            fprintf(stderr, "ExFAT: Invalid continuations for a file entry!\n");
                            goto OUT;
                        }
                        
                        --continuations;
                        break;
                        
                    default:
                        break;
                }
                
                ++entry;
            } // while
        } while (EXFAT_CLUSTER_END != (cluster = exfat_next_cluster(&iter))); // do
    } // while
    
    
OUT:
    if(cluster_data) free(cluster_data);
    if(iter.fat) exfat_free_iterator(&iter);
    
    return 0;
}
*/


static void exfat_free_volume(struct fs_volume* vol)
{
	free(vol->info);
}


int exfat_init_volume(struct fs_volume* vol)
{
	exfat_bootsect_t bs;
	memset(&bs, 0, sizeof(bs));
	
	if(sizeof(bs) != fs_read_data(vol, (char*)&bs, 0, sizeof(bs))) {
		FSLIB_DEBUG_LOG("ExFAT: failed reading bootsector\n");
		return EIO;
	}
	
	if(0 != strncmp(bs.oemname, EXFAT_OEM_STR, strlen(bs.oemname))) {
		FSLIB_DEBUG_LOG("ExFAT: invalid OEM name\n");
		return EINVAL;
	}
	
	if(bs.magic != EXFAT_BOOTSIG) {
		FSLIB_DEBUG_LOG("ExFAT: invalid bootsector signature\n");
		return EINVAL;
	}
	
	// volume ok.
	exfat_superblock_t* super = (exfat_superblock_t*) malloc(sizeof(exfat_superblock_t));
	if(!super) {
		FSLIB_DEBUG_LOG("ExFAT: no memory to allocate superblock\n");
		return ENOMEM;
	}
	
	memcpy(&super->bs, &bs, sizeof(bs));
    
	vol->offset             = 0;
	vol->block_size         = ssize(&bs) * csize(&bs);
	vol->display            = exfat_display_volume;
	vol->info               = super;
	vol->get_free_space     = exfat_get_free_space;
	vol->get_file_by_id     = NULL;
	vol->get_file_by_name	= NULL;//exfat_get_file_by_name;
	vol->cleanup            = exfat_free_volume;
    
	return 0;
}



