/*
 *  exfat_format.h
 *  Copyright 2011 Jetico, Inc. All rights reserved.
 *
 *  Definition of on-disk structure of exfat volume
 */

#ifndef FSLIB_EXFAT_FORMAT_H
#define FSLIB_EXFAT_FORMAT_H

#include <stdint.h>

#define EXFAT_OEM_STR           "EXFAT   "
#define EXFAT_BOOTSIG           0xAA55

// FAT entry values
#define EXFAT_FIRST_CLUSTER     2           /* first valid cluster number in FAT */
#define EXFAT_CLUSTER_FREE      0           /* free cluster */
#define EXFAT_CLUSTER_BAD       0xfffffff7	/* cluster contains bad sector */
#define EXFAT_CLUSTER_END       0xffffffff	/* final cluster of file or directory */

#pragma pack(push, 1)


/**
 * First sector on exfat volume.
 */
typedef struct exfat_bootsect
{
        uint8_t		jmpcode[3];         // jmp bytecode
        char		oemname[8];         // FS creator signature
        uint8_t		_reserved[53];      // Reserved area set to 0
        uint64_t	sector;             // First partition sector
        uint64_t	numsectors;         // Total partition sectors
        uint32_t	fatsector;          // First FAT table sector
        uint32_t	numfatsectors;      // Total FAT table sectors
        uint32_t	clustsector;        // First mapped cluster area sector
        uint32_t	numclusters;        // Total mapped clusters
        uint32_t	rootcluster;        // First _cluster_ of the root dirent
        uint32_t	serial;             // Serial number
        uint8_t		minor;              // Minor fs version
        uint8_t		major;              // Major fs version
        uint16_t	stateflags;         // Volume status flags
        uint8_t		ssize_bit;          // Sector size as (1 << n)
        uint8_t		csize_bit;          // Cluster size in sectors as (1 << n)
        uint8_t		numfats;            // Total FAT tables
        uint8_t		drive_no;           // Drive number, 0x80
        uint8_t		allocated_percent;	// Percentage of allocated space (bytes? clusters? sectors?)
        uint8_t		bootcode[397];      // Space reserved for boot code
        uint16_t	magic;              // 0xaa55
} exfat_bootsect_t;


#define EXFAT_ENTRY_VALID       0x80
#define EXFAT_ENTRY_CONTINUED	0x40
#define EXFAT_ENTRY_EOD         0x00
#define EXFAT_ENTRY_BITMAP      (0x01 | EXFAT_ENTRY_VALID)
#define EXFAT_ENTRY_UPCASE      (0x02 | EXFAT_ENTRY_VALID)
#define EXFAT_ENTRY_LABEL       (0x03 | EXFAT_ENTRY_VALID)
#define EXFAT_ENTRY_FILE        (0x05 | EXFAT_ENTRY_VALID)
#define EXFAT_ENTRY_FILE_INFO	(0x00 | EXFAT_ENTRY_VALID | EXFAT_ENTRY_CONTINUED)
#define EXFAT_ENTRY_FILE_NAME	(0x01 | EXFAT_ENTRY_VALID | EXFAT_ENTRY_CONTINUED)

#define EXFAT_ENAME_MAX         15

#define EXFAT_ATTRIB_RO         0x01
#define EXFAT_ATTRIB_HIDDEN     0x02
#define EXFAT_ATTRIB_SYSTEM     0x04
#define EXFAT_ATTRIB_VOLUME     0x08
#define EXFAT_ATTRIB_DIR        0x10
#define EXFAT_ATTRIB_ARCH       0x20


// Generic container for all entries
struct exfat_entry					
{
	uint8_t type;					// EXFAT_ENTRY_XXX
	uint8_t data[31];               // entry data
};


// EXFAT_ENTRY_BITMAP
// Data clusters bitmap. Use this for correct free space info instead of the actual FAT (!)
struct exfat_entry_bitmap
{
	uint8_t     _unknown1[19];
	uint32_t    start_cluster;
	uint64_t    size;               // in bytes
};


// EXFAT_ENTRY_UPCASE
// upper case translation table
struct exfat_entry_upcase
{
	uint8_t     __unknown1[3];
	uint32_t    checksum;
	uint8_t     __unknown2[12];
	uint32_t    start_cluster;
	uint64_t    size;
};


// EXFAT_ENTRY_LABEL
// Volume label
struct exfat_entry_label
{
	uint8_t     length;					// number of characters
	uint16_t    name[EXFAT_ENAME_MAX];	// in UTF-16LE
};


// EXFAT_ENTRY_FILE_NAME
// file or directory _name_
struct exfat_entry_name		
{
	uint8_t     __unknown;
	uint16_t    name[EXFAT_ENAME_MAX];	/* in UTF-16LE */
};


// EXFAT_ENTRY_FILE
// file or directory entry, first part
struct exfat_entry_file
{
	uint8_t     continuations;
	uint16_t    checksum;
	uint16_t    attrib;					/* combination of EXFAT_ATTRIB_xxx */
	uint16_t    __unknown1;
	uint16_t    crtime, crdate;			/* creation date and time */
	uint16_t    mtime, mdate;			/* latest modification date and time */
	uint16_t    atime, adate;			/* latest access date and time */
	uint8_t     crtime_cs;				/* creation time in cs (centiseconds) */
	uint8_t     mtime_cs;				/* latest modification time in cs */
	uint8_t     __unknown2[10];
};


#define EXFAT_FLAG_ALWAYS1		(1u << 0)
#define EXFAT_FLAG_CONTIGUOUS	(1u << 1)


// EXFAT_ENTRY_FILE_INFO
// file or directory entry, second part
struct exfat_entry_file_info
{
	uint8_t     flags;		
	uint8_t     __unknown1;
	uint8_t     name_length;
	uint16_t    name_hash;
	uint16_t    __unknown2;
	uint64_t    real_size;				/* in bytes, equals to size */
	uint8_t     __unknown3[4];
	uint32_t    start_cluster;
	uint64_t    size;					/* in bytes, equals to real_size */
};




#pragma pack(pop)

#endif

