/*
 *  Copyright 2011 Jetico, Inc. All rights reserved.
 *
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "fat.h"
#include "fat_format.h"
#include "fslib.h"


// Return sector size in bytes.
static inline unsigned sect_size(fat_bootsect_t* bs) {
	return bs->ssize;
}

// Return cluster size in sectors.
static inline unsigned cluster_size(fat_bootsect_t* bs) {
	return bs->csize;
}

// Get volume size in sectors.
// This returns absolute volume size including reserved sectors for FAT structures.
static inline uint32_t total_sectors(fat_bootsect_t* bs) {
    return (bs->sectors16 ? bs->sectors16 : bs->sectors32);
}

static inline unsigned fat_size(fat_bootsect_t* bs);

// Get FAT table first sector
static inline unsigned fat_offset(fat_bootsect_t* bs) {
//    uint32_t rootdirsectors = (bs->numroot * 32 + bs->ssize - 1);
//    uint32_t fatsize = bs->sectperfat16 ? bs->sectperfat16 : bs->ext.f32.sectperfat32;
//    uint32_t totalsectors = total_sectors(bs);
//    uint32_t datasector = totalsectors - (bs->numreserved + (bs->numfat * fatsize) + rootdirsectors);
//    return datasector - fat_size(bs) * bs->numfat; // First of the fat tables.
    return bs->numreserved;
}

// Returns 1 if volume is in fat32 format, 0 for fat16/12.
static inline int is_fat32(fat_bootsect_t* bs) {
    uint32_t rootdirsectors =((bs->numroot * 32) + (bs->ssize - 1)) / bs->ssize;
    uint32_t fatsize = bs->sectperfat16 ? bs->sectperfat16 : bs->ext.f32.sectperfat32;
    uint32_t totalsectors = total_sectors(bs);
    uint32_t datasector = totalsectors - (bs->numreserved + (bs->numfat * fatsize) + rootdirsectors);
//    uint32_t datasector = fat_offset(bs);
    uint32_t countofsectors = datasector / bs->csize;
    return countofsectors >= 65525;
    //return (bs->ext.f32.extsig == FAT_EXTBLOCK_SIG1);
}

// Get FAT table total sectors (for one table among a number of duplicates).
static inline unsigned fat_size(fat_bootsect_t* bs) {
	return (is_fat32(bs) ? bs->ext.f32.sectperfat32 : bs->sectperfat16);
}

// Get data region first sector
static inline unsigned data_offset(fat_bootsect_t* bs) {
	return fat_offset(bs) + fat_size(bs) * bs->numfat;
}




/*====================================================================
 *
 *	FAT Volume init and release.
 *
 *====================================================================*/


int fat_init_volume(fs_volume_t *vol)
{
    // read volume boot sector. It is always located in the first sector on FAT volumes.
	fat_bootsect_t bs;
	memset(&bs, 0, sizeof(bs));
	
	if(sizeof(bs) != fs_read_data(vol, (void*)&bs, 0, sizeof(fat_bootsect_t))) {
		FSLIB_DEBUG_LOG("FAT: Failed reading volume boot sector\n");
		return EINVAL;
	}
	
	if(FAT_BOOTSECT_MAGIC != bs.magic) {
		FSLIB_DEBUG_LOG("FAT: Invalid boot sector signature\n");
		return EINVAL;
	}


	// Volume ok, allocate and init superblock.
	fat_superblock_t* sb = (fat_superblock_t*) malloc(sizeof(fat_superblock_t));
	if(!sb) {
		FSLIB_DEBUG_LOG("FAT: Failed to allocate superblock structure\n");
		return ENOMEM;
	}
	
	memcpy(&sb->bs, &bs, sizeof(bs));
    if (!bs.ssize) {
        FSLIB_DEBUG_LOG("FAT: sector size cannot be zero.\n");
        return EINVAL;
    }
	
	// Fill out superblock convinience values.
	if(is_fat32(&bs)) {
//		if(0 != strncmp(bs.ext.f32.fs_type, FAT32_TYPE, strlen(FAT32_TYPE))) {
//			free(sb);
//			FSLIB_DEBUG_LOG("FAT: Invalid fat type\n");
//			return EINVAL;
//		}
		
		sb->fat_mask = FAT32_MASK;
		sb->fat_skip = FAT32_SKIP_SIZE;
		
		// On FAT32 root directory is not reserved and is stored in a data region mapped by table entries.
		sb->data_offset = bs.numreserved + (bs.ext.f32.sectperfat32 * bs.numfat);
		
	} else {
		if(0 == strncmp(bs.ext.f16.fs_type, FAT16_TYPE, strlen(FAT16_TYPE))) {
			sb->fat_mask = FAT16_MASK;
			sb->fat_skip = FAT16_SKIP_SIZE;
		} else if(0 == strncmp(bs.ext.f16.fs_type, FAT12_TYPE, strlen(FAT16_TYPE))) {
			sb->fat_mask = FAT12_MASK;
			sb->fat_skip = FAT12_SKIP_SIZE;
		} else {
			free(sb);
			FSLIB_DEBUG_LOG("FAT: Invalid fat type\n");
			return EINVAL;
		}
		
		// On FAT12/16 root directory is reserved and we should skip it to see where data region mapped by FAT table starts
		uint32_t rootsec = bs.numreserved + (bs.sectperfat16 * bs.numfat);
		uint32_t rootsize = ((bs.numroot * sizeof(fat_dirent_t)) + bs.ssize - 1) / bs.ssize; /* in sectors */
		sb->data_offset = rootsec + rootsize;
	}
	
	sb->maxdataclust = (total_sectors(&bs) - sb->data_offset) / bs.csize;

	// Finalize volume structure with our implementation.
    vol->offset             = 0; //sb->data_offset;
	vol->block_size         = bs.csize * bs.ssize;
	vol->info               = sb;
	vol->get_file_by_id     = NULL; // not supported yet.
	vol->get_file_by_name	= NULL; // not supported yet.
	vol->get_free_space     = fat_get_free_space;
	vol->display            = fat_display_volume;
	vol->cleanup            = fat_free_volume;
	
	return 0;
}


void fat_free_volume(fs_volume_t* vol)
{
	free(vol->info);
	vol->get_file_by_id	= NULL;
	vol->get_file_by_name	= NULL;
	vol->get_free_space	= NULL;
	vol->cleanup		= NULL;
}


void fat_display_volume(fs_volume_t* vol)
{
	fat_superblock_t* super = (fat_superblock_t*) vol->info;
	fat_bootsect_t* bs = &super->bs;
	
	if(is_fat32(bs)) {
		printf("FAT Type: %.8s\n", bs->ext.f32.fs_type);
		printf("Volume label: %.11s\n", bs->ext.f32.vol_lab);
	} else {
		printf("FAT Type: %.8s\n", bs->ext.f16.fs_type);
		printf("Volume label: %.11s\n", bs->ext.f16.vol_lab);
	}
	
	printf("Sector size (bytes): %d\n", bs->ssize);
	printf("Cluster size (sectors): %d\n", bs->csize);
	
	uint32_t nsectors = (bs->sectors16 ? bs->sectors16 : bs->sectors32);
	printf("Total sectors: %d\n", nsectors);
	
	uint32_t fat_off = bs->numreserved;
	uint32_t fat_len = (is_fat32(bs) ? bs->ext.f32.sectperfat32 : bs->sectperfat16);
	uint32_t data_off = fat_len * bs->numfat + fat_off;
	
	printf("Reserved: 0 - %d\n", fat_off - 1);
	printf("FAT Tables: %d - %d\n", fat_off, data_off - 1);
	
	unsigned i;
	for(i = 0; i < bs->numfat; ++i) {
		uint32_t cur_fat_off = fat_off + (i * fat_len);
		printf("-> FAT %d: %d - %d\n", i + 1, cur_fat_off, cur_fat_off + fat_len - 1);
	}
	
	printf("Data region: %d - %d\n", data_off, nsectors - 1);
	if(!is_fat32(bs)) {
		uint32_t rootsec = bs->numreserved + (bs->sectperfat16 * bs->numfat);
		uint32_t rootsize = ((bs->numroot * sizeof(fat_dirent_t)) + bs->ssize - 1) / bs->ssize;
		printf("-> Root dirent: %d - %d\n", rootsec, rootsec + rootsize - 1);
		printf("-> Mapped clusters: %d - %d\n", rootsec + rootsize, nsectors - 1);
	}
	
	printf("First data region sector: %d\n"
	       "Total data region clusters: %d\n",
	       super->data_offset,
	       super->maxdataclust);	
}


/*====================================================================
 *
 *	Free space mapping.
 *
 *====================================================================*/


fs_extent_t* fat_get_free_space(fs_volume_t* volume)
{
	fat_superblock_t* sb = volume->info;
	fat_bootsect_t* bootsect = &sb->bs;
	
    unsigned offset = fat_offset(bootsect) * sect_size(bootsect);
	unsigned size = fat_size(bootsect) * sect_size(bootsect);
	
	// Read fat table.
	char* fat = (char*) fs_get_data(volume, offset, size);
	if(!fat) {
		return NULL;
	}

	// Skip reserved clusters.
    char* data = fat + sb->fat_skip;
	fs_extent_t* freespace = NULL;
	fs_extent_t* cur_ext = NULL;
	
	unsigned data_offset = sb->data_offset * bootsect->ssize;
	
	uint32_t value = 0;
	int in_region = 0;	// flag indicates that we are currently inside contigious free space region.
	
	// Loop over fat table and look for empty entries
	uint32_t cluster;
    for(cluster = 0; cluster < sb->maxdataclust; ++cluster) {
		
		// Get table entry value based on FAT type and advance the pointer correspondingly.
		switch(sb->fat_mask) {
			case FAT32_MASK:
				value = *(uint32_t*) data;
				data += sizeof(uint32_t);
				break;
			case FAT16_MASK:
				value = *(uint16_t*) data;
				data += sizeof(uint16_t);
				break;
			case FAT12_MASK: {
				value = *(uint16_t*) data;
				++data;
				if(cluster & 1) {
					value >>= 4;
				} else {
					value &= 0x0FFF;
					++data;
				}
				
				break;
			}
		};
		
		value &= sb->fat_mask;
		
		if(!FAT_ENTRY_ISFREE(value)) {
			// contigious region chain is broken.
			in_region = 0;
			continue;
		}
		
		if(in_region) {
			cur_ext->size += volume->block_size;
		} else {
			// New region, allocate and attach a new vector.
			fs_extent_t* ext = (fs_extent_t*) malloc(sizeof(fs_extent_t));
			if(!ext) {
				fs_put_extent_list(freespace);
				freespace = NULL;
				goto NOMEM;
			}
			
			ext->start = cluster * volume->block_size + data_offset;
			ext->size = volume->block_size;
			ext->next = NULL;
			if(cur_ext) {
				cur_ext->next = ext;
			} else {
				freespace = ext;
			}

			cur_ext = ext;
			in_region = 1;
		}
	}

	
NOMEM:	// jump here before loop completes only if extent alloc failed.
	fs_put_data(fat);
	return freespace;
}


/*====================================================================
 *
 *	File mapping.
 *
 *====================================================================*/


static inline uint32_t get_fat32(uint8_t* fat, uint32_t cluster) {
	return *((uint32_t*) fat + cluster);
}

static inline uint32_t get_fat16(uint8_t* fat, uint32_t cluster) {
	return *((uint16_t*) fat + cluster);
}

static inline uint32_t get_fat12(uint8_t* fat, uint32_t cluster) {
	fat += cluster * 2 / 3;
	return (cluster & 1) ? *fat & 0x0FFF : *fat >> 4;
}


// Build a list of extents following the entry cluster chain at specified position.
static fs_extent_t* get_cluster_chain(fat_superblock_t* super, uint8_t* fat, uint32_t start)
{
	fs_extent_t* chain = (fs_extent_t*) malloc(sizeof(fs_extent_t));
	if(!chain) {
		return NULL;
	}
	
	chain->start = start;
	chain->size = 1;
	chain->next = NULL;
	
	fs_extent_t* cur = chain;
	
	uint32_t value;
	uint32_t cluster = start;
	while(cluster < super->maxdataclust) {
		
		// Get table entry value based on FAT type and advance the pointer correspondingly.
		switch(super->fat_mask) {
			case FAT32_MASK:
				value = get_fat32(fat, cluster);
				break;
			case FAT16_MASK:
				value = get_fat16(fat, cluster);
				break;
			case FAT12_MASK:
				value = get_fat12(fat, cluster);
				break;
		};
		
		value &= super->fat_mask;
		if(FAT_ENTRY_ISEOF(value, super->fat_mask)) {
			// cluster chain end.
			break;
		} else if(FAT_ENTRY_ISUSED(value, super->fat_mask)) {
			
			if(value == (cluster + 1)) {
				// current file extent continues 
				++cur->size;
			} else {
				// new file extent
				fs_extent_t* ext = (fs_extent_t*) malloc(sizeof(fs_extent_t));
				if(!ext) {
					fs_put_extent_list(chain);
					chain = NULL;
					break;
				}
				
				ext->start = value;
				ext->size = 1;
				ext->next = NULL;
				
				cur->next = ext;
				cur = ext;
			}
			
		} else {
			// invalid FAT table, return.
			fs_put_extent_list(chain);
			return NULL;
		}

		// set next cluster value
		cluster = value;
	}
	
	return chain;
}
	

