/*
 *  Created by Evgeny Yakovlev on 7/22/11.
 *  Copyright 2011 Jetico, Inc. All rights reserved.
 *
 *  FAT12/16/32 on-disk format structure.
 */

#ifndef LIBFS_FAT_FORMAT_H
#define LIBFS_FAT_FORMAT_H

#include <stdint.h>

#define FAT_BOOTSECT_MAGIC	0xaa55
#define FAT_EXTBLOCK_SIG1	0x29
#define FAT_EXTBLOCK_SIG2	0x28

// fs_type values in extended block.
#define FAT12_TYPE		"FAT12  "
#define FAT16_TYPE		"FAT16  "
#define FAT32_TYPE		"FAT32  "

// FAT table entry masks
#define	FAT12_MASK		0x00000fff	/* mask for 12 bit cluster numbers */
#define	FAT16_MASK		0x0000ffff	/* mask for 16 bit cluster numbers */
#define	FAT32_MASK		0x0fffffff	/* mask for FAT32 cluster numbers */

// First legal cluster number.
#define FAT_FIRST_CLUSTER	2

// Reserved fat entries skip sizes in bytes.
#define FAT12_SKIP_SIZE		(FAT_FIRST_CLUSTER * 3 / 2 /* 2 entries in 3 bytes*/ ) 
#define FAT16_SKIP_SIZE		(FAT_FIRST_CLUSTER * sizeof(uint16_t))
#define FAT32_SKIP_SIZE		(FAT_FIRST_CLUSTER * sizeof(uint32_t))

// FAT entry constants
#define FAT_ENTRY_FREE      0           // last free entry
#define FAT_ENTRY_BAD		0x0ffffff7	// bad block
#define FAT_ENTRY_EOFS		0x0ffffff8	// end-of-file region start
#define FAT_ENTRY_EOFE		0x0fffffff	// end-of-file region end

// Entry parsing macros based on FAT type
#define FAT_ENTRY_ISFREE(val)		((val) == FAT_ENTRY_FREE)
#define FAT_ENTRY_ISEOF(val, mask)	((val >= (FAT_ENTRY_EOFS & mask)) && (val <= (FAT_ENTRY_EOFE & mask)))
#define FAT_ENTRY_ISBAD(val, mask)	((val) == (FAT_ENTRY_BAD & mask))
#define FAT_ENTRY_ISUSED(val, mask)	(!(FAT_ENTRY_ISFREE(val) || FAT_ENTRY_ISEOF(val, mask)))

#pragma pack(push, 1)

/**
 * On-disk boot sector structure.
 */
typedef struct fat_bootsect 
{
        uint8_t		jmpcode[3];         // jump instruction to bootcode.
        char		oemname[8];         // FS creator signature, usually is set to "MSWIN4.1", but does no matter.
        uint16_t	ssize;              // bytes per sector, can be only 512,1024,2048 and 4096.
        uint8_t		csize;              // sectors per cluster, power of 2, > 0, <= 128. ssize * csize shall be <= 32k
        uint16_t	numreserved;        // number of reserved sectors for boot sectors, starting at the first sector of the volume. Shall not be 0. For FAT12/16 must be 1, for FAT32 is typically = 32.
        uint8_t		numfat;             // Number of FAT tables, should always contain 2.
        uint16_t	numroot;            // count of 32-byte dir entries in the root directory. For FAT32 should be set to 0.
        uint16_t	sectors16;          // count of sectors on volume for FAT12/16. For FAT32 should be set to 0. If 0, then sectors32 should be used.
        uint8_t		mediadesc;          // media descriptor. Usually = 0xF8 for non-removable drives. Allowed 0xF0, 0xF8, 0xF9, 0xFA-0xFF.
        uint16_t	sectperfat16;       // size of a single FAT table, in sectors, for FAT12/16. On FAT32 should be 0.
        uint16_t	sectpertrack;       // size of the track in sectors for volumes with "geometry"
        uint16_t	numheads;           // number of heads for volumes with geometry
        uint32_t	prevsect;           // number of hidden sectors before FS partition. Should always be zero on volumes without partitions.
        uint32_t	sectors32;          // count of sectors. If 0, then sectors16 must be used.
	
        /* The following are different extended BIOS parameter blocks for fat12/fat16 and fat32 */
        union {

		/* FAT12/16 extended BIOS parameter block */
		struct {
            uint8_t		drivenum;       // physical drive number for int 0x13.
            uint8_t		_reserved;      // Should be set to 0.
            uint8_t		extsig;         // extended boot signature, =0x29, indicates that following 3 fields in the boot sector are present.
            uint32_t	vol_id;         // volume id, used to track that the wrong disk is inserted in a removable drive.
            uint8_t		vol_lab[11];    // volume label.
            char		fs_type[8];     // informational-only string, can be not set, or set incorrectly.
			uint8_t		bootcode[448];
		} f16;
		
		/* FAT32 extended BIOS parameter block */
		struct {
            uint32_t	sectperfat32;   // count of sectors per one FAT, only defined for FAT32
            uint16_t	ext_flag;       // bits 0-3 - number of active FAT, only valid if no mirroring.
                                        // 4-6      - reserved
                                        // 7        - 0 - FAT is mirrored at runtime into all FATs, 1 - only one FAT is active.
            uint16_t	fs_ver;         // high byte - major revision number, low byte - minor revision number. Should be 0:0, otherwise will not operate with this volume.
            uint32_t	rootclust;      // first cluster where root directory is stored
            uint16_t	fsinfo;         // fat_fsinfo location
            uint16_t	bs_backup;      // boot sector backup sector number in the reserved area. Usually is set to 6.
			uint8_t		_reserved1[12];
            uint8_t		drvnum;         // physical drive number for 0x13.
			uint8_t		_reserved2;
            uint8_t		extsig;         // extended boot signature. Same as for FAT12/16.
			uint32_t	vol_id;
			uint8_t		vol_lab[11];
			char		fs_type[8];
			uint8_t		bootcode[420];
		} f32;
        } ext;
	
        uint16_t magic;       // MAGIC for all versions, =0xAA55
} fat_bootsect_t;


/**
 * FS information sector.
 * FAT32 only.
 */
typedef struct fat_fsinfo {
        uint8_t magic1[4];      // value = 0x41615252, used to validate that this is fs-info sector.
        uint8_t f1[480];        // reserved field. All bytes should be = 0.
        uint8_t magic2[4];      // value = 0x61417272
        uint8_t freecnt[4];     // contains the last known count of free clusters, or 0xfffffffff if unknown
        uint8_t nextfree[4];    // next free cluster, or 0xffffffff, if unknown.
        uint8_t f2[12];         // reserved field, all bytes should be = 0.
        uint8_t magic3[4];      // value = 0xAA550000, used to validate that this is fs-info sector.
} fat_fsinfo_t;


/**
 * Directory entry structure.
 * Also see long name structure for later versions.
 */
typedef struct fat_dirent {
        uint8_t name[8];        // short name
        uint8_t ext[3];         // /\
                                // name[0] == 0xE5 - directory entry is free
                                // name[0] == 0x00 - entry is free and last (no allocated entries after that)
                                // name[0] == 0x05 - actual byte is 0xE5 (Japan)

        uint8_t attrib;         // 0x01 - read-only
                                // 0x02 - hidden
                                // 0x04 - system file
                                // 0x08 - volume id, should be only one file with this attribute in each volume. Name of the file is a volume label.
                                // 0x10 - directory
                                // 0x20 - archive
                                // 0x0E - file have long name.

        uint8_t lowercase;      // reserved for WinNT
        uint8_t ctimeten;       // thenths of milliseconds of the creation time. (interval 0-199 actually)
        uint8_t ctime[2];       // creation time
        uint8_t cdate[2];       // creation date
        uint8_t adate[2];       // access time
        uint8_t highclust[2];   // high word of this entry's first cluster number (always 0 for FAT12/16)
        uint8_t wtime[2];       // last write time
        uint8_t wdate[2];       // last write date
        uint8_t startclust[2];  // low word of this entry's first cluster number.
        uint8_t size[4];        // file size in bytes. Not used for dirs ( = 0).
} fat_dirent_t;


/**
 * Long file name entry.
 * Those are fake entries preceding a short name entry they are associated with.
 *
 */
typedef struct fat_long_name_entry {
        uint8_t seq;            // the order of the entry in a sequence. If masked with 0x40 (0x40 | N), this is the last. Cannot be 0x00 (last free entry) and 0xE5 (entry is free).
        uint8_t part1[10];      // symbols 1-5 of the long name.
        uint8_t attributes;     // must be 0x0E
        uint8_t reserved1;
        uint8_t chksum;         // checksum of name in the short dir entry at the end of the long dir set.
        uint8_t part2[12];      // symbols 6-11 of the long name.
        uint8_t reserved2[2];   // should be = 0.
        uint8_t part3[4];       // symbols 12-13 of the long name.
} fat_long_name_entry_t;


#pragma pack(pop)

#endif
