//
//  Copyright (c) 2011 Jetico, Inc. All rights reserved.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>

#include "fs_disk.h"
#include "mbr.h"
#include "gpt.h"

/**
 * Disk type init proc.
 * Each implemented partition type implements this.
 */
typedef int (*fslib_disk_init_proc) (int fd, fs_disk_t* out_disk);

static int parse_mbr(int fd, fs_disk_t* disk);
static int parse_gpt(int fd, fs_disk_t* disk);
static int parse_raw(int fd, fs_disk_t* disk);

static const fslib_disk_init_proc gDiskInitializers[] = {
    parse_gpt, // this is important: we need to try guid before mbr since guid does have a protective mbr sector
    parse_mbr,
    parse_raw, // this must be the last one: no partitions found
};


#define DISKS_TOTAL ( sizeof(gDiskInitializers) / sizeof(*gDiskInitializers) )


static int parse_mbr(int fd, fs_disk_t* out_disk)
{
    
    memset(out_disk, 0, sizeof(*out_disk));
    
    struct mbr_sector mbr;
    if(sizeof(mbr) != pread(fd, &mbr, sizeof(mbr), 0)) {
        return EINVAL;
    }
    
    if(FAT_BOOTSECT_MAGIC != mbr.signature) {
        return EINVAL;
    }
    
    memcpy(out_disk->signature, mbr.disk_signature, sizeof(mbr.disk_signature));
    out_disk->total_partitions = 0;
    out_disk->type = "MBR";
    
    // read partition table info
    struct fs_partition parts[16]; // maximum number of partitions on an mbr disk including extended.
    struct mbr_partition* partition;
    unsigned i;
    
    for(i = 0; i < 4; ++i) {
        partition = &mbr.partition_table[i];
        if(MBR_IS_VALID_PARTITION(partition)) {
            parts[out_disk->total_partitions].start_sector = partition->first_sector_lba;
            parts[out_disk->total_partitions].total_sectors = partition->total_sectors;
            parts[out_disk->total_partitions].fstype = partition->type;
               
            out_disk->total_partitions++;
        }
    }
    
    if(out_disk->total_partitions) {
        out_disk->partitions = (struct fs_partition*) calloc(out_disk->total_partitions, sizeof(struct fs_partition));
        if(!out_disk->partitions) {
            return ENOMEM;
        }
        
        memcpy(out_disk->partitions, parts, out_disk->total_partitions * sizeof(struct fs_partition));
    } else {
        // no partitions found: return error because FAT filesystem
        // has the same 0x55AA signature as a MBR sector
        return EINVAL;
    }
    
    return 0;
}

// fallback procedure.
// interpret a disk as unpartitioned space
static int parse_raw(int fd, fs_disk_t* out_disk)
{
    // TODO:
    // 1. find out disk size
    off_t   pos;
    pos = lseek(fd, 0, SEEK_END);
    lseek(fd, 0, SEEK_SET);

    // 2. output single partition covering whole disk
    out_disk->partitions = (struct fs_partition*) calloc(1, sizeof(struct fs_partition));
    if(!out_disk->partitions) {
        return ENOMEM;
    }

    out_disk->total_partitions = 1;
    out_disk->type = "RAW";
    out_disk->partitions[0].start_sector = 0LL;
    out_disk->partitions[0].total_sectors = pos / 512;
//    out_disk->partitions[0].fstype =


    return 0;
}


static int gpt_partition_type(GUID* guid)
{
    if(0 == memcmp(guid, &gGUID_EFI, sizeof(GUID))) {
        return kFSPartition_EFI;
    } else if(0 == memcmp(guid, &gGUID_HFSP, sizeof(GUID))) {
        return kFSPartition_HFSP;
    } else {
        return kFSPartition_Unknown;
    }
}


static int parse_gpt(int fd, fs_disk_t* out_disk)
{
    memset(out_disk, 0, sizeof(*out_disk));
    struct gpt gpt;
    if(sizeof(gpt) != pread(fd, &gpt, sizeof(gpt), 0)) {
        return EINVAL;
    }

    struct gpt_header* header = &gpt.header;
    
    if(0 != memcmp(header->signature, GPT_SIGNATURE, sizeof(header->signature))) {
        return EINVAL;
    }

    memcpy(out_disk->signature, header->disk_guid, sizeof(header->disk_guid));
    out_disk->total_partitions = 0;
    out_disk->type = "GPT";
    
    // read partition table info
    unsigned i;
    struct fs_partition parts[128]; // maximum possible entries
    struct gpt_entry* entry;
    for(i = 0; i < header->total_entries; ++i) {
        entry = &gpt.entries[i];
        if(0 == memcmp(&entry->type_guid, &gGUID_Empty, sizeof(GUID))) {
            break;
        }
        
        parts[i].start_sector = entry->start_sector;
        parts[i].total_sectors = entry->last_sector - entry->start_sector + 1;
        parts[i].fstype = gpt_partition_type(&entry->type_guid);
        
        out_disk->total_partitions++;
    }
    
    if(out_disk->total_partitions) {
        out_disk->partitions = (struct fs_partition*) calloc(out_disk->total_partitions, sizeof(struct fs_partition));
        if(!out_disk->partitions) {
            return ENOMEM;
        }
    
        memcpy(out_disk->partitions, parts, out_disk->total_partitions * sizeof(struct fs_partition));
    }
    
    return 0;
}


int fslib_parse_disk(const char* whole_dev, fs_disk_t* out_disk)
{
    memset(out_disk, 0, sizeof(*out_disk));
	
	struct stat statbuf;
	int rc = stat(whole_dev, &statbuf);
	if(rc != 0) {
		fprintf(stderr, "stat failed: %s\n", strerror(errno));
		return EIO;
	}
	
    int fd = open(whole_dev, O_RDONLY);
    if(fd < 0) {
        return errno;
    }
    
    
	// Iterate disk initializers and parse partition table format
	unsigned i = 0;
	while(i < DISKS_TOTAL) {
		fslib_disk_init_proc disk_init = gDiskInitializers[i++];
		if( 0 == disk_init(fd, out_disk) ) {
            close(fd);
			return 0;
		}
	}
    
    close(fd);
    
	fprintf(stderr, "Unknown disk format\n");
    return EINVAL;

}


void fslib_put_disk(fs_disk_t* disk)
{
    free(disk->partitions);
}

