/*
 *  fslib.c
 *  Copyright 2011 Jetico, Inc. All rights reserved.
 *
 *  Top-level library interface implementation.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <fcntl.h>
#include <errno.h>
#include <fstab.h>
#include <sys/stat.h>

#include "fslib.h"
#include "fat/fat.h"
#include "hfsplus/hfsp.h"
#include "exfat/exfat.h"

typedef int (*fs_init_volume_t) (fs_volume_t* volume);

static const fs_init_volume_t g_vol_initializers[] = {
	hfsp_init_volume,
	fat_init_volume,
	exfat_init_volume,
};

#define FS_TOTAL (sizeof(g_vol_initializers) / sizeof(*g_vol_initializers))


int fslib_open_volume(const char* path, int mode, fs_volume_t* out_volume)
{
	memset(out_volume, 0, sizeof(*out_volume));
	
	struct stat statbuf;
	int rc = stat(path, &statbuf);
	if(rc != 0) {
		fprintf(stderr, "stat failed: %s\n", strerror(errno));
		return EIO;
	}
	/*
	if( !(S_ISCHR(statbuf.st_mode) || S_ISBLK(statbuf.st_mode)) ) {
		fprintf(stderr, "Invalid volume path %s\n", path);
		return EINVAL;
	}
	*/
    
	out_volume->fd = open(path, mode);
	if(out_volume->fd < 0) {
		fprintf(stderr, "Failed to open device '%s': %s\n", path, strerror(errno));
		return EIO;
	}
	
	// Iterate volume initializers and parse fs format
	unsigned i = 0;
	while(i < FS_TOTAL) {
		fs_init_volume_t vol_init = g_vol_initializers[i++];
		if( 0 == vol_init(out_volume) ) {
			return 0;
		}
	}

	fprintf(stderr, "Unknown volume format\n");
	fslib_close_volume(out_volume);
	return EINVAL;
}

int fslib_open_volume_fd(int fd, fs_volume_t* out_volume)
{
	memset(out_volume, 0, sizeof(*out_volume));
 
	out_volume->fd = fd;
	if(out_volume->fd < 0) {
		return EIO;
	}
	
	// Iterate volume initializers and parse fs format
	unsigned i = 0;
	while(i < FS_TOTAL) {
		fs_init_volume_t vol_init = g_vol_initializers[i++];
		if( 0 == vol_init(out_volume) ) {
			return 0;
		}
	}
    
	fprintf(stderr, "Unknown volume format\n");
	fslib_close_volume(out_volume);
	return EINVAL;
}

void fslib_close_volume(fs_volume_t* volume)
{
	if(volume->cleanup) {
		volume->cleanup(volume);
	}
	
	close(volume->fd);
}

void* fs_get_data(fs_volume_t *v, uint64_t start, int size)
{
    char	*ptr;
    ssize_t	ret;
    
    assert(v);
    
    ptr = calloc(1, size);
    if (!ptr) {
        fprintf(stderr, "1\n");
        return NULL;
    }
    
    ret = fs_read_data(v, ptr, start, size);
    if (0 > ret) {
        fprintf(stderr, "2\n");

        free(ptr);
        ptr = NULL;
    }
    return ptr;
}

int	fs_read_data(fs_volume_t *v, char *ptr, uint64_t start, int size)
{
    uint64_t	offset;
    ssize_t	ret, x = 0;
    
    assert(v);
    assert(ptr);
    
    offset = v->offset+start;

    for (x = 0; x < size; ) {
        
        ret = pread(v->fd, ptr+x, size-x, offset+x);
        if (-1 == ret) {
            if (EINTR == errno || EAGAIN == errno || EWOULDBLOCK == errno) {
                continue;
            }
            
            fprintf(stderr, "pread(%lu, %lld) failed: %s\n", (unsigned long)size-x, offset+x, strerror(errno));
            return -1;
        }
        
        if (0 == ret) {
            return x;
        } else {
            x += ret;
        }
    }
    
    return x;
}

int64_t fs_bmap_offset(fs_volume_t* vol, fs_extent_t* ext, uint64_t start)
{
    fs_extent_t *p;
    
    assert(vol);
    
    for (p = ext; p; p = p->next) {
        if (start > p->size) {
            start -= p->size;
            continue;
        }
        
        return p->start+start;
    }
    
    // out of range
    return -2;
}


int fs_read_extent(fs_volume_t *vol, char *buf, fs_extent_t *ext, uint64_t start, int size)
{
    fs_extent_t *p;
    int		sz, len, ret;
    
    assert(vol);
    assert(buf);
    
    len = 0;
    for (p = ext; p; p = p->next) {
        if (start > p->size) {
            start -= p->size;
            continue;
        }
        sz = start+size > p->size ? p->size - start : size;
        ret = fs_read_data(vol, buf+len, p->start+start, sz);
        if (0 > ret) {
            return -1;
        }
        
        len += ret;
        size -= ret;
        start = 0;
        if (sz != ret || 0 == size) {
            return len;
        }
        
    }
    // out of range
    return -2;
}

void fs_put_data(void *ptr)
{
    assert(ptr);
    free(ptr);
} 


void fs_put_extent_list(fs_extent_t *ptr)
{
    fs_extent_t *p;
    
    while (ptr) {
        p = ptr;
        ptr = ptr->next;
        free(p);
    }
}

void fs_print_extent_list(fs_extent_t *ptr, int blksz)
{
    if (!ptr) {
        fprintf(stderr, "The list is empty\n");
    }
    while (ptr) {
        if (blksz) {
            fprintf(stderr, "%16lld \t0x%llx \t0x%llx\n", ptr->size/blksz, ptr->start/blksz, (ptr->start+ptr->size-1)/blksz);	
        } else {
            fprintf(stderr, "%16lld \t0x%llx \t0x%llx\n", ptr->size, ptr->start, ptr->start+ptr->size-1);
        }
        ptr = ptr->next;
    }
}

void fs_put_file(fs_file_t* file)
{
    if(!file) {
        return;
    }
    
    fs_area_t* area = file->areas;
    fs_area_t* next;
    while(area) {
        next = area->next;
        fs_put_extent_list(area->extents);
        free(area);
        area = next;
    }
    
    file->areas = NULL;
}

