/*
 *  Copyright 2011 Jetico, Inc. All rights reserved.
 */

#ifndef FSLIB_FSLIB_H
#define FSLIB_FSLIB_H

#include <stdint.h>

#ifdef FSLIB_DEBUG
# include <stdio.h>
# include <assert.h>
# define FSLIB_DEBUG_LOG(format, args...) fprintf(stderr, (format), ## args)
# define FSLIB_ASSERT(pred) assert( (pred) )
#else
# define FSLIB_DEBUG_LOG(format, args...) 
# define FSLIB_ASSERT(pred) 
#endif

#pragma pack(push)
#pragma pack(1)

/**
 * Extent vector.
 * List of those describes volume free space and file data areas layout.
 * Everything is in byte units.
 */
typedef struct fs_extent {
    uint64_t            start;      // first byte of the extent
    uint64_t            size;       // total bytes if the extent.
    struct fs_extent	*next;      // next extent in list.
} fs_extent_t;



/**
 * File area layout.
 * Possible areas are fs-dependant and usually include file data streams forks. 
 */
typedef struct fs_area {
    char                tag[8];     // fs-specific tag
    fs_extent_t*        extents;    // layout
    struct fs_area*     next;       // next area
} fs_area_t;



/**
 * File info structure.
 */
typedef struct fs_file {
    fs_area_t*          areas;      // file data and metadata areas NULL-terminated list. 
} fs_file_t;


/**
 * FS Volume structure.
 * Describes parsed volume format.
 * Implemented by each supported file system format code.
 * All function implementation is optional, set to NULL if not implemented.
 */
typedef struct fs_volume {
    int                 fd;         // volume file descriptor
    uint64_t            offset;     // volume start offset in bytes
    uint64_t            size;       // volume size in sectors
    unsigned            block_size;	// volume allocation block size in bytes.
    const char          *path;		// volume file name
    void                *info;		// fs-specific volume information
    
    
    /**
     * Display information about parsed volume.
     * @vol         Owning volume.
     */
    void (*display) (struct fs_volume* vol);
    
    /**
     * Get info for file with this name.
     * @p vol         Volume.
     * @p fname       File name.
     * @p out_file    Initialized file info on success. Should be freed with fs_put_file.
     * @return      
     */
    int (*get_file_by_name) (struct fs_volume* vol, const char* fname, fs_file_t* out_file);
    
    /**
     * Get info for file with this id.
     * @p vol         Volume.
     * @p fname       File id, usually an inode number returned by stat(2).
     * @p out_file    Initialized file info on success. Should be freed with fs_put_file.
     * @return      
     */
    int (*get_file_by_id) (struct fs_volume* vol, uint32_t id, fs_file_t* out_file);
    
    /**
     * Build volume free space extent list.
     * @p vol       Volume.
     * @return      Extent list on success, NULL on failure.
     */
    fs_extent_t* (*get_free_space) (struct fs_volume* vol);
    
    /**
     * Cleanup volume resouces.
     * Should not be called directly, use fslib_close_volume instead.
     */
    void (*cleanup) (struct fs_volume* vol);
                  
} fs_volume_t;

#pragma pack(pop)

#ifdef __cplusplus
extern "C" {
#endif
	
    
	/**
     * Open volume and parse its structure.
	 *
     * @p path          Volume block device or mount point path.
     * @p mode          Unix file open mode.
     * @p out_volume	Parsed and opened volume on success.
	 */
	int fslib_open_volume(const char* path, int mode, fs_volume_t* out_volume);

	/**
     * Open volume and parse its structure.
	 *
     * @p fd            Open file descriptor for the volume.
     * @p out_volume	Parsed and opened volume on success.
	 */
	int fslib_open_volume_fd(int fd, fs_volume_t* out_volume);
	
	/**
	 * Close volume handle and release all resources.
     * @p volume      Volume to release.
	 */
	void fslib_close_volume(fs_volume_t* volume);

	
    /**
     * Read data from the volume to memory. Can be implemented via malloc()+read() or
     * mmap(). So the pointer may point to read-only memory
     *
     * @p v		volume descriptor
     * @p start	start position
     * @p size		data size in bytes
     * @return:	pointer to the sector contents; must be freed by fs_put_sectors()
     */
    void* fs_get_data(fs_volume_t *v, uint64_t start, int size);
    
    
    /**
     * Read data from the volume to the supplied buffer
     * @p v         volume descriptor
     * @p buf		target buffer, must be large enough for size bytes
     * @p start     start offset
     * @p size		data size
     * @return      number of bytes read
     */
    int	fs_read_data(fs_volume_t *v, char *buf, uint64_t start, int size);
    
    
    /**
     * Read extent data from the volume to the supplied buffer
     * @p v		volume descriptor
     * @p buf		target buffer, must be large enough for size bytes
     * @p ext		extent
     * @p start	start offset
     * @p size		data size
     * @return	number of bytes read
     */
    int fs_read_extent(fs_volume_t *vol, char *buf, fs_extent_t *ext, uint64_t start, int size);
    
    
    int64_t fs_bmap_offset(fs_volume_t* vol, fs_extent_t* ext, uint64_t start);

    
    /**
     * Free resources
     * @p ptr		pointer returned by fs_get_data()
     */
    void fs_put_data(void *ptr);
    
    
    /**
     * Print extent list
     */
    void fs_print_extent_list(fs_extent_t *ptr, int blksz);
    

    /**
     * Free extent list
     */
    void fs_put_extent_list(fs_extent_t *ptr);
    
    
    /**
     * Free file info structure.
     */
    void fs_put_file(fs_file_t* file);
    
    
#ifdef __cplusplus
}
#endif

#endif
