#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "fs_endian.h"
#include "fs_unicode.h"
#include "hfsp.h"

typedef struct bt_catalog_param {
    uint32_t		stage;		
    uint32_t		fileID;		// file catalog id
    HFSUniStr255	name;
} bt_attrib_param_t;

static int compare_attrib_key(bt_context_t *ctx, void *k1, void *k2) 
{
    HFSPlusAttrKey	*key = (HFSPlusAttrKey *)k1;
    bt_attrib_param_t	*p = (bt_attrib_param_t *)k2;
    uint32_t		fileID = be32(key->fileID);
    
    if (fileID > p->fileID) {
        return 1;
    } else if (fileID < p->fileID) {
        return -1;
    }
    // we look for all attributes, so do not compare names and start blocks
    fprintf(stderr, "Attribute: ");
    print_unistr((HFSUniStr255*)(&key->attrNameLen));
    fprintf(stderr, "\t start block %d(%X)\n", be32(key->startBlock), be32(key->startBlock));
    return 0;
}

static int show_attributes(bt_context_t *ctx, void *data, bt_param_t *param) 
{
    uint32_t	*rec = (uint32_t *)data;
    const char	*s = NULL;
    //bt_attrib_param_t		*p = (bt_attrib_param_t *)(param->key);
    HFSPlusAttrData		*pi = NULL;
    HFSPlusAttrForkData		*pf = NULL;
    HFSPlusAttrExtents		*pe = NULL;
    
    fprintf(stderr, "--- show attributes called\n");
    
    switch (be32(*rec)) {
        case kHFSPlusAttrInlineData:
            pi = (HFSPlusAttrData *)data;
            s = "Inline"; 
            break;
        case kHFSPlusAttrForkData:
            pf = (HFSPlusAttrForkData *)data;	      
            s = "Fork"; 
            break;
        case kHFSPlusAttrExtents:	
            pe = (HFSPlusAttrExtents *)data;
            s = "Extents"; 
            break;
        default:	
            fprintf(stderr, "Invalid record type\n");
            return 0;
    }
    
    if (pi) {
        int64_t bmap = fs_bmap_offset(ctx->vol, ctx->ext, param->node * ctx->size);
	fprintf(stderr, "Type: %s, Size: %d, Node %u, Offset %d, Absolute offset %lld\n", 
        s, be32(pi->attrSize), param->node, (int)(param->offset+sizeof(*pi)-2), (long long)bmap);

    }
    
    if (pf) {
	fprintf(stderr, "Type: %s ", s);
	hfsp_print_fork_data(&pf->theFork);
    }
    
    if (pe) {
    fprintf(stderr, "Type: %s, Size: %u", s, (unsigned)be32(pi->attrSize));
    }
    // process next attributes
    return 0;
}

int hfsp_attrib_tree(fs_volume_t *vol, uint32_t fileID, fs_file_t* out_file) 
{
    fs_extent_t		*tree;
    hfsp_info_t		*info;
    HFSPlusForkData	*fork;
    int			ret;
    bt_context_t     	ctx;
    bt_param_t		idx;
    bt_attrib_param_t	key;
    
    assert(vol);
    assert(out_file);

    memset(&ctx, 0, sizeof(ctx));
    
    info = (hfsp_info_t *)vol->info;
    fork = &(info->head.attributesFile);
    
    fprintf(stderr, "=== Building extents for attributes file...\n");
    tree = hfsp_build_extent_list(vol, kHFSAttributesFileID, 0, fork);
    if (!tree) {
        return -1;
    }
    
    ctx.vol	= vol;
    ctx.ext	= tree;
    ctx.total	= be64(fork->logicalSize);
    
    
    key.fileID      = fileID;
    key.name.length = 0;
    key.stage       = 0;
    
    idx.height	= 0;
    idx.left 	= 0;
    idx.key     = &key;
    idx.compare	= compare_attrib_key;
    idx.leaf	= show_attributes;
    idx.result  = (void*) out_file;
    
    fprintf(stderr, ">>> Processing the attributes file...\n");
    ret = hfsp_iterate_btree(&ctx, &idx);
    fprintf(stderr, "<<< Attribute processing completed\n");
    
    fs_put_extent_list(tree);
    
    return ret;
}
