#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "fs_endian.h"
#include "hfsp.h"

static inline int get_keylen(bt_context_t *ctx, void *k, int is_index) 
{
    uint8_t	*p8;
    uint16_t	*p16;
    int		len, use16;
    
    use16 = ctx->attr & kBTBigKeysMask;
    
    if (is_index && !(ctx->attr & kBTVariableIndexKeysMask)) {
        len = ctx->keylen;
        len += use16 ? 2 : 1;	// add the lenght of the length
    } else {
        if (use16) {
            p16 = (uint16_t *)k;
            len = be16(*p16);
            len += 2;
        } else {
            p8 = (uint8_t *)k;
            len = *p8;
            len += 1;	    
        }
    }
    return len;
}

int hfsp_read_btree_header(char *p, bt_context_t *ctx) 
{
    BTNodeDescriptor	*pn;
    BTHeaderRec		*ph;
    
    pn = (BTNodeDescriptor *)p;
    if (kBTHeaderNode != pn->kind) {
        return -1;
    }
    
    ph = (BTHeaderRec *)(pn+1);
    ctx->size	= be16(ph->nodeSize);
    ctx->depth	= be16(ph->treeDepth);
    ctx->root	= be32(ph->rootNode);
    ctx->attr	= be32(ph->attributes);
    ctx->keylen = be16(ph->maxKeyLength);
    
    // node size must be power of 2
    if (ctx->size & (ctx->size-1)) {
        return -1;
    }
    
    return 0;
    
}

static int get_node_type(bt_context_t *ctx, BTNodeDescriptor *p)
{
    if (be16(p->numRecords) > (ctx->size-sizeof(BTNodeDescriptor))/sizeof(uint16_t)) {
        fprintf(stderr, "Bad records number!\n");
        return -1;
    }
    
    switch(p->kind) {
        case kBTHeaderNode:
            fprintf(stderr, "Why process header once more?\n");
            return 0;
        case kBTIndexNode:
            return 1;
            break;
        case kBTLeafNode:
            return 0;
            break;
        default:
            return -1;
    } 
}

uint32_t hfsp_process_index(bt_context_t *ctx, char *buf, uint32_t node, bt_param_t *param) 
{
    BTNodeDescriptor	*p = (BTNodeDescriptor *)buf;
    void		*key;
    int			i, offset, records, ret, keylen;
    uint16_t		*po;
    uint32_t		*pidx, next;
    
    while (node) {
        
        ret = fs_read_extent(ctx->vol, buf, ctx->ext, node * ctx->size, ctx->size);
        if (ret != ctx->size) {
            return 0;
        }
        
        hfsp_print_btree_node(buf, node, ctx->size);
        
        // validate node type
        ret = get_node_type(ctx, p);
        if (0 > ret) {
            return 0;
        }
        
        // return leaf node; check the current height later
        if (0 == ret) {
            return node;
        }
        
        // entered new level
        if (param->height != p->height) {
            param->left = 0;
        }
        
        po = (uint16_t *)(buf+ctx->size) - 1;
        records = be16(p->numRecords);
        
        for (i = 0; i < records; i++) {
            offset = be16(po[-i]);
            key    = buf+offset;
            keylen = get_keylen(ctx, key, 1);
            pidx   = (uint32_t *)(buf+offset+keylen);
            ret    = param->compare(ctx, key, param->key);
            //	    fprintf(stderr, ("INDEX %3d: %d cmp=%d len=%d, next=%d[%x]\n", i, offset, ret, keylen, be32(*pidx), be32(*pidx));
            
            // param < key
            if (0 > ret) {
                param->left = be32(*pidx);
                continue;
            }
            
            // param >= key
            if (param->left) {
                next = (2 == p->height) ? param->left : hfsp_process_index(ctx, buf, param->left, param);
                //next = hfsp_process_index(ctx, buf, param->left, param);
                return next;	    
            }
            
            // if param->left was not set, we require param==key
            if (0 == ret) {			
                param->left = be32(*pidx);
                next = (2 == p->height) ? param->left : hfsp_process_index(ctx, buf, param->left, param);
                //next = hfsp_process_index(ctx, buf, param->left, param);
                return next;
            }
            
            // param is out of range: param->left was not set, but pparam > key
            fprintf(stderr, "ERROR! the key is out of range at %d of %d\n", i, records);
            return 0;
        }
        
        // searched through this node, but did not find fileID, so try to check the next node
        node = be32(p->fLink);
    }
    
    // we are in the rightmost node
    if (param->left) {
        next = (2 == p->height) ? param->left : hfsp_process_index(ctx, buf, param->left, param);
        //next = hfsp_process_index(ctx, buf, param->left, param);
        return next;	    
    }
    
    fprintf(stderr, "ERROR! the key is out of range; right edge reached\n");
    return -1;
}

int hfsp_process_leaves(bt_context_t *ctx, char *buf, uint32_t node, bt_param_t *param) 
{
    BTNodeDescriptor	*p = (BTNodeDescriptor *)buf;
    void		*key, *data;
    int			i, offset, records, ret, keylen;
    uint16_t		*po;
    
    while (node) {
        
        ret = fs_read_extent(ctx->vol, buf, ctx->ext, node * ctx->size, ctx->size);
        if (ret != ctx->size) {
            return -1;
        }
        
        hfsp_print_btree_node(buf, node, ctx->size);
        
        // validate node type, require leaf node
        ret = get_node_type(ctx, p);
        if (0 != ret) {
            return -1;
        }
        
        po = (uint16_t *)(buf+ctx->size) - 1;
        records = be16(p->numRecords);
        param->node = node;
        
        for (i = 0; i < records; i++) {
            offset = be16(po[-i]);
            key    = buf+offset;
            keylen = get_keylen(ctx, key, 0);	// leaf node key length is aways here
            data   = buf+offset+keylen;
            ret	   = param->compare(ctx, key, param->key);
            
            //fprintf(stderr, "LEAF  node=%u #%3d: offset=%d cmp=%d len=%d\n", node, i, offset, ret, keylen);
            
            // param < key
            if (0 > ret) {
                continue;
            }
            
            // param > key
            if (0 < ret) {
                return 0;
            }
            
            param->offset = offset;
            param->length = be16(po[-i - 1]) - offset;
            ret = param->leaf(ctx, data, param);
            if (ret) {
                return ret;
            }
            
            
        }
        
        // proceed with the next node
        node = be32(p->fLink);
    }
    
    return 0;
}

int hfsp_iterate_btree(bt_context_t *ctx, bt_param_t *param) 
{
    char		*buf;
    int			ret;
    uint32_t		leaf;
    
    buf = calloc(1, HFSP_BUFSIZE);
    if (!buf) {
        return ENOMEM;
    }
    
    // read the header
    ret = fs_read_extent(ctx->vol, buf, ctx->ext, 0, 4096);
    if (ret != 4096) {
        ret = -1;
        goto exit;
    }
    
    // check the header and fill tree params
    ret = hfsp_read_btree_header(buf, ctx);
    if (ret) {
        goto exit;
    }
    
    hfsp_print_btree_node(buf, 0, 0);
    
    // find the head leaf node
    leaf = hfsp_process_index(ctx, buf, ctx->root, param);
    if (!leaf) {
        ret = ENOENT;
        goto exit;
    }
    
    // process leaf nodes
    ret = hfsp_process_leaves(ctx, buf, leaf, param);
    
exit:
    free(buf);
    return ret;
}

