#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "fs_endian.h"
#include "fs_unicode.h"
#include "hfsp.h"

typedef struct bt_catalog_param {
    uint32_t		stage;		
    uint32_t		fileID;		// file catalog id
    HFSUniStr255	name;
} bt_catalog_param_t;

static int compare_catalog_key(bt_context_t *ctx, void *k1, void *k2) 
{
    HFSPlusCatalogKey	*key = (HFSPlusCatalogKey *)k1;
    bt_catalog_param_t	*p = (bt_catalog_param_t *)k2;
    uint32_t		fileID = be32(key->parentID);
    int			ret;
    
    if (fileID > p->fileID) {
        return 1;
    } else if (fileID < p->fileID) {
        return -1;
    }
    
    ret = strcasecmp_u16be(key->nodeName.unicode, be16(key->nodeName.length), p->name.unicode, be16(p->name.length));
    /*
     fprintf(stderr, "### KEY=%d:", fileID);
     print_unistr(&key->nodeName);
     fprintf(stderr, " vs ");
     print_unistr(&p->name);
     fprintf(stderr, " ret=%d\n", ret);
     */
    return ret;
}

static int show_catalog(bt_context_t *ctx, void *data, bt_param_t *param) 
{
    uint16_t	*rec = (uint16_t *)data;
    const char	*s = NULL;
    bt_catalog_param_t		*p = (bt_catalog_param_t *)(param->key);
    HFSPlusCatalogFolder	*pd = NULL;
    HFSPlusCatalogFile		*pf = NULL;
    HFSPlusCatalogThread	*pt = NULL;
    
    fprintf(stderr, "---- show catalog called\n");
    
    switch (be16(*rec)) {
        case kHFSPlusFolderRecord:
            pd = (HFSPlusCatalogFolder *)data;
            s = "Folder"; 
            break;
        case kHFSPlusFileRecord:
            pf = (HFSPlusCatalogFile *)data;	      
            s = "File"; 
            break;
        case kHFSPlusFolderThreadRecord:	
            pt = (HFSPlusCatalogThread *)data;
            s = "Folder thread"; 
            break;
        case kHFSPlusFileThreadRecord:
            pt = (HFSPlusCatalogThread *)data;
            s = "File thread"; 
            break;
        default:	
            fprintf(stderr, "Invalid record type\n");
            return 0;
    }
    
    if (pd) { 
        fprintf(stderr, "%s: folderID=%d flags=%04hX valence=%d\n", 
                s, be32(pd->folderID), be16(pd->flags), be32(pd->valence));
    }
    
    if (pf) { 
        fprintf(stderr, "%s: fileID=%d flags=%04hX ", s, be32(pf->fileID), be16(pf->flags));
        fprintf(stderr, "BSD owner %d:%d, flags U/%02hhX S/%02hhX mode:%04hX special:%d, FndrType:%d\n", 
                be32(pf->bsdInfo.ownerID), be32(pf->bsdInfo.groupID),
                pf->bsdInfo.ownerFlags, pf->bsdInfo.adminFlags, 
                be16(pf->bsdInfo.fileMode), be32(pf->bsdInfo.special.iNodeNum),
                be32(pf->userInfo.fdType));

        fs_file_t* file = (fs_file_t*) param->result;

        if (pf->dataFork.logicalSize) {
            fprintf(stderr, "-DATA- ");
            hfsp_print_fork_data(&pf->dataFork);
            fs_area_t* area = (fs_area_t*) malloc(sizeof(fs_area_t));
            if(area) {
                strncpy(area->tag, "DATA", sizeof(area->tag));
                area->extents = hfsp_build_extent_list(ctx->vol, be32(pf->fileID), 0, &pf->dataFork);
                area->next = file->areas;
                file->areas = area;
            }
        }
        
        if (pf->resourceFork.logicalSize) {
            fprintf(stderr, "-RSRC- ");
            hfsp_print_fork_data(&pf->resourceFork);
            fs_area_t* area = (fs_area_t*) malloc(sizeof(fs_area_t));
            if(area) {
                strncpy(area->tag, "RSRC", sizeof(area->tag));
                area->extents = hfsp_build_extent_list(ctx->vol, be32(pf->fileID), 0xFF, &pf->resourceFork);
                area->next = file->areas;
                file->areas = area;
            }
        }
        
        int64_t bmap = fs_bmap_offset(ctx->vol, ctx->ext, param->node * ctx->size);
        fprintf(stderr, "Records starts in node %u, at %hu, length %hu. Absolute offset: %lld\n", param->node, param->offset, param->length, bmap);
        
        
    }
    
    if (pt) {
        fprintf(stderr, "%s: parentID=%d ", s, be32(pt->parentID));
        print_unistr(&pt->nodeName);
        fprintf(stderr, "\n");
    }
    
    if (pt && 0 == p->stage) {
        fprintf(stderr, "\nThread found. Looking for File '");
        print_unistr(&pt->nodeName);
        fprintf(stderr, "' (%d) with parent(%d)\n", p->fileID, be32(pt->parentID));
        p->stage = 1;
        p->fileID = be32(pt->parentID);
        p->name.length = pt->nodeName.length;
        memmove(p->name.unicode, pt->nodeName.unicode, be16(pt->nodeName.length)*2);
        return 1;
    }
    return 1;
}

int hfsp_catalog_tree(fs_volume_t *vol, uint32_t fileID, fs_file_t* out_file) 
{
    fs_extent_t		*tree;
    hfsp_info_t		*info;
    HFSPlusForkData	*fork;
    int			ret;
    bt_context_t     	ctx;
    bt_param_t		idx;
    bt_catalog_param_t	key;
    
    assert(vol);
    assert(out_file);

    memset(&ctx, 0, sizeof(ctx));
    memset(out_file, 0, sizeof(*out_file));
    
    info = (hfsp_info_t *)vol->info;
    fork = &(info->head.catalogFile);
    
    fprintf(stderr, "Building extents for catalog file...\n");
    tree = hfsp_build_extent_list(vol, kHFSCatalogFileID, 0, fork);
    if (!tree) {
        return -1;
    }
    
    ctx.vol	= vol;
    ctx.ext	= tree;
    ctx.total	= be64(fork->logicalSize);
    
    
    key.fileID      = fileID;
    key.name.length = 0;
    key.stage       = 0;
    
    idx.height	= 0;
    idx.left 	= 0;
    idx.key     = &key;
    idx.compare	= compare_catalog_key;
    idx.leaf	= show_catalog;
    idx.result  = (void*) out_file;
    
    fprintf(stderr, "Processing the catalog file...\n");
    ret = hfsp_iterate_btree(&ctx, &idx);
    
    if (key.stage != 1) {
        fprintf(stderr, "Thread not found\n");
    } else {
        ret = hfsp_iterate_btree(&ctx, &idx);
    }
    fs_put_extent_list(tree);
    
    return ret;
}
