#include <sys/types.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <errno.h>

#include "fs_endian.h"
#include "hfsp.h"

typedef struct bt_extent_param {
    fs_extent_t		*ext;		// data fork extents list
    uint32_t		start;		// start block to look for    
    uint32_t		fileID;		// file catalog id
    uint32_t		total;		// total blocks expected
    uint8_t		fork;		// type: 0-data, FF-rsrc
} bt_extent_param_t;

static int compare_extent_key(bt_context_t *ctx, void *k1, void *k2) 
{
    HFSPlusExtentKey 	*key = (HFSPlusExtentKey *)k1;
    bt_extent_param_t	*p = (bt_extent_param_t *)k2;
    uint32_t		fileID = be32(key->fileID);
    uint32_t		start  = be32(key->startBlock);
    
    if (fileID > p->fileID) {
        return 1;
    } else if (fileID < p->fileID) {
        return -1;
    }
    
    if (key->forkType > p->fork) {
        return 1;
    } else if (key->forkType < p->fork) {
        return -1;
    }
    
    if (start > p->start) {
        return 1;
    } else if (start < p->start) {
        return -1;
    }
    
    return 0;
}

static int append_extents(bt_context_t *ctx, void *data, bt_param_t *param) 
{
    HFSPlusExtentDescriptor	*desc = (HFSPlusExtentDescriptor *)data;
    bt_extent_param_t		*p = (bt_extent_param_t *)(param->key);
    fs_extent_t	*tail, *ptr;
    int		i, scale;
    uint32_t	start, count;
    hfsp_info_t	*info;
    
    info = (hfsp_info_t *)(ctx->vol->info);
    scale = info->blksz;
    
    tail = p->ext;
    while (tail->next) {
        tail = tail->next;
    }
    
    fprintf(stderr, "Appending (%d): ", p->start);
    for (i = 0; i < 8; i++) {
        count = be32(desc[i].blockCount);
        if (!count) {
            break;
        }
        
        ptr = calloc(1, sizeof(fs_extent_t));
        if (!ptr) {
            return ENOMEM;
        }
        
        start = be32(desc[i].startBlock);
        fprintf(stderr, "%d:%d ", start, count);
        
        ptr->start = (uint64_t)start * scale;
        ptr->size  = (uint64_t)count * scale;
        ptr->next = NULL;
        tail->next = ptr;
        tail = ptr;
        
        p->start += count;
        
    }
    fprintf(stderr, "\n");
    return 0;
}

int hfsp_extents_tree(fs_volume_t *vol, uint32_t fileID, uint8_t type, uint32_t start, uint32_t total, fs_extent_t *ext)
{
    fs_extent_t         *tree;
    hfsp_info_t         *info;
    HFSPlusForkData     *fork;
    int                 ret;
    bt_context_t     	ctx;
    bt_param_t		idx;
    bt_extent_param_t	key;
    
    assert(vol);
    assert(ext);
    
    memset(&ctx, 0, sizeof(ctx));
    
    info = (hfsp_info_t *)vol->info;
    fork = &(info->head.extentsFile);
    
    fprintf(stderr, "Building extents for extent overflow file...\n");
    tree = hfsp_build_extent_list(vol, kHFSExtentsFileID, 0, fork);
    if (!tree) {
        return -1;
    }
    
    ctx.vol	= vol;
    ctx.ext	= tree;
    ctx.total	= be64(fork->logicalSize);
    
    key.fileID	= fileID;
    key.ext  	= ext;
    key.fork	= type;
    key.start	= start;
    key.total 	= total;
    
    idx.height	= 0;
    idx.left 	= 0;
    idx.key	= &key;
    idx.compare	= compare_extent_key;
    idx.leaf	= append_extents;
    
    fprintf(stderr, "Processing the extent overflow file...\n");
    ret = hfsp_iterate_btree(&ctx, &idx);
    
    fs_put_extent_list(tree);
    
    return ret;
}

fs_extent_t * hfsp_build_extent_list(fs_volume_t *vol, uint32_t fileID, uint8_t fork, HFSPlusForkData *p)
{
    int		i, scale, ret;
    hfsp_info_t	*info;
    fs_extent_t	*head, *tail, *ptr;
    uint32_t	start, count, pos, total;
    
    total = be32(p->totalBlocks);
    if (!total) {
        return NULL;
    }
    
    head = tail = NULL;
    info = (hfsp_info_t *)vol->info;
    pos  = 0;
    scale = info->blksz;
    
    for (i = 0; i < 8; i++) {
        count = be32(p->extents[i].blockCount);
        if (!count) {
            break;
        }
        
        ptr = calloc(1, sizeof(fs_extent_t));
        if (!ptr) {
            fs_put_extent_list(head);
            return NULL;
        }
        
        start = be32(p->extents[i].startBlock);
        
        ptr->start = (uint64_t)start * scale;
        ptr->size  = (uint64_t)count * scale;
        
        if (!head) {
            head = ptr;
            tail = ptr;
            ptr->next = NULL;
        } else {
            ptr->next = NULL;
            tail->next = ptr;
            tail = ptr;
        }
        pos += count;
    }
    
    if (pos < total) {
        fprintf(stderr, "Fragmented file: %d of %d blocks added, proceed with the extents file...\n", pos, total);
        ret = hfsp_extents_tree(vol, fileID, fork, pos, total, head);
        if (ret) {
            fs_put_extent_list(head);
            return NULL;
        }
    }
    ptr = head;
    count = 0;
    while (ptr) {
        count += (uint32_t)(ptr->size/scale);
        ptr = ptr->next;
    }
    if (count != total) {
        fprintf(stderr, "ALARM!!! File metadata mismatch. Expected %u, got %u %d-blocks\n", total, count, scale);
    }
    return head;
}

