#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include "fs_endian.h"
#include "hfsp.h"

typedef int (*data_actor_f)(char *buf, int sz, uint64_t pos, void *param);

typedef struct freespace_context {
    uint64_t	total;			// total allocation blocks
    uint64_t	start;			// start of current allocation block
    fs_extent_t	*head;			// [out] extent list head
    fs_extent_t	*tail;			// [out] extent list tail
    int		error;
    int		blksz;
    uint64_t	offset;			// volume data offset in bytes
} freespace_context_t;

static inline int hfsp_add_extent(freespace_context_t *ctx, uint64_t blk) 
{
    fs_extent_t		*p;
    
    //    fprintf(stderr, "%lld[%llx] (%lld)\n", blk - 1, blk - 1, blk - ctx->start);
    p = malloc(sizeof(fs_extent_t));
    if (!p) {
        return ENOMEM;
    }
    p->start = ctx->start * ctx->blksz + ctx->offset;
    p->size  = (blk - ctx->start) * ctx->blksz + ctx->offset;
    p->next  = NULL;
    
    if (ctx->head) {
        ctx->tail->next = p;
    } else {
        ctx->head = p;
    }
    ctx->tail = p;
    ctx->start = 0;
    return 0;
}

int	hfsp_process_bitmap(char *buf, int sz, uint64_t pos, void *param)
{
    int			i, x, bit,ret;
    uint64_t 		blk;
    freespace_context_t	*ctx;
    
    ctx = (freespace_context_t *)param;
    for (i = 0; i < sz; i++) {
        blk = (pos + i)*8;
        
        // a little optimization :)
        if (ctx->total - blk > 8) {
            if (0 == buf[i] && ctx->start) {
                continue;
            }
            if (0xFF == (unsigned char)buf[i] && !ctx->start) {
                continue;
            }
        }
        
        for (x = 0; x < 8; x++, blk++) {
            // end of bitmap condition
            if (blk >= ctx->total) {
                ret = ctx->start ? hfsp_add_extent(ctx, blk) : 0;
                return ret;
            }
            
            bit = buf[i] & (1 << (7-x)) ? 1 : 0;
            if (ctx->start) {
                if (bit) {
                    ret = hfsp_add_extent(ctx, blk);
                    if (ret) {
                        return ret;
                    }
                }
            } else {
                if (!bit) {
                    ctx->start = blk;
                    //		    fprintf(stderr, "extent: %lld[%llx]-", blk, blk);
                }
            }
        }
    }
    return 0;
}

/*
 
 int	hfsp_process_bitmap_optimized(char *buf, int sz, uint64_t pos, void *param)
 {
 int			i, j, x, bit, ret, s;
 uint64_t 		blk;
 uint32_t		 *p;
 freespace_context_t	*ctx;
 
 ctx = (freespace_context_t *)param;
 
 s = sz - (sz & 0x3);
 
 for (i = 0; i < s; i += 4) {
 blk = (pos + i)*8;
 p = (uint32_t *)(buf+i);
 
 if (0 == *p && ctx->start) {
 continue;
 }
 if (0xFFFFFFFF == *p && !ctx->start) {
 continue;
 }
 for (j = 0; j < 4; j++) {
 for (x = 0; x < 8; x++, blk++) {
 // end of bitmap condition
 if (blk >= ctx->total) {
 ret = ctx->start ? hfsp_add_extent(ctx, blk) : 0;
 return ret;
 }
 
 bit = buf[i] & (1 << (7-x)) ? 1 : 0;
 if (ctx->start) {
 if (bit) {
 ret = hfsp_add_extent(ctx, blk);
 if (ret) {
 return ret;
 }
 }
 } else {
 if (!bit) {
 ctx->start = blk;
 //		    fprintf(stderr, "extent: %lld[%llx]-", blk, blk);
 }
 }
 }
 }
 }
 
 // the remainder
 for (; i < sz; i++) {
 blk = (pos + i)*8;
 
 // a little optimization :)
 if (ctx->total - blk > 8) {
 if (0 == buf[i] && ctx->start) {
 continue;
 }
 if (0xFF == (unsigned char)buf[i] && !ctx->start) {
 continue;
 }
 }
 
 for (x = 0; x < 8; x++, blk++) {
 // end of bitmap condition
 if (blk >= ctx->total) {
 ret = ctx->start ? hfsp_add_extent(ctx, blk) : 0;
 return ret;
 }
 
 bit = buf[i] & (1 << (7-x)) ? 1 : 0;
 if (ctx->start) {
 if (bit) {
 ret = hfsp_add_extent(ctx, blk);
 if (ret) {
 return ret;
 }
 }
 } else {
 if (!bit) {
 ctx->start = blk;
 //		    fprintf(stderr, "extent: %lld[%llx]-", blk, blk);
 }
 }
 }
 }
 
 return 0;
 }
 */

int	hfsp_iterate_extent_list(fs_volume_t *vol, fs_extent_t *ext, data_actor_f actor, void *param)
{
    fs_extent_t	*p;
    uint64_t	pos, total, ext_sz, ext_pos;
    int		sz, ret;
    char	*buf;
    
    buf = calloc(1, HFSP_BUFSIZE);
    if (!buf) {
        return ENOMEM;
    }
    
    p = ext;
    total = 0;
    ret = 0;
    while (p) {
        ext_sz = p->size;	// current extent size
        ext_pos = 0;		// position in the current extent
        while (ext_sz) {
            sz = ext_sz > HFSP_BUFSIZE ? HFSP_BUFSIZE : ext_sz;
            pos = p->start + ext_pos;
            ret = fs_read_data(vol, buf, pos, sz);
            if (sz != ret) {	// read failed
                ret = errno;
                break;
            }
            sz = ret;
            //	    fprintf(stderr, "Calling actor with %lld [%llX], ext_pos=%lld [%llX]\n", total, total, ext_pos, ext_pos);
            ret = actor(buf, sz, total, param);
            if (ret) {		// actor failed
                break;
            }
            
            ext_pos += sz;
            ext_sz  -= sz;
            total   += sz;
        }
        p = p->next;
    }
    
    free(buf);
    return ret;
}

fs_extent_t *	hfsp_get_free_space(fs_volume_t *vol)
{
    fs_extent_t		*bmap;
    hfsp_info_t		*info;
    HFSPlusForkData	*fork;
    freespace_context_t	ctx;
    int			ret;
    
    memset(&ctx, 0, sizeof(ctx));
    
    info = (hfsp_info_t *)vol->info;
    fork = &(info->head.allocationFile);
    bmap = hfsp_build_extent_list(vol, kHFSAllocationFileID, 0, fork);
    if (!bmap) {
        return NULL;
    }
    
    ctx.total = be32(info->head.totalBlocks);
    ctx.blksz = info->blksz;
    ctx.offset = vol->offset;
	
    ret = hfsp_iterate_extent_list(vol, bmap, hfsp_process_bitmap, &ctx);
    //    ret = hfsp_iterate_extent_list(vol, bmap, hfsp_process_bitmap_optimized, &ctx);
    
    fs_put_extent_list(bmap);
    if (ret) {
        if (ctx.head) {
            fs_put_extent_list(ctx.head);
            ctx.head = NULL;
        }
    }
    
    return ctx.head;
}

