#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdint.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <time.h>

#include "fs_endian.h"
#include "hfsp.h"
#include "hfsp_volume.h"

int hfsp_catalog_wrapper(fs_volume_t *v, uint32_t id, fs_file_t *out_file)
{
    int ret = (1 != hfsp_catalog_tree(v, id, out_file));
    hfsp_attrib_tree(v, id, out_file);
    return ret;
}

void hfsp_cleanup_volume(fs_volume_t *vol)
{
    hfsp_info_t		*info = (hfsp_info_t *)vol->info;
    
    if (info) {
        iconv_close(info->hfs2utf);
        free(info);
        vol->info = NULL;
    }
}

int hfsp_check_volume_header(fs_volume_t *vol) 
{
    HFSPlusVolumeHeader	*h;
    hfsp_info_t		*info;
    
    h = fs_get_data(vol, 1024, 1024);
    if (!h) {
        return ENOMEM;
    }
    
    if (kHFSSigWord == be16(h->signature)) {
        HFSMasterDirectoryBlock *m = (HFSMasterDirectoryBlock *)h;
        int	s, b, o;
        
        fprintf(stderr, "HFS Wrapper detected with %04hX signature\n", be16(m->drEmbedSigWord));
        
        switch (be16(m->drEmbedSigWord)) {
            case kHFSPlusSigWord:
            case kHFSXSigWord:
                break;
            default:
                fprintf(stderr, "HFS-only volumes are not supported\n");
                fs_put_data(h);
                return -1;
        }
        
        // TODO:  be16(drEmbedExtent.blockCount);
        s = be16(m->drEmbedExtent.startBlock);
        b = be32(m->drAlBlkSiz);
        o = be16(m->drAlBlSt);
        
        vol->offset += s*b + 512*o;
	    
        fs_put_data(h);
        h = fs_get_data(vol, 1024, 1024);
        if (!h) {
            return ENOMEM;
        }
    }
    
    switch(be16(h->signature)) {
        case kHFSSigWord:
            fprintf(stderr, "HFS volumes are not supported\n");
            fs_put_data(h);
            return -1;
        case kHFSPlusSigWord:
            break;
        case kHFSXSigWord:
            break;
        default:
            fprintf(stderr, "unknown volume signature '%04hX'\n", be16(h->signature));
            fs_put_data(h);
            return -1;
    }
    // check backup volume header also
    
    //hfsp_print_volume_header(h);
    
    // the volume is ok
    info = calloc(1, sizeof(hfsp_info_t));
    if (!info) {
        fs_put_data(h);
        return ENOMEM;
    }
    
    info->head = *h;
    vol->block_size = info->blksz = be32(h->blockSize);
    info->hfs2utf = iconv_open("UTF-8", "UCS-2BE");
    
    vol->info = info;
    vol->get_free_space = hfsp_get_free_space;
    vol->get_file_by_id = hfsp_catalog_wrapper;
    vol->cleanup	= hfsp_cleanup_volume;
    vol->display = hfsp_display_volume;
    vol->cleanup = hfsp_free_volume;
	
    fs_put_data(h);
    
    
    return 0;
}

int hfsp_init_volume(fs_volume_t *vol) 
{
    int		ret;
    
    ret = hfsp_check_volume_header(vol);
    
    return ret;
}

void hfsp_display_volume(fs_volume_t* vol)
{
	hfs_volume_t* hfs = (hfs_volume_t*) vol->info;
	hfsp_print_volume_header(&hfs->head);
}

void hfsp_free_volume(fs_volume_t* vol) {
	free(vol->info);
}


