#include "BCKernel.h"
#include "kmod/bc_ctl.h"
#include "../container/user.h"
#include "../spacemap/spacemap.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <unistd.h>
#include <glob.h>
#include <fcntl.h>
#include <limits.h>
#include <time.h>
#include <assert.h>
#include <dirent.h>

// Do not use this macro with function, it will be called twise! Use only with variables.
#define CHECK(err) if (err) { DEBUG_MSG("Failed at line %i.", __LINE__); return err; }
#define SUCCESS(res) (res == kBCKernel_Success)

/**
* Gets path to sysfs depending from kernel version changes.
* @return "/sys/class/misc/", if directory exists, otherwise returns "/proc/" 
*    without checking directory existance. Caller shouldn't free return value.
*/
static const char* GetSysFSPath() 
{
    const char *right_path = "/sys/class/misc/";
    const char *old_path = "/proc/";

    struct stat st;
    int ret;
    
    ret = stat(right_path, &st);
    if (0 == ret && S_ISDIR(st.st_mode))
    {
        return right_path;
    }
    else
    {
        return old_path;
    }
}

static int ReadFromFile(const char *path, const char *format, void *dest)
{
    int error = 0;
    FILE *fd = fopen(path, "r");
    if (NULL == fd)
        return -errno;
    if (1 != fscanf(fd, format, dest))
    {
        error = -errno;
        DEBUG_MSG("read_from_file error: %s\n", strerror(-error));
    }
    fclose(fd);
    return error;
}

static bc_kern_result GetErrorResult(int error_code)
{
    if (-ENOMEM == error_code)
        return kBCKernel_NoMemory;
    if (!error_code)
        return kBCKernel_Success;    
    DEBUG_MSG("Internal Error: %s\n", strerror(-error_code));
    errno = -error_code;
    return kBCKernel_InternalError;
}

static bc_kern_result CheckMode(bc_mode_id mode)
{
    if (kBCMode_CBC != mode
        && kBCMode_LRW != mode 
        && kBCMode_XTS != mode)
        return kBCKernel_UnsupportedMode;
    return kBCKernel_Success;
}

static bc_kern_result CheckProcType(bc_proc_type type)
{
    if (kBCProcess_Decrypt != type
        && kBCProcess_Encrypt != type)
        return kBCKernel_UnsupportedMode;
    return kBCKernel_Success;
}

static int GetAlgByDir(const char *alg_dir, BCAlgInfo* out_alg)
{
    assert(NULL != out_alg);
    
    char   path_buffer[PATH_MAX] = {0};
    int    error = 0;
    
    snprintf(path_buffer, PATH_MAX - 1, "%s/id", alg_dir);
    error = ReadFromFile(path_buffer, "%i", &(out_alg->id));
    CHECK(error);
    
    snprintf(path_buffer, PATH_MAX - 1, "%s/name", alg_dir);
    error = ReadFromFile(path_buffer, "%s", &(out_alg->name[0]));
    CHECK(error);
    
    snprintf(path_buffer, PATH_MAX - 1, "%s/key_length", alg_dir);
    error = ReadFromFile(path_buffer, "%lu", &(out_alg->key_length));
    CHECK(error);

    snprintf(path_buffer, PATH_MAX - 1, "%s/block_size", alg_dir);
    error = ReadFromFile(path_buffer, "%lu", &(out_alg->block_size));
    CHECK(error);
    
    snprintf(path_buffer, PATH_MAX - 1, "%s/cbc_key_size", alg_dir);
    error = ReadFromFile(path_buffer, "%lu", &(out_alg->cbc_key_size));
    CHECK(error);
    
    snprintf(path_buffer, PATH_MAX - 1, "%s/lrw_key_size", alg_dir);
    error = ReadFromFile(path_buffer, "%lu", &(out_alg->lrw_key_size));
    CHECK(error);
    
    snprintf(path_buffer, PATH_MAX - 1, "%s/xts_key_size", alg_dir);
    error = ReadFromFile(path_buffer, "%lu", &(out_alg->xts_key_size));
    CHECK(error);

    return error;
}

bc_kern_result BCKernelIsInitialized(void)
{
    int error = 0;
    struct stat st;
    const char *bc_dev_path = "/dev/"BC_CTL_NODE_NAME;

    if((0 != stat(bc_dev_path, &st)) || (!S_ISCHR(st.st_mode)))
    {
        DEBUG_MSG("Stat command failed.\n");
        return kBCKernel_NotInitialized;
    }
    
    error = access(bc_dev_path, R_OK);
    if (error)
    {
        DEBUG_MSG("Access command failed on device entry.\n");
        return kBCKernel_NotInitialized;
    }
    
    const char *sysfs_path = GetSysFSPath();
    char sysfs_entry_path[PATH_MAX] = {0};
    snprintf(sysfs_entry_path, PATH_MAX - 1, "%s/%s", sysfs_path, BC_CTL_NODE_NAME);
    
    error = lstat(sysfs_entry_path, &st);
    if (error)
    {
        DEBUG_MSG("lStat command failed.\n");
        return kBCKernel_NotInitialized;
    }
        
    if (!S_ISLNK(st.st_mode))
    {
        DEBUG_MSG("Entry is not link.\n");
        return kBCKernel_NotInitialized;
    }
    
    if (access(sysfs_entry_path, X_OK))
    {
        DEBUG_MSG("Access command failed on sys-fs entry.\n");
        return kBCKernel_NotInitialized;
    }
    return kBCKernel_Success;
}

bc_kern_result BCKernelLookupAlgId(bc_alg_id id, BCAlgInfo* out_alg)
{
    // TODO: copy-pasted from lookup_alg_name with condition changing only.
    if (NULL == out_alg)
        return kBCKernel_InvalidBuffer;

    struct dirent *entry;
    DIR *dir = NULL;
    bc_kern_result result;
    const char *sysfs_path = GetSysFSPath();
    const char *plugins_path = BC_CTL_NODE_NAME"/"BC_CTL_PLUGINS_GROUP_STR"/";
    char alg_root_path[PATH_MAX];
    char path_buffer[PATH_MAX];
    int alg_tmp_id = 0;
    int error = 0;
    
    result = BCKernelIsInitialized();
    if (!SUCCESS(result))
    {
        DEBUG_MSG("Kernel is not initialized.\n");
        return result;
    }
    
    snprintf(alg_root_path, PATH_MAX - 1, "%s/%s", sysfs_path, plugins_path);
    
    dir = opendir(alg_root_path);
    if (NULL == dir)
    {
        return kBCKernel_InternalError;
    }
    
    result = kBCKernel_UnknownAlg;
    entry = readdir(dir);
    
    if (entry == NULL)
        error = -errno;
        
    do
    {
        if ((DT_DIR != entry->d_type) || (!strcmp(entry->d_name, ".")) || (!strcmp(entry->d_name, "..")))
        {
            DEBUG_MSG("Skipping entry %s\n", entry->d_name);
            continue;
        }
        snprintf(path_buffer, PATH_MAX - 1, "%s/%s/id", alg_root_path, entry->d_name);
        error = ReadFromFile(path_buffer, "%i", &alg_tmp_id);
        if (error)
            break;
        DEBUG_MSG("Found alg with Id = %i\n", alg_tmp_id);
        if (alg_tmp_id == (int)id)
        {
            snprintf(path_buffer, PATH_MAX - 1, "%s/%s/", alg_root_path, entry->d_name);
            error = GetAlgByDir(path_buffer, out_alg);
            if (error)
                break;
            result = kBCKernel_Success;
            break;
        }
        

    } 
    while (NULL != (entry = readdir(dir)));
    
    closedir(dir);
    
    if (error)
        return GetErrorResult(error);
    
    return result;
}

bc_kern_result BCKernelLookupAlgName(const char* name, BCAlgInfo* out_alg)
{
    if (NULL == out_alg)
        return kBCKernel_InvalidBuffer;

    struct dirent *entry;
    DIR *dir = NULL;
    bc_kern_result result;
    const char *sysfs_path = GetSysFSPath();
    const char *plugins_path = BC_CTL_NODE_NAME"/"BC_CTL_PLUGINS_GROUP_STR"/";
    char alg_root_path[PATH_MAX];
    char path_buffer[PATH_MAX];
    char alg_tmp_name[kBCKernel_AlgNameSize];
    int error = 0;
    
    result = BCKernelIsInitialized();
    if (!SUCCESS(result))
        return result;
    
    snprintf(alg_root_path, PATH_MAX - 1, "%s/%s", sysfs_path, plugins_path);
    
    dir = opendir(alg_root_path);
    if (NULL == dir)
    {
        return kBCKernel_InternalError;
    }
    
    result = kBCKernel_UnknownAlg;
    entry = readdir(dir);
    
    if (entry == NULL)
        error = -errno;
        
    do
    {
        if ((DT_DIR != entry->d_type) || (!strcmp(entry->d_name, ".")) || (!strcmp(entry->d_name, "..")))
        {
            DEBUG_MSG("Skipping entry %s\n", entry->d_name);
            continue;
        }
        snprintf(path_buffer, PATH_MAX - 1, "%s/%s/name", alg_root_path, entry->d_name);
        DEBUG_MSG("Reading name from %s...\n", path_buffer);
        error = ReadFromFile(path_buffer, "%s", (&alg_tmp_name[0]));        
        if (error)
            break;
        DEBUG_MSG("Found alg with name %s\n", alg_tmp_name);
        if (0 == strncasecmp(name, alg_tmp_name, kBCKernel_AlgNameSize))
        {
            snprintf(path_buffer, PATH_MAX - 1, "%s/%s/", alg_root_path, entry->d_name);
            error = GetAlgByDir(path_buffer, out_alg);
            if (error)
                break;
            result = kBCKernel_Success;
            break;
        }
        

    } 
    while (NULL != (entry = readdir(dir)));
    
    closedir(dir);
    
    if (error)
        return GetErrorResult(error);
    
    return result;
}

bc_kern_result BCKernelListAlgs(BCAlgInfo** out_palgs, unsigned* out_total)
{
    // Iterates thru sysfs entries and loads all necessary info
    // about each of the algs loaded.

    if (out_palgs == NULL || out_total == NULL)
        return kBCKernel_InvalidBuffer;
    int error = 0;
    unsigned i = 0;
    const char *sysfs_path = GetSysFSPath();
    const char *plugins_path = BC_CTL_NODE_NAME"/"BC_CTL_PLUGINS_GROUP_STR"/";
    char alg_search_pattern[PATH_MAX];
    glob_t gr; // list of alg dirs will be placed here.
    BCAlgInfo *algs = NULL;
    bc_kern_result result;

    result = BCKernelIsInitialized();
    if (!SUCCESS(result))
        return result;
    snprintf(alg_search_pattern, PATH_MAX - 1, "%s/%s/*", sysfs_path, plugins_path);

    // TODO: Here is supposed that there are only algs folders in plugins folder.
    error = glob(alg_search_pattern, GLOB_ERR | GLOB_MARK | GLOB_NOSORT, NULL, &gr);
    if (error)
    {
        error = -errno;
        if (-ENOENT == error)
        {
            globfree(&gr);
            (*out_palgs) = NULL;
            (*out_total) = 0;
            return kBCKernel_Success;
        }
        goto LIST_ALGS_CLEANUP;
    }

    algs = (BCAlgInfo*)malloc(gr.gl_pathc * sizeof(BCAlgInfo));
    if (NULL == algs)
    {
        error = -ENOMEM;
        goto LIST_ALGS_CLEANUP;
    }

    for (i=0; i < gr.gl_pathc; i++)
    {
        error = GetAlgByDir(gr.gl_pathv[i], &(algs[i]));
        if (error)
        {
            free(algs);
            goto LIST_ALGS_CLEANUP;
        }
    }
    (*out_palgs) = algs;
    (*out_total) = gr.gl_pathc;

LIST_ALGS_CLEANUP:
    globfree(&gr);
    return GetErrorResult(error);
}


void BCKernelFreeAlgList(BCAlgInfo* list)
{
    free(list);
}

static int OpenBCDev(int *dev_d)
{
    const char *bc_dev = "/dev/"BC_CTL_NODE_NAME;
    DEBUG_MSG("Trying to open device '%s'\n", bc_dev);
    int tmp_d = open(bc_dev, O_RDONLY);
    if (-1 == tmp_d)
    {
        int error = -errno;
        fprintf(stderr, "Error opening device: %s\n", strerror(-error));
        return error;
    }
    DEBUG_MSG("Opening successful. Descriptor value is %i\n", tmp_d);
    (*dev_d) = tmp_d;
    return 0;
}

static void CloseBCDev(int dev_d)
{
    if (close(dev_d))
    {
        ERROR_MSG("Device closing failed: %s\n", strerror(-errno));
    }
}

static int BCIoctl(int command, unsigned long arg)
{
    int descriptor = 0;
    int error = 0;
    int ioctl_return = 0;
    
    error = OpenBCDev(&descriptor);
    if (error)
    {
        errno = -error;
        return -1;
    }
    DEBUG_MSG("Calling ioctl with command %x\n", command);
    if (-1 == (ioctl_return = ioctl(descriptor, command, arg)))
    {
        error = -errno;
        fprintf(stderr, "Call to ioctl failed: %s\n", strerror(-error));
    }
    
    CloseBCDev(descriptor);
    
    if (error)
        errno = -error;
    
    return ioctl_return;
}

bc_kern_result BCKernelStoreKey(const char* key, size_t key_size, bc_key_handle* out_handle)
{
    if (NULL == key || 0 == key_size)
        return kBCKernel_InvalidKey;
    if (NULL == out_handle)
        return kBCKernel_InvalidBuffer;

    int               error = 0;
    struct bc_ctl_key ctl_key;
    bc_kern_result    result;

    memset(&ctl_key, 0, sizeof(struct bc_ctl_key));

    result = BCKernelIsInitialized();
    if (!SUCCESS(result))
        return result;

    ctl_key.key.data.value = (char *)key;
    ctl_key.key.size = key_size;
    ctl_key.out_handle.value = out_handle;

    if (-1 == BCIoctl(kBCCTL_StoreKey, (unsigned long)(&ctl_key)))
    {
        error = -errno;
        if (-EBUSY == error)
            return kBCKernel_NoMemory;
        if (-EINVAL == error)
            return kBCKernel_InvalidKey;
    }

    return GetErrorResult(error);
}

bc_kern_result BCKernelFreeKey(bc_key_handle handle)
{
    int            error = 0;
    bc_kern_result result;
    
    result = BCKernelIsInitialized();
    if (!SUCCESS(result))
        return result;

    if (-1 == BCIoctl(kBCCTL_FreeKey, (unsigned long)handle))
    {
        error = -errno;
        if (-EINVAL == error)
            return kBCKernel_InvalidKey;
    }
    
    return GetErrorResult(error);
}

/*
 * Checks minimum key size for specified alg and mode.
 * */
static bc_kern_result CheckKeySize(BCAlgInfo *alg, bc_mode_id mode, const char */*key*/, size_t key_size)
{
    size_t mode_key_size = BCKernelGetModeKeySize(alg, mode);
    if (mode_key_size > key_size)
        return kBCKernel_InvalidKey;
    return kBCKernel_Success;
}

static bc_kern_result ProcessData(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc, bc_key_handle key_handle, const char *key, size_t key_size, uint64_t iv, const char *pool, char *data, size_t nbytes)
{
    bc_kern_result        result;
    struct bc_ctl_process ctl;
    BCAlgInfo                tmp_alg;
    
    memset(&ctl, 0, sizeof(ctl));
        
    result = BCKernelIsInitialized();
    CHECK(result);
    
    result = BCKernelLookupAlgId(alg_id, &tmp_alg);
    CHECK(result);
    
    if (!BCKernelIsModeSupported(&tmp_alg, mode_id)) {
        return kBCKernel_UnsupportedMode;
    }

    if (NULL == data && 0 == nbytes)
    {
        return kBCKernel_Success;
    }
    
    if (NULL == data)
    {
        DEBUG_MSG("Data buffer is null.\n");
        return kBCKernel_InvalidBuffer;
    }
    
    // Data size must be multiple of alg's block_size.
    if (nbytes % tmp_alg.block_size)
    {
        DEBUG_MSG("Data size is not multiple of block size.\n");
        return kBCKernel_InvalidBuffer;
    }

    result = CheckMode(mode_id);
    CHECK(result);
    result = CheckProcType(proc);
    CHECK(result);
    result = BCKernelIsInitialized();
    CHECK(result);
    
    ctl.data.value = data;
    ctl.nbytes = nbytes;
    ctl.alg_id = alg_id;
    ctl.mode_id = mode_id;
    if (NULL == key)
    {
        ctl.has_key = 0;
        ctl.key.handle = key_handle;
    }
    else
    {
        result = CheckKeySize(&tmp_alg, mode_id, key, key_size);
        CHECK(result);
        assert(0 != key_size);
        ctl.has_key = 1;
        ctl.key.inlined.data.value = (char*)key;
        ctl.key.inlined.size = key_size;
    }
    ctl.iv = iv;
    ctl.pool.value = (char*)pool;
    ctl.proc_type = proc;

    if (-1 == BCIoctl(kBCCTL_Process, (unsigned long)(&ctl)))
    {
        int error = -errno;
        return GetErrorResult(error);
    }
    
    return kBCKernel_Success;
}

bc_kern_result BCKernelProcess(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc, 
                    bc_key_handle key, uint64_t iv, const char* pool, 
                    char* data, size_t nbytes)
{
    return ProcessData(alg_id, mode_id, proc, key, NULL, 0, iv, pool, data, nbytes);
}

bc_kern_result BCKernelProcessWithKey(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc,
                            const char* key, size_t key_size,
                            uint64_t iv, const char* pool,
                            char* data, size_t nbytes)
{
    if (NULL == key || 0 >= key_size)
        return kBCKernel_InvalidKey;
    return ProcessData(alg_id, mode_id, proc, -1, key, key_size, iv, pool, data, nbytes);
}


/// Checks that device is appeared.
static bool IsDeviceReady(char *dev_name)
{
    struct stat st;
    if (!stat(dev_name, &st)) {
        return S_ISBLK(st.st_mode);
    }
    return false;
}


bc_kern_result BCKernelSetupDisk(const char* file, int alg, int mode, int key_handle, const char* pool,
                                 SpaceMap *map, int options, char* out_devname)
{
    int fd = -1;
    int numtries = 0;
    Extent *extents = NULL;
    bc_u64 extentCount = 0;

    DEBUG_MSG("Setting up a new disk for file '%s'\n", file);

    if(kBCKernel_Success != BCKernelIsInitialized())
    {
        return kBCKernel_NotInitialized;
    }

    if (!map)
    {
        return kBCKernel_InvalidArgument;
    }

    // Creating an extent list from the space map.
    sm_result smres = SpaceMapToExtentArray(map, &extents, &extentCount);
#ifdef BC_DEBUG
    DEBUG_MSG("Space map for file:\n");
    SpaceMapPrint(map);
    DEBUG_MSG("After conversion have %llu extents in array.\n", (unsigned long long)extentCount);
#endif
    if (kSpaceMap_Success != smres)
    {
        if (kSpaceMap_InvalidArg == smres)
            return kBCKernel_InvalidArgument;

        if (kSpaceMap_NoMem == smres)
            return kBCKernel_NoMemory;

        return kBCKernel_InternalError;
    }

    struct bc_ctl_disk ctl;
    char deviceName[BC_MAX_PATH] = { 0 };
    ctl.out_name = deviceName;
    ctl.name_bufsize = BC_MAX_PATH;
    ctl.out_minor = 0;

    if (-1 == BCIoctl(kBCCTL_CreateDisk, (unsigned long)&ctl))
    {
        return GetErrorResult(-errno);
    }

    // Form a full path to the device.
    snprintf(out_devname, BC_MAX_PATH - 1, "/dev/%s", deviceName);
    DEBUG_MSG("Waiting for %s to appear...\n", out_devname);

    // Wait some time while the device appears.
    bool result;
    while (!(result = IsDeviceReady(out_devname)))
    {
        sleep(1);
        if (++numtries > 20)
            break;
    }

    if (!result)
    {
        fprintf(stderr, "Device %s seems to hang.\n", out_devname);
        return kBCKernel_InternalError;
    }

    // Send set fd ioctl to appeared device.

    fd = open(file, (options & kBCDisk_ReadOnly ? O_RDONLY : O_RDWR));
    if (fd < 0)
    {
        fprintf(stderr, "Can't open container file: %s\n", strerror(errno));
        return kBCKernel_InternalError;
    }

    struct bc_diskctl_set_fd diskctl;
    memset(&diskctl, 0, sizeof(diskctl));

    diskctl.fd              = fd;
    diskctl.alg_id          = alg;
    diskctl.mode_id         = mode;
    diskctl.key_handle      = key_handle;
    diskctl.pool.value      = (char*)pool;
    diskctl.extents.value   = extents;
    diskctl.extent_count    = extentCount;
    diskctl.options         = options;
#ifdef BC_DEBUG
    fprintf(stderr, "Sending %lu bytes (%lu extents) of extents as IO control.\n",
            (unsigned long)diskctl.extent_count * sizeof(Extent), (unsigned long)diskctl.extent_count);
    for (bc_u64 i = 0; i < extentCount; i++)
    {
        fprintf(stderr, " Start: %lu, length: %lu\n", (unsigned long)extents[i].start,
                (unsigned long)extents[i].length);
    }

    SpaceMapPrint(map);
#endif
    numtries = 0;
    int devfd = -1;

    do
    {
        devfd = open(out_devname, O_RDONLY);
        if (devfd < 0)
        {
            // If we have no permissions, maybe they will not have been set yet, and we need to wait.
            if (((errno == EPERM) || (errno == EACCES))
                    && (++numtries < 20))
            {
                sleep(1);
                continue;
            }

            // We are here in case of a error.
            close(fd);
            close(devfd);
            fprintf(stderr, "Can't open device file: %s. Number of times: %i\n", strerror(errno), numtries);
            return kBCKernel_InternalError;
        }
        break;
    }
    while (1);
    long rc = ioctl(devfd, kBCCTL_DiskSetFd, &diskctl);
    int error = -errno;
    close(fd);
    close(devfd);

    if (-1 == rc)
    {
        fprintf(stderr, "Error returned from kernel module. See `dmesg` for details.\n");
    }

    SpaceMapFreeArray(&extents);
    return (-1 == rc) ? GetErrorResult(error) : kBCKernel_Success;
}


bc_kern_result BCKernelRemoveDisk(const char* devname, int force)
{
    bc_kern_result result = BCKernelIsInitialized();
    CHECK(result);

    int fd = open(devname, O_RDONLY);
    if(fd < 0) {
        return kBCKernel_InternalError;
    }

    int numtries = 0;

    do
    {
        long rc = ioctl(fd, kBCCTL_DiskClearFd, (unsigned long) force);
        int error = -errno;
        // If device is locked, wait for some time to unlock before returning an error.
        if ((rc == -1) && (error == -EBUSY) && (numtries < 5))
        {
            numtries++;
            sleep(1);
            continue;
        }
        close(fd);

        return (rc == -1) ? GetErrorResult(error) : kBCKernel_Success;
    }
    while (true);

    // We can't be here.
    return kBCKernel_InternalError;
}

bc_kern_result BCKernelBenchmark(bc_alg_id alg_id, bc_mode_id mode_id, unsigned long *out_nanoseconds)
{
    BCAlgInfo alg;

    bc_kern_result result;
    result = BCKernelIsInitialized();
    CHECK(result);

    result = BCKernelLookupAlgId(alg_id, &alg);
    CHECK(result);

    if (!BCKernelIsModeSupported(&alg, mode_id))
        return kBCKernel_UnsupportedMode;

    struct bc_ctl_bench bench;
    bench.alg_id = alg_id;
    bench.alg_mode = mode_id;

    if (-1 == BCIoctl(kBCCTL_Benchmark, (unsigned long)&bench))
    {
        int error = -errno;
        return GetErrorResult(error);
    }
    (*out_nanoseconds) = bench.nsec;
    return kBCKernel_Success;
}
