/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 */
/**
 * @mainpage BestCrypt Kernel library interface.
 * This describes BestCrypt Kernel library interface.
 */
/** 
 * @file BCKernel.h
 * @brief BCKernel library interace.
 * This library provides a thin wrapper over bc kernel services.
 * It allows user space client to communicate with kernel implementations.
 *
 * This file is included both in user and kernel spaces on different platforms.
 */

#ifndef BC_LIB_KERNEL_H
#define BC_LIB_KERNEL_H

#ifdef BC_KERNEL
# include <linux/types.h>
#else
# include <stdint.h>
#endif

extern const char *PROG_NAME;

/// Kernel error codes.
typedef enum {
    kBCKernel_Success = 0,      ///< Operation successful.
    kBCKernel_NotInitialized,   ///< BC kernel is not initialized.
    kBCKernel_NoMemory,         ///< Error allocating memory.
    kBCKernel_UnknownAlg,       ///< Specified alg cannot be found.
    kBCKernel_UnsupportedMode,  ///< Alg doesn't supports specified mode.
    kBCKernel_InvalidKey,       ///< Invalid key specified.
    kBCKernel_InvalidBuffer,    ///< Invalid data buffer specified.
    kBCKernel_InvalidArgument,  ///
    kBCKernel_DeviceBusy,       ///
    kBCKernel_InternalError     ///< Unknown internal error. In this case usually errno is set and may be checked.
} bc_kern_result;


/// Data processing mode.
typedef enum {
    kBCProcess_Decrypt = 0,     ///< Decrypt processing mode.
    kBCProcess_Encrypt = 1     ///< Encrypt processing mode.
} bc_proc_type;


/// Block cipher modes.
enum {
    kBCMode_CBC = (int) 0xBC000002, ///< CBC (Cipher Block chaining) mode.
    kBCMode_LRW = (int) 0xBC000001, ///< LRW (Liskov, Rivest, Wagner) mode.
    kBCMode_XTS = (int) 0xBC000004 ///< XTS (XEX-based Tweaked Codebook) mode.
};


/// Misc constants. 
/// @warning Keep in sync!
enum {
    kBCKernel_BlockSize   = 512, ///< 
    kBCKernel_PoolSize    = 512, ///< Kernel pool size.
    kBCKernel_AlgNameSize = 16,  ///< Length of algorythm name.
    kBCKernel_TotalModes  = 3,   ///< Total cipher modes supported by kernel.
    kBCKernel_MaxKeys     = 32  ///< Maximum count of keys that can be stored in kernel at one time.
};


// Disk setup options
enum {
    kBCDisk_Default     = 0, 
    kBCDisk_ReadOnly    = 0x01,
    kBCDisk_Automount   = 0x02
};


typedef int bc_key_handle;
typedef int bc_alg_id;
typedef int bc_mode_id;

#define BC_INVALID_ALG -1;

// user space only
#ifndef BC_KERNEL
#include <syslog.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <sys/types.h>
#include <stdio.h>

#include "spacemap/spacemap.h"

#define BC_KERNEL_EXPORT __attribute__ ((visibility ("default")))

#ifdef BC_DEBUG
    //#define DEBUG_MSG(msg, args...) openlog("BCKernel", LOG_CONS | LOG_PERROR, LOG_USER); syslog(LOG_INFO, (msg), ## args);  closelog();
    #define DEBUG_MSG(msg, args...) fprintf(stderr, msg, ## args);
#else
    #define DEBUG_MSG(msg, args...)
#endif

//#define ERROR_MSG(msg, args...) openlog("BCKernel", LOG_CONS | LOG_PERROR, LOG_USER); syslog(LOG_ERR, msg, ## args);  closelog();
#define ERROR_MSG(msg, args...) fprintf(stderr, msg, ## args)

#ifdef __cplusplus 
extern "C" {
#endif

    /**
     * @brief Registered kernel block cipher definition.
     */
    typedef struct BCAlgInfo
    {
        int           id;                             ///< Cipher unique id
        char          name[kBCKernel_AlgNameSize];    ///< Cipher name
        size_t        key_length;                     ///< Cipher key length in bits
        size_t        block_size;                     ///< Cipher block size in bytes
        size_t        cbc_key_size;                   ///< CBC mode key size in bytes. 0 if mode is not supported,
        size_t        lrw_key_size;                   ///< LRW mode key size in bytes. 0 if mode is not supported.
        size_t        xts_key_size;                   ///< XTS mode key size in bytes. 0 if mode is not supported.
    } BCAlgInfo;
    
    
    /**
     * Get mode human readable name for its id.
     */
    static inline const char* BCKernelModeName(bc_mode_id id)
    {
        switch (id) {
        case kBCMode_CBC: return "CBC";
        case kBCMode_LRW: return "LRW";
        case kBCMode_XTS: return "XTS";
        default: assert(0);
        }
    }

    /**
     * Gets key size for specified alg and mode.
     * @param alg       Pointer to alg.
     * @param mode      Encryption mode id.
     * @return    Mode key size or zero, if mode is not supported.
     * */
    static inline size_t BCKernelGetModeKeySize(const BCAlgInfo *alg, bc_mode_id mode)
    {
        switch (mode)
        {
            case kBCMode_CBC:
                return alg->cbc_key_size;

            case kBCMode_LRW:
                return alg->lrw_key_size;

            case kBCMode_XTS:
                return alg->xts_key_size;

            default:
                return 0;
        }
    }

    /**
     * Checks mode support by alg.
     * @param alg       Algorythm to check.
     * @param mode      Mode to check on algorythm.
     * @return          true/false value
     * */
    static inline int BCKernelIsModeSupported(const BCAlgInfo *alg, bc_mode_id mode)
    {
        return (0 != BCKernelGetModeKeySize(alg, mode));
    }

    /**
     * Get mode id fro its human readable name.
     */
    static inline bc_mode_id BCKernelModeId(const char* name)
    {
        if (0 == strcasecmp(name, "CBC")) {
            return kBCMode_CBC;
        } else if (0 == strcasecmp(name, "LRW")) {
            return kBCMode_LRW;
        } else if (0 == strcasecmp(name, "XTS")) {
            return kBCMode_XTS;
        } else {
			fprintf(stderr, "%s: Warning: unknown encryption algorinm mode '%s', CBC will be used.\n", PROG_NAME, name);
			return kBCMode_CBC;
        }
    }

    /**
     * Returns error string description.
     * @param result BC kernel function result to process.
     * @return       String representing short error description.
     * */
    static inline const char* BCKernelGetErrorString(int result)
    {
        switch(result)
        {
            case kBCKernel_Success:
                return "Success.";
            case kBCKernel_NotInitialized:
                return "Kernel is not initialized.";
            case kBCKernel_NoMemory:
                return "Not enough memory.";
            case kBCKernel_UnknownAlg:
                return "Unknown alg.";
            case kBCKernel_UnsupportedMode:
                return "Unsupported mode.";
            case kBCKernel_InvalidKey:
                return "Invalid key specified.";
            case kBCKernel_InvalidBuffer:
                return "Invalid buffer.";
            case kBCKernel_InvalidArgument:
                return "Invalid argument";
            case kBCKernel_DeviceBusy:
                return "Device busy";
            case kBCKernel_InternalError:
                return "Internal error.";
            default:
                return "Unknown result.";
        }
    }
 
    /**
     * @brief Tell if kernel modules are loaded and initialized.
     * @return    
     *            - kBCKernel_Success, if kernel is initialized. 
     *            - kBCKernel_NotInitialized, if kernel is not correctly initialized.
     */
    extern bc_kern_result BCKernelIsInitialized(void) BC_KERNEL_EXPORT;
    
    
    /**
     * Lookup kernel cipher plugin by id.
     * This function cannot guarantee that found plugin will still be availiable after it returns since kernel module can be unloaded.
     * @param id         Block cipher id.
     * @param out_alg    Initialized algorithm description on success.
     * @return     
     *                   - kBCKernel_UnknownAlg if algorithm is not found, 
     *                   - kBCKernel_InvalidBuffer, if out_alg pointer is NULL,
     *                   - kBCKernel_InternalError if other error occured. In the last case errno is set.
     */
    extern bc_kern_result BCKernelLookupAlgId(bc_alg_id id, BCAlgInfo* out_alg) BC_KERNEL_EXPORT;
    
    
    /**
     * Lookup kernel cipher plugin by name.
     * This function cannot guarantee that found plugin will still be availiable after it returns since kernel module can be unloaded.
     * @param name       Block cipher name.
     * @param out_alg    Initialized algorithm description on success.
     * @return     
     *                   - kBCKernel_UnknownAlg if algorithm is not found, 
     *                   - kBCKernel_InvalidBuffer, if out_alg pointer is NULL,
     *                   - kBCKernel_InternalError if other error occured. In this case errno is set.
     */
    extern bc_kern_result BCKernelLookupAlgName(const char* name, BCAlgInfo* out_alg) BC_KERNEL_EXPORT;
    
    
    /**
     * List all registered ciphers.
     * This function cannot guarantee that any of the listed plugins will still be availiable after it returns since kernel module can be unloaded.
     * @param out_palgs    Pointer to where the function will initialize array of block cipher descriptions. 
     *                     Caller should free it with bc_kernel_free_alg_list.
     * @param out_total    Total number of ciphers, i.e. allocated array size.
     * @return       
     *               - kBCKernel_InvalidBuffer, if out_palgs or out_total pointer is NULL,
     *               - kBCKernel_InternalError if other error occured. In this case errno is set.
     */
    extern bc_kern_result BCKernelListAlgs(BCAlgInfo** out_palgs, unsigned* out_total) BC_KERNEL_EXPORT;
    
    
    /**
     * Free an array of cipher descriptions.
     * @param list    Array to free.
     */
    extern void BCKernelFreeAlgList(BCAlgInfo* list) BC_KERNEL_EXPORT;


    /**
     * Store encryption key in kernel service memory.
     * @param key           Key data to store.
     * @param keylen        Total bytes in key data.
     * @param out_handle    Returned handle to stored key.
     */
    extern bc_kern_result BCKernelStoreKey(const char* key, size_t keylen, bc_key_handle* out_handle) BC_KERNEL_EXPORT;
    
    
    /**
     * Free encryption key memory in kernel service.
     * @param handle    Handle to stored key.
     */
    extern bc_kern_result BCKernelFreeKey(bc_key_handle handle) BC_KERNEL_EXPORT;
    
    
    /**
     * Process (encrypt/decrypt) data.
     * @param alg_id        Block cipher id.
     * @param mode_id       Encryption mode id.
     * @param proc          Processing direction (encrypt/decrypt).
     * @param key_handle    Handle to stored encryption key.
     * @param iv            Initial vector.
     * @param pool          Random data pool. NULL for zeroed pool.
     * @param data          Data to process.
     * @param nbytes        Number of bytes to process.
     */
    extern bc_kern_result BCKernelProcess(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc, 
                        bc_key_handle key_handle, uint64_t iv, const char* pool, 
                        char* data, size_t nbytes) BC_KERNEL_EXPORT;
    
    
    /**
     * Process data with explicit key without storing it.
     * Same as above but uses explicit key data instead of stored key handle.
     */
    extern bc_kern_result BCKernelProcessWithKey(bc_alg_id alg_id, bc_mode_id mode_id, bc_proc_type proc,
                        const char* key, size_t keysize,
                        uint64_t iv, const char* pool,
                        char* data, size_t nbytes) BC_KERNEL_EXPORT;
    

    /**
     * Create new virtual block device and link it to container file.
     * @param file              File name
     * @param alg               Encryption algorithm id
     * @param mode              Encryption mode
     * @param key_handle        Encryption key handle stored in kernel memory
     * @param pool              Random data pool, can be NULL.
     * @param extents           Extents indicating where encrypted data is located.
     * @param options           kBCDisk_X flags
     */
    bc_kern_result BCKernelSetupDisk(const char* file, int alg, int mode, int key_handle, const char* pool, SpaceMap *extents, int options, char* out_devname);


    /**
     * Remove existing block device.
     * @param devname           Existing block device name
     * @param force             Force removal even if device is busy.
     */
    bc_kern_result BCKernelRemoveDisk(const char* devname, int force);

    /**
     * Algorithm-mode benchmark.
     * @param alg_id            Algorithm id to benchmark.
     * @param mode_id           Algorithm mode to benchmark.
     * @param out_nanoseconds   Average time in nanoseconds needed
     *                          to encrypt one megabyte of garbage.
     */
    bc_kern_result BCKernelBenchmark(bc_alg_id alg_id, bc_mode_id mode_id, unsigned long *out_nanoseconds);

#ifdef __cplusplus
}
#endif

#endif // !BC_KERNEL
#endif // !BC_LIB_KERNEL_H

