/*
 * Copyright (c) 2010 Jetico, Inc., Finland. All rights reserved.
 * Common wrapper over all our crypto-related functionality.
 */

#ifndef BC_KERNEL_CRYPTO_H
#define BC_KERNEL_CRYPTO_H

#include "crypto/cipher.h"
#include "crypto/mode.h"

struct bc_alg {
    // precalculated for convinience.
    size_t  key_size;
    size_t  extended_key_size;

    const struct bc_cipher* cipher;
    const struct bc_mode* mode;

    void* key;
};


/**
 * Init key table structures.
 */
void bc_crypto_init_key_table(void);

/**
 * Store key data in kernel memory and return a handle to it.
 */
int bc_crypto_store_key(char* key, size_t keysize, int* out_handle);

/**
 * Free key from kernel memory.
 */
int bc_crypto_free_key(int handle);

/**
 * Release all key table resources.
 */
void bc_crypto_release_key_table(void);

/**
 * Init algorithm.
 * Will take care to increase implementing plugin refcount.
 * Release with bc_crypto_put_alg
 */
int bc_crypto_get_alg(int alg_id, int mode_id, struct bc_alg* out_alg);

/**
 * Set algorithm key.
 */
int bc_crypto_set_key(struct bc_alg* alg, int key_handle);

/**
 * Set algorithm key from explicitely supplied buffer.
 */
int bc_crypto_set_key_explicit(struct bc_alg* alg, char* key, size_t keysize);

/**
 * Encrypt a block of data inplace.
 * @param alg       Initialized algorithm with a set key.
 * @param iv        Initial vector.
 * @param pool      Random data pool. Can be NULL and a zero-filled pool will be used instead.
 * @param nbytes    Total bytes to process.
 */
void bc_crypto_encrypt(struct bc_alg* alg, uint64_t iv, char* pool, char* data, size_t nbytes);

/**
 * Decrypt a block of data inplace.
 * @param alg       Initialized algorithm with a set key.
 * @param iv        Initial vector.
 * @param pool      Random data pool. Can be NULL and a zero-filled pool will be used instead.
 * @param nbytes    Total bytes to process.
 */
void bc_crypto_decrypt(struct bc_alg* alg, uint64_t iv, char* pool, char* data, size_t nbytes);

/**
 * Free algorithm and its key if set.
 * Will also decrement plugin reference count.
 */
void bc_crypto_put_alg(struct bc_alg* alg);


// Single-call procedure to encrypt some data
int bc_encrypt(int alg_id, int mode_id, int key_handle, uint64_t iv, char* pool, char* data, size_t nbytes);

// Single-call procedure to decrypt some data
int bc_decrypt(int alg_id, int mode_id, int key_handle, uint64_t iv, char* pool, char* data, size_t nbytes);


#endif

