/*
 * Copyright (c) 2010 Jetico, Inc., Finland
 * All rights reserved.
 *
 * Header file for controller module.
 * Included in both kernel and user space implementations.
 */

#ifndef BC_KERNEL_LINUX_CONTROLLER_H
#define BC_KERNEL_LINUX_CONTROLLER_H

#ifdef BC_KERNEL
#include <linux/types.h>
#else
#include <stdint.h>
#endif

/// Current controller version tags.
#define BC_CTL_VERSION_MAJOR 2
#define BC_CTL_VERSION_MINOR 0

/// Controller character device node name in /dev and /sys/class/misc/
#define BC_CTL_NODE_NAME                    "bestcrypt"

/// Controllers plugins sysfs directory entry
#define BC_CTL_PLUGINS_GROUP			    plugins
#define BC_CTL_PLUGINS_GROUP_STR		    "plugins"
#define BC_CTL_VERSION_MAJOR_ATTR		    ver_major
#define BC_CTL_VERSION_MINOR_ATTR		    ver_minor
#define BC_CTL_BLOCK_MAJOR_ATTR             block_major

/// Controller plugin sysfs attributes.
#define BC_CTL_PLUGIN_NAME_ATTR			    name
#define BC_CTL_PLUGIN_ID_ATTR			    id
#define BC_CTL_PLUGIN_BLOCK_SIZE_ATTR		block_size
#define BC_CTL_PLUGIN_KEY_LENGTH_ATTR		key_length
#define BC_CTL_PLUGIN_CBC_KEY_SIZE_ATTR		cbc_key_size
#define BC_CTL_PLUGIN_LRW_KEY_SIZE_ATTR		lrw_key_size
#define BC_CTL_PLUGIN_XTS_KEY_SIZE_ATTR		xts_key_size


/// IOCTL codes
enum {
    // Controller device ctl
    kBCCTL_Info         = 0xBC000000,       ///< Get info
    kBCCTL_StoreKey 	= 0xBC000001,       ///< Store encryption key.
    kBCCTL_FreeKey 		= 0xBC000002,       ///< Free encryption key.
    kBCCTL_Process 		= 0xBC000004,       ///< Process (encrypt/decrypt) a block of data.
    kBCCTL_CreateDisk	= 0xBC000008,       ///< Create new blank virtual block device.
    kBCCTL_Benchmark    = 0xBC000016,       ///< Used for benchmarking encryption algorithms.

    // Virtual block device ctl
    kBCCTL_DiskSetFd    = 0xBC000010,       //< Link container file to blank block device to begin transparent filtering
    kBCCTL_DiskClearFd  = 0xBC000020       //< Unlink container file from block device
};


// Wrapper over userland pointers for cross-architecture compatibility
typedef union
{
    void* value;
    uint64_t tmp_pad;
} bc_ctl_ptr;



// Encryption key data embedded in some parameters
struct bc_key {
	bc_ctl_ptr data;
	uint64_t   size;
};


// Store key ioctl in parameter.
struct bc_ctl_key {
    struct          bc_key key;
    bc_ctl_ptr      out_handle;
};


/**
 * Process ioctl in/out parameter.
 * User can either send the key as a handle to the stored one or supply it explicitely in which case it will not be stored.
 */
struct bc_ctl_process {
    uint32_t    flags;
    char        _tmp_pad[4];
	bc_ctl_ptr  data;
	uint64_t    nbytes;

	int		    alg_id;
	int		    mode_id;
	int		    proc_type;
	int		    has_key;
	
	union {
		struct bc_key inlined;
		int           handle;
	} key;

	uint64_t   iv;
	bc_ctl_ptr pool;
};

struct bc_ctl_bench
{
    // In
    int             alg_id;     ///< Id of algorithm to benchmark.
    int             alg_mode;   ///< Algorithm mode to benchmark.

    // Out
    unsigned long   nsec;       ///< Average time in nanoseconds that needs for encryption of 1 MB random data.
};


/**
 * Virtual block device setup command argument.
 * Controller driver will create a new blank block device and return its minor number to the caller.
 */
struct bc_ctl_disk {
    int         out_minor;          ///< returned minor number of the new device so that user space can find the new node
    char        *out_name;          ///< returned name of the new device (ex. 'bcrypta'. The caller should allocate the memory.
    int         name_bufsize;       ///< the size of out_path buffer
};


struct bc_ctl_info {
    int ver_major;
    int ver_minor;
    int block_major;
};


/**
 * Disk ctl to link block device to container file.
 * Separated from kBCCTL_CreateDisk because we need an existing block device file inode to correctly set sizes.
 */
struct bc_diskctl_set_fd 
{
    int		    fd;             // container file descriptor.
    int 	    alg_id;         // encryption algorithm id.
    int		    mode_id;        // encryption mode id.
    int		    key_handle;     // driver key handle.
    char        _tmp_pad[4];
    bc_ctl_ptr  pool;           // random data pool for CBC mode.

    bc_ctl_ptr  extents;        // extents inside container that contain real data.
    uint64_t    extent_count;   // count of extents.

    //uint64_t    offset;         // first real sector of the new device.

    int         options;
};


#endif

