/*
 * Copyright (c) 2011 Jetico, Inc., Finland
 * All rights reserved.
 *
 * Trasparently encrypted block device implementation.
 * Data less or equal to 4KB is allocating by kmalloc,
 * for more than 4KB - vmalloc.
 */

#include "bc_dev.h"
#include "bc_ctl.h"
#include "bc_defs.h"
#include "BCKernel.h"
#include "spacemap/spacemap.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/vermagic.h>
#include <linux/compiler.h>
#include <linux/kernel.h>
#include <linux/hardirq.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <linux/fs_struct.h>
#include <linux/file.h>
#include <linux/cdrom.h>
#include <linux/buffer_head.h>
#include <linux/genhd.h>
#include <linux/bio.h>
#include <linux/blkdev.h>
#include <linux/hdreg.h>
#include <linux/dcache.h>
#include <linux/mount.h>
#include <linux/kthread.h>
#include <asm/uaccess.h>

#include "bc_mem.h"

static inline long BC_IS_ERR_OR_NULL(const void *ptr)
{
    return !ptr || IS_ERR_VALUE((unsigned long)ptr);
}

// Device major number initialized by controller module on startup. Defined in bc_mod.c
extern int g_dev_major;

// Maximum number of partitions for each device.
static int g_max_parts = 16;

// Maximum number of active devices.
static int g_max_devs = 20;

// Map indicating devices available.
static bool *g_avail_devs = NULL;

/**********************************************************************************
 *
 * 	Device request processing.
 *
 **********************************************************************************/


static inline void dev_enqueue_request(struct bc_device* dev, struct bio* bio) {
	spin_lock_irq(&dev->queuelock);
	bio_list_add(&dev->queue, bio);
	spin_unlock_irq(&dev->queuelock);
	up(&dev->queuesem);
}

static inline int dev_dequeue_request(struct bc_device* dev, struct bio** out_bio) {

    int result = down_interruptible(&dev->queuesem);
    if (likely(!result)) {
	    spin_lock_irq(&dev->queuelock);
	    (*out_bio) = bio_list_pop(&dev->queue);
	    spin_unlock_irq(&dev->queuelock);
        return 0;
    }
    return result;
}


/*
 * This is called asynchroniusly by kernel block level when new request for one of our devices is availiable.
 * Device instance pointer is stored inside request queue.
 */
static inline void _dev_make_request(struct request_queue *q, struct bio *bio)
{
	struct bc_device* 	dev = q->queuedata;
//    if (unlikely(blk_queue_discard(q))) {
//        BC_KERN_ERROR("discard!");
//    }


    switch(bio_rw(bio)) {

	case WRITE: {
		if(bc_dev_is_read_only(dev)) {
			BC_KERN_ERROR("attempt to write on read only device %s\n", bc_dev_name(dev));
			bio_endio(bio, -EIO);
            return;
		}
		break;
	}
    case READA:
	case READ: {
		break;
	}
	default: {
        BC_KERN_ERROR("invalid request for device %s: %lu\n", bc_dev_name(dev),
                      (long unsigned)bio_rw(bio));
		bio_endio(bio, -EIO);
        return;
    }
	};


    //BC_DEBUG_LOG("%s request for device %s\n", (bio_rw(bio) == WRITE) ? "write" : "read", bc_dev_name(dev));

	dev_enqueue_request(dev, bio);
    return;
}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,2,0))
static int dev_make_request(struct request_queue *q, struct bio *bio)
{
    _dev_make_request(q, bio);
    return 0;
}
#else
static void dev_make_request(struct request_queue *q, struct bio *bio)
{
    _dev_make_request(q, bio);
}
#endif

static int dev_process_bio(struct bc_device* dev, struct bio* bio)
{
// https://git.kernel.org/cgit/linux/kernel/git/torvalds/linux.git/commit/?id=4550dd6c6b062fc5e5b647296d55da22616123c3
#ifdef bvec_iter_page
    struct bio_vec          bvec;
    struct bvec_iter        iter;
#else
    struct bio_vec*     bvec;
    int                 iter;
#endif
    char*               buf;
    size_t              nbytes,         // Number of bytes processed on previous vectors.
                        nbytes_done,    // Number of bytes processed by this vector.
                        total = 0;      // Total number of bytes processed.
    uint64_t           iv;             // Initial Vector for cryptographic algorithm
    mm_segment_t        fs;
    Extent*             phys;           // Array with file extents to do io.
    bc_u64              phys_count;     // Count of file extents in array.
    bc_u64              li;
    sm_result           smres;
    Extent              first;
#ifdef BC_DEBUG
    bc_u64              tmp_phys_length;
#endif

#ifdef bvec_iter_page
    loff_t position   = (loff_t)bio->bi_iter.bi_sector << 9; // Number of the first byte in io request.
#else
    loff_t position   = (loff_t)bio->bi_sector << 9; // Number of the first byte in io request.
#endif

    BUG_ON(!dev->is_alive);

    bio_for_each_segment(bvec, bio, iter)
    {
        // Some functions below expect memory address from the userspace,
        // so, to be able to use them, we need to change user-space segment
        // to kernel-space segment, and, after bio segment processing,
        // set it back to the old value, that is kept in fs var.
        fs = get_fs();
        set_fs(KERNEL_DS);

        /* One uppon a time Initial Vector was 16 bit of LBA sector number
         * got 4 times. It was 65K uniq Initial Vectors only.
         * To avoid 'hash attack' problem and for backward compatibility
         * new Initial Vector became LBA+1 in order to distinct
         * just 0 from 4 times 16 bit 0.
         */
#ifdef bvec_iter_page
        iv = __cpu_to_le64(bio->bi_iter.bi_sector + (total >> 9) + 1); // iv numbering begins from 1
        nbytes  = bvec.bv_len;
#else
        iv = __cpu_to_le64(bio->bi_sector + (total >> 9) + 1); // iv numbering begins from 1
        nbytes  = bvec->bv_len;
#endif


#ifdef bvec_iter_page
        kmap(bvec.bv_page);
        buf     = page_address(bvec.bv_page) + bvec.bv_offset;
#else
        kmap(bvec->bv_page);
        buf     = page_address(bvec->bv_page) + bvec->bv_offset;
#endif
        BUG_ON(!buf);

        // Translate io request to physical extents inside container.
        // TODO: Create a buffer holding extent result to avoid allocation of it everytime!
        if (unlikely(1 != dev->extent_count)) {
            smres = SpaceMapGetPhysicalExtents(dev->extents, dev->extent_count, (bc_u64)position,
                                       (bc_u64)nbytes, &phys, &phys_count);
            if (unlikely(kSpaceMap_Success != smres))
            {
                BC_KERN_ERROR("Cannot translate io request to physical extents:%s\n",
                              SpaceMapGetErrorString(smres));
                bio_endio(bio, -EIO);
                return 0;
            }
        } else {
            // Have only one extent (usual for main part).
            first.start = (bc_u64)position + dev->extents[0].start;
            first.length = (bc_u64)nbytes;
            phys_count = 1;
            phys = &first;
        }
#ifdef BC_DEBUG
        SpaceMapGetExtentSpace(phys, phys_count, &tmp_phys_length);
        //BC_DEBUG_LOG("total length of physical extents is %lu\n", (unsigned long) tmp_phys_length);
        if (tmp_phys_length != nbytes)
        {
            BC_KERN_ERROR("Physical length is incorrect. Wanted %lu, got %lu.\n",
                          (unsigned long)bvec->bv_len, (unsigned long)tmp_phys_length);
            BC_KERN_ERROR("Extents:\n");
            for ( li = 0; li < phys_count; li++)
            {
                BC_KERN_ERROR("  start: %lu, length: %lu\n",
                              (unsigned long)phys[li].start, (unsigned long)phys[li].length);
            }
            BC_KERN_ERROR("block IO vector information: offset: %lu, length: %lu, positoin: %lu\n",
                          (unsigned long)bvec->bv_offset, (unsigned long)bvec->bv_len,
                          (unsigned long)position);
            BC_KERN_ERROR("segment number: %i\n", i);
            bio_endio(bio, -EIO);
            return 0;
        }
#endif
		switch(bio_rw(bio)) {
        case READA:
		case READ: {
            //nbytes_done = dev->filp->f_op->read(dev->filp, buf, nbytes, &pos);
            nbytes_done = 0;

            for (li = 0; li < phys_count; li++) {
                nbytes_done += dev->filp->f_op->read(dev->filp, (buf + nbytes_done),
                                                     phys[li].length, &phys[li].start);
            }

            bc_crypto_decrypt(&dev->alg, iv, dev->pool, buf, nbytes_done);
			break;
		}
		case WRITE: {
            down(&dev->writesem);
            // data in buf points to the page and can't be larger than page size.
            memcpy(dev->writebuf, buf, nbytes);
			bc_crypto_encrypt(&dev->alg, iv, dev->pool, dev->writebuf, nbytes);
            nbytes_done = 0;

            for (li = 0; li < phys_count; li++) {
                nbytes_done += dev->filp->f_op->write(dev->filp, (dev->writebuf + nbytes_done),
                                                      phys[li].length, &phys[li].start);
            }
            up(&dev->writesem);
			break;
		}
		default: BUG();
		};

        if (unlikely(1 != dev->extent_count)) {
            SpaceMapFreePhysicalExtents(&phys);
        }

#ifdef bvec_iter_page
        kunmap(bvec.bv_page);
#else
		kunmap(bvec->bv_page);
#endif

		set_fs(fs);

        if (unlikely(nbytes != nbytes_done)) {
            // Get information about the device size for easier debugging.
            smres = SpaceMapGetExtentSpace(dev->extents, dev->extent_count, &li);
            BC_KERN_ERROR("device %s io request error:"
                          "\n\twanted: %ld bytes,"
                          "\n\tcompleted %ld bytes."
                          "\n\tnumber of the first byte in io request: %ld."
                          "\n\tdevice size: %ld.\n",
                          bc_dev_name(dev), (long)nbytes,
                          (long)nbytes_done, (long)position, (long)li);
			bio_endio(bio, -EIO);
			return 0;
		}

        total += nbytes;
        position += nbytes;
	}

	bio_endio(bio, 0);
	return 0;
}


/**
 * Request consumer kthread proc.
 * Waits for new request on queue semaphore and processes them once they arrive.
 */
static int dev_thread(void *data)
{
    struct bio*         bio;
    struct bc_device*   dev = (struct bc_device*)data;
    int error;

    if (unlikely(!dev)) {
        return -1;
    }

	set_user_nice(current,  -20l);

	/*
 	 * Driver main thread will tell us to stop when it is being torn down and no more requests are coming
	 * by upping the queue semaphore without the actual request.
     * In this (and only this) case device_dequeue_request will set NULL in bio param.
	 */
    while (1) {
        error = dev_dequeue_request(dev, &bio);
        if (likely(!error)) {
            if (likely(bio)) {
                dev_process_bio(dev, bio);
            } else {
                // Null request received, ending thread.
                break;
            }
        } // else semaphore lock is interrupted. Relaunch waiting.
    }

	BC_DEBUG_LOG("consumer thread recieved NULL request, finishing\n");
	return 0;
}




/****************************************************************************
 *
 * 	Block device management.
 *
 ****************************************************************************/




static ssize_t show_device_backing_store(struct device* dev, struct device_attribute* attr, char* buf)
{
    struct gendisk* disk = dev_to_disk(dev);
    struct bc_device* bc = disk->private_data;

    ssize_t ret;
    char *p = NULL;
    if (!bc) {
        return -1;
    }

    if (bc->filp) {
        p = d_path(&bc->filp->f_path, buf, PAGE_SIZE - 1);
    }

    if (BC_IS_ERR_OR_NULL(p))
        ret = PTR_ERR(p);
    else {
        ret = strlen(p);
        memmove(buf, p, ret);
        buf[ret++] = '\n';
        buf[ret] = 0;
    }

    return ret;
}

static struct device_attribute g_dev_backing_store_attr = __ATTR(backing_store, S_IRUGO, show_device_backing_store, NULL);

// Gets next available device number and marks it as busy.
static int bc_get_avail_dev_num(void)
{
    int i = 0;
    if (unlikely(!g_avail_devs)) {
        g_avail_devs = (bool*)VMALLOC(g_max_devs * sizeof(bool));
        memset(g_avail_devs, 0, g_max_devs * sizeof(int));
    }
    for (; i<g_max_devs; ++i) {
        if (!g_avail_devs[i]) {
            BC_DEBUG_LOG("allocated device #%i\n", i);
            g_avail_devs[i] = true;
            return i;
        }
    }

    return -1;
}


// Marks device number as available.
static void bc_put_dev_num(int dev_num)
{
    BC_DEBUG_LOG("device number to free: %i\n", dev_num);
    if ((dev_num >= 0) && (dev_num < g_max_devs))
        g_avail_devs[dev_num] = false;
}

// Forms device name and marks it as unavailable.
static int bc_get_dev_name(int dev_num, char *buf, int max_size)
{
    if (-1 == dev_num)
        return -1;
    snprintf(buf, max_size, "%s%c", BC_BASE_DEVICE_NAME, 'a' + dev_num);
    BC_DEBUG_LOG("device name %s formed for number %i.\n", buf, dev_num);
    return 0;
}

// Marks device name as available.
static void bc_put_dev_name(const char *dev_name)
{
    int start_i, dev_num;
    BC_DEBUG_LOG("freeing device number for name %s.\n", dev_name);
    start_i = strlen(BC_BASE_DEVICE_NAME);
    dev_num = dev_name[start_i] - 'a';
    bc_put_dev_num(dev_num);
}

static void bc_dev_free(struct bc_device* dev)
{
    if (dev->is_alive) {
        BC_KERN_ERROR("freeing an alive device.\n");
        BUG();
        return;
    }
    BC_DEBUG_LOG("freeing block queue\n");
	if(dev->blkqueue) 	blk_cleanup_queue(dev->blkqueue);
    if(dev->gendisk) {
        // Reset private data, that contains a pointer to our device.
        dev->gendisk->private_data = NULL;

        BC_DEBUG_LOG("putting device name\n");
        bc_put_dev_name((const char*)dev->gendisk->disk_name);

        BC_DEBUG_LOG("freeing gendisk\n");
        put_disk(dev->gendisk);
    }

    BC_DEBUG_LOG("toggling write semaphore\n");
    down(&dev->writesem);

    if(dev->writebuf) {
        BC_DEBUG_LOG("freeing writing buffer\n");
        KFREE(dev->writebuf);
        dev->writebuf = NULL;
    }

    BC_DEBUG_LOG("freeing extent list for device 0x%x\n", dev->extents);
    KFREE(dev->extents);

    BC_DEBUG_LOG("freeing device information\n");
	memset(dev, 0, sizeof(*dev));
    KFREE(dev);
}



// Link initialized device to container file creating crypto-loop device.
static int bc_dev_link(struct block_device* bdev, int fd, int algid, int modeid,
                       int key_handle, char* pool, Extent* extent_list, uint64_t extent_count,
                       /*uint64_t start_sector, uint64_t nsectors,*/ int options)
{
	int             error;
	struct file*	filp;
	struct inode*	inode;
	struct dentry*	dentry;
	loff_t 		    fsize;
    int ro;
    sm_result       smres;
    bc_u64          virtual_space;
    Extent          *extents;

    struct bc_device* dev = bdev->bd_disk->private_data;

    if (!dev) {
        BC_KERN_ERROR("invalid device private data.\n");
        return -EINVAL;
    }

    if ((!extent_list) || (!extent_count)) {
        BC_KERN_ERROR("empty extent list came.\n");
        return -EINVAL;
    }

    spin_lock(&dev->lock);
    if (!dev->is_alive) {
        BC_KERN_ERROR("tried to link a file to the device which is shutting down.\n");
        return -ENODEV;
    }

    spin_unlock(&dev->lock);

    BC_DEBUG_LOG("validating file.\n");
	// validate backing store.
    ro = options & kBCDisk_ReadOnly;

	filp = fget(fd);
	if(!filp) {
        BC_KERN_ERROR("invalid file.\n");
		return -EINVAL;
	}

    BC_DEBUG_LOG("getting inode.\n");



#if (LINUX_VERSION_CODE < KERNEL_VERSION(3,15,0))
	dentry = filp->f_dentry;
#else
	dentry = filp->f_path.dentry;
#endif

	inode = dentry->d_inode;
	if(!inode || !filp->f_op || !filp->f_op->read || !(ro || filp->f_op->write)) {
        BC_KERN_ERROR("%s: cannot get inode, no fops or file doesn't support operation\n", bc_dev_name(dev));
        error = -EINVAL;
        goto DEV_LINK_PUT_FILE;
    }

    BC_DEBUG_LOG("getting the file size.\n");

	fsize = i_size_read(inode->i_mapping->host);

    BC_DEBUG_LOG("allocating extent list for device.\n");
    extents = (Extent*)KMALLOC(extent_count * sizeof(Extent));
    if (!extents) {
        BC_KERN_ERROR("failed to allocate memory for extent list with %lu elements\n",
                      (unsigned long)extent_count);
        error = -ENOMEM;
        goto DEV_LINK_PUT_FILE;
    }
    memcpy(extents, extent_list, extent_count * sizeof(Extent));

    BC_DEBUG_LOG("calculating space.\n");
    smres = SpaceMapGetExtentSpace(extents, (bc_u64)extent_count, &virtual_space);
    if (kSpaceMap_Success != smres)
    {
        BC_KERN_ERROR("can't get the length of the space map: `%s`\n", SpaceMapGetErrorString(smres));
        error = -EFAULT;
        goto DEV_LINK_FREE_EXTENTS;
    }

    BC_DEBUG_LOG("virtual space is %llu\n", (long long unsigned)virtual_space);

    if(virtual_space > fsize) {
        BC_KERN_ERROR("container size (%lld) is less than virtual space (%llu).\n", fsize, virtual_space);
        error = -EINVAL;
        goto DEV_LINK_FREE_EXTENTS;
	}

	// figure out write permissions for this backing store.
    // if we were asked to create a device with write access we need to make sure
    // that backing store can support it and block writes otherwise.

    BC_DEBUG_LOG("checking device write permissions.\n");
	ro = (options & kBCDisk_ReadOnly) || IS_RDONLY(inode) || (NULL == filp->f_op->write);
	if(S_ISBLK(inode->i_mode)) {
        ro = (ro || bdev_read_only(inode->i_bdev));
	}

    // Set read-only flag and check file write permissions, if necessary.
    BC_DEBUG_LOG("trying to set read-only flag to %i.\n", ro);
    dev->flags.readonly = ro;

	if(!dev->flags.readonly && (0 != get_write_access(inode))) {
        BC_KERN_ERROR("failed getting write access for inode.\n");
        error = -EACCES;
        goto DEV_LINK_FREE_EXTENTS;
    }

	// set encryption context transferring ownership of the key.
	// if anything happens, don't forget to remove the key.
    BC_DEBUG_LOG("initializing algorithm.\n");
	error = bc_crypto_get_alg(algid, modeid, &dev->alg);
    if(error) {
        BC_KERN_ERROR("failed initializing algorithm id 0x%x with mode 0x%x: 0x%x\n", algid, modeid, error);
        goto DEV_LINK_FREE_EXTENTS;
    }

    BC_DEBUG_LOG("store key.\n");
	error = bc_crypto_set_key(&dev->alg, key_handle);
	if(error) {
        BC_KERN_ERROR("failed setting key with handle %d: 0x%x\n", key_handle, error);
        goto DEV_LINK_PUT_ALG;
	}

    if(pool) {
        memcpy(dev->pool, pool, sizeof(dev->pool));
    } else {
        memset(dev->pool, 0, sizeof(dev->pool));
    }

    BC_DEBUG_LOG("set gendisk capacity to %lu sectors.\n", (unsigned long)(virtual_space >> 9));
    set_capacity(dev->gendisk, virtual_space >> 9); // set capacity in sectors.

    BC_DEBUG_LOG("set bdev read-only flag to %i.\n", !!dev->flags.readonly);
	set_device_ro(bdev, !!dev->flags.readonly);

	// initialize request queue and start up the consumer thread.
    BC_DEBUG_LOG("set queue_make_request method to block queue.\n");
	blk_queue_make_request(dev->blkqueue, dev_make_request);
    BC_DEBUG_LOG("creating thread.\n");
	dev->thread = kthread_create(dev_thread, dev, "%s", bc_dev_name(dev));
	if(!dev->thread) {
        BC_KERN_ERROR("failed starting device thread\n");
        error = -ENOMEM;
        goto DEV_LINK_PUT_ALG;
	}

	dev->filp 		    = filp;
    dev->extents        = extents;
    dev->extent_count 	= extent_count;
	dev->flags.readonly = ro;
	dev->dentry 		= dget(dentry);
	dev->gfpmask		= mapping_gfp_mask(dev->dentry->d_inode->i_mapping);
	dev->bdev           = bdev;
    BC_DEBUG_LOG("set gfp mask.\n");
	mapping_set_gfp_mask(dev->dentry->d_inode->i_mapping, dev->gfpmask & ~(__GFP_IO | __GFP_FS));

    BC_DEBUG_LOG("starting up consumer thread.\n");
    wake_up_process(dev->thread);

    BC_DEBUG_LOG("set bdev size to %lu bytes.\n", (unsigned long)virtual_space);
    bd_set_size(bdev, virtual_space);               // set size in bytes.

	BC_DEBUG_LOG("linked new device %s:\n", dev->gendisk->disk_name);
    BC_DEBUG_LOG(" algorithm:   0x%x\n", algid);
    BC_DEBUG_LOG(" mode:        0x%x\n", modeid);
    BC_DEBUG_LOG(" size:        %llu\n", virtual_space);
    BC_DEBUG_LOG(" read only:   %d\n", ro);
    BC_DEBUG_LOG(" extent count: %llu\n", dev->extent_count);
    BC_DEBUG_LOG(" first extent: (%llu, %llu), address: 0x%x\n",
                 dev->extents[0].start, dev->extents[0].length, dev->extents);

	return 0;
DEV_LINK_PUT_ALG:
    bc_crypto_put_alg(&dev->alg);
DEV_LINK_FREE_EXTENTS:
    KFREE(extents);
DEV_LINK_PUT_FILE:
    fput(filp);
    return error;
}


// when forced we should not fail here
static int bc_dev_unlink(struct bc_device* dev, int force)
{
	int error;
    struct block_device* bdev = NULL;

	bdev = dev->bdev;
	if(!bdev) {
		BC_KERN_ERROR("invalid block device\n");
		return -EINVAL;
	}

    spin_lock(&dev->lock);

	if((dev->refcnt > 1 /* one reference is for currently opened device to send ioctl */)) {
		BC_KERN_ERROR("device %s is busy, refcount is %d\n", bc_dev_name(dev), dev->refcnt);
        if (!force) {
            spin_unlock(&dev->lock);
    		return -EBUSY;
        }
        else
        {
            BC_KERN_ERROR("Forced removing is used.\n");
        }
	}

    BC_DEBUG_LOG("device is going to shutdown.\n");
    dev->is_alive = 0;

    BC_DEBUG_LOG("releasing spin lock.\n");
    spin_unlock(&dev->lock);

    BC_DEBUG_LOG("removing sysfs entry for device.\n");
	sysfs_remove_file(&disk_to_dev(dev->gendisk)->kobj, &g_dev_backing_store_attr.attr);

    // unlock backing file and cleanup resources
    BC_DEBUG_LOG("committing requests to file.\n");
    fsync_bdev(bdev);
    BC_DEBUG_LOG("restoring gfp mask.\n");
	mapping_set_gfp_mask(dev->dentry->d_inode->i_mapping, dev->gfpmask);
	if(!dev->flags.readonly) {
        BC_DEBUG_LOG("releasing write access to inode.\n");
		put_write_access(dev->dentry->d_inode);
	}

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,22))
	invalidate_bdev(bdev, 0);
#else
	invalidate_bdev(bdev);
#endif


    BC_DEBUG_LOG("set gendisk capacity to zero.\n");
    set_capacity(dev->gendisk, 0);

    BC_DEBUG_LOG("releasing file pointer and dentry.\n");
	dput(dev->dentry);
	fput(dev->filp);
    dev->dentry = NULL;
    dev->filp = NULL;

    BC_DEBUG_LOG("deleting gendisk.\n");
    del_gendisk(dev->gendisk);
    BC_DEBUG_LOG("do block query cleanup.\n");
    // stop new requests on queue
    blk_cleanup_queue(dev->blkqueue);
    BC_DEBUG_LOG("incrementing queue semaphore.\n");
    up(&dev->queuesem);

    BC_DEBUG_LOG("stopping consumer thread.\n");
    error = kthread_stop(dev->thread);
    if(error) {
        BC_KERN_ERROR("warning: kthread_stop returned %d\n", error);
    }
    else {
        BC_DEBUG_LOG("thread stopped\n");
    }

    dev->blkqueue = NULL;
    dev->gendisk->private_data = NULL;

    BC_DEBUG_LOG("releasing algorithm.\n");
	bc_crypto_put_alg(&dev->alg);

    BC_DEBUG_LOG("freeing device.\n");
    bc_dev_free(dev);

    BC_DEBUG_LOG("putting module.\n");
	module_put(THIS_MODULE);
    BC_DEBUG_LOG("completed unlinking device.\n");
	return 0;
}


static int bc_open(struct block_device* bdev, unsigned mode)
{
	struct bc_device* dev = bdev->bd_disk->private_data;
    if (unlikely(!dev))
        return -EINVAL;
    if (unlikely(!dev->is_alive)) {
        BC_KERN_ERROR("someone tried to open the device that is shutting down.\n");
        return -ENODEV;
    }
	spin_lock(&dev->lock);
	dev->refcnt++;
	BC_DEBUG_LOG("device %s opened, refcount %d\n", bc_dev_name(dev), dev->refcnt);
	spin_unlock(&dev->lock);

	return 0;
}


static int bc_release(struct gendisk* gendisk, unsigned mode)
{
	struct bc_device* dev = gendisk->private_data;
    if (unlikely(!dev)) {
        return -EINVAL;
    }

	spin_lock(&dev->lock);
	dev->refcnt--;
	BC_DEBUG_LOG("device %s closed, refcount %d\n", bc_dev_name(dev), dev->refcnt);
	spin_unlock(&dev->lock);

	return 0;
}


static int bc_getgeo(struct block_device *bdev, struct hd_geometry *geo)
{
	unsigned long 		i, j, size;
	unsigned char 		head, sect;
	unsigned short 		cyl;
	struct bc_device*	dev;

	dev = bdev->bd_disk->private_data;
    if (unlikely(!dev)) {
        BC_KERN_ERROR("not a bestcrypt device in getgeo.\n");
        return -EINVAL;
    }
    if (!dev->is_alive) {
        BC_KERN_ERROR("getgeo to device that is shutting down.\n");
        return -ENODEV;
    }

	if (NULL == dev->dentry)
		return -ENXIO;
	if (NULL == geo)
		return -EINVAL;

	size = get_capacity(dev->gendisk) << 1;

	for (i = -1, j = size; j > 0; j >>= 1, i++);

	if (0 > i)
		return -ENXIO;

	if (16 > i) {
		sect = 1;
		head = 1;
		cyl  = size;
	} else if (23 > i) {
		sect = 1;
		head = 1 << (i-15);
		cyl  = size >> (i-15);
	} else {
		sect = 1 << (i-22);
		head = 1 << 7;
		cyl  = size >> (i-15);
	}
	put_user (head, (u8 *)&geo->heads);
	put_user (sect, (u8 *)&geo->sectors);
	put_user (cyl, (u16 *)&geo->cylinders);
	put_user (0,   (u32 *)&geo->start);
	return 0;
}

/// Copies user data into a memory, that is newly allocated using vmalloc.
static inline char* alloc_copy_user_v(const void __user *from, size_t nbytes)
{
    char* data = (char*) VMALLOC(nbytes);
	if(!data) {
		printk(KERN_ERR "bestcrypt: can't alloc mamory for user buffer sized %zu bytes\n", nbytes);
		return NULL;
	}

	if(0 != copy_from_user(data, from, nbytes)) {
		printk(KERN_ERR "bestcrypt: can't copy user data sized %zu bytes\n", nbytes);
        VFREE(data);
		return NULL;
	}

	return data;
}

/// Copies user data into a memory, that is newly allocated using kmalloc(GFP_KERNEL).
/// @warning Cannot be used to allocate a big amount of data.
static inline char* alloc_copy_user_k(const void __user *from, size_t nbytes)
{
    char* data = (char*) KMALLOC(nbytes);
    if(!data) {
        printk(KERN_ERR "bestcrypt: can't alloc mamory for user buffer sized %zu bytes\n", nbytes);
        return NULL;
    }

    if(0 != copy_from_user(data, from, nbytes)) {
        printk(KERN_ERR "bestcrypt: can't copy user data sized %zu bytes\n", nbytes);
        KFREE(data);
        return NULL;
    }

    return data;
}


static int bc_ioctl(struct block_device *bdev, unsigned mode, u_int cmd, u_long arg)
{
	struct bc_device* dev;
    BC_DEBUG_LOG("IOCtl to bestcrypt device.\n");
    if(!bdev) {
        BC_KERN_ERROR("no block device in ioctl.\n");
		return -EINVAL;
	}

	dev = bdev->bd_disk->private_data;
	if(!dev) {
        BC_KERN_ERROR("not a BestCrypt block device.\n");
		return -ENODEV;
	}

    if (!dev->is_alive) {
        BC_KERN_ERROR("device is going down, but catched an ioctl.\n");
        return -ENODEV;
    }

	switch(cmd) {
    case kBCCTL_DiskSetFd: {
        int error;
        char* pool;
        Extent *extent_bytes;
        struct bc_diskctl_set_fd ctl;

        BC_DEBUG_LOG("ioctl: DiskSetFd. Copy struct from user-space.\n");
        // IO control struct.
        if(0 != copy_from_user(&ctl, (struct bc_diskctl_set_fd*) arg, sizeof(ctl))) {
            BC_KERN_ERROR("Could not copy set fd arguments from user.\n");
            return -EFAULT;
        }


        // Pool.
	    if(ctl.pool.value) {
            BC_DEBUG_LOG("Copy pool from user-space.\n");
            pool = alloc_copy_user_k(ctl.pool.value, 512);
		    if(!pool) {
			    BC_KERN_ERROR("can't allocate and copy user pool data\n");
			    return -ENOMEM;
		    }
	    } else {
            pool = NULL;
        }

        // Extents.
        if (!ctl.extents.value || !ctl.extent_count) {
            KFREE(pool);
            BC_KERN_ERROR("extent map is null or empty\n");
            return -EINVAL;
        }
        else {
            BC_DEBUG_LOG("Copy extents from the user-space.\n");
            // The extent bytes will be liked in device structure in case of success.
            extent_bytes = (Extent*)alloc_copy_user_k(ctl.extents.value, ctl.extent_count * sizeof(Extent));
            BC_DEBUG_LOG("copied %llu bytes with extents from user.\n", ctl.extent_count * sizeof(Extent));

            if (!extent_bytes)
            {
                BC_KERN_ERROR("can't copy extent array from user space.\n");
                KFREE(pool);
                return -ENOMEM;
            }
        }

        BC_DEBUG_LOG("Space map copied, count: %llu, first extent start: %llu, length: %llu\n",
                     (unsigned long long)ctl.extent_count, (unsigned long long)extent_bytes[0].start, (unsigned long long)extent_bytes[0].length);

        BC_DEBUG_LOG("Linking to device.\n");
        error = bc_dev_link(bdev, ctl.fd, ctl.alg_id, ctl.mode_id, ctl.key_handle, pool,
                            extent_bytes, ctl.extent_count, ctl.options);
        if (pool) {
            KFREE(pool);
            pool = NULL;
        }

        if (extent_bytes) {
            KFREE(extent_bytes);
            extent_bytes = NULL;
        }

        if(error) {
            return error;
        }

        BC_DEBUG_LOG("Completed.\n");
        return 0;
    }

	case kBCCTL_DiskClearFd: {
        BC_DEBUG_LOG("ioctl:DiskClearFd.\n");
        return bc_dev_unlink(dev, (int) arg); // this is used for forced unmount by removing disk directly from our device.
    }

#if (LINUX_VERSION_CODE < KERNEL_VERSION(2,6,16))
	case HDIO_GETGEO:	        return bc_getgeo(bdev, (struct hd_geometry *) arg);
#endif

    case BLKRRPART: {
        BC_KERN_ERROR("BLKRRPART\n");
        return -ENOTTY;
    }

    case CDROMMULTISESSION:       // ioctl 0x5310
    case CDROM_GET_CAPABILITY:    // ioctl 0x5331
    case CDROM_LAST_WRITTEN:      // ioctl 0x5395
        // silently ignore these commands. They flood kernel ('dmesg' buffer) log.
        return -ENOTTY;

	default:
        BC_KERN_ERROR("Unknown ioctl 0x%x\n", cmd);
        return -ENOTTY;
	};

	return 0;
}


/*
 * open, release and ioctl signatures have changed starting from 2.6.28.
 * use wrappers over new implementations for older kernels.
 */
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28))

static struct block_device_operations g_dev_bdops = {
	.owner 		= THIS_MODULE,
	.open 		= bc_open,
	.release 	= bc_release,
	.ioctl 		= bc_ioctl,
	.getgeo 	= bc_getgeo,
};

#else

static struct block_device_operations g_dev_bdops = {
	.owner 		= THIS_MODULE,
	.open 		= bc_open_wrapper,
	.release 	= bc_release_wrapper,
	.ioctl 		= bc_ioctl_wrapper,

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,16))
	.getgeo 	= bc_getgeo, // for < 2.6.16 this is an HDIO_GETGEO ioctl
#endif
};


static int bc_open_wrapper(struct inode *inode, struct file *file) {
	if(NULL == inode)
		return -EINVAL;

	return (file != NULL) ? bc_open(inode->i_bdev, file->f_mode) : bc_open(inode->i_bdev, 0);
}

static int bc_release_wrapper(struct inode *inode, struct file *file) {
	if (NULL == inode)
	    return -EINVAL;

	return (file != NULL) ? bc_release(inode->i_bdev->bd_disk, file->f_mode) : bc_release(inode->i_bdev->bd_disk, 0);
}

static int bc_ioctl_wrapper(struct inode *inode, struct file *file, u_int cmd, u_long arg) {
	if (NULL == inode)
		return -EINVAL;

	return (file != NULL) ? bc_ioctl(inode->i_bdev, file->f_mode, cmd, arg) : bc_ioctl(inode->i_bdev, 0, cmd, arg);
}


#endif // (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,28))

void bc_remove_device_list(void)
{
    VFREE(g_avail_devs);
}

int bc_create_disk(struct bc_device **out_dev)
{
    int dev_num, error;
    struct bc_device* dev = KMALLOC(sizeof(struct bc_device));

    BC_DEBUG_LOG("starting to create new disk.\n");
    (*out_dev) = NULL;

	if(!dev) {
        BC_KERN_ERROR("Cannot allocate device struct.\n");
		return -ENOMEM;
	}
    BC_DEBUG_LOG("getting a number for the new device.\n");
    dev_num = bc_get_avail_dev_num();
    if (-1 == dev_num) {
        BC_KERN_ERROR("current device limit exceed.\n");
        return -ENODEV;
    }

	memset(dev, 0, sizeof(*dev));

    BC_DEBUG_LOG("allocating gendisk structure.\n");
	dev->gendisk = alloc_disk(g_max_parts);
	if(!dev->gendisk) {
        BC_KERN_ERROR("not enough memory to allocate gendisk.\n");
		bc_dev_free(dev);
		return -ENOMEM;
	}

    BC_DEBUG_LOG("allocating block queue.\n");
	dev->blkqueue = blk_alloc_queue(GFP_KERNEL);
	if(!dev->blkqueue) {
        BC_KERN_ERROR("not enough memory to allocate queue.\n");
		bc_dev_free(dev);
		return -ENOMEM;
    }

    BC_DEBUG_LOG("allocating device write buffer.\n");
    dev->writebuf = KMALLOC(PAGE_SIZE);
	if(!dev->writebuf) {
        BC_KERN_ERROR("not enough memory to allocate buffer.\n");
		bc_dev_free(dev);
        return -ENOMEM;
	}

    BC_DEBUG_LOG("init for write semaphore");
    sema_init(&dev->writesem, 1);

    BC_DEBUG_LOG("initializing related kernel structures.\n");
	bio_list_init(&dev->queue);
	spin_lock_init(&dev->lock);
	spin_lock_init(&dev->queuelock);
    sema_init(&dev->queuesem, 0);

    BC_DEBUG_LOG("filling device structure.\n");
	dev->gendisk->major 		= g_dev_major;
    dev->gendisk->first_minor   = dev_num * g_max_parts;
    dev->gendisk->fops 		    = &g_dev_bdops;
	dev->gendisk->queue		    = dev->blkqueue;
	dev->blkqueue->queuedata 	= dev;

    // The flags below indicate that the device
    // accepts TRIM operations. This will work in
    // case if FS supports it.

//    queue_flag_set_unlocked(QUEUE_FLAG_DISCARD, dev->blkqueue);
//    queue_flag_set_unlocked(QUEUE_FLAG_SECDISCARD, dev->blkqueue);
	dev->gendisk->flags 		= 0;
    dev->gendisk->private_data	= dev;
    dev->extents                = NULL;
    dev->extent_count           = 0;
    dev->is_alive               = 1; // device is now ready for work.

    BC_DEBUG_LOG("getting a sysfs device name.\n");
    bc_get_dev_name(dev_num, dev->gendisk->disk_name, DISK_NAME_LEN);
    BC_DEBUG_LOG("device name is %s\n", dev->gendisk->disk_name);

    BC_DEBUG_LOG("setting device capacity to zero.\n");
	set_capacity(dev->gendisk, 0L);

    BC_DEBUG_LOG("publishing disk to kernel.\n");
    add_disk(dev->gendisk);

    BC_DEBUG_LOG("incrementing module reference.\n");
	try_module_get(THIS_MODULE);

    BC_DEBUG_LOG("creating a sysfs entry.\n");

    error = sysfs_create_file(&disk_to_dev(dev->gendisk)->kobj, &g_dev_backing_store_attr.attr);
    if(error) {
        BC_KERN_ERROR("could not publish sysfs property for new device.\n");
    }

    BC_DEBUG_LOG("finishing creating a new disk.");
    (*out_dev) = dev;
    return 0;
}

