/*
 * Copyright (c) 2010 Jetico, Inc., Finland
 * All rights reserved.
 *
 * Trasparently encrypted block device definition.
 */

#ifndef BC_KERNEL_DEV_H
#define BC_KERNEL_DEV_H

#include <linux/version.h>
#include <linux/genhd.h>
#include <linux/bio.h>
#include <linux/blkdev.h>
#include <linux/sched.h>

#include "bc_crypto.h"
#include "spacemap/extent.h"
#include "spacemap/spacemap.h"


/* bc device description */
struct bc_device 
{
    struct file*            filp;	    	// container file pointer we are emulating.
    struct bc_alg           alg;            // algorithm and key.
	char			        pool[512];		// random data pool.
//  uint64_t		        start_sector;	// sector number where container data starts.
//  uint64_t		        nsectors;   	// total number of container data sectors.
    Extent*                 extents;        // array with extents in bytes.
    uint64_t                extent_count;   // element count in array with extents.

	struct bio_list		    queue;	    	// actual request queue.
	spinlock_t		        queuelock;	    // queue access lock.
	struct semaphore	    queuesem;	    // semaphore that signals new requests.
    struct semaphore        writesem;       // semaphore locking a write buffer.
	char*			        writebuf;	    // PAGE_SIZE'd buffer to hold copy of the data during write request.
	struct task_struct*	    thread;		    // request consumer thread.

	struct block_device*    bdev;		    // whole disk block device node.
	struct gendisk*		    gendisk;	    // underlying gendisk
	struct dentry*          dentry;	        // container file dentry
	int			            gfpmask;        
	int	                    refcnt;		    // disk open count. when reaches zero and auto flag is set the device will go away.
    int                     is_alive;       // 1, if device is still working. If 0, device is in the process of unlinking.
	spinlock_t              lock;		    // protects device data changes.

	// allocated block level request queue.
	// we don't use default queue because we are not a real disk.
	// instead we allocate it to recieve requests directly from block level.
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,24))
	struct request_queue*	blkqueue;
#elif (LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0))
	request_queue_t*	    blkqueue;
#endif
	
	// device flags.
	struct bc_device_flags {
        unsigned	autoremove  :1; // device will go away as soon as last reference holder will close it.
        unsigned	configured  :1; // device is fully initialized. when not set device either goes away or otherwise not yet initialized.
        unsigned	readonly    :1; // error on all write operations.
        unsigned	reserved    :29;
	} flags;
};

typedef void(*bc_request_processor_t)(struct bc_device* dev, struct bio* bio);


/**
 * Prepare blank device node. Return value indicates error.
 */
int bc_create_disk(struct bc_device **out_dev);

/**
 * Removes list containing device names available.
*/
void bc_remove_device_list(void);

#ifdef BC_DEBUG

/**
 * Check unfreed memory link count during device unloading in debug mode.
 */
void bc_check_unfreed_mem(void);

#endif
/**
 * Tell device name.
 */
static inline const char* bc_dev_name(struct bc_device* dev) {
	return dev->gendisk->disk_name;
}

/**
 * Tell device read only status.
 */
static inline int bc_dev_is_read_only(struct bc_device* dev) {
	return dev->flags.readonly;
}

/**
 * Tell if device is configured.
 */
static inline int bc_dev_is_configured(struct bc_device* dev) {
	return dev->flags.configured;
}

/**
 * Tell if device is set for autoremove.
 */
static inline int bc_dev_is_autoremove(struct bc_device* dev) {
	return dev->flags.autoremove;
}

#endif

