#include "bc_mem.h"
#include "bc_defs.h"

#include <asm/bug.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/kernel.h>
#include <asm/uaccess.h>

#define BC_GET_CALLER_PTR __builtin_return_address(0)
#define BC_FORCE_NO_OPTIMIZATION asm("");

/// Keeps arrays of all allocated memory addresses.
struct bc_alloc {
    void *kptr[BC_ALLOC_MAX];       ///< Allocs with kmalloc.
    void *kfunc_ptr[BC_ALLOC_MAX];  ///< Caller function pointers for kmalloc.

    void *vptr[BC_ALLOC_MAX];       ///< Allocs with vmalloc.
    void *vfunc_ptr[BC_ALLOC_MAX];  ///< Caller function pointers for vmalloc.

    int initialized;
};

static struct bc_alloc g_alloc = {
    .initialized = 0
};

#define BC_PTR(alloc_virtual) ((alloc_virtual) ? g_alloc.vptr : g_alloc.kptr)
#define BC_FUNC_PTR(alloc_virtual) ((alloc_virtual) ? g_alloc.vfunc_ptr : g_alloc.kfunc_ptr)

/// Does initialization of pointer array, if necessary.
static void check_init(int alloc_virtual)
{
    void **ptr = BC_PTR(alloc_virtual);
    if (unlikely(!g_alloc.initialized)) {
        memset(ptr, 0, BC_ALLOC_MAX * sizeof(void*));
        g_alloc.initialized = 1;
    }
}

noinline
void* alloc_wrapper(size_t size, int alloc_virtual) //BC_NOINLINE
{
    void **ptr = BC_PTR(alloc_virtual);
    void **kfunc = BC_FUNC_PTR(alloc_virtual);
    int i = 0;

    BC_FORCE_NO_OPTIMIZATION;

    check_init(alloc_virtual);

    for (; i < BC_ALLOC_MAX; ++i) {
        if (likely(!ptr[i])) {
            ptr[i] = (alloc_virtual) ? vmalloc(size) : kmalloc(size, GFP_KERNEL);
            kfunc[i] = BC_GET_CALLER_PTR;
            return ptr[i];
        }
    }

    BC_KERN_ERROR("alloc_wrapper: limit exceed for %s\n", (alloc_virtual)?
                     "virtual" : "kernel");
    return NULL;
}

noinline
void free_wrapper(void *mem, int alloc_virtual) //BC_NOINLINE
{
    void **ptr = BC_PTR(alloc_virtual);
    void **kfunc = BC_FUNC_PTR(alloc_virtual);
    int i = 0;

    BC_FORCE_NO_OPTIMIZATION;

    // In case that there is wrong memory freeing before the first allocation.
    check_init(alloc_virtual);

    for (; i < BC_ALLOC_MAX; ++i) {
        if (ptr[i] == mem) {
            if (alloc_virtual) {
                vfree(mem);
            } else {
                kfree(mem);
            }
            ptr[i] = NULL;
            kfunc[i] = NULL;
            return;
        }
    }

    BC_KERN_ERROR("free_wrapper: no such adress is allocated in %s memory.\n",
                 (alloc_virtual)? "virtual" : "kernel");
    dump_stack();
    BUG();
}

void print_unfree_memory(void)
{
    int alloc_virtual = 1; // virtual

    void **ptr = BC_PTR(alloc_virtual);
    void **kfunc = BC_FUNC_PTR(alloc_virtual);
    int i = 0, total = 0;

    check_init(alloc_virtual);

    BC_KERN_ERROR("virtual:\n");

    for (; i < BC_ALLOC_MAX; ++i) {
        if (ptr[i]) {
            BC_KERN_ERROR("unfreed memory at %p...\n", ptr[i]);
            BC_KERN_ERROR("was allocated from %pS\n", kfunc[i]);
            //free_wrapper(ptr[i], alloc_virtual);
            ++total;
        }
    }

    alloc_virtual = 0; // physical

    ptr = BC_PTR(alloc_virtual);
    kfunc = BC_FUNC_PTR(alloc_virtual);

    check_init(alloc_virtual);

    BC_KERN_ERROR("physical:\n");

    for (i = 0; i < BC_ALLOC_MAX; ++i) {
        if (ptr[i]) {
            BC_KERN_ERROR("unfreed memory at %p...\n", ptr[i]);
            BC_KERN_ERROR("was allocated from %pS\n", kfunc[i]);
            //free_wrapper(ptr[i], alloc_virtual);
            ++total;
        }
    }

    BC_KERN_ERROR("total: %d unfreed memory blocks found.\n", total);
}
