/*
 * Copyright (c) 2010 Jetico, Inc., Finland
 * All rights reserved.
 *
 * Controller module.
 */

#include "BCKernel.h"

#include "bc_ctl.h"
#include "bc_defs.h"
#include "bc_dev.h"
#include "bc_crypto.h"

#include <linux/init.h>
#include <linux/module.h>
#include <linux/file.h>
#include <linux/fs.h>
#include <linux/miscdevice.h>

#include <asm/uaccess.h>

#include "bc_mem.h"

/***********************************************************************************************
 *	Cipher plugins.
 *
 *	Each cipher plugin is a separate module.
 *	On module init it should call bc_register_plugin to pass it's cipher definition.
 *	On module exit is should call bc_unregister_plugin.
 *	We increase plugin's owning module refcount when cipher is in use.
 *
 *	Each registered plugin will have it's own sysfs folder in our "plugins" group.
 *	This folder will contain files with plugin attributes like id, block size, etc.
 *
 ***********************************************************************************************/


/**
 * Plugin bookkeeping structure.
 * Each structure is a list element.
 */
struct bc_plugin {
    const struct bc_cipher*     cipher;		// exported cipher.
    struct module*              owner;		// owning module to use for reference counting.
    struct list_head            list_entry;	// entry in plugins list.
    struct kobject              kobj;		// for directory with plugin's attributes in sysfs.
};


static struct list_head 	g_plugins_list;
static rwlock_t 		g_plugins_lock;
static struct kobject* 		g_plugins_dir;


#define BC_PLUGINS_SHARED_LOCK          read_lock(&g_plugins_lock);
#define BC_PLUGINS_SHARED_UNLOCK        read_unlock(&g_plugins_lock);
#define BC_PLUGINS_EXCLUSIVE_LOCK       write_lock(&g_plugins_lock);
#define BC_PLUGINS_EXCLUSIVE_UNLOCK     write_unlock(&g_plugins_lock);


/*
 *	Sysfs attributes and kobject type.
 *	We have our own kobj_type to initialize attributes in group for new plugins.
 */


// Overloaded plugin attribute for each file.
struct bc_plugin_attribute {
	struct attribute attr;
	ssize_t (*show)(struct bc_plugin* plug, struct bc_plugin_attribute* attr, char* buf);
	ssize_t (*store)(struct bc_plugin* plug, struct bc_plugin_attribute* attr, char* buf);
};


#define BC_CONTAINING_PLUGIN(_kobj) 		container_of(_kobj, struct bc_plugin, kobj)
#define BC_CONTAINING_ATTRIBUTE(_attr) 		container_of(_attr, struct bc_plugin_attribute, attr)
#define BC_DECLARE_PLUGIN_ATTRIBUTE2(_name) 	static struct bc_plugin_attribute plug_attr_##_name = __ATTR(_name, S_IRUGO, show_plugin_##_name, NULL)
#define BC_DECLARE_PLUGIN_ATTRIBUTE(_name)	BC_DECLARE_PLUGIN_ATTRIBUTE2(_name)
#define BC_PLUGIN_ATTRIBUTE2(_name) 		plug_attr_##_name
#define BC_PLUGIN_ATTRIBUTE(_name) 		BC_PLUGIN_ATTRIBUTE2(_name)


/*
 * Generic untyped attribute handlers.
 * These are called with untyped kobject/attribute pointers for all attributes.
 * We extract real plugin attribute and forward requests to its handlers.
 */


static ssize_t plugin_show_attr(struct kobject* kobj, struct attribute* attr, char* page_buf) 
{
	struct bc_plugin* plug = BC_CONTAINING_PLUGIN(kobj);
	struct bc_plugin_attribute* plug_attr = BC_CONTAINING_ATTRIBUTE(attr);
	
	return plug_attr->show(plug, plug_attr, page_buf);
}

static ssize_t plugin_store_attr(struct kobject* kobj, struct attribute* attr, const char* page_buf, size_t nbytes) {
	BUG(); // all attributes are read-only.
	return -1;
}

static void release_plugin_kobj(struct kobject* kobj) {
	BC_DEBUG_LOG("bc_ctl: plugin kobject %p released\n", kobj);
}

/*
 * Plugin attributes.
 * Typed plugin attributes that generic requests forward to.
 */


static ssize_t show_plugin_name(struct bc_plugin* plug, struct bc_plugin_attribute *attr, char *buf) {
	return snprintf(buf, PAGE_SIZE, "%s", plug->cipher->name);
}


static ssize_t show_plugin_id(struct bc_plugin* plug, struct bc_plugin_attribute *attr, char *buf) {
	return snprintf(buf, PAGE_SIZE, "%d", plug->cipher->id);
}


static ssize_t show_plugin_block_size(struct bc_plugin* plug, struct bc_plugin_attribute *attr, char *buf) {
	return snprintf(buf, PAGE_SIZE, "%zu", plug->cipher->block_size);
}


static ssize_t show_plugin_key_length(struct bc_plugin* plug, struct bc_plugin_attribute *attr, char *buf) {
	return snprintf(buf, PAGE_SIZE, "%zu", plug->cipher->key_size);
}


static ssize_t show_plugin_mode_key_size(struct bc_plugin* plug, const struct bc_mode* mode, char* buf)
{
	if(mode->is_supported(plug->cipher)) {
		return snprintf(buf, PAGE_SIZE, "%zu", mode->key_size(plug->cipher));
	} else {
		return snprintf(buf, PAGE_SIZE, "%d", 0);
	}
}

static ssize_t show_plugin_cbc_key_size(struct bc_plugin* plug, struct bc_plugin_attribute *attr, char *buf) {
	return show_plugin_mode_key_size(plug, bc_lookup_mode(kCBC), buf);
}


static ssize_t show_plugin_lrw_key_size(struct bc_plugin* plug, struct bc_plugin_attribute *attr, char *buf) {
	return show_plugin_mode_key_size(plug, bc_lookup_mode(kLRW), buf);
}


static ssize_t show_plugin_xts_key_size(struct bc_plugin* plug, struct bc_plugin_attribute *attr, char *buf) {
	return show_plugin_mode_key_size(plug, bc_lookup_mode(kXTS), buf);
}


BC_DECLARE_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_NAME_ATTR);
BC_DECLARE_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_ID_ATTR);
BC_DECLARE_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_BLOCK_SIZE_ATTR);
BC_DECLARE_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_KEY_LENGTH_ATTR);
BC_DECLARE_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_CBC_KEY_SIZE_ATTR);
BC_DECLARE_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_LRW_KEY_SIZE_ATTR);
BC_DECLARE_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_XTS_KEY_SIZE_ATTR);

static struct sysfs_ops g_plugin_sysfs_ops = 
{
	plugin_show_attr,
	plugin_store_attr,
};


static struct attribute* g_plugin_attributes[] = 
{
  	&BC_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_NAME_ATTR).attr,
  	&BC_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_ID_ATTR).attr,
  	&BC_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_BLOCK_SIZE_ATTR).attr,
  	&BC_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_KEY_LENGTH_ATTR).attr,
  	&BC_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_CBC_KEY_SIZE_ATTR).attr,
  	&BC_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_LRW_KEY_SIZE_ATTR).attr,
  	&BC_PLUGIN_ATTRIBUTE(BC_CTL_PLUGIN_XTS_KEY_SIZE_ATTR).attr,
       	NULL,
};
	
 
// kobj_type of all plugins with its default attributes and handlers set.
static struct kobj_type g_plugin_kobj_type = 
{
	.release 	    = release_plugin_kobj,
	.sysfs_ops	    = &g_plugin_sysfs_ops,
    .default_attrs	= g_plugin_attributes,
};


static struct bc_plugin* lookup_plugin_nolock(int id) 
{
	struct bc_plugin* plug;

	list_for_each_entry(plug, &g_plugins_list, list_entry) {
		if(plug->cipher->id == id) {
			return plug;
		}
	}

	return NULL;
}


// Can be called on partially initialized plugins.
static void free_plugin(struct bc_plugin* plug) 
{
	if(!plug) {
		return;
	}

	BUG_ON(module_refcount(plug->owner));

	kobject_put(&plug->kobj);
    KFREE(plug);
}


int bc_init_plugins_register(struct kobject* parent)
{
	INIT_LIST_HEAD(&g_plugins_list);
	rwlock_init(&g_plugins_lock);	
	
	g_plugins_dir = kobject_create_and_add(BC_CTL_PLUGINS_GROUP_STR, parent);
	if(!g_plugins_dir) {
		printk(KERN_ERR "bc_ctl: failed creating sysfs plugins directory\n");
		return -ENOMEM;
	}

	return 0;
}


int bc_register_plugin(struct module* mod, const struct bc_cipher* cipher, struct bc_plugin** out_ref)
{
	int rc;
	struct bc_plugin* plug = NULL;
	
	// lock now to protect against race when two or more ciphers with the same id try to register.
	// this may seem like an overkill, but plugin register rarely.
	BC_PLUGINS_EXCLUSIVE_LOCK;

	// will not return from loop if ok.
	do {
		if(NULL != lookup_plugin_nolock(cipher->id)) {
			rc = -EBUSY;
			break;
		}

        plug = (struct bc_plugin*)KMALLOC(sizeof(struct bc_plugin));
		if(!plug) {
			rc = -ENOMEM;
			break;
		}

		memset(plug, 0, sizeof(*plug));

 		rc = kobject_init_and_add(&plug->kobj, &g_plugin_kobj_type, g_plugins_dir, "%s", cipher->name);
		if(0 != rc) {
			break;
		}

		plug->cipher = cipher;
		plug->owner = mod;
	
		list_add(&plug->list_entry, &g_plugins_list);

		BC_PLUGINS_EXCLUSIVE_UNLOCK;

		*out_ref = plug;
	
		BC_DEBUG_LOG("bc_ctl: registered cipher %s (id 0x%x)\n", cipher->name, cipher->id);
		return 0;
	
	} while(0);

	// if we reach this, it means error.

	BC_PLUGINS_EXCLUSIVE_UNLOCK;
	free_plugin(plug);

	return rc;
}

EXPORT_SYMBOL(bc_register_plugin);


void bc_unregister_plugin(struct bc_plugin* plug)
{
	BC_DEBUG_LOG("unregistered cipher %s (id 0x%x)\n", plug->cipher->name, plug->cipher->id);

	BC_PLUGINS_EXCLUSIVE_LOCK;
	list_del(&plug->list_entry);
	free_plugin(plug);
	BC_PLUGINS_EXCLUSIVE_UNLOCK;

}

EXPORT_SYMBOL(bc_unregister_plugin);


const struct bc_cipher* bc_get_plugin(int id)
{
	struct bc_plugin* plug;
	
	BC_PLUGINS_SHARED_LOCK;
	plug = lookup_plugin_nolock(id);
	if(plug) {
		try_module_get(plug->owner);
	}
	BC_PLUGINS_SHARED_UNLOCK;

    return (plug ? plug->cipher : NULL);
}


void bc_put_plugin(const struct bc_cipher* cipher)
{
	struct bc_plugin* plug;
	
	BC_PLUGINS_SHARED_LOCK;
	plug = lookup_plugin_nolock(cipher->id);
	if(plug) {
		module_put(plug->owner);
	}
	BC_PLUGINS_SHARED_UNLOCK;
}	


void bc_free_plugins_register(void)
{
	BUG_ON(!list_empty(&g_plugins_list));
	if(g_plugins_dir) kobject_put(g_plugins_dir);
}



/***********************************************************************************************
 *
 *  Controller misc device.
 *
 *  Implementation /dev/bestcrypt control device and its ioctls
 *  - provides information about registered cipher plugins to user space through sysfs node.
 *  - implements requests to store encryption keys, encrypt/decrypt data, publish blank device nodes.
 *
 ***********************************************************************************************/


// Block device major number.
// Initialised on init, exported for bc_mod.c
int g_dev_major = 0;


static inline char* alloc_copy_user_k(const void __user *from, size_t nbytes)
{
    char* data = (char*) KMALLOC(nbytes);
	if(!data) {
		printk(KERN_ERR "bestcrypt: can't alloc mamory for user buffer sized %zu bytes\n", nbytes);
		return NULL;
	}

    if(0 != copy_from_user(data, from, nbytes)) {
        printk(KERN_ERR "bestcrypt: can't copy user data sized %zu bytes\n", nbytes);
        KFREE(data);
		return NULL;
	}

	return data;
}

static inline char* alloc_copy_user_v(const void __user *from, size_t nbytes)
{
    char* data = (char*) VMALLOC(nbytes);
    if(!data) {
        printk(KERN_ERR "bestcrypt: can't alloc mamory for user buffer sized %zu bytes\n", nbytes);
        return NULL;
    }

    if(0 != copy_from_user(data, from, nbytes)) {
        printk(KERN_ERR "bestcrypt: can't copy user data sized %zu bytes\n", nbytes);
        VFREE(data);
        return NULL;
    }

    return data;
}


static int get_info(struct bc_ctl_info* out_arg)
{
    int error;
    struct bc_ctl_info info;

    info.ver_major = BC_CTL_VERSION_MAJOR;
    info.ver_minor = BC_CTL_VERSION_MINOR;
    info.block_major = g_dev_major;

    error = copy_to_user(out_arg, &info, sizeof(info));
    if(error) {
        BC_KERN_ERROR("can't copy user info argument\n");
        return error;
    }

    return 0;
}


static int store_key(struct bc_ctl_key* arg)
{
	int 			error;
	int 			handle;
	struct bc_ctl_key 	ctl;
	struct bc_key* 		key; // do not free this on success - data will be stored in table.

	error = copy_from_user(&ctl, arg, sizeof(ctl));
	if(error) {
		BC_KERN_ERROR("can't copy user argument\n");
		return -EFAULT;
	}
	
	key = &ctl.key;	
    key->data.value = alloc_copy_user_k(key->data.value, key->size);
	if(!key->data.value) {
		BC_KERN_ERROR("can't allocate and copy user key data\n");
		return -ENOMEM;
	}
		
	error = bc_crypto_store_key(key->data.value, key->size, &handle);
	if(error) {
		BC_KERN_ERROR("failed storing key in table\n");
        KFREE(key->data.value);
		return error;
	}

	error = copy_to_user(ctl.out_handle.value, &handle, sizeof(handle));
	if(error) {
		BC_KERN_ERROR("can't copy key handle to user\n");
        bc_crypto_free_key(handle);
		return error;
	}
	
	return 0;
}


static int free_key(int arg) {
	return bc_crypto_free_key(arg);
}


static int process(struct bc_ctl_process* arg)
{
	int                   error;
	struct bc_ctl_process ctl;
	struct bc_alg         alg;
	
	char*           data = NULL;
	char*           pool = NULL;
	struct bc_key*  key  = NULL; // may not be initialized if user did not supplied key with request.

	if (NULL == arg) {
	    BC_KERN_ERROR("invalid argument for processing.");
	    return -EINVAL;
	}

	error = copy_from_user(&ctl, arg, sizeof(ctl));
	if(error) {
		BC_KERN_ERROR("can't copy argument data from user\n");
		return -EFAULT;
	}

    if (NULL == ctl.data.value && 0 != ctl.nbytes) {
        BC_KERN_ERROR("invalid data buffer.");
        return -EINVAL;
    }

    data = alloc_copy_user_v(ctl.data.value, ctl.nbytes);
	if(!data) {
		BC_KERN_ERROR("can't allocate and copy user processing data with size %llu\n", ctl.nbytes);
		error = -ENOMEM;
		goto CLEANUP;
	}

	if(ctl.pool.value) {
        pool = alloc_copy_user_k(ctl.pool.value, kBCKernel_PoolSize);
		if(!pool) {
			BC_KERN_ERROR("can't allocate and copy user pool data\n");
			error = -ENOMEM;
			goto CLEANUP;
		}
	}

	error = bc_crypto_get_alg(ctl.alg_id, ctl.mode_id, &alg);
	if(error) {
		BC_KERN_ERROR("could not configure algorithm with id 0x%x and mode 0x%x\n", ctl.alg_id, ctl.mode_id);
		goto CLEANUP;
	}
 
	// set algorithm key either from user supplied data or stored handle.
	// in any case after that we should have valid crypto context or error.
	if(ctl.has_key) {
		key = &ctl.key.inlined;
        key->data.value = alloc_copy_user_k(key->data.value, key->size);
		if(!key->data.value) {
			BC_KERN_ERROR("can't allocate and copy user key data with size %llu\n", key->size);
			error = -ENOMEM;
			goto CLEANUP;
		}

		error = bc_crypto_set_key_explicit(&alg, key->data.value, key->size);
	} else {
		error = bc_crypto_set_key(&alg, ctl.key.handle);
	}	

	if(error) {
		BC_KERN_ERROR("can't set algorithm key\n");
		bc_crypto_put_alg(&alg);
		goto CLEANUP;
	}


	// finally actual processing.
	if(ctl.proc_type == kBCProcess_Encrypt) {
		bc_crypto_encrypt(&alg, ctl.iv, pool, data, ctl.nbytes);
	} else {
		bc_crypto_decrypt(&alg, ctl.iv, pool, data, ctl.nbytes);
	}

    bc_crypto_put_alg(&alg);

	copy_to_user(ctl.data.value, data, ctl.nbytes);
	
CLEANUP:
	
    if(data) VFREE(data);
    if(pool) KFREE(pool);
    if(ctl.has_key && key && key->data.value) {
		memset(key->data.value, 0, key->size);
        KFREE(key->data.value);
	}

	return error;
}

static unsigned long get_time(void)
{
    struct timespec ts;
    getnstimeofday(&ts);
    return (unsigned long) ts.tv_sec * 1000000000L + ts.tv_nsec;
}

static int benchmark(struct bc_ctl_bench* arg)
{
    // Values are doesn't matter here, so simply keep them as they are.
    char                    iv[8],
                            *buffer = NULL,
                            *pool = NULL,
                            *key_data = NULL;

    const unsigned int      number_of_bench = 16;
    const unsigned long     buffer_size = 1024L * 1024L;
    const unsigned int      key_size = 64;

    int                     error = 0;
    unsigned int            i = 0;
    struct bc_ctl_bench     bench;
    struct bc_alg           alg;

    // Here one is having time in nanoseconds.
    unsigned long           start_time = 0,
                            end_time = 0,
                            average_time = 0;

    BC_DEBUG_LOG("benchmark started");
    if (!arg) {
        BC_KERN_ERROR("argument is null");
        return -EINVAL;
    }

    // Copying benchmark params from user space.
    error = copy_from_user(&bench, arg, sizeof(bench));
    if (error) {
        BC_KERN_ERROR("can't copy argument data from user-space\n");
        return -EFAULT;
    }

    // Initializing algorithm.
    error = bc_crypto_get_alg(bench.alg_id, bench.alg_mode, &alg);
    if (error) {
        BC_KERN_ERROR("failed configuring an algorithm with id 0x%x and mode 0x%x\n", bench.alg_id, bench.alg_mode);
        return error;
    }

    // Allocating memory for key.
    key_data = (char*)KMALLOC(key_size);
    if (!key_data) {
        BC_KERN_ERROR("can't allocate memory for key\n");
        error = -ENOMEM;
        goto BC_BENCHMARK_PUT_ALG;
    }

    // Allocating pool.
    pool = (char*)KMALLOC(512);
    if (!pool) {
        BC_KERN_ERROR("can't allocate memory for pool\n");
        error = -ENOMEM;
        goto BC_BENCHMARK_PUT_ALG;
    }

    // Allocating encryption buffer.
    buffer = (char*)VMALLOC(buffer_size);
    if (!buffer) {
        BC_KERN_ERROR("can't allocate memory for buffer\n");
        goto BC_BENCHMARK_PUT_ALG;
    }

    // Load key.
    error = bc_crypto_set_key_explicit(&alg, key_data, key_size);
    if (error) {
        BC_KERN_ERROR("can't set algorithm key\n");
        goto BC_BENCHMARK_PUT_ALG;
    }

    for (; i < number_of_bench; i++) {
        start_time = get_time();
        bc_crypto_encrypt(&alg, *((uint64_t*)iv), pool, buffer, buffer_size);
        end_time = get_time();
        average_time += end_time - start_time;
    }

    average_time = average_time / ((unsigned long) number_of_bench);
    bench.nsec = average_time;

    BC_DEBUG_LOG("benchmark completed for alg 0x%x and mode 0x%x",
                 bench.alg_id, bench.alg_mode);

    if (copy_to_user(arg, &bench, sizeof(bench))) {
        BC_KERN_ERROR("error while copying results to user\n");
        error = -EFAULT;
        goto BC_BENCHMARK_FREE_DATA;
    }

BC_BENCHMARK_PUT_ALG:
    bc_crypto_put_alg(&alg);

BC_BENCHMARK_FREE_DATA:

    if (key_data)
        KFREE(key_data);

    if (pool)
        KFREE(pool);

    if (buffer)
        VFREE(buffer);

    return error;
}


static int prepare_disk(struct bc_ctl_disk* arg)
{
    int result, size;
    struct bc_ctl_disk info;
    struct bc_device* dev = NULL;

    if (!arg) {
        BC_KERN_ERROR("NULL pointer came as an argument.\n");
        return -EINVAL;
    }

    copy_from_user(&info, arg, sizeof(struct bc_ctl_disk));
    if (!info.out_name || !info.name_bufsize) {
        BC_KERN_ERROR("out_path is null.\n");
        return -EINVAL;
    }

    result = bc_create_disk(&dev);
    if(result) {
        BC_KERN_ERROR("can't setup new block device\n");
        return -EFAULT;
    }

    info.out_minor = dev->gendisk->first_minor;

    size = min(info.name_bufsize, DISK_NAME_LEN);
    copy_to_user(info.out_name, dev->gendisk->disk_name, size);

    copy_to_user(arg, &info, sizeof(struct bc_ctl_disk));

    return 0;
}


static long bc_ctl_ioctl(struct file* filp, unsigned int command, unsigned long parameter)
{
    BC_DEBUG_LOG("IO control: 0x%x\n", command);
	switch(command) {
    case kBCCTL_Info:           return get_info((struct bc_ctl_info*) parameter);
	case kBCCTL_StoreKey:       return store_key((struct bc_ctl_key*) parameter);
	case kBCCTL_FreeKey:        return free_key((int) parameter);
	case kBCCTL_Process:        return process((struct bc_ctl_process*) parameter);
    case kBCCTL_CreateDisk:     return prepare_disk((struct bc_ctl_disk*) parameter);
    case kBCCTL_Benchmark:      return benchmark((struct bc_ctl_bench*) parameter);
    default:                    BC_KERN_ERROR("Unknown ioctl: 0x%x.", command); return -ENOTTY;
	};
}

static int bc_ctl_open(struct inode* inode, struct file* filp) {
    BC_DEBUG_LOG("Device opened.");
	return 0;
}


static int bc_ctl_close(struct inode* inode, struct file* filp) {
    BC_DEBUG_LOG("Device closed.");
	return 0;
}


static struct file_operations g_ctldev_fops = {
    .owner          = THIS_MODULE,
    .open           = bc_ctl_open,
    .release        = bc_ctl_close,
	.unlocked_ioctl = bc_ctl_ioctl,
	.compat_ioctl	= bc_ctl_ioctl, // TODO: Need check.
};


static struct miscdevice g_ctldev = {
	.minor		= MISC_DYNAMIC_MINOR,
	.name		= BC_CTL_NODE_NAME,
	.fops		= &g_ctldev_fops
};


struct bc_attribute {
	struct attribute attr;
	ssize_t (*show)(struct kobject* kobj, struct attribute* attr, char* buf);
	ssize_t (*store)(struct kobject* kobj, struct attribute* attr, char* buf);
};


static ssize_t bc_show_major_version(struct kobject* kobj, struct attribute* attr, char* buf) {
	return snprintf(buf, PAGE_SIZE, "%d", BC_CTL_VERSION_MAJOR);
}


static ssize_t bc_show_minor_version(struct kobject* kobj, struct attribute* attr, char* buf) {
	return snprintf(buf, PAGE_SIZE, "%d", BC_CTL_VERSION_MINOR);
}


static ssize_t bc_show_block_major(struct kobject* kobj, struct attribute* attr, char* buf) {
    return snprintf(buf, PAGE_SIZE, "%d", g_dev_major);
}

static struct bc_attribute g_bc_major_version_attr  = __ATTR(BC_CTL_VERSION_MAJOR_ATTR, S_IRUGO, bc_show_major_version, NULL);
static struct bc_attribute g_bc_minor_version_attr  = __ATTR(BC_CTL_VERSION_MINOR_ATTR, S_IRUGO, bc_show_minor_version, NULL);
static struct bc_attribute g_bc_block_major_attr    = __ATTR(BC_CTL_BLOCK_MAJOR_ATTR, S_IRUGO, bc_show_block_major, NULL); 


static void bc_cleanup(void)
{
	if(g_dev_major) unregister_blkdev(g_dev_major, BC_BASE_DEVICE_NAME);

	sysfs_remove_file(&g_ctldev.this_device->kobj, &g_bc_major_version_attr.attr);
	sysfs_remove_file(&g_ctldev.this_device->kobj, &g_bc_minor_version_attr.attr);
	sysfs_remove_file(&g_ctldev.this_device->kobj, &g_bc_block_major_attr.attr);

	bc_free_plugins_register();
	bc_crypto_release_key_table();
    bc_remove_device_list();

	misc_deregister(&g_ctldev);
}


int __init bc_init(void)
{
	int error;

	// init device and sysfs entry.
	error = misc_register(&g_ctldev);
	if(error) {
		BC_KERN_ERROR("failed registering misc device\n");
		return error;
	}

	// this just reserves a major number for our block devices.
	g_dev_major = register_blkdev(0, BC_BASE_DEVICE_NAME);
	if(g_dev_major < 0) {
		BC_KERN_ERROR("register_blkdev failed with 0x%x\n", g_dev_major);
		error = g_dev_major;
		goto CLEANUP;
	}

	bc_crypto_init_key_table();

	error = bc_init_plugins_register(&g_ctldev.this_device->kobj);
	if(error) {
		BC_KERN_ERROR("failed initializing plugins register\n");
		goto CLEANUP;
	}

	error = sysfs_create_file(&g_ctldev.this_device->kobj, &g_bc_major_version_attr.attr);
	if(error) {
		BC_KERN_ERROR("failed creating major version attribute\n");
		goto CLEANUP;
	}

	error = sysfs_create_file(&g_ctldev.this_device->kobj, &g_bc_minor_version_attr.attr);
	if(error) {
		BC_KERN_ERROR("failed creating minor version attribute\n");
		goto CLEANUP;
	}

	error = sysfs_create_file(&g_ctldev.this_device->kobj, &g_bc_block_major_attr.attr);
	if(error) {
		BC_KERN_ERROR("failed creating minor version attribute\n");
		goto CLEANUP;
	}

	BC_DEBUG_LOG("started\n");
	return 0;
	
CLEANUP:
	bc_cleanup();
	return error;
}


void __exit bc_exit(void)
{
	bc_cleanup();
    PRINT_UNFREE_MEM();
	BC_DEBUG_LOG("stopped\n");
}


module_init(bc_init);
module_exit(bc_exit);

MODULE_AUTHOR("Jetico, Inc.");
MODULE_DESCRIPTION("BestCrypt encrypted block device driver");
MODULE_LICENSE("GPL");

