/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 *
 * CBC Encryption mode implementation.
 * BestCrypt CBC mode is different from the books.
 * We use additional 512 byte random data pool supplied with the key to xor it into input data.
 */

#include "intsize.h"
#include "mode.h"
#include "utils.h"

#include <linux/slab.h>

enum {
	kCBCPoolSize = 512,
};


static int cbc_is_supported(const bc_cipher* cipher)
{
	return 1;
}


static size_t cbc_key_size(const bc_cipher* cipher)
{
	return bc_bits_to_bytes(cipher->key_size);
}


// return context buffer size required by cbc mode.
static size_t cbc_context_size(const bc_cipher* cipher)
{
	return cipher->extended_key_size; 
}


// init cbc context from raw key data.
static void cbc_init_context(void* context, const bc_cipher* cipher, const unsigned char* key, size_t keysize)
{
	bc_cipher_extend_key(cipher, key, context);
}





//------------------------------------------------------------
//
//	Main CBC data processing.
//
//------------------------------------------------------------


static inline void cbc_transform_iv(uint32_t* iv, uint32_t* pool)
{
	unsigned int i = (iv[0] & 0x3F) << 1;
	
	iv[0] ^= pool[i];
	iv[1] ^= pool[i+1];
}


static inline void cbc_transform_buf(const uint32_t* pool, uint32_t* buf, unsigned int len)
{
	len >>= 2;
	while (len--) {
		*buf = *buf ^ *pool;
		++buf; ++pool;
	}
}


typedef void(*cbc_sector_proc)(unsigned char* key, unsigned char* pool, const bc_cipher* cipher,
                               uint64_t secnum, unsigned char* data, size_t nbytes);


static inline void cbc_encrypt_sector(unsigned char* key, unsigned char* pool,
                     const bc_cipher* cipher, uint64_t secnum, unsigned char* data,
                                      size_t nbytes)
{
    cbc_transform_iv((uint32_t*)&secnum, (uint32_t*)pool);
    cbc_transform_buf((uint32_t*)pool, (uint32_t*)data, nbytes);
    if (unlikely(NULL != cipher->encrypt_cbc))
    {
        bc_cipher_encrypt_cbc(cipher, secnum, key, data, nbytes);
    }
    else
    {
        unsigned cipher_block_size = cipher->block_size;
        unsigned nlongs = cipher_block_size / sizeof(uint32_t); // Num of uint32_t in block.
        unsigned total_blocks = nbytes / cipher_block_size;
        
        uint32_t  i, j,
            buffer[4] = {0},
           *src_buffer = (uint32_t*)data,
           *dst_buffer = (uint32_t*)data;
        
        for (i=0; i < nlongs; i++)
            buffer[i] = ((uint32_t*)&secnum)[i & 1];
        
        // foreach block
        for (i = 0; i < total_blocks; ++i)
        {
            for (j=0; j < nlongs; j++)
            {
                buffer[j] ^= src_buffer[j];
            }
            
            bc_cipher_encrypt_ecb(cipher, key, (unsigned char*)buffer);
            
            for (j=0; j < nlongs; j++)
            {
                dst_buffer[j] = buffer[j];
            }
            
            src_buffer += nlongs;
            dst_buffer += nlongs;
        }
    }
}


static inline void cbc_decrypt_sector(unsigned char* key, unsigned char* pool,
                                      const bc_cipher* cipher, uint64_t secnum,
                                      unsigned char* data, size_t nbytes)
{
    cbc_transform_iv((uint32_t*)&secnum, (uint32_t*)pool);
    if (unlikely(NULL != cipher->decrypt_cbc))
    {
        bc_cipher_decrypt_cbc(cipher, secnum, key, data, nbytes);
    }
    else
    {
        unsigned cipher_block_size = cipher->block_size;
        unsigned nlongs = cipher_block_size / sizeof(uint32_t);
        unsigned total_blocks = nbytes / cipher_block_size;
        uint32_t i, j, 
            buffer[4] = {0},
            iv[4] = {0},
            old_src[4] = {0},
            *src_buffer = (uint32_t*)data,
            *dst_buffer = (uint32_t*)data;

        for (i = 0; i < nlongs; i++)
        {
            iv[i] = ((uint32_t*)&secnum)[i & 1];
        }

        for (i = 0; i < total_blocks; i++)
        {
            for (j = 0; j < nlongs; j++)
            {
                buffer[j] = old_src[j] = src_buffer[j];
            }
            
            bc_cipher_decrypt_ecb(cipher, key, (unsigned char*)buffer);

            for (j = 0; j < nlongs; j++)
            {
                dst_buffer[j] = buffer[j] ^ iv[j];
                iv[j] = old_src[j];
            }
            
            src_buffer += nlongs;
            dst_buffer += nlongs;
        }
    }
    cbc_transform_buf((uint32_t*)pool, (uint32_t*)data, nbytes);
}


static void cbc_process(cbc_sector_proc sector_proc, void* extkey, unsigned char* pool, const bc_cipher* cipher, uint64_t secnum, unsigned char* data, size_t nblocks)
{
	size_t nbytes = nblocks * cipher->block_size;
	size_t nsectors = bc_ceil(nbytes, 512) >> 9;
	size_t sector_bytes = 0;	// number of bytes for current sector. can be < sector size
	// foreach sector
	size_t i = 0;    
	while(i < nsectors) {
		sector_bytes = bc_clamp(nbytes, 512); // 512 or nbytes
		sector_proc(extkey, pool, cipher, secnum, &data[i * 512], sector_bytes);

		nbytes -= sector_bytes;
		++secnum;
		++i;
	}
}


static void cbc_encrypt(void* key, const bc_cipher* cipher, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes)
{
	cbc_process(cbc_encrypt_sector, key, pool, cipher, iv, data, nbytes);
}


static void cbc_decrypt(void* key, const bc_cipher* cipher, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes)
{
	cbc_process(cbc_decrypt_sector, key, pool, cipher, iv, data, nbytes);
}


// exported mode def.
bc_mode cbc_exports = 
{
	kCBC, 
	"CBC",
	cbc_is_supported,
	cbc_key_size,
	cbc_context_size,
	cbc_init_context,
	cbc_encrypt,
	cbc_decrypt
};

