//
// Copyright (c) 2006, Jetico, Inc., Finland. All rights reserved.
//

#include "gfmul.h"
#include "intsize.h"

#if defined(__cplusplus)
extern "C"
{
#endif

void memset_my( unsigned char *buf, unsigned char byte, size_t size )
{
	size_t i;
	for(i=0; i<size; i++)
		buf[i] = byte;
}

void memcpy_my( unsigned char *dst, unsigned char *src, size_t size )
{
	size_t i;
	for(i=0; i<size; i++)
		dst[i] = src[i];
}


void xor128 (uint64_t *a, uint64_t *b)
{
	*a++ ^= *b++;
	*a ^= *b;
}

void xor64 (uint64_t *a, uint64_t *b)
{
	*a ^= *b;
}



#if (PLATFORM_BYTE_ORDER == BIG_ENDIAN_PLATFORM)


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Basic algorithm of multiplying in GF(128) for Most Significant Byte (MSB) first
//
///////////////////////////////////////////////////////////////////////////////////////////////


int IsBitSet128_MSB (unsigned int bit, unsigned char *a)
{
	return a[(127 - bit) / 8] & (0x80 >> ((127 - bit) % 8));
}


// bits are set as:
//
// Most Significant byte (MSB)                                             Least Significant byte (LSB)
//
//     b0         b1           b2        b3         b4        b5        ...   b15
//
// 127...120   119...112   111...104   103...96   32...39   40...47     ...  7...0
//
//    <<=
//  shl128 is the same as rightshift() in LRW description, because we shift to the most significant bit



void shl128 (unsigned char *a)
{
	int i, x = 0, xx;
	for (i = 15; i >= 0; i--)
	{
		xx = (a[i] & 0x80) >> 7;
		a[i] = (a[i] << 1) | x;
		x = xx;
	}
}


// GFMUL_Basic16 for MSB
void GFMUL_Basic16(unsigned char *a, unsigned char *b, unsigned char *p)
{
	int i;
	unsigned char la[16];
	memcpy_my (la, a, 16);
	memset_my (p, 0, 16);

	for (i = 0; i < 128; i++)
	{
		if (IsBitSet128_MSB (i, b))
			xor128 ((unsigned __int64 *)p, (unsigned __int64 *)la);

		if (la[0] & 0x80)
		{
			shl128 (la);
			la[15] ^= 0x87;
		}
		else
		{
			shl128 (la);
		}
	}
}


//////////////////////////////////////////////////////////////////////////////////
//
// Optimized algorithms to multiply in GF(128) 128-bit Key2 by 64-bit IV
// 
// Version for Most Significan Byte (MSB) first
//
//
//////////////////////////////////////////////////////////////////////////////////


// We get Key2 in form of BIG_ENDIAN_PLATFORM:  (MSB ... LSB), i.e. k15, k14 .... k0
//
// Key is 128-bit long, but we will generate only a half of 8KB table, because the table
// will be used to multiply by 64-bit IV. 


// GFMUL_InitCtxBlock16 for MSB 
void GFMUL_InitCtxBlock16( unsigned char *Key2, GfmulCtx *Ctx )
{
	int           i,j, byte_num, bit4_shift;
	unsigned char pi_xi[16];

	memset_my( pi_xi, 0, 16 );
	memset_my( &(Ctx->gf_t8k[0][0][0]), 0, 16 * 16 * 16 );

	for (i=15; i >= 0; i--) // iterate 16 sub-tables for 8kb/2 table
	{
		byte_num = i;
		bit4_shift = ((i & 1) ^ 1) << 2;

		for (j=0; j<16; j++) // iterate every of 16 16-bytes elements in sub-table
		{
			pi_xi[ byte_num ] = j << bit4_shift;

			GFMUL_Basic16( Key2, pi_xi, (unsigned char *)&(Ctx->gf_t8k[i][j][0]) );
		}

		pi_xi[ byte_num ] = 0;
	}
}



// GFMUL_MulBlock16 for MSB
void GFMUL_MulBlock16( unsigned char *Iv, unsigned char *T, GfmulCtx *Ctx )
{
	int              i, index;
	unsigned __int64 *t64, *m64;

	memset_my( T, 0, 16 );

	t64 = (unsigned __int64 *)T;

	for (i=15; i >= 0; i--) // iterate 16 sub-tables for 8kb/2 table
	{
		index = (Iv[i] >> (((i << 2) & 4) ^ 4) ) & 0xF;

		m64 = (unsigned __int64 *)(&(Ctx->gf_t8k[i][index][0]));

		t64[0] ^= m64[0];
		t64[1] ^= m64[1];
	}
}


///////////////////////////////////////////////////////////////////////////////////////////////
//
// Basic algorithm of multiplying in GF(64) for Most Significant Byte (MSB) first
//
///////////////////////////////////////////////////////////////////////////////////////////////


int IsBitSet64_MSB (unsigned int bit, unsigned char *a)
{
	return a[(63 - bit) / 8] & (0x80 >> ((63 - bit) % 8));
}


// bits are set as:
//
// Most Significant byte (MSB)                                             Least Significant byte (LSB)
//
//    b0        b1        b2        b3        b4        b5      b6     b7
//
// 63...56   55...48   47...40   39...32   31...24   23...16  15...8  7...0
//
//    <<=
//  shl64 is the same as rightshift() in LRW description, because we shift to the most significant bit



void shl64 (unsigned char *a)
{
	int i, x = 0, xx;
	for (i = 7; i >= 0; i--)
	{
		xx = (a[i] & 0x80) >> 7;
		a[i] = (a[i] << 1) | x;
		x = xx;
	}
}


// GFMUL_Basic8 for MSB
void GFMUL_Basic8(unsigned char *a, unsigned char *b, unsigned char *p)
{
	int i;
	unsigned char la[8];
	memcpy_my (la, a, 8);
	memset_my (p, 0, 8);

	for (i = 0; i < 64; i++)
	{
		if (IsBitSet64_MSB (i, b))
			xor64 ((unsigned __int64 *)p, (unsigned __int64 *)la);

		if (la[0] & 0x80)
		{
			shl64 (la);
			la[7] ^= 0x1b; // LRW mode gets all the basics from GCM mode, where polinom x^64 + x^4 + x^3 + x + 1
			               // is used. Hence, number 0x1b (00011011) is used here.
		}
		else
		{
			shl64 (la);
		}
	}
}


//////////////////////////////////////////////////////////////////////////////////
//
// Optimized algorithms to multiply in GF(64) 64-bit Key2 by 64-bit IV
// 
// Version for Most Significan Byte (MSB) first
//
//
//////////////////////////////////////////////////////////////////////////////////



// We get Key2 in form of BIG_ENDIAN_PLATFORM:  (MSB ... LSB), i.e. k7, k6 .... k0
//

// GFMUL_InitCtxBlock8 for MSB 
void GFMUL_InitCtxBlock8( unsigned char *Key2, GfmulCtx *Ctx )
{
	int           i,j, byte_num, bit4_shift;
	unsigned char pi_xi[8];

	memset_my( pi_xi, 0, 8 );
	memset_my( &(Ctx->gf_t2k[0][0][0]), 0, 16 * 16 * BLOCK_SIZE_8 );

	for (i=15; i >= 0; i--) // iterate 16 sub-tables for 2kb table
	{
		byte_num = i >> 1;
		bit4_shift = ((i & 1) ^ 1) << 2;

		for (j=0; j<16; j++) // iterate every of 16 16-bytes elements in sub-table
		{
			pi_xi[ byte_num ] = j << bit4_shift;

			GFMUL_Basic8( Key2, pi_xi, (unsigned char *)&(Ctx->gf_t2k[i][j][0]) );
		}

		pi_xi[ byte_num ] = 0;
	}
}



// GFMUL_MulBlock8 for MSB
void GFMUL_MulBlock8( unsigned char *Iv, unsigned char *T, GfmulCtx *Ctx )
{
	int              i, index;
	unsigned __int64 *t64, *m64;

	memset_my( T, 0, 8 );

	t64 = (unsigned __int64 *)T;

	for (i=15; i >= 0; i--) // iterate 16 sub-tables for 2kb table
	{
		index = (Iv[i >> 1] >> (((i << 2) & 4) ^ 4) ) & 0xF;

		m64 = (unsigned __int64 *)(&(Ctx->gf_t2k[i][index][0]));

		t64[0] ^= m64[0];
	}
}










#elif (PLATFORM_BYTE_ORDER == LITTLE_ENDIAN_PLATFORM)



///////////////////////////////////////////////////////////////////////////////////////////////
//
// Basic algorithm of multiplying in GF(128) for Least Significant Byte (MSB) first
//
///////////////////////////////////////////////////////////////////////////////////////////////


int IsBitSet_LSB (unsigned int bit, unsigned char *a)
{
	return a[bit / 8] & (0x1 << (bit % 8));
}


// bits are set as:
//
// Least Significant byte (LSB)                                     Most Significant byte (MSB)
//
//     b0      b1        b2        b3        b4        b5       ...    b15
//
//   7...0   15...8   23...16   31...24   39...32   47...40     ...  127...120
//
//    =>>
//  shr128 is the same as rightshift() in LRW description, because we shift to the most significant bit


void shr128 (unsigned char *a)
{
	int i, x = 0, xx;
	for (i = 0; i < 16; i++)
	{
		xx = (a[i] & 0x80) >> 7;

		a[i] = (a[i] << 1) | x;
		
		x = xx;
	}
}


// GFMUL_Basic16 for LSB
void GFMUL_Basic16(unsigned char *a, unsigned char *b, unsigned char *p)
{
	int i;
	unsigned char la[16];
	memcpy_my (la, a, 16);
	memset_my (p, 0, 16);

	for (i = 0; i < 128; i++)
	{
		if (IsBitSet_LSB (i, b))
			xor128 ((uint64_t *)p, (uint64_t *)la);

		if (la[15] & 0x80)
		{
			shr128 (la);
			la[0] ^= 0x87;
		}
		else
		{
			shr128 (la);
		}
	}
}



//////////////////////////////////////////////////////////////////////////////////
//
// Optimized algorithms to multiply in GF(128) 128-bit Key2 by 64-bit IV
// 
// Version for Least Significan Byte (LSB) first
//
//
//////////////////////////////////////////////////////////////////////////////////


// We get Key2 in form of LITTLE_ENDIAN_PLATFORM:  (LSB ... MSB), i.e. k0, k1 .... k15
//
// Key is 128-bit long, but we will generate only a half of 8KB table, because the table
// will be used to multiply by 64-bit IV. 


// GFMUL_InitCtxBlock16 for LSB
void GFMUL_InitCtxBlock16( unsigned char *Key2, GfmulCtx *Ctx )
{
	int           i,j, byte_num, bit4_shift;
	unsigned char pi_xi[16];

	memset_my( pi_xi, 0, 16 );
	memset_my( &(Ctx->gf_t8k[0][0][0]), 0, 16 * 16 * 16 );

	for (i=0; i<16; i++) // iterate 16 sub-tables for 8kb/2 table
	{
		byte_num = i >> 1;
		bit4_shift = (i & 1) << 2;

		for (j=0; j<16; j++) // iterate every of 16 16-bytes elements in sub-table
		{
			pi_xi[ byte_num ] = j << bit4_shift;

			GFMUL_Basic16( Key2, pi_xi, (unsigned char *)&(Ctx->gf_t8k[i][j][0]) );
		}

		pi_xi[ byte_num ] = 0;
	}
}



// GFMUL_MulBlock16 for LSB
void GFMUL_MulBlock16( unsigned char *Iv, unsigned char *T, GfmulCtx *Ctx )
{
	int              i, index;
	uint64_t *t64, *m64;

	memset_my( T, 0, 16 );

	t64 = (uint64_t *)T;

	for (i=0; i<16; i++) // iterate 16 sub-tables for 8kb/2 table
	{
		index = (Iv[i >> 1] >> ((i << 2) & 4)) & 0xF;

		m64 = (uint64_t *)(&(Ctx->gf_t8k[i][index][0]));

		t64[0] ^= m64[0];
		t64[1] ^= m64[1];
	}
}



///////////////////////////////////////////////////////////////////////////////////////////////
//
// Basic algorithm of multiplying in GF(64) for Least Significant Byte (LSB) first
//
///////////////////////////////////////////////////////////////////////////////////////////////


// bits are set as:
//
// Least Significant byte (LSB)                                       Most Significant byte (MSB)
//
//     b0      b1        b2        b3        b4        b5       b6      b7
//
//   7...0   15...8   23...16   31...24   39...32   47...40  55...48  63...56
//
//    =>>
//  shr64 is the same as rightshift() in LRW description, because we shift to the most significant bit


void shr64 (unsigned char *a)
{
	int i, x = 0, xx;
	for (i = 0; i < 8; i++)
	{
		xx = (a[i] & 0x80) >> 7;

		a[i] = (a[i] << 1) | x;
		
		x = xx;
	}
}


// GFMUL_Basic8 for LSB
void GFMUL_Basic8(unsigned char *a, unsigned char *b, unsigned char *p)
{
	int i;
	unsigned char la[8];
	memcpy_my (la, a, 8);
	memset_my (p, 0, 8);

	for (i = 0; i < 64; i++)
	{
		if (IsBitSet_LSB (i, b))
			xor64 ((uint64_t *)p, (uint64_t *)la);

		if (la[7] & 0x80)
		{
			shr64 (la);
			la[0] ^= 0x1b; // LRW mode gets all the basics from GCM mode, where polinom x^64 + x^4 + x^3 + x + 1
			               // is used. Hence, number 0x1b (00011011) is used here.
		}
		else
		{
			shr64 (la);
		}
	}
}



//////////////////////////////////////////////////////////////////////////////////
//
// Optimized algorithms to multiply in GF(64) 64-bit Key2 by 64-bit IV
// 
// Version for Least Significan Byte (LSB) first
//
//
//////////////////////////////////////////////////////////////////////////////////


// We get Key2 in form of LITTLE_ENDIAN_PLATFORM:  (LSB ... MSB), i.e. k0, k1 .... k8


// GFMUL_InitCtxBlock8 for LSB
void GFMUL_InitCtxBlock8( unsigned char *Key2, GfmulCtx *Ctx )
{
	int           i,j, byte_num, bit4_shift;
	unsigned char pi_xi[8];

	memset_my( pi_xi, 0, 8 );
	memset_my( &(Ctx->gf_t2k[0][0][0]), 0, 16 * 16 * BLOCK_SIZE_8 );

	for (i=0; i<16; i++) // iterate 16 sub-tables for 2kb table
	{
		byte_num = i >> 1;
		bit4_shift = (i & 1) << 2;

		for (j=0; j<16; j++) // iterate every of 16 16-bytes elements in sub-table
		{
			pi_xi[ byte_num ] = j << bit4_shift;

			GFMUL_Basic8( Key2, pi_xi, (unsigned char *)&(Ctx->gf_t2k[i][j][0]) );
		}

		pi_xi[ byte_num ] = 0;
	}
}



// GFMUL_MulBlock8 for LSB
void GFMUL_MulBlock8( unsigned char *Iv, unsigned char *T, GfmulCtx *Ctx )
{
	int              i, index;
	uint64_t *t64, *m64;

	memset_my( T, 0, 8 );

	t64 = (uint64_t *)T;

	for (i=0; i<16; i++) // iterate 16 sub-tables for 2kb table
	{
		index = (Iv[i >> 1] >> ((i << 2) & 4)) & 0xF;

		m64 = (uint64_t *)(&(Ctx->gf_t2k[i][index][0]));

		t64[0] ^= m64[0];
	}
}




#endif 
// (PLATFORM_BYTE_ORDER)



#if defined(__cplusplus)
}
#endif
