/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * 
 * BestCrypt LRW Encryption mode implementation.
 */

#include "gfmul.h"
#include "mode.h"
#include "intsize.h"
#include "utils.h"

typedef struct
{
	unsigned char*	primary;
	unsigned char*	tweak;
} lrw_key;


/// lrw_key key = BC_LRW_UNPACK_KEY(keybuf, cipher);
//#define BC_LRW_UNPACK_KEY(key, cipher) { (unsigned char*) (key), ((unsigned char*)(key)) + bc_bits_to_bytes((cipher)->key_size) }
#define BC_LRW_UNPACK_KEY(key, size, cipher) { (char*)(key), (char*)(key) + (size) - (cipher)->block_size }


// all data is allocated in a single buffer really.
typedef struct
{
	GfmulCtx*	gfmul;	// gfmul context
	unsigned char*	extkey;	// cipher primary extended key
} lrw_private;


/// lrw_private priv = BC_LRW_UNPACK_CONTEXT(context);
#define BC_LRW_UNPACK_CONTEXT(context) { (context), (context) + sizeof(GfmulCtx) }


// lrw supports ciphers with block size 8 or 16. basically any cipher.
static int lrw_is_supported(const bc_cipher* cipher)
{
	return ((cipher->block_size == 8) || (cipher->block_size == 16));
}

static size_t lrw_key_size(const bc_cipher* cipher)
{
	return bc_bits_to_bytes(cipher->key_size) + cipher->block_size;
}

static size_t lrw_context_size(const bc_cipher* cipher)
{
	return cipher->extended_key_size + sizeof(GfmulCtx);
}

static void lrw_init_context(void* context, const bc_cipher* cipher, const unsigned char* keydata, size_t keysize)
{
	lrw_key key = BC_LRW_UNPACK_KEY(keydata, keysize, cipher);
	lrw_private ctx = BC_LRW_UNPACK_CONTEXT(context);
	
	if(cipher->block_size == 8) {
		GFMUL_InitCtxBlock8(key.tweak, ctx.gfmul);
	} else if (cipher->block_size == 16) {
		GFMUL_InitCtxBlock16(key.tweak, ctx.gfmul);
	} else {
	}
	
	bc_cipher_extend_key(cipher, key.primary, ctx.extkey);
}


//
//	LRW processing.
//


/// Single lrw round procesudre.
typedef void (*lrw_round_proc) (void* key, GfmulCtx* gfmul, bc_ecb_proc ecb_proc, uint64_t* iv, uint64_t* t_lrw, uint64_t* data); 


static inline void lrw_do_round8(void* key, GfmulCtx* gfmul, bc_ecb_proc ecb_proc, uint64_t* iv, uint64_t* tweak, uint64_t* data)
{
	GFMUL_MulBlock8((unsigned char*) iv, (unsigned char*) tweak, gfmul);
	
	data[0] ^= tweak[0];
	ecb_proc((unsigned char*)data, (unsigned char*)data, key);
	data[0] ^= tweak[0];
}

static inline void lrw_do_round16(void* key, GfmulCtx* gfmul, bc_ecb_proc ecb_proc, uint64_t* iv, uint64_t* tweak, uint64_t* data)
{
	GFMUL_MulBlock16((unsigned char*) iv, (unsigned char*) tweak, gfmul);
	
	data[0] = data[0] ^ tweak[0];
	data[1] = data[1] ^ tweak[1];
	ecb_proc((unsigned char*)data, (unsigned char*)data, key);
	data[0] ^= tweak[0];
	data[1] ^= tweak[1];
}

static void lrw_process(void* context, const bc_cipher* cipher, bc_ecb_proc ecb_proc, uint64_t secnum, unsigned char* data, size_t nblocks)
{
	uint64_t	iv[2] = {0};
	uint64_t	tweak[2] = {0};
	uint64_t*	byteoffset = iv;
	lrw_private	priv = BC_LRW_UNPACK_CONTEXT(context);
	lrw_round_proc 	round_proc = (cipher->block_size == 8) ? lrw_do_round8 : lrw_do_round16;

	*byteoffset = secnum << 9;
	
	// for each block, do lrw round.
	while(nblocks) 
	{
		round_proc(priv.extkey, priv.gfmul, ecb_proc, byteoffset, tweak, (uint64_t*)data);
		*byteoffset	+= cipher->block_size;
		data		+= cipher->block_size;
		nblocks		-= 1;
	}
}


static void lrw_encrypt(void* context, const bc_cipher* cipher, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes)
{
	lrw_process(context, cipher, cipher->encrypt_ecb, iv, data, nbytes);
}


static void lrw_decrypt(void* context, const bc_cipher* cipher, uint64_t iv, unsigned char* pool, unsigned char* data, size_t nbytes)
{
	lrw_process(context, cipher, cipher->decrypt_ecb, iv, data, nbytes);
}


// exported interface.
struct bc_mode lrw_exports = 
{
	kLRW,
	"LRW",
	lrw_is_supported,
	lrw_key_size,
	lrw_context_size,
	lrw_init_context,
	lrw_encrypt,
	lrw_decrypt,
};


