/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 *
 */

#include "mode.h"
#include "utils.h"

enum {
	kXTSBlockSize = 16,
};


// This structure is not allocated anywhere - it is initialized from single extended keys buffer.
typedef struct xts_private
{
	unsigned char*	extkey;		// cipher extended primary key
	unsigned char*	exttweak;	// cipher extended tweak key
} xts_private;


/// xts_private priv = BC_XTS_INIT_PRIVATE(context, cipher);
#define BC_XTS_INIT_PRIVATE(context, cipher) { (context), (context) + (cipher)->extended_key_size }


static int xts_is_supported(const bc_cipher* cipher)
{
	return (cipher->block_size == 16);
}


static size_t xts_key_size(const bc_cipher* cipher)
{
	return (bc_bits_to_bytes(cipher->key_size) + cipher->block_size);
}


static size_t xts_context_size(const bc_cipher* cipher)
{
	return (cipher->extended_key_size * 2);
}


void xts_init_context(void* context, const bc_cipher* cipher, const unsigned char* key, size_t keysize)
{
	size_t cipher_key_size = bc_bits_to_bytes(cipher->key_size);

	const unsigned char* cipher_key = key;
	const unsigned char* tweak_key = key + cipher_key_size;

	unsigned char* xts_key = context;
	unsigned char* xts_tweak = context + cipher->extended_key_size;
	
	bc_cipher_extend_key(cipher, cipher_key, xts_key);
	bc_cipher_extend_key(cipher, tweak_key, xts_tweak);
}


static void xts_process(xts_private* priv, const bc_cipher* cipher, bc_ecb_proc ecb_proc, uint64_t secnum, unsigned char* data, size_t nblocks)
{

	unsigned char*	key1 = priv->extkey;
	unsigned char*	key2 = priv->exttweak;
	
	uint64_t	alpha_J_array[512 / sizeof(uint64_t)];
	uint64_t	ulPP[kXTSBlockSize / sizeof(uint64_t)];
	
	uint64_t*	ulInBuffer	= (uint64_t*)data;
	uint64_t*	ulOutBuffer	= (uint64_t*)data;
	uint64_t	ulBlockCount	= nblocks;
	
	uint64_t	ulBlockMax	= 0;
	
	uint64_t	ulStartSector	= secnum;
	uint64_t	ulTotalSectors	= nblocks * cipher->block_size / 512;
	uint64_t	ulCurSector	= 0;
	
	uint64_t	i, ul0, ul1, ulSpecial;
	
	for(ulCurSector = 0; ulCurSector <= ulTotalSectors; ++ulCurSector)
	{
		alpha_J_array[0] = ulStartSector + ulCurSector;
		alpha_J_array[1] = 0;
		
		cipher->encrypt_ecb((unsigned char *)alpha_J_array, (unsigned char *)alpha_J_array, key2);
		
		for (i = 2; i < (512 / kXTSBlockSize) * 2; i = i + 2 )
		{
			ul0 = alpha_J_array[ i - 2 ];
			ul1 = alpha_J_array[ i - 1 ];
			
			ulSpecial = 0;
			
			if (ul1 & (unsigned long long)0x8000000000000000ULL)
			{
				ulSpecial = 135;
			}
			
			ul1 <<= 1;
			
			if (ul0 & (unsigned long long)0x8000000000000000ULL)
			{
				ul1 |= 1;
			}
			
			ul0 <<= 1;
			
			ul0 ^= ulSpecial;
			
			alpha_J_array[i]   = ul0;
			alpha_J_array[i+1] = ul1;
		}
		
		if (ulBlockCount < (512 /  kXTSBlockSize))
		{
			ulBlockMax = ulBlockCount;
		}
		else
		{
			ulBlockMax = 512 / kXTSBlockSize;
		}
		
		for ( i = 0; i < (ulBlockMax << 1); i += 2 )
		{
			ulPP[0] = alpha_J_array[i]   ^ (*ulInBuffer++);
			ulPP[1] = alpha_J_array[i+1] ^ (*ulInBuffer++);
			
			ecb_proc((unsigned char*)ulPP, (unsigned char*)ulPP, key1);
			
			*ulOutBuffer++ = alpha_J_array[i]   ^ ulPP[0];
			*ulOutBuffer++ = alpha_J_array[i+1] ^ ulPP[1];
			
			ulBlockCount--;
		}
	}
	
}


static void xts_encrypt(void* key, const bc_cipher* cipher, uint64_t secnum, unsigned char* pool, unsigned char* data, size_t nblocks)
{
	xts_private priv = {key, key + cipher->extended_key_size};
	xts_process(&priv, cipher, cipher->encrypt_ecb, secnum, data, nblocks);
}


static void xts_decrypt(void* key, const bc_cipher* cipher, uint64_t secnum, unsigned char* pool, unsigned char* data, size_t nblocks)
{
	xts_private priv = {key, key + cipher->extended_key_size};
	xts_process(&priv, cipher, cipher->decrypt_ecb, secnum, data, nblocks);
}


// exproted xts defs.
struct bc_mode xts_exports = 
{
	kXTS,
	"XTS",
	xts_is_supported,
	xts_key_size,
	xts_context_size,
	xts_init_context,
	xts_encrypt,
	xts_decrypt,
};

