/*
 *	Copyright (c) 2010 Jetico, Inc., Finland
 *	All rights reserved.
 */

#include "datablock.h"

using namespace kgghost;


// rebuild contigious empty space starting at specified keyslot.
// used to describe free space and possible hidden key with key descriptor.
static void kmap_rebuild_empty_key(BCKeySlot* kmap, unsigned keynum, unsigned maxkeys, key_descriptor_t* out_kdesc)
{
	BCKeySlot* slot = &kmap[keynum];
	assert(slot->type == kBCKeyType_Empty);
	
	out_kdesc->base = keynum;
	out_kdesc->total = 1; // including the first slot we just got as an argument.
	//out_kdesc->type = kKBType_Undefined;
	
	while(++keynum < maxkeys) {
		slot = &kmap[keynum];
		if(slot->type != kBCKeyType_Empty) {
			break;
		}
		
		out_kdesc->total++;
	}
}


// rebuild known key descriptor starting at specified position.
static void kmap_rebuild_key(BCKeySlot* kmap, unsigned keynum, unsigned maxkeys, key_descriptor_t* out_kdesc)
{
	BCKeySlot* slot = &kmap[keynum];
	assert((slot->type != kBCKeyType_Empty) && (slot->type != kBCKeyType_Part));
	
	out_kdesc->base = keynum;
	out_kdesc->total = 1;
	out_kdesc->type = slot->type;
	
	while(keynum < maxkeys) {
		slot = &kmap[++keynum];
		if(slot->type != kBCKeyType_Part) {
			break;
		}
		
		out_kdesc->total++;	
	}
}


// record key descriptor in key map.
static void kmap_record_key(BCKeySlot* kmap, key_descriptor_t* kdesc, size_t size)
{
    assert((kdesc->base + kdesc->total) <= kBCV8_MaxKeys);
	
	BCKeySlot* head = &kmap[kdesc->base];
	head->type = kdesc->type;
	head->size = size;
	head->param = 0;
	
	unsigned keynum;
	BCKeySlot* slot;
	for(keynum = 1; keynum < kdesc->total; ++keynum) {
		slot = &kmap[keynum + kdesc->base];
		slot->type = kBCKeyType_Part;
	}
}


// Try to locate and describe contigious free slot area of specified size.
static bool kmap_get_free_space(BCKeySlot* kmap, size_t nslots, size_t maxslots, key_descriptor_t* out_kdesc)
{
	if(nslots > maxslots) {
		return false;
	}
	
	// find contigious vector of empty blocks of enough size
	unsigned keynum = 0;
	while(keynum < maxslots) {
		BCKeySlot* slot = &kmap[keynum];
		if(slot->type == kBCKeyType_Empty) {
			kmap_rebuild_empty_key(kmap, keynum, maxslots, out_kdesc);
			if(out_kdesc->total >= nslots) {
				out_kdesc->total = nslots;
				return true;
			} else {
				keynum += out_kdesc->total;
			}
		} else {
			++keynum;
		}
	}
	
	return false;
}


static void kmap_rebuild_empty_key_reversed(BCKeySlot* kmap, unsigned keynum, key_descriptor_t* out_kdesc)
{
	BCKeySlot* slot = &kmap[keynum];
	assert(slot->type == kBCKeyType_Empty);

	out_kdesc->base = keynum;
	out_kdesc->total = 1; // including the first slot we just got as an argument.
	
	while(keynum != 0) {
		slot = &kmap[--keynum];
		if(slot->type != kBCKeyType_Empty) {
			break;
		}
		out_kdesc->base--;
		out_kdesc->total++;
	}
}


// Get free space for hidden key.
// Look from the end of the map which is marked by the maxslot argument.
static bool kmap_get_free_space_reversed(BCKeySlot* kmap, size_t need_slots, size_t have_slots, key_descriptor_t* out_kdesc)
{
	if(need_slots > have_slots) {
		return false;
	}
	
	unsigned slotnum = have_slots - 1;
	while(slotnum >= 0) {
		BCKeySlot* slot = &kmap[slotnum];
		if(slot->type == kBCKeyType_Empty) {
			kmap_rebuild_empty_key_reversed(kmap, slotnum, out_kdesc);
			if(out_kdesc->total >= need_slots) {
				// correct base index to point to the space we need excatly from the end.
				out_kdesc->base += out_kdesc->total - need_slots;
				out_kdesc->total = need_slots;
				return true;
			} else {
				slotnum -= out_kdesc->total;
			}
		} else {
			--slotnum;
		}
	}
	
	return false;
	
}


// Mark referenced key slots as empty.
static void kmap_mark_empty(BCKeySlot* kmap, key_descriptor_t* kdesc)
{
	for(unsigned i = 0; i < kdesc->total; ++i) {
		BCKeySlot* slot = &kmap[kdesc->base + i];
		slot->size = 0;
		slot->param = 0;
		slot->type = kBCKeyType_Empty;
	}
}


// Merge src map into dst. Dst will contain all non-empty descriptors of both maps.
static void kmap_merge(BCKeySlot* src, BCKeySlot* dst)
{
	for(unsigned i = 0; i < kBCV8_MaxKeys; ++i) {
        BCKeySlot* slot = &src[i];
        if(slot->type != kBCKeyType_Empty) {
            memcpy(&dst[i], slot, sizeof(*slot));
		}
    }
}


// Get number of key slots required to contain encoded key data of this size.
static size_t slots_for_key_size(size_t keysize)
{
	return (keysize + kKeySlotSize - 1) / kKeySlotSize;
}


namespace kgghost 
{

	void DataBlock_Init(DATA_BLOCK* db, size_t dbsize, int alg_id, int mode_id, int hash_id, unsigned long long datasize)
	{
		shred((char*)db, dbsize);
		db_header* header = (db_header*) db;
		memset(header, 0, sizeof(*header));
		header->alg_id		= alg_id;
		header->mode_id		= mode_id;
		header->hash_id		= hash_id;
		header->keyGenId	= kKGID;
		header->version		= BCV8_VERSION_TAG;
        header->data_start	= kBCV8_InitialDataBlockSize;
		header->data_length	= datasize;
		memcpy(header->signature, DB_HEADER_SIGNATURE, kBCV8_SignatureSize);
		memcpy(header->volumeLabel, DB_HEADER_VOLUME_LABEL, sizeof(header->volumeLabel));
	}
	
	
	size_t DataBlock_CapacityForSize(size_t dbsize)
	{
		if(dbsize <= sizeof(DATA_BLOCK)) {
			return 0;
		}

        // Seems, that there was an error in the value of kBCV8_MaxKeys constant:
        // the maximum number of keys, calculated by the formula below, is 74.
        // But the constant is now used to be the maximum of key count, so we use
        // it this way:
        size_t capacity = ((dbsize - sizeof(DATA_BLOCK)) / kKeySlotSize);
        return capacity <= kBCV8_MaxKeys? capacity : kBCV8_MaxKeys;
	}
	
	
	size_t DataBlock_SizeForCapacity(size_t nkeys)
	{
		return (sizeof(DATA_BLOCK) + (nkeys * kKeySlotSize));
	}
	
	
	bool DataBlock_IsHeaderEncrypted(DATA_BLOCK* db)
	{
		db_header* hdr = (db_header*) db;
		return (0 != memcmp(hdr->signature, DB_HEADER_SIGNATURE, sizeof(hdr->signature)));
	}

	
	static void DataBlock_CalcHeaderDigest(db_header* hdr, int hash_id, char* out_digest)
	{
		db_header copy;
		memcpy(&copy, hdr, sizeof(copy));
		
		// nullify some service header fields that do not affect header digest.
		memset(copy.jmpCode, 0, sizeof(copy.jmpCode));
		memset(copy.occupiedTag, 0, sizeof(copy.occupiedTag));
		memcpy(copy.volumeLabel, DB_HEADER_VOLUME_LABEL, sizeof(copy.volumeLabel));
		memcpy(copy.signature, DB_HEADER_SIGNATURE, sizeof(copy.signature));
		
		calculate_digest(hash_id, (char*)&copy, sizeof(copy), out_digest);
	}
	
	
	static bool decrypt_header(db_enc_block* enblock, const char* pool, BCCipher* cipher, const key_data_t* key)
	{
		uint64_t iv = *(uint64_t*) enblock->iv;
		cipher->decrypt(iv, (const char*)key->buf, sizeof(key->buf), pool, (char*) enblock, sizeof(*enblock) - sizeof(enblock->iv));
		
		char digest[kBCV8_DigestSize] = {0};
		DataBlock_CalcHeaderDigest((db_header*) enblock, key->hash_id, digest);
		
		return (0 == memcmp(digest, enblock->digest, sizeof(digest)));
	}
	
	
	void DataBlock_UpdateDigest(DATA_BLOCK* db, const key_data_t* key)
	{
		db_enc_block* enblock = (db_enc_block*) db;
		DataBlock_CalcHeaderDigest((db_header*)enblock, ((db_header*)enblock)->hash_id, enblock->digest);
		
		// digest is always encrypted.
		uint64_t iv = *(uint64_t*) enblock->iv;
        BCCipher cipher;
        assert(kBCKernel_Success == cipher.init(key->alg_id, key->mode_id)); // we are using decoded key here, algorithm should be valid
		cipher.encrypt(iv, (const char*)key->buf, sizeof(key->buf), db->pool, enblock->digest, sizeof(enblock->digest));
	}
	
	
	bool DataBlock_VerifyDigest(DATA_BLOCK* db, const key_data_t* key)
	{
		db_enc_block* enblock = (db_enc_block*) db;
		uint64_t iv = *(uint64_t*) enblock->iv;
    
        BCCipher cipher;
        assert(kBCKernel_Success == cipher.init(key->alg_id, key->mode_id)); // we are using decoded key here, algorithm should be valid

		char digest[kBCV8_DigestSize] = {0};

		if(DataBlock_IsHeaderEncrypted(db)) {
			db_enc_block copy;
			memcpy(&copy, (db_enc_block*) db, sizeof(copy));
			
			return decrypt_header(&copy, db->pool, &cipher, key);
		} else {
			DataBlock_CalcHeaderDigest((db_header*) db, key->hash_id, digest);

			// encrypt generated digest to compare with encrypted header digest directly
			cipher.encrypt(iv, (const char*)key->buf, sizeof(key->buf), db->pool, digest, sizeof(digest));
			return (0 == memcmp(digest, enblock->digest, sizeof(digest)));
		}
	}
	
	
	void DataBlock_EncryptHeader(DATA_BLOCK* db, const key_data_t* key)
	{
		db_enc_block* encp = &db->header;
		DataBlock_CalcHeaderDigest(&encp->header, encp->header.hash_id, encp->digest);
		
        BCCipher cipher;
        assert(kBCKernel_Success == cipher.init(key->alg_id, key->mode_id)); // we are using decoded key here, algorithm should be valid

        uint64_t iv = *(uint64_t*) encp->iv;

		cipher.encrypt(iv, (const char*)key->buf, sizeof(key->buf), db->pool, (char*) encp, sizeof(*encp) - sizeof(encp->iv));
	}
	
	
	bool DataBlock_DecryptHeader(DATA_BLOCK* db, const key_data_t* key)
	{
		db_enc_block decrypted;
        memcpy(&decrypted, (db_enc_block*)db, sizeof(decrypted));

        BCCipher cipher;

        // we are using decoded key here, algorithm should be valid
        assert(kBCKernel_Success == cipher.init(key->alg_id, key->mode_id));

		if(!decrypt_header(&decrypted, db->pool, &cipher, key)) {
			return false;
		}

        memcpy((db_header*)db, (db_header*)&decrypted, sizeof(db_header));
		DataBlock_UpdateDigest(db, key);
		
		return true;
	}
	
	
    size_t DataBlock_EncodeKey(DATA_BLOCK* db, size_t dbsize, const key_data_t* key,
                               const BCPassword* pwd, BCKeySlot* hiddenMap)
	{
		assert(!DataBlock_IsHeaderEncrypted(db));
		
		int			hash = key->hash_id;
		size_t		maxslots = DataBlock_CapacityForSize(dbsize);
		
        BCCipher cipher;
        assert(cipher.init(key->alg_id, key->mode_id) == kBCKernel_Success);

		BCKeySlot* origMap = ((db_header*)db)->keymap;

		// merge current and hidden maps
		BCKeySlot mergedMap[kBCV8_MaxKeys];
		memset(mergedMap, 0, sizeof(mergedMap));
		
		kmap_merge(origMap, mergedMap);

		if(hiddenMap) {
			kmap_merge(hiddenMap, mergedMap);
		}
		
		// locate free key blocks for new key.
		// we look for space for main part keys from the first key block and for hidden part keys from the last.
		bool has_space;
		key_descriptor_t kdesc;
		if(key->is_hidden()) {
			has_space = kmap_get_free_space_reversed(mergedMap, 1, maxslots, &kdesc);
		} else {
			has_space = kmap_get_free_space(mergedMap, 1, maxslots, &kdesc);
		}
		
		if(!has_space) {
			return 0;
		}

		kdesc.type = kKBType_PBE;

		char*	kbdata = (char*) &db->keys[kdesc.base];
		size_t	kbsize = kdesc.total * kKeySlotSize;
		
		size_t encoded_size = KGEncodeKeyBlock(key, pwd, &cipher, hash, kbdata, kbsize);	
		if(!key->is_hidden()) {
			kmap_record_key(origMap, &kdesc, encoded_size);
		} else {
			// for hidden key add verification block.
			verification_data verdata;
			verification_block* verblock = (verification_block*) (kbdata + encoded_size);
            KGEncryptBlock(&verdata, *(uint64_t*) verblock->iv, &cipher, hash, (const char*) key->buf,
                           sizeof(key->buf), db->pool, &verblock->verdata);
			
			if(hiddenMap) {
				kmap_record_key(hiddenMap, &kdesc, maxslots);
			}
		}
		
		return dbsize;
	}
		
	
	// Decode key from keyblock referenced by descriptor.
    static bool decode_keyblock(DATA_BLOCK* db, const BCPassword* pwd, BCCipher* cipher,
                                int hash, key_descriptor_t* kdesc, key_data_t* out_key)
	{
		char* kbdata	= (char*) &db->keys[kdesc->base];
		size_t kbsize	= kdesc->total * kKeySlotSize;
		
        size_t decoded_size = KGDecodeKeyBlock(kbdata, kbsize, &kdesc->type,
                                               pwd, cipher, hash, out_key);
		if (0 == decoded_size) {
            return false;
        }

		// if key is hidden verify it.
		if(out_key->is_hidden()) {
			verification_data verdata;
			verification_block* verblock = (verification_block*) (kbdata + decoded_size);
            if (kbsize < (decoded_size + sizeof(*verblock))) {
				return false;
			}
				
            if (!KGDecryptBlock(&verblock->verdata, *(uint64_t*)verblock->iv, cipher, out_key->hash_id,
                                (const char*)out_key->buf, sizeof(out_key->buf), db->pool, &verdata)) {
				return false;
		    }
        }
				
		// update decoded key info.
		//kdesc->type = coders[i]->kbtype;
		kdesc->total = slots_for_key_size(decoded_size);

		return true;
	}
	
		
    bool DataBlock_DecodeKey(DATA_BLOCK* db, size_t dbsize, const BCPassword* pwd,
              BCKeySlot* hiddenMap, key_descriptor_t* out_desc, key_data_t* out_key)
	{
		db_header*	header = (db_header*) db;
		BCKeySlot*	kmap = header->keymap;
        bool        is_encrypted = DataBlock_IsHeaderEncrypted(db);

        BCCipher cipher;
        int hash;
        if(!is_encrypted) {
            hash = header->hash_id;
            if(kBCKernel_Success != cipher.init(header->alg_id, header->mode_id)) {
                return false;
            }
        }

		unsigned slotnum = 0;
		size_t maxslots = DataBlock_CapacityForSize(dbsize);
		if(maxslots == 0) {
			return false;
		}

        BCEnumAlgorithms    algs;
        BCEnumModes         modes;
        BCEnumHashes        hashes;
		
		while(slotnum < maxslots) {
			
			// for encrypted header we got (possible) cipher and hash from the caller but we don't have a key map.
			// we have to employ the same strategy as for looking hidden keys except to treat all key space as hidden.
			out_desc->base	= slotnum;
			out_desc->total	= maxslots - slotnum;
			out_desc->type	= kKBType_Unknown;

            // if header is encrypted we will have to iterate all possible alg/mode/hash configurations
            // there is a reason to iterate cipher configurations for each key block and not the other way aroud:
            // if password is correct it will most probably decode some very first key blocks.
            if (is_encrypted) {

		        for(unsigned algnum = 0; algnum < algs.total(); ++algnum) {
                    BCAlgInfo* alg = algs.get(algnum);
                    
                    for(unsigned modenum = 0; modenum < modes.total(); ++modenum) {
                        bc_mode_id mode = modes.get(modenum);

                        if (!BCKernelIsModeSupported(alg, mode)) {
                            continue; // mode not supported.
                        }
                        cipher.init(alg, mode);
                        for(unsigned hashnum = 0; hashnum < hashes.total(); ++hashnum) {
                            hash = hashes.get(hashnum);
                            if (decode_keyblock(db, pwd, &cipher, hash, out_desc, out_key)) {
                                goto DECODED;
                            }
                        }
                    }
                }

                //return false;
                ++slotnum;

            } else { // if (is_encrypted) else
                
				// for open header we can use key map for more precise decoding.
				BCKeySlot* slot = &kmap[slotnum];
				assert(slot->type != kBCKeyType_Part);
				
				if(slot->type == kBCKeyType_Empty) {
					// look for hidden key.
					out_desc->type = kKBType_Unknown;
					kmap_rebuild_empty_key(kmap, slotnum, maxslots, out_desc);
				} else {
					kmap_rebuild_key(kmap, slotnum, maxslots, out_desc);
				}
			
                if (!decode_keyblock(db, pwd, &cipher, hash, out_desc, out_key)) {
                    slotnum++;
                    continue;
                } else {
                    goto DECODED;
                }
            }
		}
DECODED: // break from all loops once we've found a correct configuration
            if (slotnum < maxslots) { // if decoded
                // if key is hidden mark it on the map
                if (out_key->is_hidden() && hiddenMap) {
                    kmap_record_key(hiddenMap, out_desc, maxslots);
                }
                return true;
            }
		
        return false;
	}
	
	/*
	size_t DataBlock_DuplicateKey(DATA_BLOCK* db, size_t dbsize, key_descriptor_t* kdesc, key_data_t* key, KGPassword* pwd, BCCipher* cipher, BCKeySlot* hiddenMap)
	{
		assert(!DataBlock_IsHeaderEncrypted(db));
		size_t rc;
		
		if(key->is_hidden()) {
			// record hidden key descriptor into the map so that DataBlock_EncodeKey will see it as occupied and erase it afterwards.
			BCKeySlot* kmap = ((db_header*)db)->keymap;
			kmap_record_key(kmap, kdesc, 0);
			rc = DataBlock_EncodeKey(db, dbsize, key, pwd, hiddenMap);
			kmap_mark_empty(kmap, kdesc);
		} else {
			rc = DataBlock_EncodeKey(db, dbsize, key, pwd, hiddenMap);
		}
		
		return rc;
	}
    */

//    size_t DataBlock_RecodeKey(DATA_BLOCK* db, size_t dbsize, key_descriptor_t* kdesc,
//                               key_data_t* key, BCPassword* pwd, BCCipher* cipher, BCKeySlot* hiddenMap)
//    {
//        DataBlock_DeleteKey(db, dbsize, kdesc);
//        return DataBlock_EncodeKey(db, dbsize, key, pwd, hiddenMap);
//    }


	
    bool DataBlock_ChangePassword(DATA_BLOCK* db, size_t dbsize, BCPassword* oldPwd,
                                  BCPassword* newPwd, BCKeySlot *hiddenMap)
    {
        key_descriptor_t keyDescriptor;
        key_data_t key;

        if (!DataBlock_DecodeKey(db, dbsize, oldPwd, hiddenMap, &keyDescriptor, &key))
            return false;

        DataBlock_DeleteKey(db, dbsize, &keyDescriptor);

        if (dbsize != DataBlock_EncodeKey(db, dbsize, &key, newPwd, hiddenMap))
            return false;

        return true;
    }

    bool DataBlock_VerifyPassword(DATA_BLOCK* db, size_t dbsize, BCPassword* pwd,
                                  BCCipher* cipher, int hash)
	{
		key_data_t tmpkey;
		key_descriptor_t tmpdesc;
		return DataBlock_DecodeKey(db, dbsize, pwd, NULL, &tmpdesc, &tmpkey);
	}

    size_t DataBlock_GetBusySlotCount(DATA_BLOCK* db, size_t dbSize)
    {
        db_header *header = &db->header.header;
        BCKeySlot *keymap = header->keymap;
        size_t maxSlots = DataBlock_CapacityForSize(dbSize),
                busyCount = 0;

        if (maxSlots == 0)
            return 0;

        unsigned slotNumber = 0;
        for (; slotNumber < maxSlots; slotNumber++) {
            BCKeySlot *slot = &keymap[slotNumber];
            if (slot->type != kBCKeyType_Empty)
                ++busyCount;
        }
        return busyCount;
    }

	void DataBlock_DeleteKey(DATA_BLOCK* db, size_t dbsize, key_descriptor_t* desc)
	{
		char*	kbdata = (char*) &db->keys[desc->base];
		size_t	kbsize = desc->total * kKeySlotSize;
		
		shred(kbdata, kbsize);
		BCKeySlot* kmap = ((db_header*)db)->keymap;
		kmap_mark_empty(kmap, desc);
	}

} // namespace kgghost

