/*
 *	Copyright (c) 2010 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Key block coders register.
 */

#include "keyblock.h"
#include "pbe.h"

using namespace kgghost;


typedef unsigned (*kb_decode_proc) (const char* kbdata, unsigned kbsize, const BCPassword* pwd, BCCipher* cipher, int hash, key_data_t* out_key);
typedef bool (*kb_verify_proc) (const char* kbdata, unsigned kbsize, const BCPassword* pwd, BCCipher* cipher, int hash);

struct kb_decoder {
    int type;
    kb_decode_proc decode;
    kb_verify_proc verify;
};

static const kb_decoder gDecoders[] = {
    { kKBType_PBE, pbe_decode, pbe_verify },
};

#define TOTAL_CODERS (sizeof(gDecoders) / sizeof(*gDecoders))

static const kb_decoder* decoder_for_type(int type) 
{
	const kb_decoder* coder;
	unsigned i;
	for(i = 0; i < TOTAL_CODERS; ++i) {
		coder = &gDecoders[i];
		if(coder->type == type) {
			return coder;
		}
	}
		
	return NULL;
}


namespace kgghost
{

    unsigned KGEncodeKeyBlock(const key_data_t* key, const BCPassword *pwd, BCCipher* cipher, int hash, char* kbdata, unsigned kbsize)
    {
        return pbe_encode(key, pwd, kbdata, kbsize);
    }


    unsigned KGDecodeKeyBlock(const char* kbdata, unsigned kbsize, int *kbtype_,
                  const BCPassword *pwd, BCCipher* cipher, int hash, key_data_t* out_key)
    {
        int kbtype = *kbtype_;
        if (kbtype != kKBType_Unknown) {
            const kb_decoder* decoder = decoder_for_type(kbtype);
            return decoder->decode(kbdata, kbsize, pwd, cipher, hash, out_key);
        } else {
	        for (unsigned i = 0; i < TOTAL_CODERS; ++i) {
                const kb_decoder *decoder = &gDecoders[i];
                unsigned rc = decoder->decode(kbdata, kbsize, pwd, cipher, hash, out_key);
                if (rc != 0) {
                    *kbtype_ = decoder->type;
                    return rc;
                }
            }

            return 0;
        }
    }


    bool KGVerifyPassword(const char* kbdata, unsigned kbsize, int kbtype, const BCPassword* pwd, BCCipher* cipher, int hash)
    {
        if (kbtype != kKBType_Unknown) {
            const kb_decoder* decoder = decoder_for_type(kbtype);
            return decoder->verify(kbdata, kbsize, pwd, cipher, hash);
        } else {
	        for (unsigned i = 0; i < TOTAL_CODERS; ++i) {
                if (gDecoders[i].verify(kbdata, kbsize, pwd, cipher, hash)) {
                    return true;
                }
            }

            return false;
        }
    }


    bool KGChangePassword(const char* kbdata, unsigned kbsize, int kbtype, const BCPassword *pwd, const BCPassword *newpwd, BCCipher* cipher, int hash)
    {
        return false;
    }
}

