/*
 *	Copyright (c) 2010 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Key block coders register and interface.
 */

#ifndef KGGHOST_KEYBLOCK_H
#define KGGHOST_KEYBLOCK_H

#include <string.h>
#include <assert.h>

#include "container/cipher.h"
#include "kgghost.h"
#include "hash.h"

#pragma pack(push, 1)

namespace kgghost 
{

    // Type of the encoded key block
	enum {
		kKBType_Unknown   = 0,    // Unknown key block, will try every supported method
		kKBType_PBE       = 1,    // Key block encoded by password
		kKBType_PKE       = 2,    // Key block encoded by public key algorithm
		kKBType_SSS       = 3,    // Key block encoded by secret sharing scheme
	};


    /**
     * Encode key using password based encryption
     * 
     * @param key           Key data to encode
     * @param pwd           Password
     * @param cipher        Configured block cipher
     * @param hash          Hash algorithm id
     * @param kbdata        Key block buffer to store encoded key in
     * @param kbsize        Size of the key block buffer
     * @return              Size of the encoded key block on success or 0 on failure.
     */
    unsigned KGEncodeKeyBlock(const key_data_t* key, const BCPassword* pwd, BCCipher* cipher,
                              int hash, char* kbdata, unsigned kbsize);
    // unsigned KGEncodeKeyBlock(const key_data_t* key, BCSecretSharingScheme* scheme, BCCipher* cipher, int hash, char* kbdata, unsigned kbsize);
    // unsigned KGEncodeKeyBlock(const key_data_t* key, BCPublicKeyAlg* pke, BCCipher* cipher, int hash, char* kbdata, unsigned kbsize);

	
    /**
     * Decode key data from key block
     * 
     * @param kbdata        Encoded key block bytes
     * @param kbsize        Encoded key block size in bytes
     * @param kbtype_       Type of the key block (PBE, SSS, PKE) if known.
     *                      If type is unknown (kKBType_Unknown) proc will try every possible decoding method.
     * @param pwd           Initial password
     * @param cipher        Configured block cipher
     * @param hash          Hash algorithm id
     * @param out_key       Decoded key data on success
     * @return              On success real encoded key block size is returned, or 0 on failure
     */
    unsigned KGDecodeKeyBlock(const char* kbdata, unsigned kbsize, int *kbtype_, const BCPassword* pwd,
                              BCCipher* cipher, int hash, key_data_t* out_key);


    /**
     * Verify password.
     * This procedure will only verify initial password without decoding the whole key.
     * 
     * @param kbdata        Encoded key block bytes
     * @param kbsize        Encoded key block size in bytes
     * @param kbtype        Type of the key block (PBE, SSS, PKE) if known.
     *                      If type is unknown (kKBType_Unknown) proc will try every possible decoding method.
     * @param pwd           Initial password
     * @param cipher        Configured block cipher
     * @param hash          Hash algorithm id
     */
    bool KGVerifyPassword(const char* kbdata, unsigned kbsize, int kbtype, const BCPassword* pwd, BCCipher* cipher, int hash);

    bool KGChangePassword(const char* kbdata, unsigned kbsize, int kbtype, const BCPassword* pwd,
                          const BCPassword* newpwd, BCCipher* cipher, int hash);

    /**
     * Helper structure to hold encrypted data of type T padded to block boundary with its digest.
     * Used in key block encoding algorithms for password verification
     */
    template < typename T >
    struct KGEncryptedBlock {
        enum { kPaddedDataSize = (sizeof(T) + kEncryptedBlockSize - 1) & ~(kEncryptedBlockSize - 1) };
        
        char data[kPaddedDataSize];
        char digest[kDigestSize];

        KGEncryptedBlock() {
			assert(0 == (sizeof(*this) % kEncryptedBlockSize));
        }
    };

    // Decrypt and verify data of type T stored in encrypted block
    template < typename T >
    bool KGDecryptBlock(const KGEncryptedBlock<T>* block, bc_u64 iv, BCCipher* cipher, int hash, const char* key, unsigned keysize, const char* pool, T* data) 
    {
        KGEncryptedBlock<T> copy;
        memcpy(&copy, block, sizeof(copy));
        cipher->decrypt(iv, key, keysize, pool, &copy, sizeof(copy));
        memcpy(data, copy.data, sizeof(*data));

		char digest[kDigestSize] = {0};
		char more_stack[kDigestSize] = {0};
		memset(more_stack, 0, sizeof(more_stack));

        calculate_digest(hash, data, sizeof(*data), digest);

        return (0 == memcmp(digest, copy.digest, sizeof(digest)));
    }

    // Encrypt and store data of type T in encrypted block
    template < typename T >
    void KGEncryptBlock(const T* data, bc_u64 iv, BCCipher* cipher, int hash, const char* key, unsigned keysize, const char* pool, KGEncryptedBlock<T>* block)
    {
        shred(block, sizeof(*block));
        memcpy(block->data, data, sizeof(*data));
        calculate_digest(hash, data, sizeof(*data), block->digest);

        cipher->encrypt(iv, key, keysize, pool, block, sizeof(*block));
    }

}

#pragma pack(pop)

#endif

