/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 *
 * KGHOST key generator interface implementation.
 * This is the top level of key generator implementation.
 * It initates user sessions and applies policies to data block key encoding/decoding mechanisms based on user input.
 */

#include <stdlib.h>
#include <string.h>

#include "container/keygen.h"
#include "container/user.h"
#include "container/file.h"

#include "datablock.h"
#include "kgghost.h"

using namespace kgghost;


typedef struct {
    DATA_BLOCK*     db;
    unsigned        dbsize;
    BCCipher*       cipher;
    int             hash;
    key_data_t*     key;
} HeaderKeyAuth;


static BCContainerResult GetContainerResultBySMResult(sm_result result)
{
    if (kSpaceMap_Success == result)
        return kBCContainer_Success;

    if (kSpaceMap_NoMem == result)
        return kBCContainer_NoMemory;

    if (kSpaceMap_InvalidArg == result)
        return kBCContainer_InvalidArgument;

    return kBCContainer_InternalError;
}


/// @brief Converts specified password into UTF16 format that is used inside key generator.
static void passwordToUTF16(BCPassword *pwd)
{
    char *buf = (char*)malloc(sizeof(pwd->buf));
    memcpy(buf, pwd->buf, sizeof(pwd->buf));
    pwd->length = BCUserUTF8ToUTF16(buf, strlen(buf) + 1, pwd->buf, sizeof(pwd->buf));
    free(buf);
}


static const char* verifyHeaderKey(const char* password, unsigned pwdlen, void* arg)
{
    HeaderKeyAuth* auth = (HeaderKeyAuth*) arg;
    
    BCPassword pwd;
    pwd.set(password);
    passwordToUTF16(&pwd);

    key_descriptor_t kdesc;

    if(!DataBlock_DecodeKey(auth->db, auth->dbsize, &pwd, NULL, &kdesc, auth->key)) {
        return "Invalid password";
    }

    if(auth->key->is_hidden()) {
        return "Please provide main part password";
    }

    return NULL;
}

static bool decodeHeaderKey(const char *message, DATA_BLOCK* db, unsigned dbsize, BCCipher* cipher, int hash, key_data_t* out_key)
{
    HeaderKeyAuth auth = {db, dbsize, cipher, hash, out_key};
    BCPassword pwd;
    return (kBCUserOk == BCUserVerifyPassword(message, verifyHeaderKey, &auth, NULL, &pwd));
}
    
static bool generateKey(int algid, int modeid, int hashid, bc_u64 offset, bc_u64 datasize,
                        unsigned keysize, key_data_t* out_key,
                        BCProgressCallback progress, void *progressArg)
{
    char seed[kSeedLength] = {0};
    if (sizeof(seed) != BCUserGenerateSeed("Generating seed for a key.", seed, sizeof(seed),
                                           progress, progressArg)) {
        return false;
    }

    generate_key(hashid, seed, kSeedLength, (char*) out_key->buf, kMaximumKeySize);

    out_key->alg_id     = algid;
    out_key->mode_id    = modeid;
    out_key->hash_id    = hashid;
    out_key->offset     = offset;
    out_key->datasize   = datasize;
    
    if (modeid == kBCMode_CBC)
        out_key->size   = sizeof(key_data_t) + keysize - kMaximumKeySize;
    else // LRW, XTS
        out_key->size   = sizeof(key_data_t);

    return true;
}

/**
  For encrypted containers, header size is initially set to the maximum size,
  as we can't know it until header is decrypted. The main purpose of this
  function is to get the correct size of the header after decryption.
*/
static bc_u64 getDataBlockSize(DATA_BLOCK *datablock, const char *file)
{
    bc_u64 fileSize = BCFileSize(file);
    if (fileSize) {
        if (DataBlock_IsHeaderEncrypted(datablock)) {
            return kBCV8_MaximumDataBlockSize;
        }

        db_header *header = (db_header*)datablock;
        // Extension keyblocks are located after the container data.
        bc_u64 dataEnd = header->data_start + header->data_length;
        bc_u64 maxFileSize = dataEnd + kBCV8_MaximumExtKeyBlockSize;

        if (fileSize > maxFileSize) {
            return 0;
        }

        // For backup header, it is possible, that file size is much smaller than data size.
        if (fileSize < dataEnd) {
            return kBCV8_InitialDataBlockSize;
        }

        // Extended key block is located after the data in a container file.
        bc_u64 extKeyBlockSize = fileSize - dataEnd;

        return kBCV8_InitialDataBlockSize + extKeyBlockSize;
    }

    return 0;
}

KGGHOST_Header::KGGHOST_Header(std::string fileName, char* datablock,
                               unsigned size, bool /*encrypted*/)
{
    shred(m_db, sizeof(m_db));
    memcpy(m_db, datablock, size);
    m_dbsize = size;
    memset(&this->m_hiddenMap, 0, sizeof(this->m_hiddenMap));
    this->m_file = fileName;
}

void KGGHOST_Header::mainVolumeInfo(struct BCVolumeInfo* info)
{
    assert(!this->isEncrypted());

    db_header* header = (db_header*) m_db;
    memset(info, 0, sizeof(*info));

    info->alg       = header->alg_id;
    info->mode      = header->mode_id;
    info->hash      = header->hash_id;
    info->size      = header->data_length;
    info->offset    = header->data_start;
    info->hidden    = 0;

    (void)BCUserConvertString(header->description, sizeof(header->description), info->description, sizeof(info->description), kBCUser_UTF16LE, kBCUser_UTF8);
}

BCContainerResult KGGHOST_Header::volumeInfoForKey(key_data_t* key, struct BCVolumeInfo* info)
{
    db_header* header = (db_header*) m_db;
    DATA_BLOCK *db = (DATA_BLOCK*) m_db;
    sm_result smres;

    memset(info, 0, sizeof(*info));

    info->alg       = key->alg_id;
    info->mode      = key->mode_id;
    info->hash      = key->hash_id;
    info->size      = key->datasize;
    info->offset    = key->offset;
    info->hidden    = key->is_hidden();

    if(!key->is_hidden()) {
        (void)BCUserConvertString(header->description, sizeof(header->description), info->description,
                                  sizeof(info->description), kBCUser_UTF16LE, kBCUser_UTF8);

        info->map = SpaceMapCreate();
        if (!info->map) {
            return kBCContainer_NoMemory;
        }

        smres = SpaceMapAddNew(info->map, info->offset, info->size);
        if (kSpaceMap_Success != smres) {
            SpaceMapFree(info->map);
            return GetContainerResultBySMResult(smres);
        }

        //fprintf(stderr, "Usual key found. Map contains %lu extents.\n", info->map->count);
    }
    else {
        // Get space map for hidden part.
        assert(key->datasize == 4096); // Mark of hidden key.
        assert(this->m_file.length() != 0);

        bc_file_handle fileHandle;
        bc_file_result fres;
        fres = BCFileOpen(this->m_file.c_str(), kBCFileMode_Read, &fileHandle);
        if (kBCFileResult_Success != fres) {
            return kBCContainer_IOError;
        }

        // Read space map from file.
        char clusterData[kBCV8_HiddenPartSpaceMapSize];

        bc_u64 bytesRead = BCFileReadWithOffset(&fileHandle, clusterData, (bc_u64)kBCV8_HiddenPartSpaceMapSize,
                                                info->offset);
        BCFileClose(&fileHandle);

        if (kBCV8_HiddenPartSpaceMapSize != bytesRead) {
            return kBCContainer_IOError;
        }

        // Decrypt space map.
        bc_kern_result kernRes = BCKernelProcessWithKey(key->alg_id, key->mode_id,
               kBCProcess_Decrypt, (const char*)key->buf, key->size, kBCV8_HiddenPartMapIV,
               (const char*)db->pool, clusterData, kBCV8_HiddenPartSpaceMapSize);
        if (kBCKernel_Success != kernRes)
        {
            return kBCContainer_InternalError;
        }

        char digest[32] = {0};
        char *spaceMapData = clusterData + sizeof(digest);

        // Digest is inside first 32 bytes of data.
        calculate_digest(key->hash_id, spaceMapData, kBCV8_HiddenPartSpaceMapSize - sizeof(digest), digest);

        if (0 != memcmp(digest, clusterData, sizeof(digest)))
        {
            return kBCContainer_Damaged;
        }

        smres = SpaceMapFromBytes((void*)spaceMapData, kBCV8_HiddenPartSpaceMapSize, &info->map);
//        fprintf(stderr, "Hidden key found. Map contains %lu extents.\n",
//                (unsigned long)info->map->count);

        if (kSpaceMap_Success != smres) {
            return GetContainerResultBySMResult(smres);
        }
    }

    return kBCContainer_Success;
}

void KGGHOST_Header::setDescription(const char* description)
{
    assert(!this->isEncrypted());

    db_header* header = (db_header*) m_db;

    memset(header->description, 0, sizeof(header->description));
    BCUserConvertString((char*)description, strlen(description) + 1, header->description, sizeof(header->description), kBCUser_UTF8, kBCUser_UTF16LE);
}

void KGGHOST_Header::markForReencryption(bool mark)
{
    assert(!this->isEncrypted());

    db_header* header = (db_header*) m_db;

    if (mark) {
        strncpy(header->volumeLabel, DB_REENCRYPTED_LABEL, sizeof(header->volumeLabel));
    } else {
        strncpy(header->volumeLabel, DB_HEADER_VOLUME_LABEL, sizeof(header->volumeLabel));
    }
}

bool KGGHOST_Header::isMarkedForReencryption()
{
    assert(!this->isEncrypted());

    db_header* header = (db_header*) m_db;
    return (0 == strncmp(header->volumeLabel, DB_REENCRYPTED_LABEL, sizeof(header->volumeLabel)));
}

bool KGGHOST_Header::isEncrypted()
{
    DATA_BLOCK* db = (DATA_BLOCK*) m_db;
    return DataBlock_IsHeaderEncrypted(db);
}

BCContainerResult KGGHOST_Header::init(int algid, int modeid, bc_u64 size,
                                       const char* description, const char* password,
                                       BCProgressCallback progress, void *progressArg)
{
    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
	db_header* header   = (db_header*) m_db;
	int	hash            = haID_SHA256;

    DataBlock_Init(db, m_dbsize, algid, modeid, hash, size);
    if (description) {
        BCUserConvertString((char*)description, strlen(description) + 1, header->description, sizeof(header->description), kBCUser_UTF8, kBCUser_UTF16LE);
    }

	BCCipher cipher;
	if(kBCKernel_Success != cipher.init(algid, modeid)) {
		return kBCContainer_InvalidAlgorithm;
	}
	
    key_data_t key;
    if(!generateKey(algid, modeid, hash, kBCV8_InitialDataBlockSize,
                    size, cipher.keySize(), &key, progress, progressArg)) {
        return kBCContainer_Cancelled;
    }

    BCPassword pwd;
    pwd.set(password);
    passwordToUTF16(&pwd);
	
    DataBlock_EncodeKey((DATA_BLOCK*)db, m_dbsize, &key, &pwd, this->m_hiddenMap);
	DataBlock_UpdateDigest((DATA_BLOCK*)db, &key);

    return kBCContainer_Success;
}

// Reserves an extent for map saving.
static bool allocateSpaceForMap(SpaceMap *map, bc_u64 *out_offset)
{
    SMExtentNode *cur = map->start;
    while (cur)
    {
        if (cur->extent.length >= kBCV8_HiddenPartSpaceMapSize)
        {
            Extent *e = &cur->extent;
            bc_u64 offset = e->start;
            e->start += kBCV8_HiddenPartSpaceMapSize;
            e->length -= kBCV8_HiddenPartSpaceMapSize;
            if (!e->length)
            {
                sm_result spaceMapResult = SpaceMapRemove(map, e);
                if (kSpaceMap_Success != spaceMapResult)
                    return false;
            }
            (*out_offset) = offset;
            return true;
        }
        cur = cur->next;
    }

    return false;
}


static bool saveSpaceMapByOffset(const char *containerFile, BCCipher *cipher, const char *pool,
                                 key_data_t *key, SpaceMap *map, bc_u64 offset)
{
    char buffer[kBCV8_HiddenPartSpaceMapSize] = { 0 };
    // Place digest at the beginning of the buffer.
    char *digestPlace = buffer;
    char *spaceMapPlace = buffer + 32;

    const int actualMapSize = kBCV8_HiddenPartSpaceMapSize - 32;

    // Convert space map to bytes.
    sm_result smres = SpaceMapToBytes(map, (void*)spaceMapPlace, actualMapSize);

    // Calculate digest for map bytes.
    calculate_digest(key->hash_id, spaceMapPlace, actualMapSize, digestPlace);

    if (kSpaceMap_Success != smres)
        return false;

    // Encrypt whole buffer by specified key.
    cipher->encrypt(kBCV8_HiddenPartMapIV, (char*)key->buf, key->size, pool, (void*)buffer,
                    kBCV8_HiddenPartSpaceMapSize);

    bc_file_handle containerHandle;
    bc_file_result fres = BCFileOpen(containerFile, kBCFileMode_ReadWrite, &containerHandle);

    if (kBCFileResult_Success != fres)
        return false;

    bc_u64 bytesRead = BCFileWriteWithOffset(&containerHandle, buffer, kBCV8_HiddenPartSpaceMapSize, offset);
    BCFileClose(&containerHandle);

    if (bytesRead != kBCV8_HiddenPartSpaceMapSize)
        return false;

    return true;
}

BCContainerResult KGGHOST_Header::createHiddenKey(SpaceMap *spaceMap,
                                const BCPassword *mainPassword_, const BCPassword* hiddenPassword_,
                                BCProgressCallback progress, void *progressArg)
{
    DATA_BLOCK*         db     = (DATA_BLOCK*) m_db;
    db_header*          header = (db_header*) db;

    key_descriptor_t    kdesc; // unused
    BCPassword          mainPassword(mainPassword_);
    BCPassword          hiddenPassword(hiddenPassword_);

    passwordToUTF16(&mainPassword);
    passwordToUTF16(&hiddenPassword);

    // decode main header key
	
    key_data_t mainKey;
    if(!DataBlock_DecodeKey(db, m_dbsize, &mainPassword,
                this->m_hiddenMap, &kdesc, &mainKey)) {
        return kBCContainer_InvalidPassword;
    }

    assert(!mainKey.is_hidden());

    bool encrypted = this->isEncrypted();
    // decrypt header with the main key to get the space map
    if (encrypted) {
        if (!this->decryptHeader(&mainKey))
            return kBCContainer_Damaged;
    }

    BCCipher cipher;
    if(kBCKernel_Success != cipher.init(header->alg_id, header->mode_id)) {
        return kBCContainer_InvalidAlgorithm;
    }

    // Calculate offset of the spacemap store.
    bc_u64 offset;
    if (!allocateSpaceForMap(spaceMap, &offset))
        return kBCContainer_InternalError;

    // generate and encode new hidden key
	key_data_t hiddenKey;
    if (!generateKey(mainKey.alg_id, mainKey.mode_id, mainKey.hash_id, offset,
                     4096, cipher.keySize(), &hiddenKey, progress, progressArg)) {
        return kBCContainer_Cancelled;
    }

    if(!DataBlock_EncodeKey((DATA_BLOCK*)db, m_dbsize, &hiddenKey,
                            &hiddenPassword, this->m_hiddenMap)) {
        return kBCContainer_NoKeySpace;
    }

    // Save the space map by offset. Use pool from the main part.
    if (!saveSpaceMapByOffset(this->m_file.c_str(), &cipher, db->pool, &hiddenKey, spaceMap, offset))
        return kBCContainer_InternalError;

    // update digest and encrypt the header if it was encrypted
	DataBlock_UpdateDigest((DATA_BLOCK*)db, &mainKey);

    if (encrypted) {
        DataBlock_EncryptHeader(db, &mainKey);
    }

    return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::verifyPassword(const char* password, struct BCVolumeInfo* out_info)
{
    DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
    db_header* header   = (db_header*) db;
	bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);

    BCPassword pwd;
    pwd.set(password);
    passwordToUTF16(&pwd);
	
    BCCipher cipher;
    if(!is_encrypted) {
        if(kBCKernel_Success != cipher.init(header->alg_id, header->mode_id)) {
            return kBCContainer_InvalidAlgorithm;
        }
    }

    key_descriptor_t kdesc;
    key_data_t key;

    if(!DataBlock_DecodeKey(db, m_dbsize, &pwd, this->m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }

    this->volumeInfoForKey(&key, out_info);

    return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::decodeKey(const char* password, struct BCVolumeInfo* out_info, int* out_handle, char* out_pool)
{
	DATA_BLOCK* db      = (DATA_BLOCK*) m_db;
    db_header* header   = (db_header*) db;
	bool is_encrypted   = DataBlock_IsHeaderEncrypted(db);

    BCPassword pwd;
    pwd.set(password);
    passwordToUTF16(&pwd);
	
	BCCipher cipher;
    key_descriptor_t kdesc;
    key_data_t key;

    if(!is_encrypted) {
	    if(kBCKernel_Success != cipher.init(header->alg_id, header->mode_id)) {
		    return kBCContainer_InvalidAlgorithm;
	    }
    }

    if(!DataBlock_DecodeKey(db, m_dbsize, &pwd, this->m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }

	if(!key.is_hidden() && !DataBlock_VerifyDigest(db, &key)) {
		return kBCContainer_Damaged;
	}

    if(kBCKernel_Success != BCKernelStoreKey((char*)key.buf, sizeof(key.buf), out_handle)) {
        return kBCContainer_NoMemory;
    }

    memcpy(out_pool, db->pool, sizeof(db->pool));
    this->volumeInfoForKey(&key, out_info);

	return kBCContainer_Success;
}

BCContainerResult KGGHOST_Header::passwd(const char* password, const char* newPassword,
                                         BCKeyOp opflag)
{
	DATA_BLOCK*	    db = (DATA_BLOCK*) m_db;
    db_header*      header = (db_header*) db;
	int		        hash = header->hash_id;
	bool		    is_encrypted = DataBlock_IsHeaderEncrypted(db);
	BCCipher	    cipher;
	
    // Key data and descriptor we will work with
    key_descriptor_t    kdesc;
    key_data_t          key;

    key_data_t          mainKey;    // In case of encrypted header this will container decoded main key
    key_data_t*         headerKey;  // Pointer to whatever key we got that can update header digest

	if(!is_encrypted) {
		if(kBCKernel_Success != cipher.init(header->alg_id, header->mode_id)) {
			return kBCContainer_InvalidAlgorithm;
		}
	}
	
    BCPassword pwd;
    pwd.set(password);
    passwordToUTF16(&pwd);

    if(!DataBlock_DecodeKey(db, m_dbsize, &pwd, this->m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }

    // We need main key for
    // 1) verifying and encoding of datablock digest
    // 2) in case of the encrypted header - to be able to decrypt it and then encrypt again.
    if(key.is_hidden()) {

        if(!decodeHeaderKey("Please enter password from the main part:", db, m_dbsize, &cipher, hash, &mainKey)) {
            return kBCContainer_InvalidPassword;
        }

        headerKey = &mainKey;

    } else {
        headerKey = &key;
    }

    if (is_encrypted) {
        assert(this->decryptHeader(headerKey)); // we've just decoded the key correctly!
    }
	
    // Invariants at this point: 1) key is decoded and 2) header is decrypted if needed.
    assert(key.size != 0);

    if (!DataBlock_VerifyDigest(db, headerKey)) {
        return kBCContainer_Damaged;
    }

    switch (opflag) {

        // change password.
        case kBCKey_Recode:	
        {
            BCPassword new_pwd;
            new_pwd.set(newPassword);
            passwordToUTF16(&new_pwd);

            if (!DataBlock_ChangePassword(db, m_dbsize, &pwd, &new_pwd, NULL))
                return kBCContainer_InternalError;
            break;
        }

        // add password.
        case kBCKey_Duplicate:	
        {
            BCPassword new_pwd;
            new_pwd.set(newPassword);
            passwordToUTF16(&new_pwd);

            if(!DataBlock_EncodeKey(db, m_dbsize, &key, &new_pwd, this->m_hiddenMap)) {
                return kBCContainer_NoKeySpace;
            }

            break;
        }

        // remove password.
        case kBCKey_Delete:
        {
            if ((DataBlock_GetBusySlotCount(db, m_dbsize) != 1) || (key.is_hidden())) {
                DataBlock_DeleteKey(db, m_dbsize, &kdesc);
            }
            else {
                // Have only one key (main), do not allow remove it.
                return kBCContainer_NoKeySpace;
            }

            break;
        }

        default: 
            return kBCContainer_Unsupported;
    }

    DataBlock_UpdateDigest(db, headerKey);

    // if header was decrypted store a copy and encrypt it back with the new changes.
    if(is_encrypted) {
        DataBlock_EncryptHeader(db, headerKey);
    }

    return kBCContainer_Success;

}

bool KGGHOST_Header::decryptHeader(key_data_t *key)
{
    if (!this->isEncrypted())
        return true;
    assert(!key->is_hidden());

    DATA_BLOCK *db = (DATA_BLOCK*)this->m_db;
    if (!DataBlock_DecryptHeader(db, key))
        return false;

    // Update the datablock size.
    bc_u64 dbSize = getDataBlockSize(db, this->m_file.c_str());
    if (dbSize)
    {
        this->m_dbsize = dbSize;
    }

    return true;
}


BCContainerResult KGGHOST_Header::addPassword(const char* password, const char* newPassword)
{
    return this->passwd(password, newPassword, kBCKey_Duplicate);
}

BCContainerResult KGGHOST_Header::changePassword(const char* password, const char* newPassword)
{
    return this->passwd(password, newPassword, kBCKey_Recode);
}

BCContainerResult KGGHOST_Header::removePassword(const char* password)
{
    return this->passwd(password, NULL, kBCKey_Delete);
}

BCContainerResult KGGHOST_Header::encrypt(const char* password)
{
    DATA_BLOCK*     db = (DATA_BLOCK*) m_db;
    db_header*      header = (db_header*) db;

	if (DataBlock_IsHeaderEncrypted(db)) {
		return kBCContainer_Success;
	}
	
	BCCipher cipher;
	if (kBCKernel_Success != cipher.init(header->alg_id, header->mode_id)) {
		return kBCContainer_InvalidAlgorithm;
	}
	
    BCPassword pwd;
    pwd.set(password);
    passwordToUTF16(&pwd);

    key_descriptor_t kdesc;
    key_data_t key;

    if (!DataBlock_DecodeKey(db, m_dbsize, &pwd, this->m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }
	
    if (key.is_hidden()) {
        return kBCContainer_InvalidPassword;
    }

    if(!DataBlock_VerifyDigest(db, &key)) {
        return kBCContainer_Damaged;
    }

    DataBlock_EncryptHeader(db, &key);
    return kBCContainer_Success;
}
 
BCContainerResult KGGHOST_Header::decrypt(const char* password)
{
    DATA_BLOCK*	    db = (DATA_BLOCK*) m_db;

    if(!DataBlock_IsHeaderEncrypted(db)) {
        return kBCContainer_Success;
    }
	
    BCPassword pwd;
    pwd.set(password);
    passwordToUTF16(&pwd);

    key_descriptor_t kdesc;
    key_data_t key;

    if (!DataBlock_DecodeKey(db, m_dbsize, &pwd, this->m_hiddenMap, &kdesc, &key)) {
        return kBCContainer_InvalidPassword;
    }
	
    if (key.is_hidden()) {
        return kBCContainer_InvalidPassword;
    }

    if (!this->decryptHeader(&key)) {
        return kBCContainer_Damaged;
    }

    return kBCContainer_Success;
}


BCContainerResult KGGHOST_Header::store(const char* file)
{
    bc_file_handle handle;
    if(0 != BCFileOpen(file, kBCFileMode_ReadWrite, &handle)) {
        return kBCContainer_IOError;
    }

    // write initial data block header
    if(kBCV8_InitialDataBlockSize != BCFileWrite(&handle, m_db, kBCV8_InitialDataBlockSize)) {
        BCFileClose(&handle);
        return kBCContainer_IOError;
    }

    // write the rest to the end of the file
    unsigned extKbSize = m_dbsize - kBCV8_InitialDataBlockSize;
    if (extKbSize) {
        bc_u64 fileSize = BCFileSize(file);
        if (!fileSize) {
            return kBCContainer_IOError;
        }

        bc_u64 extKbOffset = fileSize - extKbSize;
        if (extKbSize != BCFileWriteWithOffset(&handle, m_db + kBCV8_InitialDataBlockSize,
                                               extKbSize, extKbOffset)) {
            BCFileClose(&handle);
            return kBCContainer_IOError;
        }
    }

    BCFileClose(&handle);
    return kBCContainer_Success;
}


KGGHOST_Header::~KGGHOST_Header()
{
    shred(m_db, sizeof(m_db));
}


static IBCHeader* kgghost_create(void)
{
    char db[kBCV8_MaximumDataBlockSize] = {0};
    unsigned dbsize = kBCV8_InitialDataBlockSize;
    return new KGGHOST_Header(std::string(), db, dbsize, false);
}


static IBCHeader* kgghost_load(const char* file)
{
	char db[kBCV8_MaximumDataBlockSize] = {0};
    DATA_BLOCK *datablock = (DATA_BLOCK*)db;
    db_header* header = (db_header*) db;

    bc_file_handle handle;
    if (0 != BCFileOpen(file, kBCFileMode_Read, &handle)) {
        return NULL;
    }

    if (kBCV8_InitialDataBlockSize != BCFileRead(&handle, db, kBCV8_InitialDataBlockSize)) {
        BCFileClose(&handle);
        return NULL;
    }

    bc_u64 dbSize = getDataBlockSize(datablock, file);
    if (!dbSize) {
        BCFileClose(&handle);
        return NULL;
    }

    if (dbSize > kBCV8_InitialDataBlockSize) {

        bc_u64 extKeyBlockSize = dbSize - kBCV8_InitialDataBlockSize;
        bc_u64 extKeyBlockOffset;

        if (!DataBlock_IsHeaderEncrypted(datablock)) {
            extKeyBlockOffset = header->data_start + header->data_length;
        } else {
            // For encrypted header, read kBCV8_MaximumExtKeyBlockSize bytes from the end.
            bc_u64 fileSize = BCFileSize(file);
            if (!fileSize) {
                BCFileClose(&handle);
                return NULL;
            }

            extKeyBlockOffset = fileSize - kBCV8_MaximumExtKeyBlockSize;
        }
        if (extKeyBlockSize !=
                BCFileReadWithOffset(&handle, db + kBCV8_InitialDataBlockSize, extKeyBlockSize, extKeyBlockOffset)) {
            BCFileClose(&handle);
            return NULL;
        }
    }

//    // if header is invalid we can possibly have an encrypted header.
//    // but we can't know for sure because there might be another keygen that will recognize this container.
//    bool encrypted = false;
//    if (strncmp((const char*)header->signature, DB_HEADER_SIGNATURE, kBCV8_SignatureSize) != 0) {
//        encrypted = true;
//    } else if ((header->version != 3) || (header->keyGenId != kKGID)) {
//        BCFileClose(&handle);
//        return NULL;
//    }

//    // see how much, if any, extended key data we need to read and allocate enough data to hold everything.
//    bc_u64 fileSize	= BCFileSize(file);
//    bc_u64 extKbOffset = 0;
//    bc_u64 extKbSize = 0;

//    if (!encrypted) {
//        extKbOffset	= header->data_length + header->data_start;
//        if(extKbOffset > fileSize) {
//            // this is possible when we are dealing with backup headers.
//            extKbSize = 0;
//        } else {
//            extKbSize = fileSize - extKbOffset;
//        }
//    } else {
//        extKbSize = fileSize - kBCV8_InitialDataBlockSize;
//        if (extKbSize > kBCV8_MaximumExtKeyBlockSize) {
//            extKbSize = kBCV8_MaximumExtKeyBlockSize;
//        }
        
//        extKbOffset = fileSize - extKbSize;
//    }
	
//    if (extKbSize > kBCV8_MaximumExtKeyBlockSize) {
//        BCFileClose(&handle);
//        return NULL;
//    }
	
//    if (extKbSize > 0) {
//        if (extKbSize != BCFileReadWithOffset(&handle, db + kBCV8_InitialDataBlockSize, extKbSize, extKbOffset)) {
//            BCFileClose(&handle);
//            return NULL;
//        }
//    }

    BCFileClose(&handle);
    return new KGGHOST_Header(file, db, dbSize, false);
}

BCContainerResult kgghost_parse(const char* file, struct BCVolumeInfo* info)
{
    BCHeaderHandle header(kgghost_load(file));
    if (!header || header->isEncrypted()) {
        return kBCContainer_UnknownFormat;
    }

    header->mainVolumeInfo(info);
    return kBCContainer_Success;
}


struct BCKeygen kg_ghost = {
    kKGID,
    "kgghost",
    "Default BestCrypt version 8 key generator",
    BCV8_VERSION_TAG,
    kBCKeygen_EncryptedHeaders,

    kgghost_create,
    kgghost_load,
    kgghost_parse,
};


