/*
 * Copyright (c) 2005-2010 Jetico, Inc., Finland
 * All rights reserved.
 *
 * Contains common keygen auth session definitions.
 */

#ifndef KGGHOST_H
#define KGGHOST_H

#include "container/keygen.h"
#include "container/intsize.h"
#include "container/user.h"
#include "db_header.h"
#include "hash.h"

#include <string.h>
#include <stdlib.h>
#include <string>


namespace kgghost {
	
	// Common definitions.
	enum {
        kKGID               = 0x00000004,   // Keygen id
        kKeySlotSize        = 256,          // Single encoded key slot.
        kDigestSize         = 32,           // Digest buffer length.
        kIVSize	            = 16,           // IV buffer length.
        kSeedLength         = 128,          // random seed buffer size for key generation.
        kEncryptedBlockSize = 16,           // Block cipher encryption block size.
        kMaximumKeySize     = 64,           // Maximum stored key size.
        kPoolSize           = 512           // Random data pool size.
	};
	
	
    #pragma pack(push, 1)
	// This is the key data encoded to and decoded from key blocks by coder algorithms.
	typedef struct key_data {
		bc_u8		size;		// real size of the structure in bytes
		bc_u8		hash_id;	// Digest algorithm used for key block verification ( same as pgp id ( f.e. pgphaSHA1 or bchaSHA256 see pktdefs.h )
		bc_u32		alg_id;		// Algorithm is used for encryption of header, contInfo, hiddenContInfo, keyMap and container body
		bc_u32		mode_id;	// The mode is used for encryption for encryption algorithm above
		bc_u64		offset;		// container data offset that is encrypted with this key.
		bc_u64		datasize;	// used as a hidden key marker, will be set to 4096 if the key is hidden.
		
		bc_u8		buf[kMaximumKeySize];
		
		key_data() {
			memset(this, 0, sizeof(key_data));
		}
		
		~key_data() {
			memset(this, 0, sizeof(key_data));
		}
		
		/// Get total key structure size that contains key with the given length.
		static inline size_t calc_size_for_key_length(size_t keylen) {
			return sizeof(key_data) + keylen;
		}

		bool is_hidden() const {
			return offset != kBCV8_InitialDataBlockSize;
		}
		
	} key_data_t;
    #pragma pack(pop)	
	

    /**
     * kgghost keygen interface implementation.
     */
    class KGGHOST_Header : public IBCHeader {
    public:

        KGGHOST_Header(std::string fileName, char *datablock, unsigned size, bool);

        virtual void mainVolumeInfo(struct BCVolumeInfo* info);
        virtual bool isEncrypted();
        virtual BCContainerResult init(int algid, int modeid, bc_u64 size,
                                       const char* description, const char* password,
                                       BCProgressCallback progress, void *progressArg);
        virtual BCContainerResult createHiddenKey(SpaceMap *spaceMap,
                                const BCPassword* mainPassword_, const BCPassword* hiddenPassword_,
                                BCProgressCallback progress, void *progressArg);
        virtual BCContainerResult verifyPassword(const char* password, struct BCVolumeInfo* out_info);
        virtual BCContainerResult decodeKey(const char* password, struct BCVolumeInfo* out_info, int* out_handle, char* out_pool);
        virtual BCContainerResult addPassword(const char* password, const char* newPassword);
        virtual BCContainerResult changePassword(const char* password, const char* newPassword);
        virtual BCContainerResult removePassword(const char* password);
        virtual BCContainerResult store(const char* file);
        virtual BCContainerResult encrypt(const char* password);
        virtual BCContainerResult decrypt(const char* password);
        virtual void setDescription(const char* description);
        virtual void markForReencryption(bool mark);
        virtual bool isMarkedForReencryption();

        virtual ~KGGHOST_Header();

    protected:

        BCContainerResult volumeInfoForKey(key_data_t* key, struct BCVolumeInfo* info);
        BCContainerResult passwd(const char* password, const char* newPassword, BCKeyOp opflag);
        bool decryptHeader(key_data_t *key);

    private:
        
        char        m_db[kBCV8_MaximumDataBlockSize];   // data block buffer
        unsigned    m_dbsize;                           // size of data block as read from file
        std::string m_file;                             // full path to the container file
        BCKeySlot   m_hiddenMap[kBCV8_MaxKeys];         // key map that will contain all hidden keys during hidden part protection
    };
}

#endif

