/*
 *	Copyright (c) 2004-2010 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Password based key block encoder.
 */

#include "keyblock.h"

#define PBE_HASH_ITERATIONS	0x100
#define PBE_SALT_SIZE		8

using namespace kgghost;

// Encoded key representation format stored in keyblock.
// We encode it by generating the storage key from password and encrypting into keyblock.
#pragma pack(push, 1)
typedef struct {
    char				            salt[PBE_SALT_SIZE];    // randomly generated salt for password.
    KGEncryptedBlock<key_data_t>	keyblock;               // encoded key.
    char				            iv[kIVSize];            // unencrypted initial vector.
} pbe_format;
#pragma pack(pop)


namespace kgghost {

    unsigned pbe_encode(const key_data_t* key, const BCPassword* pwd, char* kb_data, unsigned kb_size)
    {
	    if(kb_size < sizeof(pbe_format)) {
		    return 0;
	    }

	    // shredding the key block will have a side effect of generating random data for salt and iv.
	    shred(kb_data, kb_size);
	    pbe_format* pbe = (pbe_format*) kb_data;

	    BCCipher cipher;
	    assert(cipher.init(key->alg_id, (bc_mode_id) key->mode_id) == kBCKernel_Success);
	
	    char storage_key[kMaximumKeySize] = {0};
        generate_key_from_password(key->hash_id, pwd->buf, pwd->length, pbe->salt,
                                   sizeof(pbe->salt), PBE_HASH_ITERATIONS,
                                   storage_key, sizeof(storage_key));
	
        KGEncryptBlock(key, *(uint64_t*)pbe->iv, &cipher, key->hash_id, storage_key,
                       sizeof(storage_key), NULL, &pbe->keyblock);
	    return sizeof(pbe_format);
    }


    unsigned pbe_decode(const char* kb_data, unsigned kb_size, const BCPassword* pwd,
                        BCCipher* cipher, int hash, key_data_t* out_key)
    {
	    if(kb_size < sizeof(pbe_format)) {
		    return 0;
	    }
	
	    const pbe_format* pbe = (const pbe_format*) kb_data;
	
        char storage_key[kMaximumKeySize] = { 0 };
        generate_key_from_password(hash, pwd->buf, pwd->length, pbe->salt,
                                   sizeof(pbe->salt), PBE_HASH_ITERATIONS, storage_key, sizeof(storage_key));
	
        if (KGDecryptBlock(&pbe->keyblock, *(uint64_t*) pbe->iv, cipher, hash, storage_key, sizeof(storage_key), NULL, out_key)) {
		    return sizeof(pbe_format);
	    } else {
		    return 0;
	    }
    }


    bool pbe_verify(const char* kb_data, unsigned kb_size, const BCPassword* pwd,
                    BCCipher* cipher, int hash)
    {
	    // for password encoded keys verification and decoding is the same thing since we don't ask anything else.
	    key_data_t tmpkey;
	    return (pbe_decode(kb_data, kb_size, pwd, cipher, hash, &tmpkey) != 0);
    }

}

