/*
 *	Copyright (c) 2004-2010 Jetico, Inc., Finland
 *	All rights reserved.
 *
 *	Shamir secret sharing scheme key block encoder.
 */

#include <assert.h>
#include <vector>

#include "keyblock.h"
#include "container/user.h"

#include "CertificateLib/sss/shamir/shamir.h"
#include "CertificateLib/hashAlgs/hashAlgs.h"

using namespace kgghost;

enum {
    kSSS_HashIterations = 1, 
	kMaxEncodedSize		= sizeof(key_data_t),
	kMaxPrimeSize		= kMaxEncodedSize + 1,
	kPrimeAndSaltSize	= kMaxPrimeSize - 1,
	kSaltSize		    = 8,
};


// Share format version.
enum {
	kShareFormat_Unknown	= 0x00,
	kShareFormat_Version1	= 0x01,
	kShareFormat_Version2	= 0x02,
};


// Share description
typedef struct share {
	bc_u8	version;		    // share version tag, kShareFormat_Version2
	bc_u8	threshold;		    // total number of shares required to decode the key. each share has this value.
	bc_u8	total;		        // total number of shares encoded. should be >= required shares. each share has this value.
	bc_u8	headOfPrime;		// first byte of prime.
	bc_u8	valX;			    // share order number is used for x value.
	bc_u8	valY[kMaxPrimeSize];	// 
	bc_u8	hash;			    // id of hash algorithm used for share digest operations.
	
	
	// Load valX into power integer.
	void loadX(pw_int& power_x) const {
		if(power_x.set(sizeof(valX), &valX) != pw_OK) {
			assert(0);
		}
	}
	
	
	// Store valX from power integer.
	void storeX(const pw_int& power_x) {
		if(pw_int2bytes(&valX, sizeof(valX), &power_x) != pw_OK) {
			assert(0);
		}
	}

	
	// Load valY into power integer.
	void loadY(pw_int& power_y) const {
		if(power_y.set(sizeof(valY), valY) != pw_OK) {
			assert(0);
		}
	}
	
	
	// Store valY from power integer.
	void storeY(const pw_int& power_y) {
		if(pw_int2bytes(valY, sizeof(valY), &power_y) != pw_OK) {
			assert(0);
		}
	}
	
} share_t;


// Share stored in encoded form
typedef KGEncryptedBlock<share_t> encoded_share_t;


// SSS encoded key block.
typedef struct {
	bc_u8				primeAndSalt[kPrimeAndSaltSize];
    encoded_share_t     shares[1];	// real number of shares is unknown until at least one share is decoded
} SSS_KeyBlock;


// calculate digest for a single share.
static void calculate_share_digest(const share_t* share, const char* prime, char* out_digest)
{
	HHA ha;
	if(ha_createAlgorithm(&ha, share->hash) != hae_ok) {
		assert(0);
	}
	
	unsigned int digest_size = ha_getDigestSize(ha);
	shred(out_digest, digest_size);
	
	if(ha_process(ha, (BYTE*)prime, kPrimeAndSaltSize) != hae_ok) {
		assert(0);
	}
	
	if(ha_process(ha, (BYTE*)share, sizeof(*share)) != hae_ok) {
		assert(0);
	}
	
	if(ha_final(ha, (BYTE*)out_digest) != hae_ok) {
		assert(0);
	}
		
	ha_free(ha);
}


void sss_encode_share(const share_t* share, bc_u64 share_num, const char* prime, BCCipher* cipher, const char* key, unsigned keysize, encoded_share_t* out_data)
{
    char iv[kIVSize] = {0};

	// use index of share as an iv
	iv[0] = share_num & 0xFF;
	iv[1] = (share_num >> 8) & 0xFF;

    shred((char*) out_data, sizeof(*out_data));
    memcpy(out_data->data, share, sizeof(*share));

    calculate_share_digest(share, prime, out_data->digest);

    cipher->encrypt(*(bc_u64*) iv, key, keysize, NULL, (char*) out_data, sizeof(*out_data));
}


bool sss_decode_share(encoded_share_t* data, bc_u16 share_num, const char* prime, BCCipher* cipher, const char* key, size_t keysize, share_t* out_share)
{
    char iv[kIVSize] = {0};

	// use index of share as an iv
	iv[0] = share_num & 0xFF;
	iv[1] = (share_num >> 8) & 0xFF;

    cipher->decrypt(*(bc_u64*) iv, key, keysize, NULL, (char*)data, sizeof(*data));
    memcpy(out_share, data->data, sizeof(*out_share));

	char digest[kDigestSize] = {0};
    calculate_share_digest(out_share, prime, digest);

    return (0 == memcmp(digest, data->digest, sizeof(digest)));
}


static bool sss_decode_shares_for_password(const char* kbdata, size_t kbsize, const KGPassword* pwd, BCCipher* cipher, int hash, std::vector < share_t > * out_shares)
{
    SSS_KeyBlock* kb    = (SSS_KeyBlock*)kbdata;
	char* prime         = (char*)kb->primeAndSalt;
	char* salt          = (char*)kb->primeAndSalt + sizeof(kb->primeAndSalt) - kSaltSize;
    unsigned max_shares = (kbsize - kPrimeAndSaltSize) / sizeof(encoded_share_t);

    if(max_shares == 0) {
        return false; // not enough key block size for a single share
    }

    // generate storage key from password
    char key[kMaximumKeySize] = {0};
    generate_key_from_password(hash, pwd->buf, pwd->nbytes, salt, kSaltSize, kSSS_HashIterations, key, sizeof(key));

	share_t share;
    bool found = false;

	for(unsigned i = 0; i < max_shares; ++i) {
        
        encoded_share_t* pshare = &kb->shares[i];

        if (sss_decode_share(pshare, i, prime, cipher, key, sizeof(key), &share)) {
            out_shares->push_back(share);

            // recalculate total number of shares now when we've decoded the first share
            if(!found) {
                assert(share.total <= max_shares);
                max_shares = share.total;
            }

            found = true;
        }
    }

	return found;
}


static bool sss_commit_shares(const std::vector<share_t>* shares, SSShares* table) 
{
	pw_int x,y;
	
    std::vector<share_t>::const_iterator i = shares->begin();
    std::vector<share_t>::const_iterator end = shares->end();

	while(i != end) {
        i->loadX(x);
        i->loadY(y);
    
        int error = SSSAddShare(table, &x, &y);
        if (pw_OK == error) {
            // success here means that there are enough shares to decode the key.
            // otherwise we just need to move on
            return true;
        }
	}

    return false; // not enough share to decode the key
}


typedef struct {
    const char*             kbdata;
    unsigned                kbsize;
    BCCipher*               cipher;
    int                     hash;
    std::vector<share_t>*   shares;
} DecodeSharesCtx;

static const char* decodeSharesCallback(const char* password, unsigned pwdsize, void* arg)
{
    DecodeSharesCtx* ctx = (DecodeSharesCtx*) arg;
    
    KGPassword pwd;
    pwd.set(password);

    if (!sss_decode_shares_for_password(ctx->kbdata, ctx->kbsize, &pwd, ctx->cipher, ctx->hash, ctx->shares)) {
        return "Invalid additional password";
    } else {
        return NULL;
    }
}

static bool askPasswordAndDecodeShares(const char* kbdata, unsigned kbsize, BCCipher* cipher, int hash, std::vector<share_t>* out_shares)
{
    DecodeSharesCtx ctx = {kbdata, kbsize, cipher, hash, out_shares};
    BCPassword pwd;
    return (kBCUserOk == BCUserVerifyPassword("Please provide additional password", decodeSharesCallback, &ctx, &pwd));
}

bool sss_decode(const char* kb_data, unsigned kb_size, const KGPassword* password, BCCipher* cipher, int hash, key_data_t* out_key)
{
	SSS_KeyBlock* kb = (SSS_KeyBlock*) kb_data;
	
	// decode shares for the user of the initial password so we can get total number of encoded and required shares
    std::vector<share_t> shares;
	if(!sss_decode_shares_for_password(kb_data, kb_size, password, cipher, hash, &shares)) {
        return false;
	}
	
    unsigned total_shares = shares[0].total;
    unsigned threshold = shares[0].threshold;
    unsigned decoded_shares = shares.size();

    assert(threshold <= total_shares);

    SSShares* sss;
    if (ppe_ok != SSSAllocateShares(&sss, threshold)) {
        return false;
    }

	// initialize sss runtime prime with the first byte from decoded share and the rest from raw keyblock data
	char* primeLow = (char*)kb->primeAndSalt;
	BYTE prime[kMaxPrimeSize] = {0};
	prime[0] = shares[0].headOfPrime;
	memcpy(prime + 1, primeLow, kMaxPrimeSize - 1);
	
    if (pw_OK != sss->prime.set(kMaxPrimeSize, prime)) {
        return false;
    }

	// dump initial user's share to initialized runtime table, see if that's actually enough to decode the key
    if (sss_commit_shares(&shares, sss)) {
	    SSSRestoreKey(sss, (BYTE*)out_key, sizeof(*out_key));
        SSSFree(&sss);
        return true;
    }

    // proceed with additional passwords
    while(decoded_shares < threshold) {
        shares.clear();

        if(!askPasswordAndDecodeShares(kb_data, kb_size, cipher, hash, &shares)) {
            SSSFree(&sss);
            return false;
        }

        if(sss_commit_shares(&shares, sss)) {
	        SSSRestoreKey(sss, (BYTE*)out_key, sizeof(*out_key));
            SSSFree(&sss);
            return true;   
	    }

        decoded_shares += shares.size();
    }

    // should never reach here
    return false;
}
	

bool sss_verify(const char* kbdata, unsigned kbsize, const KGPassword* password, BCCipher* cipher, int hash)
{
    std::vector<share_t> shares;
    return sss_decode_shares_for_password(kbdata, kbsize, password, cipher, hash, &shares);
}


//
//
//	REFACTOR
//
//


#if 0
int sss_share_t::encode(bc_byte_t* oBuff, unsigned int* ioBuffSize, CBCAlg_Handle alg, int hash_alg_id, WORD share_num, bc_byte_t* primeBuffer)
{
	unsigned int size = sizeof(*this);
	bc_byte_t iv[IVECTOR_LENGTH_DB] = {0};
	int rc;

	BC_ASSERT(*ioBuffSize >= size);
	if(*ioBuffSize < size)
		return ERROR_INTERNAL_PROBLEM;
			

	// fill share digest
	rc = this->calculate_digest(primeBuffer, hash_alg_id);
	if (rc != ERROR_NO)
		return rc;



	// Encode share by encrypting it into buffer
	// Use share number as initial vector
	iv[0] = LOBYTE(share_num);
	iv[1] = HIBYTE(share_num);

	memcpy(oBuff, (const char*)this, size);
	if ( !alg->encrypt(iv, oBuff, size) )
		return ERROR_INTERNAL_PROBLEM;

	*ioBuffSize = size;

	return ERROR_NO;
}

//
//	encodeShares
//	 0. Checks for different passwords in itemT table
//	 1. Generates prime p one byte more than the key size
//	 2. Generates (recNum - 1) degree polinom P(x) = Summ( Ai*(x exp i)) , where 0 <= i <= recNum -1,
//	 3. Generates shares array by Xi and Yi where P(Xi) == Yi, the array quantity is Summ of itemT[i].sharesQ, where 0 <= i < itemQ
//	 4. For every i gets N = itemT[i].sharesQ shares from generated shares array and encrypts the shares by passwords from itemT[i].pwd
//
//		Encrypted item format is:
//			encrypted:
//				Ver + RN + PrimeAlderByte + Xi + Yi + hashAlgId + reservedBytes + digest
//		where:
//			Ver ( 1 byte ) is version number
//				Ver 1 has 152 bytes size of encrypted share ( see getShareSize() )
//			RN  ( 1 byte ) is number of required shares.
//			PrimeAlderByte - prime is saved in open state but it's alder bytes has assigned alder bit,
//							it is a reason why the byte was hidden to encrypted part.
//			The values is required to recovering key,
//			but it has to be encrypted, so we save it here.
//
//			reservedBytes is random filled value it used for normal encryption with any block based algorithm 
//
//		digest size is defined as MAXIMUM_DIGEST_SIZE_BYTES
//
//		prime and Xi Yi sizes is defined by key size, see MaxPrimeSize
//
//
//		Maximum estimated shares size for 3K block:
//			V1 is -> 70 + N * ( sizeof( sShareV1 ) == 192 ) -> N = ( 3K - 70 ) ) / 192 > 15 shares
//			V2 is -> 70 + N * ( sizeof( sShareV2 ) == 112  )-> N = ( 3K - 70 ) ) / 112 > 26 shares
//
//
//	encodeSharesEx
//		walks tables with share's passwords ( itemT )
//			1. creates symmetric algorithm from password
//			2. fills shares and calculates it's digest
//				the digest is started from plain text primes low bytes 
//				and finished by prepared shares, so we may check the prime integrity
//				together with the share in decoding porcess.
//			3. encodes corresponded share quantity from generated shares table ( sss )
//
int encode_shares(char* oBuff, unsigned int* ioBuffSize, SSShares* sss_runtime)
{
	// get size of the single share
	unsigned int shareSize = sizeof(share_type);
	if ( shareSize <= 0 )
		return ERROR_INCORRECT_FORMAT_VERSION;


	unsigned int totalShares = this->total_shares_ready();
	unsigned int requiredSize = primeAndSaltSize + totalShares * shareSize;
	
	if ( *pBuffSize < requiredSize )
	{
		*pBuffSize = requiredSize;
		return ERROR_BUFFER_TOO_SMALL;
	}


	*pBuffSize = requiredSize;

	unsigned int primeSizeBytes = pw_getSizeInBytes(&sss_runtime->prime);
	if ( buffSize < primeSizeBytes )
		return ERROR_INTERNAL_PROBLEM;


	// Symmetric algorithm for share encryption, key is generated from corresponded shares' password.
/*
	cSymAlgorithm sPwdAlg;

	if ( !sPwdAlg.Init( algId, encModeId ) )
	{
		assert( false );
		return ERROR_INVALID_ALGORITHM;
	}

	int pwdKeySizeBytes = ( sPwdAlg.m_keyLengthInBites + 7 ) / 8;
*/

	//	Writes prime's low bytes to buffer (where the password key will be stored later )
	//	and save alder byte to be hidden to encrypted part.

	int buffSize = max( pwdKeySizeBytes, primeAndSaltSize + 1 );

	// Deploys prime low bytes to buffer start and saves alder byte 
	cPlainTextBlock ptb;
	if ( ptb.Allocate( buffSize ) == NULL )
	{
		assert( false );
		return ERROR_INTERNAL_PROBLEM;
	}



	pw_int2bytes( ptb.buffer, primeSizeBytes, &sss->prime );

	BYTE *pAlderByte = &ptb.buffer[0];

	int pos = 0;
	memcpy( oBuff, ptb.buffer + 1, primeSizeBytes );
	pos += primeAndSaltSize;



	// Prepare buffer for plain text share
	cPlainTextBlock ptbShare;
	if(ptbShare.Allocate(shareSize) == NULL)
		return ERROR_INTERNAL_PROBLEM;

	share_type* share = (share_type*)ptbShare.buffer;
	share->version		= this->share_version<share_type>();
	share->reqNum		= this->shares_needed();
	share->totalQ		= (BYTE)totalShares;
	share->headOfPrime	= *pAlderByte;

	if ( primeAndSaltSize - saltSize < 0 )
		return ERROR_INTERNAL_PROBLEM;


	BYTE *salt = oBuff + primeAndSaltSize - saltSize;

	unsigned int itemN = 0;	// position in itemT
	unsigned int sssN = 0;	// position in sss

	CBCAlg_Handle alg;

	for(unsigned int i; i < m_shares_total; ++i)
	{

		res = alg_init_from_password(alg, m_algId, m_encModeId, NULL /*zero pool*/, &item[i].pwd, salt, saltSize, m_hashAlgId, iteration);
		if( ERROR_NO != res ) return res;		

		res = this->encode_single_share < share_type > (share, sss_runtime, alg, );
		if( ERROR_NO != res ) return res;		
/*

		for( int i = 0; i < itemT[ itemN ].numberOfShares; i++, sssN++ )
		{
			if ( sssN >= (int)sss->arraySize )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			// Fills x and y values

			if ( !fillValue( sv->valX, sizeof( sv->valX ), &sss->array[ sssN ].x )  ||
					!fillValue( sv->valY, sizeof( sv->valY ), &sss->array[ sssN ].y ) 
					)
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			if ( pos + shareSize > requiredSize )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}
			
			// Calculates digest of
			//	1. plain text part of prime.
			//	2. the share info exluding the digest part.
			int err = sssCalculateDigest( buff, sv, (BYTE)hashAlgId );

			if ( err != ERROR_NO )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			// Encode share directly to represent buffer
			// It uses number of share as initial vector

			iv[0] = LOBYTE( WORD( sssN ) );
			iv[1] = HIBYTE( WORD( sssN ) );

			if ( !sPwdAlg.encrypt( iv, buff + pos, ( BYTE * )sv, shareSize ) )
			{
				assert( false );
				return ERROR_INTERNAL_PROBLEM;
			}

			pos += shareSize;
		}

*/
	}

	if ( pos != requiredSize )
		return ERROR_INTERNAL_PROBLEM;


	return ERROR_NO;
}

bool CSSSProcessor::encode(bc_byte_t* kbPlain, unsigned int & kbSize, key_info_t const & key, cPassword const & init_pwd)
{
	SSShares *sss_runtime = NULL;

	unsigned int totalSharesQ = this->get_total_shares();

	int res = SSSAllocateShares( &sss_runtime, totalSharesQ );
	if ( res != ppe_ok )
		return ERROR_INTERNAL_PROBLEM;
/**
	if ( version == (BYTE)-1 )
		version = defaultSSSVersion;
*/
/**
	// Version 2 uses one byte x values, so theirs are generated from order number
	if ( ( version == 2 ) && 
			!prepareOrderedX( s_table ) )
	{
		SSSFree( &s_table );
		return false;
	}
*/

	res = this->share_coder()->prepareX(sss_runtime);
	if( ERROR_NO != res ) 
	{
		SSSFree(&sss_runtime);
		return res;
	}

	int seedSize = 0;

	res = SSSGenerateShares(sss_runtime, totalSharesQ, keyBuff, keyBuffSize, NULL, &seedSize);
	if ( res != ppe_randomBufferRequired )
	{
		SSSFree(&sss_runtime);
		return false;
	}


	//
	//	Check if seed value has been generated early.
	//
	if ( seedSize > ptbSeed.bufferSize )
	{
		if ( ptbSeed.Allocate( seedSize ) == NULL )
		{
			SSSFree(&sss_runtime);
			return false;
		}

		res = User_GenerateSeed(ptbSeed.buffer, seedSize);
		if ( res != ERROR_NO )
		{
			SSSFree(&sss_runtime);
			return false;
		}
	}

	BYTE *seed = ptbSeed.buffer;

	//cProgressDlg_ShareGen dlgMT( s, reqQ, keyBuff, keyBuffSize, seed, seedSize );
	//dlgMT.DoModal( hMainInst,IDD_WaitShareGenerationDlg, hWnd );

	//res = dlgMT.m_res;

	if ( res != ppe_ok )
	{
		SSSFree( &s_table );
		return ERROR_INTERNAL_PROBLEM;
	}


	res = this->share_coder()->encode_shares( version, kbEncrypted, &kbEncryptedSize, itemT, itemQ, sss_runtime, m_algId, m_encModeId, m_hashAlgId, reqQ );

	SSSFree(&sss_runtime);

	return res;
	return ERROR_INCORRECT_FORMAT_VERSION; // stub
}

int CSSSProcessor::getSeedSize( int keyBuffSize )
{
	SSShares *s = NULL;
	int totalSharesQ = getTotalSharesQ();

	int res = SSSAllocateShares( &s, itemQ );
	if ( res != ppe_ok )
		return -1;

	if ( version == (BYTE)-1 )
		version = 2;

	// Version 2 uses one byte x values, so theirs are generated from order number
	if ( ( version == 2 ) && 
			!prepareOrderedX( s ) )
	{
		SSSFree( &s );
		return -1;
	}

	int seedSize = 0;

	res = SSSGenerateShares( s, totalSharesQ, NULL, keyBuffSize, NULL, &seedSize );
	SSSFree( &s );

	if ( res != ppe_randomBufferRequired )
	{
		return -1;
	}

	return seedSize;
}


class cProgressDlg_ShareGen : public cProgressBlock, public cDlg
{
	SSShares *m_s;
	int m_reqQ;
	BYTE *m_keyBuff;
	int m_keyBuffSize;
	BYTE *m_seed;
	int m_seedSize;

	bool m_cancelled;
	HWND m_hProgress;
//	bool m_useMarqueeMode; 

public:
	int m_res;

	cProgressDlg_ShareGen( SSShares *s, int reqQ, BYTE *keyBuff, int keyBuffSize, BYTE *seed, int seedSize )
	{
		m_s = s;
		m_reqQ = reqQ;
		m_keyBuff = keyBuff;
		m_keyBuffSize = keyBuffSize;
		m_seed = seed;
		m_seedSize = seedSize;
		m_res = -1;
		m_cancelled = false;
	}

	BOOL InitDialog()
	{
		m_hProgress = GetDlgItem( m_hDlg, IDC_PROGRESS );

		SendMessage( m_hProgress, PBM_SETRANGE, 0, MAKELPARAM( 0, 100 ) );
//		m_useMarqueeMode = SendMessage( m_hProgress, (WM_USER+10), true, 0 ); // PBM_SETMARQUEE
		SendMessage( m_hProgress, PBM_SETSTEP, 1, 0 );

		DWORD threadId;
		HANDLE hThread = CreateThread( NULL, 0, threadFunc, this, 0, &threadId );

		if ( hThread == NULL )
		{
			MessageBoxT( m_hDlg, L"Error in thread creating", LoadStringOnceT( IDS_KeyGen ), MB_ICONSTOP );
			EndDialog( m_hDlg, -1 );
			return false;
		}

		CloseHandle( hThread );

		return TRUE;
	}

	int Generate()
	{
		m_res = SSSGenerateShares( m_s, m_reqQ, m_keyBuff, m_keyBuffSize, m_seed, &m_seedSize, this );
		SendMessage( m_hProgress, PBM_SETPOS, true, 0 );
		return m_res;
	}

	static DWORD WINAPI threadFunc( void *ptr )
	{
		cProgressDlg_ShareGen *pThis = ( cProgressDlg_ShareGen * )ptr;

		pThis->Generate();

		PostMessage( pThis->m_hDlg, WM_COMMAND, IDOK, 0 );

		return 0;
	}
	
	void SetText( const char *text )
	{
		SetDlgItemText( m_hDlg, IDC_Text, text );
	}

	// Progress test result enum
	// enum { stepResult_False_ByTable = -1, stepResult_False_ByPrimeTest = 0, stepResult_True = 1 };
	
	//
	// SetCurrentStepResult
	//	Reports result of every step, see enum above.
	//	Returns false to stop the process, otherwise returns non zero.
	//
	bool SetCurrentStepResult( int res )
	{
		if ( m_cancelled )
			return false;

//		if ( res == stepResult_False_ByPrimeTest )
//			SendMessage( m_hProgress, PBM_SETPOS, 0, 0 );
//		SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );

		if ( res == stepResult_False_ByPrimeTest )
			SendMessage( m_hProgress, PBM_STEPIT, 0, 0 );

		return true;
	}

	LRESULT RealDlgProc( UINT uMsg, WPARAM wParam, LPARAM lParam )
	{
		switch (uMsg) 
		{ 
			case WM_INITDIALOG:
				return InitDialog();

	        case WM_COMMAND: 
			{
				switch(wParam) 
				{ 
					case IDOK: 
						EndDialog( m_hDlg, m_res );
						return TRUE; 
					case IDCANCEL:
					{
						int res = MessageBoxT(
									m_hDlg,
									LoadStringOnceT ( IDS_Q_StopGenerationProcess ),
									LoadStringTwiceT( IDS_KeyGen ),
									MB_ICONQUESTION | MB_YESNO | MB_DEFBUTTON2 );

						if ( res != IDYES )
								return 0;

						m_cancelled = true;
						EnableWindow( GetDlgItem( m_hDlg, IDCANCEL ), false );
						return TRUE;
					}
					default:
						return 0; 
				}
			}
			default:
				return 0; 
		}

		return 0;
	}
};

#endif

