/*
 *	Copyright (c) 1994-2010 Jetico, Inc., Finland
 *	All rights reserved.
 */

#include <stdlib.h>
#include <string.h>

#include "container/cipher.h"
#include "datablock.h"
#include "hash.h"


#define KB_HIDDEN_MASK_POSITION		KB_HIDDEN_OFFSET_POSITION - sizeof(bc_u32)
#define KB_HIDDEN_OFFSET_POSITION	kKGSHA_KeySize - sizeof(bc_u32) * 2
#define KB_IV_POSITION			kKGSHA_KeySize	// iv is the first 64 bits of digest.

#define DB_FOREACH_KEYBLOCK(_db, _kb, _index) \
	for(_index = 0, _kb = _db->keys; _index < kKGSHA_TotalKeys; ++_index, ++_kb) 


static inline void KeyBlock_SetAttribute(KEY_BLOCK* kb, int attr) {
	kb->keyAttribute = (~KB_ATTRIBUTE_MASK | attr);
}

static inline void KeyBlock_MarkEmpty(KEY_BLOCK* kb) {
	KeyBlock_SetAttribute(kb, KB_ATTRIBUTE_Empty);
}

static inline void KeyBlock_MarkOccupied(KEY_BLOCK* kb) {
	KeyBlock_SetAttribute(kb, KB_ATTRIBUTE_Occupied);
}

static inline char* KeyBlock_KeyData(KEY_BLOCK* kb) {
	return kb->keyAndDigest;
}

static inline char* KeyBlock_DigestData(KEY_BLOCK* kb) {
	return kb->keyAndDigest + kKGSHA_KeySize;
}


static inline void DataBlock_UpdateDigest(DATA_BLOCK *db) {
	size_t nbytes = sizeof(*db) - kKGSHA_DigestSize - kKGSHA_PoolSize; // do not hash the tail of the data block with pool and digest.
	KGSHA_CalculateDigest(db->header.hashAlgorithmId, (const char*)db, nbytes, db->digest);
}


static bool DataBlock_VerifyDigest(DATA_BLOCK *db) {
	char digest[kKGSHA_DigestSize] = {0};
	size_t digest_size = KGSHA_DigestSize(db->header.hashAlgorithmId);
	size_t nbytes = sizeof(*db) - kKGSHA_DigestSize - kKGSHA_PoolSize; // do not hash the tail of the data block with pool and digest.
	
	KGSHA_CalculateDigest(db->header.hashAlgorithmId, (const char*)db, nbytes, digest);
	
	return (0 == memcmp(db->digest, digest, digest_size));
}


static inline bool DataBlock_IsPoolInitialized(DATA_BLOCK* db) {
	return ((db->header.state & KB_ATTRIBUTE_MASK) == DB_STATE_PoolInitialized);
}


static VERIFICATION_BLOCK* DataBlock_FindVerificationBlock(DATA_BLOCK* db) 
{
	KEY_BLOCK* kb = db->keys;
	for(unsigned i = 0; i < kKGSHA_TotalKeys; ++i, ++kb) {
		if(KeyBlock_IsVerificationBlock(kb)) {
			return (VERIFICATION_BLOCK*) kb;
		}
	}
	
	return NULL;
}


// Init verification block for container key.
// This procedure will not do anything if verificaion block already exists.
static bool DataBlock_InitVerificationBlock(DATA_BLOCK* db, kgsha_key* key, BCCipher* cipher)
{
	KEY_BLOCK* kb = db->keys;
	VERIFICATION_BLOCK* vb = NULL;
	
	for(unsigned i = 0; i < kKGSHA_TotalKeys; ++i, ++kb) {
		if(KeyBlock_IsEmpty(kb)) {
			vb = (VERIFICATION_BLOCK*) kb;
			break;
		}
	}
	
	if(!vb) {
		return false;
	}
	
	
	// Empty key block is already filled in by random data.
	// We'll create digest for the data and encrypt it using the Key
	int hash = db->header.hashAlgorithmId;
	KGSHA_CalculateDigest(hash, vb->verifyBytesAndDigest, kKGSHA_VerBufferSize, vb->verifyBytesAndDigest + kKGSHA_VerBufferSize);
	cipher->encrypt(0, key->data, sizeof(key->data), NULL, vb->verifyBytesAndDigest, sizeof(vb->verifyBytesAndDigest));
	KeyBlock_SetAttribute((KEY_BLOCK*)vb, KB_ATTRIBUTE_Verification);
	
	return true;
}


////////////////////////////////////////////////////////////////////
//
//	Interface.
//
////////////////////////////////////////////////////////////////////


void DataBlock_Init(DATA_BLOCK* db, int alg_id, int hash_id)
{
	// randomize everything.
	KGSHA_ShredData((char*)db, sizeof(*db));
	
	db->header.formatVersion	= FORMAT_VERSION;
	db->header.encrAlgorithmId	= alg_id;
	db->header.hashAlgorithmId	= hash_id;
	db->header.dataBlockSize	= sizeof(DATA_BLOCK);
	db->header.keyBlockSize		= sizeof(KEY_BLOCK);
	db->header.keyBlockNumber	= kKGSHA_TotalKeys;
	
	// earlier versions of the kgsha initialized the random pool when the first key was stored by encrypting it.
	// now we leave the pool as it is "decrypting" it when user enters correct password.
	// so our pool is always "initialized" but earlier versions depend on this state flag so we leave it here anyway.
	// this will work because earlier version will think that our pool is encrypted and decrypt it just as we do.
	// and we will do the same for the really encrypted pool that came from earlier version and produce the same result.
	db->header.state		= DB_STATE_PoolInitialized; 
	
	strncpy(db->header.signature, KGSHA_SIGNATURE03, sizeof(db->header.signature));
	
	KEY_BLOCK* pkb = db->keys;
	unsigned kbnum;
	
	DB_FOREACH_KEYBLOCK(db, pkb, kbnum) {
		KeyBlock_MarkEmpty(pkb);
	}
	
	DataBlock_UpdateDigest(db);
}


bool DataBlock_Verify(DATA_BLOCK* db)
{
	if(db->header.formatVersion != FORMAT_VERSION)
		return false;
	
	if(db->header.dataBlockSize != sizeof(DATA_BLOCK)) 
		return false;
	
	if((db->header.hashAlgorithmId != kSHA_ID) &&
	   (db->header.hashAlgorithmId != kSHA256_ID))
		return false;
	
	if((0 != strcmp(db->header.signature, KGSHA_SIGNATURE98)) && 
	   (0 != strcmp(db->header.signature, KGSHA_SIGNATURE03)) &&
	   (0 != strcmp(db->header.signature, KGSHA_SIGNATURE04))) {
		return false;
	}
	
	if(!DataBlock_VerifyDigest(db)) 
		return false;
	
	return true;
}


unsigned DataBlock_CountFreeKeyBlocks(DATA_BLOCK* db) 
{
	unsigned nfree = 0;

	KEY_BLOCK* kb;
	unsigned kbnum;
	DB_FOREACH_KEYBLOCK(db, kb, kbnum) {
		if(KeyBlock_IsEmpty(kb)) {
			++nfree;
		}
	}
	
	return nfree;
}

/*
 * Generates random key based on user-generated seed.
 */
void DataBlock_GenerateKey(DATA_BLOCK* db, const char* seed, bool hidden, unsigned int offset, kgsha_key* out_key)
{
	out_key->shred();
	
	KGSHA_RandomGenerateKey(db->header.hashAlgorithmId, seed, kKGSHA_SeedSize, out_key->data, kKGSHA_KeySize);
	
	out_key->is_hidden = hidden;
	
	if(hidden) {
		out_key->data_offset = offset;
	} else {
		assert(offset == 0);
		out_key->data_offset = 0;
	}
}


void DataBlock_GenerateStorageKey(DATA_BLOCK* db, const char* pwdbuf, size_t pwdlen, kgsha_password* out_pwd) 
{
	out_pwd->shred();
	if(db->header.hashAlgorithmId == kSHA256_ID) {
		KGSHA_RandomGenerateKey(db->header.hashAlgorithmId, pwdbuf, pwdlen, out_pwd->storage_key, sizeof(out_pwd->storage_key));
	} else {
		KGSHA_RandomGenerate(db->header.hashAlgorithmId, pwdbuf, pwdlen, out_pwd->storage_key, sizeof(out_pwd->storage_key));
	}
}


KEY_BLOCK* DataBlock_StoreKeyWithPassword(DATA_BLOCK* db, kgsha_key* key, const kgsha_password* pwd, BCCipher* cipher)
{
	int		hash = db->header.hashAlgorithmId;
	KEY_BLOCK*	kb = NULL;
	unsigned	kbnum;
	
	if(!key->is_hidden) {
		// for plain keys find the first free keyblock.
		for(kbnum = 0; kbnum < kKGSHA_TotalKeys; ++kbnum) {
			if(KeyBlock_IsEmpty(&db->keys[kbnum])) {
				kb = &db->keys[kbnum];
				break;
			}
		}
	} else {
		// for hidden keys find the last free keyblock that is not the one where the key is currently encoded.
		for(kbnum = kKGSHA_TotalKeys - 1; kbnum >= 0; --kbnum) {
			if(KeyBlock_IsEmpty(&db->keys[kbnum])) {
				kb = &db->keys[kbnum];
				break;
			}
		}
	}

    if(!kb) {
        return NULL;
    }

	KeyBlock_EncodeKey(key, pwd->storage_key, cipher, hash, kb);

	// then find another free keyblock for verification data if it is not yet initialized.
	if(!DataBlock_FindVerificationBlock(db)) {
		if(!DataBlock_InitVerificationBlock(db, key, cipher)) {
			return NULL;
		}
	}
	DataBlock_UpdateDigest(db);
	
	return kb;
}


KEY_BLOCK* DataBlock_VerifyPassword(DATA_BLOCK* db, const kgsha_password* pwd, BCCipher* cipher)
{
	unsigned kbnum;
	KEY_BLOCK* kb;
	kgsha_key tmpkey;
	
	for(kbnum = 0; kbnum < kKGSHA_TotalKeys; ++kbnum) {
		kb = &db->keys[kbnum];
		if(KeyBlock_DecodeKey(kb, pwd->storage_key, cipher, db->header.hashAlgorithmId, &tmpkey)) {
			return kb;
		}
	}
	
	return NULL;
}


KEY_BLOCK* DataBlock_LoadKeyWithPassword(DATA_BLOCK* db, const kgsha_password* pwd, BCCipher* cipher, kgsha_key* out_key, char* out_pool)
{
	unsigned kbnum;
	KEY_BLOCK* kb = NULL;
	
	for(kbnum = 0; kbnum < kKGSHA_TotalKeys; ++kbnum) {
		kb = &db->keys[kbnum];
		if(KeyBlock_DecodeKey(kb, pwd->storage_key, cipher, db->header.hashAlgorithmId, out_key)) {
			assert(DataBlock_IsPoolInitialized(db));
			
			// "decrypt" the random pool.
			// see comments in DataBlock_Init.
			memcpy(out_pool, db->pool, kKGSHA_PoolSize);
			cipher->decrypt(out_key->iv, out_key->data, kKGSHA_KeySize, NULL, out_pool, kKGSHA_PoolSize);
			
			return kb;
		}
	}
	
	return NULL;
}


void KeyBlock_EncodeKey(const kgsha_key* key, const char* storage_key, BCCipher* cipher, int hash, KEY_BLOCK* kb)
{
	char keyAndDigest[kKGSHA_KeySize + kKGSHA_DigestSize] = {0};
	memcpy(keyAndDigest, key->data, sizeof(key->data));

	size_t keySize = (key->is_hidden) ? kKGSHA_KeySize - sizeof(bc_u32) : kKGSHA_KeySize;
	
	if(key->is_hidden) {
		*(bc_u32*)(kb->keyAndDigest + KB_HIDDEN_OFFSET_POSITION) = key->data_offset;
	}
	
	KGSHA_CalculateDigest(hash, keyAndDigest, keySize, &keyAndDigest[kKGSHA_KeySize]);
	uint64_t iv = *(uint64_t*) (keyAndDigest + kKGSHA_KeySize);
	
	try {
		cipher->encrypt(iv, storage_key, kKGSHA_KeySize, NULL, keyAndDigest, sizeof(keyAndDigest));
	} catch (std::exception& e) {
		KGSHA_ShredData(keyAndDigest, sizeof(keyAndDigest));
		throw;
	}
	
	memcpy(kb->keyAndDigest, keyAndDigest, sizeof(keyAndDigest));
	
	// hidden keys do not change the empty attribute.
	if(!key->is_hidden) {
		KeyBlock_MarkOccupied(kb);
	}
}


bool KeyBlock_DecodeKey(const KEY_BLOCK* kb, const char* storage_key, BCCipher* cipher, int hash, kgsha_key* out_key)
{
	uint64_t	iv = 0;
	size_t		digest_size = KGSHA_DigestSize(hash);
	char		digest[kKGSHA_DigestSize] = {0};
	char		kbdata[kKGSHA_KeySize + kKGSHA_DigestSize] = {0};
	
	bool		is_hidden = KeyBlock_IsEmpty(kb);
	size_t		keySize = (is_hidden) ? kKGSHA_KeySize - sizeof(bc_u32) : kKGSHA_KeySize;

	/*
	 * We do two decryptions of the key and digest data in KEY_BLOCK:
	 * First with the zero iv to get only the correct iv and second to decrypt the rest of the data with it.
	 *
	 * The trick is that during initial encryption we use unencrypted key digest for iv, i.e. plaintext.
	 * And decryption in CBC mode with zero iv and pool will peel away the actual CBC transformation layer for the plaintext that equals real iv.
	 */

	iv = 0;
	memcpy(kbdata, kb->keyAndDigest, sizeof(kbdata));
	cipher->decrypt(iv, storage_key, kKGSHA_KeySize, NULL, kbdata, sizeof(kbdata));
	
	iv = *(uint64_t*) (kbdata + kKGSHA_KeySize);
	memcpy(kbdata, kb->keyAndDigest, sizeof(kbdata));
	cipher->decrypt(iv, storage_key, kKGSHA_KeySize, NULL, kbdata, sizeof(kbdata));
	
	// now we have potentially secret data on stack in kbdata
	// shred it when it is not needed exiting.
	// ------------------------------------------------------------------
	
	KGSHA_CalculateDigest(hash, kbdata, keySize, digest);
	bool rc = (0 == memcmp(kbdata + kKGSHA_KeySize, digest, digest_size));
	KGSHA_ShredData(digest, sizeof(digest));
	
	if(rc) {
		
		memcpy(out_key->data, kbdata, sizeof(out_key->data));
		KGSHA_ShredData(kbdata, sizeof(kbdata));
		out_key->iv = iv;
		
		// if the key was found in keyblock marked as empty we get a hidden part key
		if(KeyBlock_IsEmpty(kb)) {
            out_key->is_hidden      = true;
            out_key->data_offset    = *(bc_u32*)(out_key->data + KB_HIDDEN_OFFSET_POSITION);
		} else {
            out_key->is_hidden      = false;
            out_key->data_offset    = 0;
		}
	}
	
	return rc;
}


void KeyBlock_Erase(KEY_BLOCK* kb)
{
	KGSHA_ShredData((char*)kb, sizeof(*kb));
	KeyBlock_MarkEmpty(kb);
}

