/*
 * Copyright (c) 1994-2010 Jetico, Inc., Finland. All rights reserved.
 *
 * Definition of KGSHA DATA_BLOCK structure and interface.
 * POD types that are directly serialized to/from container file are written in all caps.
 */

#ifndef BC_KGSHA_DATABLOCK_H
#define BC_KGSHA_DATABLOCK_H

#include <limits.h>
#include <assert.h>
#include <string.h>

#include "container/intsize.h"
#include "defs.h"

class BCCipher;

#pragma pack(push, 1)


// Masks out the attribute flag in keyAttribute
#define KB_ATTRIBUTE_MASK		0x0000000f


/// Key block attribute.
enum {
	KB_ATTRIBUTE_Empty          = 0x00000001,	///< Empty key block. Set by default, not reset for hidden keys.
	KB_ATTRIBUTE_Occupied       = 0x00000002,	///< Occupied key block.
	KB_ATTRIBUTE_Partial        = 0x00000003,	///< Currently unused.
    KB_ATTRIBUTE_Verification   = 0x00000004	///< Verification key block.
};


/// Data block state flags.
enum {
    DB_STATE_PoolInitialized    = 0x00000002 	//< Data block pool has been encrypted by the main part key.
};


/**
 * On-disk encrypted key storage format.
 */
typedef struct KEY_BLOCK
{
	char		keyAndDigest[kKGSHA_KeySize + kKGSHA_DigestSize];	// encryption key and its digest all encrypted by user password
	bc_u32		keyAttribute;						// key attribute flags
} KEY_BLOCK;


/**
 * Alternative KEY_BLOCK with the same size to verify keys that come from Enterprise edition server.
 * Instead of container encryption key it contains random data and it's digest encrypted by main container key.
 */
typedef struct VERIFICATION_BLOCK
{
	// buffers below must occupy the same space as keyAndDigest KEY_BLOCK.
	char	reserved[kKGSHA_VerBlockReserved];
	char	containerID[kKGSHA_ContainerIdSize];
	char	verifyBytesAndDigest[kKGSHA_VerBufferSize + kKGSHA_DigestSize];
	
	bc_u32	keyAttribute;
} VERIFICATION_BLOCK;


/**
 * KGSHA data block header.
 */
typedef struct DB_HEADER
{
	char		signature[8];
	bc_u32		formatVersion;
	bc_u32		encrAlgorithmId;
	bc_u32		hashAlgorithmId;
	bc_u32		dataBlockSize;
	bc_u32		keyBlockSize;
	bc_u32		keyBlockNumber;
	bc_u32		state;
} DB_HEADER;


/**
 * Full KGSHA data block.
 */
typedef struct DATA_BLOCK
{
	DB_HEADER	header;
	KEY_BLOCK	keys[kKGSHA_TotalKeys];
	char		pool[kKGSHA_PoolSize];
	char		digest[kKGSHA_DigestSize];
} DATA_BLOCK;


#pragma pack(pop)


static inline bool KeyBlock_IsVerificationBlock(const KEY_BLOCK* kb) {
	return ((kb->keyAttribute & KB_ATTRIBUTE_MASK) == KB_ATTRIBUTE_Verification);
}


static inline bool KeyBlock_IsEmpty(const KEY_BLOCK* kb) {
	// Verification blocks do not reset the empty flag for some reason
	return ((kb->keyAttribute & KB_ATTRIBUTE_MASK) == KB_ATTRIBUTE_Empty) && !KeyBlock_IsVerificationBlock(kb);
}


/**
 * Safe password buffer.
 * Will not store actual password text and instead generate the storage key from it.
 */
typedef struct kgsha_password
{
	char storage_key[kKGSHA_KeySize];
	
	kgsha_password() {
		this->shred();
	}
	
	~kgsha_password() {
		this->shred();
	}
	
	void shred() {
		memset(storage_key, 0, sizeof(storage_key));
	}

} kgsha_password;

				  
/**
 * Logical representation of encryption key data stored and decoded from key blocks.
 * Methods encapsulate basic mechanism of encoding and decoding keys to/from KEY_BLOCKs.
 * The policies of locating correct keyblocks, asking password, etc are elsewhere.
 */
typedef struct kgsha_key 
{
	char		data[kKGSHA_KeySize];   ///< Key data.
	bc_u32		data_offset;            ///< Container data offset that this key encodes.
	bool		is_hidden;              ///< Key was found in a key block marked as free.
	bc_u64		iv;                     ///< Key initial vector, stored here for convinience when decoded.
	
	kgsha_key() {
		memset(data, 0, sizeof(data));
		is_hidden = false;
		data_offset = 0;
	}
	
	~kgsha_key() {
		this->shred();
	}

	void shred() {
		memset(data, 0, sizeof(data));
	}
	
} kgsha_key;


/**
 * Init data block structure to default state.
 * @param db            Data block.
 * @param alg_id        Algorithm id to be used with the data block.
 * @param hash_id       Hash algorithm id to use with data block.
 */
void DataBlock_Init(DATA_BLOCK* db, int alg_id, int hash_id);


/// Verify data block integrity.
bool DataBlock_Verify(DATA_BLOCK* db);


/// Return total number of free key blocks.
unsigned DataBlock_CountFreeKeyBlocks(DATA_BLOCK* db);


/**
 * Generate new encryption key.
 *
 * @param db            Data block to generate the key for.
 * @param hidden        Is this key hidden or not.
 * @param offset        For hidden keys this specifies data offset. Ignored for normal keys.
 * @param [out]out_key  Returned initialized key on success.
 * @note Tests:
 *      - Result template test.
 */
void DataBlock_GenerateKey(DATA_BLOCK* db, const char* seed, bool hidden, unsigned int offset, kgsha_key* out_key);


/**
 * Generate key block storage key from password.
 * @param db            Data block.
 * @param pwdbuf        Password.
 * @param pwdlen        Size of the password data in bytes.
 * @param out_pwd       Generated storage key on success.
 * @note Tests:
 *      - Result template test.
 */
void DataBlock_GenerateStorageKey(DATA_BLOCK* db, const char* pwdbuf, size_t pwdlen, kgsha_password* out_pwd);


/**
 * Store the key in data block using specified password.
 * Key can be obtained by generating a new one or decoding an existing one.
 * @warning Will not try to see if supplied password was already used to encode some other key block.
 * @param db            Data block structure to store the key.
 * @param key           Key to store.
 * @param password      User password.
 * @param cipher        Configured block cipher to use for encrypting the key.
 * @param kb            If not NULL will store the new key in this keyblock. Otherwise will find a suitable empty key block.
 * @return              KEY_BLOCK where key is now encoded or NULL if there was no free keyblock.
 * @note Tests:
 *      - Returns NULL, if no empty blocks.
 *      - If not hidden key, check that it is stored in first empty block.
 *      - For hidden key, check returned block as empty and last, and empty block count isn't changing.
 *      - For storing first time, check, that empty block count is kKGSHA_TotalKeys - 2 (verification block created).
 */
KEY_BLOCK* DataBlock_StoreKeyWithPassword(DATA_BLOCK* db, kgsha_key* key, const kgsha_password* pwd, BCCipher* cipher);


/**
 * Find and decode the key from data block using specified password.
 * @param db            Data block structure that stores the key.
 * @param password      User password to use.
 * @param cipher        Configured block cipher to decrypt keyblocks.
 * @param out_key       On success decoded key will be stored here.
 * @param[out] out_pool On success decrypted pool will be stored here.
 * @return	            KEY_BLOCK where key was found or NULL if password was incorrect.
 * @note Tests:
 *      - Template test for decrypted key.
 */
KEY_BLOCK* DataBlock_LoadKeyWithPassword(DATA_BLOCK* db, const kgsha_password* pwd, BCCipher* cipher, kgsha_key* out_key, char* out_pool);


/**
 * Verify that password is correct (i.e. decodes any keyblock).
 * @param db            Data block.
 * @param pwd           Password to verify
 * @param cipher        Configured block cipher.
 * @return              KEY_BLOCK which was correctly decoded by the password or NULL.
 */
KEY_BLOCK* DataBlock_VerifyPassword(DATA_BLOCK* db, const kgsha_password* pwd, BCCipher* cipher);


/**
 * Encode key directly into key block.
 * @param key           Key to encode.
 * @param storage_key   Storage key to encrypt keyblock with. Can be generated from password.
 * @param cipher        Block cipher to use to encrypt the keyblock.
 * @param hash          Hash algorithm id to calculate key digest.
 * @param[out] out_kb   Key block to store encoded data.
 * @note Tests:
 *      - Check correct encoding by template for all platforms.
 */
void KeyBlock_EncodeKey(const kgsha_key* key, const char* storage_key, BCCipher* cipher, int hash, KEY_BLOCK* kb);


/**
 * Decode key directly from the keyblock.
 * @param kb            Key block to decode.
 * @param storage_key   Storage key to decrypt keyblock with. Can be generated from password.
 * @param cipher        Block cipher to use to decrypt the keyblock.
 * @param hash          Hash algorithm id to calculate key digest.
 * @param[out] out_key  Decoded key on success.
 * @return              true if key could be decoded, false if not.
 * @note Tests:
 *      - Check correct decoding by template for all platforms.
 */
bool KeyBlock_DecodeKey(const KEY_BLOCK* kb, const char* storage_key, BCCipher* cipher, int hash, kgsha_key* out_key);


/**
 * Reinit key block with random data and mark free
 * @param kb            Keyblock to process.
 * @note Tests:
 *      - Erased key block must be empty.
 *      - Decoding must fail.
 */
void KeyBlock_Erase(KEY_BLOCK* kb);


#endif // BC_KGSHA_KEYBLOCK_H

