/*
 *  Copyright 2010 Jetico, Inc. All rights reserved.
 */

#include "hash.h"

#include <CertificateLib/hashAlgs/hashAlgs.h>
#include <CertificateLib/hashAlgs/cHashAlg.h>
#include <CertificateLib/hashAlgs/SHA1/Sha1.h>
#include <CertificateLib/hashAlgs/SHA256/sha256.h>

#include <assert.h>
#include <time.h>

// Map internal sha id into certificatelib's
static int map_sha_id(int sha_id)
{
	switch(sha_id) {
	case kSHA_ID:		return haID_SHA1;
	case kSHA256_ID:	return haID_SHA256;
	default:		assert(0);
	};
	
	return 0;
}

static cHashAlg* get_hash(int sha_id)
{
	cHashAlg* palg;
	if(hae_invalidAlgorithmID == ha_createAlgorithm(&palg, map_sha_id(sha_id))) {
		assert(0);
	}
	
	return palg;
}
	
	
size_t KGSHA_DigestSize(int hash_id)
{
	cHashAlg* alg = get_hash(hash_id);
	if(!alg) {
		assert(0);
	}
	
	return alg->getDigestSize();
}

void KGSHA_CalculateDigest(int hash_id, const char *data, size_t length, char *out_digest)
{
	cHashAlg* alg = get_hash(hash_id);
	if(!alg) {
		assert(0);
	}
	
	alg->makeDigest((BYTE*)out_digest, (BYTE*)data, length);
}


void KGSHA_RandomGenerate(int hash_id, const char* seed, size_t seed_size, char* out_buf, size_t nbytes)
{
	// due to compatibility reasons we use direct sha implementations here.
	if(hash_id == kSHA_ID) {
		SHA1RandomGenerate((BYTE*)out_buf, nbytes, (BYTE*)seed, seed_size);
	} else if(hash_id == kSHA256_ID) {
		SHA256_RandomGenerate((BYTE*)out_buf, nbytes, (BYTE*)seed, seed_size);
	} else {
		assert(0);
	}
}

void KGSHA_RandomGenerateKey(int hash_id, const char* seed, size_t seed_length, char *out_key, size_t out_key_size)
{
	unsigned int k, s, i, kReminder;
	size_t digestSize = KGSHA_DigestSize(hash_id);
	assert(digestSize);
	
	k = out_key_size  / digestSize;
	s = seed_length / k;
		
	if ( s >= digestSize )
	{
		for ( i=0; i < k; i++ )
		{
			if ( i == (k-1) ) kReminder = out_key_size - k * digestSize;
			else            kReminder = 0;
			
			KGSHA_RandomGenerate(hash_id,
					     &(seed[ s*i ]), 
					     seed_length - s*i,
					     &(out_key[ digestSize * i ]),
					     digestSize + kReminder);

		}
	} else {
		KGSHA_RandomGenerate(hash_id, seed, seed_length, out_key, out_key_size);
	}
}


void KGSHA_ShredData(char* data, size_t nbytes)
{	
	unsigned i;
	
	if ((!data) || (nbytes <= 0)) return;
	srand( (unsigned int)time( NULL ) );
	for(i=0; i < nbytes; i++)
	{
		data[i] = (BYTE)rand();
	}
	
	KGSHA_RandomGenerate(kSHA256_ID, data, nbytes, data, nbytes);
}

