/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 *
 * KGSHA keygen library interface.
 */

#include <stdlib.h>
#include <string.h>

#include "container/keygen.h"
#include "container/cipher.h"
#include "container/user.h"
#include "container/file.h"

#include "kgsha.h"
#include "hash.h"
#include "defs.h"
#include "datablock.h"


KGSHA_Header::KGSHA_Header(HiddenSector* hs, DATA_BLOCK* db)
{
    memcpy(&m_hs, hs, sizeof(*hs));
    memcpy(&m_db, db, sizeof(*db));
}

void KGSHA_Header::mainVolumeInfo(struct BCVolumeInfo* info)
{
    memset(info, 0, sizeof(*info));

    info->alg      = m_hs.algId;
    info->mode     = kBCMode_CBC;      // we only support CBC
    info->hash     = m_db.header.hashAlgorithmId;
    info->size     = m_hs.br.totalSectorsLong * 512LL;
    info->offset   = m_hs.dwDataOffset;

    strncpy(info->description, m_hs.description, sizeof(m_hs.description));
}

BCContainerResult KGSHA_Header::volumeInfoForKey(const kgsha_key* key, struct BCVolumeInfo* out_info)
{
    sm_result spaceMapResult;
    assert(key && out_info);

    memset(out_info, 0, sizeof(*out_info));

    out_info->alg       = m_db.header.encrAlgorithmId;
    out_info->mode      = kBCMode_CBC;
    out_info->hash      = m_db.header.hashAlgorithmId;
    out_info->hidden    = key->is_hidden;
    out_info->size      = m_hs.br.totalSectorsLong * 512LL;
    out_info->map       = SpaceMapCreate();

    if (!out_info->map)
        return kBCContainer_NoMemory;

    if (key->is_hidden) {
        out_info->offset = key->data_offset * 512LL;
        out_info->size -= out_info->offset;
    } else {
        out_info->offset = m_hs.dwDataOffset;
        strncpy(out_info->description, m_hs.description, sizeof(m_hs.description));
    }

    spaceMapResult = SpaceMapAddNew(out_info->map, out_info->offset, out_info->size);

    if (kSpaceMap_Success != spaceMapResult)
        return kBCContainer_NoMemory;

    return kBCContainer_Success;
}

void KGSHA_Header::setDescription(const char* description)
{
    memset(m_hs.description, 0, sizeof(m_hs.description));
    strncpy(m_hs.description, description, sizeof(m_hs.description));
}

void KGSHA_Header::markForReencryption(bool mark)
{
    if (mark) {
        strncpy(m_hs.br.volumeLabel, BCV7_REENCRYPTED_LABEL, sizeof(m_hs.br.volumeLabel));
    } else {
        strncpy(m_hs.br.volumeLabel, BCV7_VOLUME_LABEL, sizeof(m_hs.br.volumeLabel));
    }
}

bool KGSHA_Header::isMarkedForReencryption()
{
    return (0 == strncmp(m_hs.br.volumeLabel, BCV7_REENCRYPTED_LABEL,
                         sizeof(m_hs.br.volumeLabel)));
}

BCContainerResult KGSHA_Header::init(int algid, int modeid, bc_u64 size,
                                     const char* description, const char* password,
                                     BCProgressCallback progress, void *progressArg)
{
    // Init blank header and data block structures
    memset(&m_hs, '0', sizeof(HiddenSector));
    memset(&m_db, '0', sizeof(DATA_BLOCK));

    strncpy(m_hs.br.OEMid, "LOCOS94 ", 8);
    strncpy(m_hs.br.volumeLabel, "CRYPTED_DSK", 11);

    m_hs.br.jmpCode[0] = 0xEB;
    m_hs.br.jmpCode[1] = 0x3C;
    m_hs.br.jmpCode[2] = 0x90;
    m_hs.dwKeySize = sizeof(DATA_BLOCK);
    m_hs.dwDataOffset = (512 + sizeof(DATA_BLOCK) + 511) & ~511;
    m_hs.br.totalSectorsLong = (bc_u32)(size / 512) ;
    m_hs.algId = algid;
    m_hs.keygenId = kKGSHA256_ID;
    m_hs.version = 7;

    if(description) {
        size_t len = strlen(description);
        if (len != 0) {
            memcpy(m_hs.description, description, len > sizeof(m_hs.description) ?
                       sizeof(m_hs.description) : len);
	    }
    }

    DataBlock_Init(&m_db, algid, kSHA256_ID);
    assert(DataBlock_Verify(&m_db));

    // Init container key and encode it with the password
    if (modeid != kBCMode_CBC) {
        return kBCContainer_InvalidAlgorithm;
    }
        
    BCCipher cipher;
    if (kBCKernel_Success != cipher.init(algid, modeid)) {
        return kBCContainer_InvalidAlgorithm;
    }

    char seed[kKGSHA_SeedSize] = {0};
    if (sizeof(seed) != BCUserGenerateSeed("Generating seed for storage key.", seed, sizeof(seed),
                                           progress, progressArg)) {
        return kBCContainer_Cancelled;
    }

    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, password, strlen(password), &pwd);

    kgsha_key key;
    DataBlock_GenerateKey(&m_db, seed, false, m_hs.dwDataOffset, &key);

    DataBlock_StoreKeyWithPassword(&m_db, &key, &pwd, &cipher);
    return kBCContainer_Success;
}

BCContainerResult KGSHA_Header::createHiddenKey(SpaceMap *map,
                    const BCPassword* mainPassword, const BCPassword* hiddenPassword,
                    BCProgressCallback progress, void *progressArg)
{
    BCCipher cipher;

    if (kBCKernel_Success != cipher.init(m_db.header.encrAlgorithmId, kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }

    char seed[kKGSHA_SeedSize] = {0};
    if (sizeof(seed) != BCUserGenerateSeed("Generating seed for hidden key.", seed, sizeof(seed),
                                           progress, progressArg)) {
        return kBCContainer_Cancelled;
    }

    if ((!map) || (!map->start))
        return kBCContainer_InvalidArgument;

    // The data offset is equals to the start of the first extent.
    bc_u64 offset = map->start->extent.start;

    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, hiddenPassword->buf, hiddenPassword->length, &pwd);

    kgsha_key key;
    DataBlock_GenerateKey(&m_db, seed, true, offset, &key);

    if(!DataBlock_StoreKeyWithPassword(&m_db, &key, &pwd, &cipher)) {
        return kBCContainer_NoKeySpace;
    }
    
    return kBCContainer_Success;
}

BCContainerResult KGSHA_Header::verifyPassword(const char* password, struct BCVolumeInfo* out_info)
{
    BCCipher cipher;
    if (kBCKernel_Success != cipher.init(m_db.header.encrAlgorithmId, kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }
    
    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, password, strlen(password), &pwd);

    kgsha_key key;
    char pool[512];
    if (!DataBlock_LoadKeyWithPassword(&m_db, &pwd, &cipher, &key, pool)) {
        return kBCContainer_InvalidPassword;
    }

    return this->volumeInfoForKey(&key, out_info);
}

BCContainerResult KGSHA_Header::decodeKey(const char* password, struct BCVolumeInfo* out_info,
                                          int* out_handle, char* out_pool)
{
    BCCipher cipher;
    if (kBCKernel_Success != cipher.init(m_db.header.encrAlgorithmId, kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }
    
    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, password, strlen(password), &pwd);

    kgsha_key key;
    if (!DataBlock_LoadKeyWithPassword(&m_db, &pwd, &cipher, &key, out_pool)) {
        return kBCContainer_InvalidPassword;
    }

    if (kBCKernel_Success != BCKernelStoreKey(key.data, sizeof(key.data), out_handle)) {
        return kBCContainer_NoMemory;
    }

    return this->volumeInfoForKey(&key, out_info);
}

BCContainerResult KGSHA_Header::passwd(const char* password, const char* newPassword,
                                       BCKeyOp opflag)
{
    if (!DataBlock_Verify(&m_db)) {
        return kBCContainer_Damaged;
    }
    
    BCCipher cipher;
    if (kBCKernel_Success != cipher.init(m_db.header.encrAlgorithmId, kBCMode_CBC)) {
        return kBCContainer_InvalidAlgorithm;
    }
    
    kgsha_password pwd;
    DataBlock_GenerateStorageKey(&m_db, password, strlen(password), &pwd);

    kgsha_key key;
    char pool[512] = {0};

    KEY_BLOCK* kb = DataBlock_LoadKeyWithPassword(&m_db, &pwd, &cipher, &key, pool);
    if(!kb) {
        return kBCContainer_InvalidPassword;
    }

    switch (opflag) {
        // remove password.
        case kBCKey_Delete: {
            KeyBlock_Erase(kb);
            break;
        }

        // change password.
        case kBCKey_Recode: {
            kgsha_password new_pwd;
            DataBlock_GenerateStorageKey(&m_db, newPassword, strlen(newPassword), &new_pwd);

            KeyBlock_EncodeKey(&key, new_pwd.storage_key, &cipher,
                               m_db.header.hashAlgorithmId, kb);
            break;
        }

        // add password.
        case kBCKey_Duplicate: {
            kgsha_password new_pwd;
            DataBlock_GenerateStorageKey(&m_db, newPassword, strlen(newPassword), &new_pwd);
            
            if (!DataBlock_StoreKeyWithPassword(&m_db, &key, &new_pwd, &cipher)) {
                return kBCContainer_NoKeySpace;
            }
            
            break;
        }
            
        default: {
            return kBCContainer_Unsupported;
        }
    };
    
    return kBCContainer_Success;
}

BCContainerResult KGSHA_Header::addPassword(const char* password, const char* newPassword) {
    return this->passwd(password, newPassword, kBCKey_Duplicate);
}

BCContainerResult KGSHA_Header::changePassword(const char* password, const char* newPassword) {
    return this->passwd(password, newPassword, kBCKey_Recode);
}

BCContainerResult KGSHA_Header::removePassword(const char* password) {
    return this->passwd(password, NULL, kBCKey_Delete);
}

BCContainerResult KGSHA_Header::store(const char* file)
{
    bc_file_handle handle;
    if (!BCFileOpen(file, kBCFileMode_ReadWrite, &handle)) {
        return kBCContainer_IOError;
    }

    if (sizeof(HiddenSector) != BCFileWrite(&handle, &m_hs, sizeof(HiddenSector))) {
        BCFileClose(&handle);
        return kBCContainer_IOError;
    }

    if (sizeof(DATA_BLOCK) != BCFileWrite(&handle, &m_db, sizeof(DATA_BLOCK))) {
        BCFileClose(&handle);
        return kBCContainer_IOError;
    }

    BCFileClose(&handle);
    return kBCContainer_Success;
}

KGSHA_Header::~KGSHA_Header()
{
    memset(&m_hs, 0, sizeof(m_hs));
    memset(&m_hs, 0, sizeof(m_hs));
}



///////////////////////////////////////////////////////////////////////
//
//  Keygen exports.
//
///////////////////////////////////////////////////////////////////////


static IBCHeader* kgsha_load(const char* file)
{
    HiddenSector hs;
    DATA_BLOCK db;

    bc_file_handle handle;
    bc_file_result error = BCFileOpen(file, kBCFileMode_Read, &handle);
    if (kBCFileResult_Success != error) {
        return NULL;
    }

    bc_u64 bytesRead = BCFileRead(&handle, &hs, sizeof(hs));
    if (sizeof(hs) != bytesRead) {
        BCFileClose(&handle);
        return NULL;
    }
    
    bytesRead = BCFileRead(&handle, &db, sizeof(db));
    if (sizeof(db) != bytesRead) {
        BCFileClose(&handle);
        return NULL;
    }
        
    BCFileClose(&handle);

    if (0 != strncmp(hs.br.OEMid, "LOCOS94 ", sizeof(hs.br.OEMid))) {
        return NULL;
    }
    
    if ((kKGSHA_ID != hs.keygenId) && (kKGSHA256_ID != hs.keygenId)) {
        return NULL;
    }

    if (!DataBlock_Verify(&db)) {
        return NULL;
    }

    return new KGSHA_Header(&hs, &db);
}


static IBCHeader* kgsha_create(void)
{
    HiddenSector hs;
    DATA_BLOCK db;

    return new KGSHA_Header(&hs, &db);
}


static BCContainerResult kgsha_parse(const char* file, struct BCVolumeInfo* info)
{
    BCHeaderHandle header(kgsha_load(file));
    if (!header) {
        return kBCContainer_UnknownFormat;
    }

    header->mainVolumeInfo(info);
    return kBCContainer_Success;
}

struct BCKeygen kg_sha = 
{
    kKGSHA256_ID,
    "kgsha/kgsha256",
    "Compatibility with BestCrypt version 7",
    7,
    0,
    kgsha_create,
    kgsha_load,
    kgsha_parse
};

