/*
 *	Copyright Jetico, Oy, 2011.
 *	
 *	Mount helper tool.
 *	A suid tool that helps us mount bestcrypt virtual devices and set mount point ownership.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "common.h"


// Return 1 if filesystem type supports ownership/mode options.
// Return 0 otherwise.
static int fs_supports_ownership(const char* fs)
{
	if(fs == NULL) {
		return 0;
	}

	if(0 == strcmp(fs, "ntfs")) {
		return 1;
	} else if(0 == strcmp(fs, "vfat")) {
		return 1;
	} else if(0 == strcmp(fs, "msdos")) {
		return 1;
	} else {
		return 0;
	}
}


// Process options and launch /bin/mount.
// Change mount point ownership if requested.
static int do_mount(const char* dev, const char* mpoint, const char* type, int ro, int uid, int gid, int mode, const char* opts)
{

	pid_t pid = fork();
	if(pid < 0) {
        fprintf(stderr, "Error: fork failed: %s", strerror(errno));
		return EXIT_FAILURE;
	}

	const char* env[] = {"PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL};
	if( 0 == pid ) 
	{
        // Child.
		char ownership_opts[0xff] = {0};

		snprintf(ownership_opts, sizeof(ownership_opts) - 1, "uid=%d,gid=%d,umask=%o", uid, gid, ~(mode) & 07777 /* convert to umask */);

		// pass requested ownership options.
		// not all filesystems support this, so we can fail here.
	//	bc_get_root();

        setuid(0);
        if (type)
        {
            execle(	"/bin/mount",

                // mount arguments
                "/bin/mount",
                ro ? "-r" : "-w",
                "-t",
                type,
                "-o",
                (opts ? opts : "user,exec,suid,nodev"),
                "-o",
                (fs_supports_ownership(type) ? ownership_opts : ""),
                "-o",
                "sync",
                "--",
                dev,
                mpoint,
                NULL,
                env);
        }
        else
        {
            // If mount will fail, we do second try, so simple ignore
            // error output here.
            fclose(stderr);
            execle(	"/bin/mount",

                // mount arguments
                "/bin/mount",
                ro ? "-r" : "-w",
                "-o",
                (opts ? opts : "user,exec,suid,nodev"),
                "-o",
                ownership_opts,
                "-o",
                "sync",

                "--",
                dev,
                mpoint,
                NULL,

                env);
        }
		// exec failed if we're here
		return EXIT_FAILURE;

	}

	// Parent.
	int status;
	waitpid(pid, &status, 0);

    if (status)
    {
        // If we are here, mount has failed.

        // If filesystem was specified, simply return exit code.
        if (type)
            return WEXITSTATUS(status);

        // If no filesystem was specified, maybe ownership options are not accepted,
        // so try without them.
        pid = fork();

        if (pid == 0)
        {
            // Child.
            setuid(0);
            execle( "/bin/mount",

                // mount arguments
                "/bin/mount",
                ro ? "-r" : "-w",
                "-o",
                (opts ? opts : "user,exec,suid,nodev"),
                "-o",
                "sync",

                "--",
                dev,
                mpoint,
                NULL,

                env);
            // exec failed, if we are here.
            return EXIT_FAILURE;
        }
    }

    // Parent.
    waitpid(pid, &status, 0);
	return WEXITSTATUS(status);
}


int main(int argc, char** argv)
{
	if(argc < 3) {
        fprintf(stderr, "Error: please specify a valid BestCrypt virtual block device and a mount point.\n");
		return EXIT_FAILURE;
	}

	int ro = 0;
	int uid = getuid();
	int gid = getgid();
	int mode = 0700;
    const char* type = 0;
    const char* options = 0;

	char c;
    	while ((c = getopt(argc, argv, "rt:u:g:m:o:")) != -1) {
	switch (c) {
		case 'r': ro = 1; break;
		case 't': type = optarg; break;
		case 'u': uid = atoi(optarg); break;
		case 'g': gid = atoi(optarg); break;
		case 'm': mode = atoi(optarg); break;
		case 'o': options = optarg; break;
        default: fprintf(stderr, "Error: invalid option: -%c.\n", c); return EXIT_FAILURE;
	}
	}

	argv += optind;
	const char* blkdev = argv[0];
	const char* mpoint = argv[1];

	// verify that this is our block device.
	struct stat statbuf;
	if(0 != stat(blkdev, &statbuf)) {
        fprintf(stderr, "Error: stat failed on device %s: %s\n", blkdev, strerror(errno));
		return EXIT_FAILURE;
	}

	if(!S_ISBLK(statbuf.st_mode)) {
        fprintf(stderr, "Error: %s is not a block device.\n", blkdev);
		return EXIT_FAILURE;
	}

	if(bc_get_dev_major() != major(statbuf.st_rdev)) {
        fprintf(stderr, "Error: %s not a BestCrypt block device.\n", blkdev);
        return EXIT_FAILURE;
	}

	// call mount to do the job
	int rc = do_mount(blkdev, mpoint, type, ro, uid, gid, mode, options);
	if(rc != EXIT_SUCCESS) {
		return rc;
	}

	// change mount point ownership if requested
	if(!ro) {
		bc_get_root();

        int ch_result = chown(mpoint, uid, gid);
        if (ch_result)
        {
            fprintf(stderr, "Warning: ownership changing failed: %s\n", strerror(errno));
        }

        ch_result = chmod(mpoint, mode);
        if (ch_result)
        {
            fprintf(stderr, "Warning: mode changing failed: %s\n", strerror(errno));
        }

		bc_drop_root();
	}

	return EXIT_SUCCESS;		
}


