/*
 *	Copyright Jetico, Oy, 2011.
 *	
 *	Unmount helper tool.
 *	A suid tool that helps us unmount bestcrypt virtual devices and set mount point ownership.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "common.h"

static int do_umount(const char* path, int force)
{
	pid_t pid = fork();
	if(pid < 0) {
        fprintf(stderr, "Error: fork failed: %s\n", strerror(errno));
		return EXIT_FAILURE;
	}

	if( 0 == pid ) 
	{
		const char* env[] = {"PATH=/sbin:/usr/sbin:/bin:/usr/bin", NULL};
	
		// pass requested ownership options.
		// not all filesystems support this, so we can fail here.
		setuid(0);

		if(force) {
			execle(	"/bin/umount", "/bin/umount", "-l", path, NULL, env);
		} else {
			execle(	"/bin/umount", "/bin/umount", path, NULL, env);
		}

		// exec failed if we're here
		return EXIT_FAILURE;

	}

	// parent
	int status;
	waitpid(pid, &status, 0);
	return WEXITSTATUS(status);
}


int main(int argc, char** argv)
{
	if(argc < 2) {
        fprintf(stderr, "Error: please specify a valid mount point.\n");
		return EXIT_FAILURE;
	}

	int force = 0;
	char c;
    	while ((c = getopt(argc, argv, "f")) != -1) {
	switch (c) {
		case 'f': force = 1; break;
        default: fprintf(stderr, "Error: invalid option: -%c.\n", c); return EXIT_FAILURE;
	}
	}

	const char* mpoint = argv[optind];

	// check that this is our mount point.
	struct stat statbuf;
	if(0 != stat(mpoint, &statbuf)) {
        fprintf(stderr, "Error: stat failed: %s\n", strerror(errno));
		return EXIT_FAILURE;
	}

	if(!S_ISDIR(statbuf.st_mode)) {
        fprintf(stderr, "Error: %s is not a directory.\n", mpoint);
		return EXIT_FAILURE;
	}

	if(!is_bc_device(statbuf.st_dev)) {
        fprintf(stderr, "Error: %s is not a BestCrypt mount point.\n", mpoint);
		return EXIT_FAILURE;
	}	
	
	if(!is_owner(mpoint)) {
        fprintf(stderr, "Error: %s is not owned by the current user.\n", mpoint);
		return EXIT_FAILURE;
	}
	
	return do_umount(mpoint, force);
}
