/*
 * Copyright (c) 2012 Jetico, Inc., Finland
 * All rights reserved.
 */

#include "spacemap.h"


static void my_memset(void *buf, char c, bc_u64 size)
{
    char *b = (char*)buf;
    bc_u64 i = 0;

    for (; i < size; ++i)
    {
        b[i] = c;
    }
}


static void my_memcpy(void *to, void *from, bc_u64 count)
{
    char *f = (char*) from;
    char *t = (char*) to;
    bc_u64 i = 0;

    for (; i < count; i++)
    {
        t[i] = f[i];
    }
}


const char *SpaceMapGetErrorString(sm_result result)
{
    switch(result)
    {
        case kSpaceMap_Success:
            return "Success.";
        case kSpaceMap_InvalidArg:
            return "One or more arguments are invalid.";
        case kSpaceMap_NotFound:
            return "Element was not found.";
        case kSpaceMap_NoMem:
            return "Not enough memory to complete operation.";
        default:
            return "*** -= Undefined case came =- ***";
    }
}


SpaceMap *SpaceMapCreate(void)
{
    SpaceMap *map = (SpaceMap*)ALLOC(sizeof(SpaceMap));
    if (map)
    {
        my_memset((void*)map, 0, sizeof(SpaceMap));
    }

    // Else return NULL.
    return map;
}


SpaceMap *SpaceMapCreateCopy(SpaceMap *map)
{
    SpaceMap *copy = SpaceMapCreate();
    SMExtentNode *cur = NULL;
    sm_result res;

    if ((!copy) || (!map))
        return NULL;

    cur = map->start;
    while (cur)
    {
        res = SpaceMapAddCopy(copy, &cur->extent);
        if (kSpaceMap_Success != res)
        {
            SpaceMapFree(copy);
            return NULL;
        }
        cur = cur->next;
    }

    return copy;
}


SpaceMap *SpaceMapFromExtentArray(Extent *extents, bc_u64 count)
{
    SpaceMap *map = SpaceMapCreate();
    bc_u64 i = 0;
    sm_result result = kSpaceMap_Success;

    if (!map)
        return NULL;

    // Create an empty map from an empty array.
    if ((!extents) || (!count))
        return map;

    for (; i < count; ++i)
    {
        result = SpaceMapAddCopy(map, &extents[i]);
        if (kSpaceMap_Success != result)
        {
            SpaceMapFree(map);
            return (SpaceMap*)NULL;
        }
    }

    return map;
}


void SpaceMapFree(SpaceMap *map)
{
    SMExtentNode *list, *prev;
    if (map)
    {
        // Zero all data in map and free elements.
        list = map->start;
        prev = NULL;

        while (list)
        {
            my_memset(&list->extent, 0, sizeof(Extent));
            prev = list;
            list = list->next;
            my_memset(prev, 0, sizeof(SMExtentNode));
            FREE(prev);
        }

        // Set all fields to 0 to reduce errors.
        my_memset(map, 0, sizeof(SpaceMap));

        FREE(map);
    }
}


sm_result SpaceMapGetLength(SpaceMap *map, bc_u64 *out_length)
{
    SMExtentNode *cur;
    bc_u64 length;

    if ((!map) || (!out_length))
    {
        return kSpaceMap_InvalidArg;
    }

    if (!map->count)
    {
        (*out_length) = 0;
        return kSpaceMap_Success;
    }

    cur = map->start;
    length = 0;

    while (cur)
    {
        length += cur->extent.length;
        cur = cur->next;
    }

    (*out_length) = length;
    return kSpaceMap_Success;
}


sm_result SpaceMapGetExtentSpace(Extent *array, bc_u64 count, bc_u64 *out_length)
{
    bc_u64 i, space = 0;

    if ((!array) || (!count) || (!out_length))
        return kSpaceMap_InvalidArg;

    for (i = 0; i < count; i++)
    {
        space += array[i].length;
    }

    *(out_length) = space;

    return kSpaceMap_Success;
}


sm_result SpaceMapAddNode(SpaceMap *map, SMExtentNode *item)
{
    if (!map)
    {
        PRINT("Null map came.\n");
        return kSpaceMap_InvalidArg;
    }

    if (!item)
    {
        //fprintf(stderr, "Cannot add null extent.\n");
        PRINT("Tried to add a null item to the map.\n");
        return kSpaceMap_InvalidArg;
    }

    item->next = NULL;

    // If have something in map, add new item after last element.
    if (map->start)
    {
        map->end->next = item;
        map->end = item;
    }
    else // Have nothing, so item is our first and last element.
    {
        map->start = item;
        map->end = item;
    }

    map->count++;
    map->needsNormalization = 1;

    return kSpaceMap_Success;
}


sm_result SpaceMapAddCopy(SpaceMap *map, const Extent *extent)
{
    if (!extent)
    {
        //fprintf(stderr, "Cannot add a copy of null extent.\n");
        PRINT("Cannot add a copy of null extent.\n");
        return kSpaceMap_InvalidArg;
    }
    return SpaceMapAddNew(map, extent->start, extent->length);
}


sm_result SpaceMapAddNew(SpaceMap *map, bc_u64 start, bc_u64 length)
{
    SMExtentNode * item;
    if (!map)
        return kSpaceMap_InvalidArg;

    item = (SMExtentNode*) ALLOC(sizeof(SMExtentNode));

    if (!item)
    {
        PRINT("No memory.");
        return kSpaceMap_NoMem;
    }

    my_memset(item, 0, sizeof(SMExtentNode));

    item->extent.length = length;
    item->extent.start = start;

    SpaceMapAddNode(map, item);

    return kSpaceMap_Success;
}


sm_result SpaceMapRemove(SpaceMap *map, Extent *extent)
{
    SMExtentNode *cur, *prev;

    if ((!map) || (!extent))
        return kSpaceMap_InvalidArg;

    cur = map->start;
    prev = NULL;

    while (cur)
    {
        if (&cur->extent == extent)
            break;
        prev = cur;
        cur = cur->next;
    }

    // If found an extent to remove.
    if (&cur->extent == extent)
    {
        // If have previous item...
        if (prev)
        {
            // ...fix the map...
            if (map->end == cur)
                map->end = prev;

            // ...and remove cur.
            prev->next = cur->next;
        }

        // Check for first element.
        if (map->start == cur)
            map->start = cur->next;

        FREE(cur);
        map->count--;

        return kSpaceMap_Success;
    }

    return kSpaceMap_NotFound;
}


sm_result SpaceMapToBytes(SpaceMap *map, void *buffer, bc_u64 bufferSize)
{
    SMExtentNode *cur = NULL;
    char *buf = (char*)buffer;
    bc_u64 bufferPosition = 0;

    if ((!map) || (!buffer) || (!bufferSize))
    {
        PRINT("One of the parameters is empty but it mustn't.\n");
        return kSpaceMap_InvalidArg;
    }

    if (!map->start)
    {
        PRINT("Empty map came.\n");
        return kSpaceMap_InvalidArg;
    }

    cur = map->start;


    while (cur && (bufferPosition < bufferSize))
    {
        my_memcpy(&buf[bufferPosition], &cur->extent, sizeof(Extent));
        bufferPosition += sizeof(Extent);
        cur = cur->next;
    }

    if (cur) // Exited from loop by second condition.
    {
        PRINT("Not enough space in buffer to save map.\n");
        return kSpaceMap_NoMem;
    }

    return kSpaceMap_Success;
}


sm_result SpaceMapFromBytes(void *buffer, bc_u64 bufferSize, SpaceMap **out_map)
{
    SpaceMap *map = NULL;
    Extent *extents = (Extent*)buffer,
            *e;
    bc_u64 maxSize = bufferSize / sizeof(Extent);
    bc_u64 readCount = 0;
    sm_result result = kSpaceMap_Success;

    if ((!buffer) || (!bufferSize))
    {
        PRINT("Cannot load map from an empty buffer.\n");
        return kSpaceMap_InvalidArg;
    }

    if (!out_map)
    {
        PRINT("Empty map buffer.\n");
        return kSpaceMap_InvalidArg;
    }

    map = SpaceMapCreate();
    if (!map)
    {
        return kSpaceMap_NoMem;
    }

    while (readCount <= maxSize)
    {
        e = &extents[readCount++];
        if (e->length != 0)
        {
            result = SpaceMapAddCopy(map, e);
            if (kSpaceMap_Success != result)
                break;
        }
        else
            break;
    }

    if (kSpaceMap_Success == result)
        (*out_map) = map;
    else
        SpaceMapFree(map);

    return result;
}


sm_result SpaceMapToExtentArray(SpaceMap *map, Extent **out_array, bc_u64 *out_count)
{
    Extent *array = NULL;
    SMExtentNode *cur = NULL;
    bc_u64 i = 0;

    if ((!map) || (!out_array) || (!out_count))
        return kSpaceMap_InvalidArg;

    array = (Extent*)ALLOC(map->count * sizeof(Extent));
    if (!array)
        return kSpaceMap_NoMem;

    cur = map->start;

    for (; cur && (i < map->count); cur = cur->next, i++)
    {
        my_memcpy(&array[i], &cur->extent, sizeof(Extent));
    }

    *out_array = array;
    *out_count = map->count;

    return kSpaceMap_Success;
}


sm_result SpaceMapFreeArray(Extent **array)
{
    if (!array)
        return kSpaceMap_InvalidArg;

    if ((*array) != NULL)
        FREE(*array);

    // Set pointer to NULL to avoid errors.
    *array = 0;
    return kSpaceMap_Success;
}


// Removes next node. If node_before == NULL, removes first element from map.
static void RemoveNextNode(SpaceMap *map, SMExtentNode *node_before)
{
    SMExtentNode *node = NULL;

    if (!node_before)
        node = map->start;
    else
        node = node_before->next;

    if (!node)
    {
        PRINT("Trying to remove an unallocated node.");
        return;
    }

    // If we're dealing with first or last element, fix map pointers.
    if (map->start == node)
        map->start = node->next;

    if (map->end == node)
        map->end = node_before;

    // Move previous pointer.
    if (node_before)
        node_before->next = node->next;

    // And remove element;
    map->count--;
    FREE(node);
}


static void ExchangeNextNodes(SpaceMap *map, SMExtentNode *node_before1,
                              SMExtentNode *node_before2, SMExtentNode *node1,
                              SMExtentNode *node2)
{
    SMExtentNode *tmp;

    if (node1 == node2)
        return;

    // Change previous nodes.
    if (node_before1)
        node_before1->next = node2;
    if (node_before2)
        node_before2->next = node1;

    // Change pointers to next elements.
    tmp = node1->next;
    node1->next = node2->next;
    node2->next = tmp;

    // Fix the map.
    if (map->start == node1)
        map->start = node2;
    else if (map->start == node2)
        map->start = node1;

    if (map->end == node1)
        map->end = node2;
    else if (map->end == node2)
        map->end = node1;
}

static void InternalSort(SpaceMap *map, SMExtentNode **array)
{
    bc_u64 count = map->count, i, min_index, max_index, low, high;

    SMExtentNode *min, *max, *tmp;
    min = array[0];
    min_index = 0;
    max = array[count - 1];
    max_index = count -1;
    low = 0;
    high = count - 1;

    do
    {
        for (i = low; i <= high; i++)
        {
            if (ExtentIsEndBefore(&array[i]->extent, &min->extent))
            {
                min_index = i;
                min = array[i];
            }

            if (ExtentIsEndBefore(&max->extent, &array[i]->extent))
            {
                max_index = i;
                max = array[i];
            }
        }

        if (min_index != low)
        {
            ExchangeNextNodes(map,
                              (min_index > 0)?array[min_index - 1] : (SMExtentNode*)NULL, (low == 0)? (SMExtentNode*)NULL : array[low - 1],
                              min, array[low]);

            tmp = array[low];
            array[low] = array[min_index];
            array[min_index] = tmp;

            // Check that we didn't have our max moved.
            if (low == max_index)
                max_index = min_index;
        }

        if (max_index != high)
        {
            ExchangeNextNodes(map,
                              (max_index > 0)? array[max_index - 1] : (SMExtentNode*)NULL, array[high - 1], // high > 0.
                          max, array[high]);

            tmp = array[high];
            array[high] = array[max_index];
            array[max_index] = tmp;
        }

        low++;
        high--;
    }
    while (low < high);
}


sm_result SpaceMapSort(SpaceMap *map)
{
    SMExtentNode *cur,
            **array; // Will contain pointers to all map elements.
    bc_u64 i = 0;
    sm_result result;

    if (!map)
    {
        PRINT("Null map can't be sorted.");
        return kSpaceMap_InvalidArg;
    }

    if (!map->count)
        return kSpaceMap_Success;

    // Normalize map before sorting.
    if (map->needsNormalization)
    {
        result = SpaceMapNormalize(map);

        if (kSpaceMap_Success != result)
            return result;
    }

    array = (SMExtentNode**)ALLOC(map->count * sizeof(void*));

    if (!array)
    {
        PRINT("Not enough memory for operation.");
        return kSpaceMap_NoMem;
    }

    for (cur = map->start; cur; cur = cur->next)
        array[i++] = cur;

    InternalSort(map, array);

    FREE(array);

    return kSpaceMap_Success;
}


sm_result SpaceMapNormalize(SpaceMap *map)
{
    SMExtentNode    *cur, *another_cur, *another_cur_prev;
    Extent          *e, *e1;

    if (!map)
    {
        PRINT("Null map.");
        return kSpaceMap_InvalidArg;
    }

    cur = map->start;

    while (cur)
    {
        e = &cur->extent;
        another_cur = cur->next;
        another_cur_prev = cur;

        while (another_cur)
        {
            e1 = &another_cur->extent;

            // Remove zeroed extents.
            if (0 == e1->length)
            {
                RemoveNextNode(map, another_cur_prev);
                another_cur = cur->next;
                another_cur_prev = cur;
                continue;
            }

            if (ExtentUnion(e, e, e1)) // returns true, if extents had been intersecting and was performed a union.
            {
                RemoveNextNode(map, another_cur_prev);
                another_cur = cur->next; // It has been changed during element deletion.
                another_cur_prev = cur;
                continue;
            }

            another_cur_prev = another_cur;
            another_cur = another_cur->next;
        }
        cur = cur->next;
    }

    map->needsNormalization = 0;

    return kSpaceMap_Success;
}


void SpaceMapMoveExtents(SpaceMap *map, bc_u64 offset)
{
    SMExtentNode *cur;

    if (!map)
        return;

    cur = map->start;

    while (cur)
    {
        cur->extent.start += offset;

        cur = cur->next;
    }
}


sm_result SpaceMapExtentCount(SpaceMap *map, bc_u64 *out_count)
{
    if ((!map) || (!out_count))
        return kSpaceMap_InvalidArg;
    my_memcpy(out_count, &map->count, sizeof(bc_u64));
    return kSpaceMap_Success;
}


sm_result SpaceMapContainsExtent(SpaceMap *map, const Extent *extent, bool exact)
{
    SMExtentNode *cur;
    bool (*comparer)(const Extent *, const Extent *); // Do we need exact match or only containing is enough?

    if (!map)
        return kSpaceMap_InvalidArg;

    if (!extent)
        return kSpaceMap_InvalidArg;

    cur = map->start;

    if (exact)
        comparer = ExtentAreEqual;
    else
        comparer = ExtentContains;

    for (; cur; cur = cur->next)
    {
        if (comparer(&cur->extent, extent))
            return kSpaceMap_Success;
    }
    return kSpaceMap_NotFound;
}


sm_result SpaceMapUnion(SpaceMap *result_map, SpaceMap *another_map)
{
    SMExtentNode *cur = another_map->start;
    sm_result result = kSpaceMap_Success;

    for (; cur; cur = cur->next)
    {
        result = SpaceMapAddCopy(result_map, &cur->extent);
        if (kSpaceMap_Success != result)
            return result;
    }

    result = SpaceMapNormalize(result_map);
    return result;
}


/// Returns a value less than zero, if one more element has been added to the map or element has been deleted, and sm_result value otherwise.
static int SubstractExtents(SpaceMap *from_map, SMExtentNode *node_before_from, Extent *result, Extent *from, Extent *what)
{
    // Map should already contain `from` extent. If there is a second extent after substraction,
    // it will be added at the end of map.
    bc_u64 from_start = ExtentStart(from),
            from_end = ExtentEnd(from),
            what_start = ExtentStart(what),
            what_end = ExtentEnd(what),
            result_start, result_end;
    Extent one_more;
    int sm_res = kSpaceMap_Success;

    if (!ExtentAreIntersecting(from, what))
        return kSpaceMap_Success;

    // Exact match - remove extent from map.
    if (ExtentAreEqual(from, what))
    {
        RemoveNextNode(from_map, node_before_from);
        return -1; // Send signal for caller to see that map has been changed.
    }

    if (from_start <= what_start)
    {
        // 1)
        //from: ---====---------------
        //what: ----====--------------
        //res : ---=------------------
        // 2)
        //from: ---========-----------
        //what: -----====-------------
        //res : ---==----==-----------
        result_start = from_start;
        if (!ExtentContains(from, what)) // 1)
        {
            result_start = ExtentStart(from);
            result_end = ExtentStart(what) - 1;
        }
        else // 2)
        {
            one_more.start = what_start + 1;
            one_more.length = from_end /* end of `one_more` extent */ - one_more.start + 1;
            sm_res = SpaceMapAddCopy(from_map, &one_more);

            if (kSpaceMap_Success != sm_res)
            {
                PRINT("one_more extent copy adding failed.");
                return sm_res;
            }

            result_start = from_start;
            result_end = what_start - 1;
            sm_res = -1; // Send signal for caller to see that map has been changed.
        }
    }
    else // if from_start > what_start
    {
        // 1)
        //from: -----=====-------------
        //what: ----====---------------
        //res : --------==------------
        // 2)
        //from: -----=====-------------
        //what: ----=======------------
        //res : ----------------------

        if (!ExtentContains(what, from)) // 1)
        {
            result_start = what_end + 1;
            result_end = from_end;
        }
        else // 2)
        {
            result_start = 0;
            result_end = 0;
        }
    }

    result->start = result_start;
    result->length = result_end - result_start + 1;
    return sm_res;
}

sm_result SpaceMapSubstract(SpaceMap *from_map, SpaceMap *what_map)
{
    SMExtentNode *from_cur = NULL;
    SMExtentNode *what_cur = NULL;
    SMExtentNode *prev_from_cur = NULL;
    int result = kSpaceMap_Success;

    if (!from_map)
        return kSpaceMap_InvalidArg;
    if (!what_map)
        return kSpaceMap_Success;

    if (from_map->needsNormalization)
    {
        result = SpaceMapNormalize(from_map);
        if (kSpaceMap_Success != result)
            return (sm_result)result;
    }

    if (what_map->needsNormalization)
    {
        result = SpaceMapNormalize(what_map);
        if (kSpaceMap_Success != result)
            return (sm_result)result;
    }

    from_cur = from_map->start;

    // Do substraction for each pair of extents in map.
    while (from_cur)
    {
        what_cur = what_map->start;
        while (what_cur)
        {
            if (ExtentAreIntersecting(&from_cur->extent, &what_cur->extent))
            {
                result = SubstractExtents(from_map, prev_from_cur,
                                          &from_cur->extent, &from_cur->extent, &what_cur->extent);
                if ((kSpaceMap_Success != result) && (-1 != result))
                    return (sm_result)result;
            }

            if (-1 != result)
                what_cur = what_cur->next;
            else // Map has been changed.
            {
                // Fix result to a vaild sm_result value.
                result = kSpaceMap_Success;

                // If have something in the map, to simplify algorithm do start from the beginning.
                if (from_map->start)
                {
                    from_cur = from_map->start;
                    what_cur = what_map->start;
                    continue;
                }
                else
                {
                    // Exiting from loops.
                    goto SPACEMAP_SUBSTRACT_END;
                }
            }
        }

        prev_from_cur = from_cur;
        from_cur = from_cur->next;
    }
SPACEMAP_SUBSTRACT_END:

    return (sm_result)result;
}


sm_result SpaceMapGetMapOfSize(SpaceMap *map, bc_u64 size, SpaceMap **out_mapOfSize)
{
    SpaceMap *resultMap = 0;
    bc_u64 sizeFound = 0;
    SMExtentNode *cur = 0;
    Extent part;
    sm_result result = kSpaceMap_Success;

    if (!map || !out_mapOfSize)
        return kSpaceMap_InvalidArg;

    resultMap = SpaceMapCreate();
    if (!resultMap)
        return kSpaceMap_NoMem;

    // If one wants an empty map, return it.
    if (!size)
    {
        (*out_mapOfSize) = resultMap;
        return kSpaceMap_Success;
    }

    cur = map->start;
    // Add extents while their total size is less than desired size.
    for (; cur != NULL; cur = cur->next)
    {
        if ((sizeFound + cur->extent.length) <= size)
        {
            // Add full extent.
            result = SpaceMapAddCopy(resultMap, &cur->extent);
            if (kSpaceMap_Success != result)
            {
                SpaceMapFree(resultMap);
                return result;
            }

            sizeFound += cur->extent.length;
        }
        else
        {
            // If not completed, add the part of the extent.
            if (sizeFound < size)
            {
                part.start = cur->extent.start;
                part.length = size - sizeFound;
                result = SpaceMapAddCopy(resultMap, &part);
                if (kSpaceMap_Success != result)
                {
                    SpaceMapFree(resultMap);
                    return result;
                }
                sizeFound += part.length;
            }
            break;
        }
    }

    if (sizeFound == size)
    {
        (*out_mapOfSize) = resultMap;
        return kSpaceMap_Success;
    }

    // Have not enough space in map.
    SpaceMapFree(resultMap);
    return kSpaceMap_NotFound;
}


sm_result SpaceMapGetPhysicalExtentsFromMap(SpaceMap *map, bc_u64 io_offset, bc_u64 io_count,
                                            SpaceMap **out_physicalMap)
{
    Extent *extents = 0, *physicalExtents = 0;
    bc_u64 extentCount = 0, physicalCount = 0;
    sm_result result = kSpaceMap_Success;
    SpaceMap *resultMap = 0;

    if (!out_physicalMap)
        return kSpaceMap_InvalidArg;

    result = SpaceMapToExtentArray(map, &extents, &extentCount);
    if (kSpaceMap_Success != result)
        return result;

    result = SpaceMapGetPhysicalExtents(extents, extentCount, io_offset, io_count,
                                        &physicalExtents, &physicalCount);

    if (kSpaceMap_Success != result)
    {
        goto PHYSICAL_FROM_MAP_OUT;
    }

    resultMap = SpaceMapFromExtentArray(physicalExtents, physicalCount);
    if (!resultMap)
    {
        result = kSpaceMap_NoMem;
        goto PHYSICAL_FROM_MAP_OUT;
    }

    (*out_physicalMap) = resultMap;

PHYSICAL_FROM_MAP_OUT:
    (void)SpaceMapFreeArray(&extents);
    (void)SpaceMapFreeArray(&physicalExtents);
    if (kSpaceMap_Success != result)
        SpaceMapFree(map);
    return result;
}


sm_result SpaceMapGetPhysicalExtents(Extent *virtual_extents, bc_u64 virtual_count, bc_u64 io_offset,
                                     bc_u64 io_count, Extent **out_physical_extents, bc_u64 *out_extent_count)
{
    Extent *phys;
    bc_u64 skipped_bytes, // Sum of skipped bytes before current extent.
            bytes_read, vcount, pcount;

    // * Search for extent containing io_offset.
    skipped_bytes = 0;

    for (vcount = 0; vcount < virtual_count; vcount++)
    {
        if ((skipped_bytes + virtual_extents[vcount].length) > io_offset)
        {
            break;
        }
        skipped_bytes += virtual_extents[vcount].length;
    }

    if (vcount == virtual_count)
    {
        PRINT("Start is not found.\n");
        return kSpaceMap_NotFound;
    }

    // Check that extent found is containing a starting point.
    if (!ExtentContainsPoint(&virtual_extents[vcount], virtual_extents[vcount].start + (io_offset - skipped_bytes)))
    {
        PRINT("The extent found doesn't contain the offset.\n");
        return kSpaceMap_NotFound;
    }

    // * Allocate extent array. Maximum number of extents is the map size.
    phys = (Extent*)ALLOC(virtual_count * sizeof(Extent));
    if (!phys)
        return kSpaceMap_NoMem;

    // * Create first extent. skipped_bytes - num of bytes before this extent,
    // so we must skip (io_offset - skipped_bytes) more.
    pcount = 1;
    phys[0].start = virtual_extents[vcount].start + (io_offset - skipped_bytes);

    // If all IO data is inside found extent, finish reading.
    if ((virtual_extents[vcount].length - (io_offset - skipped_bytes)) >= io_count)
        phys[0].length = io_count;
    else // Need to read more extents.
    {
        phys[0].length = virtual_extents[vcount].length - (io_offset - skipped_bytes);

        // * Read all the extents containing io_count bytes.
        // Already read first extent.
        bytes_read = phys[0].length;

        vcount++;

        for (; (bytes_read < io_count) && (vcount < virtual_count); vcount++)
        {
            phys[pcount].start = virtual_extents[vcount].start;

            if ((virtual_extents[vcount].length + bytes_read) < io_count)
            {
                // This extent have not enough bytes, so get it and go on.
                phys[pcount].length = virtual_extents[vcount].length;
                bytes_read += phys[pcount++].length;
            }
            else
            {
                // This extent is the last, write it into array and exit the loop.
                phys[pcount].length = io_count - bytes_read;
                bytes_read += phys[pcount++].length;
                break;
            }
        }

        // * Return result.
        if (bytes_read != io_count)
        {
            PRINT("Have not enough bytes to read in map.\n");
            PRINT("Read %lu bytes, physical extents count is %lu.\n", (long unsigned)bytes_read, (long unsigned)pcount);
            FREE(phys);
            return kSpaceMap_NotFound;
        }
    }
    (*out_physical_extents) = phys;
    (*out_extent_count) = pcount;

    return kSpaceMap_Success;
}


void SpaceMapFreePhysicalExtents(Extent **physical_extents)
{
    FREE(*physical_extents);
    (*physical_extents) = NULL;
}


void SpaceMapPrint(SpaceMap *map)
{
    SMExtentNode *cur;

    if (!map)
    {
        PRINT("Printed map is empty.\n");
        return;
    }

    if (!map->start || !map->count)
    {
        PRINT("Printed map is empty.\n");
        return;
    }

    cur = map->start;
    PRINT("Printing map with %lu extents:\n", (long unsigned)map->count);
    while (cur)
    {
        PRINT("  Start: %lu, length: %lu\n", (long unsigned)cur->extent.start,
              (long unsigned)cur->extent.length);
        cur = cur->next;
    }
}
