#include "bctool.h"
#include "fslib/fslib.h"
#include "spacemap/spacemap.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>

// Inserts all fs-lib extents to specified space map.
static bool AddExtentsToMap(SpaceMap *map, fs_extent_t *extents)
{
    fs_extent_t *cur = extents;
    sm_result smres = kSpaceMap_Success;

    while (cur)
    {
        smres = SpaceMapAddNew(map, cur->start, cur->size);
        if (kSpaceMap_Success != smres)
            return false;
        cur = cur->next;
    }

    return true;
}

SpaceMap *BCTool::TransformToSubmapOfSize(SpaceMap *map, bc_u64 size)
{
    if (size != 0)
    {
        // Check that the spacemap has enough space.
        bc_u64 totalFree = 0;
        sm_result result = SpaceMapGetLength(map, &totalFree);

        if (result != kSpaceMap_Success)
        {
            this->ShowError("Error while getting free space length: %s\n",
                            SpaceMapGetErrorString(result));
            return NULL;
        }

        if (totalFree < size)
        {
            this->ShowError("Free space size (%lu) is less than specified hidden part size (%lu).\n",
                            (unsigned long)totalFree, (unsigned long)size);
            return NULL;
        }

        // Getting submap of desired size.
        SpaceMap *mapOfSize = 0;
        result = SpaceMapGetMapOfSize(map, size, &mapOfSize);

        if (kSpaceMap_Success != result)
        {
            this->ShowError("Error getting map with size %lu from the free space map with size %lu: %s\n",
                            (unsigned long)size, (unsigned long)totalFree, SpaceMapGetErrorString(result));
            return NULL;
        }

        return mapOfSize;
    }

    return NULL;
}

SpaceMap *BCTool::GetFreeSpaceFromMainPart(const char *deviceName)
{
    SpaceMap *map = (SpaceMap*)0;
    fs_extent_t *freeSpace = (fs_extent_t*)0;
    fs_volume_t vol;

    int error = -1;

    char deviceToProcess[BC_MAX_PATH] = {0};

    // Search the first partition, if any. If not, use whole device.
    if (GetFirstDevicePartition(deviceName, deviceToProcess, BC_MAX_PATH))
    {
        this->ShowError("Error getting device name: %s.\n", strerror(errno));
        return NULL;
    }

    // Try to parse volume.
    error = fslib_open_volume(deviceToProcess, O_RDWR, &vol);
    if (error)
    {
        this->ShowError("Couldn't parse container volume: %s. Possible reason is that the filesystem isn't supported.\n",
                        strerror(error));
        return NULL;
    }

    if (!vol.get_free_space)
    {
        this->ShowError("Getting of free space is not supported for this file system.\n");
        fslib_close_volume(&vol);
        return NULL;
    }

    // Get free space from volume.
    freeSpace = vol.get_free_space(&vol);

    fslib_close_volume(&vol);

    if (!freeSpace)
    {
        this->ShowError("Couldn't retrieve free space information from the container device.\n");
        return NULL;
    }

    // Form space map.
    map = SpaceMapCreate();
    if (!map)
    {
        this->ShowError("Not enough memory to create an empty space map.\n");
        fs_put_extent_list(freeSpace);
        return NULL;
    }

    if (!AddExtentsToMap(map, freeSpace))
    {
        this->ShowError("Error while creating empty space map.");
        fs_put_extent_list(freeSpace);
        return NULL;
    }

    fs_put_extent_list(freeSpace);
    freeSpace = NULL;
    return map;
}

bool BCTool::CreateHiddenPart(const char *fileName, bc_u64 size, const char *fsType)
{
    BCContainerResult containerResult;
    sm_result spaceMapResult;
    bool result = true;
    BCVolume volume;
    BCPassword mainPassword;
    BCPassword hiddenPassword;
    int formatResult;
    char deviceName[BC_MAX_PATH] = { 0 };
    char tmpFsType[BC_MAX_PATH] = { 0 };

    // Link the main part of container.
    if (!this->Link(fileName, true, (BCMountParams*)NULL, true, &volume,
                    &mainPassword, deviceName))
        return false; // All errors are already displayed inside.

    // Get free space map.
    SpaceMap *map = this->GetFreeSpaceFromMainPart(deviceName);

    // Do not check return value, all errors are already displayed inside.
    this->Unlink(deviceName, false);

    bc_u64 sizeAvailable;
    if (kSpaceMap_Success != (spaceMapResult = SpaceMapGetLength(map, &sizeAvailable)))
    {
        this->ShowError("Error while getting the free space size.\n");
        return false;
    }

    if (!map)
    {
        // Errors are already displayed.
        return false;
    }

    // Move all extents to their physical places in the container file.
    SpaceMapMoveExtents(map, volume.info.offset);

    // Ask a new password, create and store hidden key.
    containerResult = BCCreateHiddenPart(fileName, size, map, &mainPassword,
                              this->ReportProgress, NULL, &hiddenPassword, tmpFsType);

    // For GUI, here will be determined a filesystem type.
    if (strlen(tmpFsType))
    {
        fsType = tmpFsType;
    }

    if (kBCContainer_Success != containerResult)
    {
        if (kBCContainer_Cancelled == containerResult) // User cancelled an operation.
            return false;
        this->ShowError("Error while trying to create hidden part: %s\n",
                        BCContainerErrorDecription(containerResult));
        result = false;
        goto CREATE_HIDDEN_OUT;
    }

    // Format hidden volume.
    memset(deviceName, 0, BC_MAX_PATH);
    if (fsType && strlen(fsType))
    {
        this->ShowDebug("Starting hidden part formatting.\n");
        // Decode hidden volume information by the password.
        BCVolume hiddenVolume;
        containerResult = BCDecodeVolumeByPassword(fileName, &hiddenPassword, &hiddenVolume);
        if (kBCContainer_Success != containerResult)
        {
            this->ShowError("Error while decoding written hidden part: %s\n",
                            BCContainerErrorDecription(containerResult));
            result = false;
            goto CREATE_HIDDEN_OUT;
        }

        this->ShowDebug("Volume is decoded. Trying to link.\n");

        // Link container file by the decoded information.
        result = this->DoLink(fileName, &hiddenVolume, deviceName, false);
        BCPutVolume(&hiddenVolume);

        if (!result)
        {
            this->ShowWarning("Failed to format the created hidden part.\n");
            goto CREATE_HIDDEN_OUT;
        }

        this->ShowDebug("Link successful. Trying to format.\n");

        formatResult = this->FormatVolume(deviceName, fsType);
        this->ShowDebug("Format completed. Trying to unlink.\n");
        this->Unlink(deviceName, false);

        if (formatResult)
        {
            this->ShowWarning("mkfs returned error %i. Hidden part may remain unformatted. Please check the log.\n", formatResult);
            result = false;
            goto CREATE_HIDDEN_OUT;
        }

        this->ShowDebug("Unlink successful.");
    }

CREATE_HIDDEN_OUT:
    if (map)
    {
        SpaceMapFree(map);
    }
    return result;
}
