#include "bctool.h"

#include <unistd.h>
#include <sys/types.h>
#include <glob.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>

/// @brief If mounted container has partitions, gets path to device representing first device partition,
///        otherwise gets main device name.
/// @param maxNameSize Maximum length of both mainDeviceName and out_mountDeviceName strings.
int BCTool::GetFirstDevicePartition(const char *mainDeviceName, char *out_partitionDeviceName, int maxNameSize)
{
    assert(out_partitionDeviceName);

    char searchPattern[BC_MAX_PATH];
    // Create search pattern.
    snprintf(searchPattern, maxNameSize, "%s?", mainDeviceName);

    glob_t searchResults;
    int globResult = glob(searchPattern, 0,  NULL, &searchResults);
    int result = 0;

    do
    {
        // If nothing was found, use main device name.
        if (globResult)
        {
            if (globResult == GLOB_NOMATCH)
            {
                strncpy(out_partitionDeviceName, mainDeviceName, maxNameSize);
                break;
            }

            // Errno will be set in this case.
            result = -1;
            break;
        }

        assert(searchResults.gl_pathc);

        // Glob results are sorted, so simply get the first result.
        strncpy(out_partitionDeviceName, searchResults.gl_pathv[0], maxNameSize);
    }
    while (0);

    globfree(&searchResults);
    return result;
}

bool BCTool::Mount(const char *file, BCMountParams *mountParams, const char *fsType,
                     int uid, int gid, const char *mountOptions, char *out_deviceName)
{
    // Check mount point.
    if (!PathExists(mountParams->mountPoint) || !IsDirectory(mountParams->mountPoint))
    {
        this->ShowError("Mount point %s does not exist or is not a directory.\n", mountParams->mountPoint);
        return false;
    }

    if (geteuid() /* root owns everything */ && !IsOwner(mountParams->mountPoint))
    {
        this->ShowError("You are not the owner of %s.\n", mountParams->mountPoint);
        return false;
    }

    // Link container file and get main device name.
    bool result = this->Link(file, mountParams->mountReadOnly, mountParams,
                             false, NULL, NULL, out_deviceName);
    if (!result)
        return false;
    do
    {
        // Get the name of device representing first partition.
        char partitionDeviceName[BC_MAX_PATH];
        int error = GetFirstDevicePartition(out_deviceName, partitionDeviceName, BC_MAX_PATH);
        if (error)
        {
            this->ShowError("Error while getting device for mounting: %s", strerror(errno));
            result = false;
            break;
        }

        // Run our suid helper to mount device.
        const char *args[0xff] = { NULL };
        const char** parg = args;
        *parg++ = "/sbin/bcmount_helper";
        *parg++ = partitionDeviceName;
        *parg++ = mountParams->mountPoint;

        if (mountParams->mountReadOnly)
        {
            *parg++ = "-r";
        }

        if (fsType)
        {
            *parg++ = "-t";
            *parg++ = fsType;
        }

        if (mountOptions)
        {
            *parg++ = "-o";
            *parg++ = mountOptions;
        }

        char uidString[10], gidString[10];

        if (uid || gid)
        {
            if (geteuid())
            {
                this->ShowError("Only root can specify mount point's owner.\n");
                result = false;
                break;
            }

            if (uid)
            {
                snprintf(uidString, 10, "%i", uid);
                *parg++ = "-u";
                *parg++ = uidString;
            }

            if (gid)
            {
                snprintf(gidString, 10, "%i", gid);
                *parg++ = "-g";
                *parg++ = gidString;
            }
        }

        error = RunProcessV("/sbin/bcmount_helper", args);

        if (error)
        {
            result = false;
            if (32 == error)
                this->ShowError("File system is not recognized.\n");
            else
                this->ShowError("Mount failed with result %d.", error);
            break;
        }
    }
    while (0);

    // Do device unlink in case of a error.
    if (!result)
    {
        this->Unlink(out_deviceName, false);
    }

    return result;
}
