#include "bctool.h"

#include <signal.h>
#include <limits.h>
#include <unistd.h>

#include "container/file.h"
#include "keygens/kgghost/kgghost.h"

#define ROUND_UP(value, blocksize) ( ((value) + (blocksize) - 1) & ~((blocksize) - 1) )

bool BCTool::New(const char *fileName, const char *algName, const char *modeName, bc_u64 size,
                   const char *description, const char *fsType)
{
    // Check file.
    if (PathExists(fileName))
    {
        if (IsBlockSpecial(fileName))
        {
            if (!this->AskUser("Specified file is block device. All data on it will be lost. Are you sure you want it? (y/[n]):", false))
            {
                return false;
            }

            if (size != 0)
            {
                this->ShowWarning("File %s is a block device, specified size will be ignored.", fileName);
            }

            size = BCFileSize(fileName);
            if (!size)
            {
                this->ShowError("Error getting size of the block device.\n");
                return false;
            }
            // Reduce size by the size of kgghost header,
            // as it has the biggest size of existent keygens.
            size -= kBCV8_MaximumDataBlockSize;
        }
        else //if (IsFile(fileName))
        {
            this->ShowError("File %s already exists and it is not a block device.\n", fileName);
            return false;
        }
//        else
//        {
//            this->ShowError("Invalid file type.\n");
//            return false;
//        }
    }

    if (!algName)
    {
        this->ShowError("Please specify algorithm name.\n");
        return false;
    }

    if (!modeName)
    {
        this->ShowWarning("Encryption mode isn't set and will default to CBC.");
        modeName = "CBC";
    }

    if (size == 0)
    {
        this->ShowError("Please specify container size.\n");
        return false;
    }
    else if (size % 512) // Check that size is a multiple to block size.
    {
        ROUND_UP(size, 512);
        this->ShowWarning("Container size is not even to block size, will round up to %llu bytes.\n", (long long unsigned)size);
    }

    // Get specified algorithm and mode.
    BCAlgInfo alg;
    if (kBCKernel_Success != BCKernelLookupAlgName(algName, &alg))
    {
        this->ShowError("Unknown algorithm: %s.\n", algName);
        return false;
    }

    // Use CBC mode by default.
    if (!modeName)
        modeName = "CBC";

    bc_mode_id modeId = BCKernelModeId(modeName);

    if (!BCKernelIsModeSupported(&alg, modeId))
    {
        this->ShowError("Algorithm %s does not support mode %s.\n", algName, modeName);
        return false;
    }

    // Install interrupt handlers and create container.
    // TODO: review and change to sigaction().
    sighandler_t intHandler = signal(SIGINT, this->HandleSignal);
    sighandler_t quitHandler = signal(SIGQUIT, this->HandleSignal);
    sighandler_t termHandler = signal(SIGTERM, this->HandleSignal);
    sighandler_t abortHandler = signal(SIGABRT, this->HandleSignal);

    bool result;

    BCVolume vol;
    int error = BCCreateContainer(fileName, alg.id, modeId, size, description,
                                  this->ReportProgress, NULL, &vol);

    if (((BCContainerResult)error) != kBCContainer_Success)
    {
        if (((BCContainerResult)error) == kBCContainer_IOError)
        {
            this->ShowError("Error while creating container. Possible reason: not enough space.\n");
        }
        else
        {
            this->ShowError("Error while creating container: %s\n", BCContainerErrorDecription((BCContainerResult)error));
        }

        result = false;
        goto BCTOOL_NEW_RETURN;
    }

    // Format new volume if user requested.
    if (fsType && strlen(fsType))
    {
        char deviceName[PATH_MAX] = { 0 };
        error = BCKernelSetupDisk(fileName, vol.info.alg, vol.info.mode, vol.key_handle, vol.pool,
                                  vol.info.map, kBCDisk_Default, deviceName);
        if (kBCKernel_Success != error)
        {
            BCPutVolume(&vol);
            this->ShowError("Could not create device with a new container: %s\n", BCKernelGetErrorString(error));
            result = false;
            goto BCTOOL_NEW_RETURN;
        }

        error = this->FormatVolume(deviceName, fsType);
        if (error)
        {
            BCPutVolume(&vol);
            this->ShowError("Could not format volume, mkfs returned %d.\n", error);
            result = false;
            goto BCTOOL_NEW_RETURN;
        }
        sync();
        error = BCKernelRemoveDisk(deviceName, 0);
        if (error)
        {
            // Show error but continue, we have already created the container.
            this->ShowError("Could not remove block device: %s. Container remains linked to device.\n",
                            BCKernelGetErrorString(error));
        }
    }
    else
    {
        this->ShowWarning("Please do not forget to format the container.\n");
    }

    BCPutVolume(&vol);

    result = true;

BCTOOL_NEW_RETURN:

    // Restore signal handlers before return.
    signal(SIGINT, intHandler);
    signal(SIGQUIT, quitHandler);
    signal(SIGTERM, termHandler);
    signal(SIGABRT, abortHandler);

    return result;
}
