#include "bctool.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

bool BCTool::RawLink(const char *file, const char *algName, const char *modeName, bool readOnly)
{
    // Check file.
    if (!BCTool::WantRead(file))
        return false;

    // Check write permissions if requested.
    if (!readOnly)
    {
        int error = access(file, W_OK);
        if (error || (errno == EROFS))
        {
            this->ShowWarning("You do not have write permissions for file %s. Will link for read only.\n", file);
            readOnly = true;
        }
    }

    char messageTemplate[] = "Attention! After this operation all current contents of '%s' will be destroyed! Are you sure you want to continue?";
    int fullMessageLength = sizeof(*messageTemplate) + strlen(file) * sizeof(char);
    char * fullMessage = new char[fullMessageLength];
    snprintf(fullMessage, fullMessageLength, messageTemplate, file);

    // bool AskUser(const char *message, bool defaultAnswer)
    bool needContinue = this->AskUser(fullMessage, false);

    delete fullMessage;
    fullMessage = (char*)NULL;
    if (!needContinue)
    {
        // User changed his mind.
        return false;
    }

    // Get specified algorithm and mode.
    BCAlgInfo alg;
    if (kBCKernel_Success != BCKernelLookupAlgName(algName, &alg))
    {
        this->ShowError("Unknown algorithm %s.", algName);
        return false;
    }

    if (!modeName)
        modeName = "CBC";

    bc_mode_id mode = BCKernelModeId(modeName);

    size_t keysize = BCKernelGetModeKeySize(&alg, mode);
    if(0 == keysize)
    {
        this->ShowError("Algorithm %s does not support mode %s.", algName, modeName);
        return false;
    }

    // Generate random key and pool.
    char* key = (char*) malloc(keysize + 512);
    if(!key)
    {
        this->ShowError("No memory to allocate random key buffer.");
        return false;
    }

    if(keysize != BCUserGenerateSeed("Generating encryption key.", key, keysize,
                                     this->ReportProgress, NULL))
    {
        free(key);
        return false;
    }
    char pool[512] = {0};
    if(kBCMode_CBC == mode)
    {
        if(sizeof(pool) != BCUserGenerateSeed("Generating pool.", pool, sizeof(pool), this->ReportProgress, NULL))
        {
            this->ShowError("Failed seed generation.");
            free(key);
            return false;
        }
    }
    int keyHandle;
    int result = 0;

    if (kBCKernel_Success != (result = BCKernelStoreKey(key, keysize, &keyHandle)))
    {
        free(key);
        this->ShowError("Could not store key: %s.", BCKernelGetErrorString(result));
        return false;
    }
    memset(key, 0, keysize);
    free(key);

    // link the file
    struct stat st;
    if (0 != stat(file, &st))
    {
        BCKernelFreeKey(keyHandle);
        this->ShowError("stat command failed for %s.", file);
        return false;
    }

    char deviceName[BC_MAX_PATH] = { 0 };
    int options = readOnly ? kBCDisk_ReadOnly : 0;
    SpaceMap *map = SpaceMapCreate();
    sm_result smres;
    if (!map)
    {
        this->ShowError("Not enough memory to allocate space map.");
        return false;
    }
    smres = SpaceMapAddNew(map, 0, st.st_size);
    if (kSpaceMap_Success != smres)
    {
        this->ShowError("Cannot add an extent to the space map: %s.", SpaceMapGetErrorString(smres));
    }
    if(kBCKernel_Success != BCKernelSetupDisk(file, alg.id, mode, keyHandle, pool, map, options, deviceName))
    {
        BCKernelFreeKey(keyHandle);
        this->ShowError("Could not link file %s.", file);
        return false;
    }

    BCKernelFreeKey(keyHandle);

    this->ShowInfo("File successfully linked to %s.", deviceName);

    return true;
}
