#include "bctool.h"

#include "disk/disk.h"

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <sys/stat.h>
#include <mntent.h>

int BCTool::DoUmount(const char* mountPoint, const char* deviceName, bool force)
{
    // Call our suid helper to do the actual umount.
    int error;
    if (force)
    {
        error = this->RunProcessL("/sbin/bcumount_helper", mountPoint, "-f", ARG_TERM);
    }
    else
    {
        error = this->RunProcessL("/sbin/bcumount_helper", mountPoint, ARG_TERM);
    }

    if (error && !force)
    {
        return error;
    }

    // All errors about device unlinking are shown inside Unlink.
    bool result = this->Unlink(deviceName, force);

    // Set error, if it wasn't been set before and must be set now.
    if ((!result) && (!error))
        error = -1;
    return error;
}

bool BCTool::Umount(const char *mountPoint, bool force)
{
    // Check that mount point is a vaild directory.
    if (!PathExists(mountPoint) || !IsDirectory(mountPoint))
    {
        this->ShowError("Mount point %s does not exist or is not a directory.\n", mountPoint);
        return false;
    }

    if (getuid() && !IsOwner(mountPoint))
    {
        this->ShowError("You are not the owner of %s.\n", mountPoint);
        return false;
    }

    if (!IsBCMount(mountPoint))
    {
        this->ShowError("%s is not a BestCrypt mount point.\n", mountPoint);
        return false;
    }

    // Find device mounted here.
    struct stat st;
    if(0 != stat(mountPoint, &st))
    {
        this->ShowError("statfs failed: %s.\n", strerror(errno));
        return false;
    }

    BCDiskRef disk;
    if(!BCDiskRef::findDiskWithMinor(minor(st.st_dev), &disk))
    {
        this->ShowError("Could not find device mounted at %s.\n", mountPoint);
        return false;
    }

    const char* device = disk.nodeName();

    sync();

    int error = DoUmount(mountPoint, device, force);
    if (error)
    {
        this->ShowError("Umount failed with result %d. Please check that you have no files or folders opened from this device and try again.\n", error);
        return false;
    }
    else
    {
        return true;
    }
}

bool BCTool::UmountAll(bool force, int *out_mountedCount, int *out_unmountedCount)
{
    // Flush file system changes, as we can do forced unmount later.
    sync();
    unsigned umounted = 0;
    unsigned total = 0;

    std::vector<bc_mount_entry> info;
    bool result = this->GetMountInfo(&info);
    if (!result)
    {
        // TODO: That's not good...
        return false;
    }

    total = info.size();
    std::vector<bc_mount_entry>::iterator i = info.begin();

    // Try to umount/unlink as much as possible.
    for (; i != info.end(); i++)
    {
        bc_mount_entry entry = *i;
        if (strlen(entry.mountPoint))
        {
            // Container is mounted.
            result &= this->Umount(entry.mountPoint, force);
        }
        else
        {
            result &= this->Unlink(entry.deviceName, force);
        }

        if (result)
            ++umounted;
    }

    if (out_mountedCount)
        (*out_mountedCount) = total;
    if (out_unmountedCount)
        (*out_unmountedCount) = umounted;

    return result;
}
