#include "bctool.h"

#include <cstdlib>

#include <assert.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <pwd.h>
#include <cstdio>
#include <stdarg.h>
#include <limits.h>
#include <grp.h>
#include <mntent.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <linux/fs.h>
#include <vector>
#include <string>

#include "container/volume.h"
#include "container/container.h"
#include "container/file.h"
#include "kernel/BCKernel.h"
#include "disk/disk.h"
#include "container/isaac.h"

bool BCTool::WantChange(const char *file)
{
    if(!BCTool::PathExists(file)) {
        this->ShowError("File %s does not exist\n", file);
        return false;
    }

    if(!BCTool::CanWrite(file)) {
        this->ShowError("You do not have write permissions to file %s\n", file);
        return false;
    }

    return true;
}

bool BCTool::WantRead(const char *file)
{
    if(!BCTool::PathExists(file)) {
        this->ShowError("File %s does not exist\n", file);
        return false;
    }

    if(!BCTool::CanRead(file)) {
        this->ShowError("You do not have read permissions to file %s\n", file);
        return false;
    }

    return true;
}

bool BCTool::ProcessHeader(const char *container, bool encrypt)
{
    if (!this->WantChange(container))
        return false;

    int error;
    if(encrypt) {
        error = BCEncryptHeader(container);
    } else {
        error = BCDecryptHeader(container);
    }

    if(error && (kBCContainer_Cancelled != error)) {
        this->ShowError("Failed to %s the header.\n", (encrypt ? "encrypt" : "decrypt"));
        return false;
    }
    else if (!error)
    {
        this->ShowInfo("The header was successfully %sed.", (encrypt ? "encrypt":"decrypt"));
    }

    return true;
}

bool BCTool::DeleteContainer(const char *container)
{
    if (!this->AskUser("After this operation container data will be unrecoverable. Are you sure you want to continue?", false))
        return false;

    if (!this->WantChange(container))
        return false;

    bc_file_handle handle;
    bc_file_result fileResult = BCFileOpen(container, kBCFileMode_ReadWrite, &handle);
    if (kBCFileResult_Success != fileResult)
    {
        this->ShowError("Error opening file %s: %s", container, strerror(errno));
        return false;
    }

    bc_u64 fileSize = BCFileSize(container);
    bc_u64 processSize = fileSize > 4096 ? 4096 : fileSize;

    char buf[1024] = { 0 };
    bc_u64 i = 0;
    bc_u64 bytesWritten = 0;
    bc_u64 bytesTotal = 0;
    unsigned t = (unsigned)time(NULL);
    const void *seed = (const void*)&t;
    unsigned int seedSize = sizeof(unsigned);
    void *isaacContext = IsaacInit(seed, seedSize);

    for (; i < processSize; i += bytesWritten)
    {
        IsaacGenerate(isaacContext, buf, sizeof(buf));
        bytesWritten = BCFileWrite(&handle, buf, sizeof(buf));
        bytesTotal += bytesWritten;

        if ((sizeof(buf) != bytesWritten) && (bytesTotal < fileSize))
        {
            this->ShowError("Error writing to a container file %s: %s", container, strerror(errno));
            BCFileClose(&handle);
            return false;
        }
    }

    IsaacRelease(isaacContext);
    BCFileClose(&handle);
    BCFileDelete(container);
    return true;
}

bool BCTool::BackupHeader(const char *file, const char *backup)
{
    if (!this->WantRead(file))
        return false;

    if(BCTool::PathExists(backup)) {
        this->ShowError("File %s exists\n", backup);
        return false;
    }

    int error = BCBackupHeader(file, backup);
    if(error) {
        this->ShowError("Failed creating header backup\n");
        return false;
    }

    return true;
}

bool BCTool::RestoreHeader(const char *container, const char *backup)
{
    if ((!container) || (!strlen(container)) || (!backup) || !strlen(backup))
    {
        this->ShowError("Container name or backup name is empty.\n");
        return false;
    }
    if (!this->WantRead(backup))
        return false;
    if (!this->WantChange(container))
        return false;

    int error = BCRestoreHeader(container, backup);
    if (error)
    {
        this->ShowError("Restoring failed!");
        return false;
    }
    return true;
}

bool BCTool::Passwd(const char *file, BCKeyOp operation)
{
    if (!this->WantChange(file))
        return false;

    if (this->CheckContainerInUse(file))
    {
        return false;
    }

    BCContainerResult error;
    switch (operation) {
    case kBCKey_Recode:
        error = BCChangePassword(file);
        break;

    case kBCKey_Duplicate:
        error = BCAddPassword(file);
        break;

    case kBCKey_Delete:
        error = BCRemovePassword(file);
        break;

    default:
        assert(0);
    };

    if(kBCContainer_Success != error) {
        if ((kBCContainer_Cancelled != error) && (kBCContainer_InvalidPassword != error)) {
            if ((error == kBCContainer_NoKeySpace) && (operation == kBCKey_Delete))
                this->ShowError("Cannot delete last main password.\n");
            else
                this->ShowError("Failed changing container password: %s\n",
                                BCContainerErrorDecription(error));
        }
        return false;
    }

    return true;
}

int BCTool::FormatVolume(const char* deviceName, const char* fsType)
{
    // Run mkfs on linked device.
    const char* mkfs = "/sbin/mkfs";
    int error;

    //printf("Formatting device %s to %s\n", deviceName, fsType);

    if (0 == strcasecmp(fsType, "vfat"))
    {
        error = this->RunProcessL(mkfs, "-t", "msdos", "-F", "32", deviceName, ARG_TERM);
    }
    else if (0 == strcasecmp(fsType, "ext2") || 0 == strcasecmp(fsType, "ext3"))
    {
        error = this->RunProcessL(mkfs, "-t", fsType, "-m", "0", deviceName, ARG_TERM);
    }
    else if (0 == strcasecmp(fsType, "ntfs"))
    {
        error = this->RunProcessL(mkfs, "-t", fsType, "-v", deviceName, ARG_TERM);
    }
    else
    {
        error = this->RunProcessL(mkfs, "-t", fsType, deviceName, ARG_TERM);
    }

    return error;
}

int BCTool::RunProcessL(const char* name, ...)
{
    char* args[1024] = { (char *)NULL };
    char** pargs = args;
    char** plast = &args[1023];

    *pargs++ = (char*)name;

    va_list list;
    va_start(list, name);

    while (NULL != (*pargs++ = va_arg(list, char*)))
    {
        assert(pargs != plast);
    }

    va_end(list);

    return this->RunProcessV(name, (const char**) args);
}

int BCTool::RunProcessV(const char *name, const char **args)
{
    pid_t pid = fork();

    if (pid < 0)
    {
        this->ShowError("Fork failed: %s\n", strerror(errno));
        return -1;
    }

    if (0 == pid)
    {
        // child
        const char* env[] = { "PATH=/sbin:/usr/sbin:/bin:/usr/bin", ARG_TERM };

        execve(name, (char* const*)args, (char* const*)env);

        // not reached if ok
        this->ShowError("execve(%s) failed: %s\n", name, strerror(errno));
        return -1;
    }

    // parent
    int status;
    waitpid( pid, &status, 0);
    return WEXITSTATUS(status);
}

BCTool::BCTool(MessageCallback messageCallback, AskUserCallback askUser,
                   SignalCallback handleSignal, BCProgressCallback reportProgress)
{
    this->messageCallback_ = messageCallback;
    this->AskUser = askUser;
    this->HandleSignal = handleSignal;
    this->ReportProgress = reportProgress;

    this->minimumMsgLevel = kBCTool_Info;
}

bool BCTool::CheckKernelInitialization()
{
    bc_kern_result result = BCKernelIsInitialized();
    if (kBCKernel_Success != result)
    {
        this->ShowError("Kernel modules are not loaded."
                        "\nIt might occur if you have a new kernel version installed."
                        "\nTo be able to use BestCrypt with a new kernel version, please go to the root of BestCrypt sources and execute the following commands:"
                        "\n\tmake module-clean"
                        "\n\tmake module"
                        "\n\tmake module-install");
        return false;
    }

    return true;
}


bool BCTool::CheckContainerInUse(const char *containerPath)
{
    // If we can't check that, consider container free, because we've tried.

    char fullPath[PATH_MAX], tmpFullPath[PATH_MAX];

    if (!realpath(containerPath, fullPath))
    {
        this->ShowError("Container is in use. Please eject it before this action.");
        return false;
    }

    std::vector<bc_mount_entry> mounts;
    if (!this->GetMountInfo(&mounts))
    {
        this->ShowError("Container is in use. Please eject it before this action.");
        return false;
    }

    std::vector<bc_mount_entry>::iterator i = mounts.begin();
    for (; i != mounts.end(); i++)
    {
        if (realpath(i->mountPoint, tmpFullPath))
        {
            if (0 == strncmp(fullPath, tmpFullPath, PATH_MAX))
                return true;
        }
    }

    return false;
}


bool BCTool::Format(const char *file, const char *fsType, bool showWarning)
{
    if (!this->WantChange(file))
        return false;

    if (this->CheckContainerInUse(file))
    {
        return false;
    }

    BCVolume vol;
    int error = BCDecodeVolume(file, NULL, false, NULL, NULL, &vol);
    if (kBCContainer_Success != error)
    {
        if ((kBCContainer_Cancelled != error) && (kBCContainer_InvalidPassword != error))
        {
            this->ShowError("Could not decode container file.\n");
        }

        return false;
    }

    if (showWarning)
    {
        if (!this->AskUser("Warning! After formatting container data will be destroyed! Are you sure you want to continue?", false))
        {
            return false;
        }
    }

    char deviceName[BC_MAX_PATH] = { 0 };
    error = BCKernelSetupDisk(vol.file, vol.info.alg, vol.info.mode, vol.key_handle,
                              vol.pool, vol.info.map, kBCDisk_Default, deviceName);
    BCPutVolume(&vol);

    if (kBCKernel_Success != error)
    {
        this->ShowError("Could not setup new device: %s\n", BCKernelGetErrorString(error));
        return false;
    }

    bool result = true;
    error = this->FormatVolume(deviceName, fsType);
    if (error)
    {
        this->ShowError("Could not format volume, mkfs returned %d\n", error);
        result = false;
    }

    error = BCKernelRemoveDisk(deviceName, 0);
    if (error)
    {
        this->ShowError("Could not remove block device: %s\n", BCKernelGetErrorString(error));
        result = false;
    }

    if (result)
    {
        this->ShowInfo("Format successful.\n");
    }

    return result;
}

bool BCTool::CheckFS(const char *file, const char *fsType)
{
    if (!this->WantChange(file))
        return false;

    if (this->CheckContainerInUse(file))
    {
        return false;
    }

    // ask password and link
    BCVolume vol;
    int error = BCDecodeVolume(file, "Enter password to check volume:", false, NULL, NULL, &vol);
    if (kBCContainer_Success != error)
    {
        if ((kBCContainer_Cancelled != error) && (kBCContainer_InvalidPassword != error))
        {
            this->ShowError("Could not decode container file.\n");
        }
        return false;
    }

    char deviceName[PATH_MAX] = { 0 };
    error = BCKernelSetupDisk(vol.file, vol.info.alg, vol.info.mode, vol.key_handle,
                              vol.pool, vol.info.map, kBCDisk_Default, deviceName);

    if (kBCKernel_Success != error)
    {
        BCPutVolume(&vol);
        this->ShowError("Could not setup new disk.\n");
        return false;
    }

    // If user did not specify fstype try to use implicit.
    if (fsType == (const char*)NULL)
    {
        error = this->RunProcessL("/sbin/fsck", deviceName, ARG_TERM);
    }
    else
    {
        error = this->RunProcessL("/sbin/fsck", "-t", fsType, deviceName, ARG_TERM);
    }

    bool succeeded = true;
    if (error)
    {
        if (error == 1)
            this->ShowWarning("Filesystem errors are corrected successfully.");
        else
        {
            this->ShowError("fsck failed with result %d. Check `man fsck` for more detail.\n", error);
            succeeded = false;
        }
    }

    // Wait one second for reference to be closed.
    sleep(1);

    if (kBCKernel_Success != BCKernelRemoveDisk(deviceName, 0))
    {
        this->ShowError("Could not remove device. Your container remains linked to %s.\n", deviceName);
        succeeded = false;
    }

    return succeeded;
}



void BCTool::ShowMessage(bctool_msg_level level, const char* msg, va_list args)
{
    if (this->messageCallback_ && (level >= this->minimumMsgLevel))
    {
        const int MaxLen = 4096;
        char str[MaxLen] = { '\0' };
        vsnprintf(str, MaxLen, msg, args);
        this->messageCallback_(level, str);
    }
}

void BCTool::ShowError(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Error, msg, args);
    va_end(args);
}

void BCTool::ShowWarning(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Warning, msg, args);
    //va_end(args);
}

void BCTool::ShowInfo(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Info, msg, args);
    va_end(args);
}

void BCTool::ShowDebug(const char *msg, ...)
{
    va_list args;
    va_start(args, msg);
    this->ShowMessage(kBCTool_Debug, msg, args);
    va_end(args);
}

bool BCTool::GetMountInfo(std::vector<bc_mount_entry> *out_entriesPtr)
{
    assert(out_entriesPtr != NULL);
    std::vector<BCDiskRef> disks;
    BCDiskRef::listDisks(&disks);

    std::vector<BCDiskRef>::iterator i = disks.begin();
    std::vector<BCDiskRef>::iterator end = disks.end();

    while (i != end)
    {
        const char *deviceName = i->nodeName();

        // Look for device mount point.
        // Configured devices may not be actually mounted.
        FILE *mtab = setmntent("/etc/mtab", "r");
        if (NULL == mtab)
        {
            this->ShowError("Could not open mtab: %s\n", strerror(errno));
            return false;
        }

        const char *mountPoint = NULL;
        struct mntent *mountEntry = NULL;

        while ((mountEntry = getmntent(mtab)) && !feof(mtab))
        {
            if (0 == strncmp(deviceName, mountEntry->mnt_fsname, strlen(deviceName)))
            {
                mountPoint = mountEntry->mnt_dir;
                break;
            }
        }

        bc_mount_entry entry;
        memset(&entry, 0, sizeof(entry));

        strncpy(entry.deviceName, deviceName, BC_MAX_PATH);
        if (i->backingStore())
        {
            strncpy(entry.containerPath, i->backingStore(), BC_MAX_PATH);

            std::string containerPathTmp(entry.containerPath);

            if (std::string::npos == containerPathTmp.find(" (deleted)"))
                entry.isFileExists = true;
//            else
//                entry.isFileExists = false;

            if (mountPoint)
            {
                strncpy(entry.mountPoint, mountPoint, BC_MAX_PATH);
            }
//            else
//                entry.mountPoint[0] = '\0';
        }
//        else
//        {
//            entry.containerPath[0] = '\0';
//            entry.isFileExists = false;
//            entry.mountPoint[0] = '\0';
//        }
        out_entriesPtr->push_back(entry);
        endmntent(mtab);
        ++i;
    }

    return true;
}

bool BCTool::GetContainerInfo(const char *path, bc_container_info *out_info)
{
    assert(path);
    assert(out_info);

    // Get information about container.
    BCVolumeInfo volumeInfo;
    if (!BCParseContainerFile(path, &volumeInfo))
        return false;

    BCAlgInfo algInfo;
    if (kBCKernel_Success != BCKernelLookupAlgId(volumeInfo.alg, &algInfo))
    {
        this->ShowError("Could not find algorithm with id 0x%x.\n", volumeInfo.alg);
        return false;
    }

    // Store information in output buffer.
    strncpy(out_info->containerPath, path, BC_MAX_PATH);
    out_info->size = volumeInfo.size;
    out_info->algId = volumeInfo.alg;
    strncpy(out_info->algName, algInfo.name, BC_ALG_NAME_SIZE);
    out_info->modeId = volumeInfo.mode;
    strncpy(out_info->modeName, BCKernelModeName(volumeInfo.mode), BC_ALG_MODE_NAME_SIZE);
    strncpy(out_info->description, volumeInfo.description, BC_DESCRIPTION_SIZE);

    return true;
}

// static
bool BCTool::PathExists(const char *path)
{
    int error = access(path, F_OK);
    return !error;
}

// static
bool BCTool::CanRead(const char *path)
{
    int error = access(path, R_OK);
    return !error;
}

// static
bool BCTool::CanWrite(const char* path)
{
    int error = access(path, W_OK);
    return !error;
}

// static
bool BCTool::IsFile(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return S_ISREG(st.st_mode);
}

// static
bool BCTool::IsDirectory(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return S_ISDIR(st.st_mode);
}

// static
bool BCTool::IsBlockSpecial(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return S_ISBLK(st.st_mode);
}

// static
bool BCTool::IsOwner(const char* path)
{
    struct stat st;
    if(0 != stat(path, &st))
    {
        return false;
    }

    return (st.st_uid == getuid());
}

// static
bc_u64 BCTool::GetBlockDeviceSize(const char* bdev)
{
    int fd = open(bdev, O_RDONLY);

    if (fd < 0)
    {
        perror("open");
        return 0;
    }

    unsigned long long size;
    if (0 != ioctl(fd, BLKGETSIZE, &size))
    {
        perror("ioctl");
        return 0;
    }

    size *= S_BLKSIZE;
    return size;
}

// static
int BCTool::GetUidByName(const char* name)
{
    struct passwd *pass;
    if ( NULL == name ) return -1;

    uid_t uid = atol(name);

    if( (pass = getpwnam(name)) || (pass = getpwuid(uid)) )
    return pass->pw_uid;

    return -1;
}

// static
int BCTool::GetGidByName(const char* name)
{
    struct group *grp;
    if ( NULL == name ) return -1;

    gid_t gid = atol(name);

    if( (grp = getgrnam(name)) || (grp = getgrgid(gid)) )
    return grp->gr_gid;

    return -1;
}

// static
long long BCTool::BCAtol(const char *string)
{
    char *endptr;
    long long result;

    result = strtoul(string, &endptr, 0);
    if (string == endptr)
    {
        return 0;
    }

    if ((string[0] != '\0') && (endptr[0] == '\0'))
    {
        return result;
    }

    if ((endptr[0] == 'K' || endptr[0] == 'k') && (endptr[1]=='\0'))
    {
        result *= 1024;
        return result;
    }

    if ((endptr[0] == 'M' || endptr[0] == 'm') && (endptr[1] == '\0'))
    {
        result *= 1024 * 1024;
        return result;
    }

    if ((endptr[0] == 'G' || endptr[0] == 'g') && (endptr[1] == '\0'))
    {
        result *= 1024 * 1024 * 1024;
        return result;
    }

    return 0L;
}

// static
bool BCTool::IsBCMount(const char* mountPoint)
{
    struct stat st;
    if (0 != stat(mountPoint, &st))
    {
        return false;
    }

    int blockMajor = BCDiskRef::blockMajor();
    return (blockMajor == (int)major(st.st_dev));
}
