#ifndef BCTOOL_H
#define BCTOOL_H

#include "container/user.h"
#include "container/volume.h"
#include "container/keygen.h"
#include "kernel/BCKernel.h"

#include <vector>
#include <stdarg.h>

#define BC_ALG_MODE_NAME_SIZE   4
#define BC_ALG_NAME_SIZE        16

#define ARG_TERM                ((char*)0)

// Enable printf-like parameter check for function in class (first parameter is 'this' pointer).
#define PRINTF_LIKE __attribute__((format(printf, 2, 3)))

extern const char *PROG_NAME;

typedef enum
{
    kBCTool_Debug = 0,
    kBCTool_Info,
    kBCTool_Warning,
    kBCTool_Error
} bctool_msg_level;

typedef void    (*MessageCallback)(bctool_msg_level level, const char *msg);
typedef bool    (*AskUserCallback)(const char *msg, bool defaultAnswer);
typedef void    (*SignalCallback)(int signal);
//typedef int     (*ProgressCallback)(const char *message, bc_u64 byteCount, bc_u64 bytesTotal, void *arg);

/// @brief Container mount entry information.
typedef struct
{
    char deviceName[BC_MAX_PATH];           ///< BestCrypt device name, ex. /dev/bcrypta
    char containerPath[BC_MAX_PATH];        ///< Container file linked to device.
    char mountPoint[BC_MAX_PATH];           ///< Container mount point. Contains empty string for containers linked without mounting.
    bool isFileExists;                      ///< True, if container file is still exists on disk, otherwise false.
} bc_mount_entry;

/// @brief Container file information.
/// This information is available for containers with unencrypted header only.
typedef struct
{
    bc_alg_id   algId;
    bc_mode_id  modeId;
    bc_u64      size;
    char        containerPath[BC_MAX_PATH];
    char        description[BC_DESCRIPTION_SIZE];
    char        algName[BC_ALG_NAME_SIZE];
    char        modeName[BC_ALG_MODE_NAME_SIZE];
} bc_container_info;

/// @brief Benchmark information item.
typedef struct
{
    char            algName[BC_ALG_NAME_SIZE];          ///< Name of the algorithm being benchmarked.
    char            modeName[BC_ALG_MODE_NAME_SIZE];    ///< Name of the algorithm mode used.
    unsigned long   averageTime;                        ///< Average time in nanoseconds to encrypt one megabyte of garbage.
} bc_bench_info;

class BCTool
{
public:
    bctool_msg_level minimumMsgLevel;

    BCTool(MessageCallback messageCallback, AskUserCallback askUser, SignalCallback handleSignal,
             BCProgressCallback reportProgress);

    bool CheckKernelInitialization();

    /**
      @brief Tries to find a BestCrypt device that uses container with specified path.
            If yes, displays a message.
      @return True, if a device was found, otherwise false.
    */
    bool CheckContainerInUse(const char *containerPath);

    bool Mount(const char *file, BCMountParams *mountParams, const char *fsType,
               int uid, int gid, const char *mountOptions, char *out_deviceName);

    bool Umount(const char *mountPoint, bool force);

    bool UmountAll(bool force, int *out_mountedCount, int *out_unmountedCount);

    bool Link(const char *file, bool readOnly, BCMountParams *mountParams, bool mainPartOnly,
              BCVolume *volume, BCPassword *out_password, char *out_deviceName);

    bool RawLink(const char *file, const char *algName, const char *modeName, bool readOnly);

    bool Unlink(const char *deviceName, bool force);

    bool New(const char* fileName, const char *algName, const char* modeName,
             bc_u64 size, const char* description, const char *fsType);

    bool CreateHiddenPart(const char *fileName, bc_u64 size, const char *fsType);
    
    bool Format(const char *file, const char *fsType, bool showWarning);

    bool CheckFS(const char* file, const char *fsType);

    bool Passwd(const char *file, BCKeyOp operation);

    bool Reencrypt(const char *file, const char *algName, const char *modeName);

    bool BackupHeader(const char *file, const char *backup);

    bool RestoreHeader(const char *container, const char *backup);

    bool ProcessHeader(const char *container, bool encrypt);

    bool DeleteContainer(const char *container);

    bool GetMountInfo(std::vector<bc_mount_entry> *out_entriesPtr);

    bool GetContainerInfo(const char *path, bc_container_info *out_info);

    bool Benchmark(const char *algName, const char *modeName, std::vector<bc_bench_info> *out_result);    
    
    static bool PathExists(const char *path);
    static bool CanRead(const char *path);
    static bool CanWrite(const char* path);
    static bool IsFile(const char* path);
    static bool IsDirectory(const char* path);
    static bool IsBlockSpecial(const char* path);
    static bool IsOwner(const char* path);
    static bc_u64 GetBlockDeviceSize(const char* bdev);
    static int GetUidByName(const char* name);
    static int GetGidByName(const char* name);
    static long long BCAtol(const char *string);
    static bool IsBCMount(const char* mountPoint);

private:
    MessageCallback     messageCallback_;

    AskUserCallback     AskUser;
    SignalCallback      HandleSignal;
    BCProgressCallback  ReportProgress;

    bool DoLink(const char *file, BCVolume *volume, char *out_deviceName, bool readOnly);
    int RunProcessL(const char *name, ...);
    int RunProcessV(const char *name, const char **args);
    int FormatVolume(const char *deviceName, const char *fsType);
    int GetFirstDevicePartition(const char *mainDeviceName, char *out_partitionDeviceName, int maxNameSize);
    SpaceMap *GetFreeSpaceFromMainPart(const char *deviceName);
    int DoUmount(const char* mountPoint, const char* deviceName, bool force);

    /**
      @brief Cuts end of the map, if map's size is bigger than specified in 'size' parameter.
    */
    SpaceMap *TransformToSubmapOfSize(SpaceMap *map, bc_u64 size);

    void ShowError(const char *msg, ...) PRINTF_LIKE;
    void ShowWarning(const char *msg, ...) PRINTF_LIKE;
    void ShowInfo(const char *msg, ...) PRINTF_LIKE;
    void ShowDebug(const char *msg, ...) PRINTF_LIKE;
    void ShowMessage(bctool_msg_level level, const char* msg, va_list args);

    bool WantChange(const char *file);
    bool WantRead(const char *file);
};

#endif // BCTOOL_H
