#include "container/user.h"
#include "container/intsize.h"
#include "passworddialog.h"
#include "hiddenpartdialog.h"
#include "model/bcoperation.h"
#include "model/helper.h"

#include <QMessageBox>
#include <QApplication>
#include <QWaitCondition>
#include <QMutex>
#include <QProgressDialog>

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <iconv.h>
#include <locale.h>
#include <langinfo.h>
#include <termios.h>


static size_t convertString(char *msg, size_t msgSize, char *out_encodedMsg, size_t bufferSize, const char *fromEncoding, const char *toEncoding)
{
    iconv_t descriptor = iconv_open(toEncoding, fromEncoding);

    if (((iconv_t)-1) == descriptor) {
        fprintf(stderr, "Error creating conversion descriptor: %s\n", strerror(errno));
        return 0;
    }

    size_t nbytesLeft = bufferSize;
    if (-1 == (int)iconv(descriptor, &msg, &msgSize, &out_encodedMsg, &nbytesLeft)) {
        fprintf(stderr, "Error while converting: %s\n", strerror(errno));
        return 0;
    }

    iconv_close(descriptor);
    return bufferSize - nbytesLeft;
}

static size_t convertToUTF8(char *msg, size_t length, char *out_resultBuffer, size_t bufferSize)
{
    // Set program's current locale to system locale and use result as source encoding.
    setlocale(LC_ALL, "");
    const char *currentEncoding = nl_langinfo(CODESET);
    const char *targetEncoding = "UTF-8";

    return convertString(msg, length, out_resultBuffer, bufferSize, currentEncoding, targetEncoding);
}

static const char* getIconvEncoding(BCUserEncoding tag)
{
    switch (tag)
    {
        case kBCUser_UTF8:
            return "UTF-8";
        case kBCUser_UTF16LE:
            return "UTF-16LE";
        case kBCUser_UTF16BE:
            return "UTF-16BE";
        default:
            assert(0);
            return 0;
    };
}

static BCUserResult ShowPasswordDlg(const char *query, bool needConfirm, BCVerifyPasswordCallback callback, void *cookie, BCMountParams *mountParams, BCPassword *out_pwd)
{
    Q_ASSERT(query);
    BCUserResult userResult;
    PasswordDialog dlg;
    dlg.SetConfirmVisible(needConfirm);
    dlg.SetMountParams(mountParams);
    dlg.SetPasswordPromptText(query);
    dlg.SetPasswordBuffer(out_pwd);
    dlg.setModal(true);
    dlg.show();

    while (true)
    {
        dlg.setEnabled(true);
        while(!dlg.HaveResult())
        {
            BCHelperSleep(10);
            QApplication::processEvents();
        }

        // Password check can be long process, so disable dialog to indicate it.
        dlg.setEnabled(false);
        QApplication::processEvents();
        userResult = dlg.GetResult();        
        if (kBCUserCancel == userResult)
            return kBCUserCancel;

        if (callback)
        {
            const char* errorText = callback(out_pwd->buf, out_pwd->length, cookie);
            if (errorText)
            {
                dlg.ShowError(errorText);
                dlg.ResetResult();

                continue;
            }
        }
        return kBCUserOk;
    }
}

BCUserResult BCUserVerifyPassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCMountParams *mountParams, BCPassword *out_pwd)
{
    if (!query)
        query = BC_DEFAULT_PASSWORD_QUERY;

    return ShowPasswordDlg(query, false, callback, cookie, mountParams, out_pwd);
}


BCUserResult BCUserCreatePassword(const char* query, BCVerifyPasswordCallback callback, void* cookie, BCPassword* out_pwd)
{
    if (!query)
        query = BC_DEFAULT_NEW_PASSWORD_QUERY;

    return ShowPasswordDlg(query, true, callback, cookie, /*mountParams = */0, out_pwd);
}

unsigned GenerateRandom(const char *message, char *out_buf, unsigned total,
                        BCProgressCallback callback, void *callbackArg, int strong)
{
    memset(out_buf, 0, total);

    FILE *f;

    if (strong) {
        f = fopen("/dev/random", "r");
    }
    else {
        f = fopen("/dev/urandom", "r");
    }

    ssize_t bytesRead = 0;
    QApplication::processEvents();

    if (!f)
    {
        BCOperationShowErrorString(QString("Cannot open /dev/%1: %0\n").arg(strerror(errno)).arg((strong)? "random" : "urandom"));
        return 0;
    }

    do
    {
        out_buf[bytesRead++] = fgetc(f);
        QApplication::processEvents();

        if (callback)
        {
            if (!callback(message, bytesRead, total, callbackArg))
            {
                fclose(f);
                return 0;
            }
        }
    }
    while (bytesRead < total);

    fclose(f);
    return total;
}


unsigned BCUserGenerateSeed(const char *message, char *out_buf, unsigned total,
                            BCProgressCallback callback, void *callbackArg)
{
    return GenerateRandom(message, out_buf, total, callback, callbackArg, 1);
}


size_t BCUserConvertString(char* in, size_t inlen, char* out, size_t outlen, BCUserEncoding fromEncoding, BCUserEncoding toEncoding)
{
    const char* from = getIconvEncoding(fromEncoding);
    const char* to = getIconvEncoding(toEncoding);
    return convertString(in, inlen, out, outlen, from, to);
}


size_t BCUserUTF16ToUTF8(char* in_utf16, size_t inlen, char* out_utf8, size_t outlen)
{
    return convertString(in_utf16, inlen, out_utf8, outlen, "UTF-16BE", "UTF-8");
}

size_t BCUserUTF8ToUTF16(char* in_utf8, size_t inlen, char* out_utf16, size_t outlen)
{
    return convertString(in_utf8, inlen, out_utf16, outlen, "UTF-8", "UTF-16BE");
}


void BCUserMessage(const char* msg, BCUserMessageLevel level)
{
    QMessageBox msgBox;
    QMessageBox::Icon icon;
    switch (level)
    {
        case kBCUserMessage_Info:
            icon = QMessageBox::Information;
            break;
        case kBCUserMessage_Warning:
            icon = QMessageBox::Warning;
            break;
        case kBCUserMessage_Error:
            icon = QMessageBox::Critical;
            break;
        default:
            Q_ASSERT(0);
            break;
    }

    msgBox.setIcon(icon);
    msgBox.setText(msg);
    msgBox.setWindowTitle("BestCrypt");
    msgBox.exec();
}

BCUserResult BCUserProtectHiddenPart(void *header, const char* container,
                    BCProtectExistingHiddenPartCallback protectCallback, SpaceMap *hiddenMap,
                    char *fsType, SpaceMap **out_resizedMap)
{
    HiddenPartDialog dlg;
    HiddenPartDialogParameters params;
    QDialog::DialogCode result;

    params.containerFile = container;
    params.header = header;
    params.protectCallback = protectCallback;
    params.freeSpaceMap = hiddenMap;
    params.fsType = fsType;
    params.resizedMap = NULL;

    dlg.SetParams(&params);

    dlg.exec();
    result = dlg.GetResult();

    if (result == QDialog::Accepted)
    {
        (*out_resizedMap) = params.resizedMap;
        return kBCUserOk;
    }

    return kBCUserCancel;
}
