#include "containerinfowidget.h"
#include "ui_containerinfowidget.h"

#include "entity/containerdescriptor.h"
#include "model/bcoperation.h"
#include "model/helper.h"
#include "additionalactionswidget.h"

#include <QFileDialog>
#include <QScrollBar>
#include <QFontMetrics>
#include <QResizeEvent>
#include <QDesktopServices>
#include <QUrl>
#include <QDir>
#include <QMessageBox>

static void SetClippedTextToLabel(QLabel *label, QString text)
{
    QFontMetrics metrix(label->font());
    int width = label->width() - 2;
    QString clippedText = metrix.elidedText(text, Qt::ElideRight, width);
    label->setText(clippedText);
}

static void SetTextOrDefault(QLabel *label, const QString& text, const QString& defaultText)
{
    if (text.count())
    {
        SetClippedTextToLabel(label, text);
    }
    else
    {
        SetClippedTextToLabel(label, defaultText);
    }
}

ContainerInfoWidget::ContainerInfoWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::ContainerInfoWidget)
{
    ui->setupUi(this);

    this->descriptor_ = 0;
    connect(this->ui->mountLocationText, SIGNAL(textChanged(QString)), this, SLOT(OnMountPointLocationChangedSlot(QString)));
    connect(this->ui->mountReadOnlyCheckbox, SIGNAL(stateChanged(int)), this, SLOT(OnMountReadOnlyCheckStateChangedSlot(int)));
    connect(this->ui->openMountLocation, SIGNAL(clicked()), this, SLOT(OnOpenMountLocationSlot()));

}

const AdditionalActionsWidget* ContainerInfoWidget::GetAdditionalActionsWidget()
{
    return this->ui->additionalActionsWidget;
}

void ContainerInfoWidget::UpdateContainerInfo(ContainerDescriptor *descriptor)
{
    // Fill controls with descriptor parameters.
    QFileInfo fileInfo(this->descriptor_->GetPath());
    SetTextOrDefault(this->ui->containerLocationLabel, fileInfo.absolutePath(), "-");

    SetTextOrDefault(this->ui->containerNameLabel, descriptor->GetFileName(), "-");

    SetTextOrDefault(this->ui->containerSizeLabel, BCHelperGetSizeString(this->descriptor_->size), "-");
    SetTextOrDefault(this->ui->containerDescriptionLabel, descriptor->description, "None");
    SetTextOrDefault(this->ui->containerMountPointLabel, descriptor->GetMountPoint(), "-");
    SetTextOrDefault(this->ui->deviceLabel, descriptor->GetDeviceName(), "-");

    if (this->descriptor_->IsLinked())
        SetTextOrDefault(this->ui->containerMountPointLabel, descriptor->GetMountPoint(), "Linked only");

    if (!this->descriptor_->IsExists())
        this->ui->errorLabel->setText("File not found.");
    else if (!this->descriptor_->IsParsed())
    {
        this->ui->errorLabel->setText("Encrypted or corrupted header.");
        this->ui->containerAlgorithmLabel->setText("-");
    }
    else
    {
        this->ui->containerAlgorithmLabel->setText(QString("%0-%1").arg(this->descriptor_->algName).arg(this->descriptor_->modeName));
    }

    // Set mount options.
    this->ui->mountReadOnlyCheckbox->setChecked(descriptor->mountReadOnly);
    this->ui->mountLocationText->setEnabled(descriptor->IsExists() && (!descriptor->IsLinked()));
    this->ui->mountLocationButton->setEnabled(this->ui->mountLocationText->isEnabled());
    this->ui->openMountLocation->setEnabled(descriptor->lastMountPoint.count());

    this->ui->mountLocationText->setText(descriptor->lastMountPoint);
}

void ContainerInfoWidget::SetData(ContainerDescriptor *descriptor)
{
    this->descriptor_ = descriptor;
    this->ui->errorLabel->setText("");
    this->ui->additionalActionsWidget->SetDescriptor(descriptor);

    if (!descriptor)
    {
        // Display welcome screen if no descriptor is selected.
        this->ui->welcomeWidget->setVisible(true);
        this->ui->containerInfoWidget->setVisible(false);
    }
    else
    {
        this->ui->welcomeWidget->setVisible(false);
        this->ui->containerInfoWidget->setVisible(true);

        UpdateContainerInfo(descriptor);
    }
}

QString ContainerInfoWidget::GetMountLocation()
{
    return this->ui->mountLocationText->text();
}

void ContainerInfoWidget::SetMountLocation(const QString &mountLocation)
{
    this->ui->mountLocationText->setText(QString(mountLocation));
}

bool ContainerInfoWidget::NeedMountReadOnly()
{
    return this->ui->mountReadOnlyCheckbox->checkState() == Qt::Checked;
}

ContainerInfoWidget::~ContainerInfoWidget()
{
    delete ui;
}

void ContainerInfoWidget::on_createButton_clicked()
{
    emit this->OnCreate();
}

void ContainerInfoWidget::on_browseButton_clicked()
{
    emit this->OnBrowse();
}

void ContainerInfoWidget::on_mountLocationButton_clicked()
{
    QString mountLocation = BCHelperGetLocation(tr("Please select location where container should be mounted:"),
                                                this->descriptor_->lastMountPoint);
    if (mountLocation.count())
    {
        this->ui->mountLocationText->setText(mountLocation);
    }
}

void ContainerInfoWidget::OnMountPointLocationChangedSlot(QString text)
{
    Q_ASSERT(this->descriptor_);
    this->descriptor_->lastMountPoint = QString(text);
    this->ui->openMountLocation->setEnabled(text.count());
}

void ContainerInfoWidget::OnMountReadOnlyCheckStateChangedSlot(int state)
{
    Q_ASSERT(this->descriptor_);
    this->descriptor_->mountReadOnly = state == Qt::Checked;
}

void ContainerInfoWidget::OnOpenMountLocationSlot()
{
    QString dir = this->ui->mountLocationText->text();
    if (QDir(dir).exists())
    {
        QDesktopServices::openUrl(QUrl(QString("file:///") + dir));
    }
    else
    {
        QMessageBox::information(this, "Incorrect Path", "Location '" + dir
                                 + "' was not found.", QMessageBox::Ok);
    }
}

void ContainerInfoWidget::resizeEvent(QResizeEvent */*evt*/)
{
    if (!this->descriptor_)
        return;
    this->UpdateContainerInfo(this->descriptor_);
//    if (this->descriptor_->IsMounted())
//        SetClippedTextToLabel(this->ui->containerMountPointLabel,
//                              this->descriptor_->GetMountPoint());

//    QFileInfo fileInfo(this->descriptor_->GetPath());
//    SetClippedTextToLabel(this->ui->containerLocationLabel, fileInfo.absolutePath());
}
