#include "containertreewidget.h"
#include "ui_containertreewidget.h"

#include "model/containermodel.h"
#include "model/bcoperation.h"
#include "entity/containerdescriptor.h"
#include "containerview.h"

#include <QLayoutItem>
#include <QKeyEvent>
#include <QMutex>

ContainerTreeWidget::ContainerTreeWidget(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::ContainerTreeWidget)
{
    ui->setupUi(this);
    this->containerModel_ = 0;
    this->selectedView_ = 0;
    this->containerViews_ = new QList<ContainerView*>();

    this->deleteFileAction_ = new QAction(tr("Delete container file"), this->ui->removeToolButton);
    this->clearListAction_ = new QAction(tr("Clear list"), this->ui->removeToolButton);

    this->ui->removeToolButton->addAction(this->clearListAction_);
    this->ui->removeToolButton->addAction(this->deleteFileAction_);

    connect(this->ui->removeToolButton, SIGNAL(clicked()), this, SLOT(RemoveButtonClickedSlot()));
    connect(this->clearListAction_, SIGNAL(triggered()), this, SLOT(ClearButtonClickSlot()));
    connect(this->deleteFileAction_, SIGNAL(triggered()), this, SIGNAL(OnDelete()));
}

ContainerTreeWidget::~ContainerTreeWidget()
{
    delete ui;

    qDeleteAll(*(this->containerViews_));
    delete this->containerViews_;

    if (this->containerModel_)
    {
        this->containerModel_->SaveToConfig();
        delete this->containerModel_;
    }
}

void ContainerTreeWidget::InitModel(ContainerModel *model)
{
    Q_ASSERT(!this->containerModel_);
    this->containerModel_ = model;
    connect(model, SIGNAL(OnDataChanged()), this, SLOT(DataChangedSlot()), Qt::DirectConnection);
    this->UpdateViewList();
}

bool ContainerTreeWidget::SelectDescriptor(ContainerDescriptor *descriptor)
{
    ContainerView *cur;
    for (int i = 0; i < this->containerViews_->count(); i++)
    {
        cur = this->containerViews_->at(i);
        if (cur->GetDescriptor() == descriptor)
        {
            this->SelectView(cur);
            return true;
        }
    }
    return false;
}

void ContainerTreeWidget::UpdateCurrentView()
{
    if (this->selectedView_)
    {
        this->selectedView_->UpdateView();
    }
}

void ContainerTreeWidget::UpdateViewList()
{
    QString selectedContainerPath;
    if (this->selectedView_)
        selectedContainerPath = this->GetSelectedDescriptor()->GetPath();
    this->selectedView_ = 0;

    if (this->containerViews_->count())
    {
        ContainerView *cur = 0;
        QLayoutItem *item = 0;

        for (int i=0; i < this->containerViews_->count(); i++)
        {
            cur = this->containerViews_->at(i);
            disconnect(cur, SIGNAL(OnSelected(ContainerView*)), this, SLOT(SelectedViewChangedSlot(ContainerView*)));
            disconnect(cur, SIGNAL(OnEject(ContainerView*)), this, SLOT(SelectedViewEjectClickedSlot(ContainerView*)));
            disconnect(cur, SIGNAL(OnMount(ContainerView*)), this, SLOT(SelectedViewMountClickedSlot(ContainerView*)));
            disconnect(cur, SIGNAL(OnSelectNext(ContainerView*)), this, SLOT(SelectedViewSelectNextSlot(ContainerView*)));
            disconnect(cur, SIGNAL(OnSelectPrevious(ContainerView*)), this, SLOT(SelectedViewSelectPreviousSlot(ContainerView*)));
        }

        while (this->ui->itemLayout->count() > 1)
        {
            item = this->ui->itemLayout->takeAt(0);
            delete item->widget();
            delete item;
        }

        this->containerViews_->clear();
    }

    QList<ContainerDescriptor *>* descriptors = this->containerModel_->GetDescriptors();
    ContainerView *view = 0;
    int indexToSelect = 0;
    ContainerDescriptor *descriptor = 0;
    for (int i = 0; i < descriptors->count(); i++)
    {
        descriptor = descriptors->at(i);
        view = new ContainerView(this->ui->scrollArea);

        view->SetDescriptor(descriptor);
        if (descriptor->GetPath() == selectedContainerPath)
            indexToSelect = i;

        connect(view, SIGNAL(OnSelected(ContainerView*)), this, SLOT(SelectedViewChangedSlot(ContainerView*)));
        connect(view, SIGNAL(OnEject(ContainerView*)), this, SLOT(SelectedViewEjectClickedSlot(ContainerView*)));
        connect(view, SIGNAL(OnMount(ContainerView*)), this, SLOT(SelectedViewMountClickedSlot(ContainerView*)));
        connect(view, SIGNAL(OnSelectNext(ContainerView*)), this, SLOT(SelectedViewSelectNextSlot(ContainerView*)));
        connect(view, SIGNAL(OnSelectPrevious(ContainerView*)), this, SLOT(SelectedViewSelectPreviousSlot(ContainerView*)));

        this->containerViews_->append(view);
        this->ui->itemLayout->insertWidget(i, view);
    }

    this->SelectContainerByIndex(indexToSelect);
}

void ContainerTreeWidget::SelectContainerByIndex(int index)
{
    if (index >= 0 && index < this->containerViews_->count())
    {
        this->SelectView(this->containerViews_->at(index));
    }
    else
    {
        this->SelectView(0);
    }
}



void ContainerTreeWidget::SelectedViewChangedSlot(ContainerView *view)
{
    this->SelectView(view);
}

void ContainerTreeWidget::SelectedViewEjectClickedSlot(ContainerView *view)
{
    this->SelectView(view);
    emit this->OnEject();
}

void ContainerTreeWidget::SelectedViewMountClickedSlot(ContainerView */*view*/)
{
    // Mount emits on double-click, so view should be selected by first click.
    emit this->OnMount();
}

void ContainerTreeWidget::SelectedViewSelectNextSlot(ContainerView *view)
{
    int i;
    for (i = 0; i < this->containerViews_->count() - 1; i++)
    {
        if (this->containerViews_->at(i) == view)
        {
            this->SelectView(this->containerViews_->at(i + 1));
            break;
        }
    }
}

void ContainerTreeWidget::SelectedViewSelectPreviousSlot(ContainerView *view)
{
    int i;
    for (i = 1; i < this->containerViews_->count(); i++)
    {
        if (this->containerViews_->at(i) == view)
        {
            this->SelectView(this->containerViews_->at(i -1));
            break;
        }
    }
}

void ContainerTreeWidget::DataChangedSlot()
{
    this->UpdateViewList();
}

void ContainerTreeWidget::ClearButtonClickSlot()
{
    this->containerModel_->Clear();
}

void ContainerTreeWidget::RemoveButtonClickedSlot()
{
    ContainerDescriptor *descriptor = this->GetSelectedDescriptor();
    if (descriptor)
    {
        this->containerModel_->Remove(descriptor);
    }
}

void ContainerTreeWidget::SelectView(ContainerView *view)
{
    // The commented code below doesn't work for some reason. It needs to be checked.
    //this->mutex_->lock();
//    if (this->selectedView_)
//        this->selectedView_->Deselect();
//    this->selectedView_ = view;
//    if (this->selectedView_)
//        this->selectedView_->Select();
    for (int i = 0; i < this->containerViews_->count(); i++)
    {
        this->containerViews_->at(i)->Deselect();
    }

    if (view)
    {
        view->Select();
    }

    this->selectedView_ = view;

    emit this->OnSelectedDescriptorChanged(this->GetSelectedDescriptor());
    //this->mutex_->unlock();
}

void ContainerTreeWidget::keyPressEvent(QKeyEvent *event)
{
    if (!this->selectedView_)
        return;

    ContainerDescriptor *descriptor = this->selectedView_->GetDescriptor();

    if (!descriptor)
        return;

    switch(event->key())
    {
        case Qt::Key_Return:
        case Qt::Key_Enter:
            if (descriptor->IsExists())
            {
                event->accept();
                if (descriptor->IsLinked())
                    emit this->OnEject();
                else
                    emit this->OnMount();
            }
            break;

        case Qt::Key_Up:
            event->accept();
            this->SelectedViewSelectPreviousSlot(this->selectedView_);
            break;

        case Qt::Key_Down:
            event->accept();
            this->SelectedViewSelectNextSlot(this->selectedView_);
            break;

        default:
            break;
    }
}
