#ifndef CONTAINERDESCRIPTOR_H
#define CONTAINERDESCRIPTOR_H

#include <QString>
#include <QFileInfo>

#include "container/cipher.h"
#include "container/intsize.h"

/**
 * @brief Represents container parameters that will be used in user interface.
 */
class ContainerDescriptor : public QObject
{
    Q_OBJECT

public:
    ContainerDescriptor()
    {
        this->state_ = kBCContainerState_Unlinked;
        this->algId = -1;
        this->modeId = -1;
        this->size = 0;
        this->isParsed = false;
        this->mountReadOnly = false;
    }
    ~ContainerDescriptor() {}

    /// @brief Container mount status.
    typedef enum
    {
        kBCContainerState_Unlinked, ///< Container is not linked.
        kBCContainerState_Linked,   ///< Container is linked to device, but not mounted.
        kBCContainerState_Mounted   ///< Container is linked to device and mounted to some folder.
    } bc_container_state;

    const QString& GetPath() const
    {
        return this->path_;
    }
    void SetPath(const QString& path)
    {
        this->path_ = path;
        QFileInfo info(this->path_);
        this->fileName_ = info.fileName();
    }

    const QString& GetFileName() const
    {
        return this->fileName_;
    }

    const QString& GetMountPoint() const
    {
        return this->mountPoint_;
    }

    const QString& GetDeviceName() const
    {
        return this->device_;
    }

    bc_container_state GetState() const
    {
        return this->state_;
    }

    bool IsExists() const
    {
        QFileInfo info(this->GetPath());
        return info.exists();
    }

    bool IsMounted() const
    {
        return this->state_ == kBCContainerState_Mounted;
    }

    bool IsLinked() const
    {
        return (this->state_ == kBCContainerState_Linked) || (this->state_ == kBCContainerState_Mounted);
    }

    bool IsParsed() const
    {
        return this->isParsed;
    }

    void SetParsed(bool isParsed)
    {
        this->isParsed = isParsed;
        emit this->OnChanged();
    }

    void Mount(const QString& device, const QString& mountPoint)
    {
        this->device_ = device;
        this->mountPoint_ = QString(mountPoint);
        this->state_ = kBCContainerState_Mounted;
        emit this->OnChanged();
    }

    void Unmount()
    {
        this->mountPoint_.clear();
        this->state_ = kBCContainerState_Linked;
        emit this->OnChanged();
    }

    void Link(const QString& device)
    {
        this->device_ = device;
        this->mountPoint_.clear();
        this->state_ = kBCContainerState_Linked;
        emit this->OnChanged();
    }

    void Unlink()
    {
        this->mountPoint_.clear();
        this->device_.clear();
        this->state_ = kBCContainerState_Unlinked;
        emit this->OnChanged();
    }

    static ContainerDescriptor *CreateForNewContainer(QString filePath, bc_u64 size, int algId, QString algName,
                                                      int modeId, QString modeName, QString fileSystem,
                                                      QString description)
    {
        ContainerDescriptor *result = new ContainerDescriptor();
        result->SetPath(filePath);
        result->size = size;
        result->algId = algId;
        result->algName = QString(algName);
        result->modeId = modeId;
        result->modeName = QString(modeName);
        result->fileSystem = fileSystem;
        result->description = description;
        result->state_ = kBCContainerState_Unlinked;
        return result;
    }

signals:
    void OnChanged();

public:
    QString             modeName;
    QString             algName;
    QString             description;
    QString             lastMountPoint;
    QString             fileSystem;
    int                 algId;
    int                 modeId;
    bool                mountReadOnly;
    bc_u64              size;

private:
    QString             path_;
    QString             fileName_;
    QString             device_;
    QString             mountPoint_;
    bool                isParsed;

    bc_container_state  state_;
};

#endif // CONTAINERDESCRIPTOR_H
