#include "settings.h"

#include "src/version.h"
#include "container/cipher.h"
#include "entity/containerdescriptor.h"

#include "model/bcoperation.h"

#include <QSettings>
#include <QVariant>
#include <QMessageBox>

// General properties.
#define BC_CFG_USE_CONFIG               "UseConfigFile"
#define BC_CFG_SAVE_CONTAINER_LIST      "SaveContainerList"
#define BC_CFG_EJECT_ALL_HOTKEY         "EjectAllHotKey"
#define BC_CFG_EJECT_ALL_HOTKEY_MODS    "EjectAllHotKeyModifiers"
#define BC_CFG_EJECT_ALL_FORCE          "EjectAllForce"
#define BC_CFG_EJECT_ALL_USE_HOTKEY     "EjectAllUseHotKey"
#define BC_CFG_MINIMIZE_TO_TRAY         "MinimizeToTray"

// Container properties.
#define BC_CFG_CONTAINER_ARRAY          "Containers"
#define BC_CFG_CONTAINER_PATH           "Path"
#define BC_CFG_MOUNT_POINT              "MountPoint"
#define BC_CFG_MOUNT_READ_ONLY          "AlwaysMountReadOnly"

// Groups.
#define BC_CFG_STRINGS_GROUP            "Strings"
#define BC_CFG_SIZES_GROUP              "Sizes"
#define BC_CFG_POINTS_GROUP             "Points"
#define BC_CFG_BOOL_GROUP               "Booleans"

// Main form properties.
#define BC_CFG_MAINFORM_GROUP           "MainForm"
#define BC_CFG_MAINFORM_SIZE            "MainFormSize"
#define BC_CFG_MAINFORM_LOCATION        "MainFormLocation"
#define BC_CFG_MAINFORM_MAXIMIZED       "MainFormIsMaximized"


#define CHECK_SAVE(ret_val)   if (!IsConfigUsed()) return ret_val;

BCSettings::BCSettings()
    : userSettings_(QSettings::UserScope, BC_VENDOR, BC_GUI_APP_NAME)
{
    // We don't use other settings, so turn it off.
    this->userSettings_.setFallbacksEnabled(false);
}

QList<ContainerDescriptor*>* BCSettings::LoadContainerList()
{
    QList<ContainerDescriptor*> *containerList = new QList<ContainerDescriptor*>();
    int containerCount = userSettings_.beginReadArray(BC_CFG_CONTAINER_ARRAY);
    ContainerDescriptor *descriptor = 0;
    CHECK_SAVE(containerList);

    for (int i = 0; i < containerCount; i++)
    {
        userSettings_.setArrayIndex(i);
        descriptor = new ContainerDescriptor();
        descriptor->SetPath(userSettings_.value(BC_CFG_CONTAINER_PATH).toString());
        descriptor->lastMountPoint = userSettings_.value(BC_CFG_MOUNT_POINT).toString();
        descriptor->mountReadOnly = userSettings_.value(BC_CFG_MOUNT_READ_ONLY).toBool();
        containerList->append(descriptor);
    }

    userSettings_.endArray();

    return containerList;
}

void BCSettings::SaveContainerList(QList<ContainerDescriptor *> *descriptorList)
{
    CHECK_SAVE();

    Q_CHECK_PTR(descriptorList);

    // Remove old container list and add new.
    userSettings_.remove(BC_CFG_CONTAINER_ARRAY);
    userSettings_.beginWriteArray(BC_CFG_CONTAINER_ARRAY);
    ContainerDescriptor *cur = 0;

    for (int i = 0; i < descriptorList->count(); i++)
    {
        cur = descriptorList->at(i);

        if (cur)
        {
            userSettings_.setArrayIndex(i);
            userSettings_.setValue(BC_CFG_CONTAINER_PATH, cur->GetPath());
            userSettings_.setValue(BC_CFG_MOUNT_READ_ONLY, cur->mountReadOnly);
            userSettings_.setValue(BC_CFG_MOUNT_POINT, cur->lastMountPoint);
        }
    }

    userSettings_.endArray();
}

bool BCSettings::IsEjectAllHotKeyEnabled()
{
    return this->userSettings_.value(BC_CFG_EJECT_ALL_USE_HOTKEY).toBool();
}

void BCSettings::SetEjectAllHotKeyEnabled(bool enabled)
{
    this->SaveVariant(BC_CFG_EJECT_ALL_USE_HOTKEY, enabled);
}

BCHotkey BCSettings::EjectAllHotKey(const BCHotkey &defaultHotkey)
{
    QVariant resultKey = this->userSettings_.value(BC_CFG_EJECT_ALL_HOTKEY, QVariant());
    QVariant resultModifiers = this->userSettings_.value(BC_CFG_EJECT_ALL_HOTKEY_MODS, QVariant());

    bool ok1, ok2;

    if (resultKey.isValid() && resultModifiers.isValid())
    {
        BCHotkey hotkey((Qt::Key)resultKey.toLongLong(&ok1), (Qt::KeyboardModifiers)resultModifiers.toLongLong(&ok2));
        if (ok1 && ok2)
            return hotkey;
    }

    return BCHotkey(defaultHotkey);
}

void BCSettings::SetEjectAllHotKey(const BCHotkey &hotkey)
{
    CHECK_SAVE();
    this->userSettings_.setValue(BC_CFG_EJECT_ALL_HOTKEY, (qlonglong)hotkey.mainKey);
    this->userSettings_.setValue(BC_CFG_EJECT_ALL_HOTKEY_MODS, (qlonglong)hotkey.modifiers);
}

void BCSettings::SetEjectAllForced(bool forced)
{
    this->SaveVariant(BC_CFG_EJECT_ALL_FORCE, forced);
}

bool BCSettings::IsEjectAllForced()
{
    return this->userSettings_.value(BC_CFG_EJECT_ALL_FORCE, false).toBool();
}

void BCSettings::SetContainerListSaving(bool save)
{
    this->SaveVariant(BC_CFG_SAVE_CONTAINER_LIST, save);
}

bool BCSettings::IsContainerListSaving()
{
    return this->userSettings_.value(BC_CFG_SAVE_CONTAINER_LIST, true).toBool();
}

void BCSettings::SetMinimizeToTray(bool use)
{
    this->SaveVariant(BC_CFG_MINIMIZE_TO_TRAY, use);
}

bool BCSettings::IsMinimizingToTray()
{
    return this->userSettings_.value(BC_CFG_MINIMIZE_TO_TRAY, true).toBool();
}

bool BCSettings::MainFormIsMaximized()
{
    this->userSettings_.beginGroup(BC_CFG_MAINFORM_GROUP);
    bool result = this->userSettings_.value(BC_CFG_MAINFORM_MAXIMIZED, false).toBool();
    this->userSettings_.endGroup();
    return result;
}

void BCSettings::SetMainFormMaximized(bool maximized)
{
    this->SaveVariant(BC_CFG_MAINFORM_MAXIMIZED, maximized, BC_CFG_MAINFORM_GROUP);
}

QSize BCSettings::MainFormSize(const QSize& defaultSize)
{
    this->userSettings_.beginGroup(BC_CFG_MAINFORM_GROUP);
    QSize size = this->userSettings_.value(BC_CFG_MAINFORM_SIZE, defaultSize).toSize();
    this->userSettings_.endGroup();
    return size;
}

void BCSettings::SetMainFormSize(const QSize &size)
{
    this->SaveVariant(BC_CFG_MAINFORM_SIZE, size, BC_CFG_MAINFORM_GROUP);
}

QPoint BCSettings::MainFormLocation(const QPoint &defaultLocation)
{
    this->userSettings_.beginGroup(BC_CFG_MAINFORM_GROUP);
    QPoint location = this->userSettings_.value(BC_CFG_MAINFORM_LOCATION, defaultLocation).toPoint();
    this->userSettings_.endGroup();
    return location;
}

void BCSettings::SetMainFormLocation(const QPoint &location)
{
    this->SaveVariant(BC_CFG_MAINFORM_LOCATION, location);
}

bool BCSettings::IsConfigUsed()
{
    QVariant value = userSettings_.value(BC_CFG_USE_CONFIG, QVariant());
    if (value.isValid())
        return value.toBool();

    return true;
}

void BCSettings::SetConfigUsed(bool used)
{
    if (!used)
        this->Clear();

    userSettings_.setValue(BC_CFG_USE_CONFIG, used);
}

void BCSettings::Clear()
{
    this->userSettings_.clear();
}

void BCSettings::SaveVariant(const QString &elementName, const QVariant &val, const QString &group)
{
    CHECK_SAVE();
    if (group.length())
        this->userSettings_.beginGroup(group);

    this->userSettings_.setValue(elementName, val);

    if (group.length())
        this->userSettings_.endGroup();
}
