#include "mainwindow.h"
#include "ui_mainwindow.h"

#include "model/containermodel.h"
#include "model/bcoperation.h"
#include "model/hotkeylistener.h"
#include "model/devicemonitor.h"
#include "model/helper.h"
#include "entity/containerdescriptor.h"
#include "entity/settings.h"
#include "src/bctool.h"
#include "containerinfowidget.h"
#include "createdialog.h"
#include "hiddenpartdialog.h"
#include "settingsdialog.h"
#include "reencryptdialog.h"
#include "formatdialog.h"
#include "trayiconmanager.h"
#include "aboutdialog.h"
#include "logdialog.h"
#include "additionalactionswidget.h"
#include "container/exception.h"
#include "rawlinkdialog.h"

#include <QMessageBox>
#include <QFileDialog>
#include <QDesktopServices>
#include <QUrl>
#include <QIcon>
#include <QMotifStyle>
#include <QSystemTrayIcon>
#include <QSystemSemaphore>
#include <QCloseEvent>
#include <QTimer>

#include <stdlib.h>

#define BC_FILETYPES_AVAILABLE tr("BestCrypt Container Files (*.jbc);;All Files (*)")
#define SHOW_EXCEPTION(err_msg, e) BCOperationShowErrorString(QString(err_msg).arg(e.what()));

MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::MainWindow)
{
    ui->setupUi(this);

    try
    {
        if (!BCOperationCheckKernelInitialization())
        {
            ::exit(EXIT_FAILURE);
        }

        this->containerInfo_ = new ContainerInfoWidget(this->ui->informationWidget);
        this->hotkeyListener_ = new HotkeyListener();

        this->containerTree_ = this->ui->containerTree;
        this->mainToolbar_ = this->ui->mainToolBar;

        this->ui->informationWidget->layout()->addWidget(this->containerInfo_);

        // Set left side to minimum, not expandable size.
        this->ui->splitter->setStretchFactor(0, 0);
        this->ui->splitter->setStretchFactor(1, 1);

        this->containerModel_ = new ContainerModel();
        this->containerTree_->InitModel(this->containerModel_);
        this->trayIcon_ = new TrayIconManager(this->containerModel_);
        this->ProcessDescriptorChanges();

        connect(this->hotkeyListener_, SIGNAL(OnHotkeyPressed()), this, SLOT(EjectAllSlot()));
// containerTree_
        connect(this->containerTree_, SIGNAL(OnSelectedDescriptorChanged(ContainerDescriptor*)),
                this, SLOT(SelectedDescriptorChangedSlot(ContainerDescriptor*)));
        connect(this->containerTree_, SIGNAL(OnEject()), this, SLOT(EjectSlot()));
        connect(this->containerTree_, SIGNAL(OnMount()), this, SLOT(MountSlot()));
        connect(this->containerTree_, SIGNAL(OnDelete()), this, SLOT(DeleteSlot()));

        connect(this->containerInfo_, SIGNAL(OnBrowse()), this, SLOT(BrowseSlot()));
        connect(this->containerInfo_, SIGNAL(OnCreate()), this, SLOT(CreateSlot()));

// mainToolbar_
        connect(this->mainToolbar_, SIGNAL(OnCreate()), this, SLOT(CreateSlot()));
        connect(this->mainToolbar_, SIGNAL(OnMount()), this, SLOT(MountSlot()));
        connect(this->mainToolbar_, SIGNAL(OnBrowse()), this, SLOT(BrowseSlot()));
        connect(this->mainToolbar_, SIGNAL(OnEject()), this, SLOT(EjectSlot()));
        connect(this->mainToolbar_, SIGNAL(OnEjectForce()), this, SLOT(EjectForceSlot()));
        connect(this->mainToolbar_, SIGNAL(OnEjectAll()), this, SLOT(EjectAllSlot()));
// actions->File
        connect(this->ui->fileCreateAction, SIGNAL(triggered()), this, SLOT(CreateSlot()));
        connect(this->ui->fileLocateAction, SIGNAL(triggered()), this, SLOT(BrowseSlot()));
        connect(this->ui->fileSettingsAction, SIGNAL(triggered()), this, SLOT(ShowSettingsSlot()));
        connect(this->ui->fileExitAction, SIGNAL(triggered()), this, SLOT(ExitSlot()));
// actions->Container
        connect(this->ui->containerLinkAction, SIGNAL(triggered()), this, SLOT(LinkSlot()));
        connect(this->ui->containerBackupHeaderAction, SIGNAL(triggered()), this, SLOT(BackupHeaderSlot()));
        connect(this->ui->containerRestoreHeaderAction, SIGNAL(triggered()), this, SLOT(RestoreHeaderSlot()));
        connect(this->ui->containerCheckFilesystemAction, SIGNAL(triggered()), this, SLOT(FsckSlot()));
        connect(this->ui->containerDecryptHeaderAction, SIGNAL(triggered()), this, SLOT(DecryptHeaderSlot()));
        connect(this->ui->containerEncryptHeaderAction, SIGNAL(triggered()), this, SLOT(EncryptHeaderSlot()));
        connect(this->ui->containerFormatAction, SIGNAL(triggered()), this, SLOT(FormatSlot()));
        connect(this->ui->containerReencryptAction, SIGNAL(triggered()), this, SLOT(ReencryptSlot()));
        connect(this->ui->containerMountAction, SIGNAL(triggered()), this, SLOT(MountSlot()));
        connect(this->ui->containerEjectAction, SIGNAL(triggered()), this, SLOT(EjectSlot()));
        connect(this->ui->containerEjectForceAction, SIGNAL(triggered()), this, SLOT(EjectForceSlot()));
        connect(this->ui->containerEjectAllAction, SIGNAL(triggered()), this, SLOT(EjectAllSlot()));
// actions->Password
        connect(this->ui->passwordAddAction, SIGNAL(triggered()), this, SLOT(AddPasswordSlot()));
        connect(this->ui->passwordChangeAction, SIGNAL(triggered()), this, SLOT(ChangePasswordSlot()));
        connect(this->ui->passwordRemoveAction, SIGNAL(triggered()), this, SLOT(RemovePasswordSlot()));
// actions->Help
        connect(this->ui->helpAboutAction, SIGNAL(triggered()), this, SLOT(AboutSlot()));
        connect(this->ui->helpAboutQtAction, SIGNAL(triggered()), this, SLOT(AboutQtSlot()));
// trayIcon_
        connect(this->trayIcon_, SIGNAL(OnShowControlPanel()), this, SLOT(TrayIconShowControlPanelSlot()));
        connect(this->trayIcon_, SIGNAL(OnQuit()), this, SLOT(ExitSlot()));
        connect(this->trayIcon_, SIGNAL(OnDescriptorSelected(ContainerDescriptor*)),
                this, SLOT(TrayIconDescriptorSelectedSlot(ContainerDescriptor*)));
// additionalActions
        const AdditionalActionsWidget *additionalActions = this->containerInfo_->GetAdditionalActionsWidget();
        connect(additionalActions, SIGNAL(OnBackupHeader()), this, SLOT(BackupHeaderSlot()));
        connect(additionalActions, SIGNAL(OnRestoreHeader()), this, SLOT(RestoreHeaderSlot()));
        connect(additionalActions, SIGNAL(OnReencrypt()), this, SLOT(ReencryptSlot()));
        connect(additionalActions, SIGNAL(OnCreateHiddenPart()), this, SLOT(CreateHiddenSlot()));
        connect(additionalActions, SIGNAL(OnEncryptHeader()), this, SLOT(EncryptHeaderSlot()));
        connect(additionalActions, SIGNAL(OnDecryptHeader()), this, SLOT(DecryptHeaderSlot()));
        connect(additionalActions, SIGNAL(OnPasswordAdd()), this, SLOT(AddPasswordSlot()));
        connect(additionalActions, SIGNAL(OnPasswordRemove()), this, SLOT(RemovePasswordSlot()));
        connect(additionalActions, SIGNAL(OnPasswordChange()), this, SLOT(ChangePasswordSlot()));
        connect(additionalActions, SIGNAL(OnLink()), this, SLOT(LinkSlot()));
        connect(additionalActions, SIGNAL(OnFormat()), this, SLOT(FormatSlot()));
        connect(additionalActions, SIGNAL(OnFsck()), this, SLOT(FsckSlot()));
        connect(additionalActions, SIGNAL(OnShowLog()), this, SLOT(ShowLogSlot()));

        this->LoadPreviousViewSettings();
        this->UpdateHotkeyListener();
        this->trayIcon_->ShowIcon();
    }
    catch(BCException e)
    {
        BCOperationShowMessage(kBCTool_Error, e.what());
        ::exit(EXIT_FAILURE);
    }
}

/////////////////////
//      SLOTS       /
/////////////////////

void MainWindow::SelectedDescriptorChangedSlot(ContainerDescriptor *)
{
    this->ProcessDescriptorChanges();
}

void MainWindow::TrayIconDescriptorSelectedSlot(ContainerDescriptor *descriptor)
{
    if (this->containerTree_->SelectDescriptor(descriptor))
    {
        if (descriptor->IsLinked()) // If linked or mounted.
        {
            this->EjectSlot();
        }
        else
        {
            this->MountSlot();
        }
    }
    else
    {
        BCOperationShowErrorString(tr("Couldn't find descriptor."));
    }
}

void MainWindow::TrayIconShowControlPanelSlot()
{
    this->show();
    this->activateWindow();
    this->raise();
}

void MainWindow::BrowseSlot()
{
    try
    {
        QStringList selectedFiles = BCHelperLoadFiles(tr("Container Files Selection"),
                                        BC_FILETYPES_AVAILABLE, true);

        QStringList alreadyAddedFiles;

        for (int i = 0; i < selectedFiles.count(); i++)
        {
            ContainerDescriptor *descriptor = new ContainerDescriptor();
            descriptor->SetPath(selectedFiles.at(i));
            BCOperationUpdateDescriptor(descriptor);
            bool alreadyAdded = this->containerModel_->IndexOfContainer(descriptor->GetPath()) != -1;

            if (!alreadyAdded)
                this->containerModel_->Add(descriptor);
            else
                alreadyAddedFiles.append(selectedFiles.at(i));
        }

        int alreadyAddedCount = alreadyAddedFiles.count();

        // Show list of files that haven't been added.
        if (alreadyAddedCount)
        {
            // TODO: Select the first of the already added files.

        }
    }
    catch(BCException ex)
    {
        BCOperationShowMessage(kBCTool_Error, ex.what());
    }
}

void MainWindow::MountSlot()
{
    try
    {
        ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
        BCOperationMountContainer(descriptor);
        this->ProcessDescriptorChanges();
    }
    catch(BCException e)
    {
        //BCOperationShowErrorString(QString("Error while mounting: %0.").arg(e.what()));
        SHOW_EXCEPTION("Error while mounting: %0", e);
    }
}

void MainWindow::LinkSlot()
{
    try
    {
        BCOperationLinkContainer(this->containerTree_->GetSelectedDescriptor(), false);
        this->ProcessDescriptorChanges();
    }
    catch(BCException e)
    {
        SHOW_EXCEPTION("Error while linking: %0", e);
    }
}

void MainWindow::RawLinkSlot()
{
    // Select file to link.
    QStringList fileToRawLink = BCHelperLoadFiles(tr("Select file to raw link:"),
                                   "All files (*)", false);

    if (!fileToRawLink.count())
        return;

    RawLinkDialog dlg;
    RawLinkParameters p;
    p.file = fileToRawLink.at(0);

    dlg.SetParameters(&p);
    dlg.exec();
    if (dlg.result() == QDialog::Accepted)
    {
        // TODO: Linking read-only with random password is unuseful.
        BCOperationRawLinkFile(p.file, p.algName, p.modeName, false);
    }
}

void MainWindow::CreateSlot()
{
    bool dmIsListening = DeviceMonitor::GetInstance()->IsListening();
    if (dmIsListening)
        DeviceMonitor::GetInstance()->StopListening();
    try
    {
        bool needContinue = true;
        bool isContainerCreated = false;
        ContainerDescriptor *descriptor = 0;
        CreateDialog createDialog(this);        

        // Run non-modal dialog and process it's results here.
        // Similar construction is used in BCUser password-working methods.
        createDialog.setModal(true);
        createDialog.show();

        do
        {
            while (!createDialog.HaveResult())
            {
                QApplication::processEvents();
                BCHelperSleep(10);
            }

            needContinue = false;
            if (QDialog::Accepted == createDialog.GetResult())
            {
                descriptor = createDialog.GetData();

                QString containerPath = BCHelperSaveFile(QString(), tr("Enter Container File Name"),
                                                         BC_FILETYPES_AVAILABLE);
                if (!containerPath.count())
                {
                    needContinue = true;
                }
                else
                {
                    // If file exists, it will be destroyed at this point,
                    // but user has been asked a confirmation in the dialog.
                    BCHelperDeleteFileIfExists(containerPath);

                    descriptor->SetPath(containerPath);

                    QStringList errors;

                    if (!BCOperationCheckNewDescriptor(descriptor, &errors))
                    {
                        QString errorString = errors.join("\n");
                        BCOperationShowErrorString(tr("Please correct errors below and try again: \n%0").arg(errorString));
                        needContinue = true;
                    }
                    else
                    {
                        isContainerCreated = BCOperationCreateNewContainer(descriptor);
                    }
                }
            }

            if (needContinue)
            {
                createDialog.ResetResult();
                if (descriptor)
                {
                    delete descriptor;
                    descriptor = 0;
                }
            }
        }
        while (needContinue);

        if (isContainerCreated)
        {
            // Here we are after create was successful, so we have a valid descriptor.
            this->containerModel_->Add(descriptor);
            createDialog.close();
        }
        else if (descriptor)
        {
            delete descriptor;
        }
    }
    catch(BCException e)
    {
        SHOW_EXCEPTION("Error creating container: %0", e);
    }
    if (dmIsListening)
        DeviceMonitor::GetInstance()->StartListening();
}

void MainWindow::ShowSettingsSlot()
{
    SettingsDialog dlg;
    dlg.exec();
    this->UpdateHotkeyListener();
}

void MainWindow::EjectForceSlot()
{
    try
    {
        ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
        if (descriptor->IsMounted())
            BCOperationUmountContainter(descriptor, true);
        else if (descriptor->IsLinked())
            BCOperationUnlinkContainer(descriptor, true);

        this->ProcessDescriptorChanges();
    }
    catch(BCException e)
    {
        SHOW_EXCEPTION("Error ejecting: %0", e);
    }
}

void MainWindow::EjectSlot()
{
    try
    {
        ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
        if (!descriptor)
            return;

        if (descriptor->IsMounted())
        {
            BCOperationUmountContainter(descriptor, false);
            this->ProcessDescriptorChanges();
        }
        else if (descriptor->IsLinked())
        {
            BCOperationUnlinkContainer(descriptor, false);
            this->ProcessDescriptorChanges();
        }
    }
    catch (BCException e)
    {
        SHOW_EXCEPTION("Error ejecting: %0", e);
    }
}

void MainWindow::EjectAllSlot()
{
    BCOperationUmountAll(this->settings_.IsEjectAllForced());
    this->containerModel_->Update();
}

void MainWindow::ExitSlot()
{
    this->close();
}

void MainWindow::DeleteSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (!descriptor)
        return;
    if (BCOperationDeleteContainer(descriptor))
    {
        this->containerModel_->Remove(descriptor);
    }
}

void MainWindow::EncryptHeaderSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
        BCOperationEncryptHeader(descriptor);
    BCOperationUpdateDescriptor(descriptor);
    this->ProcessDescriptorChanges();
}

void MainWindow::DecryptHeaderSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
        BCOperationDecryptHeader(descriptor);
    BCOperationUpdateDescriptor(descriptor);
    this->ProcessDescriptorChanges();
}

void MainWindow::BackupHeaderSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
        BCOperationBackupHeader(descriptor);
}

void MainWindow::RestoreHeaderSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
        BCOperationRestoreHeader(descriptor);
    BCOperationUpdateDescriptor(descriptor);
}

void MainWindow::AboutSlot()
{
    AboutDialog about;
    about.exec();
}

void MainWindow::AboutQtSlot()
{
    QMessageBox::aboutQt(this);
}

void MainWindow::ReencryptSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (!descriptor)
        return;

    ReencryptDialog dlg(this);
    dlg.setModal(true);
    dlg.show();

    while (true)
    {
        while (!dlg.HaveResult())
        {
            BCHelperSleep(10);
            QApplication::processEvents();
        }

        if (dlg.GetResult() == QDialog::Accepted)
        {
            // We're here, if user clicked Ok.
            QString algName = dlg.GetAlgorythmName();
            QString modeName = dlg.GetModeName();
            if (!BCOperationReencryptContainer(descriptor, algName, modeName))
            {
                dlg.ResetResult();
                continue;
            }

            return;
        }

        // We're here, if user clicked Cancel.
        return;
    }
}

void MainWindow::CreateHiddenSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
    {
        BCOperationCreateHiddenPart(descriptor);
    }
}

void MainWindow::AddPasswordSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
    {
        BCOperationAddPassword(descriptor);
    }
}

void MainWindow::RemovePasswordSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
    {
        BCOperationRemovePassword(descriptor);
    }
}

void MainWindow::ChangePasswordSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    if (descriptor)
    {
        BCOperationChangePassword(descriptor);
    }
}

void MainWindow::FormatSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();

    if (!descriptor)
        return;

    FormatDialog dlg;
    dlg.setModal(true);
    dlg.show();

    while (true)
    {
        while(!dlg.HaveResult())
        {
            QApplication::processEvents();
            BCHelperSleep(10);
        }

        if (dlg.GetResult() == QDialog::Accepted)
        {
            QString fsType = dlg.GetFileSystem();
            BCOperationFormat(descriptor, fsType);
        }

        // We're here in case the user pressed Cancel.
        return;
    }
}

void MainWindow::FsckSlot()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();

    if (!descriptor)
        return;

    BCOperationFsck(descriptor, QString());
}

void MainWindow::ShowLogSlot()
{
    LogDialog log;
    log.exec();
}

MainWindow::~MainWindow()
{
    this->SaveViewSettings();
    delete this->hotkeyListener_;
    delete ui;
}

void MainWindow::UpdateHotkeyListener()
{
    bool result = true;

#define CHECK_RESULT(msg) if (!result) { BCOperationShowMessage(kBCTool_Error, msg); return; }

    if (this->settings_.IsEjectAllHotKeyEnabled())
    {
        if (this->hotkeyListener_->IsListening())
            result = this->hotkeyListener_->StopListening();

        CHECK_RESULT("Error while stopping global hotkey listener.");

        BCHotkey hotkey = this->settings_.EjectAllHotKey();
        result = this->hotkeyListener_->SetHotkey(hotkey);

        CHECK_RESULT("Error while setting global hotkey.");

        result = this->hotkeyListener_->StartListening();

        CHECK_RESULT("Error while starting global hotkey listening.");
    }

#undef CHECK_RESULT
}

/////////////////////////////////////////////
//      Window Events and View Settings     /
/////////////////////////////////////////////

void MainWindow::closeEvent(QCloseEvent * /*event*/)
{
    this->SaveViewSettings();
    qApp->exit(0);
}

void MainWindow::changeEvent(QEvent *e)
{
    if (e->type() == QEvent::WindowStateChange)
    {
        if (this->isMinimized() && this->settings_.IsMinimizingToTray())
        {
            this->hide();
            e->ignore();
        }
    }
    QMainWindow::changeEvent(e);
}

void MainWindow::ProcessDescriptorChanges()
{
    ContainerDescriptor *descriptor = this->containerTree_->GetSelectedDescriptor();
    this->containerInfo_->SetData(descriptor); // Update descriptor information in right panel.
    this->containerTree_->UpdateCurrentView();
    QAction *containerSelectedActions[] =
    {
        this->ui->containerBackupHeaderAction,
        this->ui->containerRestoreHeaderAction,
        this->ui->containerEncryptHeaderAction,
        this->ui->containerDecryptHeaderAction,
//        this->ui->containerLinkAction,
//        this->ui->containerMountAction,
//        this->ui->containerEjectAction,
//        this->ui->containerEjectForceAction,
        this->ui->containerFormatAction,
        this->ui->containerCheckFilesystemAction,
        this->ui->containerReencryptAction,

        this->ui->passwordAddAction,
        this->ui->passwordChangeAction,
        this->ui->passwordRemoveAction
    };

    int containerSelectedActionCount = 10, i;

    bool isLinked = false;

    // Update button state.
    if (descriptor)
    {
        isLinked = descriptor->IsLinked();
    }

    for (i = 0; i < containerSelectedActionCount; ++i)
    {
        containerSelectedActions[i]->setEnabled(!!descriptor);
    }

    // Do not allow mounting for linked containers here.
    this->mainToolbar_->SetMountEnabled(descriptor && !isLinked);
    this->mainToolbar_->SetEjectEnabled(descriptor && isLinked);

    this->ui->containerLinkAction->setEnabled(descriptor && !isLinked);
    this->ui->containerMountAction->setEnabled(descriptor && !isLinked);
    this->ui->containerEjectAction->setEnabled(descriptor && isLinked);
    this->ui->containerEjectForceAction->setEnabled(descriptor && isLinked);
    this->ui->containerEncryptHeaderAction->setEnabled(descriptor && descriptor->IsParsed());
    this->ui->containerDecryptHeaderAction->setEnabled(descriptor && !descriptor->IsParsed());
    this->containerTree_->setVisible(this->containerTree_->GetSelectedDescriptor());

    this->trayIcon_->UpdateMenu();
}

void MainWindow::LoadPreviousViewSettings()
{
    bool isMaximized;
    isMaximized = this->settings_.MainFormIsMaximized();
    if (isMaximized)
    {
        this->showMaximized();
        return; // Do not load size in this case,
                // because it will be size of maximized window.
    }
    else
        this->showNormal();

    QSize size = this->size();
    QPoint point = this->pos();

    size = this->settings_.MainFormSize(size);
    point = this->settings_.MainFormLocation(point);

    // Set minimum position, so window border can't be upper than zero point.
    if (point.x() < 70)
        point.setX(70);
    if (point.y() < 70)
        point.setY(70);

    this->setGeometry(point.x(), point.y(), size.width(), size.height());
}

void MainWindow::SaveViewSettings()
{
    this->settings_.SetMainFormLocation(this->pos());
    this->settings_.SetMainFormMaximized(this->isMaximized());
    this->settings_.SetMainFormSize(this->size());
}
