#include "bcoperation.h"

#include "container/container.h"
#include "container/keygen.h"
#include "container/exception.h"
#include "kernel/BCKernel.h"
#include "entity/containerdescriptor.h"
#include "src/version.h"
#include "disk/disk.h"
#include "model/helper.h"

#include <signal.h>
#include <stdio.h>
#include <mntent.h>
#include <sys/stat.h>
#include <string.h>
#include <errno.h>
#include <limits.h>
#include <semaphore.h>
#include <fcntl.h>
#include <execinfo.h>
#include <glob.h>

#include <QMessageBox>
#include <QStyle>
#include <QUrl>
#include <QFileDialog>
#include <QMutex>
#include <QWaitCondition>
#include <QProgressDialog>
#include <QDesktopServices>
#include <QApplication>

#define BC_BACKUP_FILENAMES "Backup files (*.bak);;BestCrypt Containers (*.jbc);; All Files (*)"


static void HandleSignal(int signal)
{
    qDebug("HandleSignal: signal #%i.", signal);
}


static int ReportProgress(const char *message, bc_u64 byteCount, bc_u64 bytesTotal, void */*arg*/)
{
    static QProgressDialog * progressDialog = 0;

    int percent = 0, percentMax = 100;
    if (bytesTotal != 0)
        percent = (int)(byteCount * 100LL / bytesTotal);

    // If progress dialog is already exists, change it's parameters.
    if (!progressDialog)
    {
        // Create a new progress dialog.
        QWidget *activeWindow = QApplication::activeWindow();
        progressDialog = new QProgressDialog(message, "Cancel", 0,
                                             percentMax, activeWindow);
        progressDialog->setWindowModality(Qt::WindowModal);
        progressDialog->setWindowTitle("BestCrypt");
    }

    progressDialog->setLabelText(message);
    progressDialog->setValue(percent);

    progressDialog->update();
    QApplication::processEvents();

    if (progressDialog->wasCanceled())
    {
        delete progressDialog;
        progressDialog = 0;
        return false;
    }
    if (byteCount != bytesTotal)
        return true;

    delete progressDialog;
    progressDialog = 0;

    return true;
}


static BCTool *CreateUI()
{
    BCTool *ui = new BCTool(BCOperationShowMessage, BCOperationAskUser, HandleSignal, ReportProgress);
    return ui;
}


bool BCOperationAskUser(const char *msg, bool defaultAnswer)
{
    QMessageBox msgBox;
    msgBox.setStandardButtons(QMessageBox::Yes | QMessageBox::No);
    msgBox.setText(msg);
    msgBox.setIcon(QMessageBox::Question);
    msgBox.setDefaultButton(defaultAnswer? QMessageBox::Yes : QMessageBox::No);
    msgBox.setWindowTitle("BestCrypt Control Panel");
    return msgBox.exec() == QMessageBox::Yes;
}


const char *BCOperationCreateMessage(const char *msg, ...);


static void UpdateMountState(ContainerDescriptor *descriptor)
{
    if (!descriptor->IsExists())
    {
        descriptor->Unlink();
        return;
    }
    BCTool *ui = CreateUI();
    std::vector<bc_mount_entry> mountEntries;
    bool succeeded = ui->GetMountInfo(&mountEntries);
    delete ui;

    if (!succeeded)
    {
        // Error had already been showed to user inside previous function call.
        return;
    }

    std::vector<bc_mount_entry>::iterator i = mountEntries.begin();
    std::vector<bc_mount_entry>::iterator end = mountEntries.end();
    std::string descriptorPath = descriptor->GetPath().toStdString();
    char descriptorAbsPath[PATH_MAX], tmpAbsPath[PATH_MAX];

    if (!realpath(descriptorPath.c_str(), descriptorAbsPath))
    {
        int error = errno;
        BCOperationShowErrorString(QString("Error while resolving path %0: %1").arg(descriptorPath.c_str()).arg(strerror(error)));
        return;
    }

    for (; i != end; i++)
    {
        bc_mount_entry entry = *i;

        if (!entry.isFileExists)
            continue;

        if (!realpath(entry.containerPath, tmpAbsPath))
        {
            int error = errno;
            BCOperationShowErrorString(QString("Error while resolving path %0: %0").arg(entry.containerPath).arg(strerror(error)));
            return;
        }
        if (0 == strcmp(descriptorAbsPath, tmpAbsPath))
        {
            // Found container linked to device.
            if (strlen(entry.mountPoint))
            {
                // This container is also mounted, not only linked.
                descriptor->Mount(entry.deviceName, entry.mountPoint);
                return;
            }
            else
            {
                // This container is linked, but not mounted.
                descriptor->Link(entry.deviceName);
                return;
            }
        }
    }

    // All mount entries had been checked, but no such file found.
    descriptor->Unlink();
    return;
}

/*
  ***********************************************
                Interface
  ***********************************************
*/

bool BCOperationCheckKernelInitialization()
{
    BCTool *ui = CreateUI();
    bool result = ui->CheckKernelInitialization();
    delete ui;
    return result;
}

const char *BCOperationCreateMessage(const char *msg, ...)
{
    const int MaxLen = 4096;
    static char message[MaxLen];
    va_list list;
    va_start(list, msg);
    vsnprintf(message, MaxLen - 1, msg, list);
    va_end(list);
    message[MaxLen - 1] = '\0';
    return message;
}


void BCOperationShowErrorString(const QString &str)
{
    BCOperationShowMessageString(kBCTool_Error, str);
}


void BCOperationShowWarningString(const QString &str)
{
    BCOperationShowMessageString(kBCTool_Warning, str);
}


void BCOperationShowMessage(bctool_msg_level level, const char* msg)
{
    BCOperationShowMessageString(level, msg);
}


void BCOperationShowMessageString(bctool_msg_level level, const QString &msg)
{
    QString caption;
    QMessageBox::Icon icon;

    switch (level)
    {
        case kBCTool_Info:
            icon = QMessageBox::Information;
            caption = "Information";
            break;

        case kBCTool_Warning:
            icon = QMessageBox::Warning;
            caption = "Warning";
            break;

        case kBCTool_Error:
            icon = QMessageBox::Critical;
            caption = "Error";
            break;

        case kBCTool_Debug:
            // TODO: Ignore debug messages here.
            break;

        default:
            Q_ASSERT(0);
    }

    QMessageBox msgBox;
    msgBox.setText(msg);
    msgBox.setWindowTitle(caption);
    msgBox.setIcon(icon);
    msgBox.exec();
}


bool BCOperationCheckNewDescriptor(ContainerDescriptor *descriptor, QStringList *out_errors)
{
    Q_ASSERT(0 != out_errors);

    QFileInfo fileInfo(descriptor->GetPath());
    //QFile file(descriptor->GetPath());
    std::string containerPath = descriptor->GetPath().toStdString();
    bool result = true;

    if (fileInfo.exists())
    {
        QMessageBox msgBox(QMessageBox::Warning, "File already exists", "File with same name already exists. Are you sure you want to replace this file?", QMessageBox::Ok | QMessageBox::Cancel);
        if (msgBox.exec() != QMessageBox::Ok)
        {
            out_errors->append(QObject::tr("* File with same name is already exists in directory."));
            return false;
        }
        // else clean file.
    }

    // Try to create file in specified location.
    FILE *f = fopen(containerPath.c_str(), "w");
    if (!f)
    {
        int error = errno;
        out_errors->append(QObject::tr("* Cannot create file '%0': %1")
                           .arg(descriptor->GetPath()).arg(strerror(error)));
        result = false;
    }
    else
    {
        fclose(f);
        remove(containerPath.c_str());
    }

    if (descriptor->description.count() >= BC_DESCRIPTION_SIZE)
    {
        out_errors->append(QObject::tr("* Description is too long (%0 symbols). Maximum length is %1.")
                           .arg(descriptor->description.count()).arg(BC_DESCRIPTION_SIZE));
        result = false;
    }

    return result;
}

void BCOperationUpdateDescriptor(ContainerDescriptor *descriptor)
{
    Q_ASSERT(descriptor);
    BCVolumeInfo containerInfo;
    BCAlgInfo algInfo;
    bc_kern_result kernelResult;

    std::string tmpPath(descriptor->GetPath().toStdString());

    if (descriptor->IsExists() && BCParseContainerFile(tmpPath.c_str(), &containerInfo))
    {
        kernelResult = BCKernelLookupAlgId(containerInfo.alg, &algInfo);

        if (kBCKernel_Success != kernelResult)
            throw BCException(BCOperationCreateMessage("Error while getting algorithm name: %s", BCKernelGetErrorString(kernelResult)));

        descriptor->algId = algInfo.id;
        descriptor->size = containerInfo.size;
        descriptor->algName = QString(algInfo.name);
        descriptor->modeId = containerInfo.mode;
        descriptor->modeName = QString(BCKernelModeName(containerInfo.mode));
        descriptor->description = QString(containerInfo.description);
        descriptor->SetParsed(true);
    }
    else
    {
        descriptor->SetParsed(false);
    }

    // Set descriptor current mount state.
    UpdateMountState(descriptor);
}


bool BCOperationEncryptHeader(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->ProcessHeader(path.c_str(), true);
    delete ui;
    return result;
}


bool BCOperationDecryptHeader(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->ProcessHeader(path.c_str(), false);
    delete ui;
    return result;
}


bool BCOperationBackupHeader(ContainerDescriptor *descriptor)
{
    QString backupFile = BCHelperSaveFile(descriptor->GetFileName(), QObject::tr("Backup file name:"),
                       BC_BACKUP_FILENAMES);
    if (!backupFile.length())
        return false;
    std::string backupStr = backupFile.toStdString();
    std::string path = descriptor->GetPath().toStdString();

    BCTool *ui = CreateUI();
    bool result = ui->BackupHeader(path.c_str(), backupStr.c_str());
    delete ui;
    return result;
}


bool BCOperationRestoreHeader(ContainerDescriptor *descriptor)
{
    QStringList backupFiles = BCHelperLoadFiles(QObject::tr("Select backup file:"),
                                           BC_BACKUP_FILENAMES, false);
    if (!backupFiles.count())
        return false;
    QString backupFile = backupFiles.at(0);
    std::string backup = backupFile.toStdString();
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->RestoreHeader(path.c_str(), backup.c_str());
    return result;
}


bool BCOperationCreateNewContainer(ContainerDescriptor *descriptor)
{
    std::string fileName = descriptor->GetPath().toStdString();
    std::string algName = descriptor->algName.toStdString();
    std::string modeName = descriptor->modeName.toStdString();
    std::string description = descriptor->description.toStdString();
    std::string fileSystem = descriptor->fileSystem.toStdString();
    bc_u64 size = descriptor->size;

    BCTool *ui = CreateUI();
    bool succeeded = ui->New(fileName.c_str(), algName.c_str(), modeName.c_str(),
                             size, description.c_str(), fileSystem.c_str());
    delete ui;

    BCOperationUpdateDescriptor(descriptor);

    return succeeded;
}

static bool DoMount(ContainerDescriptor *descriptor, const QString& mountPoint,
                    const QString& fsType, int uid, int gid,
                    bool readOnly, const QString& mountOptions)
{

    std::string pathStr(descriptor->GetPath().toStdString());
    std::string mountPointStr(mountPoint.toStdString());
    std::string fsTypeStr(fsType.toStdString());
    std::string mountOptionsStr(mountOptions.toStdString());
    char deviceName[PATH_MAX] = { 0 };
    BCMountParams mountParams;
    const char *mountOptionsBuffer = 0, *fsTypeBuffer = 0;;

    mountParams.mountReadOnly = readOnly;
    strncpy(mountParams.mountPoint, mountPointStr.c_str(), BC_MAX_PATH);
    mountParams.useDefaultMountPoint = false;

    if (fsType.count())
        fsTypeBuffer = fsTypeStr.c_str();

    if (mountOptions.count())
        mountOptionsBuffer = mountOptionsStr.c_str();

    BCTool *ui = CreateUI();
    bool isMounted = ui->Mount(pathStr.c_str(), &mountParams,
                               fsTypeBuffer, uid, gid,
                               mountOptionsBuffer, deviceName);

    delete ui;

    if (isMounted)
    {
        descriptor->lastMountPoint = QString(mountPoint);
    }
    UpdateMountState(descriptor);

    return isMounted;
}

bool BCOperationMountContainer(ContainerDescriptor *descriptor)
{
    BCOperationUpdateDescriptor(descriptor);
    if (descriptor->IsMounted())
        return true;

    if (!descriptor)
    {
        qDebug("Descriptor is null in BCOperationMountContainer().");
        return false;
    }
    QString mountLocation = descriptor->lastMountPoint;
    if (!mountLocation.count())
    {
        mountLocation = BCHelperGetLocation(QObject::tr("Mount Location for %0").arg(descriptor->GetFileName()), QString());
        if (!mountLocation.count())
        {
            // User canceled.
            return false;
        }
    }
    bool readOnly = descriptor->mountReadOnly;
    if (DoMount(descriptor, mountLocation, QString(), 0, 0, readOnly, QString()))
    {
        //QDesktopServices::openUrl(QUrl("file:///" + mountLocation));
        return true;
    }
    return false;
}

bool BCOperationLinkContainer(ContainerDescriptor *descriptor, bool readOnly)
{
    BCOperationUpdateDescriptor(descriptor);

    if (descriptor->IsLinked())
        return true;

    readOnly |= descriptor->mountReadOnly;
    std::string containerPath = descriptor->GetPath().toStdString();

    BCTool *ui = CreateUI();
    bool succeeded = ui->Link(containerPath.c_str(), readOnly, NULL, false, NULL, NULL, NULL);
    delete ui;

    UpdateMountState(descriptor);

    return succeeded;
}

bool BCOperationRawLinkFile(const QString& file, const QString& algName,
                            const QString& modeName, bool readOnly)
{
    std::string fileStr = file.toStdString();
    std::string algStr = algName.toStdString();
    std::string modeNameStr = modeName.toStdString();
    BCTool *ui = CreateUI();
    bool succeeded = ui->RawLink(fileStr.c_str(), algStr.c_str(), modeNameStr.c_str(),
                                 readOnly);
    delete ui;
    return succeeded;
}

bool BCOperationUmountContainter(ContainerDescriptor *descriptor, bool force)
{
    BCOperationUpdateDescriptor(descriptor);
    if (!descriptor->IsMounted())
        return true;
    std::string mountPointStr = descriptor->GetMountPoint().toStdString();

    Q_ASSERT(mountPointStr.length());

    BCTool *ui = CreateUI();
    bool result = ui->Umount(mountPointStr.c_str(), force); // Already showed all errors inside.
    delete ui;

    UpdateMountState(descriptor);

    return result;
}

void BCOperationUmountAll(bool force)
{
    BCTool *ui = CreateUI();
    int totalMounted, unmounted;
    (void) ui->UmountAll(force, &totalMounted, &unmounted);
    delete ui;
    // All detailed error messages are already shown inside.
    if (!totalMounted)
    {
        BCOperationShowMessage(kBCTool_Info, "There are no containers to process.");
    }
    else if (totalMounted != unmounted)
    {
        BCOperationShowErrorString("Some of your containers are still unmounted.\nIf you really need to remove them right now, please reboot your computer.");
    }
    else
    {
        BCOperationShowMessage(kBCTool_Info, "All containers are unmounted successfully.");
    }
}

bool BCOperationUnlinkContainer(ContainerDescriptor *descriptor, bool force)
{
    BCOperationUpdateDescriptor(descriptor);
    if (!descriptor->IsLinked())
        return true;

    std::string deviceName = descriptor->GetDeviceName().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->Unlink(deviceName.c_str(), force);
    delete ui;
    UpdateMountState(descriptor);
    return result;
}


bool BCOperationGetAlgList(QList<BCAlgInfo> *out_algList)
{
    Q_ASSERT(out_algList);

    BCAlgInfo *algs = NULL;
    unsigned algsTotal;
    bc_kern_result result = kBCKernel_Success;

    result = BCKernelListAlgs(&algs, &algsTotal);
    if (kBCKernel_Success != result)
    {
        if (kBCKernel_InternalError != result)
        {
            BCOperationShowErrorString(QObject::tr("Couldn't load list of supported ciphers: %0").arg(BCKernelGetErrorString(result)));
        }
        else
        {
            int error = errno;
            BCOperationShowErrorString(QObject::tr("Couldn't load list of supported ciphers: %0").arg(strerror(error)));
        }

        return false;
    }

    for (unsigned i = 0; i < algsTotal; i++)
    {
        out_algList->append(algs[i]);
    }

    BCKernelFreeAlgList(algs);
    return true;
}

bool BCOperationIsModeSupported(const BCAlgInfo& alg, bc_mode_id modeId)
{
    return (bool)BCKernelIsModeSupported(&alg, modeId);
}

QString BCOperationGetModeName(bc_mode_id modeId)
{
    return QString(BCKernelModeName(modeId));
}

bc_mode_id BCOperationGetModeId(QString modeName)
{
    std::string modeStr = modeName.toStdString();
    return BCKernelModeId(modeStr.c_str());
}

bool BCOperationGetDescriptorsFromDevices(QList<ContainerDescriptor*>* descriptors)
{
    BCTool *ui = CreateUI();
    std::vector<bc_mount_entry> mountInfo;
    bool result = ui->GetMountInfo(&mountInfo);
    delete ui;

    if (result)
    {
        std::vector<bc_mount_entry>::iterator i = mountInfo.begin();
        for (; i != mountInfo.end(); i++)
        {
            bc_mount_entry entry = *i;
            ContainerDescriptor *descriptor = new ContainerDescriptor();
            descriptor->SetPath(entry.containerPath);
            BCOperationUpdateDescriptor(descriptor);
            descriptors->append(descriptor);
        }
    }

    return result;
}

bool BCOperationReencryptContainer(ContainerDescriptor *descriptor, QString algName, QString modeName)
{
    BCTool *ui = CreateUI();
    std::string algNameStr = algName.toStdString();
    std::string modeNameStr = modeName.toStdString();
    std::string path = descriptor->GetPath().toStdString();

    bool result = ui->Reencrypt(path.c_str(), algNameStr.c_str(), modeNameStr.c_str());

    delete ui;
    BCOperationUpdateDescriptor(descriptor);
    return result;
}

bool BCOperationCreateHiddenPart(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->CreateHiddenPart(path.c_str(), (bc_u64)0, NULL);
    delete ui;
    return result;
}


bool BCOperationAddPassword(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();

    BCTool *ui = CreateUI();
    bool result = ui->Passwd(path.c_str(), kBCKey_Duplicate);
    delete ui;

    return result;
}


bool BCOperationRemovePassword(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->Passwd(path.c_str(), kBCKey_Delete);
    delete ui;

    return result;
}


bool BCOperationChangePassword(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->Passwd(path.c_str(), kBCKey_Recode);
    delete ui;
    return result;
}


bool BCOperationDeleteContainer(ContainerDescriptor *descriptor)
{
    std::string path = descriptor->GetPath().toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->DeleteContainer(path.c_str());
    delete ui;
    BCOperationUpdateDescriptor(descriptor);
    return result;
}


bool BCOperationFormat(ContainerDescriptor *descriptor, const QString& fsType)
{
    std::string path = descriptor->GetPath().toStdString();
    std::string fsTypeStr = fsType.toStdString();

    BCTool *ui = CreateUI();
    bool result = ui->Format(path.c_str(), fsTypeStr.c_str(), true);
    delete ui;
    return result;
}


bool BCOperationFsck(ContainerDescriptor *descriptor, const QString& fsType_)
{
    std::string path = descriptor->GetPath().toStdString();
    std::string fsType = fsType_.toStdString();
    BCTool *ui = CreateUI();
    bool result = ui->CheckFS(path.c_str(), fsType.c_str());
    delete ui;
    return result;
}
