/**
  Contains functions used for working with OS and BestCrypt itself, independently from GUI.
*/

#ifndef BCOPERATION_H
#define BCOPERATION_H

#include <QString>
#include <QList>

class ContainerDescriptor;
class QIcon;

#include "kernel/BCKernel.h"
#include "container/intsize.h"
#include "src/bctool.h"

/**
  @brief Checks that BestCrypt kernel modules are loaded.
*/
bool BCOperationCheckKernelInitialization();

/**
  @brief Checks that parameters of new container are valid.
*/
bool BCOperationCheckNewDescriptor(ContainerDescriptor *descriptor, QStringList *out_errors);

/**
  @brief Fills specified list with encryption algorithms supported by BC kernel.
*/
bool BCOperationGetAlgList(QList<BCAlgInfo> *out_algList);

/**
  @brief Checks that specified algorithm supports specified mode.
*/
bool BCOperationIsModeSupported(const BCAlgInfo& alg, bc_mode_id modeId);

/**
  @brief Gets encryption mode name by mode id.
*/
QString BCOperationGetModeName(bc_mode_id modeId);

/**
  @brief Gets encryption mode id by using it's name.
*/
bc_mode_id BCOperationGetModeId(QString modeName);

/**
  @brief Fills specified list with descriptors for currently mounted containers.
*/
bool BCOperationGetDescriptorsFromDevices(QList<ContainerDescriptor*>* descriptors);

/*
  ************************************************
            GUI Helpers
  ************************************************
*/

/**
  @brief Operates like sprintf(), but uses internal static buffer and returns a pointer to it.
  Result is not needed to be freed.
*/
const char* BCOperationCreateString(const char *msg, ...);

/**
  @brief Asks user an yes/no question and returns the answer.
*/
bool BCOperationAskUser(const char *msg, bool defaultAnswer);

/**
  @brief Shows message of specified level.
*/
void BCOperationShowMessage(bctool_msg_level level, const char* msg);

/**
  @brief Shows message of specified level using QString parameter.
*/
void BCOperationShowMessageString(bctool_msg_level level, const QString& msg);

/**
  @brief Shows error report to user using QString parameter.
*/
void BCOperationShowErrorString(const QString& str);

/**
  @brief Shows warning to user using QString parameter.
*/
void BCOperationShowWarningString(const QString& msg);


/*
  ************************************************
                Container Operations
  ************************************************
*/

bool BCOperationEncryptHeader(ContainerDescriptor *descriptor);

bool BCOperationDecryptHeader(ContainerDescriptor *descriptor);

bool BCOperationBackupHeader(ContainerDescriptor *descriptor);

bool BCOperationRestoreHeader(ContainerDescriptor *descriptor);

/**
  @brief Reencrypts container with specified alg and mode.
  If alg and mode are the same as current, warning will be shown and
  `false` will be returned.
*/
bool BCOperationReencryptContainer(ContainerDescriptor *descriptor, QString algName, QString modeName);

bool BCOperationCreateHiddenPart(ContainerDescriptor *descriptor);

bool BCOperationAddPassword(ContainerDescriptor *descriptor);

bool BCOperationRemovePassword(ContainerDescriptor *descriptor);

bool BCOperationChangePassword(ContainerDescriptor *descriptor);

bool BCOperationDeleteContainer(ContainerDescriptor *descriptor);

/**
  @brief Links the container to a device and creates a new file system on it.
*/
bool BCOperationFormat(ContainerDescriptor *descriptor, const QString& fsType);

/**
  @brief Links the container to a device and checks it's file system.
*/
bool BCOperationFsck(ContainerDescriptor *descriptor, const QString &fsType);

/**
  @brief Creates new container with parameters specified in descriptor.
*/
bool BCOperationCreateNewContainer(ContainerDescriptor *descriptor);

/**
  @brief Mounts specified container file to mount point specified in descriptor.
*/
bool BCOperationMountContainer(ContainerDescriptor *descriptor);

/**
  @brief Links specified container to device without mounting.
*/
bool BCOperationLinkContainer(ContainerDescriptor *descriptor, bool readOnly);

/**
  @brief Does raw linking of specified file to some BestCrypt device.
        After this operation all data in file will be destroyed.
*/
bool BCOperationRawLinkFile(const QString& file, const QString& algName,
                            const QString& modeName, bool readOnly);

/**
  @brief Unmounts specified container file.
*/
bool BCOperationUmountContainter(ContainerDescriptor *descriptor, bool force);

/**
  @brief Unlinks specified container from device.
*/
bool BCOperationUnlinkContainer(ContainerDescriptor *descriptor, bool force);

/**
  @brief Updates container descriptor by file path.
  @param descriptor     Descriptor of the container to process.
*/
void BCOperationUpdateDescriptor(ContainerDescriptor *descriptor);

/**
  @brief Reencrypts container by new key.
  @param descriptor     Descriptor of the container to process.
*/
bool BCOperationReencryptContainer(ContainerDescriptor *descriptor, QString algName, QString modeName);

/**
  @brief Ejects all mounted containers.
*/
void BCOperationUmountAll(bool force);


#endif // BCOPERATION_H
