#include "containermodel.h"
#include "entity/containerdescriptor.h"
#include "container/exception.h"
#include "entity/settings.h"
#include "bcoperation.h"
#include "devicemonitor.h"

#include <stdio.h>

ContainerModel::ContainerModel(QObject *parent)
    :QObject(parent)
{
    this->ignoreEvents = false;
    this->descriptors_ = new QList<ContainerDescriptor *>();
    this->LoadFromConfig();
    this->LoadFromDevices();

    // Subscribe to device events in system.
    connect(DeviceMonitor::GetInstance(), SIGNAL(OnDiskEvent()), this, SLOT(DiskEventSlot()));
    DeviceMonitor::GetInstance()->StartListening();
}

ContainerModel::~ContainerModel()
{
    disconnect(DeviceMonitor::GetInstance(), SIGNAL(OnDiskEvent()), this, SLOT(DiskEventSlot()));
    qDeleteAll(*(this->descriptors_));
    delete this->descriptors_;
}

void ContainerModel::Add(ContainerDescriptor *descriptor)
{
    this->InternalAdd(descriptor);
    EmitDataChanged();
}

void ContainerModel::Remove(ContainerDescriptor *descriptor)
{
    int index = this->IndexOfContainer(descriptor->GetPath());
    if (-1 != index)
    {
        this->descriptors_->removeAt(index);
        EmitDataChanged();
    }
}

int ContainerModel::IndexOfContainer(const QString& fullPath)
{
    ContainerDescriptor *cur = 0;
    for (int i=0; i<this->descriptors_->count(); i++)
    {
        cur = this->descriptors_->at(i);
        if (cur->GetPath() == fullPath)
        {
            return i;
        }
    }

    return -1;
}

void ContainerModel::Update()
{
    for (int i = 0; i < this->descriptors_->count(); i++)
    {
        try
        {
            BCOperationUpdateDescriptor(this->descriptors_->at(i));
        }
        catch (BCException ex)
        {
            BCOperationShowErrorString(QString("Error while updating descriptor for %0: %1").arg(this->descriptors_->at(i)->GetPath()).arg(ex.what()));
        }
    }
    EmitDataChanged();
}

void ContainerModel::Clear()
{
    this->descriptors_->clear();
    EmitDataChanged();
}

void ContainerModel::LoadFromConfig()
{
    BCSettings settings;
    QList<ContainerDescriptor *> *descriptorList = settings.LoadContainerList();

    this->LoadFromList(descriptorList);
    this->Update();

    delete descriptorList; // Descriptors will be deleted with model.
}

void ContainerModel::LoadFromDevices()
{
    QList<ContainerDescriptor *> descriptors;
    if (BCOperationGetDescriptorsFromDevices(&descriptors))
    {
        this->LoadFromList(&descriptors);
    }
}

void ContainerModel::SaveToConfig()
{
    BCSettings settings;
    settings.SaveContainerList(this->descriptors_);
}

void ContainerModel::InternalAdd(ContainerDescriptor *descriptor)
{
    // Add only containers that are not in model yet.
    int index = this->IndexOfContainer(descriptor->GetPath());
    if (-1 == index)
    {
        this->descriptors_->append(descriptor);
    }
}

void ContainerModel::EmitDataChanged()
{
    if (!this->ignoreEvents)
        emit this->OnDataChanged();
}

void ContainerModel::DiskEventSlot()
{
    fprintf(stderr, "Disk event came, reloading list of devices...\n");
    bool isEventsIgnored = this->ignoreEvents;
    this->ignoreEvents = false;

    // Load mounted descriptors.
    this->LoadFromDevices();

    // Update present descriptors.
    this->Update();

    this->ignoreEvents = isEventsIgnored;
    EmitDataChanged();
}

void ContainerModel::LoadFromList(QList<ContainerDescriptor *> *descriptors)
{
    if (!descriptors)
        return;

    if (!descriptors->count())
        return;

    ContainerDescriptor *descriptor = 0;
    for (int i = 0; i < descriptors->count(); i++)
    {
        descriptor = descriptors->at(i);
        // Add only existing containers.
        if (descriptor->IsExists())
            this->InternalAdd(descriptors->at(i));
        else
        {
            // Remove descriptor from list and from memory.
            descriptors->removeAt(i);
            delete descriptor;
        }
    }

    EmitDataChanged();
}
