#include "hotkeylistener.h"

#include <QApplication>
#include <QX11Info>
#include <QMutex>
#include <QKeyEvent>
#include <QWaitCondition>

#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>
#include <X11/Xproto.h>

// Fix Qt vs X11 name conflicts.
#ifdef KeyPress
    const int XKeyPress = KeyPress;
    const int XKeyRelease = KeyRelease;
    #undef KeyPress
    #undef KeyRelease
#endif

// Define usual names for X modifiers.
#define AltMask         Mod1Mask
#define NumLockMask     Mod2Mask
#define ScrollLockMask  Mod5Mask
#define CapsLockMask    LockMask

static bool g_XHasError;

static int GetNativeKey(Qt::Key key)
{
    QKeySequence sequence(key);
    qDebug("GetNativeKey: native keycode for %s", sequence.toString().toLatin1().data());
    return XKeysymToKeycode(QX11Info::display(), XStringToKeysym(QKeySequence(key).toString().toLatin1().data()));
}

static int GetNativeModifiers(Qt::KeyboardModifiers modifiers)
{
    int nativeModifiers = 0;
    QString str;
    if (modifiers & Qt::ControlModifier)
    {
        str.append(" Ctrl");
        nativeModifiers |= ControlMask;
    }
    if (modifiers & Qt::ShiftModifier)
    {
        nativeModifiers |= ShiftMask;
        str.append(" Shift");
    }
    if (modifiers & Qt::AltModifier)
    {
        str.append(" Alt");
        nativeModifiers |= AltMask;
    }

    qDebug("GetNativeModifiers: modifiers are set to%s", str.toLatin1().data());
    return nativeModifiers;
}

static int MyXErrorHandler(Display *, XErrorEvent *e)
{
    switch(e->error_code)
    {
    case BadValue:
    case BadWindow:
    case BadAccess:
        if ((e->request_code == X_GrabKey) || (e->request_code == X_UngrabKey))
        {
            qDebug("MyXErrorHandler: error while grabbing or ungrabbing.");
            g_XHasError = true;
        }
        break;
    default:
        qDebug("MyXErrorHandler: some error occured...");
        break;
    }
    return 0;
}

static bool SetKeyGrab(Qt::Key key, Qt::KeyboardModifiers modifiers)
{
    int nativeKey = GetNativeKey(key);
    if (NoSymbol == nativeKey)
    {
        qDebug("SetKeyGrab: no symbol.");
        return false;
    }
    int nativeModifiers = GetNativeModifiers(modifiers);

    g_XHasError = false;

    XErrorHandler originalXErrorHandler = XSetErrorHandler(MyXErrorHandler);

    XGrabKey(QX11Info::display(), nativeKey, nativeModifiers,
             QX11Info::appRootWindow(), false, GrabModeAsync, GrabModeAsync);

    // NumLock is used as one of the modifiers, so add one more grab.
    XGrabKey(QX11Info::display(), nativeKey, nativeModifiers | NumLockMask,
             QX11Info::appRootWindow(), false, GrabModeAsync, GrabModeAsync);
    XSync(QX11Info::display(), false);
    XSetErrorHandler(originalXErrorHandler);

    return !g_XHasError;
}

static bool RemoveKeyGrab(Qt::Key key, Qt::KeyboardModifiers modifiers)
{
    int nativeKey = GetNativeKey(key);
    int nativeModifiers = GetNativeModifiers(modifiers);

    g_XHasError = false;

    XErrorHandler originalXErrorHandler = XSetErrorHandler(MyXErrorHandler);
    XUngrabKey(QX11Info::display(), nativeKey, nativeModifiers, QX11Info::appRootWindow());
    XUngrabKey(QX11Info::display(), nativeKey, nativeModifiers, QX11Info::appRootWindow());
    XFlush(QX11Info::display());
    XSetErrorHandler(originalXErrorHandler);

    return !g_XHasError;
}

HotkeyListener::HotkeyListener()
{
    this->isListening_ = false;
    this->isInitialized_ = false;
}

HotkeyListener::~HotkeyListener()
{
    this->StopListening();
}

bool HotkeyListener::SetHotkey(const BCHotkey& hotkey)
{
    // Can't set hotkey while listening.
    if (this->IsListening())
    {
        qDebug("SetHotkey failed: already listening");
        return false;
    }

    this->mainKey_ = hotkey.mainKey;
    this->modifiers_ = hotkey.modifiers;

    this->isInitialized_ = true;
    return true;
}

bool HotkeyListener::StartListening()
{
    if (this->IsListening())
        return true;
    if (!this->isInitialized_)
    {
        qDebug("StartListening: not initialized");
        return false;
    }

    // Get key without modifiers.
    bool result = SetKeyGrab(this->mainKey_, this->modifiers_);

    // If settings key grab is failed, event filter cannot help us.
    if (!result)
    {
        qDebug("StartListening: key grabbing failed.");
        return false;
    }

    qApp->installEventFilter(this);

    this->isListening_ = true;
    qDebug("StartListening: success.");

    return true;
}

bool HotkeyListener::StopListening()
{
    if (!this->IsListening())
    {
        qDebug("StopListening: listening is not started.");
        return true;
    }
    Q_ASSERT(this->isInitialized_);

    qApp->removeEventFilter(this);

    bool result = RemoveKeyGrab(this->mainKey_, this->modifiers_);
    this->isListening_ = false;
    if (result)
        qDebug("StopListening: success");

    return result;
}

bool HotkeyListener::eventFilter(QObject */*watched*/, QEvent *event)
{
    // Ignore all events except KeyPress.
    if (event->type() != QEvent::KeyPress)
        return false;

    QKeyEvent *keyEvent = static_cast<QKeyEvent*>(event);
    if (keyEvent->key() == this->mainKey_ && keyEvent->modifiers() == this->modifiers_)
    {
        qDebug("Hotkey pressed.");
        emit this->OnHotkeyPressed();
        return true;
    }

    return false;
}
