#include "passworddialog.h"
#include "ui_passworddialog.h"

#include "model/helper.h"
#include <QFileDialog>
#include <QTimer>

PasswordDialog::PasswordDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::PasswordDialog)
{
    ui->setupUi(this);
    this->ui->errorLabel->setText("");
    this->ui->okButton->setDefault(true);
    this->SetMountParams(0);

    QObject::connect(this->ui->cancelButton, SIGNAL(clicked()), this, SLOT(OnCancelSlot()));
    QObject::connect(this->ui->okButton, SIGNAL(clicked()), this, SLOT(OnOkSlot()));
    QObject::connect(this, SIGNAL(rejected()), this, SLOT(OnRejectSlot()));
    QObject::connect(this->ui->passEdit, SIGNAL(textChanged(QString)), this, SLOT(OnPassTextChangedSlot()));
    QObject::connect(this->ui->passConfirmEdit, SIGNAL(textChanged(QString)), this, SLOT(OnPassTextChangedSlot()));
    BCHelperFixSize(this);
    this->haveResult_ = false;
}

PasswordDialog::~PasswordDialog()
{
    delete ui;
}

void PasswordDialog::SetPasswordPromptText(const QString &text)
{
    this->ui->passPromptLabel->setText(text);
    BCHelperFixSize(this);
}

void PasswordDialog::SetConfirmVisible(bool visible)
{
    this->ui->passConfirmEdit->setVisible(visible);
    this->ui->passConfirmLabel->setVisible(visible);
    BCHelperFixSize(this);
}

void PasswordDialog::ShowError(const QString &error)
{
    this->ui->errorLabel->setText(error);
    BCHelperFixSize(this);
    this->ui->passEdit->selectAll();
    QTimer::singleShot(0, this->ui->passEdit, SLOT(setFocus()));
}

void PasswordDialog::SetMountParams(BCMountParams *params)
{
    if (params)
    {
        this->mountParams_ = params;
        this->ui->mountOptionsGroupBox->setVisible(true);
        this->ui->mountPointText->setText(params->mountPoint);
        this->ui->mountReadOnlyCheckbox->setChecked(params->mountReadOnly);
    }
    else
    {
        this->mountParams_ = 0;
        this->ui->mountOptionsGroupBox->setVisible(false);
    }
    BCHelperFixSize(this);
}

void PasswordDialog::SetPasswordBuffer(BCPassword *password)
{
    this->password_ = password;
}

void PasswordDialog::OnCancelSlot()
{
    this->result_ = kBCUserCancel;
    this->haveResult_ = true;
}

void PasswordDialog::OnRejectSlot()
{
    this->OnCancelSlot();
}

void PasswordDialog::OnPassTextChangedSlot()
{
    this->ui->errorLabel->setText("");
}

void PasswordDialog::OnOkSlot()
{
    QString pass = this->ui->passEdit->text();

    // If need password confirmation, we are creating a new password.
    if (!this->ui->passConfirmLabel->isHidden())
    {
        QString passConfirm = this->ui->passConfirmEdit->text();

        // Compare.
        if (pass.compare(passConfirm))
        {
            this->ShowError(tr("Passwords do not match."));
            return;
        }
        // Check password length.
        if (pass.length() >= kBCPasswordMax / 2)
        {
            this->ShowError(tr("Password is too long. Max password length is %0.").arg(kBCPasswordMax / 2));
            return;
        }
        if (pass.length() < kBCPasswordMin)
        {
            this->ShowError(QString("Password length should be at least %0.").arg(kBCPasswordMin));
            return;
        }
    }

    this->password_->set(pass.toUtf8());
    this->result_ = kBCUserOk;
    this->SetButtonsEnabled(false);

    this->MountParamsToStruct();
    this->haveResult_ = true;
}

void PasswordDialog::SetButtonsEnabled(bool enabled)
{
    this->ui->okButton->setEnabled(enabled);
    this->ui->cancelButton->setEnabled(enabled);
}

void PasswordDialog::MountParamsToStruct()
{
    // Copy mount option values from dialog to struct.
    QString mountPoint = this->ui->mountPointText->text();
    std::string str = mountPoint.toStdString();

    if (this->mountParams_)
    {
        strncpy(this->mountParams_->mountPoint, str.c_str(), BC_MAX_PATH);
        this->mountParams_->mountReadOnly = this->ui->mountReadOnlyCheckbox->checkState() == Qt::Checked;
        this->mountParams_->useDefaultMountPoint = false;
    }
}

void PasswordDialog::on_mountLocationButton_clicked()
{
    QFileDialog folderDialog;
    folderDialog.setFileMode(QFileDialog::Directory);
    folderDialog.setOptions(QFileDialog::ShowDirsOnly);

    if (folderDialog.exec())
    {
        QStringList selectedFolders = folderDialog.selectedFiles();
        if (!selectedFolders.count())
            return;

        QString mountFolder = selectedFolders.at(0);
        this->ui->mountPointText->setText(mountFolder);
    }
}
