/*
 * Copyright 2012 Jetico, Inc. All rights reserved.
 * BestCrypt command line interface implementation and main.
 */

#include <pwd.h>
#include <grp.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <signal.h>
#include <mntent.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/vfs.h>

#include <iostream>
#include <vector>
#include <iomanip>

#include "kernel/BCKernel.h"
#include "container/container.h"
#include "container/user.h"
#include "disk/disk.h"
#include "bctool.h"

#include "version.h"

const char *PROG_NAME = "bestcrypt";

#define DIE(message, args...) { fprintf(stderr, message, ## args); exit(EXIT_FAILURE); }
#define ROUND_UP(value, blocksize) ( ((value) + (blocksize) - 1) & ~((blocksize) - 1) )

// Options
static char*            o_description = 0;      /* -d description of container */
static const char*      o_fstype = 0;           /* -t type of used file system */
static const char*      o_alg = 0;              /* -a algorithm name */
static const char*      o_algmode = 0;          /* -M algorithm mode */
static char*            o_options = 0;          /* -o additional mount(2) options */
static long long int    o_cont_size = 0;        /* -s container size in bytes */
//static int              o_verbose = 0;          /* -v verbose */
static int              o_ro = 0;               /* -r read only */
static int              o_mode = 0;             /* -m mountpoint mode */
static int              o_uid = 0;              /* -u user id for mount point (root only)*/
static int              o_gid = 0;              /* -g group id for mount point (root only)*/
static int              o_force = 0;            /* -f force operation (like umount) */
//static int              o_all = 0;              /* -A umnount all containers for umount command*/

static bool gInterrupted = false;

// Ask a yes/no question.
static bool ask_user(const char* message, bool default_answer) {
	char c;
	fprintf(stderr, "%s", message);
	c = tolower(fgetc(stdin));

	if ('\n' == c) return default_answer;

	while ('\n' != fgetc(stdin));

	if ('y' == c) return true;
	else return false;
}

static void show_message(bctool_msg_level level, const char *msg) {
    switch (level) {
        case kBCTool_Debug:
            printf("Debug: ");
            break;

        case kBCTool_Info:
            //printf("Info: ");
            break;

        case kBCTool_Warning:
            fprintf(stderr, "Warning: ");
            break;

        case kBCTool_Error:
            fprintf(stderr, "Error: ");
            break;

        default:
            assert(0);
    }

    printf("%s\n", msg);
}

// catch termination signals when writing to container file to delete it.
static void sighandler(int sig) {
    if (ask_user("Do you really want to terminate process? (y/[n]):", false)) {
        gInterrupted = true;
    }
}

static int progress_report(const char *message, bc_u64 nbytes, bc_u64 total, void* arg) {
    if(gInterrupted) {
        return false;
    }

    printf("%s: %llu\\%llu bytes (%u%%)  \r", message, (long long unsigned)(nbytes),
           (long long unsigned)(total), (unsigned)(((nbytes + 1) * 100) / total));

    if(nbytes == total) {
        printf("\n");
    }
    
    return true;
}


// Creates new instance of object representing possible UI commands. Caller should destroy it by 'delete' operator.
static BCTool *create_ui() {
    BCTool *ui = new BCTool(show_message, ask_user, sighandler, progress_report);
    return ui;
}

static int get_uid_by_name(const char* name)
{
	struct passwd *pass;
   	if ( NULL == name ) return -1;

	uid_t uid = atol(name);

    if( (pass = getpwnam(name)) || (pass = getpwuid(uid)) )
	return pass->pw_uid;
	
	return -1;
}


static int get_gid_by_name(const char* name)
{
	struct group *grp;
   	if ( NULL == name ) return -1;

	gid_t gid = atol(name);

   	if( (grp = getgrnam(name)) || (grp = getgrgid(gid)) )
	return grp->gr_gid;

	return -1;
}

/******************************************************************************************
 * bctool info file 
 *
 * Display information about container file.
 * Should never ask any passwords and thus ignore encrypted header possibility
 ******************************************************************************************/

static int bc_info(int argc, char** argv) {
    const char* path = argv[1];

	if(!path) {
        fprintf(stderr, "Missing container path.\n\nPlease specify container path to see information about container,"
               " or use 'bctool benchmark' to see all available algorithms and encryption modes"
               " and their speed.\n");
		return EXIT_FAILURE;
	}

    bc_container_info info;

    BCTool *ui = create_ui();
    int result = EXIT_FAILURE;

    if (ui->GetContainerInfo(path, &info)) {
        printf("Container information: \n");
        printf("  file:         %s\n", info.containerPath);
        printf("  size:         %lu bytes\n", (unsigned long)info.size);
        printf("  algorithm:    %s-%s\n", info.algName, info.modeName);
        printf("  description:  %s\n", info.description);

        result = EXIT_SUCCESS;
    }
    else {
        fprintf(stderr, "Unable to get information about %s: not a container file or has encrypted header.\n", path);
        result = EXIT_FAILURE;
    }
    delete ui;
    return result; 
}



/******************************************************************************************
 * bctool list 
 *
 * Lists all currently configured devices and their mount points if mounted.
 ******************************************************************************************/

static int bc_list(int argc, char** argv) {
    BCTool *ui = create_ui();
    std::vector<bc_mount_entry> entries;
    int result = EXIT_FAILURE;
    if (ui->GetMountInfo(&entries)) {
        std::vector<bc_mount_entry>::iterator i = entries.begin();
        std::vector<bc_mount_entry>::iterator end = entries.end();

        while (i != end) {
            printf("%s\t\t", i->deviceName);
            if (strlen(i->mountPoint)) {
                printf("%s", i->mountPoint);
            }
            else {
                printf("Not mounted.");
            }
            printf("\t\t%s\n", i->containerPath);
            ++i;
        }

        result = EXIT_SUCCESS;
    }
    delete ui;

    return result;
}


/******************************************************************************************
 * bctool new -a algid -M modeid -s size [-d description] [-v] file 
 *
 * Display information about container file.
 * Should never ask any passwords and thus ignore encrypted header possibility
 ******************************************************************************************/

static int bc_new(int argc, char** argv, const char* algname, const char* modename, unsigned long long size, const char* description, const char* fstype)
{
    if (argc < 2) {
        DIE("Please provide container file name.\n");
    } else if (argc > 2) {
        DIE("Too many arguments.\n");
    }

    const char* file = argv[1];

    BCTool *ui = create_ui();
    bool succeeded = ui->New(file, algname, modename, size, description, fstype);
    delete ui;

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool link file [-r]
 *
 * Parse container file and create a new crypto loop binded to it
 ******************************************************************************************/

static int bc_link(int argc, char** argv, bool ro)
{
    if (argc < 2) {
        DIE("Please specify container name");
    } else if (argc > 3) {
        DIE("Too many arguments");
    }

    char device_name[BC_MAX_PATH];

    const char* cfile = argv[1];
    BCTool *ui = create_ui();
    bool succeeded = ui->Link(cfile, ro, NULL, false, NULL, NULL, device_name);

    delete ui;

    if (succeeded)
        printf("File is successfully linked to %s\n", device_name);

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool link_raw [-r] -a algname -m modename file
 *
 * Generate random key and create a new crypto loop device binded to arbitrary file
 ******************************************************************************************/

static int bc_link_raw(int argc, char** argv, const char* algname, const char* modename, bool ro)
{
	if (argc < 2) {
        DIE("Please specify file name to link.\n");
	} else if (argc > 2) {
        DIE("Too many arguments.\n");
	}

    if (!algname) {
        DIE("Please specify algorithm name.\n");
    }

    const char* fname = argv[1];

    BCTool *ui = create_ui();
    bool succeeded = ui->RawLink(fname, algname, modename, ro);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool unlink device
 *
 * Unlink crypto loop device previously linked manually.
 ******************************************************************************************/

static int bc_unlink(int argc, char** argv)
{
	if( argc < 2 ) {
        DIE("Please specify device name\n");
	} else if (argc > 2) {
        DIE("Too many arguments\n");
	}

    const char* devname = argv[1];

    BCTool *ui = create_ui();
    bool succeeded = ui->Unlink(devname, (bool)o_force);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/********************************************************************************************************
 * bctool mount [-r] [-t fstype] [-u owner] [-g group] [-m mode] [-o options] file mount_point
 *
 * Mount container at the specified mount point.
 *
 * -u, -g and -m options allow the user to set mount point owner, group and mode respectively.
 * This is to support the case when root is mounting the container for a different user
 *
 * -o option passes its value directly to mount(1) -o 
 *
 * -t option allows user to specify file system type so we dont have to guess
 *
 * -r is for reading only
 ********************************************************************************************************/

static int bc_mount(int argc, char** argv, const char* fstype, int uid, int gid, int mode, bool ro, const char* opts)
{
	if (argc < 3) {
        DIE("Please specify container file and mount point.\n");
	} else if (argc > 3) {
        DIE("Too many arguments.\n");
	}

    const char* file = argv[1];
    const char* mpoint = argv[2];
    char device_name[BC_MAX_PATH];
    BCMountParams mount_params;
    mount_params.mountReadOnly = ro;
    mount_params.useDefaultMountPoint = false;
    strncpy(mount_params.mountPoint, mpoint, BC_MAX_PATH);
    BCTool *ui = create_ui();

    bool succeeded = ui->Mount(file, &mount_params, fstype, uid, gid, opts, device_name);
    delete ui;

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}


/********************************************************************************************************
 * bctool umount [-f] mount_point
 *
 * Unmount container at specified mount point
 * -f flag forces unmount even if container is busy.
 ********************************************************************************************************/

bool bc_umount(int argc, char** argv, bool force)
{
	if (argc < 2) {
        DIE("Please specify mount point.\n");
	} else if (argc > 2) {
        DIE("Too many arguments.\n");
	}

    const char* mpoint = argv[1];

    BCTool *ui = create_ui();
    bool succeeded = ui->Umount(mpoint, force);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/********************************************************************************************************
 * bctool umountall [-f]
 *
 * Unmount all mounted containers
 * -f flag forces unmount even if container is busy.
 ********************************************************************************************************/

bool bc_umount_all(int argc, char** argv, bool force)
{
	if (argc > 1) {
        DIE("Too many arguments\n");
	}
    int total_mounted, umounted;
    BCTool *ui = create_ui();
    bool succeeded = ui->UmountAll(force, &total_mounted, &umounted);
    delete ui;
    if (total_mounted)
    {
        printf("%d of %d containers successfully unmounted.\n", umounted, total_mounted);
        if ((umounted != total_mounted) && force)
        {
            printf("Warning! Some containers aren't unmounted!\nIf you really need to unmount them right now, please reboot your computer.");
        }
    }
    else
    {
        printf("There are no containers to unmount.\n");
    }

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool format -t fstype file
 *
 * Link and create a file system on container.
 ******************************************************************************************/

static int bc_format(int argc, char** argv, const char* fstype)
{
	if (argc < 2) {
        DIE("Please specify container name\n");
	} else if (argc > 2) {
        DIE("Too many arguments\n");
	} else if (!fstype) {
        DIE("Please specify file system\n");
    }

	const char* file = argv[1];
    BCTool *ui = create_ui();
    bool succeeded = ui->Format(file, fstype, false);
    delete ui;
    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}


/******************************************************************************************
 * bctool fsck [-t fstype] file
 *
 * Check formatted container file system
 ******************************************************************************************/

static int bc_fsck(int argc, char** argv, const char* fstype)
{
	if (argc < 2) {
        DIE("Please specify container name\n");
	} else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    const char* file = argv[1];
    BCTool *ui = create_ui();
    bool succeeded = ui->CheckFS(file, fstype);
    delete ui;

    return succeeded? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool passwd|add_passwd|del_passwd file
 *
 * Change, remove or add container password
 ******************************************************************************************/

static int bc_passwd(int argc, char** argv, BCKeyOp op)
{
    // check container file write access
    if (argc < 2) {
        DIE("Please specify container file\n");
    } else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    const char* file = argv[1];

    BCTool *ui = create_ui();
    bool result = ui->Passwd(file, op);
    delete ui;

    return result? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool encrypt_header|decrypt_header file
 *
 * Encrypt or decrypt container header.
 ******************************************************************************************/

static int bc_header(int argc, char** argv, bool encrypt)
{
    if (argc < 2) {
        DIE("Please specify container file\n");
    } else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    const char* file = argv[1];

    BCTool *ui = create_ui();
    bool result = ui->ProcessHeader(file, encrypt);
    delete ui;

    return result? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool backup container file 
 *
 * Backup container header into backup file.
 ******************************************************************************************/

static int bc_backup(int argc, char** argv)
{
    if (argc < 3) {
        DIE("Please specify container and backup files\n");
    } else if (argc > 3) {
        DIE("Too many arguments\n");
    }

    const char* container = argv[1];
    const char* backup = argv[2];

    BCTool *ui = create_ui();
    bool result = ui->BackupHeader(container, backup);
    delete ui;

    return result? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool restore container file 
 *
 * Restore container header from backup file.
 ******************************************************************************************/

static int bc_restore(int argc, char** argv)
{
    if (argc < 3) {
        DIE("Please specify container and backup files\n");
    } else if (argc > 3) {
        DIE("Too many arguments\n");
    }

    const char* container = argv[1];
    const char* backup = argv[2];

    BCTool *ui = create_ui();
    bool result = ui->RestoreHeader(container, backup);
    delete ui;

    return result? EXIT_SUCCESS : EXIT_FAILURE;
}



/******************************************************************************************
 * bctool make_hidden file -s size -t fstype
 *
 * Create container hidden part of given size.
 ******************************************************************************************/

static int bc_hidden(int argc, char** argv, unsigned long long size, const char* fstype)
{
    const char* file = argv[1];
    if (argc < 1) {
        DIE("Please specify the hidden part size.");
    }
    if (argc > 2) {
        DIE("Too many arguments.");
    }
    
    BCTool *ui = create_ui();
    bool result = ui->CreateHiddenPart(file, size, fstype);
    delete ui;

    if (!result)
        return EXIT_FAILURE;

    return EXIT_SUCCESS;
}



/******************************************************************************************
 * bctool reencrypt file -a algorithm -M mode
 *
 * Reencrypt container with new algorithm, mode and key.
 ******************************************************************************************/

static int bc_reencrypt(int argc, char** argv, const char* algname, const char* modename)
{
    if (argc < 2) {
        DIE("Please specify container file\n");
    } else if (argc > 2) {
        DIE("Too many arguments\n");
    }

    const char* file = argv[1];
    
    BCTool *ui = create_ui();
    bool result = ui->Reencrypt(file, algname, modename);
    delete ui;

    return (result)? EXIT_SUCCESS : EXIT_FAILURE;
}


static int bc_benchmark(int argc, char **argv, const char *algname, const char *modename)
{
    std::vector<bc_bench_info> bench_result;

    printf("Benchmarking...\n");
    BCTool *ui = create_ui();
    bool succeeded = ui->Benchmark(algname, modename, &bench_result);
    delete ui;

    if (!succeeded)
        return EXIT_FAILURE;

    printf("Benchmark results: \n");
    std::vector<bc_bench_info>::iterator i = bench_result.begin();
    std::cout.precision(2);

    for (; i != bench_result.end(); i++)
    {
        bc_bench_info *info = &(*i);
        std::cout << std::setw(15) << info->algName << std::setw(5) << info->modeName << std::setw(10) << std::fixed << ((double)1e9 / (double)info->averageTime) << " MB/s" << std::endl;
    }

    return EXIT_SUCCESS;
}

static int bc_is_kernel_initialized()
{
    BCTool *ui = create_ui();
    bool result = ui->CheckKernelInitialization();
    delete ui;

    if (!result)
        return 0;
    return 1;
}


static void usage(void) {
    const char *usage_string="Usage: bctool command [options] arguments\n\
\n\
Commands available:\n\
 * addpasswd file_name - add additional password for container\n\
 * backup container_file backup_file - backup container header to file\n\
 * benchmark [-a alg_name] [-M alg_mode] - benchmark algorithms\n\
 * decryptheader file_name - decrypt container header\n\
 * delpasswd file_name - delete additional password(s) from container\n\
 * encryptheader file_name - encrypt container header\n\
 * format file_name -t filesystem_type - format container\n\
 * fsck file_name [-t fs_type] - check container file system\n\
 * info file_name - print information about container\n\
 * link file_name [-r] [-f] - link container to device\n\
 * list - list all linked devices, their container files and mount points\n\
 * makehidden file_name -s size [-t fs_type] [-v] - (temporary disabled) make hidden part in container\n\
 * mount file_name mount_point [-t fstype] [-m mode] [-u user] [-g group] [-r] - mount container\n\
 * new file_name -a alg_name -s size -M alg_mode [-d description] [-t fstype] [-v] - create container\n\
 * passwd file_name - change container's password\n\
 * rawlink file_name device_name -a alg_name [-r] - raw link file to BestCrypt device\n\
 * restore container_file backup_file - restore container header from backup\n\
 * reencrypt file_name [-a alg_name] [-M alg_mode] [-v] - reencrypt container\n\
 * umount mount_point [-f] - unmount container\n\
 * umountall [-v] [-f] - unmount all mounted containers\n\
 * unlink device_name - unlink container from device\n\
\n\
options:\n\
   -t fstype   - to specify filesystem type (default: msdos)\n\
   -a alg_name - to specify algorithm name\n\
   -M alg_mode - to specify algorithm mode\n\
                 alg_mode can be CBC, LRW or XTS (default is CBC).\n\
                 Not all modes are supported by all algorithms.\n\
   -s size     - to specify container size\n\
                 example:\n\
                 -s 32768 - 32768 bytes (32 kilobytes);\n\
                 -s 10K   - 10 Kb (10240 bytes);\n\
                 -s 10M   - 10 Mb (10240 kilobytes);\n\
                 -s 10G   - 10 Gb (10240 megabytes)\n\
   -d text     - to specify description of container (up to 64 symbols)\n\
   -r          - to mount read only\n\
   -m mode     - to specify mountpoint mode in octal (default 0700)\n\
   -u user     - to specify mountpoint's owner (only root can do that)\n\
   -g group    - to specify mountpoint's group\n\
   -V          - print version and exit\n\
\n";

    fprintf(stderr, "%s\n", usage_string);
}


static int run_command(const char* cmd, int argc, char** argv)
{
    if (0 == strcmp(cmd, "addpasswd"))      return bc_passwd(argc, argv, kBCKey_Duplicate);
    if (0 == strcmp(cmd, "backup"))         return bc_backup(argc, argv);
    if (0 == strcmp(cmd, "benchmark"))      return bc_benchmark(argc, argv, o_alg, o_algmode);
    if (0 == strcmp(cmd, "decryptheader"))  return bc_header(argc, argv, false);
    if (0 == strcmp(cmd, "delpasswd"))      return bc_passwd(argc, argv, kBCKey_Delete);
    if (0 == strcmp(cmd, "encryptheader"))  return bc_header(argc, argv, true);
    if (0 == strcmp(cmd, "format"))         return bc_format(argc, argv, o_fstype);
    if (0 == strcmp(cmd, "fsck"))           return bc_fsck(argc, argv, o_fstype);
    if (0 == strcmp(cmd, "info"))           return bc_info(argc, argv);
    if (0 == strcmp(cmd, "link"))           return bc_link(argc, argv, o_ro);
    if (0 == strcmp(cmd, "list"))           return bc_list(argc, argv);
    if (0 == strcmp(cmd, "makehidden"))     return bc_hidden(argc, argv, o_cont_size, o_fstype);
    if (0 == strcmp(cmd, "mount"))          return bc_mount(argc, argv, o_fstype, o_uid, o_gid, o_mode, o_ro, o_options);
	if (0 == strcmp(cmd, "new"))            return bc_new(argc, argv, o_alg, o_algmode, o_cont_size, o_description, o_fstype);
    if (0 == strcmp(cmd, "passwd"))         return bc_passwd(argc, argv, kBCKey_Recode);
    if (0 == strcmp(cmd, "rawlink"))        return bc_link_raw(argc, argv, o_alg, o_algmode, o_ro);
    if (0 == strcmp(cmd, "restore"))        return bc_restore(argc, argv);
    if (0 == strcmp(cmd, "reencrypt"))      return bc_reencrypt(argc, argv, o_alg, o_algmode);
    if (0 == strcmp(cmd, "umount"))         return bc_umount(argc, argv, o_force);
    if (0 == strcmp(cmd, "umountall"))      return bc_umount_all(argc, argv, o_force);
    if (0 == strcmp(cmd, "unlink"))         return bc_unlink(argc, argv);



    fprintf(stderr, "Unknown command %s\n", cmd);
    usage();

    return EXIT_FAILURE;
}


int main(int argc, char** argv) 
{
    if (!bc_is_kernel_initialized())
        return EXIT_FAILURE;

    setvbuf(stderr, (char*)NULL, _IONBF, 0);
    setvbuf(stdout, (char*)NULL, _IONBF, 0);

    char c;
    while ((c = getopt(argc, argv, "Vfhprt:a:o:s:d:m:M:u:g:")) != -1)
	{
        switch (c) {
        case 'V':
            fprintf(stdout, "BestCrypt version %d.%d-%d rev%d\nCopyright Jetico, Inc. See LICENSE for details.\n", 
                    VERSION_MAJOR, VERSION_MINOR, VERSION_PATCH, VERSION_REVISION);
            exit(EXIT_SUCCESS);

//        case 'v':
//            o_verbose = 1;
//	        break;

//        case 'A':
//            o_all = 1;
//            break;

        case 'r':
            o_ro = 1;
	        break;

        case 'f':
            o_force = 1;
	        break;

        case 't':
            o_fstype = optarg;
	        break;

        case 'a':
            o_alg = optarg;
            break;

        case 'M':
            o_algmode = optarg;
            break;

        case 'o':
            o_options = optarg;
            break;

        case 'd':
            o_description = optarg;
 	        break;

        case 's':
            o_cont_size = BCTool::BCAtol(optarg);
          	if (o_cont_size == 0) {
                fprintf(stderr, "Invalid container size\n");
                exit(EXIT_FAILURE);
            } 
            break;

        case 'm':
            o_mode = strtol(optarg, NULL, 8);
            o_mode = o_mode & 07777;
            if (0 == o_mode) {
                fprintf(stderr, "Invalid mode, using 0700\n");
                o_mode = 0700;
            }
            break;

        case 'u':
          	if (0 != getuid()) {
                fprintf(stderr, "Only root can specify mountpoint's owner\n");
                exit(EXIT_FAILURE);
            }

            o_uid = get_uid_by_name(optarg);
            
            if (-1 == o_uid) {
                fprintf(stderr, "Invalid user name %s\n", optarg);
                exit(EXIT_FAILURE);
            }

            break;
            
        case 'g':
          	o_gid = get_gid_by_name(optarg);
          	if (-1 == o_gid) {
                fprintf(stderr, "Invalid group name %s\n", optarg);
                exit(EXIT_FAILURE);
            }
                	
            break;
            
        case 'h':
           	usage();
          	exit(EXIT_SUCCESS);

        default:
           	usage();
          	exit(EXIT_FAILURE);

        }
    }

    argc -= optind;
    argv += optind;
    const char* cmd = argv[0];

    if (NULL == cmd) {
        fprintf(stderr, "Missing command name\n");
        usage();
        exit(EXIT_FAILURE);
    }

	return run_command(cmd, argc, argv);
}

