#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <linux/fs.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>

void PrintUsage(const char *programName)
{
    printf("%s device-path <options>\n\
Possible options:\n\
    -f      - fill device blocks by sector numbers.\n\
    -c      - check that device blocks are filled by sector numbers.\n", programName);
}

int ProcessDevice(const char *devicePath, int fill)
{
    int mode = fill? W_OK: R_OK;
    if (access(devicePath, mode))
    {
        fprintf(stderr, "%s: access check failed (%s).\n", devicePath, strerror(errno));
        return EXIT_FAILURE;
    }

    if (fill)
    {
        printf("Are you sure you want to erase all data on %s?(y/[n])", devicePath);
        char ch = getchar();
        if ('y' != ch && 'Y' != ch)
            return EXIT_FAILURE;
    }
    
    int fd = open(devicePath, O_RDWR);
    if (-1 == fd)
    {
        fprintf(stderr, "Error opening %s: %s.\n", devicePath, strerror(errno));
        return EXIT_FAILURE;
    }

    unsigned long blockCount = 0;
    char block[512];
    char tmpBlock[512];
    uint64_t blockNumber = 0;
    int result = EXIT_SUCCESS;

    if (0 < ioctl(fd, BLKGETSIZE, &blockCount))
    {
        printf("IO ctl error: %s.\n", strerror(errno));
        close(fd);
        return EXIT_FAILURE;
    }
    printf("Blocks total: %lu (last block #%lu)\n", blockCount, blockCount - 1);

    while(1)
    {
        memset(block, '\0', 512);
        snprintf(block, 511, "%lu", blockNumber);
        block[511] = '\0';
        printf("Processing block %lu...                                     \r", blockNumber);
        if (fill)
        {
            if (512 != write(fd, block, 512))
            {
                fprintf(stderr, "\nError writing to device: %s.\n", strerror(errno));
                result = EXIT_FAILURE;
                break;
            }
        }
        else
        {
            ssize_t bytesRead = read(fd, tmpBlock, 512); 
            if (512 != bytesRead)
            {
                fprintf(stderr, "\nError while reading block %s. \n\tBytes read: %llu, \n\tsystem error: %s\n", 
                        block, (unsigned long long)bytesRead, strerror(errno));
                result = EXIT_FAILURE;
            }
            if (memcmp(block, tmpBlock, 512))
            {
                fprintf(stderr, "\nBlock #%s is not valid.\n", block);
                result = EXIT_FAILURE;
            }
        }
        blockNumber++;
        if (blockNumber == blockCount)
            break;
    }
    
    close(fd);
    printf("\nExiting.\n");
    return result;
}

int main(int argc, char **argv)
{
    if (argc < 3)
    {
        PrintUsage(argv[0]);
        exit(EXIT_FAILURE);
    }

    char *devicePath = argv[1];
    struct stat st;
    
    if (stat(devicePath, &st))
    {
        printf("Stat error: %s\n", strerror(errno));
        return EXIT_FAILURE;
    }

    if (!S_ISBLK(st.st_mode))
    {
        printf("%s is not a block device.\n", devicePath);
        return EXIT_FAILURE;
    }
    
    char ch = 0;
    while ((ch = getopt(argc, argv, "fc")) != -1)
    {
        switch(ch)
        {
            case 'f':
                return ProcessDevice(devicePath, 1);
            case 'c':
                return ProcessDevice(devicePath, 0);
            default:
                printf("Unknown parameter: -%c", ch);
                return EXIT_FAILURE;
        }
    }
    PrintUsage(argv[0]);
    return EXIT_FAILURE;
}
