#include "../../kernel/BCKernel.h"

#include <stdio.h>
#include <limits.h>

#define CHECK(res) if (res) { printf("Failed at line %i with result '%s'\n", __LINE__, BCKernelGetErrorString(res)); return res; }
#define ASSERT(condition) if (!(condition)) { printf("Condition failed at line %i\n", __LINE__); }

const char *GetModeName(bc_mode_id mode)
{
    switch(mode)
    {
        case kBCMode_CBC:
            return "CBC";
        case kBCMode_LRW:
            return "LRW";
        case kBCMode_XTS:
            return "XTS";
        default:
            return "Unsupported mode.";
    }
}

void WriteFileString(const char *buffer, size_t buffer_size, const char *file_path)
{
    int i;
    FILE *f = fopen(file_path, "w");
    for (i = 0; i < buffer_size; i++)
    {
        if (i % 16 == 0)
            fprintf(f, "\n");
        fprintf(f, "0x%02hhx, ", (unsigned char)buffer[i]);
    }
    fclose(f);
}

int GetRandomString(char *buffer, size_t buffer_length)
{
    FILE *f = fopen("/dev/urandom", "r");
    if (NULL == fgets(buffer, buffer_length, f))
        return -1;
    fclose(f);
}

bc_kern_result ProduceAlgKey(BCAlgInfo *alg, bc_mode_id mode, char *key_buffer, size_t key_buf_size, size_t *out_key_size)
{
    size_t key_size = BCKernelGetModeKeySize(alg, mode);
    ASSERT(key_size <= key_buf_size);
    if (GetRandomString(key_buffer, key_size))
        return kBCKernel_InternalError;
    *(out_key_size) = key_size;
    return kBCKernel_Success;
}

bc_kern_result ProduceAlgData(BCAlgInfo *alg, bc_mode_id mode, char *data_buffer, size_t data_buffer_size, size_t *out_data_size)
{
    //size_t block_size = alg->block_size;
    size_t data_size = 512;
    ASSERT(data_size <= data_buffer_size);
    if (GetRandomString(data_buffer, data_size))
        return kBCKernel_InternalError;
    *(out_data_size) = data_size;
    return kBCKernel_Success;
}

bc_kern_result ProduceAlgPool(char *pool, size_t pool_size)
{
    if (GetRandomString(pool, pool_size))
        return kBCKernel_InternalError;
    return kBCKernel_Success;
}

bc_kern_result ProduceAlgIV(char *iv, size_t iv_size)
{
    if (GetRandomString(iv, iv_size))
        return kBCKernel_InternalError;
    return kBCKernel_Success;
}

int ProduceFilesForAlg(BCAlgInfo *alg, bc_mode_id mode)
{
    const char    *out_path = "cipher-tests";
    const size_t   KeyBufferSize = 512;
    const size_t   DataBufferSize = 512;
    const size_t   IVSize = 8;
    char           path_buffer[PATH_MAX] = {0};
    char           key_buffer[KeyBufferSize];
    char           data_buffer[DataBufferSize];
    char           pool[kBCKernel_PoolSize];
    char           iv[IVSize];
    size_t         key_size;
    size_t         data_size;
    
    // Skip unsupported modes.
    if (kBCKernel_Success != BCKernelIsModeSupported(alg, mode))
        return kBCKernel_UnsupportedMode;
    
    // Key.
    snprintf(path_buffer, PATH_MAX - 1, "%s/%s_%skey.txt", out_path, alg->name, GetModeName(mode));
    CHECK(ProduceAlgKey(alg, mode, key_buffer, KeyBufferSize, &key_size));
    WriteFileString(key_buffer, key_size, path_buffer);
    
    // Data.
    snprintf(path_buffer, PATH_MAX - 1, "%s/%s_%sindata.txt", out_path, alg->name, GetModeName(mode));
    CHECK(ProduceAlgKey(alg, mode, data_buffer, DataBufferSize, &data_size));
    WriteFileString(data_buffer, data_size, path_buffer);
    
    // Pool.
    snprintf(path_buffer, PATH_MAX - 1, "%s/%s_%spool.txt", out_path, alg->name, GetModeName(mode));
    CHECK(ProduceAlgPool(pool, kBCKernel_PoolSize));
    WriteFileString(pool, kBCKernel_PoolSize, path_buffer);
    
    // Initial vector.
    snprintf(path_buffer, PATH_MAX - 1, "%s/%s_%siv.txt", out_path, alg->name, GetModeName(mode));
    CHECK(ProduceAlgPool(iv, IVSize));
    WriteFileString(iv, IVSize, path_buffer);
    
    // Encrypted data.
    snprintf(path_buffer, PATH_MAX - 1, "%s/%s_%soutdata.txt", out_path, alg->name, GetModeName(mode));
    CHECK(BCKernelProcessWithKey(alg->id, mode, kBCProcess_Encrypt, key_buffer, key_size, *(uint64_t*)iv, pool, data_buffer, data_size));
    WriteFileString(data_buffer, data_size, path_buffer);
    
    return kBCKernel_Success;
}

int main()
{
    BCAlgInfo        *alg_list         = NULL;
    unsigned       alg_count        = 0;
    bc_kern_result result           = kBCKernel_Success;
    int            i                = 0;
    
    CHECK(BCKernelIsInitialized());
    
    CHECK(BCKernelListAlgs(&alg_list, &alg_count));
    
    printf("Loaded alg count: %i\n", alg_count);
    
    for (i = 0; i < alg_count; i++)
    {
        printf("\nProcessing %s...\n", alg_list[i].name);
        
        printf(" * CBC...");
        result = bc_kern_result(ProduceFilesForAlg(&(alg_list[i]), kBCMode_CBC));
        if (result == kBCKernel_UnsupportedMode)
            printf("unsupported mode.\n");
        else
        {
            CHECK(result);
            printf("success.\n");
        }
        printf(" * LRW...");
        result = bc_kern_result(ProduceFilesForAlg(&(alg_list[i]), kBCMode_LRW));
        if (result == kBCKernel_UnsupportedMode)
            printf("unsupported mode.\n");
        else
        {
            CHECK(result);
            printf("success.\n");
        }
        printf(" * XTS...");
        result = bc_kern_result(ProduceFilesForAlg(&(alg_list[i]), kBCMode_XTS));
        if (result == kBCKernel_UnsupportedMode)
            printf("unsupported mode.\n");
        else
        {
            CHECK(result);
            printf("success.\n");
        }
        
    }
    
    printf("\nCompleted.\n");
    
    BCKernelFreeAlgList(alg_list);
    
    return 0;
}
