#ifndef TESTS_ENCRYPTION_C
#define TESTS_ENCRYPTION_C
#include "../tests-common.c"

void TestBCKernelStoreFreeKey(void)
{
    const char    *key = "test key,,,";
    size_t         key_len = strlen(key);
    bc_key_handle  key_handle = 0;
    bc_kern_result result;
    int            i = 0;
    FILE          *f;
    const int      key_size = 50;
    char           random_keys[kBCKernel_MaxKeys + 1][key_size];
    char          *big_key;
    const int      big_key_size = 4097;
    
    // Tests without kernel module loaded.
    result = BCKernelStoreKey(key, key_len, &key_handle);
    ASSERT_RESULT(kBCKernel_NotInitialized, result);
    
    InstallBCKernel();
    
    // Invalid key handle pointer.
    result = BCKernelStoreKey(key, key_len, NULL);
    ASSERT_RESULT(kBCKernel_InvalidBuffer, result);
    
    // Invalid key.
    result = BCKernelStoreKey(NULL, key_len, &key_handle);
    ASSERT_RESULT(kBCKernel_InvalidKey, result);
    
    // Invalid key size.
    result = BCKernelStoreKey(key, 0, &key_handle);
    ASSERT_RESULT(kBCKernel_InvalidKey, result);
    
    // Tests with kernel module loaded.
    result = BCKernelStoreKey(key, key_len, &key_handle);
    ASSERT_RESULT(kBCKernel_Success, result);
    
    result = BCKernelFreeKey(key_handle);
    ASSERT_RESULT(kBCKernel_Success, result);
    
    // Freeing key twise.
    result = BCKernelFreeKey(key_handle);
    ASSERT_RESULT(kBCKernel_InvalidKey, result);
    
    // Random handle.
    result = BCKernelFreeKey(1234);
    ASSERT_RESULT(kBCKernel_InvalidKey, result);
    
    // Invalid handle.
    result = BCKernelFreeKey(-1);
    ASSERT_RESULT(kBCKernel_InvalidKey, result);
    
    // Big key test.
    big_key = (char *)malloc(big_key_size * sizeof(char));
    
    result = BCKernelStoreKey(big_key, big_key_size, &key_handle);
    ASSERT_RESULT(kBCKernel_Success, result);
    
    result = BCKernelFreeKey(key_handle);
    ASSERT_RESULT(kBCKernel_Success, result);
    
    free(big_key);
    
    // Keys' overflowing test.
    f = fopen("/dev/urandom", "r");
    if (NULL == f)
    {
        UninstallBCKernel();
        CU_FAIL_FATAL("Cannot open /dev/urandom device.");
    }
    
    for (i=0; i < kBCKernel_MaxKeys + 1; i++)
    {
        fgets((char *)&(random_keys[i]), key_size, f);
        result = BCKernelStoreKey((char *)&(random_keys[i]), key_size, &key_handle);
        if (i < kBCKernel_MaxKeys)
        {
            CU_ASSERT(kBCKernel_Success == result);
        }
        else // For last key.
        {
            CU_ASSERT(kBCKernel_NoMemory == result);
        }
    }
    
    fclose(f);
    UninstallBCKernel();
}

void InstallKernelModules()
{
    InstallBCKernel();
    if (Install3DESModule())
    {
        UninstallBCKernel();
        CU_FAIL_FATAL("Cannot install 3DES module.");
    }
    
    if (InstallAESModule())
    {
        Uninstall3DESModule();
        UninstallBCKernel();
        CU_FAIL_FATAL("Cannot install Rijndael module.");
    }
}

void UninstallKernelModules()
{
    Uninstall3DESModule();
    UninstallAESModule();
    UninstallBCKernel();
}

static void TestAlg(BCAlgInfo *alg, bc_mode_id mode, int key_handle, const char *key, size_t key_size)
{
    CU_ASSERT_FATAL(NULL != alg);
    
    size_t mode_key_size = BCKernelGetModeKeySize(alg, mode);
    if (0 == mode_key_size) // Mode is not supported by alg.
        return;
    
    bc_kern_result result;
    const char *initial_data = "This is test....asdfghjqwertyuio";
    const char *wrong_length_data = "This data has wrong length.  ";
    const char *short_key = "Short key";
    int iv = 0;
    char *pool = NULL;
    char data[34] = {0};
    size_t data_size = strlen(initial_data);
    
    strncpy(data, initial_data, data_size);
    CU_ASSERT(data_size % alg->block_size == 0);
    
    // Short key failure check.
    
    result = BCKernelProcessWithKey(alg->id, mode, kBCProcess_Encrypt, short_key, strlen(short_key), iv, pool, data, data_size);
    ASSERT_RESULT(kBCKernel_InvalidKey, result);
    CU_ASSERT(strncmp(data, initial_data, data_size) == 0); // data was not changed.
    
    // Checking wrong data size.
    data_size = strlen(wrong_length_data);
    strncpy(data, wrong_length_data, data_size);
    
    CU_ASSERT(data_size % alg->block_size != 0);
    
    result =  BCKernelProcess(alg->id, mode, kBCProcess_Encrypt, key_handle, iv, pool, data, data_size);
    ASSERT_RESULT(kBCKernel_InvalidBuffer, result);
    CU_ASSERT(strncmp(data, wrong_length_data, data_size) == 0); // data was not changed.
    
    result = BCKernelProcessWithKey(alg->id, mode, kBCProcess_Encrypt, key, key_size, iv, pool, data, data_size);
    ASSERT_RESULT(kBCKernel_InvalidBuffer, result);
}

void TestBCKernelEncryptionCommon(void)
{
    // Using 3DES alg for testing.
    const char *key = "This is test....asdfghjqwertyuiuhrtqyejdgshetdjhdkjweyrkandhqj";
    size_t key_size = strlen(key) * sizeof(char);

    bc_kern_result result;
    bc_mode_id mode = kBCMode_CBC;
    bc_key_handle key_handle = 0;
    uint64_t iv = 0;
    int i = 0;
    int alg_id = 208;
    const char *pool = NULL;
    const char *initial_data = "This is test....asdfghjqwertyuio";
    char data[34] = {0};
    size_t data_size = 0;
    
    strncpy(data, initial_data, 33);
    data_size = strlen(data);
    // Tests without kernel module loaded.
    // Kernel is not initialized.
    result = BCKernelProcess(alg_id, mode, kBCProcess_Encrypt, key_handle, iv, pool, data, data_size);
    ASSERT_RESULT(kBCKernel_NotInitialized, result);
    
    result = BCKernelProcessWithKey(alg_id, mode, kBCProcess_Encrypt, key, key_size, iv, pool, data, data_size);
    ASSERT_RESULT(kBCKernel_NotInitialized, result);
    
    InstallKernelModules();
    
    // Data can be null...
    result = BCKernelProcess(alg_id, mode, kBCProcess_Encrypt, 123, iv, NULL, NULL, 0);
    ASSERT_RESULT(kBCKernel_Success, result);
    
    // ...but must not be null with non-zero size.
    result = BCKernelProcess(alg_id, mode, kBCProcess_Encrypt, 123, iv, NULL, NULL, data_size);
    ASSERT_RESULT(kBCKernel_InvalidBuffer, result);
    
    result = BCKernelProcessWithKey(alg_id, mode, kBCProcess_Encrypt, key, key_size, iv, pool, NULL, data_size);
    ASSERT_RESULT(kBCKernel_InvalidBuffer, result);

    // Checking result with valid parameters for each alg.
    result = BCKernelStoreKey(key, key_size, &key_handle);
    ASSERT_RESULT(kBCKernel_Success, result);
    BCAlgInfo *alg_list = NULL;
    unsigned int alg_count = 0;
    
    result = BCKernelListAlgs(&alg_list, &alg_count);
    ASSERT_RESULT(kBCKernel_Success, result);
    CU_ASSERT(0 != alg_count && NULL != alg_list);
    
    for (i = 0; i < alg_count; i++)
    {
        TestAlg(&(alg_list[i]), kBCMode_CBC, key_handle, key, key_size);
        TestAlg(&(alg_list[i]), kBCMode_LRW, key_handle, key, key_size);
        TestAlg(&(alg_list[i]), kBCMode_XTS, key_handle, key, key_size);
    }
    
    result = BCKernelFreeKey(key_handle);
    ASSERT_RESULT(kBCKernel_Success, result);
    
    UninstallKernelModules();
}



#endif
