#include "../tests-common.c"

#include "spacemap/extent.h"
#include "spacemap/spacemap.h"

#include <stdio.h>

/***********************Extent********************************/

static void TestExtentStartEnd()
{
    Extent zero = {0, 0};
    Extent ten = {5, 10};

    CU_ASSERT(ExtentStart(&zero) == 0);
    CU_ASSERT(ExtentEnd(&zero) == 0);

    CU_ASSERT(ExtentStart(&ten) == 5);
    CU_ASSERT(ExtentEnd(&ten) == 14);
}

static void TestExtentContainsPoint()
{
    Extent zero = {0, 0};
    Extent ten = {0, 10};
    bc_u64 i = 0;

    // Test border values.
    CU_ASSERT(ExtentContainsPoint(&ten, 0));
    CU_ASSERT(ExtentContainsPoint(&ten, 9));
    CU_ASSERT(!ExtentContainsPoint(&ten, 10));

    // And check all the values inside.
    for (i = 0; i < 10; i++)
    {
        if (ExtentContainsPoint(&ten, i))
            continue;
        else
        {
            printf("ExtentContainsPoint failed with i = %llu\n", i);
            CU_FAIL("ExtentContainsPoint failed.\n");
        }
    }

    CU_ASSERT(!ExtentContainsPoint(&zero, 0));
}

static void TestExtentContains()
{
    Extent twenty = {20, 20};
    Extent first = {20, 1};
    Extent last = {39, 1};
    Extent afterLast = {40, 1};
    Extent beforeFirst = {19, 1};
    Extent insideTwenty;
    bc_u64 i = 0;

    // Check border values.
    CU_ASSERT(ExtentContains(&twenty, &first));
    CU_ASSERT(!ExtentContains(&first, &twenty));

    CU_ASSERT(ExtentContains(&twenty, &last));
    CU_ASSERT(!ExtentContains(&last, &twenty));

    CU_ASSERT(!ExtentContains(&twenty, &beforeFirst));
    CU_ASSERT(!ExtentContains(&beforeFirst, &twenty));

    CU_ASSERT(!ExtentContains(&twenty, &afterLast));
    CU_ASSERT(!ExtentContains(&afterLast, &twenty));

    // Must contain itself.
    CU_ASSERT(ExtentContains(&twenty, &twenty));

    // Check all that lay inside.
    for (i = 1; i < 20; i++)
    {
        insideTwenty.start = 20 + i;
        insideTwenty.length = 20 - i;
        if (ExtentContains(&twenty, &insideTwenty)
                && (!ExtentContains(&insideTwenty, &twenty)))
            continue;
        fprintf(stderr, "Failed for extent with i = %llu\n", i);
        CU_FAIL("ExtentContains failed");
    }
}

static void TestExtentAreIntercepting()
{
    Extent ten = {10, 10};
    Extent first = {10, 1};
    Extent last = {19, 1};
    Extent afterLast = {20, 1};
    Extent beforeFirst = {9, 1};

    Extent intersected = {5, 10};
    Extent container = {9, 11};
    Extent insider = {15, 2};

    // Check border values.
    CU_ASSERT(ExtentAreIntersecting(&ten, &first));
    CU_ASSERT(ExtentAreIntersecting(&first, &ten));

    CU_ASSERT(ExtentAreIntersecting(&ten, &last));
    CU_ASSERT(ExtentAreIntersecting(&last, &ten));

    CU_ASSERT(!ExtentAreIntersecting(&ten, &afterLast));
    CU_ASSERT(!ExtentAreIntersecting(&afterLast, &ten));

    CU_ASSERT(!ExtentAreIntersecting(&ten, &beforeFirst));
    CU_ASSERT(!ExtentAreIntersecting(&beforeFirst, &ten));

    // Check intersections.
    CU_ASSERT(ExtentAreIntersecting(&ten, &intersected));
    CU_ASSERT(ExtentAreIntersecting(&intersected, &ten));

    CU_ASSERT(ExtentAreIntersecting(&ten, &container));
    CU_ASSERT(ExtentAreIntersecting(&container, &ten));

    CU_ASSERT(ExtentAreIntersecting(&ten, &insider));
    CU_ASSERT(ExtentAreIntersecting(&insider, &ten));
}

static void TestExtentUnion()
{
    Extent result;
    Extent ten = {10, 11}; // 10...20
    Extent twenty = {20, 10}; // 20...29
    Extent unn = {10, 20}; // 10...29
    Extent five = {5, 5};

    // Normal test.
    memset(&result, 0, sizeof(Extent));
    CU_ASSERT(ExtentAreIntersecting(&ten, &twenty));
    CU_ASSERT(ExtentUnion(&result, &ten, &twenty));
    CU_ASSERT(result.start == unn.start);
    CU_ASSERT(result.length == unn.length);

    // Self-union test.
    memset(&result, 0, sizeof(Extent));
    CU_ASSERT(ExtentUnion(&result, &ten, &ten));
    CU_ASSERT(result.start == ten.start);
    CU_ASSERT(result.length == ten.length);

    // Non-intersecting extents.
    CU_ASSERT(!ExtentUnion(&result, &twenty, &five));
}

/***************************************************************/

void TestExtentFunctions()
{
    TestExtentStartEnd();
    TestExtentContainsPoint();
    TestExtentContains();
    TestExtentAreIntercepting();
    TestExtentUnion();
}

/***********************SpaceMap********************************/

#define CHECK_RESULT(func) result = (func); if (kSpaceMap_Success != result) { fprintf(stderr, "Failed with result '%s'\n", SpaceMapGetErrorString(result)); CU_FAIL(""); break; }
#define CREATE_MAP(map) map = SpaceMapCreate(); if (!map) { CU_FAIL("Map creation failed."); return; }

static void TestSpaceMapContainsExtent()
{
    Extent e = {1, 2};
    Extent similar = {1, 2};
    Extent uncontained = {1, 3};
    Extent anotherUncontained = {2, 2};
    Extent andAnotherUncontained = {2, 1};
    sm_result result;
    SpaceMap *map = NULL;

    CREATE_MAP(map);

    // Check function behaviour with incorrect parameters.
    if (kSpaceMap_InvalidArg != SpaceMapContainsExtent(NULL, &e, true))
    {
        CU_FAIL("With invalid map.");
        return;
    }

    if (kSpaceMap_InvalidArg != SpaceMapContainsExtent(map, NULL, true))
    {
        CU_FAIL("With invalid extent.");
        return;
    }

    do
    {
        // Add a copy of the extent to the map.
        CHECK_RESULT(SpaceMapAddCopy(map, &e));

        // Check that map contains added extent.
        CHECK_RESULT(SpaceMapContainsExtent(map, &e, true));
        CHECK_RESULT(SpaceMapContainsExtent(map, &similar, true));

        CU_ASSERT(kSpaceMap_NotFound == SpaceMapContainsExtent(map, &uncontained, true));
        CU_ASSERT(kSpaceMap_NotFound == SpaceMapContainsExtent(map, &anotherUncontained, true));
        CU_ASSERT(kSpaceMap_NotFound == SpaceMapContainsExtent(map, &andAnotherUncontained, true));
    }
    while (false);

    SpaceMapFree(map);
}

static void TestSpaceMapNormalize()
{
    SpaceMap *map = NULL;
    Extent normal = {1, 2};
    Extent zeroed = {2, 0};
    Extent intersected1 = {5, 3};
    Extent intersected2 = {6, 4};
    Extent unioned = {5, 5};
    sm_result result;
    bc_u64 count;

    // Try with incorrect parameters.
    CU_ASSERT(kSpaceMap_InvalidArg == SpaceMapNormalize(NULL));

    CREATE_MAP(map);

    do
    {
        // Add our extents to the map.
        CHECK_RESULT(SpaceMapAddCopy(map, &normal));
        CHECK_RESULT(SpaceMapAddCopy(map, &intersected1));
        CHECK_RESULT(SpaceMapAddCopy(map, &intersected2));
        CHECK_RESULT(SpaceMapAddCopy(map, &zeroed));

        // Do normalization and see the results.
        CHECK_RESULT(SpaceMapNormalize(map));

        CHECK_RESULT(SpaceMapExtentCount(map, &count));

        CU_ASSERT(count == 2);

        CHECK_RESULT(SpaceMapContainsExtent(map, &normal, true));
        CHECK_RESULT(SpaceMapContainsExtent(map, &unioned, true));
    }
    while (false);

    SpaceMapFree(map);
}

void TestSpaceMapToArray()
{
    SpaceMap *map = NULL;
    Extent e1 = {1,1}, e2 = {2, 1}, e3 = {3, 1}, e4 = {4, 1};
    Extent *array = NULL;
    bc_u64 count = 0;
    sm_result result;

    CREATE_MAP(map);

    do
    {
        CHECK_RESULT(SpaceMapAddCopy(map, &e1));
        CHECK_RESULT(SpaceMapAddCopy(map, &e2));
        CHECK_RESULT(SpaceMapAddCopy(map, &e3));
        CHECK_RESULT(SpaceMapAddCopy(map, &e4));

        CHECK_RESULT(SpaceMapToExtentArray(map, &array, &count));

        CU_ASSERT(count == 4);
        CU_ASSERT(array != NULL);

        CU_ASSERT(ExtentAreEqual(&array[0], &e1));
        CU_ASSERT(ExtentAreEqual(&array[1], &e2));
        CU_ASSERT(ExtentAreEqual(&array[2], &e3));
        CU_ASSERT(ExtentAreEqual(&array[3], &e4));
    }
    while (false);
    SpaceMapFreeArray(&array);
    SpaceMapFree(map);
}

void TestSpaceMapSort()
{
    SpaceMap *map = NULL;
    sm_result result;
    Extent first = {1, 2};              // 1,2
    Extent second = {3, 1};             // 3
    Extent third = {4, 3};              // 4,5,6
    Extent thirdIntersected = {5, 3};   // 5,6,7
    Extent thirdFinal = {4, 4};
    Extent *array = NULL;
    bc_u64 count = 0;

    // Add extents to map in inapropriate order.
    CREATE_MAP(map);

    do
    {
        CHECK_RESULT(SpaceMapAddCopy(map, &thirdIntersected));
        CHECK_RESULT(SpaceMapAddCopy(map, &first));
        CHECK_RESULT(SpaceMapAddCopy(map, &third));
        CHECK_RESULT(SpaceMapAddCopy(map, &second));

        CHECK_RESULT(SpaceMapSort(map));
        CHECK_RESULT(SpaceMapToExtentArray(map, &array, &count));
        CU_ASSERT(count == 3);
        CU_ASSERT(ExtentAreEqual(&array[0], &first));
        CU_ASSERT(ExtentAreEqual(&array[1], &second));
        CU_ASSERT(ExtentAreEqual(&array[2], &thirdFinal));
    }
    while (false);

    SpaceMapFree(map);
    SpaceMapFreeArray(&array);
}

void TestSpaceMapSubstract()
{
    SpaceMap *from_map = NULL, *what_map = NULL;
    sm_result result;
    Extent one_three = {1, 3},
            four_six = {4, 3},
            seven_nine = {7, 3},
            eight_ten = {8, 3},
            seven = {7, 1},
            twelve_seventin = {12, 6},
            fourteen_fifteen = {14, 2},
            twelve_thirteen = {12, 2},
            sixteen_seventeen = {16, 2};

    CREATE_MAP(from_map);
    CREATE_MAP(what_map);

    do
    {
        // 1) Substracting a whole map.
        CHECK_RESULT(SpaceMapAddCopy(from_map, &one_three));
        CHECK_RESULT(SpaceMapAddCopy(what_map, &one_three));
        CHECK_RESULT(SpaceMapSubstract(from_map, what_map));

        // From-map must be fresh here.
        CU_ASSERT(from_map->count == 0);
        CU_ASSERT(from_map->start == NULL);
        CU_ASSERT(from_map->end == NULL);

        // And what-map must be the same as before.
        CU_ASSERT(what_map->count == 1);
        CHECK_RESULT(SpaceMapContainsExtent(what_map, &one_three, false));

        SpaceMapFree(from_map);
        SpaceMapFree(what_map);

        CREATE_MAP(from_map);
        CREATE_MAP(what_map);

        // 2) No substraction.
        CHECK_RESULT(SpaceMapAddCopy(from_map, &one_three));
        CHECK_RESULT(SpaceMapAddCopy(what_map, &four_six));
        CHECK_RESULT(SpaceMapSubstract(from_map, what_map));

        CHECK_RESULT(SpaceMapContainsExtent(from_map, &one_three, false));
        CU_ASSERT(from_map->count == 1);

        CHECK_RESULT(SpaceMapContainsExtent(what_map, &four_six, false));
        CU_ASSERT(from_map->count == 1);

        SpaceMapFree(from_map);
        SpaceMapFree(what_map);

        CREATE_MAP(from_map);
        CREATE_MAP(what_map);

        // 3) Intersected extents.
        CHECK_RESULT(SpaceMapAddCopy(from_map, &seven_nine));
        CHECK_RESULT(SpaceMapAddCopy(what_map, &eight_ten));
        CHECK_RESULT(SpaceMapSubstract(from_map, what_map));

        CHECK_RESULT(SpaceMapContainsExtent(from_map, &seven, false));
        CHECK_RESULT(SpaceMapContainsExtent(what_map, &eight_ten, false));

        SpaceMapFree(from_map);
        SpaceMapFree(what_map);

        CREATE_MAP(from_map);
        CREATE_MAP(what_map);

        // 4) Extent inside.
        CHECK_RESULT(SpaceMapAddCopy(from_map, &twelve_seventin));
        CHECK_RESULT(SpaceMapAddCopy(what_map, &fourteen_fifteen));
        CHECK_RESULT(SpaceMapSubstract(from_map, what_map));

        CU_ASSERT(from_map->count == 2);
        CHECK_RESULT(SpaceMapContainsExtent(from_map, &twelve_thirteen, false));
        CHECK_RESULT(SpaceMapContainsExtent(from_map, &sixteen_seventeen, false));

        CU_ASSERT(what_map->count == 1);
        CHECK_RESULT(SpaceMapContainsExtent(what_map, &fourteen_fifteen, false));

        SpaceMapFree(from_map);
        SpaceMapFree(what_map);

        CREATE_MAP(from_map);
        CREATE_MAP(what_map);

        // And do them all one more time :)
        CHECK_RESULT(SpaceMapAddCopy(from_map, &one_three));
        CHECK_RESULT(SpaceMapAddCopy(from_map, &seven_nine));
        CHECK_RESULT(SpaceMapAddCopy(from_map, &twelve_seventin));

        CHECK_RESULT(SpaceMapAddCopy(what_map, &one_three));
        CHECK_RESULT(SpaceMapAddCopy(what_map, &four_six));
        CHECK_RESULT(SpaceMapAddCopy(what_map, &eight_ten));
        CHECK_RESULT(SpaceMapAddCopy(what_map, &fourteen_fifteen));

        CHECK_RESULT(SpaceMapSubstract(from_map, what_map));

        CU_ASSERT(from_map->count == 3);
        CU_ASSERT(kSpaceMap_NotFound == SpaceMapContainsExtent(from_map, &one_three, false));
        CHECK_RESULT(SpaceMapContainsExtent(from_map, &seven, false));
        CHECK_RESULT(SpaceMapContainsExtent(from_map, &twelve_thirteen, false));
        CHECK_RESULT(SpaceMapContainsExtent(from_map, &sixteen_seventeen, false));
    }
    while (false);

    SpaceMapFree(from_map);
    SpaceMapFree(what_map);
}

static void TestSpaceMapGetPhysicalExtents()
{
    //------6----------5---------------9-----
    //---4=====9----13====17-----22========30
    Extent virt[3],
            *phys = NULL;
    bc_u64 vcount = 3,  // virtual extent count.
            pcount = 0, // count of physical extents.
            length;     // virtual space length.
    sm_result result;

    virt[0].start = 4;
    virt[0].length = 6;

    virt[1].start = 13;
    virt[1].length = 5;

    virt[2].start = 22;
    virt[2].length = 9;

    do
    {
        CHECK_RESULT(SpaceMapGetExtentSpace(virt, 3, &length));

        CU_ASSERT(length == (bc_u64)20);

        // Will read all 20 bytes.
        CHECK_RESULT(SpaceMapGetPhysicalExtents(virt, vcount, 0, length, &phys, &pcount));
        CU_ASSERT(phys != NULL);
        CU_ASSERT(pcount == 3);
        CU_ASSERT(!memcmp(phys, virt, vcount * sizeof(Extent)));
        SpaceMapFreePhysicalExtents(&phys);

        // Will read inside one extent.
        CHECK_RESULT(SpaceMapGetPhysicalExtents(virt, vcount, 2, 2, &phys, &pcount));
        CU_ASSERT(pcount == 1);
        CU_ASSERT(phys[0].start == 6);
        CU_ASSERT(phys[0].length == 2);
        SpaceMapFreePhysicalExtents(&phys);

        // Will read on the border of two extents.
        CHECK_RESULT(SpaceMapGetPhysicalExtents(virt, vcount, 5, 4, &phys, &pcount));
        CU_ASSERT(pcount == 2);
        CU_ASSERT(phys[0].start == 9);
        CU_ASSERT(phys[0].length == 1);
        CU_ASSERT(phys[1].start == 13);
        CU_ASSERT(phys[1].length == 3);
        SpaceMapFreePhysicalExtents(&phys);

        // Try to read more than available.
        result = SpaceMapGetPhysicalExtents(virt, vcount, 5, 20, &phys, &pcount);
        CU_ASSERT(result == kSpaceMap_NotFound);
        SpaceMapFreePhysicalExtents(&phys);
        return;
    }
    while(false);

    SpaceMapFreePhysicalExtents(&phys);
}

#undef CHECK_RESULT

void TestSpaceMap()
{
    TestSpaceMapContainsExtent();
    TestSpaceMapNormalize();
    TestSpaceMapToArray();
    TestSpaceMapSort();
    TestSpaceMapSubstract();
    TestSpaceMapGetPhysicalExtents();
}
