/*
 * Copyright 2010 Jetico, Inc. All rights reserved.
 * Keygen module interface.
 */

#ifndef BC_KEYGEN_H
#define BC_KEYGEN_H

#include "bctypes.h"
#include "error.h"
#include "password.h"
#include "memory.h"
#include "spacemap/spacemap.h"

#include <stdio.h>

#define BC_DESCRIPTION_SIZE 128

// Encoded volume info
struct BCVolumeInfo {
    int         alg;                                // Container encryption algorithm id.
    int         mode;                               // Container encryption mode id.
    int         hash;                               // Keygen hash algorithm id.
    int         hidden;                             // Is this a hidden volume.
    char        description[BC_DESCRIPTION_SIZE];   // User description utf8 string. Not set for hidden volumes.
    bc_u64      size;                               // Size of the main part of the container.
    bc_u64      offset;                             // Offset of the main part of the container.
    SpaceMap*   map;                                // Extent map, used for hidden keys.

    BCVolumeInfo()
    {
        memset(this, 0, sizeof(BCVolumeInfo));
    }

    ~BCVolumeInfo()
    {
        if (this->map)
        {
            SpaceMapFree(this->map);
        }
    }
};


// Key decoding operation type.
typedef enum {
	kBCKey_Recode,			        // Decode key and recode it inplace with new user secret.
	kBCKey_Duplicate,		        // Decode key and encode its copy with new secret.
    kBCKey_Delete			        // Decode key and delete it.
} BCKeyOp;


struct BCKeygen;

/**
 * Container header interface.
 * Header should contain information about main encoded volume and passwords
 * Each keygen specifies its own header format and implements its interface. 
 */
class IBCHeader {
public:
    
    // owner
    const struct BCKeygen* keygen;

    /**
     * Get main volume info.
     * Behaviour is undefined if header is encrypted.
     * @param info          Initialized main volume info.
     */
    virtual void mainVolumeInfo(struct BCVolumeInfo* info) = 0;

    /**
     * Initialize keygen.
     * @param algid         Encryption algorithm id
     * @param modeid        Encryption mode id
     * @param size          Encrypted data size in bytes
     * @param description   Optional container description
     * @param password      Initial container password
     */
    virtual BCContainerResult init(int algid, int modeid, bc_u64 size,
                                   const char* description, const char* password,
                                   BCProgressCallback progress, void *progressArg) = 0;

    /**
     * Generate and store a new hidden key.
     * @param spaceMap          Hidden part space map.
     * @param mainPassword      Main part password.
     * @param hiddenPassword    New hidden part password.
     */
    virtual BCContainerResult createHiddenKey(SpaceMap *spaceMap,
                            const BCPassword *mainPassword, const BCPassword* hiddenPassword,
                            BCProgressCallback progress, void *progressArg) = 0;

	/**
	 * Ask user secret (password, etc) and decode volume key.
     * @param password      UTF8 encoded initial user password.
     * @param out_info      Initialized decoded volume info on success.
     * @param out_key       Decoded volume key on success.
	 * @param out_pool	    Decrypted random data pool on success.
	 * @return	            kBCKeygen_Success if user provided correct password, error otherwise.
	 */
    virtual BCContainerResult decodeKey(const char* password, struct BCVolumeInfo* out_info,
                                        int* out_key, char* out_pool) = 0;

    /**
     * Verify a single password without decoding anything.
     * @param password      Password to verify.
     * @param out_info      Volume info for correct password.
     */
    virtual BCContainerResult verifyPassword(const char* password, struct BCVolumeInfo* out_info) = 0;

    /**
     * Add new container password
     * @param password      Existing password.
     * @param newPassword   New password to add.
     */
    virtual BCContainerResult addPassword(const char* password, const char* newPassword) = 0;

    /**
     * Change existing password
     * @param password      Existing password.
     * @param newPassword   New password.
     */
    virtual BCContainerResult changePassword(const char* password, const char* newPassword) = 0;

    /**
     * Remove existing password
     * @param password      Existing password to remove
     */
    virtual BCContainerResult removePassword(const char* password) = 0;

    /**
     * Tell if container header is encrypted.
     */
    virtual bool isEncrypted() = 0;

    /**
     * Encrypt container header.
     */
    virtual BCContainerResult encrypt(const char* password) = 0;

    /**
     * Decrypt container header.
     */
    virtual BCContainerResult decrypt(const char* password) = 0;

    /**
     * Update container description.
     * Each format has a maximum description length.
     */
    virtual void setDescription(const char* description) = 0;

    /**
     * Mark container as undergoing reencryption.
     * @param mark          Set or remove mark.
     */
    virtual void markForReencryption(bool mark) = 0;

    /**
     * Tell if container is marked for reencryption.
     */
    virtual bool isMarkedForReencryption() = 0;
    
    /**
     * Store keygen data into a file.
     * File might be a container file or a blank file in case of a backup.
     * @param file          File name to write to.
     */
    virtual BCContainerResult store(const char* file) = 0;

    /**
     * Destroy keygen.
     */
    virtual ~IBCHeader() {}

protected:

    IBCHeader() {}
    IBCHeader(const IBCHeader& /*rhv*/) {}
    IBCHeader& operator = (const IBCHeader& /*rhv*/) { return *this; }

};


// Scoped header pointer for resource management
typedef BCScopedPtr<IBCHeader> BCHeaderHandle;


enum {
    kBCKeygen_EncryptedHeaders = 0x01 // keygen supports encrypted headers
};


/**
 * Known keygen module description.
 */
struct BCKeygen {
	int		    id;             // Unique id
	const char*	name;           // Name string
	const char*	description;    // Detailed description
	int		    version;        // Supported container version
    int         flags;


    /**
     * Create blank keygen.
     */
    IBCHeader*      (*kgcreate) (void);

    
    /**
     * Load keygen from file.
     * @param file          Container file path.
     */
    IBCHeader*      (*kgload) (const char* file);


	/**
	 * Parse main volume info from container file.
	 * @param file	        Container file path
	 * @param info	        Initialized keygen structure on success.
	 */
	BCContainerResult	(*kgparse) (const char* file, struct BCVolumeInfo* info);
    
};


#endif

